package cox.jmatt.java.MathTools;

import java.util.*;
import java.text.NumberFormat;

/** <p>This class provides methods for creating and manipulating numbers digit-wise. It also provides methods for identifying whole and fractional place values 
    and a method to convert a long into words.</p>
    <p>Arithmetic also contains several methods that, while not specifically arithmetic-related, are handy to have. These methods generate a (person's) name and can 
    supply the properly-gendered pronouns for that name.</p>
    <h3>Decimal Arrays and Digit Arrays</h3>
    <p>Arithmetic allows digit-wise manipulation of numbers via the <code>createNumber()</code> method, decimal arrays, and digit arrays. A digit array is nothing 
    more than an integer broken down into an array corresponding to its place values. Digit arrays should be read 'backwards;' element zero is ones, element 1 is 
    tens, element 2 is hundreds, etc. The index of the array element equals the (positive) exponent of 10 in its place value. Each element of a digit array holds 
    a single digit: 0 through 9. Most methods that deal with digit arrays enforce this by modding by ten as needed.</p>
    <p>Decimal arrays work exactly as do digit arrays, but backward. The element index still represents an exponent, but in this case they are <i>negative</i>. 
    Element 0 still represents ones, but in a decimal array this is the <i>whole</i> part of the number, so 'mod 10' is not applied to this place ONLY. Element 1 
    represents the tenths place, or 10^-1.</p>
    <p>Use decimal and digit arrays to simplify rounding, addition-with-carry and subtraction-with-borrow problem creation.</p>
*/
public class Arithmetic {

  private static final String[] NUMBER_NAMES = "zero one two three four five six seven eight nine ten eleven twelve thirteen fourteen fifteen sixteen seventeen eighteen nineteen twenty".split(" ");
  private static final String[] TENS_NAMES = "zero ten twenty thirty forty fifty sixty seventy eighty ninety".split(" ");
  private static final String[] PLACE_VALUES = ("one:ten:hundred:thousand:ten thousand:hundred thousand:million:ten million:hundred million:billion:ten billion:hundred billion:"
    + "trillion:ten trillion:hundred trillion:quadrillion:ten quadrillion:hundred quadrillion").split(":");
  private static final String[] SOME_NAMES = ("Allan Betty Charlie Dianne Edward Francis Gary Helen Ivan Jane Ken Laura Mike Nora Oscar Penny Quincy Rita Sam Traci " 
    + "Vernon Alice Bob Cindy Dave Ellen Fred Gail Henry Ingrid Jack Kelly Louis Mary Ned Olivia Pete Quisha Ralph Sarah Ted Veronica").split(" ");

  private static final NumberFormat myNumberFormat = NumberFormat.getInstance();

  /** Constructor provided to generate instances for scripting apps. */
  public Arithmetic() { super(); }

  static {
    myNumberFormat.setRoundingMode(java.math.RoundingMode.HALF_UP);
    myNumberFormat.setMinimumFractionDigits(0);
  }

  /** Retrieve a place value up to hundred quadrillion. Outside that range the value '10^{pIndex} is returned. The place value is interpreted as a power of ten. 
      If the index is negative it is a fractional place, if positive, whole. If desired, the proper suffix ('-s' or '-ths') can be appended to the place value.
      @param pIndex The exponent of the place value to be returned.
      @param pSuffix true to include the suffix, false otherwise.
      @param pPlural true to pluralize the suffix, false to leave it singular.
      @return The name of the place value associates with the given index.
  */
  public static String getPlaceValue(int pIndex, boolean pSuffix, boolean pPlural) {
    String theSuffix = ((pIndex < 0)?"th":"") + (pPlural?"s":"");
    String retval = "10^{"+pIndex+"}";
    pIndex = java.lang.Math.abs(pIndex);
    if(pIndex < PLACE_VALUES.length) retval = PLACE_VALUES[pIndex] + (pSuffix?theSuffix:"");
    return retval;
  }

  /** Convenience method for <code>getPlaceValue(pIndex, pSuffix, true)</code>. */
  public static String getPlaceValue(int pIndex, boolean pSuffix) { return getPlaceValue(pIndex, pSuffix, true); }

  /** Convert a long into words. Accurate to the extent of type long. */
  public static String toWords(long pVal) {
    return toWords(""+pVal);
  }

  /** <p>Convert a String containing a number into words. Decimal points are allowed here! If the String starts with a digit, the whole part will be stated even if 
      it is zero. If the String starts with a decimal point, only the decimal part will be stated. If the number contains no decimal point there will be no 
      fractional part stated. Any conversion errors are reported at Debug level and the return value is 'zero'. If pNumber contains more than one decimal point, 
      the second point and everything afterward is ignored.</p>
      @param pNumber The String containing the number to 'speak.' If null or blank, 'zero' is returned.
      @return The words representing the number given.
  */
  public static String toWords(String pNumber) {
    if(!Question.validString(pNumber)) return "zero";
    String retval = "", sAnd = "";
    try {
      if(pNumber.indexOf(".") > -1) {
        String[] tokenz = pNumber.split("[.]");
        if(Question.validString(tokenz[0])) {
          retval = numToWords(tokenz[0]);
          sAnd = " and";
        }
        if(tokenz.length > 1 ) {
          while(tokenz[1].endsWith("0") && (tokenz[1].length() > 1)) tokenz[1] = tokenz[1].substring(0, tokenz[1].length() - 1);
          if(!("0".equals(tokenz[1]))) {
            String sDecPlace = getPlaceValue(-1*tokenz[1].length(), true, !("1".equals(tokenz[1])));
            retval = (retval + sAnd).trim()+" "+numToWords(tokenz[1])+" "+sDecPlace;
          }
        }
      } else {
        retval = numToWords(pNumber);
      }
    } catch(Exception ex) {
      CapCom.debug("Error converting to words", ex);
      retval = "zero";
    }
    return retval;
  }

  //Make sure the number is valid, handle the sign and put the number into a decimal array.
  private static String numToWords(String pNum) {
    if(!Question.validString(pNum)) return "zero";
    String retval = "retval", sMinus = "";
    pNum = pNum.trim();
    if(pNum.startsWith("-")) {
      sMinus = "negative ";
      if(pNum.length() > 0) pNum = pNum.substring(1);
    }
    if(pNum.length() < 1) return "zero";
    StringBuffer sbDigits = new StringBuffer();
    for(int i=0; i < pNum.length(); i++) sbDigits.append(pNum.charAt(i)).append(" ");
    pNum = sbDigits.toString();
    retval = numToWords(splint(pNum, true, false), NUMBER_NAMES, TENS_NAMES);
    return (sMinus + retval).trim();
  }

  // Convert a decimal array into words.
  private static String numToWords(int[] pNum, String[] pNumberNames, String[] pTensNames) {
    if((pNum == null) || (pNum.length < 1)) return "zero";
    int iLen = pNum.length, iVal = 0;
    String retval = "", curPlace = " ";
    StringBuffer valTemp = new StringBuffer();
    for(int iPtr = 0; iPtr < iLen; iPtr += 3) {
      iVal = pNum[iPtr];
      if(iLen > (iPtr + 1)) iVal += 10*pNum[iPtr + 1];
      if(iLen > (iPtr + 2)) iVal += 100*pNum[iPtr + 2];
      if((iPtr > 0) && (iVal > 0)) curPlace = getPlaceValue(iPtr, false, false);
      if(iVal > 99) valTemp.append(pNumberNames[iVal/100]).append(" hundred");
      iVal %= 100;
      if(iVal > 0) {
        if(iVal < 20) {
          valTemp.append(" ").append(pNumberNames[iVal]);
        } else {
          valTemp.append(" ").append(pTensNames[iVal/10]);
          if(iVal%10 > 0) valTemp.append("-").append(pNumberNames[iVal % 10]);
        }
      }
      retval = (valTemp.append(" ").append(curPlace.trim()).append(" ").append(retval)).toString().trim();
      valTemp = new StringBuffer();
      iVal = 0;
      curPlace = " ";
    }
    retval = retval.trim();
    return(Question.validString(retval)?retval:"zero");
  }

  /** <p>Generate a number digit-wise per a special format String. The position of a formatting character is the position in which that rule will be applied. 
      The format characters are:
      <ul>
        <li>'#' generates a random digit from 0 to 9.</li>
        <li>'?' generates a random digit from 1 to 9.</li>
        <li>'+' generates a random digit from 5 to 9.</li>
        <li>'-' generates a random digit from 0 to 4.</li>
      </ul></p>
      <p>Characters other than the format characters are copied verbatim, thus allowing dollar amounts or decimals to be generated in-place. To include a format 
      character, concatenate a String around it! If pFormat is null or empty it is returned unaltered.</p>
      @param pFormat The String containing one or more formatting characters.
      @return A String containing the specified number.
  */
  public static String createNumber(String pFormat) {
    if(!Question.validString(pFormat)) return pFormat;
    StringBuffer retval = new StringBuffer();
    for(int i=0; i < pFormat.length(); i++) {
      switch(pFormat.charAt(i)) {
        case '#': retval.append(MathEngine.random(0, 10)); break;
        case '?': retval.append(MathEngine.random(1, 10)); break;
        case '+': retval.append(MathEngine.random(5, 10)); break;
        case '-': retval.append(MathEngine.random(0, 5)); break;
        default: retval.append(pFormat.charAt(i));
      }
    }
    return retval.toString();
  }
  /** This method takes an int and turns it into an int array. Element zero is the digit in the ones place, element 1 is the tens, etc.
      @param pNum The int to split into a digit array.
      @return An int array consisting of pNum's digits, starting with the ones at element 0.
  */
  public static int[] toDigitArray(int pNum) {
    int[] retval = new int[(""+pNum).length()];
    for(int i=0; i < retval.length; i++) {
      retval[i] = pNum % 10;
      pNum /= 10;
    }
    return retval;
  }
  /** <p>This method takes a space-separated String of numbers and converts them into an int[] array. If pDigitArray is true ALL numbers are modded down to single 
      digits. If pDecimalArray is true the modding only occurs from element 1 up. If both are true, pDigitArray takes precedence. If an error occurs parsing one 
      of the numbers it becomes zero. If pData is null or empty the zero array is returned. Parsing errors are reported at Debug level.</p>
      <p>NOTE: If neither flag is set the String is split, parsed as int, and returned unaltered.</p>
      @param pData A space-separated String of ints to break into an array.
      @param pDigitArray true to parse single digits only, false to leave them alone.
      @param pDecimalArray true to parse single digits starting at element 1.
  */
  public static int[] splint(String pData, boolean pDigitArray, boolean pDecimalArray) {
    int[] retval = {0};
    if(!Question.validString(pData)) return retval;
    String [] sTokens = pData.split(" ");
    retval = new int[sTokens.length];
    for(int i=0; i < retval.length; i++) {
      try { retval[i] = Integer.parseInt(sTokens[i]); } catch(Exception ex) { CapCom.debug("splint() error parsing ("+pData+") at ["+i+"]"); }
    }
    if(pDigitArray) {
      int[] iaTemp = Arrays.copyOf(retval, retval.length);
      int iNdx = 0;
      for(int i=iaTemp.length - 1; i >= 0; i--) {
        retval[i] = iaTemp[iNdx];
        iNdx++;
      }
      retval[0] %= 10;
    }
    if(pDigitArray || pDecimalArray) {
      for(int i=1; i < retval.length; i++) { retval[i] %= 10; }
    }
    return retval;
  }
  /** This method takes a digit (int[]) array and converts it to an integer type (long). All values are made positive and modded by 10 to ensure validity of the 
      resulting number.
      @param pAry The digit array to convert. This is ENFORCED!
      @return The number made up by the digit array.
  */
  public static long fromDigitArray(int[] pAry) {
    long retval = 0;
    for(int i=pAry.length - 1; i >= 0; i--) {
      retval = retval * 10 + (java.lang.Math.abs(pAry[i]) % 10);
    }
    return retval;
  }
  /** This method turns a digit array into a String. No other formatting is applied. */
  public static String daToString(int[] pDigits) {
    if((pDigits == null) || (pDigits.length < 1)) return "0";
    StringBuffer retval = new StringBuffer();
    for(int i = pDigits.length - 1; i >= 0; i--) retval.append(java.lang.Math.abs(pDigits[i]) % 10);
    return retval.toString();
  }
  /** Convert a decimal array to a String. The decimal point is inserted between elements 0 and 1. If the array is null or empty, '0.0' is returned. If it contains 
      only one element, it is returned as a String. NOTE: Digits from position 1 up are modded by 10 to ensure that they <b>are</b> single digits.
  */
  public static String fromDecimalArray(int[] pDecArr) {
    if((pDecArr == null) || (pDecArr.length == 0)) return "0.0";
    StringBuffer retval = new StringBuffer(""+pDecArr[0]);
    if(pDecArr.length == 1) return retval.toString();
    retval.append(".");
    for(int i=1; i < pDecArr.length; i++) { retval.append((pDecArr[i] % 10)); }
    return retval.toString();
  }
  
  /** <p>This method takes a number, a String with a decimal point somewhere, and converts it into an int[] array. The index of each element indicates its 
      <i>fractional</i> place value. Element 0 is the whole part, element 1 is ten<u>ths</u>, element two is hundred<u>ths</u>, etc. If the element does not contain 
      a decimal part the entire thing is assigned to element zero.</p>
      <p>If the conversion chokes on a position that digit becomes zero and the rest of the String is still parsed. If pDec is null or empty the zero array, '{0}', 
      is returned. Parsing errors are reported at the DEBUG level. If the number contains more than one decimal point, whatever precedes the first (even nothing!) 
      becomes the whole part and whatever lies between the first and second points becomes the decimal part.</p>
      @param pDec A String containing a decimal point somewhere.
      @return A decimal array with the whole part in element 0 and moving up in index (down in place value) from there.
  */
  public static int[] toDecimalArray(String pDec) {
    int[] retval = {0};
    if(!Question.validString(pDec)) return retval;
    String[] sParts = pDec.split("[.]");
    //CapCom.debug("sParts.length "+sParts.length);
    String wholePart = "0", decPart = "";
    if(sParts.length > 1) {
      wholePart = sParts[0]; decPart = sParts[1];
    } else if(pDec.startsWith(".")) {
      decPart = pDec.substring(1);
    } else {
      wholePart = sParts[0];
    }
    retval = new int[decPart.length() + 1];
    try { retval[0] = Integer.parseInt(wholePart); } catch(Exception ex) { CapCom.debug("Error parsing "+pDec+": whole part.", ex); }
    for(int i=1; i < retval.length; i++) {
      try { retval[i] = Integer.parseInt("" + decPart.charAt(i - 1)); } catch(Exception ex) { CapCom.debug("Error parsing "+pDec+": whole part.", ex); }
    }
    return retval;
  }

  /** This method takes a decimal or digit array and rounds to the specified place. If the specified rounding place is outside the array (or equal to its length), 
      the array is returned unchanged. If the data array is null or length zero, the zero array is returned. If pData has only one element, it is returned. if 
      pPlace is zero, decimal arrays are rounded to the nearest whole number and digit arrays are returned unaltered.
      @param pData The int[] array to be rounded.
      @param pPlace The place value to round to. If negative, it is made positive.
      @param pDecArray true for a decimal array, false for a digit array.
      @return The properly rounded and truncated array.
  */
  public static int[] round(int[] pData, int pPlace, boolean pDecArray) {
    int[] retval = {0};
    boolean roundUp = false;
    if((pData == null) || (pData.length < 1)) return retval;
    if(pPlace < 0) pPlace = -pPlace;
    if((pPlace >= pData.length) || (pData.length == 1)) return pData;
    if(pDecArray) {
      if(pPlace == pData.length - 1) return pData;
      retval = new int[pPlace + 1];
      CapCom.debug("Retval length: "+retval.length);
      roundUp = pData[pPlace + 1] >= 5;
      while(pPlace > 0) {
        retval[pPlace] = roundUp?pData[pPlace] + 1:pData[pPlace];
        roundUp = retval[pPlace] > 9;
        retval[pPlace] %= 10;
        pPlace --;
      }
      retval[0] = (roundUp?pData[0] + 1:pData[0]);
    } else {
      if(pPlace == 0) return pData;
      retval = new int[pData.length];
      roundUp = (pData[pPlace - 1] >= 5);
      for(int i=0; i < pPlace; i++) { retval[i] = 0; }
      while(pPlace < pData.length) {
        retval[pPlace] = roundUp?pData[pPlace] + 1:pData[pPlace];
        roundUp = retval[pPlace] > 9;
        retval[pPlace] %= 10;
        pPlace ++;
      }
      if(roundUp) {
        retval = Arrays.copyOf(retval, pData.length + 1);
        retval[retval.length - 1] = 1;
      }
    }
    return retval;
  }
  /** This method takes an int[] array and extends it to the specified length, filling in any extra values with the padding constant. If the supplied array is 
      equal to or longer than the one sent in no padding occurs. If pAry is null it is returned and no new array is created, otherwise the new array will be a 
      copy.
      @param pAry The array to pad to the given length.
      @param pLength The minimum length of the returned array.
      @param pPad The pad value to use for any extended elements.
  */
  public static int[] pad(int[] pAry, int pLength, int pPad) {
    if(pAry == null) return pAry;
    int[] retval = new int[Math.max(pLength, pAry.length)];
    Arrays.fill(retval, pPad);
    for(int i=0; i < pAry.length; i++) { retval[i] = pAry[i]; }
    return retval;
  }
  /** This method is identical to <code>pad(int[], int, int)</code> except that it works on <b>double</b> arrays.
      @param pAry The double array to be padded.
      @param pLength The minimum length of the returned array.
      @param pPad The padding constant.
  */
  public static double[] pad(double[] pAry, int pLength, double pPad) {
    if(pAry == null) return pAry;
    double[] retval = new double[Math.max(pLength, pAry.length)];
    Arrays.fill(retval, pPad);
    for(int i=0; i < pAry.length; i++) { retval[i] = pAry[i]; }
    return retval;
  }
  /** Pick a random name and gender-appropriate pronouns. The String[] array returned has nine elements: [0] is the name, [1] is 'his/her,' [2] is 'him/her,' 
      [3] is 'he/she,' and [4] is 'his/hers.' Elements [5] through [8] exactly mirror [1] through [4] but the words are capitalized.
  */
  public static String[] getPerson() {
    int i = MathEngine.random(0, SOME_NAMES.length);
    String[] retval = new String[9];
    retval[0] = SOME_NAMES[i];
    if((i % 2) == 0) {
      retval[1] = "his"; retval[2] = "him"; retval[3] = "he"; retval[4] = "his";
      retval[5] = "His"; retval[6] = "Him"; retval[7] = "He"; retval[8] = "His";
    } else {
      retval[1] = "her"; retval[2] = "her"; retval[3] = "she"; retval[4] = "hers";
      retval[5] = "Her"; retval[6] = "Her"; retval[7] = "She"; retval[8] = "Hers";
    }
    
    return retval;
  }
  /** At times, a number needs formatting. This method returns a plain instance of java.text.NumberFormat for just such occasions. Its rounding mode is automatically 
      set to HALF_UP.
  */
  public static NumberFormat newNumberFormat() {
    NumberFormat retval = NumberFormat.getInstance();
    retval.setRoundingMode(java.math.RoundingMode.HALF_UP);
    return retval;
  }
  /** Sometimes a floating-point number needs to have its number of decimal places fixed. This method does that and returns the number as a String. If pPlace is 
      less than zero the number is returned unaltered. If zero, zero places are fixed. Grouping may be used or not; pPlace must be zero or more or the number 
      will not be formatted at all.
      @param pNumber The number to be rounded.
      @param pPlace The number of decimal places to allow.
      @param pGroup Flag passed to <code>NumberFormat.setGroupingUsed()</code>.
  */
  public static String fix(double pNumber, int pPlace, boolean pGroup) {
    if(pPlace < 0) return ""+pNumber;
    myNumberFormat.setMaximumFractionDigits(pPlace);
    myNumberFormat.setGroupingUsed(pGroup);
    return myNumberFormat.format(pNumber);
  }
  /** Calculate the Greatest Common Factor of the numbers given. If either is negative it is made positive before calculation. If either value is 0, 1 is returned. */
  public static int gcf(int pL, int pR) {
    if((pL == 0) || (pR == 0)) return 1;
    pL = (pL < 0)?-1*pL:pL; pR = (pR < 0)?-1*pR:pR;
    int nStop = ((int)Math.max(pL, pR)) +1;
    int retval = 1;
    for(int i=1; i < nStop; i++) {
      if(((pL % i) == 0) && ((pR % i) == 0)) retval = i;
    }
    return retval;
  }
  /** Find the Least Common Multiple of the given numbers. If either is negative it is made positive. If either is zero, zero is returned.*/
  public static int lcm(int pL, int pR) {
    if((pL == 0) || (pR == 0)) return 0;
    pL = (pL < 0)?-1*pL:pL; pR = (pR < 0)?-1*pR:pR;
    return (pL * pR) / gcf(pL, pR);
  }

  /** Find a number not divisible by a given divisor from a range of candidates. Returns 1 if none of the numbers are indivisible or if the divisor is 1, -1 or 0. 
      If pHigh is less than or equal to pLow the method returns 1 immediately. If more than one indivisible value exists within the range, one is picked randomly. 
      Regardless or the range values, zero will never be returned.
      @param pLow The low value from the candidate range. Included. Must be less than 'pHigh'.
      @param pHigh The high value from the candidate range. Excluded.
      @param pDivisor The number that does NOT divide the one selected.
      @return A number from [pLow, pHigh) that is NOT divisible by pDivisor or 1 if one cannot be found.
  */
  public static int indivisible(int pLow, int pHigh, int pDivisor) {
    if((pLow >= pHigh) || ((-1 <= pDivisor) && (pDivisor <= 1))) return 1;
    String retval = "";
    for(int i = pLow; i < pHigh; i++) {
      if((-1 <= i) && (i <= 1)) continue;
      if((i % pDivisor) != 0) retval = retval + " " + i;
    }
    retval = Question.validString(retval)?retval:"1";
    return MathEngine.splickInt(retval.trim());
  }

  /** Pick a number from a range that has no common factors with the given number greater than 1. If no such value exists, if the low value is greater than or 
      equal to the high value or if the given number is -1, 0 or 1, the method returns 1 immediately.
      @param pLow The low value from the candidate range. Included. Must be less than 'pHigh'.
      @param pHigh The high value from the candidate range. Excluded.
      @param pCoprime The number that has no common factors with the one selected.
      @return A number from [pLow, pHigh) that is relatively prime to pCoprime or 1 if one cannot be found.
  */
  public static int coprime(int pLow, int pHigh, int pCoprime) {
    if((pLow >= pHigh) || ((-1 <= pCoprime) && (pCoprime <= 1))) return 1;
    String retval = "";
    for(int i = pLow; i < pHigh; i++) {
      if((-1 <= i) && (i <= 1)) continue;
      if((gcf(i, pCoprime)) == 1) retval = retval + " " + i;
    }
    retval = Question.validString(retval)?retval:"1";
    return MathEngine.splickInt(retval.trim());
  }

  /** Return a SCROB instance of this class. */
  public cox.jmatt.java.MathTools.scrob.Arithmetic_SCROB getSCROB() { return new cox.jmatt.java.MathTools.scrob.Arithmetic_SCROB(); }

}
