package cox.jmatt.java.MathTools;

import cox.jmatt.java.MathTools.test.*;
import cox.jmatt.java.MathTools.util.*;
import cox.jmatt.java.MathTools.demo.gpgdemo.*;
import java.io.*;
import java.util.*;
import javax.naming.Context;

/** <p>This class, consisting solely of static methods, provides a central logging point and data repository. Its constructor is private and final as CapCom should 
    NEVER be instantiated or subclassed. The other classes in MathTools use CapCom to report log messages and to store and retrieve data used internally. 
    Applications built using MathTools may also use CapCom and are encouraged to do so! It provides several services toward that end that the MathTools classes 
    themselves don't use.</p>
    <p>CapCom, from NASA's Capsule Command, serves as a centralized point of communication between MathTools classes and the outside world. This communication 
    is based around a global MathPrinter which is set by the <code>setPrinter()</code> method. The MathPrinter interface itself is lightweight, minimalistic, and 
    flexible enough to suit the needs of all MathTools classes: MathTools has no external dependencies for logging or communication. For apps that do, MathPrinter 
    is very easy to implement and its methods were named to (hopefully) avoid collision with other logging packages. CapCom provides static convenience methods 
    for logging debugging information and errors, as well as a <code>println()</code> method and the <code>mathPrintln()</code> method. Most MathTools methods make 
    use of the <code>error()</code> method to report user-correctible goofs.</p>
    <p>NOTE: CapCom implements all MathPrinter methods to allow re-sending to the global printer. It does NOT implement the MathPrinter interface because it 
    should NEVER, NEVER be set as the Global Printer!!!</p>
    <p>In addition to logging and communication, CapCom provides a central and somewhat anonymous data repository service. It provides The Shelf for storing, 
    fetching and removing Objects. None of the MathTools classes use this but applications extending them might need such. Finally, CapCom provides facilities 
    for the Global Pizza. The Global Pizza is a <code>java.util.Properties</code> Object that can be set and retrieved via static methods. This is primarily intended 
    for scripting applications built using MathTools. Conceptually, the application sets the Global Pizza before executing a script. The script is then allowed 
    access to the Global Pizza for fetching String parameters or configuration information. MathConsole provides exactly this capability.</p>
    <p>For classes that potentially require code outside the MathTools JAR, CapCom provides a place to store and retrieve a <code>ClassLoader</code>. The 
    MathClassLoader is set automatically by <code>ScriptPanel</code> from its <code>URLLoaderPanel</code>, whether or not it uses the ClassLoader for its 
    ScriptEngineFactories. Other MathTools classes that might require such access either use this ClassLoader automatically OR have methods to trigger its use.</p>
    <p>To facilitate MathTools scripting, CapCom defines the <code>installStandardObjects()</code> method. This method takes as a parameter a <code>javax.script.ScriptEngine</code>
    and installs the MathTools classes along with a MathConsole into it. See the {@link cox.jmatt.java.MathTools.demo.ScriptPanel ScriptPanel} docs for the tools 
    installed and their names.</p>
*/
public class CapCom {

  private CapCom() { super(); }

  private static MathPrinter thePrinter = null;
  private static Properties thePizza = null;
  private static MAuthenticator theChecker = null;
  private static long numHits = 0;
  private static Hashtable<String, Object> theShelf = new Hashtable<String, Object>();
  private static ClassLoader myClassLoader = null;
  private static Context myContext = null;
  private static Object theScriptReturn = null;

  /** This field controls whether or not a <code>MathConsole</code> instance is allowed to set the Global Authenticator. By default it is false. If this is not 
      desirable it can be changed from within a MathTools application. <code>ScriptPanel</code> also sets this field; its default value is true but is easily 
      changed via a Config Panel checkbox.
  */
  public static boolean allowConsoleAuthenticatorAccess = false;

  /** This field determines whether or not <code>CapCom</code> will create accessories on demand. It is false by default but <code>ScriptPanel</code> resets it 
      to true automatically. Other applications should set it appropriately otherwise accessory classes will not be available.
  */
  public static boolean allowAccessories = false;


  /** <p>The MathTools ClassLoader is maintained by CapCom for MathTools (or application!) classes that might need one. By default <code>URLLoaderPanel</code> 
      sets it before script execution. If a ClassLoader is set this method returns it, otherwise it returns the loader used by the MathTools package.</p>
      <p>NOTE: This method <b>always</b> returns a valid <code>ClassLoader</code>! If no loader is set or it is set to null this method automatically resets it 
      to <code>Class.forName('..CapCom').getClassLoader()</code>.</p>
  */
  public static ClassLoader getMathClassLoader() {
    if(myClassLoader == null) {
      try {
        myClassLoader = Class.forName("cox.jmatt.java.MathTools.CapCom").getClassLoader();
      } catch(Exception ex) { error("Exception setting Math ClassLoader", ex); }
    }
    return myClassLoader;
  }

  /** <p>This method is used to set the MathTools ClassLoader. If set to null the next call to <code>getClassLoader()</code> sets the loader used by the MathTools 
      package and not the bootstrap loader.</p>
      <p>IMPORTANT NOTE: The ClassLoader maintained here is <b>not</b> copied. Set with care!</p>
  */
  public static void setMathClassLoader(ClassLoader pLoader) {
    debug("MathClassLoader "+((pLoader == null)?"cleared.":"set: "+pLoader.toString()));
    myClassLoader = pLoader;
  }

  /** Use this method to retrieve a JNDI naming Context if one exists. Although <code>MathDBC</code> is the only class that currently uses it such a thing might 
      be retrofitted into other classes, or user applications may also require it.
  */
  public static Context getNamingContext() { return myContext; }

  /** This method sets a JNDI naming Context usable by any of the MathTools classes or applications embedding them.
      @param pContext The <code>javax.naming.Context</code> to be set.
  */
  public static void setNamingContext(Context pContext) { myContext = pContext; }

  /** This method sets or clears a global <code>MAuthenticator</code>. The fully-qualified class name must be given and the MathTools ClassLoadar is automatically 
      used. If the class is successfully created it is set as the global authenticator. If the supplied String is null or empty the authenticator is cleared. Any 
      errors are reported at Error level.
      @param pAuthClass The fully-qualified classname of the <code>MAuthenticator</code> implementation to use.
      @param pContext The context Object sent to initialize the Authenticator.
  */
  public static void setMAuthenticator(String pAuthClass, Object pContext) {
    theChecker = null;
    if(!Question.validString(pAuthClass)) return;
    try {
      theChecker = (MAuthenticator) (Class.forName(pAuthClass, true, getMathClassLoader())).newInstance();
      if(theChecker != null) theChecker.initMAuthenticator(pContext);
    } catch(Exception ex) { error("Error setting MAuthenticator", ex); }
  }

  /** Check an authentication code and level using the global Authenticator. If no <code>MAuthenticator</code> instance is set the return value is automatically 
      false. If one is set it is checked and this method returns whatever value it does.
      @param pLevel The AUTH_ level constant of the code being checked.
      @param pCode The authentication code actually being checked.
      @return false if no <code>MAuthenticator</code> is set, or the value returned by checking the authenticator that IS set.
  */
  public static boolean authenticate(byte pLevel, String pCode) {
    if(theChecker == null) return false;
    return theChecker.mAuthenticate(pLevel, pCode);
  }

  /** If an Authenticator is set this method calls on it to generate an array of authentication codes. If no <code>MAuthenticator</code> is set the return value 
      is an empty array.
      @param pLevel The AUTH_ level of the codes to generate.
      @param pNumber The number of codes to generate.
      @param pParams Implementation-specific Object parameters.
      @return Whatever the global authenticator returns, or an empty array if none is set.
  */
  public static String[] generateAuthenticationCodes(byte pLevel, int pNumber, Object... pParams) {
    String[] retval = {};
    if(theChecker == null) return retval;
    retval = theChecker.generateCodes(pLevel, pNumber, pParams);
    return retval;
  }

  /** Create a <code>MathToolsAccessory</code> from its full class name. If <code>allowAccessories</code> is false or the class cannot be created null is 
      returned. If the name is null or blank the return value is null.
      @param pAccessory The fully-qualified classname of the accessory to create
      @return A new instance of the accessory or null if not allowed or not found.
  */
  public static MathToolsAccessory createAccessory(String pAccessory) {
    MathToolsAccessory retval = null;
    if(!Question.validString(pAccessory)) return retval;
    try {
      retval = (MathToolsAccessory) Class.forName(pAccessory, true, getMathClassLoader()).newInstance();
    } catch(Exception ex) { error("Error creating: "+pAccessory); }
    return retval;
  }

  /** Use this method to set a global MathPrinter for all MathTools classes to use. If the parameter is null (default), no MathPrinter is set and any 
      previously-set MathPrinter is removed.
  */
  public static void setPrinter(MathPrinter pPrinter) { thePrinter = pPrinter; }

  /** Check to see if a Global Printer is set. */
  public static boolean isPrinterSet() { return !(thePrinter == null); }

  /** Send a message to the printer's mathPrintln() method, if a printer is set.*/
  public static void mathPrintln(String pMesg) {
    if(thePrinter == null) return;
    thePrinter.mathPrintln(pMesg);
  }
  /** Convenience method for mathPrintln(). */
  public static void println(String pMesg) { mathPrintln(pMesg); }

  /** This method reflects processMathObject() calls to the global MathPrinter, if such exists. If the printer is defined, processMathObject() returns whatever it 
      returns. If not, it returns pData unaltered.
      @param pData The Object to be processed.
      @return Whatever the global printer does, or pData if it does not exist.
  */
  public static Object processMathObject(Object pData) {
    if(thePrinter == null) return pData;
    return thePrinter.processMathObject(pData);
  }

  /** Echo a log message to the global MathPrinter, if it is set. In addition to this method, two convenience methods are also defined for each log level: Info, 
      Debug, and Error. The first method accepts a String parameter, the second a String and Throwable. The convenience methods all reflect to this one. This 
      method itself exactly mimics the <code>mathLog()</code> method specified in MathPrinter. If no MathPrinter is set this method returns silently.
  */
  public static void mathLog(byte pLevel, String pMesg, Throwable pThr) {
    if(thePrinter == null) return;
    thePrinter.mathLog(pLevel, pMesg, pThr);
  }
  public static void error(String pMesg) { mathLog(MathPrinter.LOG_ERROR, pMesg, null); }
  public static void error(String pMesg, Throwable pThr) { mathLog(MathPrinter.LOG_ERROR, pMesg, pThr); }
  public static void debug(String pMesg) { mathLog(MathPrinter.LOG_DEBUG, pMesg, null); }
  public static void debug(String pMesg, Throwable pThr) { mathLog(MathPrinter.LOG_DEBUG, pMesg, pThr); }
  public static void info(String pMesg) { mathLog(MathPrinter.LOG_INFO, pMesg, null); }
  public static void info(String pMesg, Throwable pThr) { mathLog(MathPrinter.LOG_INFO, pMesg, pThr); }

  /** Utility method to extract the stack trace from a Throwable and return it as a String. */
  public static String getStackTrace(Throwable pThr) {
    if(pThr == null) return "";
    String retval = "CapCom: Error extracting Stack Trace.";
    try {
      ByteArrayOutputStream baOS = new ByteArrayOutputStream();
      PrintStream psOut = new PrintStream(baOS, true);
      pThr.printStackTrace(psOut);
      psOut.flush();
      psOut.close();
      baOS.flush();
      baOS.close();
      retval = baOS.toString();
    } catch(Exception ex) {}
    return retval;
  }

  /** Set in place a Global Pizza. The Global Pizza is a Properties object useful to scripting applications. Setting the Global Pizza to null clears it. This 
      method does make a copy of pPizza if it is not null, so altering the original will not affect the Global Pizza.
  */
  public static void setPizza(Properties pPizza) {
    if(pPizza == null) {
      thePizza = null;
      return;
    }
    thePizza = new Properties();
    thePizza.putAll(pPizza);
  }
  /** Retrieve a copy of the Global Pizza. If it is null, an empty Properties object is returned. If the Global Pizza is set, a copy of it is returned. */
  public static Properties getPizza() {
    if(thePizza == null) return new Properties();
    Properties retval = new Properties();
    retval.putAll(thePizza);
    return retval;
  }
  /** CapCom maintains a simple counter useful for generating application-unique numbers. It's simply an internal variable that increments each time this 
      method is called.
  */
  public static long hitCount() { return numHits++; }

  /** This method provides a place for MathTools classes, or applications that use them, to store arbitrary data (The Shelf). If the key is not valid or the data 
      is null both are ignored. It is backed by an internal static <code>Hashtable&lt;String, Object&gt;</code>.
      @param pKey The String key by which the data is to be stored.
      @param pData The data Object to be stored.
  */
  public static void storeObject(String pKey, Object pData) {
    if(Question.validString(pKey) && (pData != null)) theShelf.put(pKey, pData);
  }
  /** Fetch a previously-stored Object from the Shelf. If the key is not valid or the shelf doesn't have the item this method returns null.
      @param pKey The name of the Object to pull off the Shelf.
      @return The Object shelved under pKey, or null if it does not exist.
  */
  public static Object getObject(String pKey) {
    if(!Question.validString(pKey) || !theShelf.containsKey(pKey)) return null;
    return theShelf.get(pKey);
  }
  /** Pull something off the Shelf and remove it from storage completely. Returns null if the key is not valid or the Object isn't there.
      @param pKey The key of the Object to remove from the Shelf.
      @return The specified Object or null if it's not there.
  */
  public static Object removeObject(String pKey) {
    if(!Question.validString(pKey)) return null;
    return theShelf.remove(pKey);
  }
  /** Set the Object a script returns. This is the method {@link cox.jmatt.java.MathTools.util.MathConsole#returns MathConsole.returns()} uses.
      @param pRetObj The (script return) Object set by the MathConsole method.
  */
  public static void setScriptReturnObject(Object pRetObj) { theScriptReturn = pRetObj; }
  /** Retrieve the script return Object set by <code>MathConsole</code>. */
  public static Object getScriptReturnObject() { return theScriptReturn; }
  /** <p>This method installs the Standard Objects into the ScriptEngine provided. See the {@link cox.jmatt.java.MathTools.demo.ScriptPanel ScriptPanel} docs for 
      details. NOTE: The <b>static</b> versions of these objects are the ones installed! Classes with SCROB aliases all have a <code>getSCROB()</code> method to 
      acquire one.</p>
      <p><code>CapCom</code> exposes the following MathTools classes as its Standard Objects:
      <ul>
        <li>'Console' is a MathConsole instance.</li>
        <li>'MEngine' is a MathEngine instance. (SCROB)</li>
        <li>'Arithmetic' is an Arithmetic instance. (SCROB)</li>
        <li>'MathXML' is a MathXML instance. (SCROB)</li>
        <li>'Question' is a Question instance. (XML)</li>
        <li>'Polynomial' is a default (zero) Polynomial. (XML)</li>
        <li>'MFraction' is an MFraction representing 0/1. (XML)</li>
        <li>'MStat' is an empty MathStat. (XML)</li>
        <li>'MEquation' is a MEquation instance. (XML)</li>
        <li>'MRadical' is a default MRadical. (XML)</li>
        <li>'MComplex' is a default MComplex. (XML)</li>
        <li>'MSet' is a Real Number <code>MathSet</code> instance.</li>
        <li>'MMatrix' is a default MMatrix. (XML)</li>
        <li>'MBuilder' is a MMatrixBuilder instance. (SCROB)</li>
        <li>'Percent' is a Percent instance.</li>
        <li>'MTest' is an MTest instance. (XML)</li>
        <li>'MTFormatter' is a new MathTestFormatter instance.</li>
        <li>'COFormat' is a COFormatterFactory instance.</li>
        <li>'MTXPro' is a MTXProcessor().</li>
        <li>'MGraph' is a GraphEngineFactory instance.</li>
        <li>'MFile' is a base MathFileIO instance. (Filesystem)</li>
        <li>'MathDB' is a MathDBC instance.</li>
        <li>'MDraw' is a MathDrawSVG instance.</li>
        <li>'MathHTM' is a MathGenHTML instance.</li>
        <li>'SimpleTag' is a SimpleTagImpl instance from the markup package.</li>
        <li>'MLatex' is a MathGenLatex instance. (AWT)</li>
        <li>'Notebook' is a GraphPanelFactory from the demo.gpgdemo package.</li>
      </ul>
      Classes marked 'SCROB' are the classes with SCROBs. If the standard (static) instances of these classes do not function, each of these classes has a 
      getSCROB() method that returns an instance of its SCROB.</p>
      <p>Classes marked 'XML' have a <code>toXML()</code> method to generate MathTools XML tags. Instances of these classes may be created with specific data and 
      written to a file (or String) that MTXProcessor can convert back into their MathTools class instances. MFormatElement, Problem, and Answer also appear as XML 
      tags but these are contained within other classes. <b>Note</b>: The instances exposed to ScriptPanel are <i>default</i> instances; the data within them is 
      not particularly interesting!</p>
      <p>The 'pAWT' parameter is present to allow suppression of tools that depend on <code>java.awt.*</code> classes. The 'pFiles' parameter allows suppression 
      of <code>MathFileIO</code>.</p>
      <p>CRITICAL NOTE: The name-Object bindings listed here are <i>fragile</i>!! Executing the statement 'Console = null' in a script REMOVES THE CONSOLE OBJECT!! 
      Assign with care! This is done purposefully to allow such code as <code>Arithmetic = Arithmetic.getSCROB()</code> to function properly and allow scripts to be 
      more portable.</p>
      
  */
  public static final void installStandardObjects(javax.script.ScriptEngine pEngine, boolean pAWT, boolean pFiles) {
    if(pEngine == null) return;
    pEngine.put("Console", new MathConsole());
    pEngine.put("MEngine", new MathEngine());
    pEngine.put("Question", new Question());
    pEngine.put("Arithmetic", new Arithmetic());
    pEngine.put("MathXML", new MathXML());
    pEngine.put("Polynomial", new Polynomial());
    pEngine.put("Percent", new Percent());
    pEngine.put("MFraction", new MFraction());
    pEngine.put("MStat", new MathStat());
    pEngine.put("MBuilder", new MMatrixBuilder());
    pEngine.put("MMatrix", new MMatrix());
    pEngine.put("MComplex", new MComplex());
    pEngine.put("MSet", MathSet.realNumbers());
    pEngine.put("MTest", new MTest());
    pEngine.put("MEquation", new MEquation());
    pEngine.put("MRadical", new MRadical());
    pEngine.put("MTXPro", new MTXProcessor());
    pEngine.put("MGraph", cox.jmatt.java.MathTools.graphing.GraphEngineFactory.newInstance());
    if(pFiles) pEngine.put("MFile", MathFileIO.newInstance());
    pEngine.put("MDraw", new cox.jmatt.java.MathTools.markup.svg.MathDrawSVG());
    pEngine.put("MathHTM", new cox.jmatt.java.MathTools.markup.html.MathGenHTML());
    pEngine.put("SimpleTag", new cox.jmatt.java.MathTools.markup.SimpleTagImpl());
    pEngine.put("MathDB", new MathDBC());
    pEngine.put("MTFormatter", new MathTestFormatter());
    pEngine.put("COFormat", new COFormatterFactory());
    if(pAWT) pEngine.put("Notebook", new cox.jmatt.java.MathTools.demo.gpgdemo.GraphPanelFactory());
    pEngine.put("MLatex", new cox.jmatt.java.MathTools.markup.latex.MathGenLatex());

  }
}
