package cox.jmatt.java.MathTools;

import java.util.*;
import java.util.regex.*;

/** <p>This class provides MathTools support for imaginary and complex numbers. As with <code>MRadical</code> and <code>MathStat</code> this is primarily a formatting 
    class although it does provide some basic numerical manipulation. Once created a <code>MComplex</code> is immutable and any method that would change it returns 
    a new instance. Since <code>MComplex</code> is primarily for formatting and simple manipulation the components are represented as type int. Three formal 
    constructors are defined along with static and instance <code>newInstance()</code> methods and the <code>reParse()</code> method can build an instance from a 
    String of the form 'A + Bi'.</p>
    <p><code>MComplex</code> defines six condition-testing methods, two of which involve another instance. A complex number can be tested as to whether it is (purely) 
    real or imaginary, zero or complex. It can be tested against another <code>MComplex</code> for equality and for being its conjugate.</p>
    <p>In terms of numerical manipulation there are addition, subtraction and two multiplication methods. There is a method to form the complex conjugate and one 
    to raise 'i' to an arbitrary (integer) power. All of these methods return <code>MComplex</code> and, since instances are immutable, the original is not changed 
    by doing so. The <code>abs()</code> method returns the absolute value (or <i>magnitude</i>) of the number as type double. This is also available as a formatting 
    token.</p>
    <h3>Formatting</h3>
    <p>This is the true strength of this class. For technologies with special symbols for the imaginary number, 'i' can be set to any letter or String. The 
    <code>format()</code> method defines a plethora of replacement tokens, many of which deal with the imaginary part, its sign and the 'i' symbol. There are three 
    pre-defined formats: DEFAULT, STANDARD and TYPICAL. The former always presents a complex number in the form 'A + Bi', even when one part is zero. STANDARD is 
    either 'A + Bi' or 'A - Bi' depending on the sign of B. TYPICAL format omits a component that is zero, including the middle sign if the imaginary part is zero. 
    The <code>toString()</code> and <code>toXML()</code> methods return a constructor-like representation and a MathTools-XML tag, respectively.</p>
    <p>The <code>divide()</code> method is a hybrid between presentation and manipulation. Since complex division <i>might</i> result in a fractional value the 
    return type for this method is String. After it is accomplished the components of the division are presented as typewriter-style fractions that can be converted 
    to <code>MFraction</code>s if necessary. To facilitate this the middle sign and 'i' symbol appear as String arguments, allowing them to be omitted or replaced 
    with whatever is appropriate for further processing.</p>
*/

public class MComplex {

  /** String constant for the default format: A + Bi. Both components appear, no special formatting is done. */
  public static final String FORMAT_DEFAULT = "<0> + <1><2>";

  /** String constant for standard form: A + Bi or A - Bi, depending on B's sign. Zero parts are displayed. */
  public static final String FORMAT_STANDARD = "<0> <9> <7><2>";

  /** String constant for the typical representation of a complex number. If either part is zero it is not displayed. */
  public static final String FORMAT_TYPICAL = "<4> <11> <13><6>";

  private static Pattern patComplex = Pattern.compile("\\s*(-?\\d+) *([+-]) *[(]?(-?\\d+)[}]?i[)]?.*");
  private static String defaultFormat = FORMAT_DEFAULT, my_i = "i";

  private int myReal = 0, myImag = 0;
  private String myFormat = defaultFormat;

  /** Basic scripting constructor. It creates the complex number 0 + 0i. */
  public MComplex() { super(); }

  /** Standard constructor with both components specified. */
  public MComplex(int pReal, int pImag) {
    super();
    myReal = pReal;
    myImag = pImag;
  }

  /** Copy constructor. */
  public MComplex(MComplex pCplx) {
    super();
    if(pCplx != null) {
      myReal = pCplx.real();
      myImag = pCplx.imag();
    }
  }

  /** <p>Attempt to parse a complex number of the form 'A + Bi'. If the String cannot be parsed zero is returned. If the input String cannor be parsed or one of 
      the components is not an Integer a debug-level message is given. If 'pCplx' is null or blank zero is returned immediately.</p>
      <p>In order to parse properly the complex number <b>must</b> be of the form 'A + Bi'. The middle sign may be plus or minus and may have spaces around it 
      but the signs on the numbers, if present, must <b>not</b> have any space between. Also, if one component is zero it must still be present! The 're' in this 
      method name is for <i>regular expression</i> and that is how the value is parsed. The 'i' is a part of the pattern as is the no-space requirement between a 
      component's sign and the number itself.</p>
  */
  public static MComplex reParse(String pCplx) {
    MComplex retval = new MComplex();
    if(!Question.validString(pCplx)) return retval;
    Matcher mNum = patComplex.matcher(pCplx);
    if(mNum.matches()) {
      CapCom.debug("Complex number parsed: "+pCplx);
      try {
        int iReal = Integer.parseInt(mNum.group(1)), iImag = Integer.parseInt(mNum.group(3));
        if("-".equals(mNum.group(2))) iImag *= -1;
        retval = new MComplex(iReal, iImag);
      } catch(Exception ex) { CapCom.debug("Error parsing complex values", ex); }
    } else {
      CapCom.debug("Cannot parse complex number from: "+pCplx);
    }
    return retval;
  }

  /** Instance method to parse a complex number. */
  public MComplex re_Parse(String pCplx) { return reParse(pCplx); }

  /** Create and return a new <code>MComplex</code>. */
  public static MComplex newInstance(int pReal, int pImag) { return new MComplex(pReal, pImag); }

  /** Create and return a new <code>MComplex</code>, instance version. */
  public MComplex new_Instance(int pReal, int pImag) { return new MComplex(pReal, pImag); }

  /** Create a complex number from a Knobs String. If 'pKnobs' is null or blank zero is returned. If there are not enough numbers zeros will be used. Any leftovers 
      are silently ignored.
  */
  public static MComplex fromKnobs(String pKnobs) {
    if(!Question.validString(pKnobs)) return new MComplex();
    int[] vals = MathEngine.setKnobs(pKnobs);
    vals = Arithmetic.pad(vals, 2, 0);
    return new MComplex(vals[0], vals[1]);
  }

  /** Instance Knob-creator. */
  public MComplex from_Knobs(String pKnobs) { return fromKnobs(pKnobs); }

  /** Set a new default format for subsequent complex numbers. If null or blank FORMAT_DEFAULT is used. */
  public void setDefaultFormat(String pFormat) {
    defaultFormat = Question.validString(pFormat)?pFormat:FORMAT_DEFAULT;
  }

  /** Instance version. */
  public void setNewDefaultFormat(String pFormat) { setDefaultFormat(pFormat); }

  /** Set the format for this instance. If null or blank FORMAT_DEFAULT is used. */
  public void setFormat(String pFormat) {
    myFormat = Question.validString(pFormat)?pFormat:FORMAT_DEFAULT;
  }

  /** <p>Set the global (static) symbol used to represent the imaginary number 'i'. The default is 'i' but this method allows it to be changed to accommodate 
      formatting applications that have a specific symbol or character sequence for it. Setting null or blank resets the default.</p>
      <p>NOTE: Setting this value affects <i>formatting</i> only! The <code>reParse(), toString()</code> and <code>toXML()</code> methods <b>always</b> use 'i'.</p>
      @param pNewI The new 'i' symbol.
  */
  public static void set_i(String pNewI) { my_i = (Question.validString(pNewI)?pNewI:"i"); }

  /** Instance flavored 'set_i'. */
  public void setNew_i(String pNewI) { set_i(pNewI); }

  /** Raise 'i' to a power. If pExp is negative it is computed by rationalizing the reciprocal of 'i' to that power. */
  public static MComplex iPow(int pExp) {
    MComplex retval = new MComplex(1,0);
    int iExp = Math.abs(pExp) % 4;
    if(pExp < 0) iExp = 4 - iExp;
    switch(iExp) {
      case 1: retval = new MComplex(0,1); break;
      case 2: retval = new MComplex(-1,0); break;
      case 3: retval = new MComplex(0,-1);break;
    default:
    }
    return retval;
  }

  /** Instance method to raise 'i' to a power. */
  public MComplex i_Pow(int pExp) { return iPow(pExp); }

  /** @deprecated Use {@link #kaPow kaPow} instead. */
  public static String iKaPow(int pExp) { return iPow(pExp).format(FORMAT_TYPICAL); }

  /** This is an extreme laziness method. It returns <code>iPow(<i>pExp</i>).format(FORMAT_TYPICAL)</code> because that can be a pain to type. */
  public static String kaPow(int pExp) { return iPow(pExp).format(FORMAT_TYPICAL); }

  /** @deprecated Use {@link #ka_Pow ka_Pow} instead. */
  public String i_KaPow(int pExp) { return iPow(pExp).format(FORMAT_TYPICAL); }

  /** Extremely lazy instance Ka-POW. */
  public String ka_Pow(int pExp) { return iPow(pExp).format(FORMAT_TYPICAL); }

  /** Return the real part of this complex number. */
  public int real() { return myReal; }

  /** Return the imaginary part of this number. */
  public int imag() { return myImag; }

  /** Return true if the imaginary part of this number is zero and the real part isn't. */
  public boolean isReal() { return ((myReal != 0) && (myImag == 0)); }

  /** Return true if the real part of this number is zero and the imaginary part is not.. */
  public boolean isImaginary() { return ((myReal == 0) && (myImag != 0)); }

  /** Return true if both parts of this number are nonzero. */
  public boolean isComplex() { return ((myReal != 0) && (myImag != 0)); }

  /** Return true if both parts of this number are zero. */
  public boolean isZero() { return ((myReal == 0) && (myImag == 0)); }

  /** Calculate the absolute value (magnitude) of this number. */
  public double abs() { return Math.sqrt(myReal*myReal + myImag*myImag); }

  /** Return true if the real and imaginary parts of this and 'pRHS' are equal. Returns false if 'pRHS' is null. */
  public boolean equals(MComplex pRHS) {
    if(pRHS == null) return false;
    return ((myReal == pRHS.real()) && (myImag == pRHS.imag()));
  }

  /** Return true if 'pRHS' is the complex conjugate of this number. Returns false if 'pRHS' is null. */
  public boolean isConjugate(MComplex pRHS) {
    if(pRHS == null) return false;
    return ((myReal == pRHS.real()) && (myImag == (-1*pRHS.imag())));
  }

  /** Add this to another complex number. Neither is changed, a new instance is returned. If 'pRHS' is null this is returned. */
  public MComplex add(MComplex pRHS) {
    if(pRHS == null) return this;
    return new MComplex(myReal + pRHS.real(), myImag + pRHS.imag());
  }

  /** Subtract another complex number from this one. If 'pRHS' is null nothing happens and this is returned. */
  public MComplex subtract(MComplex pRHS) {
    if(pRHS == null) return this;
    return new MComplex(myReal - pRHS.real(), myImag - pRHS.imag());
  }

  /** Multiply this by another complex number. Neither is changed, if 'pRHS' is null this is returned. */
  public MComplex multiply(MComplex pRHS) {
    if(pRHS == null) return this;
    return new MComplex(myReal*pRHS.real() - myImag*pRHS.imag(), myImag*pRHS.real() + myReal*pRHS.imag());
  }

  /** Multiply this number by a scalar (int) constant. A new MComplex is created. */
  public MComplex multiply(int pVal) { return new MComplex(myReal*pVal, myImag*pVal); }

  /** Form the complex conjugate of this number: A + Bi becomes A - Bi. */
  public MComplex conjugate() { return new MComplex(myReal, -1*myImag); }

  /** <p>Format this <code>MComplex</code> number per the template supplied. The following replacement tokens are defined:
      <ul>
        <li>&lt;0&gt; - The real part.</li>
        <li>&lt;1&gt; - The imaginary part.</li>
        <li>&lt;2&gt; - The current 'i' symbol.</li>
        <li>&lt;3&gt; - The absolute value (magnitude) of this number.</li>
        <li>&lt;4&gt; - The implicit real part: blank if it is zero.</li>
        <li>&lt;5&gt; - The implicit imaginary part: blank if it is zero.</li>
        <li>&lt;6&gt; - The implicit 'i': blank if the imaginary part is zero.</li>
        <li>&lt;7&gt; - The absolute value of the imaginary part.</li>
        <li>&lt;8&gt; - The implicit absolute value of the imaginary part: blank if zero.</li>
        <li>&lt;9&gt; - The explicit sign of the imaginary part.</li>
        <li>&lt;10&gt; - The nonzero sign of the imaginary part: blank if it is zero, explicit otherwise.</li>
        <li>&lt;11&gt; - The smart middle sign: blank if the real part is zero and the imaginary part positive, otherwise the same as &lt;10&gt;.</li>
        <li>&lt;12&gt; - The smart imaginary part: blank if it is -1, 0 or 1; visible otherwise.</li>
        <li>&lt;13&gt; - The unsigned smart imaginary part: exactly like &lt;12&gt; but unsigned.</li>
        <li>&lt;14&gt; - The absolute value of the real part.</li>
        <li>&lt;15&gt; - The implied absolute value of the real part; blank if it is zero.
        <li>&lt;16&gt; - The explicit sign of the real part.</li>
        <li>&lt;17&gt; - The nonzero real sign. Blank if the real part is zero, &lt;15&gt; otherwise.</li>
      </ul></p>
      <p>If the template is null or blank it is returned.</p>
      @param pTemplat A <code>Question.fillTemplate()</code>-style formatting String.
      @return The supplied String template, filled using the defined tokens.
  */
  public String format(String pTemplat) {
    if(!Question.validString(pTemplat)) return pTemplat;
    String[] tokenz = new String[18];
    tokenz[0] = ""+myReal; tokenz[1]=""+myImag; tokenz[2]=my_i; tokenz[3]=""+abs();
    tokenz[4] = (myReal == 0)?"":tokenz[0]; tokenz[5] = (myImag == 0)?"":tokenz[1]; tokenz[6] = (myImag == 0)?"":my_i;
    tokenz[7]=""+Math.abs(myImag); tokenz[8]=(myImag == 0)?"":tokenz[7]; tokenz[9]=(myImag < 0)?"-":"+"; tokenz[10]=(myImag == 0)?"":tokenz[9];
    tokenz[11] = ((myReal == 0) && (myImag > 0))?"":tokenz[10]; tokenz[12] = ((-1 <= myImag) && (myImag <= 1))?"":tokenz[1];
    tokenz[13] = ((-1 <= myImag) && (myImag <= 1))?"":tokenz[7]; tokenz[14] = ""+Math.abs(myReal); tokenz[15] = (myReal == 0)?"":tokenz[14];
    tokenz[16] = (myReal < 0)?"-":"+"; tokenz[17] = (myReal == 0)?"":tokenz[15];
    return Question.fillTemplate(pTemplat, tokenz).trim();
  }

  /** Format this number per the internal format. */
  public String format() { return format(myFormat); }

  /** <p>This is a formatting method. Since division might result in a non-integer number this method returns a String consisting of one or two typewriter-style 
      fractions along with some embellishments. The 'pPlus' argument is for the sign (or whatever) between the two parts. the 'pImag' argument is the symbol used 
      for 'i'. If numerical manipulation is still required each part can be parsed into a <code>MFraction</code>; simply use the default values.</p>
      <p>The division itself is accomplished by multiplying this MComplex by the conjugate of the divisor. Each part is then expressed as a simple fraction of 
      the product part over the (real-only!) product of the divisor and its conjugate.</p>
      <p>If the divisor is null or zero an error String is returned.</p>
      @param pDivisor The MComplex to divide this one by.
      @param pPlus The separator between the two parts. Defaults to a single space.
      @param pImag The 'i' symbol. Defaults to blank.
  */
  public String divide(MComplex pDivisor, String pPlus, String pImag) {
    if((pDivisor == null) || pDivisor.isZero()) return ("Error: Divisor was null or zero.");
    pPlus = Question.validString(pPlus)?pPlus:" ";
    pImag = Question.validString(pImag)?pImag:"";
    int denom = pDivisor.real()*pDivisor.real() + pDivisor.imag()*pDivisor.imag();
    pDivisor = pDivisor.conjugate();
    pDivisor = multiply(pDivisor);
    return (""+pDivisor.real()+"/"+denom+pPlus+pDivisor.imag()+"/"+denom+pImag).trim();
  }

  /** Return a constructor-like representation of this complex number. */
  public String toString() { return "MComplex("+myReal+", "+myImag+");"; }

  /** Create a simplified (MathTools) XML representation of this complex number. The 'data' attribute contains a simple String representation and the 'pID' argument, 
      if present, is for an optional 'id' attribute. All of this is within an empty 'MComplex' tag.
  */
  public String toXML(String pID) {
    StringBuffer retval = new StringBuffer("<MComplex ");
    if(Question.validString(pID)) retval.append("id=\"").append(pID).append("\" ");
    retval.append("data=\"").append(myReal).append("+").append(myImag).append("i\" />");
    return retval.toString();
  }

  /** Return a copy of this complex number. */
  public MComplex copyOf() { return new MComplex(myReal, myImag); }

}

