package cox.jmatt.java.MathTools;

import cox.jmatt.java.MathTools.util.MathPrinter;
import cox.jmatt.java.MathTools.test.MTest;
import java.util.*;

/** <p>MEquation's purpose is to provide a fast and easy way to generate simple equations commonly found on algebra tests. This is by NO means an exhaustive list 
    of all the different equations or configurations of equations in the current world. Rather, it presents some of the equation types found on equation-solving 
    tests. The heart and soul of this class is the Equation enum. Each TYPE_ constant serves to <i>classify</i> the type of equation it represents, identify 
    placeholders (NOT variables) that need to be filled, and provide standard formats for presenting that equation. This class is not written to be the end-all 
    solver of every type of equation or inequality imaginable. Rather, it provides a <i>very</i> fast way to generate simple equations found on basic algebra 
    tests.</p>
    <p>A note on TYPE_ constants and their content. There is no particular rationale behind why <i>this</i> type of equation received <i>that</i> designation. The 
    TYPE_ constants simply appear in the order in which they were conceived.</p>
    <p>The main production method of this class, and its power method, is createQuestion(). This takes an Equation, a Knobs string, a variable, a relation, and a 
    format key. It grabs the Equation, sets the knobs, preformats, gets the appropriate format, sets the relation, puts the variable and preformat numbers 
    into a String[] array and calls Question.fillTemplate(). The resulting Problem is put into the problem part and a simple Answer of the form 'x = N' is put into 
    the Answer part. The variable and relation are adjusted for those input and the thing is formatted in proper MathML if that is the key chosen. The instance 
    version of createQuestion() is newQuestion() and it takes the same parameters.</p>
    <p>The other utility methods are getByType(), its instance version getEquationByType() and checkEquation(). The first two are safety methods for getting a TYPE_ 
    Equation without worrying about Exceptions. If one is thrown it is caught and reported at Debug level and TYPE_ERROR is returned. The checkEquation() method 
    is primarily for debugging the equations and experimenting with them. It prints the solution in place of the variable for the selected Equation, giving a quick 
    way to check.</p>

    <h3>Equations and Inequalities</h3>
    <p>Even though the name of the class is <code>M<u>Equation</u></code> it can be used to generate inequalities. The <code>createQuestion()</code> method 
    doesn't care what relation is sent in, it just formats. It is possible to reverse the equation: x + A = B becomes B = x + A. This is a very naive reversal 
    which can be problematic to inequalities.</p>
    <p>To solve this potential problem MEquation and the Equation enum implement BIRD. This is <b>B</b>lind <b>I</b>nequality <b>R</b>everse <b>D</b>etection and 
    it works as follows. Each Equation in the enum defines the <code>bird()</code> method. This is valid ONLY after a call to <code>preformat()</code> and it 
    indicates whether or not an inequality should be reversed based on the last numbers sent in. <code>createQuestion()</code> handles this automatically but it 
    must be addressed when building <i>inequalities</i> manually.</p>
    <p>Concerning reversal, <code>createQuestion()</code> applies NO intelligence to the relation sent in. That is, the method does NOT in any way alter the 
    relation itself. It <b>does not</b> change the sense of the inequality, only the format. Regardless of the relation, if BIRD says to reverse then the 
    <b>Answer</b> will be reversed. So, for the Problem '-2x &lt; 10' the Answer is formatted '-5 &lt; x'. BIRD still works when 'reverse' is requested. If reverse 
    is requested, the previous Problem would appear '10 &lt; -2x' and the Answer would be 'x &lt; -5'.</p>

    <h3>Preformatting and formatting</h3>
    <p>Each equation comes equipped with four built-in presentation formats: Standard, OpenOffice, MathML and LaTeX. While this is convenient for those who use 
    these technologies, it presents a problem for those who do not. For that reason each TYPE_ of equation has two <code>preformat()</code> methods. One accepts 
    a String of Knobs and the other an int[] array.</p>
    <p>The <code>preformat()</code> methods generate a String consisting of the <i>data</i> required for each type of equation: the values calculated for the 
    placeholders (capital letters) that define the equation and its solution. The preformat format is 'Answer : A B C... : <i>BIRD</i>'. This, along with the 
    'typewriter-style' representation of each equation is all that's needed to create the equation itself. MEquation's actual formatting methods do exactly this: 
    split the answer away from the constants and coefficients (placeholders) then create a Question.fillTemplate() token array. The variable is, of course, in 
    &lt;0&gt; and the letters go up in order from there. I.e. A = &lt;1&gt;, B = &lt;2&gt;, and so on up to the number of placeholders required. The last element 
    in the array (after the second colon in the preformat String) indicates whether or not an inequality needs reversing.</p>
    <p>In addition to the <code>preformat()</code> methods two <code>tokens()</code> methods are also available. One takes an int[] array and the other a String 
    of Knobs but both return the preformat data as a String[] array. The first element in the array (tokens()[0]) is the <b>answer</b> and the placeholders 
    start at element 1 and move upward in order; that is, tokens[1] = 'A', tokens[2] = 'B' etc. The <b>last</b> element in the array is the BIRD value 'true' or 
    'false'. These methods were added to facilitate easier custom formatting via <code>Question.fillTemplate()</code>.</p>
    <p>If desired, negative placeholders can be enclosed in parentheses. The <code>setNegativeParentheses()</code> method does this. It accepts two Strings that 
    correspond to the left and right parenthesis. These values work independently and if eithr is set to a non-null, non-blank value it will be applied to its 
    side of a negative placeholder. IMPORTANT NOTE: If set, these indicators <b>do</b> apply to isolated placeholders; e.g. the single constant on one side of 
    an equation. Setting either value to null or blank clears it and any values set remain until changed.</p>

    <h3>Calculations, Placeholders, and Knobs</h3>
    <p>All of the equations listed have placeholders and require Knobs. Even though the equations typically require as many Knobs as they have placeholders, the 
    two are not directly connected! That is to say calculations are performed on the <i>Knobs</i> in order to calculate values for the <i>placeholders</i>, the 
    constants or coefficients. The way the solution and placeholders are calculated is listed in the <code>Equation</code> docs.</p>
    <p>EXAMPLE: Equation.TYPE_400, 'Ax + B = C', requires 3 Knobs. Knob 0 is the answer ('x'), Knob 1 is the value of 'A' and Knob 2 is the value of 'B'. The 
    'C' placeholder is calculated as 'A*x + B'. The best way to get a feel for this is to examine the <code>enum</code> docs and experiment with the checkEquation() 
    method.</p>

    <h3>Making Nasty Equations</h3>
    <p><code>MEquation</code> and its <code>preformat()</code> methods were written to generate <i>nice</i> equations. Using these methods guarantees that no matter 
    what values are passed in, the answer will be an integer and <b>not</b> a fractional value.</p>
    <p>The <code>Equation.solveFor()</code> and the <code>MEquation.solveEquation()</code> methods allow the generation of NON-nice equations. The heart of these 
    methods is an int[] array consisting of the <b>placeholders</b> in order: [0] = 'A', [1] = 'B' etc. These values are substituted for the placeholders and the 
    equation solved. There is NO GUARANTEE that the result will be nice, though! The <code>Equation.solveFor()</code> method presents the answer as a String 
    consisting of one or two answers - one if no fractions are involved, a typewriter-style fraction followed by the double value otherwise - followed by the 
    literal String '(REV)' if an inequality needs reversing.</p>
    <p>The <code>MEquation.solveEquation()</code> methods (and their instance shadows) provide an easy way to access the raw 'solve-string' or to generate a 
    <code>Question</code>. In the latter case the variable, relation and format must be specified, BIRD is applied if necessary and a Question is returned.</p>

    <h3>The MEquation Container</h3>
    <p>Though primarily designed only to support the Equation enum, instances of MEquation can act as containers for Equation data. The <code>createMEquation()
    </code> method does this, and it or its instance version are the only ways such a container should be created. The primary rationale for this is to allow 
    XML or some other form of Object persistence. Once created, a MEquation is immutable.</p>
    <p>NOTE: <code>MEquation</code> <i>instances</i> can only be created for nice equations, not nasty ones.</p>
*/
public class MEquation {

  /** <p>This enum specifies the standard types of equations contained within the MEquations class. Each equation is documented in 'standard form' but contains 
      additional formats for MathML, OpenOffice and LaTeX.  This format can be accessed via the 'getFormat()' method using one of the specified keys.</p>
      <p>The calculations are listed for each Equation. Elements starting with 'K' represent Knobs: 'K0' is Knob[0], 'K1' is Knob[1] etc.</p>
      <p>The actual formats for each equation are standard <code>Question.fillTemplate()</code> Strings. The first element in the replacement token array ('[0]') 
      holds the variable and the rest hold the placeholders in order: [1] is A, [2] is B etc. The relation stored in each format is equals ('=') and it is the 
      <b>only</b> occurrence of this character in the raw templates.</p>
  */
  public static enum Equation {

    /** TYPE_ constant used as an error marker, '-1' Knobs. It is also used to debug several internal methods. */
    TYPE_ERROR (-1, "Error", "Error", "<mtext>Error</mtext>", "Error") {
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, 4, 0);
        CapCom.debug("Excluding "+Knob[0]+" from ["+Knob[1]+","+Knob[2]+"]: "+_exclude(Knob[0], Knob[1], Knob[2]));
        CapCom.debug("Decorating "+Knob[0]+": "+_decorate(Knob[0]));
        CapCom.debug("Raw preformat: "+_preformat(Knob[3], false, Knob[0], Knob[1], Knob[2]));
        return "ERROR:0 0";
      }
      public boolean bird() { return false; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 5, 0);
        return "Error!";
      }
    },
    /** Equation type: X + A = B, 2 Knobs. X = K0, A = K1, B = X + A. */
    TYPE_100 (2, " <0> + <1> = <2> ", " <0> + <1> = <2> ","<mi><0></mi> <mo>+</mo> <mn><1></mn> <mo>=</mo> <mn><2></mn>", "<0>+<1>=<2>") {
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x = Knob[0], A = Knob[1], B = x + A;
        return _preformat(x, false, A, B);
      }
      public boolean bird() { return false; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 2, 0);
        return ""+(PH[1] - PH[0]);
      }
    },
    /** Equation type: X - A = B, 2 Knobs. X = K0, A = K1, B = X - A. */
    TYPE_150 (2, " <0> - <1> = <2> ", " <0> - <1> = <2> ", "<mi><0></mi> <mo>-</mo> <mn><1></mn> <mo>=</mo> <mn><2></mn>", "<0>-<1>=<2>") {
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x = Knob[0], A = Knob[1], B = x - A;
        return _preformat(x, false, A, B);
      }
      public boolean bird() { return false; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 2, 0);
        return ""+(PH[0] + PH[1]);
      }
    },
    /** Equation type: Ax = B, 2 Knobs. X = K0, A = K1, B = A * X. */
    TYPE_200 (2, " <1><0> = <2> ", " <1><0> = <2> ", "<mn><1></mn><mi><0></mi> <mo>=</mo> <mn><2></mn>","<1><0>=<2>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x = Knob[0], A = Knob[1], B = x * A;
        revAns = (A < 0);
        return _preformat(x, revAns, A, B);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 2, 0);
        return ""+PH[1]+"/"+PH[0]+" = "+((double) PH[1]/PH[0])+((PH[0]<0)?" (REV)":"");
      }
    },
    /** Equation type: X/A = B, 2 Knobs. A = K0 (forced nonzero), B = K1, x = A * B. */
    TYPE_300 (2, " <0>/<1> = <2> ", " <0> over <1> = <2> ", "<mfrac><mi><0></mi><mn><1></mn></mfrac> <mo>=</mo> <mn><2><mn>", "\\frac{<0>}{<1>} = <2>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int A = Knob[0], B = Knob[1];
        A = _exclude(A, -1, 1);
        revAns = (A < 0);
        return _preformat(A*B, revAns, A, B);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 2, 0);
        return ""+(PH[1]*PH[0])+((PH[0] < 0)?" (REV)":"");
      }
    },
    /** Equation type: X/A + B = C, 3 Knobs. A = K0 (not [-1,1]), B = K1, C = K2, X = A(C - B). */
    TYPE_325 (3, " <0>/<1> + <2> = <3> ", " <0> over <1> + <2> = <3> ", "<mfrac><mi><0></mi><mn><1></mn></mfrac> <mo>+</mo> <mn><2><mn> <mo>=</mo> <mn><3></mn",
      "\\frac{<0>}{<1>} + <2> = <3>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int A = _exclude(Knob[0], -1, 1), B = Knob[1], C = Knob[2];
        revAns = (A < 0);
        return _preformat(A*(C - B), revAns, A, B, C);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 3, 0);
        return ""+(PH[0]*(PH[2] - PH[1]))+((PH[0] < 0)?" (REV)":"");
      }
    },
    /** Equation type: X/A - B = C, 3 Knobs. A = K0 (not [-1,1]), B = K1, C = K2, X = A(C + B). */
    TYPE_350 (3, " <0>/<1> - <2> = <3> ", " <0> over <1> - <2> = <3> ", "<mfrac><mi><0></mi><mn><1></mn></mfrac> <mo>-</mo> <mn><2><mn> <mo>=</mo> <mn><3></mn",
      "\\frac{<0>}{<1>} - <2> = <3>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int A = _exclude(Knob[0], -1, 1), B = Knob[1], C = Knob[2];
        revAns = (A < 0);
        return _preformat(A*(C + B), revAns, A, B, C);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 3, 0);
        return ""+(PH[0]*(PH[2] + PH[1]))+((PH[0] < 0)?" (REV)":"");
      }
    },
    /** Equation type: Ax + B = C, 3 Knobs. x = K0, A = K1, B = K2, C = A*x + B. */
    TYPE_400 (3, " <1><0> + <2> = <3> ", " <1><0> + <2> = <3> ", "<mn><1></mn><mi><0></mi> <mo>+</mo> <mn><2></mn> <mo>=</mo> <mn><3></mn>", "<1><0>+<2>=<3>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x = Knob[0], A = Knob[1], B = Knob[2];
        revAns = (A < 0);
        return _preformat(x, revAns, A, B, A * x + B);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 3, 0);
        return ""+(PH[2] - PH[1])+"/"+PH[0]+" = "+((double) (PH[2]-PH[1])/PH[0])+((PH[0] < 0)?" (REV)":"");
      }
    },
    /** Equation type: Ax - B = C, 3 Knobs. x = K0, A = K1, B = K2, C = A*x - B. */
    TYPE_450 (3, " <1><0> - <2> = <3> ", " <1><0> + <2> = <3> ", "<mn><1></mn><mi><0></mi> <mo>-</mo> <mn><2></mn> <mo>=</mo> <mn><3></mn>", "<1><0>-<2>=<3>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x = Knob[0], A = Knob[1], B = Knob[2];
        revAns = (A < 0);
        return _preformat(x, revAns, A, B, A * x - B);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 3, 0);
        return ""+(PH[2] + PH[1])+"/"+PH[0]+" = "+((double) (PH[2]+PH[1])/PH[0])+((PH[0] < 0)?" (REV)":"");
      }
    },
    /** Equation type: A(x + B) = C, 3 Knobs. x = K0, A = K1, B = K2, C = A*(x + B). */
    TYPE_500 (3, " <1>(<0> + <2>) = <3> ", " <1>(<0> + <2>) = <3> ", "<mn><1></mn><mo>(</mo><mi><0></mi> <mo>+</mo> <mn><2></mn><mo>)</mo> <mo>=</mo> <mn><3></mn>",
        "<1>(<0>+<2>)=<3>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x = Knob[0], A = Knob[1], B = Knob[2];
        revAns = (A < 0);
        return _preformat(x, revAns, A, B, A * (x + B));
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 3, 0);
        return ""+(PH[2] - PH[0]*PH[1])+"/"+PH[0]+" = "+((double) (PH[2] - PH[0]*PH[1])/PH[0])+((PH[0] < 0)?" (REV)":"");
      }
    },
    /** Equation type: A(x + B) + C = D, 4 Knobs. x = K0, A = K1, B = K2, C = K3, D = A(x + B) + C. */
    TYPE_515 (4, " <1>(<0> + <2>) + <3> = <4> ", " <1>(<0> + <2>) + <3> = <4> ", "<mn><1></mn><mo>(</mo><mi><0></mi> <mo>+</mo> <mn><2></mn><mo>)</mo> <mo>+</mo> <mn><3></mn> <mo>=</mo><mn><4></mn>",
        "<1>(<0>+<2>)+<3>=<4>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x = Knob[0], A = Knob[1], B = Knob[2], C = Knob[3];
        revAns = (A < 0);
        return _preformat(x, revAns, A, B, C, A*(x + B) + C);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 4, 0);
        int numer = PH[3] - PH[2] - PH[0]*PH[1];
        return ""+(numer)+"/"+PH[0]+" = "+((double) numer / PH[0])+((PH[0] < 0)?" (REV)":"");
      }
    },
    /** Equation type: A(x + B) - C = D, 4 Knobs. x = K0, A = K1, B = K2, C = K3, D = A(x + B) - C. */
    TYPE_530 (4, " <1>(<0> + <2>) - <3> = <4> ", " <1>(<0> + <2>) - <3> = <4> ", "<mn><1></mn><mo>(</mo><mi><0></mi> <mo>+</mo> <mn><2></mn><mo>)</mo> <mo>-</mo> <mn><3></mn> <mo>=</mo><mn><4></mn>",
        "<1>(<0>+<2>)-<3>=<4>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x = Knob[0], A = Knob[1], B = Knob[2], C = Knob[3];
        revAns = (A < 0);
        return _preformat(x, revAns, A, B, C, A*(x + B) - C);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 4, 0);
        int numer = PH[3] + PH[2] - PH[0]*PH[1];
        return ""+(numer)+"/"+PH[0]+" = "+((double) numer / PH[0])+((PH[0] < 0)?" (REV)":"");
      }
    },
    /** Equation type: A(x - B) = C, 3 Knobs. x = K0, A = K1, B = K2, C = A*(x - B). */
    TYPE_550 (3, " <1>(<0> - <2>) = <3> ", " <1>(<0> - <2>) = <3> ", "<mn><1></mn><mo>(</mo><mi><0></mi> <mo>-</mo> <mn><2></mn><mo>)</mo> <mo>=</mo> <mn><3></mn>",
        "<1>(<0>-<2>)=<3>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x = Knob[0], A = Knob[1], B = Knob[2];
        revAns = (A < 0);
        return _preformat(x, revAns, A, B, A * (x - B));
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 3, 0);
        return ""+(PH[2] + PH[0]*PH[1])+"/"+PH[0]+" = "+((double) (PH[2] + PH[0]*PH[1])/PH[0])+((PH[0] < 0)?" (REV)":"");
      }
    },
    /** Equation type: A(x - B) + C = D, 4 Knobs. x = K0, A = K1, B = K2, C = K3, D = A(x - B) + C. */
    TYPE_565 (4, " <1>(<0> - <2>) + <3> = <4> ", " <1>(<0> - <2>) + <3> = <4> ", "<mn><1></mn><mo>(</mo><mi><0></mi> <mo>-</mo> <mn><2></mn><mo>)</mo> <mo>+</mo> <mn><3></mn> <mo>=</mo><mn><4></mn>",
        "<1>(<0>-<2>)+<3>=<4>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x = Knob[0], A = Knob[1], B = Knob[2], C = Knob[3];
        revAns = (A < 0);
        return _preformat(x, revAns, A, B, C, A*(x - B) + C);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 4, 0);
        int numer = PH[3] - PH[2] + PH[0]*PH[1];
        return ""+(numer)+"/"+PH[0]+" = "+((double) numer / PH[0])+((PH[0] < 0)?" (REV)":"");
      }
    },
    /** Equation type: A(x - B) - C = D, 4 Knobs. x = K0, A = K1, B = K2, C = K3, D = A(x - B) - C. */
    TYPE_580 (4, " <1>(<0> - <2>) - <3> = <4> ", " <1>(<0> - <2>) - <3> = <4> ", "<mn><1></mn><mo>(</mo><mi><0></mi> <mo>-</mo> <mn><2></mn><mo>)</mo> <mo>-</mo> <mn><3></mn> <mo>=</mo><mn><4></mn>",
        "<1>(<0>-<2>)-<3>=<4>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x = Knob[0], A = Knob[1], B = Knob[2], C = Knob[3];
        revAns = (A < 0);
        return _preformat(x, revAns, A, B, C, A*(x - B) - C);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 4, 0);
        int numer = PH[3] + PH[2] + PH[0]*PH[1];
        return ""+(numer)+"/"+PH[0]+" = "+((double) numer / PH[0])+((PH[0] < 0)?" (REV)":"");
      }
    },
    /** Equation type: (Ax + B)/C = D, 3 Knobs. A = K0 (not in [-1,1]), x = K1, C = K2 (not in [-1,1]), D = (Ax + B)/C, B in {1,2...} such that C divides Ax + B */
    TYPE_600 (3, " (<1><0> + <2>)/<3> = <4> ", " { <1><0> + <2> } over <3> = <4> ", "<mfrac><mrow><mn><1></mn><mi><0></mi> <mo>+</mo> <mn><2></mn></mrow><mn><3></mn></mfrac> <mo>=</mo> <mn><4></mn>",
        "\\frac{<1><0>+<2>}{<3>}=<4>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int A = _exclude(Knob[0], -1, 1),B = 1, C = _exclude(Knob[2], -1, 1), x = Knob[1], D = 0, iTemp = A * x;
        while(((A*x + B) % C) != 0) B++;
        if((A*x+B) == 0) B += Math.abs(C);
        D = (A*x + B)/C;
        revAns = ((A*C) < 0);
        return _preformat(x, revAns, A, B, C, D);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 4, 0);
        return ""+(PH[3]*PH[2] - PH[1])+"/"+PH[0]+" = "+((double) (PH[3]*PH[2] - PH[1])/PH[0])+(((PH[0]*PH[2])<0)?" (REV)":"");
      }
    },
    /** Equation type: (Ax - B)/C = D, 3 Knobs. A = K0 (not in [-1,1]), x = K1, C = K2 (not in [-1,1]), D = (Ax + B)/C, B in {1,2...} such that C divides Ax - B */
    TYPE_650 (3, " (<1><0> - <2>)/<3> = <4> ", " { <1><0> - <2> } over <3> = <4> ", "<mfrac><mrow><mn><1></mn><mi><0></mi> <mo>-</mo> <mn><2></mn></mrow><mn><3></mn></mfrac> <mo>=</mo> <mn><4></mn>",
      "\\frac{<1><0>-<2>}{<3>}=<4>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int A = _exclude(Knob[0], -1, 1),B = 1, C = _exclude(Knob[2], -1, 1), x = Knob[1], D = 0, iTemp = A * x;
        while(((A*x - B) % C) != 0) B++;
        if((A*x-B) == 0) B += Math.abs(C);
        D = (A*x - B)/C;
        revAns = ((A*C) < 0);
        return _preformat(x, revAns, A, B, C, D);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 4, 0);
        return ""+(PH[3]*PH[2] + PH[1])+"/"+PH[0]+" = "+((double) (PH[3]*PH[2] + PH[1])/PH[0])+(((PH[0]*PH[2])<0)?" (REV)":"");
      }
    },
    /** Equation type: (Ax + B)/C + D = E, 4 Knobs. x = K0, A = K1 (not [-1,1]), C = K2 (not [-1,1]), D = K3, E = (Ax + B)/C + D. B calculated such that C|(Ax + B).*/
    TYPE_700 (4, "(<1><0> + <2>)/<3> + <4> = <5>", "{<1><0> + <2>} over {<3>} + <4> = <5>", "<mfrac><mrow><mn><1></mn><mi><0></mi><mo>+</mo><mn><2></mn></mrow><mn><3></mn></mfrac><mo>+</mo><mn><4></mn><mo>=</mo><mn><5></mn>",
      "\\frac{<1><0>+<2>}{<3>}+<4>=<5>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x=Knob[0], A=_exclude(Knob[1], -1, 1), B=1, C=_exclude(Knob[2], -1, 1), D=Knob[3], E=0;
        while(((A*x + B) % C) != 0) B++;
        E = (A*x + B)/C + D;
        revAns = ((A*C) < 0);
        return _preformat(x, revAns, A, B, C, D, E);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 5, 0);
        int theNum = PH[2]*(PH[4] - PH[3]) - PH[1];
        return ""+theNum+"/"+PH[0]+" = "+((double) theNum/PH[0])+(((PH[0] * PH[2]) < 0)?" (REV)":"");
      }
    },
    /** Equation type: (Ax + B)/C - D = E, 4 Knobs. x = K0, A = K1 (not [-1,1]), C = K2 (not [-1,1]), D = K3, E = (Ax + B)/C - D. B calculated such that C|(Ax + B).*/
    TYPE_725 (4, "(<1><0> + <2>)/<3> - <4> = <5>", "{<1><0> + <2>} over {<3>} - <4> = <5>", "<mfrac><mrow><mn><1></mn><mi><0></mi><mo>+</mo><mn><2></mn></mrow><mn><3></mn></mfrac><mo>-</mo><mn><4></mn><mo>=</mo><mn><5></mn>",
      "\\frac{<1><0>-<2>}{<3>}-<4>=<5>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x=Knob[0], A=_exclude(Knob[1], -1, 1), B=1, C=_exclude(Knob[2], -1, 1), D=Knob[3], E=0;
        while(((A*x + B) % C) != 0) B++;
        E = (A*x + B)/C - D;
        revAns = ((A*C) < 0);
        return _preformat(x, revAns, A, B, C, D, E);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 5, 0);
        int theNum = PH[2]*(PH[4] + PH[3]) - PH[1];
        return ""+theNum+"/"+PH[0]+" = "+((double) theNum/PH[0])+(((PH[0] * PH[2]) < 0)?" (REV)":"");
      }
    },
    /** Equation type: (Ax - B)/C + D = E, 4 Knobs. x = K0, A = K1 (not [-1,1]), C = K2 (not [-1,1]), D = K3, E = (Ax + B)/C + D. B calculated such that C|(Ax + B).*/
    TYPE_750 (4, "(<1><0> - <2>)/<3> + <4> = <5>", "{<1><0> - <2>} over {<3>} + <4> = <5>", "<mfrac><mrow><mn><1></mn><mi><0></mi><mo>-</mo><mn><2></mn></mrow><mn><3></mn></mfrac><mo>+</mo><mn><4></mn><mo>=</mo><mn><5></mn>",
      "\\frac{<1><0>-<2>}{<3>}+<4>=<5>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x=Knob[0], A=_exclude(Knob[1], -1, 1), B=1, C=_exclude(Knob[2], -1, 1), D=Knob[3], E=0;
        while(((A*x - B) % C) != 0) B++;
        E = (A*x - B)/C + D;
        revAns = ((A*C) < 0);
        return _preformat(x, revAns, A, B, C, D, E);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 5, 0);
        int theNum = PH[2]*(PH[4] - PH[3]) + PH[1];
        return ""+theNum+"/"+PH[0]+" = "+((double) theNum/PH[0])+(((PH[0] * PH[2]) < 0)?" (REV)":"");
      }
    },
    /** Equation type: (Ax - B)/C - D = E, 4 Knobs. x = K0, A = K1 (not [-1,1]), C = K2 (not [-1,1]), D = K3, E = (Ax + B)/C + D. B calculated such that C|(Ax + B).*/
    TYPE_775 (4, "(<1><0> - <2>)/<3> - <4> = <5>", "{<1><0> - <2>} over {<3>} - <4> = <5>", "<mfrac><mrow><mn><1></mn><mi><0></mi><mo>-</mo><mn><2></mn></mrow><mn><3></mn></mfrac><mo>-</mo><mn><4></mn><mo>=</mo><mn><5></mn>",
      "\\frac{<1><0>-<2>}{<3>}-<4>=<5>") {
      private boolean revAns = false;
      public String preformat(int[] Knob) {
        Knob = Arithmetic.pad(Knob, numKnobs(), 0);
        int x=Knob[0], A=_exclude(Knob[1], -1, 1), B=1, C=_exclude(Knob[2], -1, 1), D=Knob[3], E=0;
        while(((A*x - B) % C) != 0) B++;
        E = (A*x - B)/C - D;
        revAns = ((A*C) < 0);
        return _preformat(x, revAns, A, B, C, D, E);
      }
      public boolean bird() { return revAns; }
      public String solveFor(int... PH) {
        PH = Arithmetic.pad(PH, 5, 0);
        int theNum = PH[2]*(PH[4] + PH[3]) + PH[1];
        return ""+theNum+"/"+PH[0]+" = "+((double) theNum/PH[0])+(((PH[0] * PH[2]) < 0)?" (REV)":"");
      }
    };

    /** Constant used for Standard format. */
    public static final int STANDARD = 0;

    /** Constant used for OpenOffice format. */
    public static final int OPEN_OFFICE = 1;

    /** Constant used for MathML format. The elements are not enclosed in &lt;mrow&gt; or &lt;math&gt; tags. */
    public static final int MATHML = 2;

    /** Constant used for RAW LaTeX format. This is just the equation and no enclosing formatting. */
    public static final int LATEX = 3;

    /* Exclude a value from an interval. */
    private static final int _exclude(int pVal, int pLow, int pHigh) {
      if(pHigh < pLow) {
        int iTemp = pHigh;
        pHigh = pLow;
        pLow = iTemp;
      }
      while((pLow <= pVal) && (pVal <= pHigh)) { pVal++; }
      return pVal;
    }

    /* Enclose a negative value in parentheses, if they are defined. */
    private static final String _decorate(int pVal) {
      if(pVal >=0) return ""+pVal;
      return ((Question.validString(myLParen)?myLParen:"") + pVal + (Question.validString(myRParen)?myRParen:""));
    }

    /* Create the preformat String. */
    private static final String _preformat(int pAns, boolean pBIRD, int... pCoefs) {
      if((pCoefs == null) || (pCoefs.length < 1)) return "0:0";
      StringBuffer retval = new StringBuffer();
      retval.append(pAns).append(":");
      for(int i=0; i < pCoefs.length; i++) retval.append(_decorate(pCoefs[i])).append(" ");
      retval.setCharAt(retval.length() - 1, ':');
      retval.append(pBIRD);
      return retval.toString().trim();
    }

    private String[] myFormats = { "x = 1" };
    private boolean reverseAnswer = false;
    private int myNumberOfKnobs = -1;
    Equation(int pKnobs, String... pFormats) {
      myFormats = pFormats;
      reverseAnswer = false;
      myNumberOfKnobs = pKnobs;
    }

    /** Query a particular TYPE_ as to the number of Knobs it requires. */
    public int numKnobs() { return myNumberOfKnobs; }

    /** This method is used generically to fetch any intrinsic format. If no key matches, STANDARD is returned. Although this class and the enum types imply 
        equations, inequalities can also be returned. The relation parameter determines what relation will actually be returned. If null or empty the default 
        is equality.
        @param pKey One of the format constants listed above.
        @param pRelation The relation (equality or inequality) type to return. Defaults to '='.
    */
    public String getFormat(int pKey, String pRelation) {
      pRelation = Question.validString(pRelation)?pRelation:"=";
      if((pKey < 0) || (pKey >= myFormats.length)) pKey = 0;
      if((pKey == MATHML) && myFormatEscape) pRelation = MathXML.escapeXML(pRelation);
      return myFormats[pKey].replaceAll("=", pRelation);
    }

    /** Preformat the data from the equation type. This method takes an int[] array and feeds it directly into the computations. <b>IMPORTANT NOTE:</b> If the 
        array supplied is not long enough it is padded to the appropriate length with zero values.
        @param pVals An array containing raw data for the computations.
    */
    public abstract String preformat(int[] pVals);

    /** <p>Preformat the data from the equation type. The return value is in the form 'Answer:A B C...:<i>BIRD</i>' up to the number of placeholders in the 
        equation type. <i>BIRD</i> will be 'true' or 'false' and indicates whether or not an inequality should be reversed.</p>
        @param pKnobs A String containing the requisite number of Knobs.
        @return A pre-formatted String containing the answer to the equation and the placeholder values, in order.
    */
    public String preformat(String pKnobs) {
      if(!Question.validString(pKnobs)) pKnobs = "=0 0";
      int[] myKnobs = MathEngine.setKnobs(pKnobs);
      return preformat(myKnobs);
    }

    /** <p>This method returns the <code>preformat</code> data as a String array. Element [0] is the answer and the placeholders increase in order from 
        Element [1]. Inequality reverse detection <b>IS</b> set with this method and the last element in the array (<b>after</b> all placeholders!) is either 
        'true' or 'false', depending on the call to <code>bird()</code></p>
        @param pVals An array containing raw data for the equation type.
    */
    public String[] tokens(int[] pVals) {
      String[] retval = {"0"};
      try {
        String[] anser = preformat(pVals).split(":");
        String[] vars = anser[1].split(" ");
        retval = new String[vars.length + 2];
        retval[0] = anser[0];
        retval[retval.length - 1] = anser[2];
        for(int i=0; i < vars.length; i++) { retval[i+1] = vars[i]; }
      } catch(Exception ex) { CapCom.error("Error forming Equation tokens", ex); }
      return retval;
    }

    /** Preformat to tokens but with Knobs for data. Inequality reverse detection <b>IS</b> set with this method.
        @param pKnobs A String containing the requisite number of Knobs.
        @return The preformat data as a String[] array: {Answer, A, B, ... , <i>BIRD</i>}.
    */
    public String[] tokens(String pKnobs) {
      if(!Question.validString(pKnobs)) pKnobs = "=0 0";
      return tokens(MathEngine.setKnobs(pKnobs));
    }

    /** <p><b>B</b>lind <b>I</b>nequality <b>R</b>everse <b>D</b>etection. This method indicates whether or not an inequality should be reversed based on the 
        numbers used to build it. It is 'blind' because it makes no assumptions about whether or not it actually IS building an inequality, it always presumes so.</p>
        <p>NOTE: This method is valid immediately after a call to <code>preformat()</code> or <code>tokens()</code> <b>ONLY</b>!! Any other time its value is 
        meaningless.</p>
    */
    public abstract boolean bird();

    /** <p>This method allows an <code>Equation enum</code> member to be used to create an equation or inequality that does NOT solve nicely. The arguments are the 
        placeholders, in order. The return value is String to allow fractions to be presented (typewriter format) and to allow inequality reversal to be indicated. 
        If the values indicate that an inequality should be reversed the String '(REV)' appears at the end of the return value.</p>
        @param pPlaceHolders The placeholders required for the <code>Equation</code> desired.
    */
    public abstract String solveFor(int... pPlaceHolders);
  
  }

  /* * * * * * * * * * ENUMERATION STOPS HERE * * * * * * * * * */

  /** Copy of Equation.STANDARD. */
  public static final int STANDARD = Equation.STANDARD;

  /** Copy of Equation.OPEN_OFFICE. */
  public static final int OPEN_OFFICE = Equation.OPEN_OFFICE;

  /** Copy of Equation.MATHML. */
  public static final int MATHML = Equation.MATHML;

  /** Copy of Equation.LATEX */
  public static final int LATEX = Equation.LATEX;

  private static String myLParen = null, myRParen = null;
  private static boolean myFormatEscape = false;

  private Equation myType = Equation.TYPE_ERROR;
  private String myKnobs = "1 1 1 1 1 1 1 1 1 1";
  private String myVariable = "x";
  private String myRelation = "=";
  private boolean myReverse = false;

  /** Constructor for script access. */
  public MEquation() { super(); }

  /* Private constructor for holding values. */
  private MEquation(String pType, String pKnobs, String pVar, String pRel, String pReverse) {
    try {
      myType = Equation.valueOf(pType);
      myReverse = "true".equalsIgnoreCase(pReverse);
      myKnobs = MathEngine.isoKnobs(Question.validString(pKnobs)?pKnobs:"1 1 1 1 1 1 1 1 1 1");
      myVariable = Question.validString(pVar)?pVar:"x";
      myRelation = Question.validString(pRel)?unescapeRelation(pRel):"=";
    } catch(Exception ex) { CapCom.debug("Error creating MEquation container.", ex); }
  }

  /* Escape a relation for use in an XML attribute. */
  private static String escapeRelation(String pRel) {
    if(!Question.validString(pRel)) return "=";
    String retval = pRel.replaceAll("[<]", "L");
    retval = retval.replaceAll("[>]", "G");
    return retval;
  }
  /* Unscape a relation for use in an XML attribute. */
  private static String unescapeRelation(String pRel) {
    if(!Question.validString(pRel)) return "=";
    String retval = pRel.replaceAll("L", "<");
    retval = retval.replaceAll("G", ">");
    return retval;
  }

  /** Set or clear the parenthesis symbols used to enclose negative placeholder values. Each symbol is independent of the other so one or both may be set or 
      cleared. If set, the left value is prepended to a negative placeholder and the right value is appended. Setting null or blank clears them. These values 
      remain until re-set.
      @param pLeft The left parenthesis symbol to enclose negative numbers.
      @param pRight The right parenthesis symbol.
  */
  public static void setNegativeParentheses(String pLeft, String pRight) { myLParen = pLeft; myRParen = pRight; }

  /** Instance shadow to set parentheses. */
  public void setNewNegativeParentheses(String pLeft, String pRight) { setNegativeParentheses(pLeft, pRight); }

  /** Use this method to XML-escape the relations sent into a MathML format. Once set this value remains until un-set. The default is false. */
  public static void setEscapeFormat(boolean pEnable) { myFormatEscape = pEnable; }

  /** Instance version. */
  public void setNewEscapeFormat(boolean pEnable) { myFormatEscape = pEnable; }

  /** Check the values for a particular type of equation. The answer, enclosed within parentheses, is substituted in place of the variable in STANDARD format. 
      This allows a particular equation to be investigated and its difficulty adjusted. ANY Exceptions are caught and reported at Error level.
      @param pType The equation TYPE_ to check.
      @param pVals An int[] array containing values for the Knobs the equationr requires.
      @return A STANDARD-format statement to check the equation.
  */
  public static String checkEquation(Equation pType, int[] pVals) {
    String retval = "Error occurred!";
    try {
      String[] preForm = pType.preformat(pVals).split(":");
      String[] preTem = preForm[1].split(" ");
      String[] tokenz = new String[preTem.length + 1];
      tokenz[0] = "("+preForm[0]+")";
      for(int i=0; i < preTem.length; i++) { tokenz[i+1] = preTem[i]; }
      retval = Question.fillTemplate(pType.getFormat(Equation.STANDARD, "="), tokenz);
    } catch(Exception ex) { CapCom.error("Equation check error.", ex); }
    return retval;
  }

  /** <p>Generate a Question from one of the Equation types, a Knobs string, a relation and a format. This is the only actual production method MEquation has. The 
      newQuestion() method reflects to this one and the checkEquation() method is primarily meant for experimentation and debugging. The 'get Equation' methods 
      simply provide an error-proof way to obtain TYPE_ constants.</p>
      <p>If desired, the equation (inequality) can be reversed. This simply swaps sides around the relation, so the equation 'x + A = B' becomes 'B = x + A.' CAUTION: 
      reversing the <i>sides</i> of an inequality does NOT reverse the inequality! Send the proper one in. The Answer is reversed based on the 'pReverse' argument 
      and BIRD.</p>
      @param pEqn The TYPE_ of equation to generate.
      @param pKnobs The Knobs string for the equation.
      @param pVar The variable to use in place of 'x'. If null or empty, one will be selected.
      @param pRel The relation to use. Defaults to equality.
      @param pFormat One of the format constant keys to use. Defaults to STANDARD.
      @param pReverse true to reverse the sides of the equation, false to leave them as written.
  */
  public static Question createQuestion(Equation pEqn, String pKnobs, String pVar, String pRel, int pFormat, boolean pReverse) {
    Question retval = new Question();
    if(pEqn == null) return retval;
    boolean isMathML = (Equation.MATHML == pFormat);
    pRel = Question.validString(pRel)?pRel:"=";
    String pFmt = pEqn.getFormat(pFormat, ":");
    String[] preForm = pFmt.split(":");
    if(isMathML) {
      preForm = pFmt.split("<mo>:</mo>");
      pRel = MathXML.escapeXML(pRel);
    }
    if(pReverse) {
      pFmt = preForm[1] + pRel + preForm[0];
      if(isMathML) pFmt = preForm[1] + "<mo>" + pRel + "</mo>" + preForm[0];
    } else {
      pFmt = preForm[0] + pRel + preForm[1];
      if(isMathML) pFmt = preForm[0] + "<mo>" + pRel + "</mo>" + preForm[1];
    }
    preForm = pEqn.preformat(pKnobs).split(":");
    boolean revAns = pEqn.bird();
    String[] preTem = preForm[1].split(" ");
    String[] tokenz = new String[preTem.length + 1];
    tokenz[0] = Question.validString(pVar)?pVar:MathEngine.splick("x y z s t m n x y z a b");
    for(int i=0; i < preTem.length; i++) { tokenz[i+1] = preTem[i]; }
    boolean bReverse = (pReverse || revAns) && !(pReverse && revAns);
    CapCom.debug("Need reverse: "+bReverse);
    if(isMathML) {
      if(bReverse) {
        retval.setAnswer("<mrow><mn>"+preForm[0]+"</mn><mo>"+pRel+"</mo><mi>"+tokenz[0]+"</mi></mrow>");
      } else {
        retval.setAnswer("<mrow><mi>"+tokenz[0]+"</mi><mo>"+pRel+"</mo><mn>"+preForm[0]+"</mn></mrow>");
      }
      retval.setProblem("<mrow>"+Question.fillTemplate(pFmt, tokenz)+"</mrow>");
    } else {
      retval.setProblem(Question.fillTemplate(pFmt, tokenz));
      if(bReverse) {
        retval.setAnswer(preForm[0] + " "+pRel+" "+tokenz[0]);
      } else {
        retval.setAnswer(tokenz[0] + " "+pRel+" "+preForm[0]);
      }
    }
    return retval;
  }

  /** Instance version of createQuestion(). */
  public Question newQuestion(Equation pEqn, String pKnobs, String pVar, String pRel, int pFormat, boolean pReverse) { return createQuestion(pEqn, pKnobs, pVar, pRel, pFormat, pReverse);}

  /** This is <code>createQuestion()</code> with an int for the Equation type, otherwise it is identical. */
  public static Question createQuestion(int pEqn, String pKnobs, String pVar, String pRel, int pFormat, boolean pReverse) {
    return createQuestion(getByType(pEqn), pKnobs, pVar, pRel, pFormat, pReverse);
  }

  /** Instance shadow of int-typed <code>createQuestion()</code>. */
  public Question newQuestion(int pEqn, String pKnobs, String pVar, String pRel, int pFormat, boolean pReverse) {
    return createQuestion(getByType(pEqn), pKnobs, pVar, pRel, pFormat, pReverse);
  }

  /** Use this method when custom-formatting an <code>Equation enum</code> equation that doesn't necessarily solve nicely. The 'phVals' values are the placeholders 
      in the order they appear in the description: A, B, C.... The return value is the <code>solveFor()</code> String for the equation selected.
      @param pType The <code>Equation</code> type to solve.
      @param phVals The placeholder values.
  */
  public static String solveEquation(int pType, int... phVals) { return getByType(pType).solveFor(phVals); }

  /** Instance <code>String solveEquation()</code> shadow. */
  public String solve_Equation(int pType, int... phVals) { return getByType(pType).solveFor(phVals); }

  /** <p>This is the full-featured nasty-equation formatting method. It accepts a type, a format constant, a variable, a relation and a placeholder array. It 
      returns a <code>Question</code> consisting of the formatted equation for the Problem part and the answer for the Answer part. BIRD is applied to the 
      Answer.</p>
      <p>Default values are supplied for the variable ('x') and the relation ('=').</p>
      @param pType The TYPE_ number of the equation to format.
      @param pFormat The constant for the desired format. Use -1 for STANDARD.
      @param pVar The variable to use in the equation.
      @param pRel The relation to use in the equation.
      @param phVals The placeholder values to plug in.
  */
  public static Question solveEquation(int pType, int pFormat, String pVar, String pRel, int... phVals) {
    Question retval = new Question();
    pVar = Question.validString(pVar)?pVar:"x";
    pRel = Question.validString(pRel)?pRel:" = ";
    Equation myEqn = getByType(pType);
    String templat = myEqn.getFormat(pFormat, pRel);
    String[] tokenz = new String[phVals.length + 1];
    tokenz[0] = pVar;
    for(int i=0; i < phVals.length; i++) tokenz[i+1] = ""+phVals[i];
    retval.setProblem(Question.fillTemplate(templat, tokenz));
    tokenz = myEqn.solveFor(phVals).split("[(]RE");
    if(tokenz.length > 1) {
      retval.setAnswer(tokenz[0] + pRel + pVar);
    } else {
      retval.setAnswer(pVar + pRel + tokenz[0]);
    }
    return retval;
  }

  /** <code>Question</code>-valued instance version of <code>solveEquation()</code>. */
  public Question solve_Equation(int pType, int pFormat, String pVar, String pRel, int... phVals) { return solveEquation(pType, pFormat, pVar, pRel, phVals); }

  /** This method creates a MEquation to hold the data supplied. The Knobs are evaluated and stored 'frozen' to ensure immutability. Though this is a <b>static</b> 
      method, it returns a MEquation <b>instance</b>. The output format is NOT stored; it is specified upon formatting!
      @param pType A String corresponding to the type of equation desired.
      @param pKnobs The Knobs to use for the equation.
      @param pVar The variable to use for the equation. The default is 'x.'
      @param pRel The relation to use. Defaults to '=.'
      @param pReverse Whether or not to reverse the equation.
  */
  public static MEquation createMEquation(String pType, String pKnobs, String pVar, String pRel, String pReverse) {
    return new MEquation(pType, pKnobs, pVar, pRel, pReverse);
  }

  /** Instance version of createMEquation(). */
  public MEquation newMEquation(String pType, String pKnobs, String pVar, String pRel, String pReverse) {
    return new MEquation(pType, pKnobs, pVar, pRel, pReverse);
  }

  /** (Re-)Create a MEquation from a <code>dataString()</code>. If the string is not properly formatted a default MEquation is returned. */
  public static MEquation createMEquation(String pDataString) {
    MEquation retval = new MEquation();
    if(!Question.validString(pDataString)) return retval;
    String[] partz = pDataString.split(":");
    if(partz.length != 5) return retval;
    return new MEquation(partz[0], partz[1], partz[2], partz[3], partz[4]);
  }

  /** Instance flavor of <code>createMEquation(DataString)</code>. */
  public MEquation newMEquation(String pDataString) { return createMEquation(pDataString); }

  /** This method provides an Exception-proof way to fetch a particular TYPE_ of Equation. If anything terrible happens, TYPE_ERROR is returned. If this is the case, 
      it is reported at DEBUG level.
      @param pTypeNumber An int corresponding to the numeric type of Equation desired; so '550' would create a TYPE_550 Equation.
  */
  public static Equation getByType(int pTypeNumber) {
    Equation retval = Equation.TYPE_ERROR;
    try { retval = Equation.valueOf("TYPE_"+pTypeNumber); } catch(Exception ex) { CapCom.debug("Error getting TYPE_"+pTypeNumber); }
    return retval;
  }

  /** Instance method for getByType(). */
  public Equation getEquationByType(int pTypeNum) { return getByType(pTypeNum); }

  /** This method provides an Exception-proof way to fetch a particular TYPE_ of Equation from a String. If the String supplied cannot be resolved to a TYPE_ of 
      Equation, TYPE_ERROR is returned and the incident is reported at Debug level.
      @param pType The String label of an Equation type; 'TYPE_100' would return a TYPE_100 Equation.
  */
  public static Equation getByType(String pType) {
    Equation retval = Equation.TYPE_ERROR;
    try { retval = Equation.valueOf(pType); } catch(Exception ex) { CapCom.debug("Error getting "+pType); }
    return retval;
  }

  /** Instance version of getByType(String). */
  public Equation getEquationByType(String pType) { return getByType(pType); }

  /** Return the specified format for the given equation with the relation supplied.
      @param pType The TYPE_ number of the desired equation.
      @param pFormat The format desired. Defaults to STANDARD.
      @param pRel The relation to set. Defaults to '='.
  */
  public static String getFormat(int pType, int pFormat, String pRel) {
    return getByType(pType).getFormat(pFormat, pRel);
  }

  /** Instance <code>getFormat()</code>. */
  public String get_Format(int pType, int pFormat, String pRel) {
    return getByType(pType).getFormat(pFormat, pRel);
  }

  /** Return the internal data for this MEquation in a simple, colon-separated format. The order of the elements is per the <code>createMEquation()</code> 
      method.
  */
  public String dataString() {
    StringBuffer retval = new StringBuffer(myType.toString());
    retval.append(":");
    retval.append(MathEngine.isoKnobs(myKnobs));
    retval.append(":").append(myVariable).append(":").append(escapeRelation(myRelation)).append(":").append(myReverse);
    return retval.toString();
  }

  /** Return <code>dataString()</code> with some added text. */
  public String toString() { return "MEquation( \""+dataString()+"\");"; }

  /** Convert this MEquation to XML. The data String is basically a colon-separated collection of the values required for the <code>createMEquation()</code> 
      method. This format should <b>not</b> be hand-edited as it can break the parsing process! The Knobs represented are in <code>MathEngine.isoKnob()</code> 
      format to preserve the immutability of the data between marshalling (to XML) and unmarshalling. The relation String is escaped and restored in a very 
      simple fashion: '&lt;' is replaced with 'L' and '&gt;' is replaced with 'G'. The (output) format is <b>not</b> a part of the tag since it is a part of the 
      formatting process ONLY.
      @param pID An optional ID attribute to be added to the tag. Ignored if null or empty.
  */
  public String toXML(String pID) {
    StringBuffer retval = new StringBuffer("<MEquation ");
    if(Question.validString(pID)) retval.append("id=\""+pID+"\" ");
    retval.append("data=\"").append(dataString());
    retval.append("\" />");
    return retval.toString();
  }

  /** Output the current MEquation as a Question using the format supplied. */
  public Question toQuestion(int pFormat) { return createQuestion(myType, myKnobs, myVariable, myRelation, pFormat, myReverse); }

  /** Output the current MEquation as a Question using STANDARD format. */
  public Question toQuestion() { return createQuestion(myType, myKnobs, myVariable, myRelation, Equation.STANDARD, myReverse); }

}
