package cox.jmatt.java.MathTools;

import java.util.*;

/** <p>This class encapsulates the behavior and manipulation of fractions. It provides static and instance methods to create, add, subtract, multiply, divide, raise, 
    and reduce fractions. It also provides a rich capability to format these fractions many ways. The createMFraction() method can take a fraction in a 
    commonly-typed format, e.g. '-2 4/9', and create an MFraction from it. From there it can be easily manipulated.</p>
    <p>Once created, an MFraction is immutable. The only things that change it are methods that do not change the value of the underlying fraction. The raise() and 
    reduce() methods, for instance, multiply or divide both the numerator and denominator by the same (non-zero!) quantity. This may change the <i>appearance</i> of 
    the fraction but it does not change the fraction itself.</p>
    <p>Internally the fraction is maintained as a sign multiplier, a numerator, and a denominator. There are methods to test whether the fraction is proper or 
    improper, and methods to produce the numbers needed to format as mixed numbers or improper fractions. MFraction's versatile format() methods allow access to 
    all of these values, raw or 'cooked.'</p>
*/
public final class MFraction {

  /** Default fraction bar symbol: a single slash ('/'). */
  public static final String DEFAULT_FRACTION_BAR = "/";
  /** Standard (default) format for a fraction: &lt;0&gt/&lt;1&gt;. */
  public static final String FORMAT_STANDARD = "<0>/<1>";
  /** Standard mixed format: &lt;4&gt; &lt;5&gt/&lt;1&gt;. */
  public static final String FORMAT_STANDARD_MIXED = "<4> <5>/<1>";
  /** Typical format: mixed if improper, single number if whole or unit, proper if proper. The bar used is the one currently set. */
  public static final String FORMAT_TYPICAL = "<7><17> <14><13><16>";
  /** OpenOffice fraction format, per STANDARD with (plain) numerator and denominator. */
  public static final String FORMAT_OO_STANDARD = " {<0>} over {<1>} ";
  /** OpenOffice fraction, per STANDARD_MIXED. */
  public static final String FORMAT_OO_MIXED = " <4> {<5>} over {<1>} ";
  /** Typical format for OpenOffice. CRITICAL NOTE: For this format to work the 'bar' value <b>must</b> be set to 'over'. */
  public static final String FORMAT_OO_TYPICAL = "<7><17> {<14>} <13> {<16>}";
  /** MathML presentation standard 'mfrac' tag with just the numerator() and denominator() present. */
  public static final String FORMAT_MATHML = "<mfrac><mn> <0> </mn><mn> <1> </mn></mfrac>";
  /** MathML mixed number with explicit whole part and fractional numerator. (Uses mrow and mfrac.) */
  public static final String FORMAT_MATHML_MIXED = "<mrow> <mn><3></mn> <mfrac><mn> <5> </mn><mn> <1> </mn></mfrac> </mrow>";
  /** Typical format for MathML. No requirements for the 'bar' value. */
  public static final String FORMAT_MATHML_TYPICAL = "<mrow> <mo><7></mo><mn><17></mn> <mfrac><mn> <14> </mn><mn> <16> </mn></mfrac> </mrow>";
  /** LaTeX fraction, (signed) numerator over denominator. */
  public static final String FORMAT_LATEX = "\\frac{<0>}{<1>}";
  /** LaTeX mixed fraction. Implicit whole part, mixed numerator over denominator. */
  public static final String FORMAT_LATEX_MIXED = "<4> \\frac{<5>}{<1>}";
  /** Typical LaTeX format. No bar requirements. */
  public static final String FORMAT_LATEX_TYPICAL = "<7><17> \\frac{<14>}{<16>}";
  /** Dump ALL replacement tokens. Not intended for production use. */
  public static final String FORMAT_DEBUG = "Numerator:\t<0>\nDenominator:\t<1>\n|Numerator|:\t<2>\nWhole Part:\t<3>\nImp. Whole Part:\t<4>\nMixed Numer.:\t<5>\n"
  + "Exp. Sign:\t<6>\nImp. Sign:\t<7>\nSignum:\t\t<8>\nDec. Value:\t<9>\n|Whole Part|:\t<10>\n|Mixed. Numer.|:\t<11>\n"
  + "Bar:\t\t<12>\nImp Bar:\t<13>\nImp Mx Nm\t<14>\nAbs Imp Mx Nm:\t<15>\nImp. Den.:\t<16>\nAbs. Imp Whole:\t<17>";

  private static final String[] COMPARE_SYMBOLS = {" < ", " = ", " > "};

  private static String DEFAULT_FORMAT = FORMAT_STANDARD;
  private static String DEFAULT_BAR = DEFAULT_FRACTION_BAR;

  private int myNumerator = 0;
  private int myDenominator = 1;
  private int mySignum = 1;
  private String myFormat = DEFAULT_FORMAT;
  private String myBar = DEFAULT_BAR;

  /** Zero-argument constructor. Gives the zero fraction: 0/1. */
  public MFraction() { super(); }
  
  /** Standard constructor: numerator and denominator specified. If the denominator is zero, an ArithmeticException is thrown.
      @param pNumerator The numerator of the fraction.
      @param pDenominator The denominator of the fraction.
      @throws ArithmeticException if the denominator is zero. NOT ALLOWED!
  */
  public MFraction(int pNumerator, int pDenominator) {
    super();
    if(pDenominator == 0) throw new ArithmeticException("Fraction denominator of zero.");
    myNumerator = Math.abs(pNumerator);
    myDenominator = Math.abs(pDenominator);
    mySignum = (pNumerator == 0)?0:(isNeg(pNumerator) * isNeg(pDenominator));
  }
  
  /** Mixed-number constructor: whole part, numerator and denominator. Can throw an ArithmeticException on a zero denominator!
      @param pWhole The whole part of the mixed number.
      @param pNumerator The numerator of the fractional part.
      @param pDenominator The denominator of the fractional part.
      @throws ArithmeticException if the denominator is zero.
  */
  public MFraction(int pWhole, int pNumerator, int pDenominator) {
    if(pDenominator == 0) throw new ArithmeticException("Fraction denominator of zero.");
    myDenominator = Math.abs(pDenominator);
    myNumerator = Math.abs(pWhole) * myDenominator + Math.abs(pNumerator);
    mySignum = ((pWhole == 0) && (pNumerator == 0))?0:(isNeg(pWhole)*isNeg(pNumerator)*isNeg(pDenominator));
  }
  // Used in calculating the signum
  private static final int isNeg(int pNum) { return (pNum < 0)?-1:1; }

  /** Constructor for a String containing a single-line-formatted fraction or mixed number. See MFraction.createMFraction() for details. If the denominator is zero 
      an ArithmeticException is thrown but if the String is invalid the zero fraction is returned.
      @throws ArithmeticException if the denominator is zero.
  */
  public MFraction(String pFraction) {
    super();
    MFraction mfTemp = createMFraction(pFraction);
    if(mfTemp.denominator() == 0) throw new ArithmeticException("Fraction denominator of zero: "+pFraction);
    myNumerator = mfTemp.absNumerator();
    myDenominator = mfTemp.denominator();
    mySignum = mfTemp.signum();
  }
  /** Copy constructor.
      @throws NullPointerException if pMFraction is null.
  */
  public MFraction(MFraction pMFraction) {
    super();
    if(pMFraction == null) throw new NullPointerException("Cannot copy a null fraction.");
    myNumerator = pMFraction.absNumerator();
    myDenominator = pMFraction.denominator();
    mySignum = pMFraction.signum();
    myBar = pMFraction.getBar();
  }

  /** <p>Parse a typically-formatted String into an MFraction. A typically-formatted String is a string containing the typical typewriter or keyboard formats of 
      fractionnal notation: '4/9' or '3 1/5'. Because of the way this method parses a String, the space delimiter will work in both places. For instance, '3 1/5' 
      could also be 3 1 5'. Slashes also work in both places.</p>
      <p>If an error occurs during parsing it is reported at Error level and the zero fraciton is returned. That is, this method guarantees a valid MFraction, 
      just not necessarily the one in the String.</p>
      @param pFraction A String consisting of a single-line representation of a fraction, e.g. '3/8 or -12 5/7'.
      @return A valid MFraction. Either the one specified, or, if an error occurs, the fraction '0/1'.
  */
  public static final MFraction createMFraction(String pFraction) {
    MFraction retval = new MFraction();
    if(!Question.validString(pFraction)) return retval;
    int iWhole = 0, iNumer = 0, iDenom = 1;
    try {
      String[] sTokens = pFraction.split("[ /]");
      switch(sTokens.length) {
        case 3:
          iWhole = Integer.parseInt(sTokens[0]);
          iNumer = Integer.parseInt(sTokens[1]);
          iDenom = Integer.parseInt(sTokens[2]);
        break;
        case 2:
          iNumer = Integer.parseInt(sTokens[0]);
          iDenom = Integer.parseInt(sTokens[1]);
        break;
        case 1:
          iNumer = Integer.parseInt(sTokens[0]);
        break;
        default:
      }
      retval = new MFraction(iWhole, iNumer, iDenom);
    } catch(Exception ex) { CapCom.error("Error creating MFraction: "+pFraction); }
    return retval;
  }
  /** Instance version of createMFraction(). */
  public MFraction newMFraction(String pFraction) { return createMFraction(pFraction); }

  /** Create a MFraction given the whole part, numerator and denominator as type int. Any errors are reported at Error level and the return is '0/1'.
      @param pWhole The whole part of the fraction. May be zero.
      @param pNum The numerator of the fraction. May be zero.
      @param pDen The denominator of the fraction. MUST NOT BE ZERO!!!
  */
  public static MFraction createMFraction(int pWhole, int pNum, int pDen) {
    MFraction retval = new MFraction();
    try { retval = new MFraction(pWhole, pNum, pDen); } catch(Exception ex) { CapCom.error("Error creating MFraction.", ex); }
    return retval;
  }
  /** Instance version of <code>createMFraction(int, int, int)</code>. */
  public MFraction newMFraction(int pWhole, int pNum, int pDen) { return createMFraction(pWhole, pNum, pDen); }

  /** Create a random MFraction whose denominator is 2<sup>pDen</sup> and whose numerator is an odd number between 1 and one less than the denominator. If 'pDen' 
      is zero the zero fraction is returned. If it is '1' the fraction is '1/2'. If 'pDen' is negative the zero fraction is returned.
      @param pDen The power of two for the denominator.
  */
  public static MFraction createDyadic(int pDen) {
    if(pDen < 1) return new MFraction();
    if(pDen == 1) return new MFraction(0, 1, 2);
    pDen = (int) Math.pow(2, pDen);
    int numer = MathEngine.random(0, pDen/2)*2 + 1;
    return new MFraction(0, numer, pDen);
  }
  /** Instance version of <code>createDyadic()</code>. */
  public MFraction newDyadic(int pDen) { return createDyadic(pDen); }

  /** Create a fraction that cannot be reduced. This uses the {@link cox.jmatt.java.MathTools.Arithmetic#coprime Arithmetic.coprime()} method to generate the 
      numerator so all range and return rules apply. Note: the range/divisor parameters here are in the same order as the 'coprime()' method.
      @param pWhole The whole part of the fraction.
      @param pDenom The denominator of the fraction.
      @param pLow The low value for the numerator range, inclusive.
      @param pHigh The high value for the numerator range, exclusive.
  */
  public static MFraction createIrreducible(int pWhole, int pLow, int pHigh, int pDenom) {
  return new MFraction(pWhole, Arithmetic.coprime(pLow, pHigh, pDenom), pDenom);
  }

  /** Instance shadow for an irreducible fraction. */
  public MFraction newIrreducible(int pWhole, int pLow, int pHigh, int pDenom) {
  return new MFraction(pWhole, Arithmetic.coprime(pLow, pHigh, pDenom), pDenom);
  }

  /** Format an MFraction according to the specified format string. It is a Question.fillTemplate() String with the following tokens supplied to the data array:
      <ul>
        <li>&lt;0&gt; - The (signed if negative) numerator, per numerator().</li>
        <li>&lt;1&gt; - The denominator, which does not change by fraction type (mixed/improper).</li>
        <li>&lt;2&gt; - The absNumerator(), no sign.</li>
        <li>&lt;3&gt; - The wholePart() of the fraction. If it is not zero, the sign will be here.</li>
        <li>&lt;4&gt; - The implicit whole part: if the fraction is proper this token is blank.</li>
        <li>&lt;5&gt; - The mixedNumerator(). If the fraction is proper this is the same as &lt;0&gt;.</li>
        <li>&lt;6&gt; - The explicit sign, '+' or '-'.</li>
        <li>&lt;7&gt; - The implicit sign, '' if positive or '-' if negative.</li>
        <li>&lt;8&gt; - The signum, explicitly.</li>
        <li>&lt;9&gt; - The decimal (double) value.</li>
        <li>&lt;10&gt; - The absolute value of the whole part.</li>
        <li>&lt;11&gt; - The absolute value of the mixed numerator.</li>
        <li>&lt;12&gt; - The current bar String, explicitly.</li>
        <li>&lt;13&gt; - The implicit bar: blank if the fraction is whole or unit.</li>
        <li>&lt;14&gt; - The implicit mixed numerator: blank if the fraction is whole or unit.</li>
        <li>&lt;15&gt; - The implicit unsigned mixed numerator: blank if whole or unit, &lt;11&gt; otherwise.</li>
        <li>&lt;16&gt; - The implicit denominator: blank if the fraction is whole or unit.</li>
        <li>&lt;17&gt; - The implicit unsigned whole part: blank if the fraction is proper, the absolute value of the whole part otherwise. </li>
      </ul></p>
      <p>If something goes horribly wrong the String '0/1' is returned. </p>
  */
  public static String formatMFraction(MFraction pFraction, String pFormat) {
    if((pFraction == null) || !Question.validString(pFormat)) return "0/1";
    String[] sTokens = new String[18];
    boolean wholeUnit = pFraction.isUnit() || pFraction.isWhole();
    sTokens[0] = ""+pFraction.numerator(); sTokens[1] = ""+pFraction.denominator(); sTokens[2] = ""+pFraction.absNumerator();
    sTokens[3] = ""+pFraction.wholePart(); sTokens[4] = (pFraction.isProper()?"":""+pFraction.wholePart()); sTokens[5] = ""+pFraction.mixedNumerator();
    sTokens[6] = (pFraction.signum() < 0)?"-":"+"; sTokens[7] = (pFraction.signum() < 0)?"-":""; sTokens[8] = ""+pFraction.signum();
    sTokens[9] = "" + pFraction.toDouble(); sTokens[10] = "" + Math.abs(pFraction.wholePart()); sTokens[11] = ""+Math.abs(pFraction.mixedNumerator());
    sTokens[12] = pFraction.getBar(); sTokens[13] = (wholeUnit?"":sTokens[12]); sTokens[14] = (wholeUnit?"":sTokens[5]);
    sTokens[15] = (wholeUnit?"":sTokens[11]); sTokens[16] = (wholeUnit?"":sTokens[1]);
    sTokens[17] = (pFraction.isProper()?"":""+Math.abs(pFraction.wholePart()));
    return Question.fillTemplate(pFormat, sTokens);
  }
  /** Instance version of formatMFraction(). */
  public String format_MFraction(MFraction pFraction, String pFormat) { return formatMFraction(pFraction, pFormat); }

  /** Set the default format for all MFractions created after this is set. If pFormat is null or blank, FORMAT_STANDARD is set. */
  public static void setDefaultFormat(String pFormat) {
    if(Question.validString(pFormat)) {
      DEFAULT_FORMAT = pFormat;
    } else {
      DEFAULT_FORMAT = FORMAT_STANDARD;
    }
  }
  /** Instance alias for setDefaultFormat(). */
  public void setNewDefaultFormat(String pFormat) { setDefaultFormat(pFormat); }

  /** Set the format for this MFraction. If pFormat is null or empty it is set to the current default format. Returns a reference to this to allow method chaining. */
  public MFraction setFormat(String pFormat) {
    if(Question.validString(pFormat)) {
      myFormat = pFormat;
    } else {
      myFormat = DEFAULT_FORMAT;
    }
    return this;
  }

  /** Set the String used for the default fraction bar. If null or blank DEFAULT_FRACTION_BAR is set. */
  public static void setDefaultBar(String pBar) {
    DEFAULT_BAR = Question.validString(pBar)?pBar:DEFAULT_FRACTION_BAR;
  }

  /** Instance default bar set method. */
  public void setNewDefaultBar(String pBar) { setDefaultBar(pBar); }

  /** Set the fraction bar String for this instance. If 'pBar' is null or blank the current default is set. */
  public void setBar(String pBar) {
    myBar = Question.validString(pBar)?pBar:DEFAULT_BAR;
  }

  /** Return the current bar value, just in case it's needed. */
  public String getBar() { return myBar; }

  /** Format this MFraction per the supplied format. If it is blank or null the currently-set format is used. */
  public String format(String pFormat) { return formatMFraction(this, (Question.validString(pFormat)?pFormat:myFormat)); }

  /** Format this per its current furmat. */
  public String format() { return formatMFraction(this, myFormat); }

  /** This method takes care of the MFraction's signs. Ordinarily, if the fraction is negative the numerator should receive the sign. This method makes sure that 
      happens. Most methods that might alter things call this one automatically, but it never hurts to do it manually, too. A reference to this MFraction is 
      returned.
  */
  public MFraction normalize() {
    int newNum = Math.abs(myNumerator), newDen = Math.abs(myDenominator);
    mySignum = (newNum == 0)?0:((newNum * newDen) / (myNumerator * myDenominator));
    myNumerator = newNum; myDenominator = newDen;
    return this;
  }
  /** Return 1 if the fraction is positive, 0 if it is zero, and -1 if it is negative. */
  public int signum() { return mySignum; }

  /** Get the fraction's numerator. If the fraction is negative, this value is negative. */
  public int numerator() { return mySignum * myNumerator; }

  /** Get the fraction's denominator. */
  public int denominator() { return myDenominator; }

  /** Return true if the numerator is (strictly!) greater than the denominator. */
  public boolean isImproper() { return myNumerator > myDenominator; }

  /** Return true if the numerator is evenly divisible by the denominator. */
  public boolean isWhole() { return (myNumerator % myDenominator) == 0; }

  /** Return true if the numerator equals the denominator. */
  public boolean isUnit() { return (myNumerator == myDenominator); }

  /** Return true if the numerator is (strictly) less than the denominator. */
  public boolean isProper() { return myNumerator < myDenominator; }

  /** Return the whole part of the fraction expressed as a mixed number. If it is nonzero the whole part will have the sign. */
  public int wholePart() { return (mySignum * (myNumerator / myDenominator)); }

  /** Return the NUMERATOR of the fractional part of this fraction expressed as a mixed number. If the fraction is proper this is the same as numerator(), i.e. it 
      does have the sign. If the fraction is improper it returns the integer remainder of (unsigned) numerator / denominator.
  */
  public int mixedNumerator() {
    if(myNumerator < myDenominator) return numerator();
    return myNumerator % myDenominator;
  }

  /** Return the unsigned (absolute) value of the numerator. Regardless of the fraction's sign, this value is guaranteed non-negative. */
  public int absNumerator() { return myNumerator; }

  /** Return the double value of this MFraction. */
  public double toDouble() { return ((double) mySignum * myNumerator) / ((double) myDenominator); }

  /** Return a constructor-like representation of this MFraction. */
  public String toString() { return "MFraction(\""+numerator()+"/"+myDenominator+"\");";}

  /** Test for equality between this MFraction and another. This is the proper equality test: A/B = C/D iff AD = BC.
      @throws NullPointerException if pMFrac is null.
  */
  public boolean equals(MFraction pMFrac) {
    if(pMFrac == null) throw new NullPointerException("NULL does not equal.");
    return ( numerator() * pMFrac.denominator()) == (pMFrac.numerator() * myDenominator);
  }

  /** Compare this MFraction to another. The result is -1 if this fraction is less, 0 if they are equal, and 1 if this is greater.
      @throws NullPointerException if pMFrac is null.
  */
  public int compareTo(MFraction pMFrac) {
    if(pMFrac == null) throw new NullPointerException("NULL does not compare.");
    int nLeft = numerator() * pMFrac.denominator(), nRight = pMFrac.numerator() * myDenominator;
    if(nLeft < nRight) return -1;
    if(nLeft > nRight) return 1;
    return 0;
  }

  /** Compare per <code>compareTo()</code> but returns '&lt;,' '=' or '&gt;' instead of a number. If pRHFrac is null the return is '*ERROR*.' */
  public String sCompareTo(MFraction pRHFrac) {
    if(pRHFrac == null) return "*ERROR*";
    return COMPARE_SYMBOLS[compareTo(pRHFrac) + 1];
  }

  /** Add this MFraction to another. Neither fraction is changed, a new one is created. If pFrac is null, this fraction is returned. */
  public MFraction add(MFraction pFrac) {
    if(pFrac == null) return this;
    return new MFraction(numerator() * pFrac.denominator() + pFrac.numerator() * myDenominator, myDenominator * pFrac.denominator());
  }

  /** Subtract another fraction from this one. Neither fraction is changed, a new one is created. If pFrac is null a reference to this is returned. */
  public MFraction subtract(MFraction pFrac) {
    if(pFrac == null) return this;
    return new MFraction(numerator() * pFrac.denominator() - pFrac.numerator() * myDenominator, myDenominator * pFrac.denominator());
  }

  /** Multiply this MFraction by another. Neither is changed, a new one is created. If pFrac is null, this fraction is returned. */
  public MFraction multiply(MFraction pFrac) {
    if(pFrac == null) return this;
    return new MFraction(numerator() * pFrac.numerator(), myDenominator * pFrac.denominator());
  }

  /** Divide this fraction by another MFraction. A new MFraction is creaed. If pFrac is null a reference to this is returned. If pFrac is zero an ArithmeticException 
      is thrown.
      @throws ArithmeticException if pFrac is zero.
  */
  public MFraction divide(MFraction pFrac) {
    if(pFrac == null) return this;
    return new MFraction(numerator() * pFrac.denominator(), myDenominator * pFrac.numerator());
  }

  /** Create and return the reciprocal of this fraction if it exists. If the (current) numerator is zero an ArithmeticException is thrown.
      @throws ArithmeticException if the reciprocal would have a zero denominator.
  */
  public MFraction reciprocal() {
    if(myNumerator == 0) throw new ArithmeticException("Zero numerator, reciprocal does not exist!");
    return new MFraction(mySignum * myDenominator, myNumerator);
  }

  /** Multiply the numerator and denominator by the same constant. If pNum is zero, this is returned. If pNum is not zero, a new fraction is created. */
  public MFraction raise(int pNum) {
    if(pNum == 0) return this;
    return new MFraction( pNum * numerator(), pNum * myDenominator);
  }

  /** Reduce the fraction by pNum. If pNum is zero or does not evenly divide both the numerator and denominator, this is returned. Otherwise, a new MFraction 
      is created.
  */
  public MFraction reduce(int pNum) {
    if((pNum == 0) || !((myNumerator % pNum) ==0) || !((myDenominator % pNum) == 0)) return this;
    return new MFraction(numerator() / pNum, myDenominator / pNum);
  }

  /** Reduce completely. The numerator and denominator are divided by their GCF. */
  public MFraction reduce() { return reduce(Arithmetic.gcf(myNumerator, myDenominator)); }

  /** Return a simple XML format for this MFraction. It consists of an empty MFraction tag with the attributes 'numerator=' and 'denominator='. If pID is not null 
      or empty it becomes an 'id=' attribute.
      @param pID The id attribute on the resulting tag. If null or empty no 'id' attribute is set.
  */
  public String toXML(String pID) {
    StringBuffer retval = new StringBuffer("<MFraction ");
    if(Question.validString(pID)) retval.append("id=\""+pID+"\" ");
    retval.append("numerator=\""+myNumerator+"\" ").append("denominator=\""+myDenominator+"\" />");
    return retval.toString();
  }

}

