package cox.jmatt.java.MathTools;

import java.util.*;
import java.io.*;
import java.text.NumberFormat;
import cox.jmatt.java.MathTools.util.MMatrixBuilder;

/** <p>MMatrix represents a mathematical matrix, a rectangular array of numbers. This class provides tools for creating and manipulating matrices, for doing arithmetic 
    operations with and to them, and for presenting them nicely. The companion to this class is {@link cox.jmatt.java.MathTools.util.MMatrixBuilder MMatrixBuilder}, 
    a class designed solely to create MMatrices from various styles of input data (many involving Strings!).</p>
    <p>Once created, a MMatrix is immutable. None of its methods change it in any way, nor are any of its critical parts subject to change, inadvertent or otherwise. 
    The heart of a MMatrix is a double[][] array. This array is NEVER exposed directly! Any operation which might result in a change to the elements of the matrix 
    returns either a copy of the elements or a copy of the MMatrix with the required changes made.</p>
*/
public class MMatrix {

  private static NumberFormat DEFAULT_NUMBER_FORMAT = null;
  private static String theLatexEnv = "tabular", theLatexEndEnv = "tabular", theLatexOptions = null, defaultRelation = " = ";
  private static boolean enableColSpec = true;

  double[][] myCoefs = { {0} };
  int myRows = 1, myCols = 1;
  String[] myVars = {"x0"};
  private NumberFormat myNFormat = DEFAULT_NUMBER_FORMAT;

  /** Standard ScriptingObject constructor. */
  public MMatrix() { super(); }

  /** Method to create a MMatrix from a double[][] array. This is the only true constructor for this class since the MMatrixBuilder class has plenty of parsing 
      and construction methods.
  */
  public MMatrix(double[][] pAry) {
    super();
    myCoefs = MMatrixBuilder.smoothCopy(pAry);
    myRows = myCoefs.length;
    myCols = myCoefs[0].length;
  }

  /** Copy constructor. It does NOT copy the variables set for the MMatrix being copied. */
  public MMatrix(MMatrix pMatrix) {
    super();
    if(pMatrix != null) {
      myCoefs = pMatrix.elements();
      myRows = pMatrix.rows();
      myCols = pMatrix.columns();
    }
  }

  /* Private constructor! Used to make alterations w/o multiple calls to smoothCopy() */
  private MMatrix(double[][] pAry, int pRow, int pCol, double pValue) {
    super();
    try { pAry[pRow][pCol] = pValue; } catch(Exception ignored) {}
    myCoefs = pAry;
    myRows = myCoefs.length;
    myCols = myCoefs[0].length;
  }

  /** Create a MMatrix from a formatted String. The row separator is a semicolon and the element separator is a comma. This is also the default toString() format. */
  public static MMatrix createMMatrix(String pFormat) { return MMatrixBuilder.createMMatrix(pFormat, ";", ","); }

  /** Instance version of createMMatrix(). */
  public MMatrix newMMatrix(String pFormat) {  return MMatrixBuilder.createMMatrix(pFormat, ";", ","); }

  /** This is the zero matrix; the default return value for many error conditions. It consists of the double[][] '{{0}}'. */
  public static final double[][] ZERO() { double[][] retval = new double[1][1]; retval[0][0] = 0.0; return retval; }

  /** Instance version of ZERO(), should it be required. */
  public double[][] getZero() { return ZERO(); }

  /** Return the number of rows in this MMatrix. */
  public int rows() { return myRows; }

  /** Return the number of columns in this MMatrix. */
  public int columns() { return myCols; }

  /** Get the elements in this MMatrix. A <b>copy</b> of the actual elements array is returned! */
  public double[][] elements() { return MMatrixBuilder.smoothCopy(myCoefs); }

  /** Return the order of the MMatrix. This will be a String of the form 'RxC' where R is the rows and C is the columns. */
  public String order() { return (""+rows()+"x"+columns()); }

  /** ADD this matrix to another. If the orders are different the zero MMatrix is returned; no exception is thrown. If the other MMatrix is null, a copy of this 
      is returned.
  */
  public MMatrix add(MMatrix pSummand) {
    if(pSummand == null) return new MMatrix(elements(), -1, -1, 0);
    if((myRows != pSummand.rows()) || (myCols != pSummand.columns())) return new MMatrix(ZERO(), -1, -1, 0);
    double[][] daRHS = pSummand.elements(), daResult = new double[myRows][myCols];
    for(int i=0; i < myRows; i++) {
      for(int j=0; j < myCols; j++) { daResult[i][j] = myCoefs[i][j] + daRHS[i][j]; }
    }
    return new MMatrix(daResult, -1, -1, 0);
  }

  /** SUBTRACT another matrix from this one. If the orders are different the zero MMatrix is returned; no exception is thrown. If the other MMatrix is null, a 
      copy of this one is returned.
  */
  public MMatrix subtract(MMatrix pSubtrahend) {
    if(pSubtrahend == null) return new MMatrix(elements(), -1, -1, 0);
    if((myRows != pSubtrahend.rows()) || (myCols != pSubtrahend.columns())) return new MMatrix(ZERO(), -1, -1, 0);
    double[][] daRHS = pSubtrahend.elements(), daResult = new double[myRows][myCols];
    for(int i=0; i < myRows; i++) {
      for(int j=0; j < myCols; j++) { daResult[i][j] = myCoefs[i][j] - daRHS[i][j]; }
    }
    return new MMatrix(daResult, -1, -1, 0);
  }

  /** MULTIPLY this matrix by another one. If the these columns are different from pFactor's rows, the zero MMatrix is returned; no exception is thrown. If the 
      other MMatrix is null, a copy of this one is returned.
  */
  public MMatrix multiply(MMatrix pFactor) {
    if(pFactor == null) return new MMatrix(elements(), -1, -1, 0);
    if(myCols != pFactor.rows()) return new MMatrix(ZERO(), -1, -1, 0);
    int nCols = pFactor.columns();
    double[][] daRHS = pFactor.elements(), daResult = MMatrixBuilder.createArray(myRows, nCols, 0.0);
    for(int i=0; i < myRows; i++) {
      for(int j=0; j < nCols; j++) {
        for(int k=0; k < myCols; k++) { daResult[i][j] += myCoefs[i][k] * daRHS[k][j]; }
      }
    }
    return new MMatrix(daResult, -1, -1, 0);
  }

  /** Multiply the elements in this MMatrix by a scalar constant. The original matrix is not changed; a copy with the products is returned.
      @param pScalar The scalar value to multiply by each element in the matrix.
      @return A copy of this MMatrix with the multiplications done.
  */
  public MMatrix multiply(double pScalar) {
    double[][] daProduct = new double[myRows][myCols];
    for(int i=0; i < myRows; i++) {
      for(int j=0; j < myCols; j++) { daProduct[i][j] = pScalar * myCoefs[i][j]; }
    }
    return new MMatrix(daProduct, -1, -1, 0.0);
  }

  /** Calculate and return the determinant of a SQUARE matrix. This is done recursively by minors with no shortcutting. WARNING: This method is <b>brittle</b>!! 
      If the matrix is not square, if any row or column is not defined or any other error occurs the return value is <code>Double.NaN</code>.
  */
  public double determinant() { return determinant(myCoefs); }

  private double determinant(double[][] pAry) {
    double retval = 0;
    try {
      if(pAry.length == 1) return pAry[0][0];
      if(pAry.length == 2) return (pAry[0][0]*pAry[1][1] - pAry[0][1]*pAry[1][0]);
      for(int i=0; i < pAry.length; i++) {
        retval += Math.pow(-1,i) * pAry[i][0] * determinant(_minor(pAry, i, 0));
      }
    } catch(Exception ex) {
      CapCom.debug("Determinant Error", ex);
      retval = Double.NaN;
    }
    return retval;
  }

  private double[][] _minor(double[][] pAry, int xRow, int xCol) {
    double[][] retval = null;
    int nSize = pAry.length - 1, iRow = 0, jCol = 0;
    retval = new double[nSize][];
    for(int i=0; i < pAry.length; i++) {
      if(i == xRow) continue;
      retval[iRow] = new double[nSize];
      jCol = 0;
      for(int j=0; j < pAry.length; j++) {
        if(j == xCol) continue;
        retval[iRow][jCol] = pAry[i][j];
        jCol++;
      }
      iRow++;
    }
    return retval;
  }

  /** Form and return the transpose of this MMatrix. Columns become rows and rows become columns. */
  public MMatrix transpose() { return new MMatrix(MMatrixBuilder.transposeArray(elements())); }

  /** This method prints the MMatrix elements in one long string. The row- and column- separators can be specified. The default values are a semicolon for the 
      row separator and a comma for the column separator.
  */
  public String toString(String pRowSep, String pColSep) {
    StringBuffer retval = new StringBuffer();
    String sRowSep = (pRowSep != null)?pRowSep:";";
    String sColSep = (pColSep != null)?pColSep:",";
    for(int i=0; i < myCoefs.length; i++) {
      for(int j=0; j < myCoefs[i].length; j++) {
        retval.append(myCoefs[i][j]);
        if(j < myCoefs[i].length - 1) retval.append(sColSep);
      }
      if(i < myCoefs.length - 1) retval.append(sRowSep);
    }
    return retval.toString();
  }

  /** Print the MMatrix in simplified XML form. This is an empty &lt;MMatrix&gt; tag with the attribute 'data='. The data string uses a semicolon for the row 
      separator and a comma for the column separator. If the ID parameter is not blank or null, it appears as an 'id=' attribute.
      @param pID String ID for the XML tag. Ignored if null or empty.
  */
  public String toXML(String pID) {
    StringBuffer retval = new StringBuffer("<MMatrix ");
    if(Question.validString(pID)) retval.append("id=\"").append(pID).append("\" ");
    retval.append("data=\"").append(toString(";",",")).append("\" />");
    return retval.toString();
  }

  /** This method calls toString(null, null) and wraps some extra text around the result. */
  public String toString() {
    StringBuffer retval = new StringBuffer("MMatrix(\"");
    retval.append(toString(null, null)).append("\")");
    return retval.toString();
  }

  /** This is a system of equations method. It sets the variables to be used when the system is printed. The variables used can be changed as they are not a part 
      of the actual matrix represented by the MMatrix. If there are fewer columns than variables any extras are ignored. If there are fewer variables than columns, 
      the variables are padded with 'x#' where '#' is the column number. A null or empty String sets each variable to its default value.
      @param pVars A String of space-separated values representing the variables for the system.
      @return A reference to this MMatrix.
  */
  public MMatrix setVariables(String pVars) {
    if(!Question.validString(pVars)) pVars = "x0";
    String[] sVars = pVars.split(" ");
    int nVars = Math.min(sVars.length, myCols);
    myVars = new String[myCols]; for(int i=0; i < myCols; i++) { myVars[i] = "x"+i; }
    for(int i=0; i < nVars; i++) { myVars[i] = sVars[i]; }
    return this;
  }

  /** Interpret the matrix as a system of equations and evaluate each row (equation) against the values given for each column (variable). The values for the variables 
      are given as a space-separated String of numbers. If there are more columns than values the array is padded with zeros. If there are more values than variables, 
      the leftovers are ignored. Any errors generated during parsing are reported at Debug level.
  */
  public double[] evaluate(String pVarVals) {
    if(!Question.validString(pVarVals)) pVarVals = "0";
    String[] sVarVals = MathEngine.spill(pVarVals, " ", myCols, "0.0");
    double[] dVarVals = new double[sVarVals.length];
    for(int i=0; i < dVarVals.length; i++) {
      try { dVarVals[i] = Double.parseDouble(sVarVals[i]); } catch(Exception ex) { CapCom.debug("Error parsing double: "+sVarVals[i], ex); }
    }
    double[] retval = new double[myRows]; Arrays.fill(retval, 0.0);
    for(int i=0; i < myRows; i++) {
      for(int j=0; j < myCols; j++) {
        retval[i] += myCoefs[i][j] * dVarVals[j];
      }
    }
    return retval;
  }

  /** View the MMatrix as a system of equations and evaluate each equation (row) against the values given. The variable values are given as type Number to allow 
      an array of Integer or Double.
  */
  public double[] evaluate(Number[] pVals) {
    double[] retval = new double[myCoefs.length], dVals = new double[myCoefs[0].length];
    Arrays.fill(retval, 0.0);
    Arrays.fill(dVals, 0.0);
    if(pVals != null) {
      int iMax = Math.min(dVals.length, pVals.length);
      for(int i=0; i < iMax; i++) {
        dVals[i] = (pVals[i] == null)?0.0:pVals[i].doubleValue();
      }
    }
    for(int iRow = 0; iRow < myCoefs.length; iRow++) {
      for(int iCol = 0; iCol < myCoefs[iRow].length; iCol++) {
        retval[iRow] += myCoefs[iRow][iCol] * dVals[iCol];
      }
    }
    return retval;
  }

  /** <p>Format this MMatrix. This is the formatting uber-method; it can format as a plain matrix, an augmented matrix or a system of equations. The parameters 
      provide full control over every aspect of the process.</p>
      <h3>Plain Matrix</h3>
      <p>To format as a plain matrix pass <code>null</code> for the constants matrix (pRHVM), the variables, the relations and the augmentation separator. This 
      leaves only the row and column separators along with the prefix and suffix, which are valid for all options. If pPrefix and pSuffix are null they are 
      ignored. If pRowSep is null its default is a semicolon. If pColSep is null it defaults to a comma. No other value-detection is performed so whitespace 
      characters can be used.</p>
      <h3>Augmented Matrix</h3>
      <p>To format as an augmented matrix pass a non-null augmentation separator and make sure pRHVM is not null or empty. If so then no constants will appear. 
      If pRHVM has more elements than the matrix has rows any leftovers are ignored; and if it has too few then rows past the end will have no constants. The 
      row separator does appear at the end of each row, after the augmentation value, and the column separator appears between the coefficient elements but 
      <b>not</b> between the last coefficient and the augmentation separator or value.</p>
      <h3>System of Equations</h3>
      <p>To format a system of equations or inequalities at least one of 'pVars' or 'pRels' must be non-null. The constant array (pRHVM) should also be defined 
      and contain sufficient elements; see 'Augmented Matrix' for details. If there are not enough relations for the number of rows the shortfall defaults to 
      equality ('=') and if there are too many the extras are ignored. Likewise with the variable names; the default values are 'x0,' 'x1' and so on.</p>
      <h3>Augmented Matrix and Systems of Equations Notes</h3>
      <p>If all argument values are specified and not null then priority goes to System of Equations. In all cases the column and row separators are used; for 
      systems of equations it is appended after the variable and before the sign of the next term. The row separator is appended to the end of every row except 
      the last row. The prefix and suffix Strings, if present, appear absolutely first and absolutely last, respectively.</p>
      <p>This method makes no assumptions about the correctness of the constants passed in; it does not evaluate the numbers, it simply formats them. This allows 
      for construction of inconsistent or dependent systems BUT, as just stated, no mathematical error-checking is performed!</p>
      <p>This method deprecates all <code>Question</code>-valued formatting methods. These methods will be removed by stable version 1.5 or 1.6.</p>
      @param pRHVM The constants matrix augmented onto the coefficient matrix.
      @param pRowSep The row separator String. Defaults to a semicolon if null.
      @param pColSep The column separator. Defaults to a comma if null.
      @param pVars The variables to use when formatting as a system.
      @param pRels The relations to use when formatting as a system.
      @param pAugSep The separator between the last coefficient and the augmented constant. Defaults to a blank if null.
      @param pPrefix String value prepended to the output String, once. Ignored if null.
      @param pSuffix String appended to the end of the output, once. Ignored if null.
  */
  public String format(double[] pRHVM, String pRowSep, String pColSep, String[] pVars, String[] pRels, String pAugSep, String pPrefix, String pSuffix) {
    boolean isSystem = !(pVars == null) || !(pRels == null);
    pPrefix = (pPrefix == null)?"":pPrefix;
    pSuffix = (pSuffix == null)?"":pSuffix;
    pAugSep = (pAugSep == null)?"":pAugSep;
    pColSep = (pColSep == null)?",":pColSep;
    pRowSep = (pRowSep == null)?";":pRowSep;
    StringBuffer retval = new StringBuffer(pPrefix);
    if(pRHVM == null) pRHVM = new double[0];
    String[] sRHVM = new String[myCoefs.length], sVars = new String[0], sRels = new String[0];
    for(int i=0; i < sRHVM.length; i++) sRHVM[i] = "";
    for(int i=0; (i < pRHVM.length) && ( i < sRHVM.length); i++) {
      sRHVM[i] = (myNFormat == null)?"" + pRHVM[i]:myNFormat.format(pRHVM[i]);
    }
    if(isSystem) {
      if(pVars == null) pVars = new String[0];
      sVars = new String[myCoefs[0].length];
      for(int i=0; i < myCoefs[0].length; i++) {
        sVars[i] = (i < pVars.length)?pVars[i]:"x"+i;
      }
      if(pRels == null) pRels = new String[0];
      sRels = new String[myCoefs.length];
      for(int i=0; i < myCoefs.length; i++) {
        sRels[i] = (i < pRels.length)?pRels[i]:defaultRelation;
      }
      for(int iRow = 0; iRow < myCoefs.length; iRow++) {
        retval.append((myCoefs[iRow][0] < 0)?"- ":"");
        for(int iCol = 0; iCol < myCoefs[iRow].length; iCol++) {
          if(myCoefs[iRow][iCol] != 0) retval.append((myNFormat == null)?Math.abs(myCoefs[iRow][iCol]):myNFormat.format(Math.abs(myCoefs[iRow][iCol])) ).append(sVars[iCol]);
          if(iCol < myCoefs[iRow].length - 1) {
            retval.append(pColSep);
            switch((int) Math.signum(myCoefs[iRow][iCol+1])) {
              case 1:
                retval.append(" + ");
                break;
              case -1:
                retval.append(" - ");
              default:
            }
          }
        }
        retval.append(sRels[iRow]);
        retval.append(sRHVM[iRow]);
        if(iRow < myCoefs.length - 1) retval.append(pRowSep);
      }
    } else {
      for(int iRow = 0; iRow < myCoefs.length; iRow++) {
        for(int iCol = 0; iCol < myCoefs[iRow].length; iCol++) {
          retval.append((myNFormat == null)?myCoefs[iRow][iCol]:myNFormat.format(myCoefs[iRow][iCol]));
          if(iCol < myCoefs[iRow].length - 1) retval.append(pColSep);
        }
        retval.append(pAugSep).append(sRHVM[iRow]);
        if(iRow < myCoefs.length - 1) retval.append(pRowSep);
      }
    }
    retval.append(pSuffix);
    return retval.toString();
  }

   /** This method is a souped-up version of toString(String, String). It allows for the inclusion of a prefix- and suffix-String and it formats the numbers in the 
      element array.
      @param pRowSep The row separator. Defaults to a newline.
      @param pColSep The column separator. Defaults to a single space.
      @param pPrefix The prefix, which appears ONCE and first. Ignored if null or empty.
      @param pSuffix The suffix tacked on at the end. Ignored if empty or null.
      @return The MMatrix formatted as a String.
  */
  public String formatMatrix(String pRowSep, String pColSep, String pPrefix, String pSuffix) {
    StringBuffer retval = new StringBuffer();
    String sRowSep = Question.validString(pRowSep)?pRowSep:"\n";
    String sColSep = Question.validString(pColSep)?pColSep:" ";
    if(Question.validString(pPrefix)) retval.append(pPrefix);
    for(int i=0; i < myCoefs.length; i++) {
      for(int j=0; j < myCoefs[i].length; j++) {
        retval.append(((myNFormat == null)?myCoefs[i][j]:myNFormat.format(myCoefs[i][j])));
        if(j < myCoefs[i].length - 1) retval.append(sColSep);
      }
      if(i < myCoefs.length - 1) retval.append(sRowSep);
    }
    if(Question.validString(pSuffix)) retval.append(pSuffix);
    return retval.toString();
  }

  /** Each MMatrix has an internal NumberFormat object. If not null it is applied to the numbers in ALL format__() methods. Use this method to set the 
      NumberFormat for this MMatrix.
      @param pNFormat A NumberFormat to use when formatting matrices. Set to null to switch it off.
  */
  public void setNumberFormat(NumberFormat pNFormat) { myNFormat = pNFormat; }

  /** Use this method to set the default NumberFormat for all subsequent MMatrix instances. Setting null clears it. */
  public static void setDefaultNumberFormat(NumberFormat pNFormat) { DEFAULT_NUMBER_FORMAT = pNFormat; }

  /** Instance shadow to set default NumberFormat. */
  public void setNewDefaultNumberFormat(NumberFormat pNFormat) { DEFAULT_NUMBER_FORMAT = pNFormat; }

  /** Render this MMatrix as MathML. The pLFence and pRFence elements are parsed but should only contain a fence character. If either fence character is null or 
      empty it is not added.
      @param pLFence The left fence character.
      @param pRFence The right fence character.
      @return A MathML-formatted MMatrix.
  */
  public String toMathML(String pLFence, String pRFence) {
    StringBuffer retval = new StringBuffer();
    if(Question.validString(pLFence) || Question.validString(pRFence)) retval.append("<mrow>");
    if(Question.validString(pLFence)) retval.append(MathXML.parseMathML(pLFence)).append("\n");
    retval.append("<mtable>\n");
    for(int i=0; i < myRows; i++) {
      retval.append("  <mtr> ");
      for(int j=0; j < myCols; j++) {
        retval.append("<mtd><mn>").append(((myNFormat == null)?myCoefs[i][j]:myNFormat.format(myCoefs[i][j]))).append("</mn></mtd>");
      }
      retval.append(" </mtr>\n");
    }
    retval.append("</mtable>\n");
    if(Question.validString(pRFence)) retval.append(MathXML.parseMathML(pRFence));
    if(Question.validString(pLFence) || Question.validString(pRFence)) retval.append("</mrow>\n");
    return retval.toString();
  }
  /** Convenience method with square braces for fences: toMathML('[', ']');. */
  public String toMathML() { return toMathML("[", "]"); }

  /** <p>Set the global LaTeX environment used to render the MMatrix. These settings are global and once set remain in effect until changed. The default LaTeX 
      environment used with the '/begin{}' statement is 'tabular' and setting pEnv to null restores this.</p>
      <p>The 'pColSpec' parameter determines whether or not column alignment specifications are appended to the end of the '/begin{}' statement. This is true by 
      default but can be changed. The 'pOptions' parameter is for any square-bracket options between '/begin' and the opening curly brace. If this String is 
      defined it will be included verbatim otherwise it is ignored. By default it is null.</p>
      <p>To summarize, the parameter placement is:<br />
        <center>/begin[<i>options</i>]{<i>environment</i>}{<i>colSpec</i>}</center></p>
      @param pEnv The LaTeX environment used to format the MMatrix. Defaults to 'tabular.'
      @param pColSpec true to include the column alignment specification (default), false to omit it.
      @param pOptions LaTeX 'square bracket' options. Ignored if null.
  */
  public static void setLatexEnvironment(String pEnv, boolean pColSpec, String pOptions) {
    theLatexEnv = (Question.validString(pEnv)?pEnv:"tabular");
    theLatexEndEnv = theLatexEnv;
    if(theLatexEndEnv.indexOf("}") > -1) { theLatexEndEnv = theLatexEndEnv.substring(0, theLatexEndEnv.indexOf("}")); }
    enableColSpec = pColSpec;
    theLatexOptions = pOptions;
  }

  /** Instance shadow to set LaTex environment. */
  public void setNewLatexEnvironment(String pEnv, boolean pColSpec, String pOptions) { setLatexEnvironment(pEnv, pColSpec, pOptions); }

  /** <p>Print this MMatrix using LaTeX notation. Parameters are per the {@link #format format()} method. The Row/Column separators are handled internally as is 
      the Augmentation separator. The prefix and suffix Strings are set to the appropriate LaTeX <code>/begin{}</code> and <code>/end{}</code> statements. 
      All LaTeX configuration settings are global and are set with the <code>setLatexEnvironment()</code> methods.</p>
      <h3>LaTeX-ing MMatrices</h3>
      <p>(N.B. When discussing LaTeX statements here the forward slash ('/') is used in place of the backslash to avoid ISSUES.)</p>
      <p>This method contains the most sophisiticated LaTeX processing of all the core tools. It is designed to output LaTeX markup for a plain matrix, an 
      augmented matrix or a system of equations. As stated above it makes use of the standard <code>format()</code> method with LaTeX-specific values provided for 
      the required arguments. Since LaTeX allows quite a few options these values can be configured via the {@link #setLatexEnvironment setLatexEnvironment()} 
      methods.</p>
      <p>By default the environment used is 'tabular' with column-alignment specification enabled and no (<i>square-bracket</i>) options. These are easily changed 
      but, once set, remain at the set values until they are reset! It is also possible through clever use of the 'environment' parameter to include a second 
      'curly-brace' structure after the environment. Simply disable column alignment, put a closing curly brace after the environment name, an opening curly brace 
      and whatever is desired within the second pair. This method automatically closes the last curly brace.</p>
      <p>The machinery here uses a least common denominator approach: it supplies the bare minimum LaTeX to allow for maximum flexibility. It supplies the 
      'begin... \end' block with the settings detailed above but <i>nothing else.</i> Specifically it doesn't apply any fence characters so these should be included 
      in the Question format.</p>
      <p>Respectively the arguments are the right-side constant matrix (as a double[] array), the variables to be used for a system of equations and the relations 
      used in the system of equations, both of the latter as String[] arrays. Alignment characters can be included with the variables or the relations in order to 
      control system-of-equations alignment more precisely. Exactly what is formatted depends on the arguments. More specifically, it depends on which ones are 
      defined and which are not.</p>
      <p>If all three arguments are null the method prints a plain matrix. If the constant matrix is non-null and both other arguments are null the result is an 
      augmented matrix. The '|c' IS added if the column specifier is enabled so the vertical line is supplied. If the first argument is non-null and <i>either</i> 
      of the other two is non-null a system of equations is attempted. In order for this attempt to succeed the environment must be changed to 'eqnarray' or 
      something equivalent and the column alignment specifier must be disabled. For systems of equations (or inequalities!) both the second and third arguments 
      should be specified; however, the default relation for this method is '&amp;=&amp;'.</p>
      <p><b>IMPORTANT NOTES:</b> In order to format an augmented matrix both the second and third arguments MUST be null. For either an augmented matrix or a 
      system the first argument must be non-null. For a system of equations the environment MUST be changed! If fancier formatting options or more granular control 
      is desired the plain <code>format()</code> method must be used.</p>
      @param pRHVM The constant matrix for augmented matrices or systems. Must be non-null for either.
      @param pVars The variables used when formatting a system of equations. If non-null a system is attempted.
      @param pRels The relations used when formatting a system. If non-null this is attempted.
  */
  public String toLatex(double[] pRHVM, String[] pVars, String[] pRels) {
    String lRowSep = "\\\\\n", lColSep = " & ", lAugSep = null, lSuffix="\n\\end{"+theLatexEndEnv+"}\n", retval = null;
    StringBuffer lPrefix = new StringBuffer("\\begin");
    if(Question.validString(theLatexOptions)) lPrefix.append("[").append(theLatexOptions).append("]");
    lPrefix.append("{").append(theLatexEnv);
    if(enableColSpec) {
      lPrefix.append("}{");
      for(int i=0; i < myCols; i++) { lPrefix.append("c"); }
    }
    if(pRHVM != null) { // Augmented Matrix or System
      if((pVars == null) && (pRels == null)) {
        // Augmented Matrix
        lPrefix.append("|c");
        lAugSep = " & ";
      } else {
        // System of Equations
        lColSep = " ";
      }
    }
    lPrefix.append("}\n");
    defaultRelation = "&=&";
    retval = format(pRHVM, lRowSep, lColSep, pVars, pRels, lAugSep, lPrefix.toString(), lSuffix);
    defaultRelation = " = ";
    return retval;
  }

}
