package cox.jmatt.java.MathTools;

import java.util.*;

/** <p>The MRadical class provides an easy way to create and present radicals without requiring double approximation or other theatrics. Once created, MRadicals 
    are immutable. The sign, radicand, and index can be examined but not changed. MRadical also has methods for comparing two MRadicals. Apart from these methods 
    MRadical is primarily a formatting class.</p>
    <p>The true power of this class is in the format() method. It provides tokens for the 'raw' radical as well as the simplified radical. If a radicand can be 
    simplified, i.e. it contains a perfect-power factor, the perfect power is factored out, represented as a coefficient to the radical, and the radicand reduced 
    appropriately. This is an automatic part of the formatting process.</p>
*/
public class MRadical {

  /** Format string for debugging purposes. NOT an official format! */
  public static final String FORMAT_DEBUG = "MRadical\nRadicand: <0>\nIndex: <1>\nAbsRadicand: <2>\nSign: <3>\nImp. sign: <4>\nSignum: <5>\nValue: <6>" + 
  "\nSimp. Radicand: <7>\nAbs. coef: <8>\nCoef: <9>\nImp.Coef: <10>\nImp. signed coef: <11>\nImp. index: <12>\n- - -";
  /** Standard raw form for radicals: radical(index, radicand). */
  public static final String FORMAT_STANDARD = "radical(<1>, <0>)";
  /** Standard simplified form for radicals: coef radical(index, simp. radicand). */
  public static final String FORMAT_STANDARD_SIMP = "<9> radical(<1>, <7>)";
  /** OpenOffice.org, raw form. */
  public static final String FORMAT_OO = "nroot {<12>} {<0>}";
  /** OpenOffice.org, simplified . */
  public static final String FORMAT_OO_SIMP = "<9> nroot {<12>} {<7>}";
  /** LaTeX raw form. */
  public static final String FORMAT_LATEX = "\\sqrt[<12>]{<0>}";
  /** LaTeX simplified. */
  public static final String FORMAT_LATEX_SIMP = "<9> \\sqrt[<12>]{<7>}";

  private static String DEFAULT_FORMAT = FORMAT_STANDARD;
  private static final String[] COMPARES = {"<", "=", ">"};

  private int myRadicand = 0;
  private int myIndex = 1;
  private byte mySign = 1;
  private String myFormat = DEFAULT_FORMAT;

  /** Standard script-exposure constructor. The default MRadical is zero, or, the first root of zero. */
  public MRadical() { super(); }

  /** Constructor to specify all critical parts of the radical: radicand, and index. Negative values for the index are made positive. If the radicand is 
      negative and the index is even, the default radical is created. If the index is zero, the radical becomes 1.
      @param pRadicand The radicand.
      @param pIndex The index.
  */
  public MRadical(int pIndex, int pRadicand) {
    super();
    if(pIndex == 0) {
      myRadicand = 1;
    }
    if(pIndex < 0) pIndex = -pIndex;
    if((pRadicand < 0) && (pIndex % 2 == 0)) return;
    myIndex = pIndex;
    if(pRadicand < 0) mySign = (byte) -1;
    if(pRadicand == 0) mySign = (byte) 0;
    myRadicand = Math.abs(pRadicand);
  }

  /** Copy constructor. */
  public MRadical(MRadical pRadical) {
    super();
    if(pRadical != null) {
      myRadicand = pRadical.absRadicand();
      myIndex = pRadical.index();
      mySign = (byte) pRadical.signum();
    }
  }

  /** Static method to create a MRadical given its radicand and index. */
  public static MRadical createMRadical(int pIndex, int pRadicand) { return new MRadical(pIndex, pRadicand); }

  /** Instance method to create a new MRadical. */
  public MRadical newMRadical(int pIndex, int pRadicand) { return new MRadical(pIndex, pRadicand); }

  /** Static method to create a square root. Index is automatically 2! */
  public static MRadical createMRadical(int pRadicand) { return new MRadical(2, pRadicand); }

  /** Instance method to create a square root. */
  public MRadical newMRadical(int pRadicand) { return new MRadical(2, pRadicand); }

  /** Method to construct a MRadical with a 'coefficient,' e.g. 2 sqrt(5) which is equivalent to sqrt(20).
      @param pCoef The number multiplied onto the radical.
      @param pIndex The index (root) of the radical.
      @param pRadicand The radicand.
  */
  public static MRadical createMRadical(int pCoef, int pIndex, int pRadicand) {
    return new MRadical(pIndex, ((int) Math.pow(pCoef, pIndex) * pRadicand));
  }

  /** Instance version of the three-int factory method. */
  public MRadical newMRadical(int pCoef, int pIndex, int pRadicand) { return createMRadical(pCoef, pIndex, pRadicand); }

  /** This method returns a constructor-like representation of the MRadical. */
  public String toString() { return("MRadical("+myIndex+", "+(mySign*myRadicand)+");"); }

  /** Evaluate the radical as a double. */
  public double toDouble() {
    if(myIndex == 0) return 1.0;
    return(mySign * Math.pow((double) myRadicand, 1.0/myIndex));
  }

  /** Get the value of the radicand. */
  public int radicand() { return mySign*myRadicand; }

  /** Get the absolute value of the radicand. */
  public int absRadicand() { return myRadicand; }

  /** Get the value of the index. */
  public int index() { return myIndex; }

  /** Return the sign of the MRadical: -1 if negative, 0 if zero, +1 if positive. */
  public int signum() { return mySign; }

  /** Check to see if the two MRadicals represent the same value. It checks actual value and NOT identity, so the square root of four (2) WILL test as equal to 
      the cube root of 8 (also 2).
  */
  public boolean equals(MRadical pRadical) {
    if(pRadical == null) return false;
    return ( ((int) Math.pow(myRadicand, pRadical.index())) == ((int) Math.pow(pRadical.radicand(), myIndex)));
  }

  /** This method tests to see if the two MRadicals are <b>exactly identical</b>, not just equal in value. */
  public boolean identicalTo(MRadical pRadical) {
    if(pRadical == null) return false;
    return ((radicand() == pRadical.radicand()) && (myIndex == pRadical.index()));
  }

  /** Compare this MRadical to another. Returns -1 if this is less than pRadical, 0 if both are equal, and 1 if this is greater than pRadical.
      @throws NullPointerException If pRadical is null.
  */
  public int compareTo(MRadical pRadical) {
    if(pRadical == null) throw new NullPointerException("Compared MRadical was null");
    if(this.equals(pRadical)) return 0;
    return (toDouble() < pRadical.toDouble())?-1:1;
  }

  /** Compare this MRadical to another but return a single-character String: '&lt;', '=' or '&gt'. This method will return null if the argument is null. */
  public String sCompareTo(MRadical pRadical) {
    if(pRadical == null) return null;
    return COMPARES[compareTo(pRadical) + 1];
  }

  /** <p>Format this MRadical per the template supplied. This is a standard <code>Question.fillTemplate()</code> with the following tokens supplied:
      <ul>
        <li>&lt;0&gt; is the radicand with the sign.</li>
        <li>&lt;1&gt; is the index.</li>
        <li>&lt;2&gt; is the absolute value of the radicand.</li>
        <li>&lt;3&gt; is the explicit sign: '+' or '-'.</li>
        <li>&lt;4&gt; is the implicit sign: '-' if negative, '' otherwise.</li>
        <li>&lt;5&gt; is the signum of the radicand.</li>
        <li>&lt;6&gt; is the double value of the radical.</li>
        <li>&lt;7&gt; is the simplified radicand.</li>
        <li>&lt;8&gt; is the coefficient of the simplified radical without the sign.</li>
        <li>&lt;9&gt; is the signed coefficient of the simplified radical.</li>
        <li>&lt;10&gt; is the implicit coefficient. If the radical doesn't simplify it is ''.</li>
        <li>&lt;11&gt; is the implicit signed coefficient. If the coefficient is +1 this is blank, if -1 it is '-'.</li>
        <li>&lt;12&gt; is the implicit index. Blank if the index is 2, the same as &lt;1&gt; otherwise.
      </ul>
      NOTE: Tokens 7 to 11 result from the radical being simplified. The greatest 'perfect power' of the index is factored out of the radicand and interpreted 
      as a coefficient for the radical expression. If there is no perfect power factor inside the radicand, the simplified tokens equal their unsimplified 
      equivalents.</p>
      <p>MRadical maintains the sign separate from the radicand itself. Tokens 10 and 11 reflect this through the simplification process. The negative sign, if 
      present, appears on the coefficient. When using the simplified tokens, if there is a chance the radicand might be negative <b>token 2</b> should be used for 
      the actual radicand, otherwise the negative sign may show up twice!</p>
      @param pTemplate The template to use when formatting the MRadical. If null or empty, FORMAT_STANDARD is used.
      @return The freshly-formatted MRadical.
  */
  public String format(String pTemplate) {
    if(!Question.validString(pTemplate)) pTemplate = FORMAT_STANDARD;
    String[] tokenz = new String[13];
    tokenz[0] = ""+radicand(); tokenz[1] = ""+myIndex; tokenz[2] = ""+myRadicand; tokenz[3] = (mySign < 0)?"-":"+"; tokenz[4] = (mySign < 0)?"-":"";
    tokenz[5] = ""+mySign; tokenz[6] = ""+toDouble();
    
    int theCoef = 0, theFactor = 0, theRad = myRadicand, retCoef = 1;
    while(theFactor < myRadicand) {
      theCoef++;
      theFactor = (int) Math.pow(theCoef, myIndex);
      if(myRadicand % theFactor == 0) { retCoef = theCoef; theRad = myRadicand / theFactor; }
    }
    tokenz[7] = ""+theRad; tokenz[8] = ""+retCoef; tokenz[9] = tokenz[4]+retCoef;
    tokenz[10] = (retCoef > 1)?""+retCoef:""; tokenz[11] = tokenz[4]+tokenz[10];
    tokenz[12] = (myIndex == 2)?"":tokenz[1];
    return Question.fillTemplate(pTemplate, tokenz);
  }

  /** Express the radical using MathML. No formatting is done, so it's just the radicand and the index. If the index is 2, &lt;msqrt&gt; tags are used, otherwise 
      the tags are &lt;mroot&gt;.
  */
  public String toMathML() {
    if(myIndex == 2) return MathXML.mSqrt(""+radicand());
    return MathXML.mRoot(""+radicand(), ""+myIndex);
  }

  /** Express the radical in simplified XML notation. It is an empty &lt;MRadical&gt; tag with attributes 'radicand=' and 'index='. */
  public String toXML(String pID) {
    StringBuffer retval = new StringBuffer("<MRadical ");
    if(Question.validString(pID)) retval.append("id=\"").append(pID).append("\" ");
    retval.append("radicand=\"").append(radicand()).append("\" ");
    retval.append("index=\"").append(myIndex).append("\" />");
    return retval.toString();
  }

  /** Set the format String for this MRadical. A blank or null value resets to the default format. */
  public void setFormat(String pFormat) {
    myFormat = (Question.validString(pFormat))?pFormat:DEFAULT_FORMAT;
  }

  /** Set the default format for all MRadicals created after this method is called. An empty or null value resets to FORMAT_STANDARD. */
  public static void setDefaultFormat(String pFormat) {
    DEFAULT_FORMAT = (Question.validString(pFormat))?pFormat:FORMAT_STANDARD;
  }

  /** Instance version of <code>setDefaultFormat()</code>. */
  public void setNewDefaultFormat(String pFormat) { setDefaultFormat(pFormat); }

  /** Format this MRadical per its currently-set format String. */
  public String format() { return format(myFormat); }

}
