package cox.jmatt.java.MathTools;

import java.util.*;
import cox.jmatt.java.MathTools.test.*;

/** <p>This is the power-class of the MathTools package. It provides services to all other packages except Question and contains several powerful and flexible 
    power-methods. It also contains a pseudorandom number generator that can be re-seeded dynamically to allow for 'deterministically random' numbers. Along with 
    this is a re-seed method that takes its seed from the system clock, so 'more' randomness is also available. Even though all MEngine methods are static, 
    a constructor is provided to allow exposure to scripts.</p>
    <h3>Strings and Parameter Blocks</h3>
    <p>In addition to its mathematical methods, <code>MathEngine</code> also contains equipment for working with Strings and parameter blocks. A parameter block 
    is a group of two or more parameters (Global Pizza) that share a common prefix and sequential numbers. The prefix must be <b>identical</b> on all elements 
    and the numbering must start at zero. I.e. topping.0 = cheese, topping.1 = onions, topping.2 = pepperoni, etc. Param block detection stops when an attempt to 
    retrieve a number fails, so numbers cannot be skipped.</p>
    <p>The <code>splick()</code> methods split a string into tokens and return one at random. The <code>splickInt()</code> method does the same, but assumes a 
    space-separated String of ints and returns its choice as an int. The <code>spill()</code> power-method breaks a String into an array, guarantees a minimum 
    length, and pads any 'leftover' elements with a designated fill value.</p>
*/
public class MathEngine {

  /** This is the maximum prime number available to the <code>rPrime()</code> methods: 1013. */
  public static final int MAX_PRIME = 1013;

  /** This is the prefix attached to quantities stored from <code>parseRPN()</code>. */
  public static final String MEM_PREFIX = "RPNC_";

  private static final long MODULUS = 1051*1699;
  private static final byte MAX_RANDOM_RADIUS = 3;
  private static final String SOME_PRIMES = "2 3 5 7 11 13 17 19 23 29 31 37 41 43 47 53 59 61 67 71 73 79 83 89 97 101 103 107 109 113 127 131 137 139 149 151 "
        + "157 163 167 173 179 181 191 193 197 199 211 223 227 229 233 239 241 251 257 263 269 271 277 281 283 293 307 311 313 317 331 337 347 349 353 359 367 "
        + "373 379 383 389 397 401 409 419 421 431 433 439 443 449 457 461 463 467 479 487 491 499 503 509 521 523 541 547 557 563 569 571 577 587 593 599 601 "
        + "607 613 617 619 631 641 643 647 653 659 661 673 677 683 691 701 709 719 727 733 739 743 751 757 761 769 773 787 797 809 811 821 823 827 829 839 853 "
        + "857 859 863 877 881 883 887 907 911 919 929 937 941 947 953 967 971 977 983 991 997 1009 1013";
  private static long curSeed = System.currentTimeMillis() % MODULUS;
  private static boolean adjRand = false;
  private static double adjPower = 1.0, adjRoot = 1.0;
  private static int theDelta = 1;
  private static byte randomRadius = 0;
  private static Random myRandomGenerator = new Random();

  /** Constructor provided to generate instances for scripting. */
  public MathEngine() { super(); }
  /**<p>This is the principal (pseudo-) random number function. It provides random numbers from zero (inclusive) to one (exclusive). It is probably not strong 
      enough for serious cryptographic work but it should suffice for most math-test uses.</p>
      <p>This method was changed in stable release 1.4 to make use of a <code>java.util.Random</code> object. Very basic code was inserted to assure appropriate 
      randomness in the form of limiting runs of high or low values.</p>
      @return A pseudorandom number on the interval [0,1).
  */
  public static double random() {
    double retval = myRandomGenerator.nextDouble();
    while(((randomRadius <= -MAX_RANDOM_RADIUS) && (retval < 0.5)) || ((randomRadius >= MAX_RANDOM_RADIUS) && (retval > 0.5))) retval = myRandomGenerator.nextDouble();
    if((retval < 0.5) && (randomRadius > -MAX_RANDOM_RADIUS)) randomRadius--;
    if((retval > 0.5) && (randomRadius < MAX_RANDOM_RADIUS)) randomRadius++;
    if(adjRand) retval = java.lang.Math.pow(retval, adjPower / adjRoot);
    return retval;
  }
  /** For ADVANCED USERS. This method allows 'tweaking' of the numbers generated by <code>random()</code>, hence ALL random numbers generated here. If adjustment 
      is enabled the 'raw' random number generated is raised to the power (pPower/pRoot). That is: <code>Adjusted = Math.pow(Raw, pExponent/pRoot)</code>. If 
      either argument is zero it is set to 1. If both are zero adjustment is disabled.
      @param pPower The power used to adjust the 'raw' random generated. Minimum value is 1.
      @param pRoot The root applied to the raw random generated. Minimum value of 1.
  */
  public static void adjustRandom(double pPower, double pRoot) {
    if((pPower < 1.0) && (pRoot < 1.0)) {
      adjRand = false;
      return;
    }
    adjRand = true;
    adjPower = java.lang.Math.max(1.0, pPower);
    adjRoot = java.lang.Math.max(1.0, pRoot);
  }
  /** Reseed the random number generator with the given value. WARNING: Reseeding with the same number yields THE SAME sequence of randoms! Use with care.
      @param pSeed The new seed for the random generator.
  */
  public static void reseed(long pSeed) { myRandomGenerator.setSeed(pSeed); }
  /** Reseed the random number generator from the system clock. These sequences are harder to guess but not (quite) impossible. */
  public static void reseed() { myRandomGenerator.setSeed(System.currentTimeMillis()); }
  /** Generate a random int from a range. The number generated will be from the lowest in the range (inclusive) to the highest (exclusive). If the high value is 
      less than the low the two are swapped before calculation. If the numbers are the same, it is returned. If the difference is 1, the low value is returned.
      @param pLow The low number in the range. Included in the possible return values.
      @param pHigh The high number in the range. Excluded from the possible return values.
      @return A random int from the interval [pLow, pHigh).
  */
  public static int random(int pLow, int pHigh) {
    if((pLow == pHigh) || (Math.abs(pHigh - pLow) == 1)) return pLow;
    int retval = pLow;
    if(pHigh < pLow) { pLow = pHigh; pHigh = retval; }
    retval = ((int) ((pHigh - pLow)*random())) + pLow;
    return retval;
  }
  /** Generate a nonzero random from a range. If the high and low ends are the same number, it is returned. If the difference in values is one, the low value 
      is returned. IMPORTANT NOTE: Calling <code>nzRandom(0,0)</code> <b>will</b> yield zero. Calling <code>nzRandom(0, 1)</code> yields a value of one!
      @param pLow The low int in the range (inclusive).
      @param pHigh The high end of the range (exclusive).
      @return A nonzero number from the range, provided the difference in values is greater than 1.
  */
  public static int nzRandom(int pLow, int pHigh) {
    if(pLow == 0) {
      if(pHigh == 0) return 0;
      if(pHigh == 1) return 1;
    }
    if((pLow == pHigh) || (Math.abs(pHigh - pLow) == 1)) return pLow;
    int retval = 0;
    while(retval == 0) { retval = random(pLow, pHigh); }
    return retval;
  }
  /** Pick a random prime number. The return will be a prime number between 2 and MAX_PRIME inclusive. */
  public static int rPrime() { return splickInt(SOME_PRIMES); }
  /** Pick a random prime number guaranteed to be less than the number given. If pMax is less than 10 it is set at 10.
      @param pMax The upper limit for the prime number picked. If less than 10 it becomes 10.
  */
  public static int rPrime(int pMax) {
    int retval = rPrime();
    if(pMax < 10) pMax = 10;
    while(retval >= pMax) { retval = rPrime(); }
    return retval;
  }

  /** This method reflects <code>CapCom.hitCount()</code>. It is useful for generating application-unique numbers. */
  public static long hitCount() { return CapCom.hitCount(); }

  /** <p>A Knob is a random number selected from a range. What distinguishes it from any other random is that Knobs are generally picked more than one at a time. 
      This method takes a space-separated String of numbers, breaks them into individual numbers, and interprets them pair-wise as low-end high-end values for 
      a random function. One random value is generated for each <i>pair</i> of numbers in the parameter String. If the number of values in the String is odd, 
      the last number is ignored.</p>
      <p>By default this method generates nonzero randoms. To change this behavior place an equals sign in front of the low value in the range. E.g. '=-5 6' would 
      choose a possibly-zero value between -5 and 5. NOTE: Since the Knob string is <i>space</i>-separated there must be NO SPACE between the equals sign and the 
      number or negative sign!</p>
      <p>If pKnobs is not formed properly (pairs of numbers, no non-numbers except the optional prefix) the return value is indeterminate. Any errors are reported 
      at the Error level.</p>
      @param pKnobs A String of space-separated numbers.
      @return An int[] array containing random values.
  */
  public static int[] setKnobs(String pKnobs) {
    int[] retval = new int[1]; retval[0] = 0;
    int iLow, iHigh;
    pKnobs = pKnobs.trim();
    if(!Question.validString(pKnobs)) return retval;
    boolean allowZero = false;
    String[] sRanges = pKnobs.split(" ");
    if(sRanges.length < 2) return retval;
    retval = new int[sRanges.length / 2];
    for(int i=0; i < (sRanges.length / 2); i++) {
      allowZero = false;
      if(sRanges[2*i].startsWith("=")) { allowZero = true; sRanges[2*i] = sRanges[2*i].substring(1); }
      if(sRanges[2*i+1].startsWith("=")) { allowZero = true; sRanges[2*i+1] = sRanges[2*i+1].substring(1); }
      try {
        iLow = Integer.parseInt(sRanges[2*i]); iHigh = Integer.parseInt(sRanges[2*i+1]);
        retval[i] = allowZero?random(iLow, iHigh):nzRandom(iLow, iHigh);
      } catch(Exception ex) {
        CapCom.error("Error setting Knob "+i+" in: "+pKnobs, ex);
        retval[i] = 0;
      }
    }
    return retval;
  }

  /** This method returns a String such that <code>isoKnobs(pKnobs) == isoKnobs(isoKnobs(pKnobs))</code>. That is, calling setKnobs on the resulting string yields 
      the same values each time.
      @param pKnobs The original Knobs string.
      @return A String that repeated calls to <code>setKnobs()</code> on will yield the same results.
  */
  public static String isoKnobs(String pKnobs) {
    int[] knobz = setKnobs(pKnobs);
    StringBuffer retval = new StringBuffer();
    for(int i=0; i < knobz.length; i++) { retval.append(" ").append(knobz[i]).append(" ").append(knobz[i]); }
    return retval.toString().substring(1);
  }
  /** This method echoes <code>System.currentTimeMillis()</code> should such be desirable. */
  public static long getMTime() { return System.currentTimeMillis(); }
  /** This method checks to see if one double is 'close to' another. The computation returns true if <code>abs(pLHS - pRHS) < 5*Math.pow(10.0, -pDelta)</code>.
      @param pLHS The first double to check.
      @param pRHS The second double to check.
      @param pDelta The order of magnitude to compare. Will be made positive if negative.
      @return true if pLHS is within (5.0 * 10^(-pDelta)) of pRHS, false otherwise.
  */
  public static boolean closeTo(double pLHS, double pRHS, int pDelta) {
    return (Math.abs(pLHS - pRHS) < (5.0 * Math.pow(10, -Math.abs(pDelta))));
  }
  /** This method sets the magnitude for the two-argument <code>closeTo()</code>. If pDelta is negative it is made positive when set. The default value is 1.
      @param pDelta The magnitude for the two-double method.
  */
  public static void setCloseTo(int pDelta) { theDelta = pDelta; }
  /** Compare the difference of two doubles against the internal magnitude. Otherwise like the three-argument <code>closeTo()</code>. */
  public static boolean closeTo(double pLHS, double pRHS) { return closeTo(pLHS, pRHS, theDelta); }
  /** <p>This method splits sData based on sToken and randomly picks one element from the resulting array. The name comes from 'split-pick'. If pData is null or 
      empty it is returned unchanged. If pToken is null or empty, pData is split on spaces.</p>
      <p>NOTE: This method utilizes the <code>String.split()</code> method so regex collisions can occur! If such a thing happens the data String is returned whole 
      and the error is reported at Error level.</p>
      @param pData The String to be split.
      @param pToken The token around which pData is to be split.
      @return One randomly-chosen element from the array resulting from the split.
  */
  public static String splick(String pData, String pToken) {
    if(!Question.validString(pData)) return pData;
    if(!Question.validString(pToken)) pToken = " ";
    try {
      String[] sTemp = pData.split(pToken);
      pData = sTemp[random(0, sTemp.length)];
    } catch(Exception ex) { CapCom.error("Error splicking: "+pData, ex); }
    return pData;
  }

  /** Convenience method for splick() that splits on a single space character. */
  public static String splick(String pData) { return splick(pData, " "); }

  /** This method splits a space-separated String of integers, picks one, and returns it as an int. Errors are reported and the return value for such is zero. */
  public static int splickInt(String pData) {
    int retval = 0;
    try { retval = Integer.parseInt(splick(pData, " ")); } catch(Exception ex) { CapCom.error("splickInt error on: "+pData, ex); }
    return retval;
  }

  /** This method splits a String on the specified break-String and, if the resulting array is shorter than the length requested, pads the remaining elements 
      with the specified fill String. If the split array is longer than requested it is returned as-is. If pData is null or empty a new String[] array of the 
      specified length is created, filled with the fill value, and returned. The method name comes from 'split-fill.'
      @param pData The String to be split.
      @param pSplitter The break-string to use for splitting pData. If null or blank, a space is used.
      @param pLength The guaranteed minimum length of the returned array.
      @param pFill The String used to pad the array if the splitting didn't give enough elements. If null, an empty String is used.
      @return A String[] array of at least pLength elements, padded at the end with pFill if needed.
  */
  public static String[] spill(String pData, String pSplitter, int pLength, String pFill) {
    String[] retval = new String[1]; retval[0] = "";
    if(!Question.validString(pData)) pData = " ";
    pSplitter = Question.validString(pSplitter)?pSplitter:" ";
    retval = pData.split(pSplitter);
    if(retval.length >= pLength) return retval;
    String[] nuRetval = new String[pLength];
    if(pFill == null) pFill = "";
    for(int i=0; i < pLength; i++) {
      if(i < retval.length) {
        nuRetval[i] = retval[i];
      } else {
        nuRetval[i] = new String(pFill);
      }
    }
    return nuRetval;
  }

  /** <p>This method rearranges the elements in an Object[] array into random order. The array sent in is shuffled and returned. The shuffling method used is 
      simple. The array is iterated over forward and backward five times. With each iteration there is a 55% chance that any element will be swapped with the 
      adjascent element below it.</p>
      <p><b>NOTE:</b> This method <i>does</i> alter the array that is sent in. The element order is changed, the array is not copied!</p>
      @param pAry The array to shuffle.
      @return The same array with elements in random order.
  */
  public static Object[] shuffle(Object[] pAry) {
    Object oTemp = null;
    for(int i=0; i < 5; i++) {
      for(int j = 1; j < pAry.length; j++) {
        if(random() < 0.45) continue;
        oTemp = pAry[j-1];
        pAry[j-1] = pAry[j];
        pAry[j] = oTemp;
      }
      for(int j = pAry.length - 1; j > 0; j--) {
        if(random() < 0.45) continue;
        oTemp = pAry[j-1];
        pAry[j-1] = pAry[j];
        pAry[j] = oTemp;
      }
    }
    return pAry;
  }

  /** <code>shuffle()</code> an array of <b>int</b>s. */
  public static int[] iShuffle(int[] pAry) {
    int oTemp;
    for(int i=0; i < 5; i++) {
      for(int j = 1; j < pAry.length; j++) {
        if(random() < 0.45) continue;
        oTemp = pAry[j-1];
        pAry[j-1] = pAry[j];
        pAry[j] = oTemp;
      }
      for(int j = pAry.length - 1; j > 0; j--) {
        if(random() < 0.45) continue;
        oTemp = pAry[j-1];
        pAry[j-1] = pAry[j];
        pAry[j] = oTemp;
      }
    }
    return pAry;
  }

  /** The <code>shuffle()</code> a double[] array. */
  public static double[] dShuffle(double[] pAry) {
    double oTemp;
    for(int i=0; i < 5; i++) {
      for(int j = 1; j < pAry.length; j++) {
        if(random() < 0.45) continue;
        oTemp = pAry[j-1];
        pAry[j-1] = pAry[j];
        pAry[j] = oTemp;
      }
      for(int j = pAry.length - 1; j > 0; j--) {
        if(random() < 0.45) continue;
        oTemp = pAry[j-1];
        pAry[j-1] = pAry[j];
        pAry[j] = oTemp;
      }
    }
    return pAry;
  }

  /** <p>This method reads a block of sequentially-numbered lines from the Global Pizza and returns them as a String[] array. The lines <b>must</b> be of the form 
      '<i>name</i>.0= text', '<i>name</i>.1= text', ... '<i>name</i>.N= text', where 'name' is the common name for all of the lines. The numbering must start at 
      zero and move up by one as the parsing stops with the first empty line. Once parsed, the lines are put into a String[] array and returned.</p>
      <p>The name comes from 'block-lines.' If there are no such lines an empty array is returned.</p>
      @param pName The common name of the block of lines to read.
      @return A String[] array containing the lines enumerated from '<i>pName</i>.0 up to the last consecutive line so named.
  */
  public static String[] blines(String pName) {
    String[] retval = new String[0];
    pName = pName.trim();
    if(!Question.validString(pName)) return retval;
    Properties myPizza = CapCom.getPizza();
    if(myPizza.isEmpty()) return retval;
    int iNdx = 0;
    ArrayList<String> myList = new ArrayList<String>();
    String sLine = myPizza.getProperty(pName+"."+iNdx, "");
    while(Question.validString(sLine)) {
      myList.add(sLine);
      iNdx++;
      sLine = myPizza.getProperty(pName+"."+iNdx, "");
    }
    retval = myList.toArray(retval);
    return retval;
  }

  /** Read a block of sequential parameter lines into an array then return one element chosen at random. If pName is null or empty an empty String is returned. 
      If the array contains only one element, it is returned. The name is 'blines-pick.'
      @param pName The name of the block of lines to parse.
      @return One of the lines chosen at random.
  */
  public static String blick(String pName) {
    String[] saTemp = blines(pName);
    if(saTemp.length < 1) return "";
    if(saTemp.length == 1) return saTemp[0];
    return saTemp[random(0, saTemp.length)];
  }

  /** This method reads a block of sequentially numbered lines from the Global Pizza and concatenates them as a single String using the supplied String as a 
      delimiter. If pName is null or empty, an empty String is returned. If there is only one line the delimiter does not appear after it. The name is from 'block 
      read.'
      @param pName The name of the block of lines to parse.
      @param pDelim The concatenation delimiter String. If null or empty, a blank String ('') is used.
      @return The specified block of lines as a single String.
  */
  public static String bread(String pName, String pDelim) {
    String[] sLines = blines(pName);
    if(sLines.length < 1) return "";
    pDelim = (pDelim == null)?"":pDelim;
    StringBuffer retval = new StringBuffer(sLines[0]);
    for(int i=1; i < sLines.length; i++) { retval = retval.append(pDelim).append(sLines[i]); }
    return retval.toString();
  }

  /** This method checks each String in an array for a specified substring. The return value is a String[] array containing only those Strings containing the 
      specified substring. If pRaw is null or empty, pMatch is null or empty, or none of the Strings contain pMatch the return value is an empty array.
      @param pRaw The String[] array from which to extract the desired Strings.
      @param pMatch The String to check against each element of pRaw.
      @return A String[] array of the pRaw elements containing pMatch.
  */
  public static String[] filter(String[] pRaw, String pMatch) {
    String[] retval = {};
    if((pRaw == null) || (pRaw.length < 1) || !Question.validString(pMatch)) return retval;
    ArrayList<String> arlTemp = new ArrayList<String>();
    for(int i=0; i < pRaw.length; i++) {
      if(Question.validString(pRaw[i]) && (pRaw[i].indexOf(pMatch) > -1)) arlTemp.add(pRaw[i]);
    }
    if(arlTemp.size() > 0) {
      retval = new String[arlTemp.size()];
      retval = arlTemp.toArray(retval);
    }
    return retval;
  }

  /** Shortcut method for <code>filter(bLines(pPrefix), pMatch)</code>. That is, it filters a block of parameters from the Global Pizza.
      @param pPrefix The common block prefix per <code>blines()</code>.
      @param pMatch The extraction substring.
  */
  public static String[] filter(String pPrefix, String pMatch) { return filter(blines(pPrefix), pMatch); }

  /** This method filters a String[] array per <code>filter()</code> and then randomly picks one of the resulting elements. The method name is loosely derived 
      from 'filter-pick.'
      @param pRaw The original array.
      @param pMatch The substring to match.
      @return One element from the resulting array, or null if none matched.
  */
  public static String flick(String[] pRaw, String pMatch) {
    String[] saTemp = filter(pRaw, pMatch);
    if(saTemp.length < 1) return null;
    return saTemp[random(0, saTemp.length)];
  }

  /** The <code>flick(String[], String)</code> method that operates on the Global Pizza. */
  public static String flick(String pPrefix, String pMatch) {
    String[] saTemp = filter(pPrefix, pMatch);
    if(saTemp.length < 1) return null;
    return saTemp[random(0, saTemp.length)];
  }

  /** Concatenate an int array int String. The last element separator can be a comma or the word 'and', and the numbers can be converted to words. If the array 
      is empty an empty String is returned.
      @param pNums The int array to concatenate.
      @param pAnd true to set the last separator as 'and', false to leave it as a comma.
      @param pWords true to convert the numbers to words before concatenating.
  */
  public static String arCat(int[] pNums, boolean pAnd, boolean pWords) {
    if(pNums == null) return "";
    if(pNums.length < 2) return (pWords?Arithmetic.toWords(pNums[0]):"" + pNums[0]);
    StringBuffer retval = new StringBuffer();
    String lastSep = pAnd?" and":",";
    for(int i = 0; i < pNums.length; i++) {
      retval.append(pWords?Arithmetic.toWords(pNums[i]):" "+pNums[i]);
      if(i < pNums.length - 1) retval.append((i == pNums.length - 2)?lastSep:",");
    }
    return retval.toString();
  }

  /** Concatenate a double array into String, per <code>arCat(int)</code>. */
  public static String arCat(double[] pNums, boolean pAnd, boolean pWords) {
    if(pNums == null) return "";
    if(pNums.length < 2) return (pWords?Arithmetic.toWords("" + pNums[0]):"" + pNums[0]);
    StringBuffer retval = new StringBuffer();
    String lastSep = pAnd?" and":",";
    for(int i = 0; i < pNums.length; i++) {
      retval.append(pWords?Arithmetic.toWords("" + pNums[i]):" "+pNums[i]);
      if(i < pNums.length - 1) retval.append((i == pNums.length - 2)?lastSep:",");
    }
    return retval.toString();
  }

  /** <p>Concatenate a String array into a long String. The element separator and final element separator can be specified as can the direction of the 
      concatenation. If the array sent in is null or empty a blank String is returned.</p>
      <p>If 'pSep' is null it is replaced by a blank String (''). If 'pAnd' is null the value of 'pSep' is used. No other validity checking is done; if either 
      of these Strings is defined that is what is used. If 'pDescending' is true the concatenation starts with the highest element in the array and moves down, 
      otherwise it starts at zero and moves up.</p>
      @param pAry The String[] array to concatenate.
      @param pSep The separator to place between the elements. If null it defaults to a blank.
      @param pAnd The separator to place between the final two elements. If null it defaults to the value of 'pSep'.
      @param pDescending 'true' to start at the top value and move down, 'false' to start at zero and move up.
  */
  public static String arCat(String[] pAry, String pSep, String pAnd, boolean pDescending) {
    if((pAry == null) || (pAry.length < 1)) return "";
    StringBuffer retval = new StringBuffer();
    if(pSep == null) pSep = "";
    if(pAnd == null) pAnd = pSep;
    if(pDescending) {
      for(int i = pAry.length - 1; i >= 0; i--) {
        retval.append(pAry[i]);
        if(i > 0) retval.append((i == 1)?pAnd:pSep);
      }
    } else {
      for(int i=0; i < pAry.length; i++) {
        retval.append(pAry[i]);
        if(i < pAry.length - 1) retval.append((i == pAry.length - 2)?pAnd:pSep);
      }
    }
    return retval.toString();
  }

  /** This method takes an int[] array and boxes it into Integer[]. It is provided in case auto-boxing/unboxing doesn't work. It will always return an array, 
      even if it is empty.
  */
  public static Integer[] boxInt(int[] pAry) {
    Integer[] retval = {};
    if(pAry.length < 1) return retval;
    retval = new Integer[pAry.length];
    for(int i=0; i < retval.length; i++) retval[i] = new Integer(pAry[i]);
    return retval;
  }

  /** This method takes an int[] array and boxes it into Double[]. It is provided in case auto-boxing/unboxing doesn't work. It will always return an array, 
      even if it is empty.
  */
  public static Double[] boxDouble(double[] pAry) {
    Double[] retval = {};
    if(pAry.length < 1) return retval;
    retval = new Double[pAry.length];
    for(int i=0; i < retval.length; i++) retval[i] = new Double(pAry[i]);
    return retval;
  }

  /** <p>Parse an RPN-formatted String and evaluate it numerically. The format for the string is RPN, so the calculation '4+7' would be entered '4 7 +'. The format 
      uses spaces to separate the calculation tokens. This method implements a very simple stack-based calculator and the operations are evaluated as the String 
      is parsed from left to right. Conceptually the calculator has two registers in addition to the Stack: R1 and R2. So the calculation above would be evaluated 
      as follows:
      <ol>
        <li>Push 4 onto the Stack</li>
        <li>Push 7</li>
        <li>Pop 7 into R2</li>
        <li>Pop 4 into R1</li>
        <li>Evaluate R1 + R2</li>
        <li>Push the result (11) onto the Stack</li>
      </ol></p>
      <p>In all cases R1 is the left-hand value and R2 is the right-hand value. That is, when an operation token is encountered, R2 is popped first followed by 
      R1. The operation is evaluated and the result pushed back onto the stack. When all operations are finished, ONE (numerical!) value should remain in the 
      Stack: this value is popped and returned as a String. <b>NOTE:</b> Any errors during the process will leave the Stack in an indeterminate state! If this 
      happens, the return value CANNOT be guaranteed!</p>
      <p>This method has two modes of operation: integer and floating-point. Integer mode is the default and should be used whenever possible. Floating-point mode, 
      as implied by the name, uses floating-point numbers and is subject to all their foibles. Floating-point mode can be explicitly activated by making the first 
      token '.' but it will be triggered by ANY decimal point ANYWHERE in the calculation!</p>
      <p>The operations defined are:
      <ul>
        <li>'+' evaluates: R1 + R2</li>
        <li>'-' evaluates: R1 - R2</li>
        <li>'*' evaluates: R1 * R2</li>
        <li>'/' evaluates: R1 / R2. In Integer mode any remainder is lost.</li>
        <li>'%' evaluates: R1 % R2. Floating-point mode uses the java.lang.math.IEEEremainder() method.</li>
        <li>'^' evaluates: R1 ^ R2. Specifically, java.lang.math.pow(R1, R2), re-cast as int if necessary.</li>
        <li>'neg' pops the Stack, negates it, and pushes the result.</li>
        <li>'abs' pops the Stack, applies absolute value, and pushes the result.</li>
        <li>'PI' pushes java.lang.math.PI onto the stack (floating-point mode only).</li>
        <li>'E' pushes java.lang.math.E onto the stack (floating-point mode only).</li>
        <li>'fix' calls <code>Arithmetic.fix(R1, R2)</code> (floating-point mode only, fix 0 does NOT switch to integer mode).</li>
        <li>'.' switches to floating-point mode.</li>
      </ul></p>
      <p>The following operations work on the last value pushed onto the stack. That number is popped (as R1), the function applied, and the result pushed. These 
      functions are from the <code>java.lang.Math</code> package and they only operate in floating-point mode.
      <ul>
        <li>'sin' = sin(R1). R1 is assumed to be an angle measured in radians.</li>
        <li>'cos' = cos(R1). R1 is assumed to be an angle measured in radians.</li>
        <li>'tan' = tan(R1). R1 is assumed to be an angle measured in radians.</li>
        <li>'sinh' = sinh(R1). Hyperbolic sine.</li>
        <li>'cosh' = cosh(R1).</li>
        <li>'tanh' = tanh(R1).</li>
        <li>'deg' converts R1 from radians to (approximate) degrees.</li>
        <li>'rad' converts R1 from degrees to (approximate) radians.</li>
      </ul></p>
      <p><b>Memory Operations</b> These operations make use of CapCom's Shelf for memory operations that can persist from one method call to another. There are 
      three memory operations and one non-operation character defined:
      <ul>
        <li>'$<i>Label</i>' - This is used to define the label under which a value is stored and later retrieved. It MUST NOT contain spaces!</li>
        <li>'@s' - <b>Store</b> a value under the previously-defined label.</li>
        <li>'@r' - <b>Recall</b> a previously-stored value. The value remains on The Shelf.</li>
        <li>'@d' - <b>Delete</b> a previously-stored value. The value <i>is</i> recalled, per '@r', but it is removed from The Shelf</li>
      </ul>
      The store operation pops the top value off the stack, stores it on The Shelf, and puts it back. It has NO EFFECT on any calculations before or after it; 
      it leaves the stack exactly as it was. The recall and delete operations pull the value off The Shelf (and delete, if '@d') and push it onto the stack. To 
      use these operations first make sure there is something on the stack to store if storing. Next define the label: a valid label consists of at least one 
      non-blank character after the '$'. Finally, immediately after the label (and space) put the operation.</p>
      <p><b>EXAMPLE:</b> The calculation '42 $life @s' stores the number '42' under the label 'Life'. Now '$Life @r' recalls '42' until it is deleted ('@d').</p>
      <br />
      <p>Although this method is useful by itself, its true power comes when combined with the <code>Question.fillTemplate()</code> method. Complex calculations 
      can be pre-formatted with token elements, filled in at runtime and evaluated multiple times. This allows arbitrary computations to be created and altered 
      without recompiling a program! Particularly useful when combined with a scripting language.</p>
      @param pCalc A String containing RPN-formatted calculations.
      @return A String containing the result of evaluating the calculations.
  */
  public static String parseRPN(String pCalc) {
    if(!Question.validString(pCalc)) return null;
    pCalc = pCalc.trim();
    boolean floatMode = (pCalc.indexOf(".") > -1);
    StringTokenizer stChomp = new StringTokenizer(pCalc);
    String retval=null, reg2=null, memLabel = null;
    Stack<String> theStack = new Stack<String>();
    Object oTemp = null;
    theStack.push("ERROR");
    try {
      while(stChomp.hasMoreTokens() && !theStack.empty()) {
        retval = stChomp.nextToken();
        switch(retval.charAt(0)) {
        case '+':
        case '*':
        case '/':
        case '%':
        case '^':
          reg2 = theStack.pop();
          if(floatMode) {
            theStack.push(exEvalFloat(theStack.pop(), reg2, retval.charAt(0)));
          } else {
            theStack.push(exEvalInt(theStack.pop(), reg2, retval.charAt(0)));
          }
          break;
        case '-':
          if(retval.length() == 1) { // Operation: subtraction
            reg2 = theStack.pop();
            if(floatMode) {
              theStack.push(exEvalFloat(theStack.pop(), reg2, '-'));
            } else {
              theStack.push(exEvalInt(theStack.pop(), reg2, '-'));
            }
          } else {
            theStack.push(retval); // Negative number
          }
          break;
        case 'n':
        case 'N':
          if("neg".equals(retval.toLowerCase())) {
            if(floatMode) {
              theStack.push(exEvalFloat(theStack.pop(), "0", 'n'));
            } else {
              theStack.push(exEvalInt(theStack.pop(), "0", 'n'));
            }
          }
          break;
        case 'a':
        case 'A':
          if("abs".equals(retval.toLowerCase())) {
            if(floatMode) {
              theStack.push(exEvalFloat(theStack.pop(), "0", 'a'));
            } else {
              theStack.push(exEvalInt(theStack.pop(), "0", 'a'));
            }
          }
          break;
        case '.':
          floatMode = true;
          if(retval.length() > 1) theStack.push("0"+retval); // Starts with a decimal, make it a number!
          break;
        case 'p':
        case 'P':
          if(floatMode && "pi".equals(retval.toLowerCase())) theStack.push(""+java.lang.Math.PI);
          break;
        case 'e':
        case 'E':
          if(floatMode && (retval.length() == 1)) theStack.push(""+java.lang.Math.E);
          break;
        case 'f':
        case 'F':
          if(floatMode && "fix".equals(retval.toLowerCase())) {
            reg2 = theStack.pop();
            theStack.push(Arithmetic.fix(Double.parseDouble(theStack.pop()), (int) Double.parseDouble(reg2), false));
          }
          break;
        case 's':
        case 'S':
          if(!floatMode) break;
          if("sin".equalsIgnoreCase(retval)) { theStack.push(""+Math.sin(Double.parseDouble(theStack.pop()))); }
          if("sinh".equalsIgnoreCase(retval)) { theStack.push(""+Math.sinh(Double.parseDouble(theStack.pop()))); }
          break;
        case 'c':
        case 'C':
          if(!floatMode) break;
          if("cos".equalsIgnoreCase(retval)) { theStack.push(""+Math.cos(Double.parseDouble(theStack.pop()))); }
          if("cosh".equalsIgnoreCase(retval)) { theStack.push(""+Math.cosh(Double.parseDouble(theStack.pop()))); }
          break;
        case 't':
        case 'T':
          if(!floatMode) break;
          if("tan".equalsIgnoreCase(retval)) { theStack.push(""+Math.tan(Double.parseDouble(theStack.pop()))); }
          if("tanh".equalsIgnoreCase(retval)) { theStack.push(""+Math.tanh(Double.parseDouble(theStack.pop()))); }
          break;
        case 'd':
        case 'D':
          if(!floatMode) break;
          if("deg".equalsIgnoreCase(retval)) { theStack.push(""+Math.toDegrees(Double.parseDouble(theStack.pop()))); }
          break;
        case 'r':
        case 'R':
          if(!floatMode) break;
          if("rad".equalsIgnoreCase(retval)) { theStack.push(""+Math.toRadians(Double.parseDouble(theStack.pop()))); }
          break;
        case '$':
          if(retval.length() > 1) memLabel = retval.substring(1);
          break;
        case '@':
          if(!Question.validString(memLabel) || (retval.length() < 2)) break;
          reg2 = null;
          switch(retval.charAt(1)) {
            case 'S': case 's':
              reg2 = theStack.pop();
              if(Question.validString(reg2)) {
                CapCom.storeObject(MEM_PREFIX+memLabel, reg2);
                CapCom.debug("parseRPN() stored: "+memLabel+" = "+reg2);
              }
              break;
            case 'R': case 'r':
              oTemp = CapCom.getObject(MEM_PREFIX+memLabel);
              if((oTemp != null) && (oTemp instanceof String)) reg2 = (String) oTemp;
              oTemp = null;
              memLabel = null;
              break;
            case 'D': case 'd':
              oTemp = CapCom.removeObject(MEM_PREFIX+memLabel);
              if((oTemp != null) && (oTemp instanceof String)) reg2 = (String) oTemp;
              oTemp = null;
              memLabel = null;
              break;
            default:
          }
          if(Question.validString(reg2)) theStack.push(reg2);
          break;

        default: theStack.push(retval); // If not an op, must be a number
        }
      }
      retval = theStack.pop();
    } catch(Exception ex) {CapCom.error("parseRPN parsing", ex); }
    
    return retval;
  }
  private static String exEvalInt(String lhs, String rhs, char op) {
    long retval = Long.MAX_VALUE - 10, LHS=0, RHS=0;
    try {
      LHS = Long.parseLong(lhs); RHS = Long.parseLong(rhs);
    } catch(Exception ex) { CapCom.error("parseRPN: int mode eval", ex); }
    switch(op) {
      case '+': retval = LHS + RHS; break;
      case '-': retval = LHS - RHS; break;
      case '*': retval = LHS * RHS; break;
      case '/': retval = LHS / RHS; break;
      case '%': retval = LHS % RHS; break;
      case '^': retval = java.lang.Math.round(java.lang.Math.pow(LHS, RHS)); break;
      case 'n': retval = -1 *LHS; break;
      case 'a': retval = java.lang.Math.abs(LHS); break;
      default:
    }
    return ""+retval;
  }
  private static String exEvalFloat(String lhs, String rhs, char op) {
    double retval = Double.MAX_VALUE - 10, LHS=0, RHS=0;
    try {
      LHS = Double.parseDouble(lhs); RHS = Double.parseDouble(rhs);
    } catch(Exception ex) { CapCom.error("parseRPN: float mode eval", ex); }
    switch(op) {
      case '+': retval = LHS + RHS; break;
      case '-': retval = LHS - RHS; break;
      case '*': retval = LHS * RHS; break;
      case '/': retval = LHS / RHS; break;
      case '%': retval = java.lang.Math.IEEEremainder(LHS, RHS); break;
      case '^': retval = java.lang.Math.pow(LHS, RHS); break;
      case 'n': retval = -1 *LHS; break;
      case 'a': retval = java.lang.Math.abs(LHS); break;
      default:
    }
    return ""+retval;
  }
  /** <p>This method creates an algebraic or other mathematical expression based on a variable, a Template, and a set of Knobs. The variable has a default value 
      but the Template and Knobs must be specified. With all elements present, a Question is returned containing the Problem and Answer, properly formatted and 
      correct.</p>
      <p>The Template specifies the presentation and calculation elements of the expression. The basic format is 'Problem @@ <i>calc_1:calc_2: ... :calc_n</i> @@ 
      Answer'. The calculation component may contain Knob tokens or literals: in operation the calculations are fed (as a single unit) into the Question.fillTemplate() 
      method along with the values specified by the Knobs component. The calculation String is then split along the colons. The resulting array then supplies 
      the fillTemplate() method again for both the Problem and Answer components.</p>
      <p>IMPORTANT NOTE: For <b>this method only</b> the Knobs begin at &lt;1&gt;! That is to say, ordinarily the first element in the Knob array is element [0] but 
      for this method ONLY the Knob number and the calculation number are the same! Knob 0 is <i>always</i> zero. In the Problem/Answer part <i>token</i> zero is 
      the variable.</p>
      <p>The exact sequence in which things occur is as follows:
        <ol>
          <li>The Knobs and Template Strings are checked for validity. If either is invalid an error Question is returned.</li>
          <li>The Knobs are evaluated.</li>
          <li>The Template is broken into its components: Problem, calculation, and Answer.</li>
          <li>The calculation component is sent to <code>Question.fillTemplate()</code> with the Knobs array (converted to String).</li>
          <li>The calculation String is split into individual calculations (String[] array split on ':').</li>
          <li>Each calculation is sent to <code>parseRPN()</code> and evaluated.</li>
          <li>A new token array (String[]) is constructed: element[0] is set to the variable and the calculation results are copied starting at element[1].</li>
          <li>The Problem component and the token array are sent to <code>Question.fillTemplate()</code>.</li>
          <li>The Answer component and the token array are sent to <code>Question.fillTemplate()</code>.</li>
          <li>The filled Problem and Answer Strings are formed into a Question.</li>
          <li>The fresh Question is returned.</li>
        </ol></p>
      <p>Example: The template is '&lt;0&gt; - &lt;1&gt; = &lt;2&gt; @@ &lt;1&gt; : &lt;2&gt; : &lt;1&gt; &lt;2&gt; + @@ &lt;0&gt; = &lt;3&gt;'. If the Knobs String 
      is '5 15 20 60' and yields Knob[1]=7 and Knob[2]=41, the calculation String would be '7 : 41 : 7 41 +'. Suppose the variable given is 'Z'. The resulting 
      String array is: token[] = {'Z', '7', '41', '48'}. The 48 comes from parseRPN('7 41 +'.) The filled-in Problem becomes 'Z - 7 = 41' and the Answer will be 
      'Z = 48'. These components are then placed into a Question and returned.</p>
      <p>Any errors in the process result in an error Question being returned. While this method does not itself generate errors, calls to setKnobs(), parseRPN() 
      or fillTemplate() might.</p>
      @param pVar The variable that replaces token 0. If null or blank a default variable ('x') is used.
      @param pKnobs The Knobs to be set for the result.
      @param pTemplate The template containing the parts of the expression.
      @return A Question filled with the elements making up the expression.
  */
  public static Question createExpression(String pVar, String pKnobs, String pTemplate) {
    Question retval = new Question("Err(1)", "Bad Question", "Bad Question");
    if(!Question.validString(pKnobs) || !Question.validString(pTemplate)) return retval;
    if(!Question.validString(pVar)) pVar = "x";
    int[] myKnobs = setKnobs(pKnobs);
    String[] sTokens = new String[myKnobs.length+1];
    sTokens[0]="0";
    for(int i = 0; i < myKnobs.length; i++) { sTokens[i+1] = ""+myKnobs[i]; } // Calculation Knobs
    String[] sParts = pTemplate.split("@@");
    if(sParts.length < 3) return retval;
    sParts[1] = Question.fillTemplate(sParts[1], sTokens);
    String[] sCalcs = sParts[1].split(":");
    sTokens = new String[sCalcs.length +1];
    sTokens[0] = pVar;
    for(int i=1; i < sTokens.length; i++) {
      sTokens[i] = parseRPN(sCalcs[i-1]);
      CapCom.debug("STokens: "+i);
    }
    sParts[0] = Question.fillTemplate(sParts[0], sTokens);
    sParts[2] = Question.fillTemplate(sParts[2], sTokens);
    retval = new Question(sParts[0], sParts[2]);
    return retval;
  }
  /** Return a SCROB instance of this class. */
  public cox.jmatt.java.MathTools.scrob.MathEngine_SCROB getSCROB() { return new cox.jmatt.java.MathTools.scrob.MathEngine_SCROB(); }

  /** Create and return a blank Question. */
  public static Question newQuestion() { return new Question(); }
  /** Create and return an Arithmetic instance. */
  public static Arithmetic newArithmetic() { return new Arithmetic(); }
  /** Create and return a MathConsole anchored to CapCom. */
  public static cox.jmatt.java.MathTools.util.MathConsole newConsole() { return new cox.jmatt.java.MathTools.util.MathConsole(); }
  /** Create and return a Polynomial from a formatting String. String format is per the Polynomial class. */
  public static Polynomial newPolynomial(String pData) { return new Polynomial(pData); }
  /** Create and return a Percent instance. */
  public static Percent newPercent() { return new Percent(); }
  /** Create and return a MFraction. */
  public static MFraction newMFraction() { return new MFraction(); }
  /** Create and return a MathStat instance. */
  public static MathStat newMathStat() { return new MathStat(); }
  /** Create and return a MMatrix. */
  public static MMatrix newMMatrix() { return new MMatrix(); }
  /** Create and return a MMatrixBuilder. */
  public static cox.jmatt.java.MathTools.util.MMatrixBuilder newMMBuilder() { return new cox.jmatt.java.MathTools.util.MMatrixBuilder(); }
  /** Create and return a MathXML() instance. */
  public static MathXML newMathXML() { return new MathXML(); }
  /** Create and return a MEquation() instance. */
  public static MEquation newMEquation() { return new MEquation(); }
  /** Create and return a MRadical() instance. */
  public static MRadical newMRadical() { return new MRadical(); }
  /** Create a MComplex number with the given components. */
  public static MComplex newMComplex(int pReal, int pImag) { return new MComplex(pReal, pImag); }
  /** Create and return a MTXProcessor() instance. */
  public static cox.jmatt.java.MathTools.util.MTXProcessor newMTXPro() { return new cox.jmatt.java.MathTools.util.MTXProcessor(); }
  /** Create and return a blank MTest. */
  public static MTest newMTest() { return new MTest(); }
  /** Create and return a <code>MathTestFormatter</code>. */
  public static MathTestFormatter newMTFormatter() { return new MathTestFormatter(); }
  /** Create a <code>MathToolsAccessory</code>. If accessories are disabled the return value will be null. */
  public static cox.jmatt.java.MathTools.util.MathToolsAccessory getAccessory(String pName) { return CapCom.createAccessory(pName); }
  /** Create a <code>COFormatterFactory</code>. */
  public static COFormatterFactory newCOFormatter() { return new COFormatterFactory(); }
  /** Create a <code>MathDBC</code> instance. */
  public static cox.jmatt.java.MathTools.util.MathDBC newMathDBC() { return new cox.jmatt.java.MathTools.util.MathDBC(); }
  /** Create a new <code>MathGenHTML</code>. */
  public static cox.jmatt.java.MathTools.markup.html.MathGenHTML newMathGenHTML() { return new cox.jmatt.java.MathTools.markup.html.MathGenHTML(); }
  /** Create a new <code>MathDrawSVG</code>. */
  public static cox.jmatt.java.MathTools.markup.svg.MathDrawSVG newSVG() { return new cox.jmatt.java.MathTools.markup.svg.MathDrawSVG(); }
  /** Create a new <code>MathGenLatex</code>. */
  public static cox.jmatt.java.MathTools.markup.latex.MathGenLatex newMLatex() { return new cox.jmatt.java.MathTools.markup.latex.MathGenLatex(); }
  /** Create a new <code>SimpleTagImpl</code>. */
  public static cox.jmatt.java.MathTools.markup.SimpleTagImpl newSimpleTag() { return new cox.jmatt.java.MathTools.markup.SimpleTagImpl(); }
  /** Create a <code>GraphEngineFactory</code>. */
  public static cox.jmatt.java.MathTools.graphing.GraphEngineFactory newGraphEngine() { return cox.jmatt.java.MathTools.graphing.GraphEngineFactory.newInstance(); }
  /** Create a real-number <code>MathSet</code>.*/
  public static MathSet newMathSet() { return MathSet.realNumbers(); }

}
