package cox.jmatt.java.MathTools;

import java.util.*;
import java.io.*;

/** <p>This class acts as an immutable container for sets and provides equipment for formatting questions involving them. Even though the class is named 
    <code>MathSet</code>, it is in truth a container for holding sets and for preserving superset-subset relationships between them. As such it is primarily 
    designed to format and present set/subset questions and not to manipulate them.</p>
    <p>When working with this class, a <b>chain</b> of sets means a collection of contained sets moving from the most general (universal) set ('largest') to the 
    most specific ('smallest') in the group. For instance, the Real number chain is<br /><center> 'R &#8835; Q &#8835; Z &#8835; W &#8835; N'<br /></center> The 
    <i>Universal Set</i> is the ONE superset of all other sets and there can be only one! <code>MathSet</code> automatically takes the first set added as the 
    universal set, so build with care.</p>
    <p>A freshly-created <code>MathSet</code> can have sets added to it but that is all. Once it is closed, either explicitly via the 'closeSet()' method or 
    implicitly by another method that calls it, the container becomes immutable. No more data can be added and all subset and superset relationships are 
    calculated. In addition to sets, a <code>MathSet</code> can have a name and a description. These serve no mathematical purpose and exist only to allow naming 
    and description of an instance.</p>
    <p>Once closed, any set within the <code>MathSet</code> can be <i>selected</i> via the 'select()' method. This method allows access to the selected set's 
    supersets, subsets and formatting symbols. The Universal Set is selected by default.</p>

    <h2>Creating and Adding Sets</h2>
    <p>A set added to a <code>MathSet</code> instance <i>must</i> have two things: a name and a position somewhere within a chain. It <i>may</i> also have two or 
    more symbols associated with it for formatting purposes. All of this is accomplished via <i>Set Descriptor Notation</i> which is used to add sets to a 
    <code>MathSet</code>.</p>
    <p>Set Descriptor Notation uses colons and '@' signs, so these characters may not appear within a set's name. Spaces and other characters are allowed and are 
    ignored. Colons are used to separate set names within a chain and the '@' character is used to delimit a set's symbols:<pre>
    Real:Rational:Integer
    Integer:Whole@W@Whole Numbers
    </pre>
    The first line defines three sets: Real, Rational and Integer. If this is the first descriptor added to the <code>MathSet</code> then 'Real' becomes the 
    Universal Set. The second line defines the Whole numbers as a subset of the Integers and also defines three symbols: the name of the set, the single-letter 
    symbol and the formal name of the set.</p>
    <p>A <code>MathSet</code> can hold as many sets as desired and each set may define as many symbols as needed. When defining symbols, though, the name of the 
    set MUST be the first symbol since that is how symbols are associated with sets. Sets can be added in any order, provided there is at least one link to the 
    other sets somewhere in the data. <code>MathSet</code> does a good job figuring out the order of set chains as long as the links can be deduced from the data 
    sent in.</p>
    <h3>Formatting and <i>Formatting</i></h3>
    <p>The <code>format()</code> methods are where set symbols come into play. When a set is selected, its symbols become the replacement token array for a call 
    to <code>Question.fillTemplate()</code>. The symbols appear in the same order as they are defined. For the set defined above, '&lt;2&gt;' would be replaced 
    with 'Whole Numbers'. This means that token '&lt;0&gt;' always contains the set's name and it will always be defined!</p>
    <h3>Set Pizza</h3>
    <p>The heart of a <code>MathSet</code> is a <code>java.util.Properties</code> object. The 'toPizza()' method closes the set (if necessary) and dumps the 
    contents of this object to a properly-formatted String. The demo panel for this class, <code>MathSetPizza</code>, contains selections to parse descriptors or 
    pizzas into an instance. There are also two selections that will create a <code>MathSet</code> from the demo panel and store it on <code>CapCom</code>'s 
    Shelf, thus allowing script access to a demo-defined instance.</p>
    <p>The pizza generated by a <code>MathSet</code> includes a LOT of work 'under the hood,' so these files should not be edited manually! The 'toPizza()' and 
    'fromPizza()' methods should be employed to allow definition and subsequent re-use of large or complicated chains and trees of sets. An instance created from 
    the demo can be accessed from within a script through <code>Console.getObject(MSet.SHELF_KEY)</code>.</p>

    <h2>Subsets and Supersets</h2>
    <p>This is the primary raison-d'etre for this class. Therefore there are many methods for dealing with them. The 'select()' is the foundational method; it 
    focuses all attention (and subsequent method results) on a particular set. All super- and subsets are in relation to the selected set as are the various 
    formatting methods.</p>
    <h3>Predefined Sets</h3>
    <p>Because they are so common, the Real Numbers and the Complex Numbers come packaged within <code>MathSet</code>. There are two static and two instance 
    methods for creating these sets and each set has five symbols defined on it, including one <code>MathTestFormatter</code>-style token key. These values are 
    also defined as static String constants, so customization of these symbols for specific software package is simplicity itself.</p>
    <h3>My Own Clone</h3>
    <p>Each <code>MathSet</code> instance maintains an internal copy of itself. Method documentation refers to this as the <i>copy</i> or the <i>clone</i> and 
    it is used primarily in the 'createQuestion()' method. For the most part it is inaccessible, directly, but there are some methods usable to manipulate it.</p>
    <p>The internal self-copy is not created until the first method requiring it is called, so any defaults (ie formats) are applied at that time.</p>
*/

public class MathSet {

  /** This String is used to represent the empty set. The literal value is '{}'.*/
  public static final String EMPTY = "{}";

  /** The simplest default format possible: &lt;0&gt;. */
  public static final String DEFAULT_FORMAT = "<0>";

  /** This is the String key used by the <code>MathSet</code> demo panel to store a parsed or pizza'd set on The Shelf. Literal value is 'MathSet_SHELF_KEY'. */
  public static final String SHELF_KEY = "MathSet_SHELF_KEY";

  /** This constant is defined in the Complex Number <code>MathSet</code> symbols as a <code>MathTestFormatter</code> replacement token; i.e. it is bracketed 
      with '${' and '}'. If using these tokens, add the symbol value to the replacement set using this key.
  */
  public static final String SYM_COMPLEX = "COMPLEX_SYMBOL_TOKEN";

  /** Replacement token for purely imaginary numbers. */
  public static final String SYM_IMAGINARY = "IMAGINARY_SYMBOL_TOKEN";

  /** Replacement token for the set of Real numbers. */
  public static final String SYM_REAL = "REAL_SYMBOL_TOKEN";

  /** Replacement token for the set of Rational numbers. */
  public static final String SYM_RATIONAL = "RATIONAL_SYMBOL_TOKEN";

  /** Replacement token for the set of Integers. */
  public static final String SYM_INTEGER = "INTEGER_SYMBOL_TOKEN";

  /** Replacement token for the set of Whole numbers. */
  public static final String SYM_WHOLE = "WHOLE_SYMBOL_TOKEN";

  /** Replacement token for the set of Natural numbers. */
  public static final String SYM_NATURAL = "NATURAL_SYMBOL_TOKEN";

  /** Replacement token for the set of Irrational numbers. */
  public static final String SYM_IRRATIONAL = "IRRATIONAL_SYMBOL_TOKEN";

  /** Debug/testing format used with 'createQuestion()' to display token values. Main set tokens are in this String. */
  public static final String DEBUG_PROBLEM_FORMAT = "<2> hasSuper(<3>): <4>, hasSub(<3>): <6>, randSuper: <8>, randSub: <10>, format(): <0>, randSet: <12>";

  /** Debug/testing format used with 'createQuestion()' to display token values. Clone set tokens are in this String. */
  public static final String DEBUG_ANSWER_FORMAT = "<3> hasSuper(<2>): <5>, hasSub(<2>): <7>, randSuper: <9>, randSub: <11>, format(): <1>, randSet: <13>";

  /** Debugging format for the Question created using the above formats. */
  public static final String DEBUG_QUESTION_FORMAT = "Main Set:\t<1>\nClone Set:\t<2>";
  
  /** Problem template for a simple superset Question involving the selected sets. */
  public static final String SIMPLE_SUPER_PROB = "Is the set of <0> a superset of the set of <1>?";

  /** Answer template for SIMPLE_SUPER_PROB. */
  public static final String SIMPLE_SUPER_ANS = "It <20>.";

  /** Problem template for a simple subset Question using selected sets. */
  public static final String SIMPLE_SUB_PROB = "Is the set of <0> a subset of the set of <1>?";

  /** Answer template for SIMPLE_SUB_PROB. */
  public static final String SIMPLE_SUB_ANS = "It <22>";

  /** This is the delimiter String used when building one LONG data string: ':@:'. */
  public static final String DATA_STRING_DELIMITER = ":@:";

  private static String defaultFormat = DEFAULT_FORMAT;
  private static long safetyValve = 10000;

  private String myUSet = null, myCurrentSet = null, myFormat = defaultFormat, myName = null, myDesc = null, mySupersets = "", myCurrentSymbolString = "";
  private Properties mySets = null;
  private boolean amIClosed = false;
  private HashSet<String> mySetList = new HashSet<String>();
  private String[] myCurrentSymbols = {}, mySubsets = {}, mySetArray = {};
  private HashSet<String> hsSource = new HashSet<String>();
  private MathSet myOwnClone = null;

  private MathSet() {
    mySets = new Properties();
  }
  
  private MathSet(String pUSet) {
    pUSet = Question.validString(pUSet)?pUSet:EMPTY;
    mySets = new Properties();
    setUSet(pUSet);
  }

  private MathSet(Properties pPizza) {
    mySets = new Properties();
    for(String sKey : pPizza.stringPropertyNames()) { mySets.setProperty(sKey, pPizza.getProperty(sKey)); }
    mySetList = new HashSet<String>();
    mySetArray = mySets.getProperty("MathSet.contents", "").split(":");
    for(String sTemp : mySetArray) mySetList.add(sTemp);
    amIClosed = true;
    getName(); getDescription();
    select(getUSet());
  }

  /** Create a <code>MSet</code> using the parameter string supplied. If the pSets is null or blank, the empty set is returned. */
  public static final MathSet createInstance(String pSets) {
    if(!Question.validString(pSets)) return new MathSet(EMPTY).closeSet();
    MathSet retval = new MathSet();
    retval.addSets(pSets);
    return retval;
  }

  /** Instance shadow to create a <code>MathSet</code>. */
  public final MathSet newInstance(String pSets) { return createInstance(pSets); }

  /** Create a <code>MathSet</code> from a Properties pizza. It may be raw String data, loaded from a file or loaded from a resource stream.
      @param pPizza The String containing pizza data OR the name of the file to load.
      @param pIsFile Is 'pPizza' a filename (true) or raw String (false).
      @param pGRAS If true, load the file using 'getResourceAsStream()'. Ignored if 'pIsFile' is false.
  */
  public static final MathSet createFromPizza(String pPizza, boolean pIsFile, boolean pGRAS) {
    MathSet retval = emptySet();
    Properties newPizza = new Properties();
    try {
      if(pIsFile) {
        InputStream isIn = null;
        if(pGRAS) {
          isIn = MathSet.class.getResourceAsStream(pPizza);
        } else {
          isIn = new FileInputStream(pPizza);
        }
        newPizza.load(isIn);
        isIn.close();
      } else {
        StringReader srIn = new StringReader(pPizza);
        newPizza.load(srIn);
        srIn.close();
      }
      retval = new MathSet(newPizza);
    } catch(Exception ex) { CapCom.error("Error creating MathSet from pizza.", ex); }
    return retval;
  }

  /** Instance shadow pizza-creation method. */
  public MathSet fromPizza(String pPizza, boolean pIsFile, boolean pGRAS) { return createFromPizza(pPizza, pIsFile, pGRAS); }

  /** Create an instance from the Global Pizza. This method <b>absolutely requires</b> a valid prefix since it uses <code>MathEngine.bread()</code> to load the 
      data from the Global Pizza.
      @param pPrefix The Global Pizza block parameter prefix used to create an instance.
  */
  public static final MathSet createFromGlobalPizza(String pPrefix) {
    MathSet retval = emptySet();
    if(!Question.validString(pPrefix)) return retval;
    retval = createFromPizza(MathEngine.bread(pPrefix, "\n"), false, false);
    return retval;
  }

  /** Instance Global Pizza creation method. */
  public final MathSet fromGlobalPizza(String pPrefix) { return createFromGlobalPizza(pPrefix); }

  /** <p>Create a (closed!) <code>MathSet</code> containing the number sets Real, Rational, Integer, Whole, Natural and Irrational. Since the Real Numbers and 
      Complex Numbers are two of the most commonly-used sets, both are defined by default.</p>
      <p>In addition to the subset containment, these <code>MathSet</code>s each define symbols for use as replacement tokens. The tokens defined are:
      <ul>
        <li> &lt;0&gt; - The name of the specific set, capitalized and singular. ('Integer').</li>
        <li> &lt;1&gt; - The name of the specific set, capitalized and plural. ('Integers').</li>
        <li> &lt;2&gt; - The name of the specific set, lowercase and singular. ('integer').</li>
        <li> &lt;3&gt; - The name of the specific set, lowercase and plural. ('integers').</li>
        <li> &lt;4&gt; - The set symbol using keyboard characters. ('Z'). Irrational is denoted '~Q' and the purely imaginary numbers, 'I'.</li>
        <li> &lt;5&gt; - The <code>MathTestFormatter</code> replacement token key for the set. (SYM_INTEGER). See below for details.</li>
      </ul></p>
      <p>Symbol (replacement token) &lt;5&gt; is designed for the (likely) event that none of the predefined tokens will work. The token 5 symbol consists of 
      the literal text of the corresponding 'SYM_' constant, bracketed between '${' and '}'. This way, <code>MathTestFormatter</code> can be loaded with the 
      String containing the appropriate symbol-generation characters under the 'SYM_' key. If, for instance, the text to generate the symbol for the Rational 
      numbers is '{{RAT}}', instantiate a <code>MathTestFormatter</code>, call <code>addProperty(MathSet.SYM_RATIONAL, '{{RAT}}')</code> and run the Strings 
      generated by <code>MathSet</code> through the formatter. Scripts access the key as <code>MSet.SYM_RATIONAL</code>.</p>
  */
  public static final MathSet realNumbers() {
    MathSet retval = new MathSet();
    retval.addSets("Real:Rational:Integer:Whole:Natural");
    retval.addSets("Real:Irrational");
    addNumberSets(retval, false);
    retval.setName("Real Numbers");
    retval.setDescription("The set of real numbers.");
    retval.closeSet();
    return retval;
  }

  /** Instance shadow. */
  public MathSet getRealNumbers() { return realNumbers(); }

  /** Create a (closed!) <code>MathSet</code> containing the number sets Complex, Real, Rational, Integer, Whole, Natural and Irrational. It also includes the 
      (Complex!) subset of purely imaginary numbers.
  */
  public static final MathSet complexNumbers() {
    MathSet retval = new MathSet();
    retval.addSets("Complex:Real:Rational:Integer:Whole:Natural");
    retval.addSets("Complex:Real:Irrational");
    retval.addSets("Complex:Imaginary");
    addNumberSets(retval, true);
    retval.setName("Complex Numbers");
    retval.setDescription("The set of complex numbers.");
    retval.closeSet();
    return retval;
  }

  /** Complex numbers instance shadow. */
  public MathSet getComplexNumbers() { return complexNumbers(); }

  /** <p>Create an empty <code>MathSet</code>. It contains no subsets or formatting symbols; it is EMPTY! Each call to this method creates a 'new' empty set, they 
      will NOT test as equal!</p>
      <p>This set is the return value for some creation methods when an error occurs.</p>
  */
  public static final MathSet emptySet() {
    MathSet retval = new MathSet(EMPTY);
    retval.setName("EMPTY");
    retval.setDescription("This set contains nothing!");
    retval.closeSet();
    return retval;
  }

  /** Empty set shadow. */
  public MathSet getEmptySet() { return emptySet(); }

  /* Now add the number sets. */
  private static final void addNumberSets(MathSet pSet, boolean pComplex) {
    if(pSet == null) return;
    if(pComplex) {
      pSet.addSets("Complex@Complex@complex@complex@C@${"+SYM_COMPLEX+"}");
      pSet.addSets("Imaginary@Imaginaries@imaginary@imaginaries@I@${"+SYM_IMAGINARY+"}");
    }
    pSet.addSets("Real@Reals@real@reals@R@${"+SYM_REAL+"}");
    pSet.addSets("Rational@Rationals@rational@rationals@Q@${"+SYM_RATIONAL+"}");
    pSet.addSets("Integer@Integers@integer@integers@Z@${"+SYM_INTEGER+"}");
    pSet.addSets("Whole@Wholes@whole@wholes@W@${"+SYM_WHOLE+"}");
    pSet.addSets("Natural@Naturals@natural@naturals@N@${"+SYM_NATURAL+"}");
    pSet.addSets("Irrational@Irrationals@irrational@irrationals@~Q@${"+SYM_IRRATIONAL+"}");
  }

  /** <p>Add a set or set-chain to the current instance. Each set added is reported at Debug level.</p>
      <p>To allow for text file parsing, this method defines two macro commands and a comment character. They are:
      <ul>
        <li>#N<i>Name</i> - Assign the instance a name.</li>
        <li>#D<i>Description</i> - Set the description.</li>
        <li># - Indicates a comment. Everything on this line is ignored.</i>
      </ul>
      Each special command starts with '#' followed by a single letter. Lowercase 'n' and 'd' will also work. If including comments, leave at least one space 
      after the '#' or the name/description might become corrupted.</p>
      @param pSets The set String to add to the current instance.
      @return true if the sets are added, false otherwise.
  */
  public boolean addSets(String pSets) {
    if(amIClosed || !Question.validString(pSets)) return false;
    boolean retval = true;
    String[] sLines = pSets.split(DATA_STRING_DELIMITER);
    for(int i=0; retval && (i < sLines.length); i++) retval = retval & _addSets(sLines[i]);
    return retval;
  }
  private final boolean _addSets(String pSets) {
    if( (pSets.trim().length() > 2) && (pSets.trim().startsWith("#N") || pSets.trim().startsWith("#n")) ) {
      setName(pSets.trim().substring(2));
      return true;
    }
    if( (pSets.trim().length() > 2) && (pSets.trim().startsWith("#D") || pSets.trim().startsWith("#d")) ) {
      setDescription(pSets.trim().substring(2));
      return true;
    }
    if(pSets.trim().startsWith("#")) return true;
    CapCom.debug("Raw set string: "+pSets);
    String[] theSets = pSets.split(":");
    String setChain = "";
    for(int i=0; i < theSets.length; i++) {
      CapCom.debug("Adding set: "+theSets[i]);
      theSets[i] = addSymbols(theSets[i]);
      setUSet(theSets[i]);
      mySetList.add(theSets[i]);
      if(Question.validString(theSets[i])) setChain = setChain + ":" + theSets[i];
      CapCom.debug("Added set: "+theSets[i]);
    }
    if(setChain.length() > 1) setChain = setChain.substring(1);
    setChain = setChain.trim();
    CapCom.debug("Set chain: "+setChain);
    theSets = setChain.split(":");
    switch(theSets.length) {
    case 0: case 1: break;
    case 2:
      hsSource.add(setChain);
      break;
    default:
      for(int i=0; i < theSets.length - 1; i++) hsSource.add(theSets[i]+":"+theSets[i+1]);
    }
    return true;
  }
  /* Now do the actual work! */
  private String addSymbols(String pSet) {
    if(!Question.validString(pSet)) return "";
    String[] theSet = pSet.split("@");
    if(theSet.length > 1) {
      String symbols = "";
      for(String sym : theSet) { if(Question.validString(sym)) symbols = symbols + "@" + sym; }
      if(symbols.length() > 1) {
        symbols = symbols.substring(1);
        if(Question.validString(symbols)) mySets.setProperty(theSet[0].trim()+".symbols", symbols);
        CapCom.debug(theSet[0]+".symbols: "+symbols);
      }
    } else {
      if(mySets.getProperty(theSet[0]+".symbols", null) == null) mySets.setProperty(theSet[0].trim()+".symbols", theSet[0]);
    }
    return theSet[0].trim();
  }

  /** <p>Close this <code>MathSet</code>. When that happens no new sets may be added and any methods that would normally do so return silently. Returns a
      self-reference for method chaining.</p>
  */
  public final MathSet closeSet() {
    if(amIClosed) return this;
    amIClosed = true;
    StringBuffer sbTemp = new StringBuffer();
    for(String iter : mySetList) sbTemp.append(":").append(iter);
    if((sbTemp.length() > 0) && (sbTemp.charAt(0) == ':')) sbTemp.deleteCharAt(0);
    if(sbTemp.length() > 0) mySets.setProperty("MathSet.contents", sbTemp.toString());
    mySetArray = sbTemp.toString().split(":");
    myCurrentSet = myUSet;
    myCurrentSymbolString = mySets.getProperty(myUSet+".symbols", myUSet);
    myCurrentSymbols = myCurrentSymbolString.split("@");
    calculateSupersetChains();
    calculateSubsets();
    calculateSupersets();
    hsSource = null;
    return this;
  }

  /* Calculate superset chains. */
  private final void calculateSupersetChains() {
    int noSwaps = 0, iNdx = 0;
    long safetyGauge = 0;
    String[] setChains = null, saTemp = null;
    String sHead = null, sTail = null, sBody = null;
    HashSet<String> hsDest = null;
    CapCom.debug("Source size: "+hsSource.size());
    //CapCom.debug("--- Source content ---");
    while((safetyGauge < safetyValve) && (noSwaps < 1000)) {
      hsDest = new HashSet<String>();
      setChains = new String[hsSource.size()];
      iNdx = 0;
      for(Iterator<String> iter = hsSource.iterator(); iter.hasNext(); ) {
        setChains[iNdx] = iter.next();
        //CapCom.debug("Iteration: "+safetyGauge+"\tSetChain["+iNdx+"]: "+setChains[iNdx]);
        iNdx++;
      }
      for(int i=0; i < setChains.length; i++) {
        sBody = setChains[i];
        saTemp = breakString(sBody); sHead = saTemp[0]; sBody = saTemp[1]; sTail = saTemp[2];
        for(int j=0; j < setChains.length; j++) {
          if(i == j) continue;
          if(setChains[j].endsWith(sHead)) {
            sBody = fuseChains(setChains[j], sBody);
            saTemp = breakString(sBody); sHead = saTemp[0]; sBody = saTemp[1]; sTail = saTemp[2];
          } else if(setChains[j].startsWith(sTail)) {
            sBody = fuseChains(sBody, setChains[j]);
            saTemp = breakString(sBody); sHead = saTemp[0]; sBody = saTemp[1]; sTail = saTemp[2];
          }
          hsDest.add(sBody);
        }
      }
      safetyGauge++;
      if(hsDest.size() == hsSource.size()) noSwaps++;
      hsSource = hsDest;
    }
    CapCom.debug("--- Dumping final set chains ---");
    for(Iterator<String> iter = hsSource.iterator(); iter.hasNext(); ) CapCom.debug(iter.next());
    CapCom.debug("Safety gauge: "+safetyGauge);
  }

  /* Break a colon-separated String into the head element, the body and the tail. */
  private final String[] breakString(String pSets) {
    String[] retval = new String[3], saParts = pSets.split(":");
    retval[0] = saParts[0]; retval[1] = pSets; retval[2] = saParts[saParts.length - 1];
    return retval;
  }

  /* Fuse two chains into one. Assumes the last element in LHS is the same as the first in RHS*/
  private final String fuseChains(String LHS, String RHS) {
    String[] saL = LHS.split(":"), saR = RHS.split(":"), saRes;
    StringBuffer retval = new StringBuffer();
    for(int i=0; i < saL.length - 1; i++) retval.append(":").append(saL[i]);
    for(int i=0; i < saR.length; i++) retval.append(":").append(saR[i]);
    if(retval.charAt(0) == ':') retval = retval.deleteCharAt(0);
    return retval.toString();
  }

  /* Calculate subsets. */
  private final void calculateSubsets() {
    String[] saSets = null;
    Hashtable<String, HashSet<String>> theSubsets = new Hashtable<String, HashSet<String>>();
    StringBuffer sbSubsets = new StringBuffer();
    for(String iter : hsSource) {
      saSets = iter.split(":");
      for(int i=0; i < saSets.length - 1; i++) {
        if(!theSubsets.containsKey(saSets[i])) theSubsets.put(saSets[i], new HashSet<String>());
        for(int j = (i+1); j < saSets.length; j++) (theSubsets.get(saSets[i])).add(saSets[j]);
      }
    }
    for(String iter : theSubsets.keySet()) {
      for(String ssets : (theSubsets.get(iter))) sbSubsets.append(":").append(ssets);
      if(sbSubsets.charAt(0) == ':') sbSubsets.deleteCharAt(0);
      mySets.setProperty(iter+".subsets", sbSubsets.toString());
      sbSubsets = new StringBuffer();
    }
  }

  /* Calculate supersets. */
  private final void calculateSupersets() {
    Hashtable<String, HashSet<String>> theSupersets = new Hashtable<String, HashSet<String>>();
    CapCom.debug("--- Calculating subsets ---");
    int lastColon = 0;
    String lastSet = null, superSets = null;
    for(String iter : hsSource) {
      lastColon = iter.lastIndexOf(":");
      superSets = iter;
      while(lastColon > 0) {
        lastSet = (superSets+" ").substring(lastColon+1).trim();
        superSets = superSets.substring(0, lastColon);
        CapCom.debug("Set: "+lastSet+"\tSupersets: "+superSets);
        if(!theSupersets.containsKey(lastSet)) theSupersets.put(lastSet, new HashSet<String>());
        (theSupersets.get(lastSet)).add(superSets);
        lastColon = superSets.lastIndexOf(":");
      }
    }
    for(String iter : theSupersets.keySet()) {
      superSets = "";
      for(String curSet : (theSupersets.get(iter))) {
        if(curSet.length() > superSets.length()) superSets = curSet;
      }
      mySets.setProperty(iter+".supersets", superSets);
    }
  }
  
  /** Test to see whether or not this set is closed. If so, nothing more can be added to it. If not, it can still be edited. */
  public final boolean isClosed() { return amIClosed; }

  /** Return the name of the Universal Set for this <code>MathSet</code>. */
  public String getUSet() {
    if(!Question.validString(myUSet)) myUSet = mySets.getProperty("MathSet.universalSet");
    return myUSet;
  }

  /* Set the universal set. Ignored if it's closed or already set to a valid String. */
  private void setUSet(String pSet) {
    if(amIClosed || Question.validString(myUSet)) return;
    myUSet = pSet;
    if(Question.validString(myUSet)) mySets.setProperty("MathSet.universalSet", myUSet);
  }

  /** Set a short String description of this <code>MathSet</code>. Ignored if null or empty OR the set has been closed.
      @param pDesc The description of what the set it.
  */
  public void setDescription(String pDesc) {
    if(amIClosed || !Question.validString(pDesc)) return;
    mySets.setProperty("MathSet.description", pDesc);
    myDesc = pDesc;
  }

  /** Retrieve the description, if set. If not, a blank String is returned. */
  public String getDescription() {
    if(myDesc == null) myDesc = mySets.getProperty("MathSet.description", "");
    return myDesc;
  }

  /** Give this <code>MathSet</code> a name. This has NO effect on its contents or manipulation, it is meta-data. */
   public void setName(String pName) {
    if(amIClosed || !Question.validString(pName)) return;
    mySets.setProperty("MathSet.name", pName);
    myName = pName;
  }

  /** Get the name, if set. If not, a default name (NOT saved to pizza!) is returned. */
  public String getName() {
    if(myName == null) myName = mySets.getProperty("MathSet.name", null);
    if(myName == null) myName = "__MathSet__"+CapCom.hitCount();
    return myName;
  }

  /** Query as to whether or not this <code>MathSet</code> contains the set specified. The names must match exactly and it must not contain a colon. This method 
      does close the set.
      @param pSet The set to check for existence (presence within this instance).
  */
  public final boolean contains(String pSet) {
    closeSet();
    return (Question.validString(pSet) && mySetList.contains(pSet));
  }

  /** Select one of the sets contained within this <code>MathSet</code> as the current set. If the supplied name is null, blank or not contained, the Universal 
      Set is selected instead. This method DOES close the set!
      @param pSet The set to select. Must exist!
  */
  public MathSet select(String pSet) {
    closeSet();
    if(contains(pSet)) {
      myCurrentSet = pSet;
    } else {
      myCurrentSet = myUSet;
    }
    myCurrentSymbolString = mySets.getProperty(myCurrentSet+".symbols", myCurrentSet);
    myCurrentSymbols = myCurrentSymbolString.split("@");
    mySubsets = mySets.getProperty(myCurrentSet+".subsets", myCurrentSet).split(":");
    mySupersets = mySets.getProperty(myCurrentSet+".supersets", myCurrentSet);
    return this;
  }

  /** Select a particular set in the internal copy. If the internal copy is not yet defined, this method defines it. The instance returned is this one, not the 
      copy.
  */
  public MathSet selectInCopy(String pSet) {
    closeSet();
    makeClone();
    myOwnClone.select(pSet);
    return this;
  }

  /** Return the currently-selected set after closing. */
  public String getSelectedSet() {
    closeSet();
    return myCurrentSet;
  }

  /** Return an array containing all supersets of the currently-selected set. The sets are in order from universal set [0] to closest. The set is closed first. */
  public String[] getSupersets() {
    closeSet();
    return mySupersets.split(":");
  }

  /** Return a colon-separated list of supersets, per <i>Set Descriptor</i> format. Close the set first. */
  public String getSupersetString() {
    closeSet();
    return mySupersets;
  }

  /** Return the symbols for the current set as a single '@'-delimited String, closing if needed. */
  public String getSymbolString() {
    closeSet();
    return myCurrentSymbolString;
  }

  /** Return the current symbols as a String array. This closes the set. */
  public String[] getSymbols() {
    closeSet();
    String[] retval = new String[myCurrentSymbols.length];
    for(int i=0; i < myCurrentSymbols.length; i++) retval[i] = myCurrentSymbols[i];
    return retval;
  }

  /** Close the set and return an array (copied!) containing all subsets of the selected set. There is no particular order to them. */
  public String[] getSubsets() {
    closeSet();
    String[] retval = new String[mySubsets.length];
    for(int i=0; i < mySubsets.length; i++) retval[i] = mySubsets[i];
    return retval;
  }

  /* Check to see if a String array contains a specific element. */
  private final boolean arrayHasElement(String pElement, String[] pAry) {
    if(!Question.validString(pElement) || (pAry == null) || (pAry.length < 1)) return false;
    boolean retval = false;
    for(int i=0; (i < pAry.length) && !retval; i++) retval = pElement.equals(pAry[i]);
    return retval;
  }

  /** Check to see if the selected set has the specified superset. This method closes the set.
      @param pSet The potential superset to check.
  */
  public boolean hasSuperset(String pSet) {
    closeSet();
    return arrayHasElement(pSet, mySupersets.split(":")) || myCurrentSet.equals(pSet);
  }

  /** Check to see if a specified set has a given superset. This differs from the one-argument method in that no selection is required. The internal copy is 
      used to check.
      @param pBaseSet The set whose supersets to check.
      @param pSet The potential superset to check.
  */
  public boolean hasSuperset(String pBaseSet, String pSet) {
    selectInCopy(pBaseSet);
    return myOwnClone.hasSuperset(pSet);
  }

  /** Check to see if the selected set has the specified subset. This method closes the set.
      @param pSet The potential subset to check.
  */
  public boolean hasSubset(String pSet) {
    closeSet();
    return arrayHasElement(pSet, mySubsets) || myCurrentSet.equals(pSet);
  }

  /** Check for subset containment without requiring selection. Uses the clone.
      @param pBaseSet The set whose subsets to check.
      @param pSet The potential subset to check.
  */
  public boolean hasSubset(String pBaseSet, String pSet) {
    selectInCopy(pBaseSet);
    return myOwnClone.hasSubset(pSet);
  }

  /** Select a random superset from the selected set's supersets. This method does close the set. If this set has no supersets it returns itself. */
  public String randomSuperset() {
    closeSet();
    String[] saTemp = mySupersets.split(":");
    return saTemp[MathEngine.random(0, saTemp.length)];
  }

  /** Send a random superset of the currently-selected set to the inner copy. Close and clone this set if necessary. Returns a self-reference. */
  public MathSet randomSupersetToClone() {
    makeClone();
    myOwnClone.select(randomSuperset());
    return this;
  }

  /** Select a random superset from the inner copy and return a self-reference. */
  public MathSet randomSupersetFromClone() {
    makeClone();
    select(myOwnClone.randomSuperset());
    return this;
  }

  /** Return a random subset of the currently-selected set. This method closes the set. If this set has no subsets it returns itself. */
  public String randomSubset() {
    closeSet();
    return mySubsets[MathEngine.random(0, mySubsets.length)];
  }

  /** Select a random subset of this set in the inner copy. Close and clone as needed. Return a self-reference. */
  public MathSet randomSubsetToClone() {
    makeClone();
    myOwnClone.select(randomSubset());
    return this;
  }

  /** Select a random subset from the inner copy. Close, clone, return a self-reference. */
  public MathSet randomSubsetFromClone() {
    makeClone();
    select(myOwnClone.randomSubset());
    return this;
  }

  /** <p>This method works one of two ways, depending on its parameter. If 'sub2sup' is false, a random subset is sent to the inner copy and a random superset 
      is selected from it, thus guaranteeing a random superset-to-subset relationship between this set and the inner copy. If 'sub2sup' is true, a random superset 
      is sent to the inner copy and a random subset of that one selected here, yielding a random subset-to-superset connection.</p>
      <p>This method was designed to provide the ULTIMATE in random, guaranteed set/subset selection. To wit, assume the Strings 'pProb' and 'pAns' contain the 
      Problem and Answer formats for a set-subset question and 'bSubSup' contains the desired 'sub2sup' value. 'Middle Set' is a set close to the middle 
      of the chain and is constant! E.g. 'Integer' for the Complex numbers.<pre>
      mySpecialSet.select('Middle Set').shakeShakeShake(bSubSup).createQuestion(pProb, pAns);
      </pre>
      This simple code snippet creates a Question with random sets and a guaranteed-true 'subset of' or 'superset of' answer!<p>
      @param sub2sup Set to determine which relationship is desired, as described above.
      @return A self-reference.
  */
  public MathSet shakeShakeShake(boolean sub2sup) {
    makeClone();
    if(sub2sup) {
      randomSupersetToClone();
      randomSubsetFromClone();
    } else {
      randomSubsetToClone();
      randomSupersetFromClone();
    }
    return this;
  }

  /** Fetch a random set from ALL the sets contained within this <code>MathSet</code>. */
  public String randomSet() { return mySetArray[MathEngine.random(0, mySetArray.length)]; }

  /** Print some information about the sets contained within this <code>MathSet</code>. Does not close the set but may not display valid information until closure. */
  public final String dumpInfo() {
    StringBuffer retval = new StringBuffer("--- MathSet Meta-Information ---\n");
    retval.append("MathSet name:\t").append(mySets.getProperty("MathSet.name", "Not set")).append("\n");
    retval.append("Description:\t").append(mySets.getProperty("MathSet.description", "Not set")).append("\n");
    retval.append("Universal set:\t").append(mySets.getProperty("MathSet.universalSet", "Not set")).append("\n");
    retval.append("Set contents:\t").append(mySets.getProperty("MathSet.contents", "Not set")).append("\n");
    return retval.toString();
  }

  /** Dump the contents of this <code>MathSet</code> into pizza (<code>Properties</code>) format. This method DOES close the set! */
  public String toPizza() {
    closeSet();
    StringWriter retval = new StringWriter();
    try {
      mySets.store(retval, "MathSet Pizza");
      retval.flush();
      retval.close();
    } catch(Exception ex) { CapCom.error("Error creating MSet pizza.", ex); }
    return retval.toString();
  }

  /** Create and return a copy of this <code>MathSet</code>. This set is closed if necessary. */
  public final MathSet copyOf() {
    closeSet();
    MathSet retval = new MathSet(mySets);
    retval.select(getSelectedSet());
    retval.setFormat(myFormat);
    return retval;
  }

  /** Set a default format. If null or blank, defaults to DEFAULT_FORMAT. */
  public static final void setDefaultFormat(String pFormat) { defaultFormat = Question.validString(pFormat)?pFormat:DEFAULT_FORMAT; }

  /** Instance shadow. */
  public final void setNewDefaultFormat(String pFormat) { setDefaultFormat(pFormat); }

  /** Set the internal format for this <code>MathSet</code>. Null or blank sets to the current default format. */
  public final void setFormat(String pFormat) { myFormat = Question.validString(pFormat)?pFormat:defaultFormat; }

  /** Set the format for the internal copy. This method does create the copy if it isn't already there and it will close the set. */
  public final void setFormatInCopy(String pFormat) {
    closeSet();
    makeClone();
    myOwnClone.setFormat(pFormat);
  }

  /** Convenience method to set the same internal format for both the instance and the internal clone. */
  public final void setFormatInBoth(String pFormat) {
    setFormatInCopy(pFormat);
    setFormat(pFormat);
  }

  /* Create myOwnClone. */
  private final void makeClone() {
    if(myOwnClone == null) myOwnClone = copyOf();
  }

  /** Remove (set to null) the internal clone. Does not close the set, if it is still open. */
  public void unClone() { myOwnClone = null; }

  /** <p>Select a set and format it per the supplied template String. Formatting is per <code>Question.fillTemplate()</code> with the supplied template String 
      and the symbols of the selected set as the replacement array. That is, the replacement token array is the result of splitting the appropriate set 
      descriptor on the character ;@'.</p>
      <p>Since the first action taken is to 'select()' the specified set, this method DOES close the set!</p>
      @param pSet The set to format. If null, blank or absent, the Univeral Set is used.
      @param pTemplate The formatting template. If null or blank, the default format is used.
  */
  public String format(String pSet, String pTemplate) {
    select(pSet);
    pTemplate = (Question.validString(pTemplate)?pTemplate:myFormat);
    return Question.fillTemplate(pTemplate, myCurrentSymbols);
  }

  /** Format the currently-selected set per the supplied template. If 'pTemplate' is null or blank the default is used. */
  public String format(String pTemplate) {
    pTemplate = (Question.validString(pTemplate)?pTemplate:myFormat);
    return Question.fillTemplate(pTemplate, myCurrentSymbols);
  }

  /** Format the currently-selected set per the current internal template. */
  public String format() {return Question.fillTemplate(myFormat, myCurrentSymbols); }

  /** <p>This is <code>MathSet</code>'s power-formatting method. Even though <code>MathSet</code> defines few actual formatting tokens, this method defines 
      many. For parameters, this method accepts the set to select in the main (parent) instance, the set to select in the internal copy (clone), the Problem 
      template and the Answer template for the Question returned.</p>
      <p>The replacement tokens defined are detailed below. The left column is for the parent and the right for the clone. Methods that require a parameter have 
      it supplied from the other instance. For instance, the parent 'hasSubset()' call is applied to the set selected in the clone and vise versa. These tokens 
      have an asterisk between the parentheses in the description.The 'format' method calls are the zero-argument versions. In the two-argument methods, the set 
      on which the method is called supplies the first argument, even though both happen from the parent set.</p>
      <p>Tokens #20 through #27 are derived from the <code>hasSuperset()</code> and <code>hasSubset()</code> tokens, but from the opposite point of view. In these 
      cases the token text reads 'is a superset' or 'is not a superset', meaning the order of the tokens and the sets is more intuitive. For instance, with token 
      #6, 'true' means the parent set does have a <i>subset</i> of the clone's set, which actually means the clone set is the subset! Token #22 is more intuitive; 
      it asserts that the parent set 'is a subset of' the clone set or 'is not a subset of' the clone set. Text-based tokens are supplied for selected and random 
      sets (not random super/sub-sets) and both superset and subset tokens are provided.</p>
      <p><table>
      <tr><td /><td>Main set<br />Main.call(Clone)</td><td>Clone<br />Clone.call(Main)</td></tr>
      <tr><td><code>format()</code></td><td> &lt;0&gt; </td><td> &lt;1&gt; </td></tr>
      <tr><td><code>getSelectedSet()</code></td><td> &lt;2&gt; </td><td> &lt;3&gt; </td></tr>
      <tr><td><code>hasSuperset(*)</code></td><td> &lt;4&gt; </td><td> &lt;5&gt; </td></tr>
      <tr><td><code>hasSubset(*)</code></td><td> &lt;6&gt; </td><td> &lt;7&gt; </td></tr>
      <tr><td><code>randomSuperset()</code></td><td> &lt;8&gt; </td><td> &lt;9&gt; </td></tr>
      <tr><td><code>randomSubset()</code></td><td> &lt;10&gt; </td><td> &lt;11&gt; </td></tr>
      <tr><td><code>randomSet()</code></td><td> &lt;12&gt; </td><td> &lt;13&gt; </td></tr>
      <tr><td><code>#13, #14 format()</code></td><td> &lt;14&gt; </td><td> &lt;15&gt; </td></tr>
      <tr><td><code>hasSuperset(*,*)</code><br />Called on #12 and #13</td><td> &lt;16&gt; </td><td> &lt;17&gt; </td></tr>
      <tr><td><code>hasSubset(*,*)</code><br />Called on #12 and #13</td><td> &lt;18&gt; </td><td> &lt;19&gt; </td></tr>
      <tr><td colspan="3">&quot;_(L)_ is (not) a [sub|super]set of _(R)_&quot;</td></tr>
      <tr><td>#0&#8835;#1?, #1&#8835;#0?</td><td> &lt;20&gt; </td><td> &lt;21&gt; </td></tr>
      <tr><td>#0&#8834;#1?, #1&#8834;#0?</td><td> &lt;22&gt; </td><td> &lt;23&gt; </td></tr>
      <tr><td>#12&#8835;#13?, #13&#8835;#12?</td><td> &lt;24&gt; </td><td> &lt;25&gt; </td></tr>
      <tr><td>#12&#8834;#13?, #13&#8834;#12?</td><td> &lt;26&gt; </td><td> &lt;27&gt; </td></tr>
      
      <tr><td><code>getUSet()</code></td><td> &lt;28&gt; </td><td> N/A </td></tr>
      <tr><td><code>getName()</code></td><td> &lt;29&gt; </td><td> N/A </td></tr>
      <tr><td><code>getDescription()</code></td><td> &lt;30&gt; </td><td> N/A </td></tr>
      </table></p>
      @param pSet The set to select in the main instance.
      @param pCloneSet The set to select in the internal copy.
      @param pProblem The Problem template for the returned Question.
      @param pAnswer The Answer template for the returned Question.
  */
  public Question createQuestion(String pSet, String pCloneSet, String pProblem, String pAnswer) {
    Question retval = new Question();
    selectInCopy(pCloneSet);
    select(pSet);
    pProblem = Question.validString(pProblem)?pProblem:DEBUG_PROBLEM_FORMAT;
    pAnswer = Question.validString(pAnswer)?pAnswer:DEBUG_ANSWER_FORMAT;
    String[] tokenz = _makeFormatTokens();
    retval.setProblem(Question.fillTemplate(pProblem, tokenz));
    retval.setAnswer(Question.fillTemplate(pAnswer, tokenz));
    return retval;
  }

  /** This is the basic <code>Question</code> creation method. It assumes both this set and the inner copy are ready, it simply formats and returns a Question. 
      replacement tokens are per the power-formatting method and only the Problem/Answer formats are required.
      @param pProb The Problem formatting template.
      @param pAns The Answer template.
  */
  public Question createQuestion(String pProb, String pAns) {
    Question retval = new Question();
    pProb = Question.validString(pProb)?pProb:DEBUG_PROBLEM_FORMAT;
    pAns = Question.validString(pAns)?pAns:DEBUG_ANSWER_FORMAT;
    String[] tokenz = _makeFormatTokens();
    retval.setProblem(Question.fillTemplate(pProb, tokenz));
    retval.setAnswer(Question.fillTemplate(pAns, tokenz));
    return retval;
  }

  private final String[] _makeFormatTokens() {
    String[] tokenz = new String[31];
    String mainSet = getSelectedSet(), cloneSet = myOwnClone.getSelectedSet();
    
    tokenz[0] = format(); tokenz[1] = myOwnClone.format(); tokenz[2] = mainSet; tokenz[3] = cloneSet;
    tokenz[4] = ""+hasSuperset(cloneSet); tokenz[5] = ""+myOwnClone.hasSuperset(mainSet); tokenz[6] = ""+hasSubset(cloneSet); tokenz[7] = ""+myOwnClone.hasSubset(mainSet);
    tokenz[8] = randomSuperset(); tokenz[9] = myOwnClone.randomSuperset(); tokenz[10] = randomSubset(); tokenz[11] = myOwnClone.randomSubset();
    tokenz[12] = randomSet(); tokenz[13] = myOwnClone.randomSet(); tokenz[14] = myOwnClone.select(tokenz[12]).format(); tokenz[15] = myOwnClone.select(tokenz[13]).format();
    tokenz[16] = ""+hasSuperset(tokenz[12], tokenz[13]); tokenz[17] = ""+hasSuperset(tokenz[13], tokenz[12]);
    tokenz[18] = ""+hasSubset(tokenz[12], tokenz[13]); tokenz[19] = ""+hasSubset(tokenz[13], tokenz[12]);

    tokenz[20] = (tokenz[5].equalsIgnoreCase("true")?"is ":"is not ")+"a superset"; tokenz[21] = (tokenz[4].equalsIgnoreCase("true")?"is ":"is not ")+"a superset";
    tokenz[22] = (tokenz[7].equalsIgnoreCase("true")?"is ":"is not ")+"a subset"; tokenz[23] = (tokenz[6].equalsIgnoreCase("true")?"is ":"is not ")+"a subset";
    tokenz[24] = (tokenz[17].equalsIgnoreCase("true")?"is ":"is not ")+"a superset"; tokenz[25] = (tokenz[16].equalsIgnoreCase("true")?"is ":"is not ")+"a superset";
    tokenz[26] = (tokenz[19].equalsIgnoreCase("true")?"is ":"is not ")+"a subset"; tokenz[27] = (tokenz[18].equalsIgnoreCase("true")?"is ":"is not ")+"a subset";

    tokenz[28] = getUSet(); tokenz[29] = getName(); tokenz[30] = getDescription();
    return tokenz;
  }

  /** Create a Question of the form 'Is the set of X a [subset, superset] of the set of Y'. The boolean parameter controls the type of question and the parent's 
      selected set is not altered. Both sets are selected at random.
      @param pSuper 'true' for a superset question, 'false' for subset.
  */
  public Question createRandomQuestion(boolean pSuper) {
    String pProb = "Is the set of <14> a subset of the set of <15>?", pAns = "It <26>";
    if(pSuper) {
      pProb = "Is the set of <14> a superset of the set of <15>?";
      pAns = "It <24>.";
    }
    return createQuestion(getSelectedSet(), null, pProb, pAns);
  }

  /** Create a Question using the SIMPLE_?_PROB and SIMPLE_?_ANS formats. This method differs from <code>createRandomQuestion()</code> in that it uses the selected 
      sets rather than random ones.
      @param pSet The set to select in the main instance.
      @param pCloneSet The set to select in the inner copy.
      @param pSuper 'true' for a superset question, 'false' for subset.
  */
  public Question createSimpleQuestion(String pSet, String pCloneSet, boolean pSuper) {
    Question retval = null;
    if(pSuper) {
      retval = createQuestion(pSet, pCloneSet, SIMPLE_SUPER_PROB, SIMPLE_SUPER_ANS);
    } else {
      retval = createQuestion(pSet, pCloneSet, SIMPLE_SUB_PROB, SIMPLE_SUB_ANS);
    }
    return retval;
  }

  /** Create a data String for this <code>MathSet</code>. This String can be one (VERY) long String or a newline-separated group of them. Each set is listed, 
      complete with its superset chain and symbols. In the multiline version each set has its own line. In the single-long-String, the individual set descriptors 
      are separated with the character sequence ':@:'. The set is closed before the data String is constructed.
      @param pLongString Set 'true' for one long String, 'false' for multiline.
  */
  public String dataString(boolean pLongString) {
    closeSet();
    StringBuffer retval = new StringBuffer();
    select(null);
    String sDelim = pLongString?DATA_STRING_DELIMITER:"\n";
    String sTemp = mySets.getProperty("MathSet.name", null);
    if(sTemp != null) retval.append("#N").append(sTemp).append(sDelim);
    sTemp = mySets.getProperty("MathSet.description", null);
    if(sTemp != null) retval.append("#D").append(sTemp).append(sDelim);
    retval.append(getSymbolString()).append(sDelim);
    String[] theSubsets = getSubsets();
    for(int i=0; i < theSubsets.length; i++) {
      select(theSubsets[i]);
      retval.append(getSupersetString()).append(":").append(getSymbolString()).append(sDelim);
    }
    return retval.toString();
  }

}

