package cox.jmatt.java.MathTools;

import java.util.*;
import cox.jmatt.java.MathTools.util.MMatrixBuilder;

/** <p>This class gives MathTools its statistical component. In addition to static calculation methods <code>MathStat</code> also acts as an immutable container 
    for the raw data.</p>
    <p>As a container <code>MathStat</code> can hold up to two data sets represented as double[] arrays. These are, respectively, the primary or first data set 
    and the secondary or second data set. Many of the methods imply X- and Y-coordinates for these data sets but that is NOT the case. Methods and calculations 
    requiring X and Y values do interpret them as such but others do not.</p>
    <p>If a <code>MathStat</code> is created around two data sets there must be two valid data sets defined. If the first data set in the two-argument constructor 
    is null or empty and the second is not, the single data set defined becomes the primary data set and the empty is ignored.</p>
*/
public class MathStat {

  /** This format template is used for debugging and development. NOT an offical format! */
  public static final String FORMAT_DEBUG = "Item:\tDS1\tDS2\nCount:\t<0>\t<1>\nMean\t<2>\t<3>\nP-Var\t<4>\t<5>\ns-Var\t<6>\t<7>\np-SD\t<8>\t<9>\ns-SD\t<10>\t<11>\n" +
        "Corr:\t<12>\nRegr. Slope:\t<13>\nRegr y-int:\t<14>\nRegr eq'n:\t<15>\nData 1:\t<16>\nData 2:\t<17>\n";

  private double[] myDataX = {}, myDataY = {};
  private double myCorrelation = 0, myRegressionSlope = 0;
  private boolean gotCorr = false, gotRegSlope = false;

  /** Standard scripting constructor. */
  public MathStat() { super(); }

  /** Construct a <code>MathStat</code> instance around a single data set (array).
      @param pAry The double[] array containing data values.
  */
  public MathStat(double[] pAry) {
    super();
    if(validArray(pAry)) myDataX = Arrays.copyOf(pAry, pAry.length);
  }

  /** Construct a <code>MathStat</code> instance around two data sets. If pXData is null or empty and pYData is not, pYData becomes the primary (first) data set 
      and the second data set is ignored.
      @param pXData The double[] array containing the first data set.
      @param pYData The array containing the second data set.
  */
  public MathStat(double[] pXData, double[] pYData) {
    super();
    if(validArray(pXData)) myDataX = Arrays.copyOf(pXData, pXData.length);
    if(validArray(pYData)) myDataY = Arrays.copyOf(pYData, pYData.length);
    if(validArray(myDataY) && !validArray(myDataX)) {
      myDataX = myDataY;
      myDataY = new double[0];
    }
  }

  /** Construct a <code>MathStat</code> around up to two Strings of space-separated numbers.
      @param pXData The String containing the first data set.
      @param pYData The String containing the second data set.
  */
  public MathStat(String pXData, String pYData) {
    this(sparse(pXData), sparse(pYData));
  }

  /** Create a new <code>MathStat</code> given up to two data sets.
      @param pXVals The first data set.
      @param pYVals The second data set.
      @return A new MathStat containing the given data.
  */
  public MathStat newInstance(double[] pXVals, double[] pYVals) { return new MathStat(pXVals, pYVals); }

  /** Create a new <code>MathStat</code> given up to two space-separated String data sets.
      @param pXVals The first data set as a String of space-separated numbers.
      @param pYVals The second data set as a String.
      @return A new MathStat containing the given data.
  */
  public MathStat newInstance(String pXVals, String pYVals) { return new MathStat(sparse(pXVals), sparse(pYVals)); }

  /** <p>This method creates a <code>MathStat</code> from a collection of points. The instance automatically has both data sets defined because the input data 
      is in the form of points.</p>
      <p>The input String must be of the form 'x0,y0 x1,y1 x2,y2 ...' where x<i>#</i> goes to the first data set and y<i>#</i> goes to the second. It is critical 
      that there be NO SPACES between the numbers and commas and single spaces between the pairs. This method works by splitting the input first on spaces and 
      the individual elements by comma. The actual parsing is error-protected so one bad number won't choke the entire process but it WILL corrupt the data.</p>
      <p>When parsing the values are parsed in order. If the first item can't be converted both it and the second will become zero. If the first item is parsed 
      as double successfully but the second fails, the first number will be accurate but the second will be zero. Any values absent from the resulting arrays 
      are automatically replaced with zeros.</p>
      <p>Any parse errors are reported at Debug level. If the input String is null or empty an empty <code>MathStat</code> is returned.</p>
      @param pPoints The String containing space-separated point-pairs consisting of two comma-separated numbers.
  */
  public static MathStat fromPoints(String pPoints) {
    if(!Question.validString(pPoints)) return new MathStat();
    String[] pointz = pPoints.split(" "), tokenz = null;
    double[] xVals = new double[pointz.length], yVals = new double[pointz.length];
    Arrays.fill(xVals, 0.0); Arrays.fill(yVals, 0.0);
    for(int i=0; i < pointz.length; i++) {
      if(!Question.validString(pointz[i])) continue;
      tokenz = pointz[i].split(",");
      try {
        xVals[i] = Double.parseDouble(tokenz[0]);
        if(tokenz.length > 1) yVals[i] = Double.parseDouble(tokenz[1]);
      } catch(Exception ex) { CapCom.debug("Error parsing point: "+pointz[i], ex); }
    }
    return new MathStat(xVals, yVals);
  }

  /** Instance version. See static method for details. */
  public MathStat from_Points(String pPoints) { return fromPoints(pPoints); }

  /** Static method to create a <code>MathStat</code> with random values. If 'pYNum' is zero no second set is created. The '_Float' arguments are used for 
      floating-point (double) values.
      @param pNum Number of values in the first data set.
      @param pLow The minimum value for the first data set (inclusive).
      @param pHigh The maximum for the first data set (exclusive).
      @param pFloat true for decimals, false to force integer (as double).
      @param pYNum Number of values in the first data set. Zero for no second data set.
      @param pYLow The minimum value for the first data set (inclusive).
      @param pYHigh The maximum for the first data set (exclusive).
      @param pYFloat true to allow fractions false for int.
  */
  public static MathStat newRandomInstance(int pNum, double pLow, double pHigh, boolean pFloat, int pYNum, double pYLow, double pYHigh, boolean pYFloat) {
    double[] xVals = {}, yVals = {};
    if(pNum < 1) pNum = 1;
    xVals = new double[pNum];
    if(pYNum > 0) yVals = new double[pNum];
    return new MathStat(_randArr(xVals, pLow, pHigh, pFloat), _randArr(yVals, pYLow, pYHigh, pYFloat));
  }

  /** Instance method to generate a <code>MathStat</code> with random values, two data sets. Parameters are per <code>newRandomInstance()</code>. */
  public MathStat newRandom(int pNum, double pLow, double pHigh, boolean pFloat, int pYNum, double pYLow, double pYHigh, boolean pYFloat) {
    return newRandomInstance(pNum, pLow, pHigh, pFloat, pYNum, pYLow, pYHigh, pYFloat);
  }

  /** Convenience instance method for a random single-set <code>MathStat</code>. */
  public MathStat newRandom(int pNum, double pLow, double pHigh, boolean pFloat) {
    return newRandomInstance(pNum, pLow, pHigh, pFloat, 0, 0.0, 1.0, true);
  }

  /** <p>Use this method to create regression problems that <i>should</i> have a fairly high correlation. It generates 'pNum' points around the line from (pX1, pY1) 
      to (pX2, pY2). The 'pDelta' parameter generates a random percentage by which the generated y-values vary from the actual line. The actual y-value is the 
      calculated y-value plus or minus (pDelta/100). The x-values can be forced to be int but the y-values cannot.</p>
      @param pNum The number of data points in the set.
      @param pX1 The x-coordinate of the first endpoint.
      @param pY1 The y-coordinate of the first endpoint.
      @param pX2 The x-coordinate of the second endpoint.
      @param pY2 The y-coordinate of the second endpoint.
      @param pDelta The percent by which the actual y-values vary. If less than 2 it becomes 2.
      @param pFloat 'true' to allow floating-point values, false for int only.
  */
  public static MathStat newLinearInstance(int pNum, double pX1, double pY1, double pX2, double pY2, int pDelta, boolean pFloat) {
    if(pDelta < 2) pDelta = 2;
    if(pNum < 1) pNum = 1;
    double[] xVals = new double[pNum], yVals = new double[pNum];
    xVals = new double[pNum];
    xVals = _randArr(xVals, pX1, pX2, pFloat);
    double dSlope = (pY2 - pY1)/(pX2 - pX1), dInt = pY2 - dSlope*pX2;
    for(int i=0; i < pNum; i++) {
      yVals[i] = (dSlope*xVals[i] + dInt)*((MathEngine.random() < 0.5)?(1.0 - ((double) MathEngine.random(0, pDelta)/100)):(1.0 + ((double) MathEngine.random(0, pDelta)/100)));
    }
    return new MathStat(xVals, yVals);
  }

  /** Instance method to generate a linear regression <code>MathStat</code>. */
  public MathStat newLinear(int pNum, double pX1, double pY1, double pX2, double pY2, int pDelta, boolean pFloat) {
    return newLinearInstance(pNum, pX1, pY1, pX2, pY2, pDelta, pFloat);
  }

  /* Generate a random array. */
  private static double[] _randArr(double[] pVals, double pLow, double pHigh, boolean pFloat) {
    if(!validArray(pVals)) return pVals;
    if(pHigh < pLow) {
      double dTemp = pHigh;
      pHigh = pLow;
      pLow = dTemp;
    }
    if(!pFloat) {
      pVals = MMatrixBuilder.createRandArray(1, pVals.length, (int)pLow, (int) pHigh, true)[0];
    } else {
      double dRange = pHigh - pLow;
      for(int i=0; i < pVals.length; i++) pVals[i] = dRange*MathEngine.random()+pLow;
    }
    return pVals;
  }

  /** Check to make sure the supplied array is not null or empty.
      @param pAry The double[] array to check.
      @return true if the array is defined and at least length 1, false otherwise.
  */
  public static final boolean validArray(double[] pAry) {
    return ((pAry != null) && (pAry.length > 0));
  }

  /** Instance version of <code>validArray()</code>. */
  public final boolean isValidArray(double[] pAry) { return validArray(pAry); }

  /** Find the simple arithmetic mean of a double[] array. If pAry is null or length zero the return value is zero.
      @param pAry The array of values to average.
  */
  public static double simpleMean(double[] pAry) {
    if(!validArray(pAry)) return 0;
    double dSum = 0;
    for(int i=0; i < pAry.length; i++) { dSum += pAry[i]; }
    return dSum / ((double) pAry.length);
  }

  /** This method calculates the variance of the numbers in the array sent in. If the array is null or its length is less than 2 the return value is zero. The 
      variance is calculated as sum((pAry[i] - mean)^2)/N where N is pAry.length or, if pSample is true, N is pAry.length - 1.
      @param pAry The numbers whose variance is to be calculated.
      @param pSample true for sample variance, false for population variance.
      @return The sample or population variance as calculated above.
  */
  public static double variance(double[] pAry, boolean pSample) {
    if((pAry == null) || (pAry.length < 2)) return 0;
    double dMean = simpleMean(pAry), dSum = 0;
    double dSize = (double) (pAry.length - (pSample?1:0));
    for(int i=0; i < pAry.length; i++) { dSum += Math.pow(pAry[i] - dMean, 2); }
    return (dSum / dSize);
  }

  /** This method converts a double[] array into a separated String. If the array is null or empty a blank but non-null String is returned. The separator String 
      can be specified. If null or blank a single space is used. The separator is not appended after the last value.
      @param pAry The array to stringify.
      @param pSep The separator to put between the elements.
  */
  public static String dataString(double[] pAry, String pSep) {
    if(!validArray(pAry)) return "";
    String sSep = Question.validString(pSep)?pSep:" ";
    StringBuffer retval = new StringBuffer();
    for(int i=0; i < pAry.length; i++) {
      retval.append(pAry[i]);
      if(i < pAry.length - 1) retval.append(sSep);
    }
    return retval.toString().trim();
  }

  /** Instance version. */
  public String data_String(double[] pAry, String pSep) { return dataString(pAry, pSep); }

  /** This method splits a String of space-separated doubles and parses them into an array. If any of the resulting Strings does not parse properly it is replaced 
      with 0.0 and the process continues. Any parse errors are reported at the Debug level. If the String sent in is null or empty an empty array is returned. The 
      name is from 'split-parse'.
      @param pVals A String of space-separated numbers parsed in as doubles.
      @return A double[] array consisting of the parsed values.
  */
  public static double[] sparse(String pVals) {
    double[] retval = {};
    if(!Question.validString(pVals)) return retval;
    String[] numz = pVals.split(" ");
    retval = new double[numz.length];
    for(int i=0; i < retval.length; i++) {
      try { retval[i] = Double.parseDouble(numz[i]); } catch(Exception ex) { CapCom.debug("Error parsing double: "+numz[i], ex); }
    }
    return retval;
  }

  /** Instance version of <code>sparse()</code>. NOTE: The instance method capitalizes the 'P'! */
  public double[] sParse(String pVals) { return sparse(pVals); }

  /** Return the number of values in the first data set. */
  public int count() { return myDataX.length; }

  /** Return the number of values in the second data set. */
  public int yCount() { return myDataY.length; }

  /** Return the simple mean of the first data set. */
  public double mean() { return simpleMean(myDataX); }

  /** Calculate the simple mean of the second data set. */
  public double yMean() { return simpleMean(myDataY); }

  /** Calculate the population variance of the first data set. */
  public double pVar() { return variance(myDataX, false); }

  /** Calculate the population variance of the second data set. */
  public double ypVar() { return variance(myDataY, false); }

  /** Calculate the sample variance of the first data set. */
  public double sVar() { return variance(myDataX, true); }

  /** Calculate the sample variance of the second data set. */
  public double ysVar() { return variance(myDataY, true); }

  /** Calculate the population standard deviation of the first data set. */
  public double pDev() { return Math.sqrt(variance(myDataX, false)); }

  /** Calculate the population standard deviation of the second data set. */
  public double ypDev() { return Math.sqrt(variance(myDataY, false)); }

  /** Calculate the sample standard deviation of the first data set. */
  public double sDev() { return Math.sqrt(variance(myDataX, true)); }

  /** Calculate the sample standard deviation of the second data set. */
  public double ysDev() { return Math.sqrt(variance(myDataY, true)); }

  /** <p>This method calculates the correlation between the two data sets. It only works if:
      <ol>
        <li>Both data sets are defined.</li>
        <li>Both data sets are the same size.</li>
        <li>Both data sets have more than 2 values.</li>
      </ol></p>
      <p>If either of these conditions is false the return value is zero.</p>
  */
  public double correlation() {
    if(gotCorr) return myCorrelation;
    if(!validArray(myDataX) || !validArray(myDataY)) return 0;
    if((myDataX.length != myDataY.length) || (myDataX.length < 3)) return 0;
    double dNum = (double) (myDataX.length - 1), sdX = sDev(), sdY = ysDev(), xBar = mean(), yBar = yMean(), dSum = 0;
    for(int i=0; i < myDataX.length; i++) {
      dSum = dSum + ((myDataX[i] - xBar) * (myDataY[i] - yBar));
    }
    myCorrelation = dSum / (dNum * sdX * sdY);
    gotCorr = true;
    return myCorrelation;
  }

  /** Linear regression: calculate the slope of the line of best fit between the points in the data set. Restrictions are per <code>correlation()</code> and the 
      return value for bad data is zero.
  */
  public double regressionSlope() {
    if(gotRegSlope) return myRegressionSlope;
    if(!validArray(myDataX) || !validArray(myDataY)) return 0;
    if((myDataX.length != myDataY.length) || (myDataX.length < 3)) return 0;
    double dNum = (double) myDataX.length, sumX = 0, sumY = 0, sumXY = 0, sumX2 = 0;
    for(int i=0; i < myDataX.length; i++) {
      sumX += myDataX[i]; sumY += myDataY[i];
      sumXY = sumXY + (myDataX[i] * myDataY[i]);
      sumX2 = sumX2 + (myDataX[i] * myDataX[i]);
    }
    myRegressionSlope = ((dNum * sumXY) - (sumX * sumY)) / ((dNum * sumX2) - (sumX * sumX));
    gotRegSlope = true;
    return myRegressionSlope;
  }

  /** Linear regression: calculate the y-intercept of the regression line. This method does call <code>regressionSlope()</code> if necessary. If no regression is 
      possible the return value is zero.
  */
  public double regressionIntercept() {
    if(!gotRegSlope) regressionSlope();
    if(!gotRegSlope) return 0;
    return (yMean() - myRegressionSlope * mean());
  }

  /** Linear regression: calculate and print the slope and y-intercept of the regression line in the form Y = mX + b. */
  public String regressionLine() { return "Y = ("+regressionSlope()+")X + "+regressionIntercept(); }

  /** Return a constructor-like representation of this instance. */
  public String toString() {
    StringBuffer retval = new StringBuffer("MathStat(");
    if(validArray(myDataX)) retval.append(" \"").append(dataString(myDataX, null)).append("\"");
    if(validArray(myDataY)) retval.append(", \"").append(dataString(myDataY, null)).append("\"");
    retval.append(");");
    return retval.toString();
  }

  /** Return a simple XML representation of this class: an empty MathStat tag with a 'data=' and 'data2=' attributes. If there is no secondary data set the second 
      attribute will not be present.
      @param pID The value of the tag's 'id=' attribute. Ignored if null or empty.
      @return A simple XML representation of the data in this class.
  */
  public String toXML(String pID) {
    StringBuffer retval = new StringBuffer("<MathStat ");
    if(Question.validString(pID)) retval.append("id=\"").append(pID).append("\" ");
    if(validArray(myDataX)) retval.append("data=\"").append(dataString(myDataX, null)).append("\" ");
    if(validArray(myDataY)) retval.append("data2=\"").append(dataString(myDataY, null)).append("\" ");
    retval.append("/>");
    return retval.toString();
  }

  /** <p>This is <code>MathStat</code>'s power-formatting method. Every single piece of data, raw or calculated, is available. The formatting process itself uses 
      <code>Question.fillTemplate()</code> with the following tokens:
      <ul>
        <li>&lt;0&gt; is <code>count()</code>, the number of values in the first data set.</li>
        <li>&lt;1&gt; is <code>yCount()</code>, the count for the second data set.</li>
        <li>&lt;2&gt; is <code>mean()</code>, the mean of the first data set.</li>
        <li>&lt;3&gt; is <code>yMean()</code>, the second data set mean.</li>
        <li>&lt;4&gt; is <code>pVar()</code>, the population variance for the first data set.</li>
        <li>&lt;5&gt; is <code>ypVar()</code>, the second set population variance.</li>
        <li>&lt;6&gt; is <code>sVar()</code>, the sample variance for the first set.</li>
        <li>&lt;7&gt; is <code>ysVar()</code>, the second set's sample variance.</li>
        <li>&lt;8&gt; is <code>pDev()</code>, the population standard deviation for the first data set.</li>
        <li>&lt;9&gt; is <code>ypDev()</code>, the population std. deviation for the second set..</li>
        <li>&lt;10&gt; is <code>sDev()</code>, the sample standard deviation on the first data set.</li>
        <li>&lt;11&gt; is <code>ysDev()</code>, the second set sample standard deviation.</li>
        <li>&lt;12&gt; is <code>correlation()</code>, the correlation coefficient between the two data sets.</li>
        <li>&lt;13&gt; is <code>regressionSlope()</code>, the slope of the regression line for both data sets.</li>
        <li>&lt;14&gt; is <code>regressionIntercept()</code>. the y-intercept of the regression line.</li>
        <li>&lt;15&gt; is <code>regressionLine()</code>, the String containing the regression linear equation.</li>
        <li>&lt;16&gt; is <code>dataString(Set_1, pSepX)</code>, data set one separated by pSepX.</li>
        <li>&lt;17&gt; is <code>dataString(Set_2, pSepY)</code>, the data set 2 values separated by pSepY.</li>
      </ul></p>
      <p>Tokens 16 and 17 use, respectively, pSepX and pSepY as separators when concatenating the data values. If pTemplate is null or empty an empty String is 
      returned.</p>
      @param pTemplate The template String for <code>Question.fillTemplate()</code>.
      @param pSepX The separator String to use when concatenating the first data set.
      @param pSepY The separator for the second data set.
  */
  public String format(String pTemplate, String pSepX, String pSepY) {
    if(!Question.validString(pTemplate)) return "";
    String[] tokenz = new String[18];
    tokenz[0] = ""+count(); tokenz[1] = ""+yCount(); tokenz[2] = ""+mean(); tokenz[3] = ""+yMean(); tokenz[4] = ""+pVar(); tokenz[5] = ""+ypVar();
    tokenz[6] = ""+sVar(); tokenz[7] = ""+ysVar(); tokenz[8] = ""+pDev(); tokenz[9] = ""+ypDev(); tokenz[10] = ""+sDev(); tokenz[11] = ""+ysDev();
    tokenz[12] = ""+correlation(); tokenz[13] = ""+regressionSlope(); tokenz[14] = ""+regressionIntercept(); tokenz[15] = regressionLine();
    tokenz[16] = dataString(myDataX, pSepX); tokenz[17] = dataString(myDataY, pSepY);
    return Question.fillTemplate(pTemplate, tokenz);
  }

  /** This <code>format()</code> method accepts two templates and returns a <code>Question</code>. The first template is for the Problem component and the second 
      is for the Answer. The pSepX and pSepY Strings are per <code>format(String, String, String)</code>.
      @param pProb The Problem template String.
      @param pAns The Answer template String.
      @param pSepX The first data set separator.
      @param pSepY The second data set separator.
      @return A Question formed by sending the problem and answer templates through the other <code>format()</code> method.
  */
  public Question format(String pProb, String pAns, String pSepX, String pSepY) {
    return new Question(format(pProb, pSepX, pSepY), format(pAns, pSepX, pSepY));
  }

  /** Create and return a <code>java.util.Map&lt;String, String&gt;</code> of this data. The values entered are per the replacement tokens in <code>format()</code> 
      and the keys are the method names with the following exceptions: the data Strings are 'dsX' and 'dsY' respectively. A prefix may be added to each key but 
      is not required. If something catastrophic happens it is reported at the Debug level and the returned Map will not be reliable.
      @param pPrefx The prefix to add to the key String. Ignored if null.
      @param pSepX The separator for the first data set.
      @param pSepY The separator String for the second data set.
  */
  public Map toMap(String pPrefx, String pSepX, String pSepY) {
    HashMap<String, String> retval = new HashMap<String, String>();
    if(pPrefx == null) pPrefx = "";
    try {
      retval.put((pPrefx+"count").trim(), ""+count()); retval.put((pPrefx+"yCount").trim(), ""+yCount()); retval.put((pPrefx+"mean").trim(), ""+mean());
      retval.put((pPrefx+"yMean").trim(), ""+yMean()); retval.put((pPrefx+"pVar").trim(), ""+pVar()); retval.put((pPrefx+"ypVar").trim(), ""+ypVar());
      retval.put((pPrefx+"sVar").trim(), ""+sVar()); retval.put((pPrefx+"ysVar").trim(), ""+ysVar()); retval.put((pPrefx+"pDev").trim(), ""+pDev());
      retval.put((pPrefx+"ypDev").trim(), ""+ypDev()); retval.put((pPrefx+"sDev").trim(), ""+sDev()); retval.put((pPrefx+"ysDev").trim(), ""+ysDev());
      retval.put((pPrefx+"correlation").trim(), ""+correlation()); retval.put((pPrefx+"regressionSlope").trim(), ""+regressionSlope());
      retval.put((pPrefx+"regressionIntercept").trim(), ""+regressionIntercept()); retval.put((pPrefx+"regressionLine").trim(), regressionLine()); 
      retval.put((pPrefx+"dsX").trim(), dataString(myDataX, pSepX)); retval.put((pPrefx+"dsY").trim(), dataString(myDataY, pSepY)); 
    } catch(Exception ex) { CapCom.debug("Error loading MathStat Map", ex); }
    return retval;
  }

}

