package cox.jmatt.java.MathTools;

import java.util.*;

/** <p>This class provides XML and MathML Presentation markup utilities to the other MathTools classes. It contains quick methods to wrap number, operator, and 
    identifier (variable) tags around single terms, as well as a method to wrap content in &lt;mrow&gt; tags. It contains two methods for generating fraction 
    markup and the wrapMath() method which generates &lt;math&gt; tags complete with the MathML xmlns URI.</p>
    <p>The uber-method here is <code>parseMathML()</code>. It takes a String, optionally containing special formatting characters, and turns it into MathML. 
    It can distinguish between numbers, operators, and identifiers and will tag them appropriately. For multiplication it uses the Unicode 'dot operator' sequence, 
    and for division, the 'division sign' sequence. The <code>wrapXML()</code> method can generate any tagged content and the <code>escapeXML()</code> method 
    replaces certain common characters with their XML entity representations.</p>
    <p>Note on Method Naming: Methods beginning with 'wrap' DO NOT parse their input before formatting. Methods named directly after MathML tags, e.g. mSub, mRoot, 
    DO <code>parseMathML()</code> their parameters before adding them. This is done primarily because the <code>parseMathML()</code> method makes use of some of 
    the 'wrap__()' methods. If subclassing, take this into account!</p>
    <p>The name of this class, MathXML, was chosen to avoid possible conflict and confusion with other objects that might bear the name 'MathML.' This class is not 
    and was not designed to be the end-all XML class. Rather, it provides sufficient XML/MathML to support the other MathTools classes. As with them, MathXML 
    contains only static methods but provides a constructor for scripting applications.</p>
*/
public class MathXML {

  /** <p>This <code>enum</code> is used to hold Unicode math symbols and to provide them and their values in an easily accessible way. The names are the ones 
      Unicode uses for the symbols with some of them slightly abbreviated. For letter symbols (e.g. Greek) the '_LC' denotes the lowercase letter, unadorned is 
      uppercase.</p>
  */
  public static enum MathSymbol {

    /** This is the 'does not exist' character (backward 'E' with a slash) but is also used for <b>unknown</b>. */
    NOT_EXIST("2204"),
    FOR_ALL("2200"),
    INTEGRAL("222B"),
    INFINITY("221E"),
    UNION("222A"),
    INTERSECTION("2229"),
    /** Large uppercase sigma. */
    N_ARY_SUM("2211"),
    /** Large uppercase PI. */
    N_ARY_PRODUCT("220F"),
    CIRCLE_PLUS("2295"),
    CIRCLE_MINUS("2296"),
    CIRCLE_TIMES("2297"),
    CIRCLE_SLASH("2298"),
    CIRCLE_DOT("2299"),
    SUBSET("2282"),
    SUPERSET("2283"),
    NOT_SUBSET("2284"),
    NOT_SUPERSET("2285"),
    ELEMENT("2208"),
    NOT_ELEMENT("2209"),
    PLUS_MINUS("00B1"),
    GAMMA("0393"),
    DELTA("0394"),
    ALPHA_LC("03B1"),
    BETA_LC("03B2"),
    DELTA_LC("03B4"),
    THETA_LC("03B8"),
    LAMBDA_LC("03BB"),
    MU_LC("03BC"),
    PI_LC("03C0"),
    SIGMA_LC("03C3"),
    RADICAL("221A"),
    CUBE_ROOT("221B"),
    FOURTH_ROOT("221C"),
    NATURAL_NUMS("2115"),
    INTEGER_NUMS("2124"),
    RATIONAL_NUMS("211A"),
    REAL_NUMS("211D"),
    COMPLEX_NUMS("2102"),
    DIVIDE_BY("00F7");

    private int myValue = 0;
    MathSymbol(String pValue) {
      try { myValue = Integer.parseInt(pValue, 16); } catch(Exception ignored) {}
    }
    /** Fetch the decimal value of the symbol. */
    public int decValue() { return myValue; }
    /** Fetch the symbol as type char. */
    public char charValue() { return ((char) myValue); }
  }

  /* * * * * END OF ENUM * * * * */

  /** String containing the MathML namespace URI. */
  public static final String NSURI = "http://www.w3.org/1998/Math/MathML";

  /** Constructor provided for exposing to scripts. */
  public MathXML() { super(); }

  /** Replace certain characters with their XML escapes. This method replaces &lt;, &gt;, &amp;, &quot; and &apos; only. If pData is null or empty it is returned 
      unaltered.
      @param pData The String to XML-escape.
      @return The same String with all dangerous characters replaced.
  */
  public static String escapeXML(String pData) {
    if(!Question.validString(pData)) return pData;
    String retval = pData.replaceAll("&", "&amp;");
    retval = retval.replaceAll(">", "&gt;");
    retval = retval.replaceAll("<", "&lt;");
    retval = retval.replaceAll("\"", "&quot;");
    retval = retval.replaceAll("'", "&apos;");
    return retval;
  }
  /** Look up a <code>MathSymbol</code> constant based on a String. This is an error-save way to do a lookup; if the symbol is not found the result is the 
      constant 'NOT_EXIST'. Any errors are reported at Debug level.
      @param pSymbol The name of the MathSymbol to retrieve. It must EXACTLY MATCH the name of the symbol.
      @return The requested symbol or 'NOT_EXIST' if it could not be found.
  */
  public static MathSymbol getSymbol(String pSymbol) {
    MathSymbol retval = MathSymbol.NOT_EXIST;
    try {
      retval = MathSymbol.valueOf(pSymbol);
    } catch(Exception ex) { CapCom.debug("Exception finding: "+pSymbol, ex); }
    return retval;
  }
  /** This method returns a Unicode/XML entity escape for the requested symbol. Basically it wraps the (decimal) value between '&amp;#' and a semicolon.
      @param pSymbol The symbol to look up.
      @return The XML-entity value as a String.
  */
  public static String getEntity(String pSymbol) { return "&#"+(getSymbol(pSymbol).decValue())+";"; }
  /** Wrap the &lt;math&gt; tags around MathML content. The content itself is not escaped, so that should be done before calling this method. The full 'xmlns=' 
      namespace URI is supplied. If pContent is null or empty an empty tag with a unique ID attribute is returned. The display should be 'block', 'inline', or 
      null. If null it is ignored, otherwise it is added exactly as it is passed. Use with care!
      @param pContent XML content to be wrapped in math tags.
      @param pDisplay The 'display=' attribute: 'inline', 'block', or null.
      @return The content wrapped in &lt;math&gt; tags.
  */
  public static String wrapMath(String pContent, String pDisplay) {
    if(!Question.validString(pContent)) return "<math xmlns=\"http://www.w3.org/1998/Math/MathML\" id=\"mt"+CapCom.hitCount()+"\" />";
    String sDisp = (Question.validString(pDisplay)?" display=\""+pDisplay+"\" ":" ");
    return "<math xmlns=\"http://www.w3.org/1998/Math/MathML\""+sDisp+">\n  "+pContent+"\n</math>";
  }

  /** Create an XML tag around the supplied content. If the tag name is null or empty the method returns null or empty. If the content is null or empty it returns 
      an empty tag of the type specified with a unique ID attribute. All wrap_() methods except <code>wrapMath()</code> echo back to this one.
      @param pTag The name of the tag to create.
      @param pContent The content to include in the tag.
  */
  public static String wrapTag(String pTag, String pContent) {
    if(!Question.validString(pTag)) return pTag;
    if(!Question.validString(pContent)) return "<"+pTag+" id=\""+pTag+CapCom.hitCount()+"\" />";
    return "<"+pTag+"> "+pContent+" </"+pTag+">";
  }

  /** Wrap something in number tags, &lt;mn&gt;. This is the MathML tag used to enclose a number. */
  public static String wrapMN(String pData) { return wrapTag("mn", pData); }

  /** Wrap something in identifier tags, &lt;mi&gt;. This is the MathML tag used to enclose an identifier (variable). */
  public static String wrapMI(String pData) { return wrapTag("mi", pData); }

  /** Wrap something in operator tags, &lt;mo&gt;. This is the MathML tag used to enclose an operation symbol. */
  public static String wrapMO(String pData) { return wrapTag("mo", pData); }

  /** Wrap a math row, &lt;mrow&gt;. This is used for multiple terms that should be written in a horizontal row. */
  public static String wrapRow(String pData) { return wrapTag("mrow", pData); }

  /** <p>This is MathXML's power method. It parses a String containing one or more formatting characters and generates a MathML expression. The data String is 
      broken down into space-separated tokens and parsed from left to right. The MathML is generated dynamically as the tokens are iterated.</p>
      <p>The formatting characters are:
      <ul>
        <li>@r/@R Open/close an &lt;mrow&gt; tag.</li>
        <li>@f/@F Open/close an &lt;mfrac&gt; tag.</li>
        <li>@q/@Q Open/close an &lt;msqrt&gt; tag.</li>
        <li>@c/@C Open/close an &lt;mroot&gt; tag.</li>
        <li>@@ Literal '@ sign; used to escape '@'.</li>
        <li>@&lt; Literal tag delimiter: append everyting except the '@' sign. Beware of spaces when embedding tags!</li>
        <li>@= Literal content. Append everything after the '=' sign.</li>
        <li>@_ Literal space. Append one space character. (The character is the underscore.)</li>
        <li>&amp; Interpreted as a Unicode character or XML-escaped character. Wrapped in &lt;mo&gt; tags.</li>
        <li>$$ Append a <b>raw</b> <code>MathSymbol</code> escape based on the <i>String</i> value after the '$' sign.</li>
        <li>$i Append a <code>MathSymbol</code> escape as an <b>identifier</b>.</li>
        <li>$o Append a <code>MathSymbol</code> escape as an <b>operator</b>.</li>
        <li>0..9 Interpreted as a number and enclosed in &lt;mn&gt; tags.</li>
        <li>+ Interpreted as the addition operator.</li>
        <li>*, / Interpreted as operators but given Unicode characters 'dot operator' and 'division sign.'</li>
        <li>- (Only) Interpreted as the operation of subtraction.</li>
        <li>- (In front of a number) Interpreted as unary negative of the number.</li>
        <li>=, &lt;, &gt; Interpreted as operators.</li>
        <li>[, (, {, |, and their closes are interpreted as operators.</li>
      </ul>
      Tokens that start with a letter are assumed to be identifiers (variables).</p>
      <p>If the input String is null or empty an empty &lt;mrow&gt; tag with a unique ID is returned.</p>
      <p><b>CAUTION:</b>Using more than one space to separate the tokens may confuse the parser!</p>
      @param pData The data String containing raw data to be parsed.
      @return MathML per formatting characters and the rules cited above.
  */
  public static String parseMathML(String pData) {
    if(!Question.validString(pData)) return wrapTag("mrow", "");
    StringBuffer retval = new StringBuffer();
    String[] tokens = pData.split(" ");
    for(int i=0; i < tokens.length; i++) {
      tokens[i] += " ";
      switch(tokens[i].charAt(0)) {
        case '@':
          switch(tokens[i].charAt(1)) {
            case 'r': retval.append("<mrow>"); break;
            case 'R': retval.append("</mrow>"); break;
            case 'f': retval.append("<mfrac>"); break;
            case 'F': retval.append("</mfrac>"); break;
            case 'q': retval.append("<msqrt>"); break;
            case 'Q': retval.append("</msqrt>"); break;
            case 'c': retval.append("<mroot>"); break;
            case 'C': retval.append("</mroot>"); break;
            case '@': retval.append("@"); break;
            case '<': retval.append(tokens[i].substring(1)); break;
            case '=': retval.append(tokens[i].substring(2)); break;
            case '_': retval.append(" "); break;
            default: // switch on '@' at 1
          }
          break;
        case '$':
          switch(tokens[i].charAt(1)) {
            case '$': retval.append(getEntity(tokens[i].substring(2).trim())).append(" "); break;
            case 'i': retval.append("<mi>").append(getEntity(tokens[i].substring(2).trim())).append("</mi>"); break;
            case 'o': retval.append("<mo>").append(getEntity(tokens[i].substring(2).trim())).append("</mo>"); break;
            default:
          }
          break; // switch on '$' at 1
        case '&':
          retval.append(wrapMO(tokens[i]));
          break;
        case'0': case'1': case'2': case'3': case'4': case'5': case'6': case'7': case'8': case'9':
          retval.append(wrapMN(tokens[i]));
          break;
        case '+': retval.append("<mo>+</mo>"); break;
        case '-':
          switch(tokens[i].charAt(1)) {
            case ' ': retval.append("<mo>-</mo>"); break;
            case'0': case'1': case'2': case'3': case'4': case'5': case'6': case'7': case'8': case'9':
              retval.append(wrapMN(tokens[i]));
              break;
            default: // switch on '-' at 1
              retval.append(wrapMI(tokens[i]));
          }
          break;
        case '*': retval.append("<mo>&#8901;</mo>"); break;
        case '/': retval.append("<mo>&#247;</mo>"); break;
        case '=': case '<': case '>': case '(': case '{': case '[': case ']': case '}': case ')': case '|':
          retval.append(wrapTag("mo", tokens[i]));
          break;
        default: // switch at 0
          retval.append(wrapMI(tokens[i]));
      }
    }
    return retval.toString();
  }
  /** <p>This method generates an &lt;mfrac&gt; expression. If the whole part is not null or empty, the resulting expression is in a row. If so, just the &lt;mfrac&gt; 
      tag is generated. All three parts are sent to <code>parseMathML()</code> before being included: if one of the components should be in a row it must be enclosed 
      in '@r' and '@R' to ensure proper parsing. The whole part need not be 'rowed' as it is automatically a part of a mrow tag.</p>
      <p>NOTE: This <i>method</i> should not be confused with the 'MFraction' <i>class</i>. The former is a MathXML method for formatting fractions per MathML 
      Presentation specification. The latter is a MathTools class encompassing the behavior AND presentation of fractions.</p>
      @param pWhole The whole part of the fraction. May be null or empty.
      @param pNumerator The numerator of the fraction.
      @param pDenominator The denominator of the fraction.
      @return A freshly-created mfrac expression.
  */
  public static String mFrac(String pWhole, String pNumerator, String pDenominator) {
    StringBuffer retval = new StringBuffer();
    if(Question.validString(pWhole)) retval.append("<mrow>"+parseMathML(pWhole));
    retval.append("<mfrac>").append(parseMathML(pNumerator)).append(parseMathML(pDenominator)).append("</mfrac>");
    if(Question.validString(pWhole)) retval.append("</mrow>");
    return retval.toString();
  }
  /** Convenience method for mFrac(null, pNumerator, pDenominator). */
  public static String mFrac(String pNumerator, String pDenominator) { return mFrac(null, pNumerator, pDenominator); }

  /** Superscript a term. Both operands are fed to <code>parseMathML()</code> before processing, so parser markup is allowed. If the superscript is null or empty 
      it is ignored and only pBase is parsed.
      @param pBase The base to which the superscript is attached.
      @param pSuper The superscript to attach.
      @return A properly parsed and formatted &lt;msup&gt; element.
  */
  public static String mSup(String pBase, String pSuper) {
    if(!Question.validString(pSuper)) return parseMathML(pBase);
    return ("<msup> "+parseMathML(pBase)+parseMathML(pSuper)+" </msup>");
  }

  /** Subscript a term. Both operands are fed to <code>parseMathML()</code> before processing, so parser markup is allowed. If the subscript is null or empty it 
      is ignored and only pBase is parsed.
      @param pBase The base to which the subscript is attached.
      @param pSub The subscript to attach.
      @return A properly parsed and formatted &lt;msub&gt; element.
  */
  public static String mSub(String pBase, String pSub) {
    if(!Question.validString(pSub)) return parseMathML(pBase);
    return ("<msub> "+parseMathML(pBase)+parseMathML(pSub)+" </msub>");
  }

  /** Create a &lt;msubsup&gt; element. If either the subscript or superscript is null or empty the other is sent to the appropriate method. All parameter Strings 
      are parsed before adding so parser markup will be valid.
      @param pBase The base to which the super- and subscript are attached.
      @param pSub The subscript to attach. Ignored if null or blank.
      @param pSuper The superscript to attach. Ignored if null or blank.
      @return A fresh &lt;msubsup&gt; element.
  */
  public static String mSubSup(String pBase, String pSub, String pSuper) {
    if(!Question.validString(pSub)) return mSup(pBase, pSuper);
    if(!Question.validString(pSuper)) return mSup(pBase, pSub);
    return ("<msub> "+parseMathML(pBase)+parseMathML(pSub)+parseMathML(pSuper)+" </msub>");
  }

  /** Format a monomial (single polynomial term). This method operates one of two ways. In 'raw' mode everything present is a part of the result. The only things 
      omitted are things that are null or blank. In 'smart' mode exponents of 1 are excluded and the exponent (0) <i>and</i> variable are excluded from the 
      constant term. Neither mode encloses the monomial in &lt;mrow&gt; tags as it is assumed the <u>mono</u>mial is in fact part of a <u>poly</u>nomial which 
      will be wrapped in its entirety. All entries are parsed before being tagged, so parser markup is supported.
      @param pCoef The coefficient of the term.
      @param pVar The variable of the term. If it is null or blank both it and the exponent are ignored.
      @param pExp The exponent of the variable. If this can't be parsed as an int, the method reverts to 'raw' mode.
      @param pSmart true to activate 'smart' mode, false for 'raw' mode.
      @return A freshly-formatted and parsed monomial term.
  */
  public static String formatMonomial(String pCoef, String pVar, String pExp, boolean pSmart) {
    StringBuffer retval = new StringBuffer();
    if(Question.validString(pCoef)) retval.append(parseMathML(pCoef));
    if(!pSmart) {
      if(Question.validString(pVar)) retval.append(mSup(pVar, pExp));
      return retval.toString();
    }
    int iExp = 10; // Force the issue, if in doubt.
    try { iExp = Integer.parseInt(pExp); } catch(Exception ignored) {}
    if(iExp == 1) {
      retval.append(parseMathML(pVar));
    } else if(iExp == 0) {
      // Nothing to do here. Constant term's already there.
    } else {
      if(Question.validString(pVar)) retval.append(mSup(pVar, pExp));
    }
    return retval.toString();
  }

  /** Overscript a term. Both operands are fed to <code>parseMathML()</code> before processing, so parser markup is allowed. If the overscript is null or empty 
      it is ignored and only pBase is parsed.
      @param pBase The base to which the overscript is attached.
      @param pOver The overscript to attach.
      @return A properly parsed and formatted &lt;mover&gt; element.
  */
  public static String mOver(String pBase, String pOver) {
    if(!Question.validString(pOver)) return parseMathML(pBase);
    return ("<mover> "+parseMathML(pBase)+parseMathML(pOver)+" </mover>");
  }

  /** Underscript a term. Both operands are fed to <code>parseMathML()</code> before processing, so parser markup is allowed. If the underscript is null or empty 
      it is ignored and only pBase is parsed.
      @param pBase The base to which the underscript is attached.
      @param pUnder The underscript to attach.
      @return A properly parsed and formatted &lt;munder&gt; element.
  */
  public static String mUnder(String pBase, String pUnder) {
    if(!Question.validString(pUnder)) return parseMathML(pBase);
    return ("<munder> "+parseMathML(pBase)+parseMathML(pUnder)+" </munder>");
  }

  /** Under- and overscript a term. All operands are fed to <code>parseMathML()</code> before processing, so parser markup is allowed. If either *script is null 
      or empty the other is processed normally.
      @param pBase The base to which the underscript is attached.
      @param pUnder The underscript to attach.
      @return A properly parsed and formatted &lt;munderover&gt; element.
  */
  public static String mUnderOver(String pBase, String pUnder, String pOver) {
    if(!Question.validString(pUnder)) return mSup(pBase, pOver);
    if(!Question.validString(pOver)) return mSup(pBase, pUnder);
    return ("<msub> "+parseMathML(pBase)+parseMathML(pUnder)+parseMathML(pOver)+" </msub>");
  }

  /** Create a &lt;msqrt&gt; tag. The radicand is parsed but does NOT need to be 'rowed' since the tag allows multiple children. */
  public static String mSqrt(String pRadicand) {
    return wrapTag("msqrt", parseMathML(pRadicand));
  }

  /** Create a &lt;mroot&gt; tag. Both parameters are parsed and MUST be rowed if they contain more than one element.
      @param pRadicand The radicand of the resulting radical.
      @param pIndex The value of the root to be taken.
      @return A freshly-formed &lt;mroot&gt; tag.
  */
  public static String mRoot(String pRadicand, String pIndex) {
    return wrapTag("mroot", (parseMathML(pRadicand)+parseMathML(pIndex)));
  }

  /** Return a SCROB instance of this class. */
  public cox.jmatt.java.MathTools.scrob.MathXML_SCROB getSCROB() { return new cox.jmatt.java.MathTools.scrob.MathXML_SCROB(); }

}

