package cox.jmatt.java.MathTools;

/** <p>The Percent class encapsulates all the machinery necessary to calculate simple percents. The methods in this class require terminology to use. In all cases 
    involving percent problems the following three words are used consistently:
    <ul>
      <li>The <b>base</b> is the number before the percent is applied. It is the starting number, or what the percent is applied to.</li>
      <li>The <b>rate</b> is the percent being applied. This class uses <i>whole</i> numbers for percent rates, not decimals.</li>
      <li>The <b>amount</b> is the number after the percent is applied. It is the ending number, or the result of applying the percent.</li>
    </ul>
    As stated previously, any time a percent rate is supplied it is assumed to be in integer form. There is no need to convert to decimal values as this is 
    handled automatically! The methods in this class are static and ALL are convenience methods! While percent calculations are not hard to do, doing a lot of 
    them at one time, especially when making up test questions, can be tedious and time-consuming. The Percent class is designed to ease this burden.</p>
    <p>Two types of methods make up this class: quick-solve convenience methods and power-formatting methods. The latter solve for the (one) unknown quantity in 
    a percent if given the other two. If any of the values is not valid it is returned as <code>Double.NaN</code>. The power-formatting methods take arguments 
    of Object type <code>Double</code> to allow null for the unknown value. That is the value that will be calculated and then fed into a String template via 
    ten pre-defined formatting characters. The result is broken into a Problem and Answer and returned as a Question.</p>
    <p>The <b>change</b> methods (<code>solveChange()</code> and <code>formatChange()</code>) are used to calculate a <i>percent change</i>. The numeric values 
    are ordered as given: the first value is the starting value and the second the ending value. The percent change is calculated from that and the sign determines 
    whether it is a percent increase (positive) or percent decrease (negative). As with other methods the percent change is given as an integral percent and not 
    a decimal. NOTE: The fix value of the percent change is separate from the other fix values!</p>
*/
public class Percent {

  /** The simplest form of percent question: '&lt;6&gt; is &lt;1&gt;&lt;3&gt; of &lt;0&gt;? @@ &lt;2&gt; is &lt;1&gt;% of &lt;0&gt;.' */
  public static final String SIMPLE_QUESTION = "<6> is <1><3> of <0>? @@ <9> is <8>% of <7>.";

  /** The simplest form of percent change question: 'What is the percent change from &lt;0&gt; to &lt;1&gt; @@ A(n) &lt;3&gt; percent &lt;4&gt;.' */
  public static final String SIMPLE_CHANGE_QUESTION = "What is the percent change from <0> to <1>? @@ A(n) <3> percent <4>.";

  private static int nFixBase = -1, nFixRate = -1, nFixAmount = -1;
  private static boolean nGroup = false;

  /** Standard constructor, useful for scripting. */
  public Percent() { super(); }

  /** Static factory method for creating new Percents. */
  public static Percent createPercent() { return new Percent(); }
  /** Instance method for a new Percent. */
  public Percent newPercent() { return new Percent(); }

  /** Solve for the Amount, given the Base and the Rate. */
  public static double solveAmount(double pBase, double pRate) { return (pBase * pRate) / 100.0; }

  /** Solve for the Rate, given the Base and the Amount. */
  public static double solveRate(double pBase, double pAmount) { return (pAmount * 100.0) / pBase; }

  /** Solve for the Base, given the Rate and the Amount. */
  public static double solveBase(double pRate, double pAmount) { return (pAmount * 100.0) / pRate; }

  /** Solve for the percent change, int version. The calculation is (pSecond - pFirst) / pFirst, multiplied by 100. If something goes terribly wrong the return 
      value is <code>Double.NaN</code>. The sign of the return value determines whether the change is a percent increase (positive) or percent decrease 
      (negative).
      @param pFirst The starting number (the number that changes).
      @param pSecond The ending number (what pFirst changes into).
      @return The percent change or <code>NaN</code> if something terrible happens).
  */
  public static double solveChange(int pFirst, int pSecond) {
    double retval = 0;
    try {
      retval = 100.0 * (pSecond - pFirst)/pFirst;
    } catch(Exception ex) { return Double.NaN; }
    return retval;
  }

  /** Calculate percent change on two doubles. Parameters and return values are per the int-argument method. */
  public static double solveChange(double pFirst, double pSecond) {
    double retval = 0;
    try {
      retval = 100.0 * (pSecond - pFirst)/pFirst;
    } catch(Exception ex) { return Double.NaN; }
    return retval;
  }

  /** Instance version of <code>solveChange(int, int)</code>. */
  public double solve_Change(int pFirst, int pSecond) { return solveChange(pFirst, pSecond); }

  /** Instance version of <code>solveChange(double, double)</code>. */
  public double solve_Change(double pFirst, double pSecond) { return solveChange(pFirst, pSecond); }

  /** Instance version with modified name. */
  public double solve_Amount(double pBase, double pRate) {  return (pBase * pRate) / 100.0; }

  /** Instance version with modified name. */
  public double solve_Rate(double pBase, double pAmount) { return (pAmount * 100.0) / pBase; }

  /** Instance version with modified name. */
  public double solve_Base(double pRate, double pAmount) { return (pAmount * 100.0) / pRate; }

  /** <p>This is the power-method of the Percent class. It takes a Question.fillTemplate()-style formatting String, fills in information, and returns a Question. 
      The basic format of the String is 'Problem @@ Answer' with the following tokens filled in to the data array:
      <ul>
        <li>&lt;0&gt; is the Base, or the word 'what' if the Base is unknown.</li>
        <li>&lt;1&gt; is the Rate, or the word 'what' if the Rate is unknown.</li>
        <li>&lt;2&gt; is the Amount, or the word 'what' if the Amount is unknown.</li>
        <li>&lt;3&gt; is the symbol '%' if the Rate is known, or the word 'percent' if the Rate is unknown.</li>
        <li>&lt;4&gt; is the Base, or the word 'What', capitalized, if the Base is unknown.</li>
        <li>&lt;5&gt; is the Rate, or the word 'What' if the Rate is unknown.</li>
        <li>&lt;6&gt; is the Amount, or the word 'What' if the Amount is unknown.</li>
        <li>&lt;7&gt; is the Base, always.</li>
        <li>&lt;8&gt; is the Rate, always.</li>
        <li>&lt;9&gt; is the Amount, always.</li>
        <li>&lt;10&gt; is Base + Amount, always.</li>
        <li>&lt;11&gt; is Base - Amount, always.</li>
      </ul></p>
      <p>The true power and flexibility of this method come from the parameters supplied. The three basic numeric quantities are all given as Double (Object version!) 
      to allow one of them to be set to null. Whichever parameter is set to null is the parameter that will be solved! The data tokens for template replacement 
      are listed above, and the resulting String is then broken down into Problem and Answer parts and returned as a Question.</p>
      <p>NOTE: Exactly <b>one</b> of pBase, pRate, and pAmt must be null! Also, the template breaks on '@@' and this is not configurable. If pTemplate is null or 
      empty SIMPLE_QUESTION is used. If either component of the Template is blank or there is no '@@' the Answer of the resulting Question will be 'Ans?'. If 
      more than one required value is null, the returned Question is blank.</p>
      <p>The formatting process allows the number of decimal places in each quantity to be specified. The fix values are explicitly supplied here: a value of -1 
      results in no place-fixing for that quantity.</p>
      @param pBase The value of the Base, or null if it is unknown.
      @param pRate The value of the Rate, or null if it is unknown.
      @param pAmt The value of the Amount, or null if it is unknown.
      @param pTemplate The 'Problem @@ Answer' String template fed to Questionl.fillTemplate().
      @param pFixBase The number of decimal places allowed for the Base, or -1 to ignore.
      @param pFixRate The number of decimal places allowed for the Rate, or -1 to ignore.
      @param pFixAmt The number of decumal places allowed for the Rate, or -1 to ignore.
      @param pGroup Determine whether or not grouping is enabled for any 'fixed' numbers.
      @return A freshly-built Question with all necessary components filled in.
  */
  public static Question format(Double pBase, Double pRate, Double pAmt, String pTemplate, int pFixBase, int pFixRate, int pFixAmt, boolean pGroup) {
    Question retval = new Question();
    int howMany = 0;
    howMany += (pBase == null)?1:0;
    howMany += (pRate == null)?1:0;
    howMany += (pAmt == null)?1:0;
    if(howMany > 1) return retval;
    if(!Question.validString(pTemplate)) pTemplate = SIMPLE_QUESTION;
    String[] sTokens = new String[12];
    pTemplate += " @@ No Data. @@ No Answer!";
    if(pBase != null) {
      sTokens[0] = pBase.toString();
      sTokens[4] = pBase.toString();
      sTokens[7] = pBase.toString();
    } else {
      sTokens[0] = "what";
      sTokens[4] = "What";
      sTokens[7] = Arithmetic.fix(solveBase(pRate, pAmt), pFixBase, pGroup);
    }
    if(pRate != null) {
      sTokens[1] = pRate.toString();
      sTokens[5] = pRate.toString();
      sTokens[8] = pRate.toString();
      sTokens[3] = "%";
    } else {
      sTokens[1] = "what";
      sTokens[5] = "What";
      sTokens[8] = Arithmetic.fix(solveRate(pBase, pAmt), pFixRate, pGroup);
      sTokens[3] = "percent";
    }
    if(pAmt != null) {
      sTokens[2] = pRate.toString();
      sTokens[6] = pRate.toString();
      sTokens[9] = pRate.toString();
    } else {
      sTokens[2] ="what";
      sTokens[6] = "What";
      sTokens[9] = Arithmetic.fix(solveAmount(pBase, pRate), pFixAmt, pGroup);
    }
    sTokens[10] = "" + (Double.parseDouble(sTokens[7]) + Double.parseDouble(sTokens[9]));
    sTokens[11] = "" + (Double.parseDouble(sTokens[7]) - Double.parseDouble(sTokens[9]));
    String sTemp = Question.fillTemplate(pTemplate, sTokens);
    CapCom.debug("Filled template: "+sTemp);
    String[] quesParts = sTemp.split("@@");
    //CapCom.debug("quesParts array length: "+quesParts.length);
    if(quesParts.length > 1) retval.setData(quesParts[0], quesParts[1]);
    //CapCom.debug("Question prob: "+retval.getProblem()+" Ans: "+retval.getAnswer());
    return retval;
  }

  /** Instance version of the full-power format() method. */
  public Question formatPercent(Double pBase, Double pRate, Double pAmt, String pTemplate, int pFixBase, int pFixRate, int pFixAmt, boolean pGroup) {
    return format(pBase, pRate, pAmt, pTemplate, pFixBase, pFixRate, pFixAmt, pGroup);
  }

  /** Format per the current 'fix' values. */
  public static Question format(Double pBase, Double pRate, Double pAmt, String pTemplate) { return format(pBase, pRate, pAmt, pTemplate, nFixBase, nFixRate, nFixAmount, nGroup); }

  /** Format per the current 'fix' values, instance version. */
  public Question formatPercent(Double pBase, Double pRate, Double pAmt, String pTemplate) { return format(pBase, pRate, pAmt, pTemplate); }

  /** <p>Formatting method for percent change problems. All parameters are specified. If anything goes wrong a Debug-level message is sent and the return value 
      is a blank Question. The starting and ending numbers are cast as <code>Number</code> to allow use of the int-argument methods. This happens only if both 
      numbers are of type Integer.</p>
      <p>The Question is formed per the standard <code>Percent.format()</code> method: the answer token is hard-coded '@@' and the template is in Problem / 
      Answer form. The replacement tokens available to this method are:
      <ul>
        <li>&lt;0&gt; is the beginning number (pFirst), unaltered.</li>
        <li>&lt;1&gt; is the ending number (pSecond), unaltered.</li>
        <li>&lt;2&gt; is the signed percent change, fixed per the number specified.</li>
        <li>&lt;3&gt; is the absolute value of the percent change.</li>
        <li>&lt;4&gt; is either 'increase' or 'decrease' depending on the sign of the percent change.</li>
      </ul></p>
      @param pFirst The first (changing) value.
      @param pSecond The second (ending) value.
      @param pTemplate The template String to format into a Question. If null or blank, SIMPLE_CHANGE_QUESTION is used.
      @param pFixFirst The decimal places for pFirst.
      @param pFixSecond The decimal places to fix in pSecond or -1.
      @param pFixChange The number of decimal places to allow in the percent change, or -1 not to fix at all.
      @param pGroup Set whether or not to use grouping if fixing the numbers.
      @return A freshly-created Question based on the template sent in or a blank Question if something goes horribly wrong.
  */
  public static Question formatChange(Number pFirst, Number pSecond, String pTemplate, int pFixFirst, int pFixSecond, int pFixChange, boolean pGroup) {
    Question retval = new Question();
    if((pFirst == null) || (pSecond == null)) return retval;
    if(!Question.validString(pTemplate)) pTemplate = SIMPLE_CHANGE_QUESTION;
    try {
      double dChange = solveChange(pFirst.doubleValue(), pSecond.doubleValue());
      boolean firstInt = false;
      if((pFirst instanceof Integer) && (pSecond instanceof Integer)) dChange = solveChange(pFirst.intValue(), pSecond.intValue());
      String[] tokenz = new String[5];
      tokenz[0] = Arithmetic.fix(pFirst.doubleValue(), pFixFirst, pGroup); tokenz[1] = Arithmetic.fix(pSecond.doubleValue(), pFixSecond, pGroup);
      tokenz[2] = Arithmetic.fix(dChange, pFixChange, pGroup); tokenz[3] = Arithmetic.fix(Math.abs(dChange), pFixChange, pGroup);
      tokenz[4] = (dChange < 0)?"decrease":"increase";
      pTemplate = Question.fillTemplate(pTemplate, tokenz);
      tokenz = pTemplate.split("@@");
      if(tokenz.length == 2) retval.setData(tokenz[0], tokenz[1]);
    } catch(Exception ex) { CapCom.debug("Error formatting percent change.", ex); }
    return retval;
  }

  /** Percent change formatting method using internal fix values: Base for pFirst, Amount for pSecond and Rate for the percent change. */
  public static Question formatChange(Number pFirst, Number pSecond, String pTemplate) { return formatChange(pFirst, pSecond, pTemplate, nFixBase, nFixAmount, nFixRate, nGroup); }

  /** Instance version of the full percent change <code>format()</code> method. */
  public Question format_Change(Number pFirst, Number pSecond, String pTemplate, int pFixFirst, int pFixSecond, int pFixChange, boolean pGroup) {
    return formatChange(pFirst, pSecond, pTemplate, pFixFirst, pFixSecond, pFixChange, pGroup);
  }

  /** Instance version of the short percent change formatting method. */
  public Question format_Change(Number pFirst, Number pSecond, String pTemplate) { return formatChange(pFirst, pSecond, pTemplate, nFixBase, nFixAmount, nFixRate, nGroup); }

  /** Set the default number of places to fix values for the Base, Rate and Amount. Once set these values are in effect until they are changed. Setting a value 
      to -1 turns off fixing for that element.
      @param pBase Decimal places to fix in the Base.
      @param pRate Decimal places to fix in the Rate.
      @param pAmt Decimal places to fix in the Amount.
      @param pGroup Enable/disable grouping.
  */
  public static void setFixValues(int pBase, int pRate, int pAmt, boolean pGroup) { nFixBase = pBase; nFixRate = pRate; nFixAmount = pAmt; nGroup = pGroup; }

  /** Instance access to setFixValues(). */
  public void setNewFixValues(int pBase, int pRate, int pAmt, boolean pGroup) { setFixValues(pBase, pRate, pAmt, pGroup); }

  /** This method is designed to convert between a (whole) percent and a decimal. The parameters are type Double to allow (exactly!) one to be null. If both 
      are null the result is '0.0'. If both are defined pDecimal is converted to a percent.
      @param pPercent The percent value to convert to decimal.
      @param pDecimal The decimal value to convert to percent.
      @param pFix The number of decimal places to fix in the answer. Ignored if -1
  */
  public static String convertRate(Double pPercent, Double pDecimal, int pFix) {
    if((pPercent == null) && (pDecimal == null)) return "0.0";
    double retval = 0.0;
    if(pDecimal == null) {
      retval = pPercent.doubleValue()/100.0;
    } else {
      retval = pDecimal.doubleValue() * 100.0;
    }
    return (pFix > -1)?Arithmetic.fix(retval, pFix, false):("" + retval);
  }
  /** Instance rate converter. */
  public String convert_Rate(Double pPercent, Double pDecimal, int pFix) { return convertRate(pPercent, pDecimal, pFix); }

}

