package cox.jmatt.java.MathTools;

import java.util.*;

/** <p>This class encapsulates polynomial behavior. It provides methods to add, subtract and multiply polynomials. Once created a Polynomial is immutable. Only 
    the names of the variable and the function have setters and they do not affect the actual polynomial, only its presentation.</p>
    <p>One critical aspect to using Polynomial is understanding its formatting String. At its heart this String lists the polynomial coefficients from left to 
    right IN DESCENDING ORDER! So the String '1 0 3 -6' would be the polynomial x^3 + 3x - 6'. To specify the variable to use when formatting the Polynomial, 
    prefix it to the format String followed by a colon. E.g. the polynomial cited above would be: 'x:1 0 3 -6'. Finally, to specify the name of the function used 
    when formatting in function notation, precede the variable with the name of the function followed by a colon. So, 'P:z:2 6 9' would, if expressed using function 
    notation, be 'P(z) = 2z^2 + 6z + 9'. Both the variable and function names are optional, but if only one is present it is taken to be the variable. So, to 
    specify a function name in the format String you must also specify the variable name.</p>
*/
public class Polynomial {

  /** Typical Polynomial format using <code>makeFormatTokens()</code> tokens: &lt;12&gt; &lt;8&gt;&lt;13&gt;&lt;17&gt;&lt;14&gt;. */
  public static final String FORMAT_TYPICAL = "<12> <8><13><17><14> ";
  /** Typical Polynomial format using <code>makeFormatTokens()</code> tokens, OpenOffice version. */
  public static final String FORMAT_OO_TYPICAL = "<12> <8>{<13>}^{<14>}~";
  /** LaTeX term format. All intelligence is applied: the lead coefficient is blank if +1, '-' if -1, visible otherwise; no exponent on first-degree or constant 
      term; no variable on constant term.
  */
  public static final String FORMAT_LATEX = " <12> <8>{<13>}^{<14>}";

  private static String DEFAULT_FORMAT = FORMAT_TYPICAL;
  private static String defaultCaret = "^";
  private static String defaultVar = "x";
  private static String defaultFName = "f";

  private int[] myCoefs = {0};
  private String myVar = defaultVar;
  private String myFName = defaultFName;
  private String myFormat = DEFAULT_FORMAT;

  /** The default constructor. It returns the zero polynomial and is useful only for creating other Polynomials. */
  public Polynomial() { super(); }
  /** Create a Polynomial from an int[] array. This array becomes the coefficients. If the array is null or of length zero the zero polynomial is created.
      @param pCoefs An int array containing the coefficients of the new Polynomial.
  */
  public Polynomial(int[] pCoefs) {
    if((!(pCoefs == null)) && (pCoefs.length > 0)) {
      myCoefs = Arrays.copyOf(pCoefs, pCoefs.length);
    }
  }
  /** Construct a Polynomial from a createPloynomial() format String. */
  public Polynomial(String pData) {
    Polynomial pTemp = createPolynomial(pData);
    myCoefs = Arrays.copyOf(pTemp.coefficients(), pTemp.coefficients().length);
    myVar = pTemp.getVariable();
    myFName = pTemp.getFName();
  }
  /** Copy constructor. If pPoly is null the zero polynomial is created. */
  public Polynomial(Polynomial pPoly) {
    if(pPoly != null) {
      myCoefs = pPoly.coefficients();
      myVar = pPoly.getVariable();
      myFName = pPoly.getFName();
    }
  }
  /** <p>This method takes a specially-formatted String and converts it into a polynomial. The format of the String is 'F:V:Cn Cn-1 ... C1 C0' where F is the name 
      of the function, V is the variable, and Cn ... C0 is the space-separated list of coefficients. If any coefficient cannot be converted to int, it is replaced 
      silently with 0. Both the function name and variable are optional; if there are only two components, however, they are assumed to be the variable and the 
      coefficient list. If the format String is null or empty the zero Polynomial is returned.</p>
      <p>This is a static factory method. It is the most common way to construct a Polynomial.</p>
      @param pData The Polynomial data String.
      @return A freshly-constructed Polynomial.
  */
  public static Polynomial createPolynomial(String pData) {
    Polynomial retval = new Polynomial();
    if(!Question.validString(pData)) return retval;
    String[] sTokens = pData.split(":");
    String sVar = defaultVar, sFName = defaultFName;
    int iNdx = 0;
    int[] newCoefs = {0};
    switch(sTokens.length) {
      case 3:
        sFName = sTokens[iNdx++];
      case 2:
        sVar = sTokens[iNdx++];
      case 1:
        sTokens = sTokens[iNdx].split(" ");
        newCoefs = new int[sTokens.length];
        iNdx = 0;
        for(int i = newCoefs.length - 1; i >= 0; i--) {
          try { newCoefs[i] = Integer.parseInt(sTokens[iNdx]); } catch(Exception ignored) { newCoefs[i] = 0; }
          iNdx++;
        }
        break;
      default:
    }
    retval = new Polynomial(newCoefs).truncate();
    retval.setVariable(sVar).setFName(sFName);
    return retval;
  }
  /** This is the instance version of the static createPolynomial() method. */
  public Polynomial newPolynomial(String pData) { return createPolynomial(pData); }

  /** Create an instance from an int array. */
  public static Polynomial createPolynomial(int[] coefs) { return new Polynomial(coefs); }
  /** Instance version of <code>createPolynomial(int[])</code>. */
  public Polynomial newPolynomial(int[] coefs) { return new Polynomial(coefs); }

  /** Return a copy of this <code>Polynomial</code>. All data is the same but the instances themselves are NOT! */
  public Polynomial copyOf() { return new Polynomial(this); }

  /** Construct a random Polynomial using a KNOBS String. The variable and function can be set per the standard Polynomial format string but the coefficients 
      conponent is fed to MathEngine.setKnobs()!
      @param pData A Polynomial-compliant formatting string BUT with the coefficients interpreted as a MathEngine setKnobs() string.
      @return A Polynomial with random coefficients.
  */
  public static Polynomial createPolyKnobs(String pData) {
    Polynomial retval = new Polynomial();
    if(!Question.validString(pData)) return retval;
    String[] sTokens = pData.split(":");
    String sVar = defaultVar, sFName = defaultFName;
    int iNdx = 0;
    int[] newCoefs = {0};
    switch(sTokens.length) {
      case 3:
        sFName = sTokens[iNdx++];
      case 2:
        sVar = sTokens[iNdx++];
      case 1:
        newCoefs = MathEngine.setKnobs(sTokens[iNdx]);
        break;
      default:
    }
    int[] revCoefs = new int[newCoefs.length];
    iNdx = newCoefs.length;
    for(int i=0; i < revCoefs.length; i++) { revCoefs[i] = newCoefs[--iNdx]; }
    retval = new Polynomial(revCoefs).truncate();
    retval.setVariable(sVar).setFName(sFName);
    return retval;
  }
  /** This is the instance version of createPolyKnobs(). */
  public Polynomial newPolyKnobs(String pData) { return createPolyKnobs(pData); }

  /** This is the only method that can alter an already-created Polynomial. It removes any leading zero coefficients and resizes the coefficient array. It returns 
      a reference to this polynomial. Most methods that might generate extraneous zeros call this one automatically but it is here if needed.
  */
  public Polynomial truncate() {
    if(myCoefs.length == 1) return this;
    int iHigh = 0;
    for(int i=0; i < myCoefs.length; i++) { iHigh = (myCoefs[i] != 0)?i:iHigh; }
    if(iHigh == myCoefs.length - 1) return this;
    int[] temp = Arrays.copyOf(myCoefs, iHigh + 1);
    myCoefs = Arrays.copyOf(temp, temp.length);
    return this;
  }
  /** Get the variable used for the Polynomial. The default is 'x.' */
  public String getVariable() { return myVar; }
  /** <p>Set the variable used for this polynomial. A null or blank value is ignored and the variable is set to the default ('x') OR the last variable set in 
      any previous Polynomial. That is, setting the variable sets the new default value which remains until specifically changed. A self-reference is returned 
      to allow method chaining.</p>
  */
  public Polynomial setVariable(String pNew) {
    if(Question.validString(pNew)) defaultVar = pNew;
    myVar = defaultVar;
    return this;
  }
  /** Get the function name used for the Polynomial. The default is 'f.' */
  public String getFName() { return myFName; }
  /** <p>Set the name of the function used when this Polynomial is printed in function form. Setting this to a non-null, non-empty value sets a new default value 
      for all subsequent instances.</p>
  */
  public Polynomial setFName(String pNew) {
    if(Question.validString(pNew)) defaultFName = pNew;
    myFName = defaultFName;
    return this;
  }

  /** The <i>caret</i> is the character used to designate an exponent. It defaults to '^' but can be changed. This method sets the default caret for all 
      instances. If pNewCaret is null or blank it resets to '^'. This is the value used in the caret and smart caret formatting tokens; there is no per-instance 
      caret.
  */
  public static void setCaret(String pNewCaret) { defaultCaret = Question.validString(pNewCaret)?pNewCaret:"^"; }
  /** Instance shadow. */
  public void setNewCaret(String pNewCaret) { setCaret(pNewCaret); }

  /* ########## Formatting methods begin here ########## */

  /** This method generates the construction String used to create instances and in <code>toString()</code> and <code>toXML()</code>. */
  public String dataString() {
    StringBuffer retval = new StringBuffer(getFName().trim()).append(":").append(getVariable().trim()).append(":");
    for(int i = myCoefs.length - 1; i >=0; i--) retval.append(myCoefs[i]).append(" ");
    return retval.toString().trim();
  }

  /** <p>This method formats the terms of the Polynomial but returns them as a String array. From there custom formatting and arrangement can be performed. The 
      replacement tokens are per the full <code>format()</code> method, all terms are included, and ascending/descending is not relevant. If pTerm is null or 
      blank the default is used.</p>
      <p>CRITICAL NOTE: This method has been updated to use <code>makeFormatTokens()</code> tokens.</p>
      @param pTerm The formatting String to be used for each term.
      @return A String[] array consisting of the terms formatted per the template sent in.
  */
  public String[] formatTerms(String pTerm) {
    pTerm = (Question.validString(pTerm))?pTerm:DEFAULT_FORMAT;
    String[] retval = new String[myCoefs.length];
    for(int i=0; i < retval.length; i++) {
      retval[i] = Question.fillTemplate(pTerm, makeFormatTokens(myCoefs[i], i, myVar, myFName, false));
    }
    return retval;
  }

  /** <p>This is the standard Polynomial formatting method. It is designed to format the Polynomial only, any 'decorations' must be added afterward. It may be 
      printed in ascending or descending order and with zero terms included or omitted. The formatting tokens available are:
      <ul>
        <li>&lt;0&gt; - The explicit variable.</li>
        <li>&lt;1&gt; - The polynomial/function name.</li>
        <li>&lt;2&gt; - The explicit coefficient.</li>
        <li>&lt;3&gt; - The explicit exponent.</li>
        <li>&lt;4&gt; - The absolute value of the coefficient.</li>
        <li>&lt;5&gt; - The implicit coefficient. Blank if +1.</li>
        <li>&lt;6&gt; - The implicit unsigned coefficient. Blank if <u>+</u>1.</li>
        <li>&lt;7&gt; - The smart coefficient. Identical to #5 but visible for the constant term.</li>
        <li>&lt;8&gt; - The smart unsigned coefficient. Identical to #6 but visible for the constant term.</li>
        <li>&lt;9&gt; - The smart negative coefficient. Always visible, enclosed in parentheses if negative.</li>
        <li>&lt;10&gt; - The explicit sign (coefficient).</li>
        <li>&lt;11&gt; - The implicit sign. Blank if positive.</li>
        <li>&lt;12&gt; - The smart sign. Blank if it is positive and on the first term, visible otherwise.</li>
        <li>&lt;13&gt; - The implicit variable. Blank on the constant term.</li>
        <li>&lt;14&gt; - The implicit exponent. Blank if it is zero or one.</li>
        <li>&lt;15&gt; - Smart plus. Blank on the first term, '+' on subsequent ones.</li>
        <li>&lt;16&gt; - Explicit caret. This is set for all instances with the <code>setCaret()</code> methods and is always visible.</li>
        <li>&lt;17&gt; - Smart caret. Blank when the exponent is zero or one, visible otherwise.</li>
      </ul></p>
      <p>These are the tokens generated by <code>makeFormatTokens()</code> applied to this Polynomial.</p>
      @param pTerm The formatting template for each term.
      @param pAscending True to print ascending, false for descending.
      @param pIncludeZero True to include zero terms, false to omit them.
  */
  public String format(String pTerm, boolean pAscending, boolean pIncludeZero) {
    if(!Question.validString(pTerm)) pTerm = FORMAT_TYPICAL;
    StringBuffer retval = new StringBuffer();
    boolean isFirst = true;
    if(pAscending) {
      for(int i=0; i < myCoefs.length; i++) {
        if((myCoefs[i] !=0) || pIncludeZero) retval.append(Question.fillTemplate(pTerm, makeFormatTokens(myCoefs[i], i, myVar, myFName, isFirst)));
        isFirst = false;
      }
    } else {
      for(int i=myCoefs.length - 1; i >= 0; i--) {
        if((myCoefs[i] !=0) || pIncludeZero) retval.append(Question.fillTemplate(pTerm, makeFormatTokens(myCoefs[i], i, myVar, myFName, isFirst)));
        isFirst = false;
      }
    }
    return retval.toString();
  }

  /** <p>This method creates the formatting token array for the data supplied. This is independent of any particular Polynomial instance so the arguments must 
      be supplied each time. The tokens supplied are listed in the <code>format()</code> method.</p>
      @param pCoef The term coefficient.
      @param pExp The term exponent.
      @param pVar The term variable. If null or blank a default is supplied.
      @param pFName The polynomial or function name. If null or blank a default is supplied.
      @param isFirst Flag to indicate the first term formatted regardless of 'direction.'
  */
  public static String[] makeFormatTokens(int pCoef, int pExp, String pVar, String pFName, boolean isFirst) {
    String[] retval = new String[18];
    pVar = Question.validString(pVar)?pVar:"x"; pFName = Question.validString(pFName)?pFName:"P";
    retval[0] = pVar; retval[1] = pFName; retval[2] = ""+pCoef; retval[3] = ""+pExp; retval[4] = "" + Math.abs(pCoef);
    retval[5] = (pCoef == 1)?"":retval[2]; retval[6] = ((pCoef == 1) || (pCoef == -1))?"":retval[4];
    retval[7] = (pExp == 0)?retval[2]:retval[5]; retval[8] = (pExp == 0)?retval[4]:retval[6];
    retval[9] = (pCoef < 0)?"("+pCoef+")":"" + pCoef;
    retval[10] = (pCoef < 0)?"-":"+"; retval[11] = (pCoef < 0)?"-":""; retval[12] = (isFirst && (pCoef > 0))?"":retval[10];
    retval[13] = (pExp == 0)?"":pVar; retval[14] = ((pExp == 1) || (pExp == 0))?"":retval[3];
    retval[15] = (isFirst)?"":"+"; retval[16] = defaultCaret; retval[17] = ((pExp < 2)?"":defaultCaret);
    return retval;
  }

  /** <p>Format this Polynomial per its internal format. Zero terms are excluded.</p>
      <p>CRITICAL NOTE: This method has been updated to use the new <code>formatTokens()</code> method.</p>
      @param pAscending true to format in ascending order, false for descending.
      @return A freshly-formatted Polynomial.
  */
  public String format(boolean pAscending) { return format(myFormat, pAscending, false); }

  /** <p>This method creates a rational expression whose numerator can be factored and reduced. It accepts a quotient and divisor and produces the dividend as 
      the product of the two. It assumes both Polynomials sent in have formats set and it uses default formatting for the dividend. The 'lFence' and 'rFence' 
      Strings are wrapped around all formatted Strings. The 'divBy' argument is placed between the dividend and the quotient with no spacing.</p>
      <p>CRITICAL NOTE: This method has been updated to the new <code>format()</code> methods.</p>
      @param pQuotient The quotient (answer!) Polynomial.
      @param pDivisor The divisor Polynomial.
      @param lFence Appended to all formatted Polynomials on the left side. Ignored if null or blank.
      @param rFence Appended to all formatted Polynomials on the right side. Ignored if null or blank.
      @param pDivBy Placed between the dividend Polynomial and the divisor. Default value is a single slash.
  */
  public static Question formatQuotient(Polynomial pQuotient, Polynomial pDivisor, String pDivBy, String lFence, String rFence) {
    Question retval = new Question();
    if((pQuotient == null) || (pDivisor == null)) return retval;
    pDivBy = Question.validString(pDivBy)?pDivBy:"/";
    lFence = (lFence == null)?"":lFence;
    rFence = (rFence == null)?"":rFence;
    Polynomial theDividend = pQuotient.multiply(pDivisor);
    StringBuffer sbTemp = new StringBuffer().append(lFence).append(theDividend.format(false)).append(rFence).append(pDivBy);
    sbTemp.append(lFence).append(pDivisor.format(false)).append(rFence);
    retval.setProblem(sbTemp.toString());
    sbTemp = new StringBuffer().append(lFence).append(pQuotient.format(false)).append(rFence);
    retval.setAnswer(sbTemp.toString());
    return retval;
  }

  /** Instance version of <code>formatQuotient()</code>. */
  public Question formatNewQuotient(Polynomial pQuotient, Polynomial pDivisor, String pDivBy, String lFence, String rFence) {
    return formatQuotient(pQuotient, pDivisor, pDivBy, lFence, rFence);
  }

  /** <p>This method creates a polynomial equation or inequality based on the two Polynomials sent in. The left-hand side of the relation is the sum of the two 
      and the right-hand side is pRHP. If null or blank 'pRelation' defaults to '='. The fence characters are put around both formatted Strings and are 
      ignored if null or blank. If either Polynomial is null the method returns null.</p>
      <p>The original intent behind this method was to generate guaranteed-factorable polynomial equations with nonzero terms on each side. Its solvability 
      depends on that of the 'pLHP' Polynomial.</p>
      @param pLHP The left-hand Polynomial.
      @param pRHP The right-hand Polynomial.
      @param pRelation The relation to use. Defaults to equality.
      @param lFence Appended to both formatted Polynomials on the left.
      @param rFence Appended to both formatted Polynomials on the right.
  */
  public static String formatEquation(Polynomial pLHP, Polynomial pRHP, String pRelation, String lFence, String rFence) {
    if((pLHP == null) || (pRHP == null)) return null;
    pRelation = Question.validString(pRelation)?pRelation:"=";
    lFence = (lFence == null)?"":lFence;
    rFence = (rFence == null)?"":rFence;
    Polynomial theSum = pLHP.add(pRHP);
    StringBuffer sbTemp = new StringBuffer();
    sbTemp.append(lFence).append(theSum.format(false)).append(rFence).append(pRelation);
    sbTemp.append(lFence).append(pRHP.format(false)).append(rFence);
    return sbTemp.toString();
  }

  /** Instance shadow for creating equations. */
  public String formatNewEquation(Polynomial pLHP, Polynomial pRHP, String pRelation, String lFence, String rFence) {
    return formatEquation(pLHP, pRHP, pRelation, lFence, rFence);
  }

  /** Super-convenience method to assemble function notation consisting of this Polynomial's name and variable or expression. The body of the polynomial is not 
      included but an relation sign may be.
      @param pExpr The expression around which to build the function notation.
      @param pRelation The relation symbol to append to the finished notation, separated by a single space. Ignored if null or blank.
  */
  public String makeFun(String pExpr, String pRelation) {
    return myFName +"("+ (Question.validString(pExpr)?pExpr:myVar) +") " + (Question.validString(pRelation)?pRelation:"");
  }

  /** Set the <b>term</b> format for this Polynomial. Null or blank defaults to the default format. */
  public void setFormat(String pFormat) {
    myFormat = (Question.validString(pFormat))?pFormat:DEFAULT_FORMAT;
  }

  /** Set the default term format for every Polynomial created after this one. Null or blank defaults to FORMAT_TYPICAL. */
  public static void setDefaultFormat(String pFormat) {
    DEFAULT_FORMAT = (Question.validString(pFormat))?pFormat:FORMAT_TYPICAL;
  }

  /** Instance flavor of <code>setDefaultFormat()</code>. */
  public void setNewDefaultFormat(String pFormat) { setDefaultFormat(pFormat); }

  /** Express the Polynomial as a String. The Polynomial may be in ascending or descending order and printed using function notation or not. Despite its name 
      this method does not present fancy formatting options. To convert to MathML or some other format, dump the coefficient array and perform the formatting 
      manually. The array index of each term is its exponent.
      @param pAscending true for ascending order, false for descending order.
      @param pFunction true to use function notation, false for standard polynomial notation.
  */
  public String printPolynomial(boolean pAscending, boolean pFunction) {
    if(myCoefs.length == 1) return ""+myCoefs[0];
    StringBuffer retval = new StringBuffer();
    String sPos = "", sNeg = " -";
    if(pFunction) retval.append(myFName+"(").append(myVar+") = ");
    if(!pAscending) {
      for(int i = myCoefs.length - 1; i >= 0; i--) {
        if(myCoefs[i] == 0) continue;
        retval.append((myCoefs[i] < 0)?sNeg:sPos);
        if( !((myCoefs[i] == 1) || (myCoefs[i] == -1)) || (i == 0)) retval.append(Math.abs(myCoefs[i]));
        if(i > 0) retval.append(myVar);
        if(i > 1) retval.append("^"+i);
        sPos = " + "; sNeg = " - ";
      }
    } else {
      for(int i = 0; i < myCoefs.length; i++) {
        if(myCoefs[i] == 0) continue;
        retval.append((myCoefs[i] < 0)?sNeg:sPos);
        if( !((myCoefs[i] == 1) || (myCoefs[i] == -1)) || (i == 0) ) retval.append(Math.abs(myCoefs[i]));
        if(i > 0) retval.append(myVar);
        if(i > 1) retval.append("^"+i);
        sPos = " + "; sNeg = " - ";
      }
    }
    return retval.toString();
  }

  /** <p>This method provides a totally different way to format the Polynomial. The argument is a standard template string BUT in this method the replacement 
      tokens are the coefficients of the Polynomial: &lt;0&gt; is the constant term, &lt;1&gt; is the first-degree coefficient up to the degree of the entire 
      Polynomial. This allows for easier generation of Polynomial expressions with out-of-order terms.</p>
      <p>To facilitate custom construction of polynomial equations with nonzero terms on both sides this method (and <b>ONLY</b> this method) defines another set 
      of replacement tokens. Placing a double minus ('--') in front of the number within the token reverses the sign of that term. That means &lt;--2&gt; is the 
      opposite of the square coefficient.</p>
      <p>Coefficients of '+1' may be printed or suppressed. The 'printOnes' argument controls this; send true to include them, false to replace them with ''.</p>
      <p>If the template is null or blank it is returned unaltered.</p>
      @param pUnTemplate The template to receive the coefficient array.
      @param printOnes True to print coefficients of 1, false to blank them.
  */
  public String unFormat(String pUnTemplate, boolean printOnes) {
    if(!Question.validString(pUnTemplate)) return pUnTemplate;
    String[] tokenz = new String[myCoefs.length];
    for(int i = 0; i < tokenz.length; i++) {
      tokenz[i] = ((myCoefs[i] != 1) || (printOnes))?""+myCoefs[i]:"";
    }
    String retval = Question.fillTemplate(pUnTemplate, tokenz);
    retval = retval.replaceAll("<--", "<");
    for(int i = 0; i < tokenz.length; i++) {
      tokenz[i] = ((myCoefs[i] != -1) || (printOnes))?""+(-myCoefs[i]):"";
    }
    return Question.fillTemplate(retval, tokenz);
  }

  /** The toXML() method returns the Polynomial's data in a createPolynomial()-style XML String: &lt;Polynomial data='F:V:Cn ... C1 C0' /&gt; for easy 
      parsing. The pID, if not null or empty, becomes an 'id=' attribute.
      @param pID The id attribute of the resulting XML tag. May be null or empty.
  */
  public String toXML(String pID) {
    StringBuffer retval = new StringBuffer("<Polynomial ");
    if(Question.validString(pID)) retval.append("id=\""+pID+"\" ");
    retval.append(" data=\"");
    retval.append(dataString()).append("\" />");
    return retval.toString();
  }
  /** Overridden to provide a constructor-style representation of this Polynomial. */
  public String toString() {
    StringBuffer retval = new StringBuffer("Polynomial(\"");
    retval.append(dataString()).append("\" );");
    return retval.toString();
  }

  /** Format the Polynomial using MathML Presentation markup. This can be ascending or descending order and 'smart' or 'raw'. In 'raw' mode everything is present 
      on every term. 'Smart' mode is exactly per smart mode on MathXML.formatMonomial(). Regardless of mode, &lt;mrow&gt; tags are added appropriately.
      @param pAscending true to print in ascending mode, false for descending.
      @param pSmart true for smart mode, false for raw mode.
      @return A MathML-formatted String of the Polynomial.
  */
  public String toMathML(boolean pAscending, boolean pSmart) {
    StringBuffer retval = new StringBuffer("<mrow>");
    if(pAscending) {
      for(int i=0; i < myCoefs.length; i++) { retval.append(MathXML.formatMonomial(""+myCoefs[i], myVar, ""+i, pSmart)); }
    } else {
      for(int i=myCoefs.length - 1; i >= 0; i--) { retval.append(MathXML.formatMonomial(""+myCoefs[i], myVar, ""+i, pSmart)); }
    }
    retval.append("</mrow>");
    return retval.toString();
  }

  /* ########## End of formatting methods. ########## */

  /** Return the degree (exponent of the highest term) of the Polynomial. */
  public int degree() { return myCoefs.length - 1; }

  /** Return a copy of the Polynomial's coefficients. */
  public int[] coefficients() { return Arrays.copyOf(myCoefs, myCoefs.length); }
  /** Return a single specific coefficient. If the coefficient specified does not exist or the positon is negative, the return is 0.
      @param pPos The position of the coefficient to fetch.
      @return The coefficient at pPos.
  */
  public int coefficient(int pPos) {
    if((pPos > myCoefs.length - 1) || (pPos < 0)) return 0;
    return myCoefs[pPos];
  }

  /** This method provides a naive equality test for <code>Polynomial</code> instances. Two polynomials test as equal if and only if they are of equal degree 
      and all coefficients are equal. If 'pRHP' is null this method returns false immediately.
      @param pRHP The <code>Polynomial</code> to test for equality.
      @return false if 'pRHP' is null, true iff the degrees <i>and</i> the coefficients are equal.
  */
  public boolean equals(Polynomial pRHP) {
    if((pRHP == null) || (degree() != pRHP.degree())) return false;
    boolean retval = true;
    int[] rCoefs = pRHP.coefficients();
    for(int i=0; (retval && (i < myCoefs.length)); i++) retval = retval && (myCoefs[i] == rCoefs[i]);
    return retval;
  }

  /** Add this to another Polynomial. Neither is changed; a copy is returned. If pPoly is null it is treated as the zero Polynomial. The variable and function name 
      are of this Polynomial.
      @param pPoly The Polynomial to add to this one.
      @return The sum of the two Polynomials.
  */
  public Polynomial add(Polynomial pPoly) {
    if(pPoly == null) return new Polynomial(this);
    int[] sCoefs = pPoly.coefficients();
    int lMax = Math.max(myCoefs.length, sCoefs.length);
    int[] newCoefs = new int[lMax];
    Arrays.fill(newCoefs, 0);
    for(int i = 0; i < lMax; i++) {
      newCoefs[i] = ((i < myCoefs.length)?myCoefs[i]:0) + ((i < sCoefs.length)?sCoefs[i]:0);
    }
    return new Polynomial(newCoefs).truncate().setVariable(getVariable()).setFName(getFName());
  }

  /** Subtract another Polynomial from this one. Neither is changed; a copy is returned. If pPoly is null it is treated as the zero Polynomial. The variable and 
      function name are of this Polynomial.
      @param pPoly The Polynomial to subtract from this one.
      @return The difference of the two Polynomials.
  */
  public Polynomial subtract(Polynomial pPoly) {
    if(pPoly == null) return new Polynomial(this);
    int[] sCoefs = pPoly.coefficients();
    int lMax = Math.max(myCoefs.length, sCoefs.length);
    int[] newCoefs = new int[lMax];
    Arrays.fill(newCoefs, 0);
    for(int i = 0; i < lMax; i++) {
      newCoefs[i] = ((i < myCoefs.length)?myCoefs[i]:0) - ((i < sCoefs.length)?sCoefs[i]:0);
    }
    return new Polynomial(newCoefs).truncate().setVariable(getVariable()).setFName(getFName());
  }

  /** Multiply this Polynomial by another. If the second Polynomial is null a copy of this one is returned.
      @param pPoly The Polynomial to multiply by this one.
      @return The product of the two Polynomials.
  */
  public Polynomial multiply(Polynomial pPoly) {
    if(pPoly == null) return new Polynomial(this);
    int[] sCoefs = pPoly.coefficients();
    if(sCoefs.length == 1) return multiply(sCoefs[0]);
    int[] newCoefs = new int[myCoefs.length + sCoefs.length - 1];
    Arrays.fill(newCoefs, 0);
    for(int i=0; i < myCoefs.length; i++) {
      for(int j=0; j < sCoefs.length; j++) {
        newCoefs[i+j] += myCoefs[i] * sCoefs[j];
      }
    }
    return new Polynomial(newCoefs).truncate().setVariable(getVariable()).setFName(getFName());
  }

  /** Multiply the terms of this Polynomial by a constant. If the scalar is zero, the zero Polynomial is returned. As with other methods, this one does not alter 
      the Polynomial to which it is applied.
      @param pConst The int constant to multiply by the coefficients.
  */
  public Polynomial multiply(int pConst) {
    int[] sCoefs = coefficients();
    for(int i=0; i < sCoefs.length; i++) { sCoefs[i] *= pConst; }
    return new Polynomial(sCoefs).truncate().setVariable(getVariable()).setFName(getFName());
  }

  /** Evaluate the current Polynomial for the given value. Returned as a long to allow for evaluation.
      @param pVal The value at which to evaluate the Polynomial.
      @return The Polynomial evaluated at pVal.
  */
  public long evaluate(int pVal) {
    if(myCoefs.length == 1) return pVal * myCoefs[0];
    long retval = 0;
    for(int i = myCoefs.length - 1; i >= 0; i--) {  retval = retval * pVal + myCoefs[i]; }
    return retval;
  }

  /** This method evaluates <i>this</i> Polynomial OF another Polynomial. This is useful for function composition. Neither of the operands is changed and the 
      resulting Polynomial has its name set to the composition of the two names of the operands using a lowercase 'O' ('o') for the composition symbol. If the 
      Polynomial sent in is null the return value is the zero Polynomial.
      @param pPoly The Polynomial to be substituted for the variable in this and evaluated.
      @return The composition of this Polynomial evaluated at pPoly, or the zero Polynomial if this one is null.
  */
  public Polynomial evaluate(Polynomial pPoly) {
    if(pPoly == null) return new Polynomial();
    int[] retCoefs = new int[degree() * pPoly.degree() + 1];
    Arrays.fill(retCoefs, 0);
    int[] pCoefs = pPoly.coefficients(), tempCoefs = {0};
    retCoefs[0] = myCoefs[0];
    Polynomial pProd = new Polynomial(pPoly);
    for(int i=1; i < myCoefs.length; i++) {
      tempCoefs = (pProd.multiply(myCoefs[i])).coefficients();
      for(int j=0; j < tempCoefs.length; j++) { retCoefs[j] += tempCoefs[j]; }
      pProd = pProd.multiply(pPoly);
    }
    return new Polynomial(retCoefs).truncate().setFName(getFName()+"o"+pPoly.getFName());
  }

  /** Calulate the first derivative of the Polynomial, because it is easy. The original Polynomial is not changed. The function name is altered: it has a 'prime' 
      mark (single quote) appended.
  */
  public Polynomial derivative() {
    if(myCoefs.length == 0) return new Polynomial();
    int[] dCoefs = new int[myCoefs.length - 1];
    for(int i=0; i < dCoefs.length; i++) { dCoefs[i] = (i+1)*myCoefs[i+1]; }
    return new Polynomial(dCoefs).truncate().setVariable(getVariable()).setFName(getFName()+"'");
  }

}
