package cox.jmatt.java.MathTools;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
/** <p>This is MathTools' basic unit of data storage and presentation. A <code>Question</code> consists of three basic parts: the ID, the Problem part and the 
    Answer part. Since MathTools is designed to simplify creation and <i>formatting</i> of tests and worksheets this class also defines multiple formatting 
    methods and plenty of formatting data, both global and local.</p>
    <p>Concerning content, every individual <code>Question</code> component can be individually set or cleared. There are multiple methods for doing this along 
    with a very broad <code>setData()</code> method to set up to everything at once. What's more, many methods that would ordinarily return void instead return 
    a self-reference to allow for method chaining.</p>
    <h3>Templates</h3>
    <p><code>Question</code>'s unquestioned power-method is <code>fillTemplate()</code>. This method provides a very simple templating engine. It accepts a 
    String template and a String array of replacement data. The replacement tokens are numbered and the token number equals the index of the array element 
    that will replace it.</p>
    <p>When called, <code>fillTemplate()</code> replaces the numbered elements in the template String with the appropriate element in the replacement array. 
    The replacement tokens are of the form '&lt;<i><b>N</b></i>&gt;' where <i><b>N</b></i> is a number. This method is useful in and of itself but it also 
    forms the basis for the other MathTools formatting methods. Other MathTools classes define a specific set of replacement tokens and accept a template 
    String as an argument for maximum flexibility and simplicity.</p>
    <p>The {@link cox.jmatt.java.MathTools.test.MathTestFormatter MathTestFormatter} class provides a much more powerful templating engine, raising the Question 
    (pun intended) 'Why here, too?' The answer is in intents. <code>MathTestFormatter</code> is test-centric in its design. It is not concerned with exactly 
    how a fraction or radical should be formatted, merely with where they should go on the test. The <code>fillTemplate()</code> method here is concerned less 
    with the overall test and most with the formatting of the individual element the Question represents.</p>
    <h3>Validating Strings</h3>
    <p>Another method (static, with an instance shadow) used ubiquitously throughout MathTools is the <code>validString()</code> method. This method accepts a 
    String argument and returns true if the String sent in is non-null and has a trimmed length greater than zero.</p>
    <h2>Formatting and The Clicker</h2>
    <p>Each <code>Question</code> instance maintains four separate internal formats: ID, entire Question, Problem-only and Answer-only. Each of these formats 
    is individually settable but the <code>Question</code> class itself maintains (static) default formats for each of these. For maximum ease of use simply set 
    whatever default(s) are appropriate and only set what needs changing.</p>
    <h3>The Answer Token</h3>
    <p>The <b>Answer Token</b> is a part of the class that isn't really presentation nor is it data. Its original intent was as a simple String that could be 
    used to <code>split()</code> the result of a formatting method's output. The default value is '@@' and it should not be changed without a valid reason. As 
    with other data the Answer Token has a settable default value along with instance setter methods.</p>
    <h3>The Clicker</h3>
    <p>The Clicker is a number that can be accessed when formatting a Question's ID. Conceptually it represents a cheap plastic (or expensive metal) counter: 
    a small box with one or more buttons and three or more digit-wheels. When a button is pressed the corresponding digit-wheel increments by one. When one of 
    the wheels passes '9' it resets to '0' and automatically increments the next-higher wheel.</p>
    <p>The Clicker, then, provides an internal 'clicky-counter' for the <code>Question</code> class. Although it defines both static and instance methods to set 
    and increment, The Clicker itself is a static value. Its companion method <code>numberToLetter()</code> converts a numeric value, presumably The Clicker, to 
    a one- or more letter combination.</p>
    <h3>ID Formatting</h3>
    <p>Manually setting a Question's ID to a valid String sets that String as the ID. Period. That can be bothersome, though, when assigning large numbers of 
    relatively similar IDs. For this reason Question maintains a (static) default ID <i>format</i>. This format has access to The Clicker as a number, as 
    lower-case letters or as UPPERCASE letters. The <i>format</i> is static but each instance's ID can be set using it. This is done automatically when a new 
    <code>Question</code> is created but it can be forced by setting the (instance!) ID to null. This replaces the instance ID with the result of formatting 
    the default ID (format) with The Clicker. Afterwards a quick call to <code>click()</code> increments The Clicker.</p>
    <h3>Version Codes</h3>
    <p>This class provides a way to generate version codes automatically. The <code>formatVCode()</code> methods provide formatting tokens based on the current 
    time (full version!) and the Clicker. A formatting template with settable default determines the result of these methods.</p>
    <p>Besides simple version codes this method can also be used for timestamps. It provides both the date and time formatted per the <code>DateFormat</code> 
    constants SHORT, MEDIUM, LONG and FULL.</p>
    <h3>Built-In Formats</h3>
    <p>Concerning formatting questions, <code>Question</code> provides four pre-built format templates. Each of these is a static String constant that starts 
    with 'FORMAT_'. Debug format is for dumping the internal data and shouldn't really be used for production. XML format is used by the <code>toXML()</code> 
    method. The Properties and Velocity formats result in a <code>java.util.Properties</code>-type two-line result. The Velocity format uses underscores instead 
    of dots to allow for easier use with the Apache Velocity templating engine.</p>
*/
public class Question implements cox.jmatt.java.MathTools.test.MTestElement {

  /** This is the default Answer Token: '@@'. */
  public static final String DEFAULT_ANSWER_TOKEN = "@@";
  /** The default format used to generate ID Strings: 'Question_&lt;0&gt;.' */
  public static final String DEFAULT_ID_FORMAT = "Question_<0>";
  /** Default format for the entire <code>Question</code>: '&lt;0&gt; &lt;3&gt; &lt;1&gt; &lt;3&gt; &lt;2&gt;'. */
  public static final String DEFAULT_QUESTION_FORMAT = "<0> <3> <1> <3> <2>";
  /** Default format template for the Problem only: '&lt;0&gt;_Prob = &lt;1&gt;'. */
  public static final String DEFAULT_PROBLEM_FORMAT = "<0>_Prob = <1>";
  /** Default format template for the Answer part only: '&lt;0&gt;_Ans = &lt;2&gt;'. */
  public static final String DEFAULT_ANSWER_FORMAT = "<0>_Ans = <2>";
  /** Format String for internal MathTools XML. */
  public static final String FORMAT_XML = "<Question id=\"<0>\" ><Problem><1></Problem><Answer><2></Answer></Question>";
  /** Simple <code>Properties</code>-style format. */
  public static final String FORMAT_PROPERTIES = "Q<0>.P = <1> \nQ<0>.A = <2>";
  /** Apache Velocity-style format String. */
  public static final String FORMAT_VELOCITY = "Q<0>_P = <1> \nQ<0>_A = <2>";
  /** Debugging format, not for production use! */
  public static final String FORMAT_DEBUG = "-- Question Data --\nID:\t<0>\nProb:\t<1>\nAns:\t<2>\nToken:\t<3>";
  /** Version code default: &lt;3&gt;&lt;6&gt;&lt;9&gt;&lt;28&gt;. */
  public static final String DEFAULT_VCODE = "<3><6><9><28>";

  private static final String _NEWLINE_TOKEN = "/N3WlIn3/", _TAB_TOKEN = "/TaBt4b/", _RETURN_TOKEN = "/R3TuRn/", _QUOTE_TOKEN = "/Qu0t3d/";
  private static final String LETTERZ = "ABCDEFGHJKLMNPQRSTUVWXYZ";

  private static int theClicker = 0;
  private static String defaultAnswerToken = DEFAULT_ANSWER_TOKEN;
  private static String defaultIDFormat = DEFAULT_ID_FORMAT;
  private static String defaultVCode = DEFAULT_VCODE;
  private static String defaultFormat = DEFAULT_QUESTION_FORMAT, defaultProblemFormat = DEFAULT_PROBLEM_FORMAT, defaultAnswerFormat = DEFAULT_ANSWER_FORMAT;
  private static boolean defaultEscapeXML = true;

  private String myAnswerToken = defaultAnswerToken;
  private String myID = formatID(defaultIDFormat);
  private String myProblem = "", myAnswer = "";
  private String myFormat = defaultFormat, myProblemFormat = defaultProblemFormat, myAnswerFormat = defaultAnswerFormat;
  private boolean myEscapeXML = defaultEscapeXML;

  /** Build a blank <code>Question</code>. */
  public Question() {
    super();
  }

  /** Create a <code>Question</code> with Problem and Answer specified. */
  public Question(String pProb, String pAns) {
    super();
    if(validString(pProb)) myProblem = pProb;
    if(validString(pAns)) myAnswer = pAns;
  }

  /** Build a <code>Question</code> with ID, Problem and Answer specified. */
  public Question(String pID, String pProb, String pAns) {
    super();
    if(validString(pID)) myID = pID;
    if(validString(pProb)) myProblem = pProb;
    if(validString(pAns)) myAnswer = pAns;
  }

  /** Copy constructor. If pQues is null an empty Question is returned. NO FORMATTING INFORMATION IS COPIED!! Only data. */
  public Question(Question pQues) {
    super();
    if(pQues != null) {
      myID = pQues.getID();
      myProblem = pQues.getProblem();
      myAnswer = pQues.getAnswer();
    }
  }

  /** Static factory method to create a blank Question. */
  public static Question createQuestion() { return new Question(); }

  /** Shadow method for <code>createQuestion()</code>. */
  public Question newQuestion() { return new Question(); }

  /** Factory method to set up to all internal data. Parameters are per <code>setData()</code>. */
  public static Question createQuestion(String... pData) { return (new Question()).setData(pData); }

  /** Instance shadow for the full <code>createQuestion()</code> method. */
  public Question newQuestion(String... pData) { return (new Question()).setData(pData); }

  /** <p>Generate a <code>Question</code> by breaking a string on a specified splitting sequence. If the resulting array has two or three elements they are formed 
      into a new <code>Question</code> (ID + Problem + Answer or Problem + Answer). If pData is null or blank or the split array is not two or three elements 
      a blank <code>Question</code> is returned. If pSplitter is null or blank the default Answer Token ('@@') is used.</p>
      <p>This method is provided to allow Questions to be stored compactly as single Strings. Of course scripts can also build the requisite String manually and 
      parse it into a Question when needed. Judicious use of the <code>unquote()</code> and <code>quote()</code> methods can tidy things up as needed.</p>
      @param pData A String containing two or three Question components.
      @param pSplitter The String on which pData is to be split.
  */
  public static Question parseQuestion(String pData, String pSplitter) {
    Question retval = new Question();
    if(!validString(pData)) return retval;
    pSplitter = validString(pSplitter)?pSplitter:DEFAULT_ANSWER_TOKEN;
    String[] partz = pData.split(pSplitter);
    if((partz.length == 2) || (partz.length == 3)) retval.setData(partz);
    return retval;
  }

  /** Instance shadow for <code>parseQuestion(String, String)</code>. */
  public Question parse_Question(String pData, String pSplitter) { return parseQuestion(pData, pSplitter); }

  /** Convenience method for <code>parseQuestion(String, null)</code>. */
  public static Question parseQuestion(String pData) { return parseQuestion(pData, null); }

  /** Instance shadow for the single-String parse method. */
  public Question parse_Question(String pData) { return parseQuestion(pData, null); }

  /** Check to see if a string is null or of zero length AFTER its trim() function is applied: whitespace characters are considered length zero!
      @param pStr The String to check.
      @return false if the String is null or of zero length, true otherwise.
  */
  public static boolean validString(String pStr) { return ((pStr != null) && (pStr.trim().length() > 0)); }
  /** This is the instance version of validString(). The name is slightly altered. */
  public boolean isValidString(String pStr) { return Question.validString(pStr); }

  /** This method replaces newlines, tabs, returns and quotes. The replacement tokens are longer than one character but can be read as to what they are. The 
      intent with this method is to allow a String containing potentially irritating formatting characters to be expressed as a single unbroken line. Because 
      of the inherent instability in trying to out-think the number of backslashes required the replacement tokens are mangled versions of the spelling of what 
      they replace. If this method behaves irrationally the replacement token characters should be checked. To view these tokens <code>unquote()</code> a String 
      with the formatting characters and print it from a script.
      @param pRaw The String containing embedded tabs, newlines etc.
      @return The same String but with escapes replacing the embedded characters.
  */
  public static String unquote(String pRaw) {
    pRaw = pRaw.replaceAll("\n", _NEWLINE_TOKEN);
    pRaw = pRaw.replaceAll("\r", _RETURN_TOKEN);
    pRaw = pRaw.replaceAll("\t", _TAB_TOKEN);
    pRaw = pRaw.replaceAll("\"", _QUOTE_TOKEN);    
    return pRaw;
  }
  /** Instance method. */
  public String un_quote(String pRaw) { return unquote(pRaw); }

  /** This method undoes <code>unquote()</code>. It restores the original characters into the supplied String.
      @param pCooked A String containing converted (escaped) control characters.
      @return The original String with the raw control characters restored.
  */
  public static String requote(String pCooked) {
    pCooked = pCooked.replaceAll(_NEWLINE_TOKEN, "\n");
    pCooked = pCooked.replaceAll(_RETURN_TOKEN, "\r");
    pCooked = pCooked.replaceAll(_TAB_TOKEN, "\t");
    pCooked = pCooked.replaceAll(_QUOTE_TOKEN, "\"");    
    return pCooked;
  }

  /** Instance shadow. */
  public String re_quote(String pCooked) { return requote(pCooked); }

  /** <p>This is Question's power method. It provides a very basic templating engine. The template itself is a string containing tokens of the form '&lt;<i>X</i>&gt;' 
      or '&lt;K<i>X</i>&gt;' where <i>X</i> is a number. Each numbered token is replaced with the corresponding element from the String[] array supplied to the 
      method. Invalid tokens are ignored as are numbers outside the array supplied. If the template is null or blank it is returned as-is. If the replacement array 
      is null or empty, the template is returned unaltered. The 'K' form of the token is there for Knob identification; either form is equally valid and both may 
      be mixed.</p>
      <p>This method does define one special token: '&lt;-1&gt;' is a blanking token; it is replaced with an empty String after all other tokens have been 
      replaced.</p>
      @param pTemplate A String template containing one or more replacement tokens as defined above.
      @param pData A String array containing replacement data for the tokens in the template.
      @return A fresh String with all tokens replaced.
  */
  public static String fillTemplate(String pTemplate, String[] pData) {
    if((!validString(pTemplate)) || (pData == null) || (pData.length < 1)) return pTemplate;
    String sEngine = new String(pTemplate);
    try {
      for(int i=0; i < pData.length; i++) {
        pData[i] = java.util.regex.Matcher.quoteReplacement(pData[i]);
        sEngine = sEngine.replaceAll("<"+i+">", pData[i]);
        sEngine = sEngine.replaceAll("<K"+i+">", pData[i]);
        sEngine = sEngine.replaceAll("<k"+i+">", pData[i]);
      }
      sEngine = sEngine.replaceAll("[<]-1[>]", "");
    } catch(Exception ex) { CapCom.error("Question.fillTemplate error.", ex); }
    return sEngine;
  }
  /** Instance version of fillTemplate() with a slightly changed name. */
  public String fill_Template(String pTemplate, String[] pData) { return fillTemplate(pTemplate, pData); }

  /** This is <code>fillTemplate()</code> designed around an array of Number instead of String. The pForceInt flag can be set to 'true' to format the Numbers as 
      integer type (long, specifically), but should only be used when the Numbers <b>are</b> integer type.
      @param pTemplate A String template containing replacement tokens.
      @param pNData A Number array with replacement values.
      @param pForceInt Set 'true' to force an integer value.
      @return A fresh String with all the numeric tokens replaced.
  */
  public static String fillTemplate(String pTemplate, Number[] pNData, boolean pForceInt) {
    if((!validString(pTemplate)) || (pNData == null) || (pNData.length < 1)) return pTemplate;
    String sEngine = new String(pTemplate), sData = null;
    try {
      for(int i=0; i < pNData.length; i++) {
        sData = (pNData[i] == null)?"<-1>":(pForceInt?""+(pNData[i].longValue()):pNData[i].toString());
        sEngine = sEngine.replaceAll("<"+i+">", sData);
        sEngine = sEngine.replaceAll("<K"+i+">", sData);
        sEngine = sEngine.replaceAll("<k"+i+">", sData);
      }
      sEngine = sEngine.replaceAll("[<]-1[>]", "");
    } catch(Exception ex) { CapCom.error("Question.fillNTemplate error.", ex); }
    return sEngine;
  }

  /** Instance version of <code>fillTemplate()</code>, Number version. */
  public String fill_Template(String pTemplate, Number[] pNData, boolean pForceInt) { return fillTemplate(pTemplate, pNData, pForceInt); }

  /** Set a new default format for all subsequent instances. If a format is null or blank DEFAULT_QUESTION_FORMAT is used. Replacement tokens are per the 
      <code>format(String)</code> method. Up to three defaults can be specified; use is per <code>setFormat(String...)</code>.
      @param pFormats The new default format templates. If null this method returns silently.
  */
  public static void setDefaultFormat(String... pFormats) {
    if(pFormats == null) return;
    switch(pFormats.length) {
      case 3:
        defaultFormat = validString(pFormats[0])?pFormats[0]:DEFAULT_QUESTION_FORMAT;
        pFormats[0] = pFormats[1]; pFormats[1] = pFormats[2];
      case 2:
        defaultProblemFormat = validString(pFormats[0])?pFormats[0]:DEFAULT_PROBLEM_FORMAT;
        defaultAnswerFormat = validString(pFormats[1])?pFormats[1]:DEFAULT_ANSWER_FORMAT;
        break;
      case 1:
        defaultFormat = validString(pFormats[0])?pFormats[0]:DEFAULT_QUESTION_FORMAT;
      default:

    }
  }
  /** Shadow method to set default format. Returns a self-reference. */
  public Question setNewDefaultFormat(String... pFormats) {
    setDefaultFormat(pFormats);
    return this;
  }

  /** Set the default value for XML-escaping Problem and Answer content. Unless this is handled explicitly it should always be 'true.' */
  public static void setDefaultEscapeXML(boolean pYesNo) { defaultEscapeXML = pYesNo; }
  /** Shadow method to set default XML escape. */
  public Question setNewDefaultEscapeXML(boolean pYesNo) {
    defaultEscapeXML = pYesNo;
    return this;
  }

  /** Increment the clicker by 1. */
  public static void incrementClicker() { theClicker++; }
  /** Shadow method for <code>incrementClicker()</code>. Returns a self-reference to allow method chaining. */
  public Question click() {
    theClicker++;
    return this;
  }

  /** Set a new value for the clicker. If less than zero it becomes zero.
      @param pNewVal The new clicker value.
  */
  public static void setClicker(int pNewVal) { theClicker = ((pNewVal < 0)?0:pNewVal); }
  /** Instance version of <code>setClicker()</code>. Returns a self-reference to allow chaining. */
  public Question setNewClicker(int pNewVal) {
    setClicker(pNewVal);
    return this;
  }

  /** Convert a number to an uppercase letter. Does not include 'I' or 'O' since this is primarily to represent question numbers as letters. If the number is too
      great for a single letter the sequence grows to two letters: 'AA,' 'AB,' 'AC' etc.
      @param pNum The number to express as a letter.
  */
  public static String numberToLetter(long pNum) {
    String retval = "";
    if(pNum < LETTERZ.length()) {
      retval = "" + LETTERZ.charAt((int) pNum);
    } else {
      retval = numberToLetter(pNum / LETTERZ.length() - 1) + numberToLetter((int) (pNum % LETTERZ.length()));
    }
    return retval;
  }

  /** Set the default version code format. Setting null or blank sets the default. */
  public static void setVCFormat(String pFormat) { defaultVCode = validString(pFormat)?pFormat:DEFAULT_VCODE; }
  /** Instance shadow. */
  public void setNewVCFormat(String pFormat) { setVCFormat(pFormat); }

  /** <p>Format a version code per the supplied template. Most of the replacement tokens are based on a <code>java.util.Calendar</code> created as soon as this 
      method is called. The <code>Calendar</code> is specifically set to lenient mode and where applicable zero-based values (offsets) are used. Each quantity 
      supplied has three tokens: the raw number, the number as UPPER case letters and as lower case.</p>
      <p>The formatting tokens are:
      <center>
      <table border="1" >
      <tr><td><b>Raw</b></td><td><b>UC</b></td><td><b>LC</b></td><td><b>Description</b></td></tr>
      <tr><td>&lt;0&gt;</td><td>&lt;1&gt;</td><td>&lt;2&gt;</td><td>Current year, 4-digit.</td></tr>
      <tr><td>&lt;3&gt;</td><td>&lt;4&gt;</td><td>&lt;5&gt;</td><td>Current year, 2-digit.</td></tr>
      <tr><td>&lt;6&gt;</td><td>&lt;7&gt;</td><td>&lt;8&gt;</td><td>Current month.</td></tr>
      <tr><td>&lt;9&gt;</td><td>&lt;10&gt;</td><td>&lt;11&gt;</td><td>Current day of month.</td></tr>
      <tr><td>&lt;12&gt;</td><td>&lt;13&gt;</td><td>&lt;14&gt;</td><td>Current hour (0 - 23).</td></tr>
      <tr><td>&lt;15&gt;</td><td>&lt;16&gt;</td><td>&lt;17&gt;</td><td>Current minute.</td></tr>
      <tr><td>&lt;18&gt;</td><td>&lt;19&gt;</td><td>&lt;20&gt;</td><td>Current second.</td></tr>
      <tr><td>&lt;21&gt;</td><td>&lt;22&gt;</td><td>&lt;23&gt;</td><td>Current millisecond.</td></tr>
      <tr><td>&lt;24&gt;</td><td>&lt;25&gt;</td><td>&lt;26&gt;</td><td>The long time in millis.</td></tr>
      <tr><td>&lt;27&gt;</td><td>&lt;28&gt;</td><td>&lt;29&gt;</td><td>Long time in millis modded to a two-letter String.</td></tr>
      <tr><td>&lt;30&gt;</td><td>&lt;31&gt;</td><td>&lt;32&gt;</td><td>Long time in millis modded to a single letter String.</td></tr>
      <tr><td>&lt;33&gt;</td><td>&lt;34&gt;</td><td>&lt;35&gt;</td><td>The Clicker.</td></tr>
      <tr><td><b>Date</b></td><td><b>Time</b></td><td /><td><b>Timestamp Style</b></td></tr>
      <tr><td>&lt;36&gt;</td><td>&lt;37&gt;</td><td /><td><code>DateFormat.SHORT</code>.</td></tr>
      <tr><td>&lt;38&gt;</td><td>&lt;39&gt;</td><td /><td><code>DateFormat.MEDIUM</code>.</td></tr>
      <tr><td>&lt;40&gt;</td><td>&lt;41&gt;</td><td /><td><code>DateFormat.LONG</code>.</td></tr>
      <tr><td>&lt;42&gt;</td><td>&lt;43&gt;</td><td /><td><code>DateFormat.FULL</code>.</td></tr>
      </table></center></p>
      <p>If 'pFormat' is null or blank DEFAULT_VCODE is used.</p>
  */
  public static String formatVCode(String pFormat) {
    pFormat = validString(pFormat)?pFormat:DEFAULT_VCODE;
    String[] tokenz = new String[44];
    Calendar calTemp = Calendar.getInstance();
    calTemp.setLenient(true);
    java.text.DateFormat dfBase = java.text.DateFormat.getInstance(), dfTemp = null;
    dfBase.setLenient(true);
    dfBase.setCalendar(calTemp);
    Date rightNow = calTemp.getTime();
    int iTemp = 0, numLet = LETTERZ.length(), numLet2 = numLet * numLet;
    iTemp = calTemp.get(Calendar.YEAR); tokenz[0] = ""+iTemp; tokenz[1] = numberToLetter(iTemp); tokenz[2] = tokenz[1].toLowerCase();
    iTemp %= 100; tokenz[3] = ""+iTemp; tokenz[4] = numberToLetter(iTemp); tokenz[5] = tokenz[4].toLowerCase();
    iTemp = calTemp.get(Calendar.MONTH); tokenz[6] = ""+iTemp; tokenz[7] = numberToLetter(iTemp); tokenz[8] = tokenz[7].toLowerCase();
    iTemp = calTemp.get(Calendar.DAY_OF_MONTH); tokenz[9] = ""+iTemp; tokenz[10] = numberToLetter(iTemp); tokenz[11] = tokenz[10].toLowerCase();
    iTemp = calTemp.get(Calendar.HOUR_OF_DAY); tokenz[12] = ""+iTemp; tokenz[13] = numberToLetter(iTemp); tokenz[14] = tokenz[13].toLowerCase();
    iTemp = calTemp.get(Calendar.MINUTE); tokenz[15] = ""+iTemp; tokenz[16] = numberToLetter(iTemp); tokenz[17] = tokenz[16].toLowerCase();
    iTemp = calTemp.get(Calendar.SECOND); tokenz[18] = ""+iTemp; tokenz[19] = numberToLetter(iTemp); tokenz[20] = tokenz[19].toLowerCase();
    iTemp = calTemp.get(Calendar.MILLISECOND); tokenz[21] = ""+iTemp; tokenz[22] = numberToLetter(iTemp); tokenz[23] = tokenz[22].toLowerCase();
    long lTemp = calTemp.getTimeInMillis(); tokenz[24] = ""+lTemp; tokenz[25] = numberToLetter(lTemp); tokenz[26] = tokenz[25].toLowerCase();
    lTemp %= numLet2; tokenz[27] = ""+lTemp; tokenz[28] = numberToLetter(lTemp); tokenz[29] = tokenz[28].toLowerCase();
    lTemp %= numLet; tokenz[30] = ""+lTemp; tokenz[31] = numberToLetter(lTemp); tokenz[32] = tokenz[31].toLowerCase();
    iTemp = theClicker; tokenz[33] = ""+iTemp; tokenz[34] = numberToLetter(iTemp); tokenz[35] = tokenz[34].toLowerCase();
    iTemp = java.text.DateFormat.SHORT; tokenz[36] = dfBase.getDateInstance(iTemp).format(rightNow); tokenz[37] = dfBase.getTimeInstance(iTemp).format(rightNow);
    iTemp = java.text.DateFormat.MEDIUM; tokenz[38] = dfBase.getDateInstance(iTemp).format(rightNow); tokenz[39] = dfBase.getTimeInstance(iTemp).format(rightNow);
    iTemp = java.text.DateFormat.LONG; tokenz[40] = dfBase.getDateInstance(iTemp).format(rightNow); tokenz[41] = dfBase.getTimeInstance(iTemp).format(rightNow);
    iTemp = java.text.DateFormat.FULL; tokenz[42] = dfBase.getDateInstance(iTemp).format(rightNow); tokenz[43] = dfBase.getTimeInstance(iTemp).format(rightNow);
    return fillTemplate(pFormat, tokenz);
  }
  /** Instance shadow. */
  public String formatNewVCode(String pFormat) { return formatVCode(pFormat); }

  /** Format a VCode per the currently-set default. */
  public static String formatVCode() { return formatVCode(defaultVCode); }
  /** Instance shadow. */
  public String formatNewVCode() { return formatVCode(defaultVCode); }

  /* Internal ID formatting method. <0> is clicker, <1> is UPPER, <2> is lowercase. */
  private static final String formatID(String pRawID) {
    String[] tokenz = new String[3];
    tokenz[0] = ""+theClicker; tokenz[1] = numberToLetter(theClicker); tokenz[2] = tokenz[1].toLowerCase();
    return fillTemplate(pRawID, tokenz);
  }
  /** <p>This method sets the default format used to generate <code>Question</code> IDs. Null or blank resets it to DEFAULT_ID_FORMAT. There are three replacement 
      tokens available:
      <ul>
        <li>&lt;0&gt; - The numeric value of the Clicker.</li>
        <li>&lt;1&gt; - The value of the Clicker as an UPPERCASE letter.</li>
        <li>&lt;2&gt; - The value of the Clicker as a lowercase letter.</li>
      </ul></p>
      <p>This value remains until it is changed.</p>
      @param pFormat The new ID format String.
  */
  public static void setIDFormat(String pFormat) {
    defaultIDFormat = validString(pFormat)?pFormat:DEFAULT_ID_FORMAT;
  }

  /** Set the ID for this <code>Question</code>. If pNewID is null or blank a default ID is assigned based on the default format and current Clicker. Returns a 
      self-reference to allow method chaining.
      @param pNewID The new ID to assign.
  */
  public Question setID(String pNewID) {
    myID = validString(pNewID)?pNewID:formatID(defaultIDFormat);
    return this;
  }
  /** Retrieve the currently-set ID. No formatting takes place here. */
  public String getID() { return myID; }

  /** Set the Problem part of this <code>Question</code>. If null a blank String is assigned.
      @param pProb The new value for the Problem.
  */
  public Question setProblem(String pProb) {
    myProblem = (pProb == null)?"":pProb;
    return this;
  }
  /** Fetch the current value of the Problem. */
  public String getProblem() { return myProblem; }

  /** Set the Answer part of this <code>Question</code>. If null a blank String is assigned.
      @param pAns The new value for the Answer.
  */
  public Question setAnswer(String pAns) {
    myAnswer = (pAns == null)?"":pAns;
    return this;
  }
  /** Fetch the current value of the Answer. */
  public String getAnswer() { return myAnswer; }

  /** Set the Answer Token for this <code>Question</code>. If pToken is null the default Answer Token is used.
      @param pToken The new value for the Answer Token.
  */
  public Question setAnswerToken(String pToken) {
    myAnswerToken = validString(pToken)?pToken:DEFAULT_ANSWER_TOKEN;
    return this;
  }
  /** Fetch the current Answer Token. */
  public String getAnswerToken() { return myAnswerToken; }

  /** <p>Set up to ALL aspects of this Question in one fell swoop. The number of arguments is valid for up to four elements. These are interpreted as follows:
      <ul>
        <li>4 arguments: ID, Problem, Answer, Answer Token.</li>
        <li>3 arguments: ID, Problem, Answer.</li>
        <li>2 arguments: Problem, Answer.</li>
        <li>1 argument:  Answer Token.</li>
      </ul></p>
      <p>Any other argument lengths are ignored. A self-reference is returned.</p>
      @param pData Up to four Strings for each component.
  */
  public Question setData(String... pData) {
    if(pData == null) return this;
    switch(pData.length) {
      case 4:
        setAnswerToken(pData[3]);
      case 3:
        setID(pData[0]);
        pData[0] = pData[1]; pData[1] = pData[2];
      case 2:
        setProblem(pData[0]);
        setAnswer(pData[1]);
        break;
      case 1:
        setAnswerToken(pData[0]);
      default:
    }
    return this;
  }

  /** <p>Format this <code>Question</code> per the supplied template String. All data is supplied as <code>fillTemplate()</code>-style replacement tokens. The 
      tokens available are:
      <ul>
        <li>&lt;0&gt; - The current ID.</li>
        <li>&lt;1&gt; - The Problem part.</li>
        <li>&lt;2&gt; - The Answer part.</li>
        <li>&lt;3&gt; - The current Answer Token.</li>
      </ul></p>
      <p>This is the base method for all other <code>Question.format()</code> methods so these are the replacement tokens available to them.</p>
      @param pTemplate A template String used to format the Question.
      @return The freshly-formatted String based on the template sent in and the current data set.
  */
  public String format(String pTemplate) {
    String[] tokenz = new String[4];
    tokenz[0] = myID; tokenz[1] = myProblem; tokenz[2] = myAnswer; tokenz[3] = myAnswerToken;
    return fillTemplate(pTemplate, tokenz);
  }

  /** <p>Set the internal formats used to format the individual components. One, two or three arguments are valid; the interpretation depends on the number:
      <ul>
        <li>3 arguments: entire Question format, Problem format, Answer format.</li>
        <li>2 arguments: Problem format, Answer format.</li>
        <li>1 argument:  entire Question format.</li>
      </ul></p>
      <p>Any other number of arguments are ignored. A self-reference is returned.</p>        
      @param pFormats Up to three Strings used for the internal format(s) for this instance.
  */
  public Question setFormat(String... pFormats) {
    if(pFormats == null) return this;
    switch(pFormats.length) {
      case 3:
        myFormat = validString(pFormats[0])?pFormats[0]:defaultFormat;
        pFormats[0] = pFormats[1]; pFormats[1] = pFormats[2];
      case 2:
        myProblemFormat = validString(pFormats[0])?pFormats[0]:defaultProblemFormat;
        myAnswerFormat = validString(pFormats[1])?pFormats[1]:defaultAnswerFormat;
        break;
      case 1:
        myFormat = validString(pFormats[0])?pFormats[0]:defaultFormat;
      default:

    }
    return this;
  }

  /** Format this <code>Question</code> per its internal format. */
  public String format() { return format(myFormat); }

  /** Format per the internal Problem format. */
  public String formatProblem() { return format(myProblemFormat); }

  /** Format per the internal Answer format. */
  public String formatAnswer() { return format(myAnswerFormat); }

  /** Return a simple constructor-like representation of this Question. */
  public String toString() {
    StringBuffer retval = new StringBuffer("Question(\"").append(unquote(myID)).append("\", ");
    retval.append("\"").append(unquote(myProblem)).append("\", ");
    retval.append("\"").append(unquote(myAnswer)).append("\");");
    return retval.toString();
  }

  /** Set the 'escape XML' state for this instance and return a self-reference. Unless the content <b>is</b> XML or escaping is handled explicitly this flag 
      should always be 'true.'
      @param pYesNo true to escape XML content, false to disable.
  */
  public Question setEscapeXML(boolean pYesNo) {
    myEscapeXML = pYesNo;
    return this;
  }

  /** Return a simple XML representation of this <code>Question</code>. This particular format is internal to MathTools and not meant to be used outside of it. 
      It consists of a &lt;Question&gt; tag with an 'id=' attribute and two child elements: &lt;Problem&gt; and &lt;Answer&gt;.
  */
  public String toXML() {
    String[] tokenz = new String[3];
    tokenz[0] = myEscapeXML?MathXML.escapeXML(myID):myID;
    tokenz[1] = myEscapeXML?MathXML.escapeXML(myProblem):myProblem;
    tokenz[2] = myEscapeXML?MathXML.escapeXML(myAnswer):myAnswer;
    return fillTemplate(FORMAT_XML, tokenz);
  }

  /** Create a copy of this Question. Only data is copied; formatting information and the Answer Token are not. */
  public Question copyOf() { return new Question(this); }

  /** This method <code>unquote()</code>'s the ID, Problem and Answer. This allows multiline content to be represented on a single, compact line. A self-reference 
      is returned.
  */
  public Question unqoute() {
    myID = unquote(myID);
    myProblem = unquote(myProblem);
    myAnswer = unquote(myAnswer);
    return this;
  }

  /** This method <code>requote()</code>'s the (presumably-escaped) ID, Problem and Answer. This restores the original formatting. */
  public Question requote() {
    myID = requote(myID);
    myProblem = requote(myProblem);
    myAnswer = requote(myAnswer);
    return this;
  }

}

