package cox.jmatt.java.MathTools.demo;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.test.*;

/** <p>This class provides a fairly comprehensive demonstration of <code>MathTestFormatter</code> and serves as a handy utility in its own right. The GUI consists 
    of the main (File) panel, an internal template panel, a Properties panel, and a regex data Panel. The main panel has TextFields for selecting template or 
    Properties files as well as a Checkbox to send the formatted template to <code>CapCom.processMathObject()</code>, a Checkbox to enable regex matching and a 
    TextField to enter the regular expression. Since <code>MTDemo</code> uses a <code>BasicMathPrinter</code> this sends the result to the clipboard.</p>
    <p>The template can be internal or external. The internal area can be loaded (if this class is built around a <code>Dialog</code> or <code>Frame</code>). 
    The Properties can be both internal <i>and</i> external. Enabling the '[]Properties file' Checkbox on the File panel loads Properties from the file specified. 
    Enabling the internal Properties Checkbox parses the Properties panel (<code>addProperties(String, null, null)</code>) and adds them to the formatter. The 
    order in which properties are set is exactly that: external first followed by internal.</p>
    <p>Once the template String and Properties are set the template is formatted using the zero-argument method. From there the result is printed and, if the 
    Checkbox is enabled, processed.</p>
    <p>To format using regular expressions enable the '[]Regex' checkbox on the file panel and enter the regular expression in that TextField. The data String(s) 
    to which the regex is to be applied should be entered or loaded in the Regex Data panel. The content is split on newlines and the array is then sent to the 
    <code>reFormat()</code> method. The results are then re-assembled into a newline-separated String and printed/processed.</p>
*/
public class MTFormatterDemo extends Panel implements ActionListener {

  /** String constant used to add the File (main) panel. */
  public static final String KEY_FILE = "File";
  /** String constant used to add the internal template panel. */
  public static final String KEY_TEMPLATE = "Template";
  /** String constant used to add the internal Properties panel. */
  public static final String KEY_PROPS = "Properties";
  /** String constant used to add the regular expression data panel. */
  public static final String KEY_REGEX = "Regex";

  private static final String BUTTON_LOAD_TEMPLATE = "Load Template";
  private static final String BUTTON_CLEAR_TEMPLATE = "Clear Template";
  private static final String BUTTON_CHOOSE_TEMPLATE = "Choose Template file...";
  private static final String BUTTON_LOAD_PIZZA = "Load Properties";
  private static final String BUTTON_CLEAR_PIZZA = "Clear Pizza";
  private static final String BUTTON_CHOOSE_PIZZA = "Choose Properties file...";
  private static final String BUTTON_LOAD_REGEX = "Load Regex Data";
  private static final String BUTTON_CLEAR_REGEX = "Clear Regex";
  private static final String BUTTON_NEXT = "|>>";
  private static final String BUTTON_PREVIOUS = "<<|";
  private static final String BUTTON_GO = "Go!";

  private Dialog myParentDialog = null;
  private Frame myParentFrame = null;
  private FileDialog myFileDialog = null;
  private CardLayout cdlMain = null;
  private TextField txfTemplateFile = new TextField(40), txfPizzaFile = new TextField(40), txfRegex = new TextField(40);
  private TextArea txaTemplate = new TextArea(), txaPizza = new TextArea(), txaRegexData = new TextArea();
  private Checkbox cbxPizzaFile = new Checkbox("Properties"), cbxTemplateFile = new Checkbox("Template");
  private Checkbox cbxLockTemplate = new Checkbox("Lock Clear", true), cbxLockProps = new Checkbox("Lock Clear", true), cbxLockRegex = new Checkbox("Lock Clear", true);
  private Checkbox cbxEnablePizza = new Checkbox("Enable internal Properties");
  private Checkbox cbxClip = new Checkbox("Call processMathObject() on result", true);
  private Checkbox cbxEnableRegex = new Checkbox("Regex");

  private boolean gotParent = false;

  /** Zero-argument version, no [Load] buttons will be present. */
  public MTFormatterDemo() { super(); }

  /** Build a demo around a Frame to allow <code>FileDialog</code>s for loading. */
  public MTFormatterDemo(Frame pParent) {
    super();
    myParentFrame = pParent;
    gotParent = (pParent != null);
  }

  /** Build a demo around a <code>Dialog</code> to allow <code>FileDialog</code>s for loading. */
  public MTFormatterDemo(Dialog pParent) {
    super();
    myParentDialog = pParent;
    gotParent = (pParent != null);
  }

  /** Overridden to handle button clicks, nothing else. */
  public void actionPerformed(ActionEvent aev) {
    String sWhat = aev.getActionCommand();
    if(BUTTON_NEXT.equals(sWhat)) {
      cdlMain.next(this);
    } else if(BUTTON_PREVIOUS.equals(sWhat)) {
      cdlMain.previous(this);
    } else if(BUTTON_CLEAR_TEMPLATE.equals(sWhat)) {
      clearPanel(KEY_TEMPLATE, !cbxLockTemplate.getState());
      cbxLockTemplate.setState(true);
    } else if(BUTTON_CLEAR_PIZZA.equals(sWhat)) {
      clearPanel(KEY_PROPS, !cbxLockProps.getState());
      cbxLockProps.setState(true);
    } else if(BUTTON_CLEAR_REGEX.equals(sWhat)) {
      clearPanel(KEY_REGEX, !cbxLockRegex.getState());
      cbxLockRegex.setState(true);
    } else if(BUTTON_LOAD_TEMPLATE.equals(sWhat)) {
      loadFile(getFileFromDialog(sWhat), KEY_TEMPLATE);
    } else if(BUTTON_LOAD_REGEX.equals(sWhat)) {
      loadFile(getFileFromDialog(sWhat), KEY_REGEX);
    } else if(BUTTON_LOAD_PIZZA.equals(sWhat)) {
      loadFile(getFileFromDialog(sWhat), KEY_PROPS);
    } else if(BUTTON_CHOOSE_TEMPLATE.equals(sWhat)) {
      txfTemplateFile.setText(getFileFromDialog(sWhat));
      cbxTemplateFile.setState(Question.validString(txfTemplateFile.getText()));
    } else if(BUTTON_CHOOSE_PIZZA.equals(sWhat)) {
      txfPizzaFile.setText(getFileFromDialog(sWhat));
      cbxPizzaFile.setState(Question.validString(txfPizzaFile.getText()));
    } else if(BUTTON_GO.equals(sWhat)) {
      processGoButton();
    } else {
      CapCom.debug("Unknown command: "+sWhat);
    }
  }

  /** Build the standard MTFormatterDemo GUI. */
  public final Panel standardGUI() {
    cdlMain = new CardLayout();
    setLayout(cdlMain);
    add(filePanel(), KEY_FILE);
    add(templatePanel(), KEY_TEMPLATE);
    add(propsPanel(), KEY_PROPS);
    add(regexPanel(), KEY_REGEX);
    return this;
  }

  /** Build the File panel. */
  public final Panel filePanel() {
    Panel retval = new Panel(new BorderLayout());
    retval.add(new Label("MTFormatter Demo", Label.CENTER), BorderLayout.NORTH);
    Button btnTemp = new Button(BUTTON_NEXT); btnTemp.addActionListener(this);
    retval.add(btnTemp, BorderLayout.EAST);
    btnTemp = new Button(BUTTON_PREVIOUS); btnTemp.addActionListener(this);
    retval.add(btnTemp, BorderLayout.WEST);
    Panel pnlTemp = new Panel(new GridLayout(1, 0));
    pnlTemp.add(new Label(" "));
    btnTemp = new Button(BUTTON_GO); btnTemp.addActionListener(this);
    pnlTemp.add(btnTemp);
    pnlTemp.add(new Label(" "));
    retval.add(pnlTemp, BorderLayout.SOUTH);

    pnlTemp = new Panel(new GridBagLayout());
    GridBagConstraints gbc = new GridBagConstraints(); gbc.insets = new Insets(2, 2, 2, 2); gbc.gridx=1; gbc.gridy=1;
    gbc.anchor = GridBagConstraints.WEST; gbc.fill = GridBagConstraints.HORIZONTAL;
    pnlTemp.add(cbxTemplateFile, gbc); gbc.gridx++; gbc.gridwidth=2;
    pnlTemp.add(txfTemplateFile, gbc); gbc.gridx+=2; gbc.gridwidth=1;
    if(gotParent) {
      btnTemp = new Button("Choose"); btnTemp.setActionCommand(BUTTON_CHOOSE_TEMPLATE);
      btnTemp.addActionListener(this);
      pnlTemp.add(btnTemp, gbc);
    }
    gbc.gridy++; gbc.gridx=1;
    pnlTemp.add(cbxPizzaFile, gbc); gbc.gridx++; gbc.gridwidth=2;
    pnlTemp.add(txfPizzaFile, gbc); gbc.gridx+=2; gbc.gridwidth=1;
    if(gotParent) {
      btnTemp = new Button("Choose"); btnTemp.setActionCommand(BUTTON_CHOOSE_PIZZA);
      btnTemp.addActionListener(this);
      pnlTemp.add(btnTemp, gbc);
    }
    gbc.gridy++; gbc.gridx=2; gbc.gridwidth=2;
    pnlTemp.add(cbxEnablePizza, gbc);
    gbc.gridy++; gbc.gridx=2; gbc.gridwidth=2;
    pnlTemp.add(cbxClip, gbc); gbc.gridwidth=1;
    gbc.gridy++; gbc.gridx=1;
    pnlTemp.add(cbxEnableRegex, gbc); gbc.gridx++; gbc.gridwidth=2;
    pnlTemp.add(txfRegex, gbc); gbc.gridwidth=1;
    retval.add(pnlTemp, BorderLayout.CENTER);
    
    return retval;
  }

  /** Build the internal template panel. */
  public final Panel templatePanel() {
    Panel retval = new Panel(new BorderLayout());
    retval.add(new Label("Internal Template", Label.CENTER), BorderLayout.NORTH);
    Button btnTemp = new Button(BUTTON_NEXT); btnTemp.addActionListener(this);
    retval.add(btnTemp, BorderLayout.EAST);
    btnTemp = new Button(BUTTON_PREVIOUS); btnTemp.addActionListener(this);
    retval.add(btnTemp, BorderLayout.WEST);
    retval.add(txaTemplate, BorderLayout.CENTER);
    Panel pnlTemp = new Panel(new GridLayout(1,0));
    if(gotParent) {
      btnTemp = new Button("Load"); btnTemp.setActionCommand(BUTTON_LOAD_TEMPLATE);
      btnTemp.addActionListener(this); pnlTemp.add(btnTemp);
    } else {
      pnlTemp.add(new Label(" "));
    }
    btnTemp = new Button(BUTTON_GO); btnTemp.addActionListener(this);
    pnlTemp.add(btnTemp);
    pnlTemp.add(new Label(" "));
    pnlTemp.add(cbxLockTemplate);
    btnTemp = new Button("Clear"); btnTemp.setActionCommand(BUTTON_CLEAR_TEMPLATE);
    btnTemp.addActionListener(this); pnlTemp.add(btnTemp);
    retval.add(pnlTemp, BorderLayout.SOUTH);
    return retval;
  }

  /** Build the internal Properties panel. */
  public final Panel propsPanel() {
    Panel retval = new Panel(new BorderLayout());
    retval.add(new Label("Internal Properties", Label.CENTER), BorderLayout.NORTH);
    Button btnTemp = new Button(BUTTON_NEXT); btnTemp.addActionListener(this);
    retval.add(btnTemp, BorderLayout.EAST);
    btnTemp = new Button(BUTTON_PREVIOUS); btnTemp.addActionListener(this);
    retval.add(btnTemp, BorderLayout.WEST);
    retval.add(txaPizza, BorderLayout.CENTER);
    Panel pnlTemp = new Panel(new GridLayout(1,0));
    if(gotParent) {
      btnTemp = new Button("Load"); btnTemp.setActionCommand(BUTTON_LOAD_PIZZA);
      btnTemp.addActionListener(this); pnlTemp.add(btnTemp);
    } else {
      pnlTemp.add(new Label(" "));
    }
    btnTemp = new Button(BUTTON_GO); btnTemp.addActionListener(this);
    pnlTemp.add(btnTemp);
    pnlTemp.add(new Label(" "));
    pnlTemp.add(cbxLockProps);
    btnTemp = new Button("Clear"); btnTemp.setActionCommand(BUTTON_CLEAR_PIZZA);
    btnTemp.addActionListener(this); pnlTemp.add(btnTemp);
    retval.add(pnlTemp, BorderLayout.SOUTH);
    return retval;
  }

  /** Build the regex data panel. */
  public final Panel regexPanel() {
    Panel retval = new Panel(new BorderLayout());
    retval.add(new Label("Regular Expression Data", Label.CENTER), BorderLayout.NORTH);
    Button btnTemp = new Button(BUTTON_NEXT); btnTemp.addActionListener(this);
    retval.add(btnTemp, BorderLayout.EAST);
    btnTemp = new Button(BUTTON_PREVIOUS); btnTemp.addActionListener(this);
    retval.add(btnTemp, BorderLayout.WEST);
    retval.add(txaRegexData, BorderLayout.CENTER);
    Panel pnlTemp = new Panel(new GridLayout(1,0));
    if(gotParent) {
      btnTemp = new Button("Load"); btnTemp.setActionCommand(BUTTON_LOAD_REGEX);
      btnTemp.addActionListener(this); pnlTemp.add(btnTemp);
    } else {
      pnlTemp.add(new Label(" "));
    }
    btnTemp = new Button(BUTTON_GO); btnTemp.addActionListener(this);
    pnlTemp.add(btnTemp);
    pnlTemp.add(new Label(" "));
    pnlTemp.add(cbxLockRegex);
    btnTemp = new Button("Clear"); btnTemp.setActionCommand(BUTTON_CLEAR_REGEX);
    btnTemp.addActionListener(this); pnlTemp.add(btnTemp);
    retval.add(pnlTemp, BorderLayout.SOUTH);
    return retval;
  }

  private String getFileFromDialog(String pKey) {
    String retval = "";
    if(!gotParent) return retval;
    if(myFileDialog == null) {
      if(myParentFrame != null) {
        myFileDialog = new FileDialog(myParentFrame);
      } else {
        myFileDialog = new FileDialog(myParentDialog);
      }
    }
    myFileDialog.setTitle(pKey);
    myFileDialog.setVisible(true);
    if(Question.validString(myFileDialog.getFile())) { retval = myFileDialog.getDirectory()+myFileDialog.getFile(); }
    return retval;
  }

  /** Clear the internal template or Properties panels. Since the [Load] button might not be present the clear functions are locked. The internal machinery clears 
      based on the state of the '[]Lock' checkboxes.
      @param pWhich KEY_TEMPLATE to clear the template panel, KEY_PROPS to clear the Properties panel, KEY_REGEX to clear the regex data.
      @param pOkToClear true to really clear it, false not to.
  */
  public final void clearPanel(String pWhich, boolean pOkToClear) {
    if(!pOkToClear) return;
    if(KEY_TEMPLATE.equals(pWhich)) {
      txaTemplate.setText("");
    } else if(KEY_PROPS.equals(pWhich)) {
      txaPizza.setText("");
    } else if(KEY_REGEX.equals(pWhich)) {
      txaRegexData.setText("");
    } else { CapCom.debug("Can't clear "+pWhich); }
  }

  /** Load a file into the internal template or Properties. The pWhich parameter determines who gets the load, the pFile determines the file to load. Any errors 
      are reported at Error level.
      @param pWhere KEY_TEMPLATE to load an internal template, KEY_PROPS to load the Properties panel, KEY_REGEX to load regex data.
      @param pFile The name of the file to load.
  */
  public final void loadFile(String pFile, String pWhere) {
    if(!Question.validString(pFile)) return;
    try {
      StringBuffer sbData = new StringBuffer();
      int chIn = 0;
      FileReader frIn = new FileReader(pFile);
      while((chIn = frIn.read()) != -1) { sbData.append((char) chIn); }
      frIn.close();
      if(KEY_TEMPLATE.equals(pWhere)) {
        txaTemplate.setText(sbData.toString());
      } else if(KEY_PROPS.equals(pWhere)) {
        txaPizza.setText(sbData.toString());
      } else if(KEY_REGEX.equals(pWhere)) {
        txaRegexData.setText(sbData.toString());
      } else {
        CapCom.println(sbData.toString());
      }
    } catch(Exception ex) { CapCom.error("Error loading file: "+pFile, ex); }
  }

  /** This method actually merges the template and data. It reads the internal template or loads, depending on which is selected, collects Properties from the 
      internal set and loads the properties file if enabled, then formats the template. The result is returned as a String. Any errors are reported at Error level 
      and the return value is a blank String. If regex matching is enabled the regex on the File panel is compiled and applied to each String in the regex data 
      area; the content of the <code>TextArea</code> is split on newlines and the resulting array sent to <code>reFormat()</code>.
  */
  public String formatTemplate() {
    String retval = "";
    try {
      MathTestFormatter myFormatter = new MathTestFormatter();
      if(cbxTemplateFile.getState()) {
        myFormatter.loadTemplate(txfTemplateFile.getText(), false);
      } else {
        myFormatter.setTemplate(txaTemplate.getText());
      }
      myFormatter.addGlobalPizza();
      if(cbxPizzaFile.getState()) { myFormatter.loadProperties(txfPizzaFile.getText(), false); }
      if(cbxEnablePizza.getState()) { myFormatter.addProperties(txaPizza.getText(), null, null); }
      if(cbxEnableRegex.getState()) {
        myFormatter.setRegex(txfRegex.getText());
        String[] tokenz = txaRegexData.getText().split("\n");
        if(tokenz.length < 2) {
          retval = myFormatter.reFormat(tokenz[0]);
        } else {
          String[] resultz = myFormatter.reFormat(tokenz);
          StringBuffer sbResults = new StringBuffer();
          for(int i=0; i < resultz.length; i++) { sbResults.append(resultz[i]).append("\n"); }
          retval = sbResults.toString();
        }
      } else {
        retval = myFormatter.format();
      }
    } catch(Exception ex) { CapCom.error("Error formatting template.", ex); }
    return retval;
  }

  /** This method is called when the [Go!] button is clicked. It calls <code>formatTemplate()</code>, prints the result, and sends it to <code>CapCom.processMathObject()</code> 
      if that is enabled. For a <code>BasicMathPrinter</code> this dumps the String to the clipboard.
  */
  public void processGoButton() {
    String sResult = formatTemplate();
    if(cbxClip.getState()) CapCom.processMathObject(sResult);
    CapCom.println(sResult);
  }

}

