package cox.jmatt.java.MathTools.demo;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.sql.*;
import javax.sql.*;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.util.*;
import cox.jmatt.java.MathTools.test.MathTestFormatter;

/** <p>This class serves as a <code>MathDBC</code> demo and as a JDBC tool. Although not visible <code>MathDBC</code> is used to create a <code>Connection</code> 
    to the specified database. The Driver class name and JDBC URL are configured on the Config panel and the main area holds SQL commands to be executed. The 
    Login Properties panel holds properties that are used when establishing a <code>Driver Connection</code> and an <code>URLLoaderPanel</code> rounds out the 
    mix. The Loader panel is used to set the <code>CapCom</code> Math ClassLoader for the <code>MathDBC</code> instance to use.</p>
    <p>If the last SQL statement executed generates a <code>ResultSet</code> it can be printed via a <code>MathTestFormatter</code>. The config panel holds a 
    TextField for the template String and the output can be sent to the clipboard. (<code>BasicMathPrinter processMathObject()</code>). No formatting tokens other 
    than the <code>ResultSet</code> or built-in ones are available but those should suffice.</p>
    <p>In its second role <code>MathDBCDemo</code> allows 'raw' SQL to be executed using JDBC. While there are many other tools dedicated solely to that purpose none 
    of them are included in MathTools! This tool allows the quick creation or manipulation of a JDBC-enabled database without needing to wrap the SQL in scripting 
    language statements. This means a database can be created quickly from nothing but SQL. It is also possible to execute queries against it, quickly and from 
    within MathTools, after it has been created.</p>
    <p>Besides quick database creation <code>MathDBCDemo</code> allows SQL experimentation. This makes it possible to 'debug' SQL statements before including them 
    in scripts, without resorting to an external tool.</p>
    <h3>Configuration and Debugging</h3>
    <p><code>MathDBCDemo</code> defines a set of {@link #processMacroCommand macro commands} that allow in-line configuration and debugging. These commands are all 
    two characters in length, the first of which is a slash. All necessary information, such as Driver class name and JDBC connection URL, can be specified 
    <i>within the SQL commands</i>. The echo command ('/e') prints a message and the break command ('/x') STOPS execution. This allows SQL debugging with single 
    line progressive granularity. This means a LONG batch of SQL can be debugged without the need for massive amounts of typing or re-loading.</p>
*/
public class MathDBCDemo extends Panel implements ActionListener {

  /** This is the constant key used to select the SQL entry panel in the main display area. */
  public static final String KEY_SQL = "SQL Commands";
  /** This is the constant key used to select the <code>connect()</code> panel. */
  public static final String KEY_PROPS = "Connection Properties";
  /** This is the constant key used to select the configuration panel. */
  public static final String KEY_CONFIG = "Config";
  /** This is the constant key used to select the URL ClassLoader. */
  public static final String KEY_URL = "URL Loader";

  private static final String FORWARD = "|>>";
  private static final String BACKWARD = "<<|";

  private Dialog myDialog = null;
  private Frame myFrame = null;
  private FileDialog myFileDialog = null;
  private CardLayout cdlMain = null;
  private URLLoaderPanel myLoaderPanel = null;
  private TextArea txaScriptCon, txaPizzaCon;
  private Button btnRunScript = new Button("Execute SQL"), btnClearScript = new Button("Clear SQL"), btnClearPizza = new Button("Clear Props");
  private Button btnLoadScript = new Button("Load"), btnLoadPizza = new Button("Load");
  private Checkbox cbxLockClear = new Checkbox("Lock clear!", true), cbxUseDS = new Checkbox("Use DataStore");
  private Checkbox cbxFormatResults = new Checkbox("MTFormat ResultSet", true), cbxClip = new Checkbox("processMathObject(formattedString)", true);
  private TextField txfDriver = new TextField(40), txfConnection = new TextField(40), txfResTemplate = new TextField(40);

  /** Zero-argument constructor. <code>FileDialog</code> functionality will not be present. */
  public MathDBCDemo() { super(); }

  /** Build the demo around a <code>Dialog</code>. All functions available. */
  public MathDBCDemo(Dialog pParent) {
    super();
    myFileDialog = new FileDialog(pParent);
    myLoaderPanel = new URLLoaderPanel(pParent);
  }

  /** Build the demo around a <code>Dialog</code>. All functions available. */
  public MathDBCDemo(Frame pParent) {
    super();
    myFileDialog = new FileDialog(pParent);
    myLoaderPanel = new URLLoaderPanel(pParent);
  }

  /** This method builds the standard GUI and returns itself as a reference. It should be used in constructing the GUI. */
  public final Panel standardGUI() {
    cdlMain = new CardLayout();
    setLayout(cdlMain);
    add(SQLEntryPanel(), KEY_SQL);
    add(configPanel(), KEY_CONFIG);
    add(loginPropsPanel(), KEY_PROPS);
    Button btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); CapCom.storeObject(URLLoaderPanel.URL_PANEL_WEST, btnTemp);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this);CapCom.storeObject(URLLoaderPanel.URL_PANEL_EAST, btnTemp);
    if(myLoaderPanel == null) myLoaderPanel = new URLLoaderPanel();
    add(myLoaderPanel.standardGUI(), KEY_URL);
    return this;
  }

  /** This is the Panel used to enter the SQL commands. */
  public final Panel SQLEntryPanel() {
    Panel retval = new Panel(new BorderLayout());
    retval.add(new Label("MathDBC Demo: SQL Entry Panel", Label.CENTER), BorderLayout.NORTH);
    Button btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.WEST);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.EAST);
    txaScriptCon = new TextArea(); retval.add(txaScriptCon, BorderLayout.CENTER);
    btnRunScript.addActionListener(this); 
    if(myFileDialog != null) {
      Panel pnlTemp = new Panel(new GridLayout(1,0));
      pnlTemp.add(btnRunScript);
      pnlTemp.add(new Label(" "));
      btnLoadScript.addActionListener(this);
      pnlTemp.add(btnLoadScript);
      retval.add(pnlTemp, BorderLayout.SOUTH);
    } else {
      retval.add(btnRunScript, BorderLayout.SOUTH);
    }
    return retval;
  }

  /** This is the configuration Panel. */
  public final Panel configPanel() {
    Panel pnlConfig = new Panel(new BorderLayout());
    pnlConfig.add(new Label("MathDBC/MTFormatter Configuration Panel", Label.CENTER), BorderLayout.NORTH);
    Button btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); pnlConfig.add(btnTemp, BorderLayout.WEST);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this); pnlConfig.add(btnTemp, BorderLayout.EAST);
    Panel retval = new Panel(new GridLayout(0,1));
    Panel pnlTemp = null;
    retval.add(new Label("JDBC Driver Class Name", Label.LEFT));
    retval.add(txfDriver);
    retval.add(new Label("JDBC Database URL", Label.LEFT));
    retval.add(txfConnection);
    retval.add(new Label("ResultSet output template",Label.LEFT));
    retval.add(txfResTemplate);
    pnlTemp = new Panel(new GridLayout(1,0));
    pnlTemp.add(cbxFormatResults);
    pnlTemp.add(cbxClip);
    retval.add(pnlTemp);
    
    pnlConfig.add(retval, BorderLayout.CENTER);
    pnlTemp = new Panel(new GridLayout(1,0));
    btnClearScript.addActionListener(this); pnlTemp.add(btnClearScript);
    btnClearPizza.addActionListener(this); pnlTemp.add(btnClearPizza);
    pnlTemp.add(cbxLockClear);
    pnlConfig.add(pnlTemp, BorderLayout.SOUTH);
    return pnlConfig;
  }

  /** This Panel holds the Properties object used to enter database login parameters. */
  public final Panel loginPropsPanel() {
    Panel retval = new Panel(new BorderLayout());
    retval.add(new Label("Connection Properties Panel", Label.CENTER), BorderLayout.NORTH);
    Button btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.WEST);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.EAST);
    txaPizzaCon = new TextArea(); retval.add(txaPizzaCon, BorderLayout.CENTER);
    Panel pnlTemp = new Panel(new GridLayout(1,0));
    pnlTemp.add(new Label(" "));
    if(myFileDialog != null) {
      btnLoadPizza.addActionListener(this);
      pnlTemp.add(btnLoadPizza);
    }
    pnlTemp.add(new Label(" "));
    retval.add(pnlTemp, BorderLayout.SOUTH);
    return retval;
  }

  /** Handle all button clicks. */
  public final void actionPerformed(ActionEvent aev) {
    String sCommand = aev.getActionCommand();
    Object oCmd = aev.getSource();
    if(sCommand.equals(FORWARD)) { cdlMain.next(this); return; }
    if(sCommand.equals(BACKWARD)) { cdlMain.previous(this); return; }
    if(oCmd == btnLoadScript) { loadFile(KEY_SQL); return; }
    if(oCmd == btnLoadPizza) { loadFile(KEY_PROPS); return; }
    if(oCmd == btnRunScript) { executeSQL(); return; }
    if(oCmd == btnClearScript) { clearSQL(!cbxLockClear.getState()); cbxLockClear.setState(true); return; }
    if(oCmd == btnClearPizza) { clearProps(!cbxLockClear.getState()); cbxLockClear.setState(true); return; }
  }

  /** Parse the config Properties object and return it. */
  public Properties getPizza() {
    Properties thePizza = new Properties();
    try {
      StringReader srIn = new StringReader(txaPizzaCon.getText());
      thePizza.load(srIn);
      srIn.close();
    } catch(Exception ex) { CapCom.error("Config Props parse error", ex); }
    return thePizza;
  }
  /** Load a script or pizza from a FileDialog. This method only works if its ScriptPanel was created around a Frame or a Dialog.
  @param pKey The KEY_ constant determining where to load.
  */
  public void loadFile(String pKey) {
    if((myFileDialog == null) || !Question.validString(pKey)) return;
    myFileDialog.setTitle("Load "+pKey);
    myFileDialog.setVisible(true);
    if(myFileDialog.getFile() == null) return;
    loadFile(myFileDialog.getDirectory()+myFileDialog.getFile(), pKey);
  }
  /** Load a script or Properties from a file. This method expects a text file as it will attempt to create a FileReader and copy it to the specified Area. Any 
      Exceptions are reported at the Error level. If the filename or key is null or empty this method returns silently. The pKey parameter determines where to 
      send the load: KEY_SQL loads SQL, KEY_PROPS loads into the config panel. If pKey is not valid or not recognized the load goes to CapCom.println();
      @param pFile The name of the file to load.
      @param pKey The place to send the loaded file.
  */
  public void loadFile(String pFile, String pKey) {
    if(!Question.validString(pFile)) return;
    CapCom.debug("Loading: "+pFile);
    try {
      BufferedReader fileIn = new BufferedReader(new FileReader(pFile));;
      StringWriter strOut = new StringWriter();
      String sLine = null;
      while((sLine = fileIn.readLine()) != null) {
        strOut.write(sLine+"\n");
      }
      strOut.flush();
      strOut.close();
      fileIn.close();
      if(KEY_SQL.equals(pKey)) {
        setSQL(strOut.toString());
      } else if(KEY_PROPS.equals(pKey)) {
        txaPizzaCon.setText(strOut.toString());
      } else { CapCom.println(strOut.toString()); }
    } catch(Exception ex) { CapCom.error("Error loading "+pFile, ex); }
  }
  /** Fetch the current SQL commands. */
  public String getSQL() { return txaScriptCon.getText(); }
  /** Put SQL commands in the SQL Entry Area. */
  public void setSQL(String pScript) { txaScriptCon.setText(pScript); }
  /** Get the current driver name. */
  public String getDriver() { return txfDriver.getText(); }
  /** Set the name of the JDBC driver. */
  public void setDriver(String pDriver) { txfDriver.setText(pDriver); }
  /** Get the current database URL. */
  public String getJDBC() { return txfConnection.getText(); }
  /** Set the database URL. */
  public void setJDBC(String pJDBC) { txfConnection.setText(pJDBC); }
  /** Get the current row format template. */
  public String getTemplate() { return txfResTemplate.getText(); }
  /** Set the row format template. */
  public void setTemplate(String pTemplate) { txfResTemplate.setText(pTemplate); }

  /** Clear the SQL area. The default implementation clears on the state of the '[]Lock all clear functions' Checkbox.
      @param pOkToClear true to REALLY clear the area, false not to.
  */
  public void clearSQL(boolean pOkToClear) {
    if(pOkToClear) txaScriptCon.setText("");
  }
  /** Clear the parameter area. The default implementation clears on the state of the '[]Lock all clear functions' Checkbox.
      @param pOkToClear true to REALLY clear the config props.
  */
  public void clearProps(boolean pOkToClear) {
    if(pOkToClear) txaPizzaCon.setText("");
  }

  /** <p>This method executes the SQL statements in the main Panel. If the last statement executed returns a ResultSet and the 'format result' Checkbox is enabled 
      it is sent to a <code>MathTestFormatter</code> for formatting. The formatting template is entered in the 'template' TextArea. If the 'processMathObject()' 
      Checkbox is enabled the resulting String is sent to <code>CapCom.processMathObject()</code>, which sends it to the clipboard.</p>
  */
  public void executeSQL() {
    String theSQL = getSQL();
    if(!Question.validString(theSQL)) return;
    String[] commandz = theSQL.split("\n");
    int iStart = 0;
    for(iStart = 0; (iStart < commandz.length) && ((commandz[iStart]+" ").startsWith("--")); iStart++) {
      if(!processMacroCommand(commandz[iStart].substring(2).trim())) return;
    }
    CapCom.setMathClassLoader(myLoaderPanel.getURLClassLoader());
    MathDBC myDBC = new MathDBC();
    Properties myProps = getPizza();
    myDBC.addProperties(myProps);
    Connection myConnection = null;
    if(cbxUseDS.getState()) {
      if(!myDBC.isDataSourceAvailable(getJDBC())) return;
      CapCom.debug("Got DataSource");
      String userName = myProps.getProperty("user", null), passWord = myProps.getProperty("password", null);
      if((Question.validString(userName) && Question.validString(passWord))) {
        myConnection = myDBC.getDSConnection(userName, passWord);
      } else {
        myConnection = myDBC.getDSConnection();
      }
    } else {
      if(!myDBC.setDriver(getDriver())) { CapCom.debug("Unable to set driver"); return; }
      myConnection = myDBC.getConnection(getJDBC());
    }
    if(myConnection == null) { CapCom.debug("Unable to get Connection"); return; }
    try {
      Statement stmt = myConnection.createStatement();
      CapCom.debug("Starting SQL execution");
      boolean trucking = true;
      for(int i=iStart; trucking && (i < commandz.length); i++) {
        if(!Question.validString(commandz[i])) continue;
        if(commandz[i].startsWith("--")) {
          trucking = processMacroCommand(commandz[i].substring(2).trim());
        } else {
          stmt.execute(commandz[i].trim());
        }
      }
      ResultSet rsDone = stmt.getResultSet();
      if(trucking && cbxFormatResults.getState() && (rsDone != null)) {
        CapCom.debug("Formatting ResultSet");
        MathTestFormatter myFormatter = new MathTestFormatter();
        myFormatter.addGlobalPizza();
        myFormatter.setTemplate(txfResTemplate.getText());
        String theFormatted = myFormatter.format(rsDone);
        if(cbxClip.getState()) CapCom.processMathObject(theFormatted);
        CapCom.println(theFormatted);
      }
    } catch(Exception ex) {
      CapCom.error("Error executing SQL", ex);
    }
    MathDBC.closeAll();
    CapCom.debug("Done");
  }

  /** <p>This method processes any macro commands embedded within SQL comments. The macro commands allow configuration values to be set. The <b>macro</b> commands 
      all start with a slash character ('/') and consist of a single letter followed by the data to be configured. The command <b>line</b> must start with a double 
      hyphen followed by a single space: '-- ' and then the command. The macro commands are as follows.
      <ul>
        <li>'/d <i>Driver.class.name</i>' sets the JDBC <code>Driver</code>.</li>
        <li>'/j <i>jdbcURL</i>' sets the JDBC connection URL.</li>
        <li>'/t <i>Template String</i>' sets the row formatting template.</li>
        <li>'/u <i>URL String</i>' adds an URL to the URLLoaderPanel.</li>
        <li>'/p <i>key = Property Value</i>' adds a property to the login Properties panel. The equals sign is critical!</li>
        <li>'/e <i>[Message]</i>' prints a line with a message (optional).</li>
        <li>'/x <i>[Message]</i>' STOPS execution. The message is printed if present.</li>
        <li>'//' is a comment. This line and any text on it is ignored.</li>
      </ul></p>
      <p>Uppercase letters will also work.</p>
      <p>NOTE: The macro commands <i>are</i> embedded within SQL line comments. In order to be executed the (trimmed) line must start with a double hyphen followed 
      by at least one space. The macro parser ignores any line that does not begin '-- ' and further ignores lines whose fourth character is not a slash.</p>
      @param pCmd The macro command to execute.
      @return false on 'break' ('/x'), true otherwise.
  */
  public boolean processMacroCommand(String pCmd) {
    if(!(Question.validString(pCmd) && pCmd.startsWith("/"))) return true;
    switch((pCmd+"  ").charAt(1)) {
      case 'd': case 'D':
        setDriver((pCmd+"   ").substring(2).trim());
        break;
      case 'j': case 'J':
        setJDBC((pCmd+"   ").substring(2).trim());
        break;
      case 't': case 'T':
        setTemplate((pCmd+"   ").substring(2).trim());
        break;
      case 'u': case 'U':
        myLoaderPanel.addURL((pCmd+"   ").substring(2));
        break;
      case 'p': case 'P':
        if(pCmd.trim().length() > 3) txaPizzaCon.append(pCmd.substring(2).trim()+"\n");
        break;
      case 'e': case 'E':
        CapCom.println("[ECHO] "+(pCmd+"  ").substring(2).trim());
        break;
      case 'x': case 'X':
        CapCom.println("[BREAK] "+(pCmd+"  ").substring(2).trim());
        return false;
      default:
    }
    return true;
  }

}

