package cox.jmatt.java.MathTools.demo;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import cox.jmatt.java.MathTools.CapCom;
import cox.jmatt.java.MathTools.Question;

/** <p>This class provides an easy way to specifiy polymorphic configurable parameters to test MathTools (or other class) methods. The parameters in the TextArea 
    can be edited to appropriate values. The program can then call an appropriate 'as__(<i>N</i>) method to retrieve the parameter. The value <i>N</i> is the 
    index of the parameter to fetch. Example: asLong(4) fetches 'par.4=...' as a long.</p>
    <p>This class may be used as-is to supply parameters to another class or it may be extended. The processGoButton() is the extension method. The default 
    implementation merely eats the pizza (Properties) in the TextArea. The getNorthPanel() method provides a quick and easy way to provide a simple GUI.</p>
    <p>NOTE: The eatPizza() method MUST be called before meaningful values can be extracted from the parameters. The default values for all as__() methods are: 
    0 for numeric methods, '' for String, null for <code>asNullString()</code> and false for boolean.</p>
*/
public class PizzaPanel extends Panel implements ActionListener {

  private TextArea txaParams = new TextArea();
  private Button btnClear = new Button("Clear"), btnReset = new Button("Reset params");
  private Button btnGo = new Button("Go!");
  private TextField txfNParams = new TextField();
  private Properties myProps = new Properties();
  

  public PizzaPanel() {
    super();
    setLayout(new BorderLayout());
    add(getNorthPanel(), BorderLayout.NORTH);
    add(txaParams, BorderLayout.CENTER);
    Panel pnlTemp = new Panel(new GridLayout(1,0));
    btnGo.addActionListener(this);
    pnlTemp.add(new Label(" ")); pnlTemp.add(btnGo);
    pnlTemp.add(new Label(" "));
    pnlTemp.add(btnClear);
    btnReset.addActionListener(this);
    pnlTemp.add(btnReset);
    pnlTemp.add(txfNParams);
    btnClear.addActionListener(this);
    add(pnlTemp, BorderLayout.SOUTH);
    txfNParams.setText("10");
    resetParams(10);
  }
  /** Clear the panel and do NOT set any parameters. This is default behavior for the '[Clear]' buttons. */
  public void clearParams() { txaParams.setText(""); }
  /** Clear the panel and set the number of (blank) parameters specified.
      @param pNPars The number of parameter spots to display.
  */
  public void resetParams(int pNPars) {
    txaParams.setText("");
    for(int i=0; i < pNPars; i++) {
      txaParams.append("par."+i+"=\n");
    }
  }
  /** Clear the panel and set the number of parameters to whatever is in the lower right TextField. This is what the '[Reset params]' button does. */
  public void resetParams() {
    int nPars = 10;
    try {
      nPars = Integer.parseInt(txfNParams.getText().trim());
    } catch(Exception ignored) {}
    resetParams(nPars);
  }
  /** Retrieve the component to be displayed in the NORTH part of the Panel. Defaults to a title Label. */
  public Panel getNorthPanel() {
    Panel retval = new Panel();
    retval.add(new Label("Parameter Panel"));
    return retval;
  }
  /** Used for the [Clr] and [Go!] buttons. */
  public void actionPerformed(ActionEvent aev) {
    if(aev.getSource() == btnClear) clearParams();
    if(aev.getSource() == btnGo) processGoButton();
    if(aev.getSource() == btnReset) resetParams();
  }
  /** Parse the currently-set parameters into the internal Properties object. This method should be called any time the parameters might have changed. Returns 
      a copy of the Properties object generated to allow access to typed-in properties that do not follow the standard pattern. (The standard pattern is 
      'par.<i>X</i> where <i>X</i> is a number.)
      @return A fresh copy of the Properties parsed in.
  */
  public Properties eatPizza() {
    myProps = new Properties();
    try {
      StringReader srIn = new StringReader("  "+txaParams.getText());
      myProps.load(srIn);
      srIn.close();
    } catch(Exception ex) { cox.jmatt.java.MathTools.CapCom.error("Error loading PizzaPanel", ex); }
    Properties retval = new Properties();
    retval.putAll(myProps);
    return retval;
  }

  /** Fetch the contents of the pizza area unaltered. */
  public String getRawPizza() { return txaParams.getText(); }

  /** Add to the content of the parameter area. If pContent is non-null and non-empty it is appended to the content of the parameter area. No line processing is 
      done so any desired newlines must be handled manually.
  */
  protected void _addToPizza(String pContent) {
    if(Question.validString(pContent)) txaParams.append(pContent);
  }
  
  /** This method is called when the [Go!] button is clicked. Default implementation calls eatPizza(). */
  public void processGoButton() { eatPizza(); }

  /** Interpret the parameter 'par.<i>iNdx</i> as a long. Defaults to 0. */
  public long asLong(int iNdx) {
    long retval = 0;
    try { retval = Long.parseLong(myProps.getProperty("par."+iNdx, "0")); } catch(Exception ex) {}
    return retval;
  }
  /** Interpret the parameter 'par.<i>iNdx</i> as an int. Defaults to 0. */
  public int asInt(int iNdx) {
    int retval = 0;
    try { retval = Integer.parseInt(myProps.getProperty("par."+iNdx, "0")); } catch(Exception ex) {}
    return retval;
  }
  /** Interpret the parameter 'par.<i>iNdx</i> as a double. Defaults to 0.0. */
  public double asDouble(int iNdx) {
    double retval = 0.0;
    try { retval = Double.parseDouble(myProps.getProperty("par."+iNdx, "0.0")); } catch(Exception ex) {}
    return retval;
  }
  /** Interpret the parameter 'par.<i>iNdx</i> as a String. Defaults to a blank String. */
  public String asString(int iNdx) {
    String retval = " ";
    try { retval = myProps.getProperty("par."+iNdx, ""); } catch(Exception ex) {}
    return retval;
  }
  /** Interpret the parameter 'par.<i>iNdx</i> as a String. Defaults to null if no value is present. */
  public String asNullString(int iNdx) {
    String retval = null;
    try { retval = myProps.getProperty("par."+iNdx, null); } catch(Exception ex) {}
    return cox.jmatt.java.MathTools.Question.validString(retval)?retval:null;
  }
  /** Interpret the parameter 'par.<i>iNdx</i> as a boolean. Defaults to false. */
  public boolean asBoolean(int iNdx) {
    boolean retval = false;
    try { retval = Boolean.parseBoolean(myProps.getProperty("par."+iNdx, "false")); } catch(Exception ex) {}
    return retval;
  }

}

