package cox.jmatt.java.MathTools.demo;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.net.*;
import javax.script.*;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.util.*;

/** <p>ScriptPanel provides a very minimal scripting framework for the MathTools packages. It uses the javax.scripting.* API. The default ScriptPanel can be 
    used as-is but it is also easily extended. The getEngineLoader(), scriptStarting() and scriptEnded() methods are designed to be overridden. The scriptStarting() 
    method is called before script execution and has a chance to veto execution. The scriptEnded() method is called after the script stops, whether through completion 
    or Exception.</p>
    <p>If a custom ClassLoader is required, override getEngineLoader(). By default this method returns null. If, however, the '[]Enable URLClassLoader' box is checked, 
    ScriptPanel attempts to build an URLClassLoader around the list of URLs provided on the URL Loader Panel.</p>
    <p>For a list of the tools exposed and their names see {@link cox.jmatt.java.MathTools.CapCom#installStandardObjects CapCom.installStandardObjects()}.</p>
    <p>Additionally, ScriptPanel sets the Global Pizza from the Pizza Panel before the script is executed. This makes parameters available through Console's 
    getProperty() methods. The '[]Enable Console.setAuthenticator()' checkbox is used to set the <code>CapCom.allowConsoleAuthenticatorAccess</code> field.</p>
    <p>When the script succeeds there are options for its return Object. The default implementation of scriptEnded() calls toString() on the (<i>non-null!</i>) 
    return Object and prints it to the console. If the []processMathObject checkbox is enabled, the 'raw' return Object is processed via 
    <code>CapCom.processMathObject()</code> prior to being String'ed. If a <code>BasicMathPrinter</code> is in place the String content of the returned Object 
    is sent to the clipboard.</p>
*/
public class ScriptPanel extends Panel implements ActionListener {

  /** This is the constant key used to select the Script Source panel in the main display area. */
  public static final String KEY_SCRIPT = "Script";
  /** This is the constant key used to select the Script Pizza (Parameters) panel in the main display area. */
  public static final String KEY_PIZZA = "Pizza";
  /** This is the constant key used to select the Script Configuration panel in the main display area. */
  public static final String KEY_CONFIG = "Config";
  /** This is the constant key used to select the URL ClassLoader panel in the display area. */
  public static final String KEY_URL = "URL Loader";

  private static final String FORWARD = "|>>";
  private static final String BACKWARD = "<<|";
  private static final String KEY_PICK_URL = "URL (filename) into list";
  private static final String KEY_LOAD_URL_LIST = "URL list from file";

  private CardLayout cdlDeck;
  private TextArea txaScriptCon, txaPizzaCon;
  private Button btnRunScript = new Button("Run Script"), btnClearScript = new Button("Clear Script Area"), btnClearPizza = new Button("Clear Parameter Area");
  private Button btnClearURLS = new Button("Clear URL List"), btnCheckEnv = new Button("Check Environment"), btnExtractPizza = new Button("Extract...");
  private Button btnSetPizza = new Button("Send to Global Pizza");
  private Button btnLoadScript = new Button("Load"), btnLoadPizza = new Button("Load");
  private Checkbox cbxLockClear = new Checkbox("Lock all clear functions!", true);
  private CheckboxGroup cbgEngine = new CheckboxGroup();
  private Checkbox cbxByName = new Checkbox("by Name", cbgEngine, true);
  private Checkbox cbxByExt = new Checkbox("by Extension", cbgEngine, false);
  private Checkbox cbxByMime = new Checkbox("by Mimetype", cbgEngine, false);
  private Checkbox cbxURL = new Checkbox("Enable URLClassLoader", false);
  private Checkbox cbxConsoleAuth = new Checkbox("Enable Console.setAuthenticator()", true);
  private TextField txfEngineBy = new TextField("javascript"), txfExtractPizza = new TextField("#PARAM");
  private Checkbox cbxProcessObject = new Checkbox(" processMathObject(result)", true);
  private FileDialog myFileDialog = null;
  private URLLoaderPanel myLoaderPanel = null;

  public ScriptPanel() {
    super();
  }
  /** Create a ScriptPanel anchored to a Frame. This provides an anchor for the FileDialog used to load things.
      @param pParent The Frame used to anchor a FileDialog.
  */
  public ScriptPanel(Frame pParent) {
    super();
    myFileDialog = new FileDialog(pParent);
    myLoaderPanel = new URLLoaderPanel(pParent);
  }
  /** Create a ScriptPanel anchored to a Dialog. This provides an anchor for the FileDialog used to load things.
      @param pParent The Dialog used to anchor a FileDialog.
  */
  public ScriptPanel(Dialog pParent) {
    super();
    myFileDialog = new FileDialog(pParent);
    myLoaderPanel = new URLLoaderPanel(pParent);
  }
  
  /** Standard GUI is provided to allow FULL construction before messing with the CardLayout. To do so in one step: '(new ScriptPanel()).standardGUI();'. ScriptPanel 
      automatically adds a Frame-anchored instance of URLLoaderPanel under KEY_URL.
  */
  public final Panel standardGUI() {
    cdlDeck = new CardLayout();
    setLayout(cdlDeck);
    add(scriptSourcePanel(), KEY_SCRIPT);
    add(scriptPizzaPanel(), KEY_PIZZA);
    add(scriptConfigPanel(), KEY_CONFIG);
    Button btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); CapCom.storeObject(URLLoaderPanel.URL_PANEL_WEST, btnTemp);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this);CapCom.storeObject(URLLoaderPanel.URL_PANEL_EAST, btnTemp);
    if(myLoaderPanel == null) myLoaderPanel = new URLLoaderPanel();
    add(myLoaderPanel.standardGUI(), KEY_URL);
    return this;
  }
  /** The Script Source panel holds the script plus machinery to run and otherwise support it. It is added to the CardLayout using KEY_SCRIPT. */
  public final Panel scriptSourcePanel() {
    Panel retval = new Panel(new BorderLayout());
    retval.add(new Label("Script Source Panel", Label.CENTER), BorderLayout.NORTH);
    Button btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.WEST);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.EAST);
    txaScriptCon = new TextArea(); retval.add(txaScriptCon, BorderLayout.CENTER);
    btnRunScript.addActionListener(this); 
    if(myFileDialog != null) {
      Panel pnlTemp = new Panel(new GridLayout(1,0));
      pnlTemp.add(btnRunScript);
      pnlTemp.add(new Label(" "));
      btnLoadScript.addActionListener(this);
      pnlTemp.add(btnLoadScript);
      retval.add(pnlTemp, BorderLayout.SOUTH);
    } else {
      retval.add(btnRunScript, BorderLayout.SOUTH);
    }
    
    return retval;
  }
  /** The Script Pizza panel holds a parameter panel. When the script is executed this is set as the Global Pizza. Select using KEY_PIZZA. */
  public final Panel scriptPizzaPanel() {
    Panel retval = new Panel(new BorderLayout());
    retval.add(new Label("Script Pizza Panel", Label.CENTER), BorderLayout.NORTH);
    Button btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.WEST);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.EAST);
    txaPizzaCon = new TextArea(); retval.add(txaPizzaCon, BorderLayout.CENTER);
    btnSetPizza.addActionListener(this);
    if(myFileDialog != null) {
      Panel pnlTemp = new Panel(new GridLayout(1,0));
      pnlTemp.add(btnSetPizza);
      btnExtractPizza.addActionListener(this);
      pnlTemp.add(btnExtractPizza);
      pnlTemp.add(txfExtractPizza);
      btnLoadPizza.addActionListener(this);
      pnlTemp.add(btnLoadPizza);
      retval.add(pnlTemp, BorderLayout.SOUTH);
    } else {
      retval.add(btnSetPizza, BorderLayout.SOUTH);
    }
    return retval;
  }
  /** The Script Config Panel has all the configuration options necessary to run ScriptPanel. Select with KEY_CONFIG. */
  public final Panel scriptConfigPanel() {
    Panel retval = new Panel(new BorderLayout());
    retval.add(new Label("Script Configuration Panel", Label.CENTER), BorderLayout.NORTH);
    Button btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.WEST);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.EAST);
    Panel pnlTemp = new Panel(new GridLayout(0,2));
    pnlTemp.add(new Label("Load engine...", Label.CENTER)); pnlTemp.add(txfEngineBy); 
    pnlTemp.add(cbxByName); pnlTemp.add(cbxByExt); pnlTemp.add(cbxByMime);
    btnCheckEnv.addActionListener(this); pnlTemp.add(btnCheckEnv);
    pnlTemp.add(cbxConsoleAuth);
    pnlTemp.add(cbxURL); pnlTemp.add(cbxProcessObject); pnlTemp.add(cbxLockClear);
    btnClearScript.addActionListener(this); pnlTemp.add(btnClearScript);
    btnClearPizza.addActionListener(this); pnlTemp.add(btnClearPizza);
    retval.add(pnlTemp, BorderLayout.CENTER);
    return retval;
  }

  /** Specifically select one of the panels to show.
      @param pKey One of the KEY_ constants to select a panel.
  */
  public void selectPanel(String pKey) {
    if(Question.validString(pKey)) cdlDeck.show(this, pKey);
  }

  /** This method runs a check on the MathTools environment. It checks available XML tools (SAX, DOM etc.) and scripting languages (ScriptEngineManager). The 
      results are printed and sent to the clipboard via <code>processMathObject()</code>.
  */
  public void checkEnvironment() {
    StringBuffer sbOut = new StringBuffer("----- Testing MathTools Environment -----\n");
    cox.jmatt.java.MathTools.util.MTXProcessor mtxPro = new cox.jmatt.java.MathTools.util.MTXProcessor();
    sbOut.append("XML\tSAX available: ").append(mtxPro.isAvailable()).append(".\n");
    sbOut.append("\tXSLT available: ").append(mtxPro.isTransformAvailable()).append(".\n");
    sbOut.append("\tDOM available: ").append(mtxPro.isDOMAvailable()).append(".\n");
    sbOut.append("\tXPath available: ").append(mtxPro.isXPathAvailable()).append(".\n");
    sbOut.append("Math ClassLoader\n");
    ClassLoader myLoader = CapCom.getMathClassLoader();
    sbOut.append("\tCurrent: ").append(myLoader.toString()).append("\n");
    if(myLoader instanceof java.net.URLClassLoader) {
      sbOut.append("\t-- URL List --\n");
      for(java.net.URL theURL : ((URLClassLoader) myLoader).getURLs()) {
        sbOut.append("\t>> ").append(theURL.toString()).append("\n");
      }
    }
    sbOut.append("Scripting Languages\n");
    for(ScriptEngineFactory sFact : (new ScriptEngineManager(myLoader)).getEngineFactories()) {
      sbOut.append("\t").append(sFact.getLanguageName()).append("\n");
    }
    sbOut.append("----- Environment Check Complete -----\n");
    CapCom.processMathObject(sbOut);
    CapCom.println(sbOut.toString());
  }

  /** Handle all button clicks. When overriding this method, be sure to call super.actiionPerformed() to keep the panel working! */
  public void actionPerformed(ActionEvent aev) {
    if(FORWARD.equals(aev.getActionCommand())) {
      cdlDeck.next(this);
    } else if(BACKWARD.equals(aev.getActionCommand())) {
      cdlDeck.previous(this);
    } else if(aev.getSource() == btnClearScript) {
      clearScript(cbxLockClear.getState());
    } else if(aev.getSource() == btnClearPizza) {
      clearPizza(cbxLockClear.getState());
    } else if(aev.getSource() == btnSetPizza) {
      deliverPizza();
    } else if(aev.getSource() == btnRunScript) {
      runScript();
    } else if(aev.getSource() == btnLoadScript) {
      loadFile(KEY_SCRIPT);
    } else if(aev.getSource() == btnLoadPizza) {
      loadFile(KEY_PIZZA);
    } else if(aev.getSource() == btnCheckEnv) {
      checkEnvironment();
    } else if(aev.getSource() == btnExtractPizza) {
      extractPizza();
    }
  }

  /** Clear the script area. The default implementation clears on the state of the '[]Lock all clear functions' Checkbox.
      @param pOkToClear true to REALLY clear the script area, false not to.
  */
  public void clearScript(boolean pOkToClear) {
    if(cbxLockClear.getState()) {
      CapCom.info("Script [Clear] is locked.");
      return;
    }
    cbxLockClear.setState(true);
    txaScriptCon.setText("");
  }
  /** Clear the parameter area. The default implementation clears on the state of the '[]Lock all clear functions' Checkbox.
      @param pOkToClear true to REALLY clear the pizza area, false not to.
  */
  public void clearPizza(boolean pOkToClear) {
    if(cbxLockClear.getState()) {
      CapCom.info("Pizza [Clear] is locked.");
      return;
    }
    cbxLockClear.setState(true);
    txaPizzaCon.setText("");
    txfExtractPizza.setText("#PARAM");
  }
  /** Parse the Script Pizza area into a Properties object and set it as the Global Pizza. */
  public void deliverPizza() {
    try {
      Properties thePizza = new Properties();
      StringReader srIn = new StringReader(txaPizzaCon.getText());
      thePizza.load(srIn);
      srIn.close();
      CapCom.setPizza(thePizza);
    } catch(Exception ex) { CapCom.error("ScriptPizza parse error", ex); }
    CapCom.debug("Global Pizza delivered.");
  }
  /** Extract parameters from the Script Panel and put them into the Pizza Panel. This is done according to the 'prefix' TextField next to the [Extract] Button.
      Anything starting with that is extracted, the prefix stripped and the rest placed into the Pizza Panel. The default value is '#PARAM'.
  */
  public void extractPizza() {
    String sPrefix = txfExtractPizza.getText();
    String sScript = getScript();
    if(!Question.validString(sPrefix)) return;
    CapCom.debug("Extracting: "+sPrefix);
    if(!Question.validString(sScript)) {
      CapCom.debug("Pizza box is empty.");
      return;
    }
    String[] linez = sScript.split("\n");
    for(int i=0; i < linez.length; i++) {
      if(Question.validString(linez[i]) && (linez[i].trim().startsWith(sPrefix))) txaPizzaCon.append((linez[i].trim().substring(sPrefix.length())).trim()+"\n");
    }
  }
  /** Load a script or pizza from a FileDialog. This method only works if its ScriptPanel was created around a Frame or a Dialog.
  @param pKey The KEY_ constant determining where to load.
  */
  public void loadFile(String pKey) {
    if((myFileDialog == null) || !Question.validString(pKey)) return;
    myFileDialog.setTitle("Load "+pKey);
    myFileDialog.setVisible(true);
    if(myFileDialog.getFile() == null) return;
    loadFile(myFileDialog.getDirectory()+myFileDialog.getFile(), pKey);
  }
  /** Load a script or pizza from a file. This method expects a text file as it will attempt to create a FileReader and copy it to the Script Area. Any Exceptions 
      are reported at the Error level. If the filename or key is null or empty this method returns silently. The pKey parameter determines where to send the load:
      KEY_SCRIPT loads a script, KEY_PIZZA loads into the parameter panel. If pKey is not valid or not recognized the load goes to CapCom.println();
      @param pFile The name of the file to load.
      @param pKey The place to send the loaded file.
  */
  public void loadFile(String pFile, String pKey) {
    if(!Question.validString(pFile)) return;
    CapCom.debug("Loading: "+pFile);
    try {
      BufferedReader fileIn = new BufferedReader(new FileReader(pFile));;
      StringWriter strOut = new StringWriter();
      String sLine = null;
      while((sLine = fileIn.readLine()) != null) {
        strOut.write(sLine+"\n");
      }
      strOut.flush();
      strOut.close();
      fileIn.close();
      if(KEY_SCRIPT.equals(pKey)) {
        setScript(strOut.toString());
      } else if(KEY_PIZZA.equals(pKey)) {
        txaPizzaCon.setText(strOut.toString());
      } else { CapCom.println(strOut.toString()); }
    } catch(Exception ex) { CapCom.error("Error loading "+pFile, ex); }
  }
  /** Fetch the script from the Script Area. */
  public String getScript() { return txaScriptCon.getText(); }
  /** Put a script in the Script Area. */
  public void setScript(String pScript) { txaScriptCon.setText(pScript); }
  /** <p>This method is called when the [Run Script] button is clicked. It first calls scriptStarting(). If scriptStarting() returns true, script execution 
      continues. First a ScriptEngineFactory is instantiated using the provided ClassLoader, if provided. A ScriptEngine is then created using the method selected 
      on the Config panel. If the data field is null or empty a Javascript engine is created. If this fails, the process stops with an Error message.</p>
      <p>The next method call is to scriptStarting(). If it returns false the process stops. Otherwise, the standard Objects are added via the Engine's put() 
      method. Once that is done, the script is executed using the engine's eval() method. This part of the code is in a try/catch block so any Exceptions thrown 
      by the script won't crash the entire application. Instead, if an Exception is thrown, it is reported to scriptEnded().</p>
      <p>The exact sequence in which things happen is:
        <ol>
          <li>The script is checked for validity. If it is null or empty this method returns silently.</li>
          <li>CapCom's MathClassLoader is set from the URLLoaderPanel.</li>
          <li>A ScriptEngineManager is instantiated, using <code>getEngineLoader()</code> if it provides one.</li>
          <li>A ScriptEngine is created by the method selected. If the TextField is empty, a Javascript engine is attempted.</li>
          <li><code>configureEngine()</code> is called.</li>
          <li>The Global Pizza is set from the Pizza Panel.</li>
          <li><code>scriptStarting()</code> is called. If it returns false execution stops and no Exception is thrown.</li>
          <li>The Standard Objects are put() into the ScriptEngine via <code>CapCom.installStandardObjects()</code>.</li>
          <li>Set <code>CapCom allowConsoleAuthenticatorAccess</code> and <code>allowAccessories</code>.</li>
          <li>The script is executed via <code>ScriptEngine.eval()</code>.</li>
          <li><code>MathDBC.closeAll()</code> is called, regardless of any Exception thrown.</li>
          <li><code>MathFileIO.closeAll()</code> is called. This happens regardless of any Exception thrown.</li>
          <li>The <code>scriptEnded()</code> method is called with the return value from the script and null for the Exception.</li>
        </ol>
      <p>If the script itself is null or empty none of this happens! <code>runScript()</code> returns silently. If an Exception occurs anywhere in the process 
      (except false from <code>scriptStarting()</code>) the <code>scriptEnded()</code> method is called immediately with null for the Object and the Exception 
      that was thrown.</p>
  */
  public void runScript() {
    String theScript = getScript();
    if(!Question.validString(theScript)) return;
    Object retval = null;
    try {
      if(myLoaderPanel != null) myLoaderPanel.setMathClassLoader();
      ClassLoader theLoader = getEngineLoader();
      ScriptEngineManager theManager = new ScriptEngineManager();
      if(theLoader != null) theManager = new ScriptEngineManager(theLoader);
      String sEngineBy = txfEngineBy.getText().trim();
      ScriptEngine theEngine = null;
      if(cbxByName.getState() && Question.validString(sEngineBy)) {
        theEngine = theManager.getEngineByName(sEngineBy);
      } else if(cbxByExt.getState() && Question.validString(sEngineBy)) {
        theEngine = theManager.getEngineByExtension(sEngineBy);
      } else if(cbxByMime.getState() && Question.validString(sEngineBy)) {
        theEngine = theManager.getEngineByMimeType(sEngineBy);
      } else {
        theEngine = theManager.getEngineByName("javascript");
      }
      theEngine = configureEngine(theEngine);
      if(theEngine == null) throw new NullPointerException("Unable to instantiate ScriptEngine");
      deliverPizza();
      if(!scriptStarting()) return;
      CapCom.installStandardObjects(theEngine, true, true);
      CapCom.allowAccessories = true;
      CapCom.allowConsoleAuthenticatorAccess = cbxConsoleAuth.getState();
      retval = theEngine.eval(theScript);
      CapCom.allowAccessories = false;
      MathDBC.closeAll();
      MathFileIO.closeAll();
    } catch(Exception ex) {
      MathDBC.closeAll();
      MathFileIO.closeAll();
      scriptEnded(null, ex);
    }
    scriptEnded(retval, null);
  }

  /** Call this method to check the state of the '[]Enable URLClassLoader' box. */
  public final boolean isClassLoaderEnabled() { return cbxURL.getState(); }

  /** ScriptPanel calls this method before instaitiating a ScriptEngineFactory. If the '[]Enable URLClassLoader' box is unchecked or an error occurs this method 
      returns null. The return value, if not null, is used to create a ScriptEngineFactory. If the 'Enable' box is checked, the ClassLoader returned is the one 
      stored by {@link cox.jmatt.java.MathTools.CapCom#setMathClassLoader CapCom}.
      @return A ClassLoader to use for creating ScriptEngineFactories or null to use the default.
  */
  public ClassLoader getEngineLoader() {
    if(!cbxURL.getState() || (myLoaderPanel == null)) return null;
    CapCom.debug("Attempting to get MTClassLoader.");
    return CapCom.getMathClassLoader();
  }

  /** Override this method to configure the ScriptEngine used to execute the script. Specifically, if extra Objects are needed by the script this is the place 
      to install them. The ScriptEngine parameter is the one created from the options selected and ScriptPanel expects a valid engine to be returned. If the 
      return value is null it will be detected and execution will stop. If the parameter supplied is null, engine creation failed. The default implementation 
      returns the ScriptEngine unchanged. NOTE: <code>installStandardObjects()</code> is called AFTER this method, so any Object stored under one of the Standard 
      names will be overwritten.
      @param pEngine The ScriptEngine created to execute the script.
      @return The freshly-configured ScriptEngine to execute the script.
  */
  public ScriptEngine configureEngine(ScriptEngine pEngine) { return pEngine; }
  /** This method is called after the ScriptEngine is created but before the script is executed. Override it to do pre-script stuff and, possibly, veto execution 
      of the script. The default implementation prints a Debug message and returns true.
      @return true to continue running the script, false to abort.
  */
  public boolean scriptStarting() {
    CapCom.debug("Script starting");
    return true;
  }

  /** <p>This method is called when a script ends, whether normally or otherwise. If it ends normally, the Object parameter contains whatever the script itself 
      returned. If execution of the script threw an Exception, the Object parameter is null and the Exception is passed. Unless <code>scriptStarting()</code> 
      returns false or there is no script, this method WILL be called.</p>
      <p>The default implementation reports any Exceptions at Error level. If the []processMathObject Checkbox is enabled, the Object returned from the script is 
      sent to <code>CapCom.processMathObject()</code> and its return value replaces that of the 'raw' return Object. For <code>BasicMathPrinter</code> this sends 
      the String content to the clipboard. If the Object present after all of this is null, this is reported at Debug level, otherwise its toString() method is 
      called and printed.</p>
      <p>NOTE: If the 'raw' return Object is null this method checks CapCom for anything stored using the <code>MathConsole.sReturn()</code> method. If 
      that Object is not null it replaces pRetObj for processing.</p>
      @param pRetObj The Object a successful execution returns, which may be null.
      @param pError The Exception thrown if something awful happens. If this is the case, pRetObj is automatically null.
  */
  public void scriptEnded(Object pRetObj, Throwable pError) {
    if(pError != null) {
      CapCom.error("Error running script.", pError);
      return;
    }
    if(pRetObj == null) pRetObj = CapCom.getScriptReturnObject();
    CapCom.setScriptReturnObject(null);

    if(cbxProcessObject.getState()) pRetObj = CapCom.processMathObject(pRetObj);
    if(pRetObj == null) {
      CapCom.debug("Return Object was null");
      return;
    }
    CapCom.println(pRetObj.toString());
    CapCom.debug("Script ended normally.");
  }

}
