package cox.jmatt.java.MathTools.demo;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;
import cox.jmatt.java.MathTools.*;

/** <p>This class wraps a simple GUI around an URLClassLoader. It provides a TextArea to list URLs, machinery to fetch these URLs as either an URL[] array or as a 
    fully-instantiated ClassLoader, and a way to pick single URLs from a file or to load a list of them.</p>
    <p>Originally a part of ScriptPanel, URLLoaderPanel was factored out to provide independent support to other classes that might need an URLClassLoader. In 
    GUI construction this class fetches two Components from the CapCom Shelf: the Forward and Back buttons to flip among ScriptPanel's CardLayout. If this class 
    is added to another application any Component can be added in their place. Simply CapCom.storeObject() these components prior to instantiation, using the keys 
    URL_PANEL_EAST and URL_PANEL_WEST. URLLoaderPanel removes these from the Shelf so using more than one requires storing these items more than once!</p>
    <p>URLLoaderPanel can be constructed with zero arguments but the URL panel will not have Buttons to pick a file URL or load a list of them. These functions 
    require a FileDialog which requires instantiation around a Dialog or a Frame.</p>
    <p>When the [Check URLs] button is clicked, the getURLS() method is called. Doing so parses the URLs individually, so any malformed URLs are reported at Error 
    level and then ignored: the other URLs are still parsed and added to the array. The getURLClassLoader() method also performs this check. Any time getURLS() 
    is called, it reports bad URLs at Error level and gives Info on how many were successfully loaded. Clicking the [Set MTCL] button creates a ClassLoader and 
    sends it to <code>CapCom.setMathClassLoader()</code>. ScriptPanel does this automatically, the button is provided for subclasses or user-created apps that 
    do not rely on ScriptPanel to do such things.</p>
    <p>The URLClassLoader is constructed using the URL[]-array constructor with nothing else. If more complexity is desired, subclass and add it or modify the 
    ClassLoader after it has been created. Or call getURLS() and create your own!</p>
*/
public class URLLoaderPanel implements ActionListener {


  /** Key used to store the EAST component on The Shelf. */
  public static final String URL_PANEL_EAST = "URLLoaderPanel.URL_PANEL_EAST";
  /** Key used to store the WEST component on The Shelf. */
  public static final String URL_PANEL_WEST = "URLLoaderPanel.URL_PANEL_WEST";


  private Dialog myDialog = null;
  private Frame myFrame = null;
  private FileDialog myFileDialog = null;

  private static final String KEY_PICK_URL = "URL (filename) into list";
  private static final String KEY_LOAD_URL_LIST = "URL list from file";

  private Button btnClearURLS = new Button("Clear");
  private Button btnPickURL = new Button("Add file URL");
  private Button btnLoadURL = new Button("Load URL list");
  private Button btnCheckURLS = new Button("Check URLs");
  private Button btnSetMTCL = new Button("Set MTCL");
  private Checkbox cbxLock = new Checkbox("Lock", true);
  private TextArea txaURLCon = null;

  /** Default constructor. If this constructor is used the [Load URL list] and [Add file URL] buttons will not be added to the GUI. */
  public URLLoaderPanel() { super(); }
  /** Constructor around a Frame. This enables the FileDialog and [Load...] and [Add...] buttons. */
  public URLLoaderPanel(Frame pParent) {
    super();
    myFileDialog = new FileDialog(pParent);
  }
  /** Constructor around a Dialog. This enables the FileDialog and [Load...] and [Add...] buttons. */
  public URLLoaderPanel(Dialog pParent) {
    super();
    myFileDialog = new FileDialog(pParent);
  }

  /** <p>The Standard GUI builds the URLLoaderPanel's GUI per standard layout: title, URL List area, and buttons. This method does check CapCom's Shelf for components 
      to add to the WEST and EAST parts of its BorderLayout. During construction, it checks the using the URL_PANEL_ keys and, if it receives a Component, adds it 
      to the appropriate spot. URLLoaderPanel uses these with ScriptPanel to acquire the card Forward and Back buttons. If Components are found on the Shelf, they 
      are removed.</p>
      <p>Note: To instantiate and get the GUI in one fell swoop use the construct '(new URLLoaderPanel(?)).standardGUI();'.</p>
  */
  public final Panel standardGUI() {
    Panel retval = new Panel((new BorderLayout()));
    retval.add(new Label("URL ClassLoader Panel", Label.CENTER), BorderLayout.NORTH);
    txaURLCon = new TextArea();
    retval.add(txaURLCon, BorderLayout.CENTER);
    retval.add(getEastPanel(), BorderLayout.EAST);
    retval.add(getWestPanel(), BorderLayout.WEST);
    if(myDialog != null) myFileDialog = new FileDialog(myDialog);
    if(myFrame != null) myFileDialog = new FileDialog(myFrame);
    Panel pnlTemp = new Panel(new GridLayout(1,0));
    if(myFileDialog != null) {
      btnPickURL.addActionListener(this);
      pnlTemp.add(btnPickURL);
    }
    btnCheckURLS.addActionListener(this); pnlTemp.add(btnCheckURLS);
    if(myFileDialog != null) {
      btnLoadURL.addActionListener(this);
      pnlTemp.add(btnLoadURL);
    }
    btnSetMTCL.addActionListener(this); pnlTemp.add(btnSetMTCL);
    pnlTemp.add(cbxLock);
    btnClearURLS.addActionListener(this); pnlTemp.add(btnClearURLS);
    retval.add(pnlTemp, BorderLayout.SOUTH);
    return retval;
  }

  /** This is the Panel placed at BorderLayout.EAST. Override to customize URLLoaderPanel's GUI. The default implementation looks for the back button, [&lt;&lt|], 
      on CapCom's Shelf. If it is found, it is wrapped in a Panel and added.
  */
  protected Panel getEastPanel() {
    Panel retval = new Panel(new GridLayout(0,1));
    try {
      Component cTemp = (Component) CapCom.removeObject(URL_PANEL_EAST);
      if(cTemp != null) retval.add(cTemp, BorderLayout.EAST);
    } catch(Exception ex) { CapCom.error("Error removing URLLoader.EAST", ex); }
    return retval;
  }

  /** This is the Panel placed at BorderLayout.WEST. The default implementation tries to remove key URL_PANEL_WEST from the CapCom shelf. If a Component is found, 
      it is added. Subclasses can use this method to customize the standard GUI.
  */
  protected Panel getWestPanel() {
    Panel retval = new Panel(new GridLayout(0,1));
    try {
      Component cTemp = (Component) CapCom.removeObject(URL_PANEL_WEST);
      if(cTemp != null) retval.add(cTemp, BorderLayout.EAST);
    } catch(Exception ex) { CapCom.error("Error removing URLLoader.WEST", ex); }
    return retval;
  }

  public void actionPerformed(ActionEvent aev) {
    if(btnClearURLS == aev.getSource()) {
      if(cbxLock.getState()) {
        CapCom.debug("[Clear] button is LOCKED!");
        return;
      }
      txaURLCon.setText(" ");
      cbxLock.setState(true);
    } else if(btnPickURL == aev.getSource()) {
      pickURL();
    } else if(btnLoadURL == aev.getSource()) {
      loadURLList();
    } else if(btnCheckURLS == aev.getSource()) {
      CapCom.info("Checking URL list");
      getURLS();
    } else if(btnSetMTCL == aev.getSource()) {
      CapCom.info("Setting MT ClassLoader");
      setMathClassLoader();
    }
  }
  /** Add an URL to the list. If pURL is null or empty it is ignored. */
  public void addURL(String pURL) {
    if(Question.validString(pURL)) txaURLCon.append(pURL.trim()+"\n");
  }
  /** Pick an URL using a FileDialog. If this class was not constructed around a Frame or Dialog, this method does nothing. */
  public void pickURL() {
    if(myFileDialog == null) return;
    myFileDialog.setTitle("Select file for list");
    myFileDialog.setVisible(true);
    if(!Question.validString(myFileDialog.getFile())) return;
    File fTemp = new File(myFileDialog.getDirectory()+myFileDialog.getFile());
    txaURLCon.append(fTemp.toURI().toString()+"\n");
  }
  /** Load a list of URLs into the TextArea. If this class has no parent, this method does nothing. */
  public void loadURLList() {
    if(myFileDialog == null) return;
    myFileDialog.setTitle("Select file for list");
    myFileDialog.setVisible(true);
    if(!Question.validString(myFileDialog.getFile())) return;
    String pFile = myFileDialog.getDirectory()+myFileDialog.getFile();
    try {
      BufferedReader fileIn = new BufferedReader(new FileReader(pFile));;
      StringWriter strOut = new StringWriter();
      String sLine = null;
      while((sLine = fileIn.readLine()) != null) {
        strOut.write(sLine+"\n");
      }
      strOut.flush();
      strOut.close();
      fileIn.close();
      txaURLCon.setText(strOut.toString());
    } catch(Exception ex) { CapCom.error("Error loading "+pFile, ex); }
  }
  
  /** This method attempts to parse the URLs in the URL ClassLoader area into an URL[] array. Any errors are reported at the Error level. The URLs are parsed 
      individually and any valid ones go into the array. If the URL Panel TextArea is empty this method returns null. If all the URLs are bad, it returns an 
      empty array.
      @return An array of valid URLs.
  */
  public final URL[] getURLS() {
    String sContent = txaURLCon.getText();
    if(!Question.validString(sContent)) return null;
    String[] sURLS = (sContent).split("\n");
    if((sURLS == null) || (sURLS.length < 1)) return null;
    URL[] retval = {};
    ArrayList<URL> ulTemp = new ArrayList<URL>();
    URL theURL;
    for(int i=0; i < sURLS.length; i++) {
      try {
        theURL = new URL(sURLS[i]);
        ulTemp.add(theURL);
      } catch(Exception ex) { CapCom.error("Bad URL: "+sURLS[i]); }
    }
    retval = ulTemp.toArray(retval);
    CapCom.info("Loaded "+retval.length+" URLs.");
    return retval;
  }
  /** Create an URLClassLoader around the URLs listed. If anyting goes wrong it is reported at Error level and null is returned. */
  public ClassLoader getURLClassLoader() {
    URL[] theURLS = getURLS();
    if((theURLS == null) || (theURLS.length < 1)) return null;
    URLClassLoader theLoader = null;
    try { theLoader = new URLClassLoader(theURLS); } catch(Exception ex) { CapCom.error("Error creating URLClassLoader", ex); }
    return theLoader;
  }

  /** Call this method to create an URLClassLoader and send it to CapCom. ScriptPanel does this automatically. If no ClassLoader is created, the ClassLoader set 
      is null.
  */
  public void setMathClassLoader() { CapCom.setMathClassLoader(getURLClassLoader()); }

}
