package cox.jmatt.java.MathTools.demo;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.util.*;

/** <p>This utility provides basic XML/XSLT processing. It uses <code>MTXProcessor</code> under the hood and allows more flexible options than the simple parsing 
    and transforms provided by selections 'K' and 'M' in <code>MathXMLPizza</code>. The GUI allows dialog-based file selection as well as loadable internal areas 
    for the XML data and XSLT stylesheet. The XSLT stylesheet can also be checked for validity.</p>
    <h3>Checking XML Structure</h3>
    <p>Both the XML data and the XSLT stylesheet can be checked for structural validity. No complex analysis is done; this is the 'checkXML()' method from 
    <code>MTXPro</code> applied to either the file specified ('[]XML File' Checkbox checked) or the internal XML. Checking the XSL Stylesheet works the same 
    way but with the '[]XSLT Sheet' Checkbox determining whether to check the file specified or the internal stylesheet.</p>
    <p>The two options that can be specified from the main panel are SAX validation and namespace awareness. Checking the appropriate box enables that feature.</p>
    <h3>XSL Transforms</h3>
    <p>XSLT processing has a lot of flexibility and options. The XML data to transform can come from a file or an internal area, as above, and the XSLT Stylesheet 
    can as well. Checking the box on the main panel reads the stylesheet from a file, unchecking it reads from the XSLT Stylesheet Panel. The Stylesheet 
    can also be checked for structural validity. The internal Stylesheet panel can be loaded and has a lockable 'Clear' button.</p>
    <p>Immediately after the internal Stylesheet panel is the XSLT Parameters Panel. Parameters are entered as '<i>key</i> = <i>value</i>' pairs with one pair 
    per line. An octothorpe ('#') indicates a comment. If present and enabled these parameters are provided to the Transformer during the process. As with the 
    Data and Stylesheet panels this one is loadable with a locked 'Clear' button.</p
    <p>Besides styleshet parameters this Panel can also set output properties. They are included with the standard parameters but the first character must be 
    an asterisk ('*'). The asterisk marks output properties and it is stripped before the property is added. Other than this output properties are identical to 
    stylesheet parameters.</p>
    <p>In addition to the basic parameter providing capability the Parameter panel can extract parameters from the actual stylesheet. Each parameter must have 
    a specific prefix ('#PARAM' is the default). Any (trimmed) line that begins with this prefix is extracted and added to the Parameter Panel starting with the 
    first character after the prefix. Output properties can also be extracted; merely include the asterisk as the first character after the parameter prefix.</p>
    <h3>XPath</h3>
    <p>In addition to XSLT, <code>XMLUtils</code> provides XPath capability. The 'XPath Expressions' panel provides a place to put one or more XPath expressions. 
    Any expressions here are applied to the XML Data (file or internal) and the results printed and placed in the clipboard. If more than one XPath expression 
    is entered (or loaded) they are applied, in order, to the data. Checking the 'Transform result' Checkbox sends the final result through the XSL Transform 
    defined on the main panel (file or internal). If the 'Transform' checkbox is not checked the result is still sent through the XSLT identity transform in order 
    to make it printable.</p>
    <h3>GUI Elements</h3>
    <p>Output from all actions (except checking documents) is printed to the screen and can be sent to the clipboard. The 'processMathObject()' Checkbox on the 
    main panel enables this. If it and the '[]Clip' checkbox (<code>Basic MathPrinter</code>) are checked the data goes to the clipboard. The output from an XSL 
    Transform (only!) can be sent to a file. If the file exists it will not be overwritten unless the '[]Overwrite...' box is checked.</p>
    <p>The '[]Enable stylesheet properties' box must be checked to include parameters in a transform and at least one parameter must be defined.</p>
    <p>The XML Data, XSLT Stylesheet, XSL Transform Parameters and XPath Expressions Panels are all basically identical: the <code>TextArea</code> is loadable from 
    a file and its 'Clear' button is locked. If <code>XMLUtils</code> is NOT instantiated around a <code>Dialog</code> or <code>Frame</code> no 'Load' buttons will 
    be present. As a part of <code>MTDemo</code> they will always be present. Clicking a 'Load' or 'Choose' button opens a <code>FileDialog</code> to allow file 
    selection.</p>
*/

public class XMLUtils extends Panel implements ActionListener {

  private static final String FORWARD = "|>>";
  private static final String BACKWARD = "<<|";
  private static final String KEY_LOAD = "Load ";
  private static final String KEY_CLEAR = "Clear ";
  private static final String KEY_CHOOSE = "Choose ";
  private static final String KEY_PARSE = "Parse ";
  private static final String KEY_TRANSFORM = "Transform ";

  private static final String KEY_MAIN = "XML Utility Panel";
  private static final String KEY_DATA = "XML Data";
  private static final String KEY_XSLT = "XSLT Stylesheet";
  private static final String KEY_OUTPUT = "Output";
  private static final String KEY_PIZZA = "XSL Transform Parameters";
  private static final String KEY_XPATH = "XPath Expressions";

  private FileDialog myFileDialog = null;
  private CardLayout cdlMain = null;
  private Button btnTemp = null;
  private Panel pnlTemp = null;
  private Checkbox cbxLockData = new Checkbox("Lock clear", true), cbxLockXSLT = new Checkbox("Lock clear", true), cbxLockPizza = new Checkbox("Lock clear", true);
  private Checkbox cbxLockXPath = new Checkbox("Lock clear", true);
  private TextArea txaData = new TextArea(), txaTransform = new TextArea(), txaPizza = new TextArea(), txaXPath = new TextArea();
  private Checkbox cbxData = new Checkbox("XML File"), cbxValidate = new Checkbox("Validate"), cbxNamespace = new Checkbox("Namespace", true);
  private Checkbox cbxTransform = new Checkbox("XSLT Sheet"), cbxOut = new Checkbox("Output File"), cbxOkToStomp = new Checkbox("Overwrite existing file");
  private Checkbox cbxClip = new Checkbox("Call processMathObject(result)", true), cbxEnablePizza = new Checkbox("Enable stylesheet properties", true);
  private Checkbox cbxXPTransform = new Checkbox("Transform result");
  private TextField txfData = new TextField(40), txfTransform = new TextField(40), txfOut = new TextField(40), txfParam = new TextField();

  private MTXProcessor myProcessor = new MTXProcessor();

  /** Construct an instance without a 'Load' button. */
  public XMLUtils() {
    setUpGUI();
  }

  /** Construct an instance around a <code>Frame</code>. The 'Load' button will be available. */
  public XMLUtils(Frame myParent) {
    super();
    if(myParent != null) myFileDialog = new FileDialog(myParent);
    setUpGUI();
  }

  /** Construct an instance around a <code>Dialog</code>. The 'Load' button will be available. */
  public XMLUtils(Dialog myParent) {
    super();
    if(myParent != null) myFileDialog = new FileDialog(myParent);
    setUpGUI();
  }

  private void setUpGUI() {
    cdlMain = new CardLayout();
    setLayout(cdlMain);
    add(pnlMain(), KEY_MAIN);
    add(pnlData(), KEY_DATA);
    add(pnlXSLT(), KEY_XSLT);
    add(pnlPizza(), KEY_PIZZA);
    clearPizza(true);
    add(pnlXPath(), KEY_XPATH);
  }

  private Panel pnlMain() {
    Panel retval = new Panel();
    retval.setLayout(new BorderLayout());
    retval.add(new Label(KEY_MAIN, Label.CENTER), BorderLayout.NORTH);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.EAST);
    btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.WEST);
    pnlTemp = new Panel(new GridLayout(1,0));
    btnTemp = new Button("Parse"); btnTemp.setActionCommand(KEY_PARSE+KEY_MAIN); btnTemp.addActionListener(this); pnlTemp.add(btnTemp);
    pnlTemp.add(new Label(""));
    btnTemp = new Button("Transform"); btnTemp.setActionCommand(KEY_TRANSFORM+KEY_MAIN); btnTemp.addActionListener(this); pnlTemp.add(btnTemp);
    btnTemp = new Button("Parse XSL"); btnTemp.setActionCommand(KEY_PARSE+KEY_TRANSFORM+KEY_MAIN); btnTemp.addActionListener(this); pnlTemp.add(btnTemp);
    retval.add(pnlTemp, BorderLayout.SOUTH);
    pnlTemp = new Panel(new GridBagLayout());
    GridBagConstraints gbc = new GridBagConstraints(); gbc.insets = new Insets(2,1,2,1);
     gbc.gridx=1; gbc.gridy=1; gbc.anchor = GridBagConstraints.WEST;
    pnlTemp.add(cbxData, gbc); gbc.gridx++; gbc.gridwidth=2;
    pnlTemp.add(txfData, gbc); gbc.gridx += 2; gbc.gridwidth=1;
    btnTemp = new Button("Choose"); btnTemp.setActionCommand(KEY_CHOOSE+KEY_DATA); btnTemp.addActionListener(this); pnlTemp.add(btnTemp, gbc);
     gbc.gridx=2; gbc.gridy++;
    pnlTemp.add(cbxValidate, gbc); gbc.gridx++; gbc.anchor = GridBagConstraints.CENTER;
    pnlTemp.add(cbxNamespace, gbc); gbc.gridx=1; gbc.gridy++; gbc.anchor = GridBagConstraints.WEST;
    pnlTemp.add(cbxTransform, gbc); gbc.gridx++; gbc.gridwidth=2;
    pnlTemp.add(txfTransform, gbc); gbc.gridx += 2; gbc.gridwidth=1;
    btnTemp = new Button("Choose"); btnTemp.setActionCommand(KEY_CHOOSE+KEY_XSLT); btnTemp.addActionListener(this); pnlTemp.add(btnTemp, gbc);
     gbc.gridx=2; gbc.gridwidth=2; gbc.gridy++;
    pnlTemp.add(cbxEnablePizza, gbc);
     gbc.gridx=1; gbc.gridy++; gbc.gridwidth=1;
    pnlTemp.add(cbxOut, gbc); gbc.gridx++; gbc.gridwidth=2;
    pnlTemp.add(txfOut, gbc); gbc.gridx += 2; gbc.gridwidth=1;
    btnTemp = new Button("Choose"); btnTemp.setActionCommand(KEY_CHOOSE+KEY_OUTPUT); btnTemp.addActionListener(this); pnlTemp.add(btnTemp, gbc);
     gbc.gridx=2; gbc.gridy++; gbc.gridwidth=1;
    pnlTemp.add(cbxOkToStomp, gbc); gbc.gridy++;
    pnlTemp.add(cbxClip, gbc); gbc.gridx=1; gbc.gridy++; gbc.anchor = GridBagConstraints.WEST;
    
    retval.add(pnlTemp, BorderLayout.CENTER);
    return retval;
  }

  private Panel pnlData() {
    Panel retval = new Panel();
    retval.setLayout(new BorderLayout());
    retval.add(new Label(KEY_DATA+" Panel", Label.CENTER), BorderLayout.NORTH);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.EAST);
    btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.WEST);

    pnlTemp = new Panel(new GridLayout(1,0));
    if(myFileDialog == null) {
      pnlTemp.add(new Label(""));
    } else {
      btnTemp = new Button("Load"); btnTemp.setActionCommand(KEY_LOAD+KEY_DATA); btnTemp.addActionListener(this);
      pnlTemp.add(btnTemp);
    }
    pnlTemp.add(new Label("")); pnlTemp.add(new Label(""));
    btnTemp = new Button("Clear"); btnTemp.setActionCommand(KEY_CLEAR+KEY_DATA); btnTemp.addActionListener(this);
    pnlTemp.add(btnTemp);
    pnlTemp.add(cbxLockData);
    retval.add(pnlTemp, BorderLayout.SOUTH);
    retval.add(txaData, BorderLayout.CENTER);
    return retval;
  }

  /** Clear the XML Data area, conditioned on the value pOkToClear. */
  public void clearData(boolean pOkToClear) {
    if(pOkToClear) {
      txaData.setText("");
    } else {
      CapCom.debug("Clear Data is LOCKED");
    }
  }

  private Panel pnlXSLT() {
    Panel retval = new Panel();
    retval.setLayout(new BorderLayout());
    retval.add(new Label(KEY_XSLT+" Panel", Label.CENTER), BorderLayout.NORTH);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.EAST);
    btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.WEST);

    pnlTemp = new Panel(new GridLayout(1,0));
    if(myFileDialog == null) {
      pnlTemp.add(new Label(""));
    } else {
      btnTemp = new Button("Load"); btnTemp.setActionCommand(KEY_LOAD+KEY_XSLT); btnTemp.addActionListener(this);
      pnlTemp.add(btnTemp);
    }
    pnlTemp.add(new Label("")); pnlTemp.add(new Label(""));
    btnTemp = new Button("Clear"); btnTemp.setActionCommand(KEY_CLEAR+KEY_XSLT); btnTemp.addActionListener(this);
    pnlTemp.add(btnTemp);
    pnlTemp.add(cbxLockXSLT);
    retval.add(pnlTemp, BorderLayout.SOUTH);
    retval.add(txaTransform, BorderLayout.CENTER);
    return retval;
  }

  /** Clear the XSLT stylesheet area, conditioned on the value pOkToClear. */
  public void clearTransform(boolean pOkToClear) {
    if(pOkToClear) {
      txaTransform.setText("");
    } else {
      CapCom.debug("Clear Data is LOCKED");
    }
  }

  private Panel pnlPizza() {
    Panel retval = new Panel();
    retval.setLayout(new BorderLayout());
    retval.add(new Label(KEY_PIZZA+" Panel", Label.CENTER), BorderLayout.NORTH);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.EAST);
    btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.WEST);

    pnlTemp = new Panel(new GridLayout(1,0));
    if(myFileDialog == null) {
      pnlTemp.add(new Label(""));
    } else {
      btnTemp = new Button("Load"); btnTemp.setActionCommand(KEY_LOAD+KEY_PIZZA); btnTemp.addActionListener(this);
      pnlTemp.add(btnTemp);
    }
    btnTemp = new Button("Extract"); btnTemp.setActionCommand(KEY_CHOOSE+KEY_PIZZA); btnTemp.addActionListener(this);
    pnlTemp.add(btnTemp);
    pnlTemp.add(txfParam);
    btnTemp = new Button("Clear"); btnTemp.setActionCommand(KEY_CLEAR+KEY_PIZZA); btnTemp.addActionListener(this);
    pnlTemp.add(btnTemp);
    pnlTemp.add(cbxLockPizza);
    retval.add(pnlTemp, BorderLayout.SOUTH);
    retval.add(txaPizza, BorderLayout.CENTER);
    return retval;
  }

  /** Clear the XSLT Parameter area, conditioned on the value pOkToClear. */
  public void clearPizza(boolean pOkToClear) {
    if(pOkToClear) {
      txaPizza.setText("");
      txfParam.setText("#PARAM");
    } else {
      CapCom.debug("Clear Data is LOCKED");
    }
  }

  private void extractPizza(String pPrefix) {
    if(!Question.validString(pPrefix)) return;
    if(cbxTransform.getState()) {
      // Parse file for params.
      try {
        BufferedReader brIn = new BufferedReader(new FileReader(txfTransform.getText()));
        String sLine = null;
        while ((sLine = brIn.readLine()) != null) {
          sLine = sLine.trim();
          if(sLine.trim().startsWith(pPrefix)) txaPizza.append(sLine.substring(pPrefix.length()).trim()+"\n");
        }
        brIn.close();
      } catch(Exception ex) { CapCom.debug("Error parsing params from file.", ex); }
    } else {
      String[] sLines = txaTransform.getText().split("\n");
      for(String sLine : sLines) {
        sLine = sLine.trim();
        if(sLine.startsWith(pPrefix)) txaPizza.append(sLine.substring(pPrefix.length()).trim()+"\n");
      }
    }
  }

  private Panel pnlXPath() {
    Panel retval = new Panel();
    retval.setLayout(new BorderLayout());
    retval.add(new Label(KEY_XPATH+" Panel", Label.CENTER), BorderLayout.NORTH);
    btnTemp = new Button(FORWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.EAST);
    btnTemp = new Button(BACKWARD); btnTemp.addActionListener(this); retval.add(btnTemp, BorderLayout.WEST);

    pnlTemp = new Panel(new GridLayout(1,0));
    if(myFileDialog == null) {
      pnlTemp.add(new Label(""));
    } else {
      btnTemp = new Button("Load"); btnTemp.setActionCommand(KEY_LOAD+KEY_XPATH); btnTemp.addActionListener(this);
      pnlTemp.add(btnTemp);
    }
    btnTemp = new Button("Apply"); btnTemp.setActionCommand(KEY_PARSE+KEY_XPATH); btnTemp.addActionListener(this);
    pnlTemp.add(btnTemp);
    pnlTemp.add(cbxXPTransform);
    btnTemp = new Button("Clear"); btnTemp.setActionCommand(KEY_CLEAR+KEY_XPATH); btnTemp.addActionListener(this);
    pnlTemp.add(btnTemp);
    pnlTemp.add(cbxLockXPath);
    retval.add(pnlTemp, BorderLayout.SOUTH);
    retval.add(txaXPath, BorderLayout.CENTER);
    return retval;
  }

  /** Clear the XPath Expressions area, conditioned on the value pOkToClear. */
  public void clearXPath(boolean pOkToClear) {
    if(pOkToClear) {
      txaXPath.setText("");
    } else {
      CapCom.debug("Clear Data is LOCKED");
    }
  }

  public void actionPerformed(ActionEvent aev) {
    String sCmd = aev.getActionCommand();
    CapCom.debug("Got command: "+sCmd);
    if(!Question.validString(sCmd)) return;
    if(FORWARD.equalsIgnoreCase(sCmd)) {
      cdlMain.next(this);
    } else if(BACKWARD.equalsIgnoreCase(sCmd)) {
      cdlMain.previous(this);
    } else if((KEY_CLEAR+KEY_DATA).equalsIgnoreCase(sCmd)) {
      clearData(!cbxLockData.getState());
      cbxLockData.setState(true);
    } else if((KEY_CLEAR+KEY_XSLT).equalsIgnoreCase(sCmd)) {
      clearTransform(!cbxLockXSLT.getState());
      cbxLockXSLT.setState(true);
    } else if((KEY_CLEAR+KEY_PIZZA).equalsIgnoreCase(sCmd)) {
      clearPizza(!cbxLockPizza.getState());
      cbxLockPizza.setState(true);
    } else if((KEY_CLEAR+KEY_XPATH).equalsIgnoreCase(sCmd)) {
      clearXPath(!cbxLockXPath.getState());
      cbxLockXPath.setState(true);
    } else if(sCmd.startsWith(KEY_LOAD)) {
      loadFile(getFileName(sCmd), sCmd);
    } else if((KEY_CHOOSE+KEY_DATA).equalsIgnoreCase(sCmd)) {
      txfData.setText(getFileName(sCmd));
      cbxData.setState(Question.validString(txfData.getText()));
    } else if((KEY_CHOOSE+KEY_XSLT).equalsIgnoreCase(sCmd)) {
      txfTransform.setText(getFileName(sCmd));
      cbxTransform.setState(Question.validString(txfTransform.getText()));
    } else if((KEY_CHOOSE+KEY_OUTPUT).equalsIgnoreCase(sCmd)) {
      txfOut.setText(getFileName(sCmd));
      cbxOut.setState(Question.validString(txfOut.getText()));
    } else if((KEY_PARSE+KEY_MAIN).equalsIgnoreCase(sCmd)) {
      parseXML();
    } else if((KEY_PARSE+KEY_TRANSFORM+KEY_MAIN).equalsIgnoreCase(sCmd)) {
      parseXSL();
    } else if((KEY_TRANSFORM+KEY_MAIN).equalsIgnoreCase(sCmd)) {
      transformData();
    } else if((KEY_CHOOSE+KEY_PIZZA).equalsIgnoreCase(sCmd)) {
      extractPizza(txfParam.getText());
    } else if((KEY_PARSE+KEY_XPATH).equalsIgnoreCase(sCmd)) {
      applyXPath();
    //} else if() {
    //} else if() {
    } else {
      CapCom.debug("Don't know command: "+sCmd);
    }
  }

  /** Get the name of a file to load. The String supplied is the title for the <code>FileDialog</code> if it exists. If the <code>FileDialog</code> does not exist 
      or the 'Cancel' button is pressed the return value is a blank String.
  */
  public String getFileName(String pTitle) {
    if(myFileDialog == null) return "";
    myFileDialog.setTitle(pTitle);
    myFileDialog.setVisible(true);
    String retval = myFileDialog.getFile();
    if(retval != null) {
      retval = myFileDialog.getDirectory()+File.separator+retval;
    } else {
      retval = "";
    }
    return retval;
  }

  private void loadFile(String pName, String pKey) {
    if(!(Question.validString(pName) && Question.validString(pKey))) return;
    CapCom.debug("Load file: "+pName);
    try {
      StringBuffer sbFile = new StringBuffer();
      String sLine = null;
      BufferedReader frIn = new BufferedReader(new FileReader(pName));
      while((sLine = frIn.readLine()) != null) sbFile.append(sLine).append("\n");
      frIn.close();
      if(pKey.indexOf(KEY_DATA) > -1) {
        txaData.setText(sbFile.toString());
      } else if(pKey.indexOf(KEY_XSLT) > -1) {
        txaTransform.setText(sbFile.toString());
      } else if(pKey.indexOf(KEY_PIZZA) > -1) {
        txaPizza.setText(sbFile.toString());
      } else if(pKey.indexOf(KEY_XPATH) > -1) {
        txaXPath.setText(sbFile.toString());
      //} else if(pKey.indexOf(KEY_) > -1) {
      //} else if(pKey.indexOf(KEY_) > -1) {
      //} else if(pKey.indexOf(KEY_) > -1) {
      } else {
        CapCom.debug("No place to send file load");
      }
    } catch(Exception ex) { CapCom.error("Unable to load file: "+pName, ex); }
  }

  private void parseXML() {
    myProcessor.setValidating(cbxValidate.getState());
    myProcessor.setNamespaceAware(cbxNamespace.getState());
    if(!myProcessor.isAvailable()) {
      CapCom.println("Unable to create SAX Parser");
      return;
    }
    CapCom.println("Checking "+(cbxData.getState()?"file":"internal data")+" for errors");
    myProcessor.checkXML((cbxData.getState()?txfData.getText():txaData.getText()), !cbxData.getState());
    CapCom.println("Done.");
  }

    private void parseXSL() {
    myProcessor.setValidating(cbxValidate.getState());
    myProcessor.setNamespaceAware(cbxNamespace.getState());
    if(!myProcessor.isAvailable()) {
      CapCom.println("Unable to create SAX Parser");
      return;
    }
    CapCom.println("Checking "+(cbxTransform.getState()?"XSLT file":"internal XSLT stylesheet")+" for errors");
    myProcessor.checkXML((cbxTransform.getState()?txfTransform.getText():txaTransform.getText()), !cbxTransform.getState());
    CapCom.println("Done.");
  }

  private boolean canTransform() {
    if(!myProcessor.isTransformAvailable()) {
      CapCom.println("XSL Transform not available.");
      return false;
    }
    String sPizza = txaPizza.getText();
    myProcessor.clearParameters();
    myProcessor.clearOutputProperties();
    if(cbxEnablePizza.getState() && Question.validString(sPizza)) {
      String[] sProps = sPizza.split("\n"), sPair = {};
      for(String sLine : sProps) {
        sLine = sLine.trim();
        if(!Question.validString(sLine)) continue;
        sPair = sLine.split("=", 2);
        switch(sLine.charAt(0)) {
        case '#': break;
        case '*':
          if(sLine.length() > 1) myProcessor.setOutputProperty(sPair[0].substring(1).trim(), sPair[1].trim());
          break;
        default:
          if(sPair.length == 2) myProcessor.setParameter(sPair[0].trim(), sPair[1].trim());
        }
      }
    }
    return true;
  }
  private void transformData() {
    if(!canTransform()) return;
    CapCom.println("Transforming "+(cbxData.getState()?"file":"internal XML")+" by "+(cbxTransform.getState()?"file":"internal stylesheet")+".");
    String rezult = myProcessor.transformData((cbxData.getState()?txfData.getText():txaData.getText()), (cbxTransform.getState()?txfTransform.getText():txaTransform.getText()), !cbxData.getState(), !cbxTransform.getState());
    CapCom.println(rezult);
    if(cbxClip.getState()) CapCom.processMathObject(rezult);
    if(cbxOut.getState()) {
      try {
        File filOut = new File(txfOut.getText());
        if(filOut.exists() && !cbxOkToStomp.getState()) {
          CapCom.println("File exists and overwrite is off.");
        } else {
          FileWriter fwOut = new FileWriter(filOut);
          fwOut.write(rezult);
          fwOut.flush();
          fwOut.close();
        }
      } catch(Exception ex) {
        CapCom.error("File output error.", ex);
      }
    }
  }

  private void applyXPath() {
    if(!myProcessor.isXPathAvailable()) {
      CapCom.println("XPath not available.");
      return;
    }
    if(!canTransform()) return;
    CapCom.println("Applying XPath(s) to "+(cbxData.getState()?"XML file.":"internal XML data."));
    org.w3c.dom.Node docNode = myProcessor.parseDOM((cbxData.getState()?txfData.getText():txaData.getText()), !cbxData.getState());
    String retval = txaXPath.getText();
    if(Question.validString(retval)) {
      String[] sLines = retval.split("\n");
      for(String sLine : sLines) if(!sLine.startsWith("#")) docNode = myProcessor.applyXPath(sLine, docNode);
    }
    if(cbxXPTransform.getState()) {
      retval = myProcessor.transformDOM(docNode, (cbxTransform.getState()?txfTransform.getText():txaTransform.getText()), !cbxTransform.getState());
    } else {
      retval = myProcessor.transformDOM(docNode, null, true);
    }
    CapCom.println(retval);
    if(cbxClip.getState()) CapCom.processMathObject(retval);
  }

}

