package cox.jmatt.java.MathTools.demo.gpgdemo;

import java.awt.*;
import java.util.Properties;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.markup.svg.*;
import cox.jmatt.java.MathTools.demo.GPGenerator;

/** <p>BPS stands for 'blood pressure (and) sugar,' and this panel generates a chart to track that. It is a square-grid chart with the vertical scale centered 
    on 100 and magnified between 80 and 120. Outside 80 and 120 the scale shrinks, then again above 140 an below 60. This allows greater detail if blood pressure 
    and blood sugar are close to where they should be while still allowing tracking of high or low numbers.</p>
    <p>To use the chart, enumerate the horizontal lines with appropriate dates and times. To record the numbers, place a dot corresponding to blood sugar. Draw 
    shorter lines (tick marks) at the numbers corresponding to systolic (high) and diastolic (low) blood pressure and draw a heavier line between them.</p>
    <p>Everything about the graph is configurable with one exception: The numbering and <code>TextPen</code> used are hard-coded. The scales and (vertical) scale 
    factor can be changed to expand or shrink the chart but the horizontal lines occur at steps of five with the tens being listed to the side. The first vertical 
    line can be different from the others. All horizontal lines are the same.</p>
*/
public class BPSPanel extends GraphPanel {

  private TextField txfLinePen = null, txfVerticalPen = null, txfTextPen = null, txfCenterline = null, txfBottom = null, txfHSpace = null, txfVScale = null;
  private TextField txfVerticalFirst = null, txfVSpace = null;

  /** Simple constructor. */
  public BPSPanel() {
    super();
    buildGUI();
    GPickle.gpgBPSPanel = this;
  }

  private void buildGUI() {
    txfLinePen = new TextField("0");
    txfVerticalPen = new TextField("0");
    txfVerticalFirst = new TextField("0");
    txfCenterline = new TextField("550");
    txfBottom = new TextField("950");
    txfHSpace = new TextField("25");
    txfVScale = new TextField("0.7");
    txfVSpace = new TextField("40");
    setLayout(new GridLayout(0,2));
    add(new Label("Horizontal spacing")); add(txfHSpace);
    add(new Label("Horizontal line pen")); add(txfLinePen);
    add(new Label("Horizontal centerline")); add(txfCenterline);
    add(new Label("Bottom limit")); add(txfBottom);
    add(new Label("Vertical line pen")); add(txfVerticalPen);
    add(new Label("First vertical pen")); add(txfVerticalFirst);
    add(new Label("Initial vertical space")); add(txfVSpace);
    add(new Label("Vertical scale factor")); add(txfVScale);
  }

  /** Reset to default values. */
  public void reset() {
    super.reset();
    txfLinePen.setText("0"); txfVerticalPen.setText("0"); txfVerticalFirst.setText("0"); txfCenterline.setText("550"); txfBottom.setText("950");
    txfHSpace.setText("25"); txfVScale.setText("0.7"); txfVSpace.setText("40");
  }

  /** Draw the chart. */
  public void drawGraph(int mTop, int mBottom, int mLeft, int mRight, SVGTag pPaper) {
    if(pPaper == null) return;
    MathDrawSVG MDraw = cox.jmatt.java.MathTools.demo.GPGenerator.getMDraw();
    int hPen = getPen(txfLinePen), vPen = getPen(txfVerticalPen), vFirst = getPen(txfVerticalFirst), hSpace = asInt(txfHSpace.getText(), 25, 5, null);
    int centerLine = asInt(txfCenterline.getText(), 550, 10, mBottom), iNdx = 0;
    double vScale = asDouble(txfVScale.getText(), 0.7, 0.1, 1.0), deltaY = asDouble(txfVSpace.getText(), 40.0, 10.0, null), yPos = centerLine - deltaY;
    boolean trucking = true;
    mLeft += 10; // Adjust for text!
    mBottom = asInt(txfBottom.getText(), 550, 100, mBottom);
    StylePen tPen = MDraw.getTextPen().setFontSize(15.0).setTextAnchor('e').setStroke("none").setFill("black").setStrokeWidth(1.0);
    MDraw.setTextStyle(tPen);
    MDraw.setLinePen(hPen);
    //pPaper.drawLine((double) mLeft, (double) mTop, (double) mRight, (double) mTop);
    pPaper.drawLine((double) mLeft, (double) centerLine, (double) mRight, (double) centerLine);
    pPaper.drawText((double) mLeft, (double) (centerLine + 5), "100"); // Adjust y by +5 to center the line vertically.
    for(iNdx = 5; trucking && (iNdx < 100); iNdx += 5) {
      pPaper.drawLine((double) mLeft,  yPos, (double) mRight,  yPos);
      if((iNdx % 10) == 0) pPaper.drawText((double) mLeft,  yPos + 5, ""+ (100 + iNdx));
      if((iNdx == 20) || (iNdx == 40)) deltaY *= vScale;
      yPos -= deltaY;
      trucking = (yPos > mTop);
    }
    deltaY = deltaY = asDouble(txfVSpace.getText(), 40.0, 10.0, null);
    yPos = centerLine + deltaY;
    trucking = true;
    for(iNdx = 5; trucking && (iNdx < 75); iNdx += 5) {
      pPaper.drawLine((double) mLeft,  yPos, (double) mRight,  yPos);
      if((iNdx % 10) == 0) pPaper.drawText((double) mLeft,  yPos + 5, ""+ (100 - iNdx));
      if((iNdx == 20) || (iNdx == 40)) deltaY *= vScale;
      yPos += deltaY;
      trucking = (yPos < mBottom);
    }
    mLeft += 10;
    MDraw.setLinePen(vFirst);
    pPaper.drawLine((double) mLeft, (double) mTop, (double) mLeft, (double) mBottom);
    MDraw.setLinePen(vPen);
    for(deltaY = mLeft + hSpace; deltaY < mRight; deltaY += hSpace) {
      pPaper.drawLine(deltaY, (double) mTop, deltaY, (double) mBottom);
    }
  }

  /** Set the pens used to draw the lines.
      @param horizPen The pen used for horizontal lines.
      @param vertPen The pen used for vertical lines after the first.
      @param vertFirstPen The pen used for the first vertical line.
  */
  public void setPens(int horizPen, int vertPen, int vertFirstPen) {
    txfLinePen.setText("" + horizPen);
    txfVerticalPen.setText("" + vertPen);
    txfVerticalFirst.setText("" + vertFirstPen);
  }

  /** Set the spacings used: horizontal, initial vertical space and vertical scale factor.
      @param pHorizontal The horizontal spacing between vertical lines.
      @param pVertInitial The initial vertical space.
      @param pVertScale The vertical scale factor.
  */
  public void setSpacing(int pHorizontal, int pVertInitial, double pVertScale) {
    txfHSpace.setText("" + pHorizontal);
    txfVSpace.setText("" + pVertInitial);
    txfVScale.setText("" + pVertScale);
  }

  /** Set the horizontal centerline and chart-wise lower margin.
      @param pCenterline The centerline value.
      @param pBottom The chart's lower limit.
  */
  public void setLimits(int pCenterline, int pBottom) {
    txfCenterline.setText("" + pCenterline);
    txfBottom.setText("" + pBottom);
  }

  
  public String toPizza(String pPrefix) {
    StringBuffer retval = new StringBuffer("# BPSPanel configuration data.");
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
      retval.append(" Prefix: ").append(pPrefix);
    } else {
      pPrefix = "";
    }
    retval.append("\n");
    retval.append(getKey(txfHSpace, pPrefix, "bps.hSpace"));
    retval.append(getKey(txfVSpace, pPrefix, "bps.vSpace"));
    retval.append(getKey(txfVScale, pPrefix, "bps.verticalScale"));
    retval.append(getKey(txfBottom, pPrefix, "bps.bottomLimit"));
    retval.append(getKey(txfCenterline, pPrefix, "bps.centerline"));
    retval.append(getKey(txfLinePen, pPrefix, "bps.pen.horizontal"));
    retval.append(getKey(txfVerticalPen, pPrefix, "bps.pen.vertical"));
    retval.append(getKey(txfVerticalFirst, pPrefix, "bps.pen.verticalFirst"));
    return retval.toString();
  }

  public void eatPizza(Properties pPizza, String pPrefix) {
    if((pPizza == null) || pPizza.isEmpty()) return;
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
    } else {
      pPrefix = "";
    }
    setKey(pPizza, txfHSpace, pPrefix, "bps.hSpace", "25");
    
    setKey(pPizza, txfVSpace, pPrefix, "bps.vSpace", "40");
    setKey(pPizza, txfVScale, pPrefix, "bps.verticalScale", "0.7");
    setKey(pPizza, txfBottom, pPrefix, "bps.bottomLimit", "950");
    setKey(pPizza, txfCenterline, pPrefix, "bps.centerline", "550");
    setKey(pPizza, txfLinePen, pPrefix, "bps.pen.horizontal", "0");
    setKey(pPizza, txfVerticalPen, pPrefix, "bps.pen.vertical", "0");
    setKey(pPizza, txfVerticalFirst, pPrefix, "bps.pen.verticalFirst", "0");
    
  }

}

