package cox.jmatt.java.MathTools.demo.gpgdemo;

import java.awt.*;
import java.util.Properties;
import java.io.*;

import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.markup.svg.*;

/** <p>This panel generates simple bar charts and line charts. It also provides very fine control over exactly how these graphs appear. There are two distinct 
    modes of operation, three types of graphs and lots of options for each combination.</p>
    <h3>Consistent Options</h3>
    <p>These are options that are consistent across all modes of operation and graph types. The x- and y-axis can be visible or not independently of each other: 
    the '[]Vertical axis pen' and '[]Horizontal axis pen' checkboxes toggle axis visibility. The pen used for each axis is set in the TextArea beside the checkboxes 
    and these are also consistent. If the y-axis is enabled and the '[]Vertical scale' checkbox is enabled there will be tick marks along the y-axis. The interval 
    between them is mode-dependent, though, so the use of the TextArea beside it is covered later. The '[]Show midpoints' checkbox toggles midpoint/data point 
    visibility for bar graphs and line graphs, respectively.</p>
    <p>The 'Graph Pen' TextArea is consistent: this is the pen used to draw the graph. The 'Bar Width:Gap' TextArea deserves special attention. This can be one 
    number or two separated by a colon. A single number is the bar width, two numbers are the bar width and the gap between bars. These numbers are <b>always</b> 
    in <i>raw user units</i> (100 per inch by default). The default value is '50:50' which gives a 50-unit (half-inch) bar followed by a 50-unit (half-inch) gap. 
    This is important for all graph types as it determines horizontal spacing.</p>
    <p>The final consistent item is the graph data area. This is the TextArea to the right of the display, just under the 'Graph pen' selector. It is a TextArea 
    containing one or more data lines. Each line can take one of two forms:
    <ol>
      <li>'<b><u>NN</u></b> <i>Label text.</i>' - This indicates a bar or data point. The number <b>NN</b> is the bar height/y-coordinate and the label text is 
        what appears below the x-axis (even if it is invisible!). There must be a single space between the number and the label; spaces within the label are 
        ignored.</li>
      <li>'<b><u>NN</u></b>' - This indicates a horizontal space. This does NOT include a bar gap afterward, only the space specified. As with bar width this is 
        in raw units.</li>
    </ol>
    The bar height depends on the mode of operation and will be explained later.</p>
    <h3>Graph Types</h3>
    <p>There are three types of graphs available: bar charts, line graphs and a thin-bar chart. The bar chart is just that: a plain vanilla bar chart. Each data 
    line generates a rectangular bar with the label centered beneath it. If midpoints are enabled they appear as the midpoint of the top of the bar. All bars and 
    gaps are the same width and the only way to change this is with horizontal spacers.</p>
    <p>The thin-bar graph is constructed like a bar graph but with a bar width of zero. Any bar width less than 1 (but greater than zero!) generates this type 
    of graph. The bar intervals are still generated along with the gaps but the 'graph' consists of a vertical line from the x-axis (visible or not) to the 
    midpoint of the (invisible!) bar. For best results use a thick pen for this.</p>
    <p>The line graph is calculated like an invisible-bar bar chart. The top midpoints are calculated and connected left to right by line segments. Intervals are 
    calculated using bar width and gaps, but without the bars. As with any graph the midpoints may be shown as (obvious!) points or not.</p>
    <h3>Modes of Operation</h3>
    <p>This is where things get a bit complicated.Bar height and the interpretation of the 'Vertical scale' setting (NOT tick-mark visibility!) depend on the mode 
    of operation. There are two of these: 'raw' mode and 'relative' mode. In <b>raw mode</b> the bar height is given in raw user units: the vertical scale is exactly 
    the same as the horizontal scale. Each is defined by the <code>SVGTag</code> 'height=', 'width=' and 'viewBox=' attributes. If the '[]Vertical scale' checkbox 
    is enabled the value set becomes the (raw-unit) distance between the y-axis tick marks. The first one appears on the x-axis and they move upward as far as 
    they can. This provides absolute control over bar height but if the margins change the graph can overflow it.</p>
    <p><b>Relative mode</b> is the default and this should NOT be changed without a good reason! In relative mode the 'Vertical scale' setting determines how many 
    units high the y-axis is. If tick marks are enabled they will appear but even if the checkbox is disabled the 'Vertical scale' setting is still used! The 
    default value is '10' which means the vertical axis is ten units, regardless of size or viewBox. The bar heights are given in terms of these relative units 
    hence the graph will scale to the size of its margins. The horizontal axis is, of course, still in raw units.</p>
    <h3>Scripting</h3>
    <p>Since this panel generates unique graphs and not 'just' graph paper its scripting capabilities are extended. The <code>reset()</code> method resets the 
    panel to its factory defaults, which includes the demo bars in the graph data area. Scripts can call the <code>clearBars()</code> method to clear the data 
    area completely. This is especially useful when combined with the <code>addBar()</code> method which adds a single line to the area. This allows graph data 
    to be built up line by line which works well with scripting.</p>
    <p>As with all the other classes in the <code>gpgdemo.*</code> package all GUI elements are totally scriptable too.</p>
*/

public class BarChartPanel extends GraphPanel {

  private static final String[] DEFAULT_BARS = "25:8 Pizza:2.5 Chicken:6 Hamburger:5 Tacos".split(":");

  /* Default pens. */
  private static final int GPEN = 1, HAPEN = 0, VAPEN = 0, TPEN = 22;

  /* Default sizes. */
  private static final int VTICKS = 10, BARSIZE = 50, BARGAP=50;

  private TextArea txaBars = new TextArea(8,30);
  private CheckboxGroup cbgType = new CheckboxGroup();
  private Checkbox cbxBars = new Checkbox("Bar Chart", cbgType, true), cbxLines = new Checkbox("Line chart", cbgType, false);
  private Checkbox cbxVTicks = new Checkbox("Vertical scale"), cbxRelSpace = new Checkbox("Use relative spacing");
  private Checkbox cbxMidpoints = new Checkbox("Show bar midpoints.", true);
  private Checkbox cbxVAxis = new Checkbox("Vertical axis pen", true), cbxHAxis = new Checkbox("Horiz. axis pen", true);
  private TextField txfVAPen = new TextField(), txfHAPen = new TextField(), txfBarPen = new TextField();
  private TextField txfTextPen = new TextField(), txfVTicks = new TextField(), txfBarSize = new TextField();

  /** Standard constructor; nothing special is needed. */
  public BarChartPanel() {
    super();
    setUpGUI();
    GPickle.gpgBarChart = this;
  }

  /** Draw the bar/line graph. */
  public void drawGraph(int mTop, int mBot, int mLeft, int mRight, SVGTag theGraph) {
    boolean isBarChart = cbxBars.getState(), drawVTicks = cbxVTicks.getState(), drawVAxis = cbxVAxis.getState(), drawHAxis = cbxHAxis.getState();
    boolean drawMidpts = cbxMidpoints.getState(), isRelative = cbxRelSpace.getState();
    int graphPen = getPen(txfBarPen, GPEN), textPen = getPen(txfTextPen, TPEN), vertPen = getPen(txfVAPen, VAPEN), horizPen = getPen(txfHAPen, HAPEN);
    int vTicks = asInt(txfVTicks.getText(), VTICKS, 1, null);
    
    String[] tokenz = (""+txfBarSize.getText()+":-1:-1:-1").split(":");
    //CapCom.debug("Bar width: "+tokenz[0]+", Bar gap: "+tokenz[1]);
    double barWidth = asDouble(tokenz[0], BARSIZE, 0.0, null), barGap = asDouble(tokenz[1], BARGAP, 0.0, null);
    double xMin = mLeft * 1.02, xMax = mRight * 0.98, yMin = mBot * 0.95, yMax = mTop * 0.98;
    //CapCom.debug("XMin: "+xMin+", XMax: "+xMax+", yMin: "+yMin+", yMax: "+yMax);
    double vScale = (yMax - yMin) / vTicks, textBaseLine = (yMin + mBot)/2;
    MathDrawSVG MDraw = cox.jmatt.java.MathTools.demo.GPGenerator.getMDraw();
    String graphName = (isBarChart?"Bar":"Line")+"Chart"+CapCom.hitCount();
    if(drawVAxis) {
      MDraw.setLinePen(vertPen);
      theGraph.drawLine(xMin, yMin, xMin, yMax);
      if(drawVTicks) {
        if(isRelative) {
          for(int i=1; i <= vTicks; i++) theGraph.drawLine(xMin - 5, yMin + (vScale*i),xMin + 5, yMin + (vScale*i));
        } else {
          for(double delta = yMin; delta > yMax; delta -= ((double) vTicks)) theGraph.drawLine(xMin - 5, delta, xMin + 5, delta);
        }
      }
    }
    
    if(!isRelative) vScale = -1.0;
    MDraw.setLinePen(horizPen);
    if(drawHAxis) theGraph.drawLine(xMin - 5, yMin, xMax + 5, yMin);
    MDraw.setLinePen(graphPen);
    MDraw.setRectanglePen(graphPen);
    MDraw.setPathPen(graphPen);
    MDraw.setTextPen(textPen);
    double curX = xMin + barGap, curMP = xMin, barHeight = 0;
    String[] barz = txaBars.getText().split("\n");
    PathTag barPath = null;
    for(int i=0; (i < barz.length) && ((curX + barWidth) < xMax); i++) {
      if(!Question.validString(barz[i])) continue;
      curMP = curX + barWidth/2;
      //CapCom.debug("Bar: "+i+", CurX: "+curX+", CurMP: "+curMP);
      tokenz = barz[i].split(" ", 2);
      if(tokenz.length == 2) {
        barHeight = yMin + (vScale * asDouble(tokenz[0], 1.0, 1.0, null));
        if(isBarChart) {
          if(barWidth < 1) {
            theGraph.drawLine(curMP, yMin, curMP, barHeight);
          } else {
            theGraph.drawRectangle(curX, barHeight, barWidth, yMin - barHeight);
          }
          if(drawMidpts) theGraph.drawPoint(curMP, barHeight);
        } else {
          if(barPath == null) {
            barPath = MDraw.getPathTag();
            barPath.MoveTo(curMP, barHeight);
          }
          barPath.LineTo(curMP, barHeight);
          if(drawMidpts) theGraph.drawPoint(curMP, barHeight);
        }
        theGraph.drawText(curMP, textBaseLine, tokenz[1]);
        curX = curX + barWidth + barGap;
      } else {
        curX += asDouble(tokenz[0], 0.0, 0.0, null);
      }
    }
    if(barPath != null) theGraph.add(barPath);
  }

  /** Set the pens used to draw the graph. Passing a negative value ignores that pen.
      @param pGraph The pen used for the bars or graph lines.
      @param pText The pen used for bar/vertex labels.
      @param pVAxis The vertical axis pen.
      @param pHAxis The horizontal axis pen.
  */
  public void setPens(int pGraph, int pText, int pVAxis, int pHAxis) {
    if(pGraph >= 0) txfBarPen.setText(""+pGraph);
    if(pText >= 0) txfTextPen.setText(""+pText);
    if(pVAxis >= 0) txfVAPen.setText(""+pVAxis);
    if(pHAxis >= 0) txfHAPen.setText(""+pHAxis);
  }

  /** Set the vertical scale and enable/disable tick marks. Setting a negative or zero value ignores the vertical scale and just enables/disables.
      @param pEnable Enable or disable vertical tick marks.
      @param pScale The vertical scale value.
  */
  public void setVerticalScale(boolean pEnable, int pScale) {
    cbxVTicks.setState(pEnable);
    if(pScale > 0) txfVTicks.setText(""+pScale);
  }

  /** Enable/disable horizontal and vertical axes and bar midpoint marks.
      @param pHoriz Enable/disable horizontal axis.
      @param pVert Enable/disable vertical axis.
      @param pMidpt Enable/disable bar midpoints.
  */
  public void setEnableAxis(boolean pHoriz, boolean pVert, boolean pMidpt) {
    cbxHAxis.setState(pHoriz);
    cbxVAxis.setState(pVert);
    cbxMidpoints.setState(pMidpt);
  }

  /** Enable/disable relative spacing. */
  public void setRelative(boolean pEnable) { cbxRelSpace.setState(pEnable); }

  /** Select the type of graph to generate: 'true' for a bar chart, 'false' for a line graph. */
  public void setBarChart(boolean isBarChart) {
    cbxBars.setState(isBarChart);
    cbxLines.setState(!isBarChart);
  }

  /** Set the bar width and space between the bars. A bar width less than zero is ignored. If the gap size is less than zero it is ignored but the bar size can 
      still be set.
      @param pWidth The width of the bars on the graph.
      @param pGap The size of the space between bars.
  */
  public void setBarWidth(double pWidth, double pGap) {
    if(!(pWidth < 0)) {
      if(!(pGap < 0)) {
        txfBarSize.setText(""+pWidth+":"+pGap);
      } else {
        txfBarSize.setText(""+pWidth);
      }
    }
  }

  /** Set the data for the bars or points. It is a String[] array with each element consisting of a single number or a number, a space and a label. Single numbers
      are interpreted as horizontal spaces, 'number-space-label' elements are bar or point height and label, depending on the graph type. A null or empty array 
      is silently ignored. Setting valid values does clear the data already there.
  */
  public void setBars(String[] pBars) {
    if((pBars == null) || (pBars.length < 1)) return;
    txaBars.setText("");
    for(String sLine : pBars) txaBars.append(sLine + "\n");
  }

  /** Clear the bar data area. This is a script-only method. */
  public void clearBars() { txaBars.setText(""); }

  /** <p>Add a bar or horizontal space to the bar area. If 'pSize' is zero or less nothing happens. If the label is null or blank a spacing gap is added.<p>
      <p>This is a script-only method.</p>
      @param pSize The height of the bar or width of the horizontal space.
      @param pLabel The bar label, if defined. If null or blank a space is added instead.
  */
  public void addBar(double pSize, String pLabel) {
    if(pSize > 0) {
      if(Question.validString(pLabel)) {
        txaBars.append(""+pSize+" "+pLabel+"\n");
      } else {
        txaBars.append(""+pSize+"\n");
      }
    }
  }

  /** Reset the panel to its defaults. */
  public void reset() {
    setPens(GPEN, TPEN, VAPEN, HAPEN);
    setVerticalScale(false, VTICKS);
    setEnableAxis(true, true, false);
    setBarChart(true);
    setBarWidth(BARSIZE, BARGAP);
    setBars(DEFAULT_BARS);
    setRelative(true);
  }

  public String toPizza(String pPrefix) {
    StringBuffer retval = new StringBuffer("# BarChartPanel Stored State.");
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
      retval.append("Prefix: ").append(pPrefix);
    }
    retval.append("\n");
    retval.append(pPrefix).append("bar.GraphType=").append(cbxBars.getState()?"Bar":"Line").append("\n");
    retval.append(getKey(txfBarSize, pPrefix, "bar.barWidth"));
    retval.append(getKey(txfBarPen, pPrefix, "bar.graphPen"));
    retval.append(getKey(txfTextPen, pPrefix, "bar.textPen"));
    retval.append(getKey(cbxVAxis, pPrefix, "bar.VAxisPen.enabled"));
    retval.append(getKey(txfVAPen, pPrefix, "bar.VAxisPen.value"));
    retval.append(getKey(cbxHAxis, pPrefix, "bar.HAxisPen.enabled"));
    retval.append(getKey(txfHAPen, pPrefix, "bar.HAxisPen.value"));
    retval.append(getKey(cbxVTicks, pPrefix, "bar.verticalScale.enabled"));
    retval.append(getKey(txfVTicks, pPrefix, "bar.verticalScale.value"));
    retval.append(getKey(cbxMidpoints, pPrefix, "bar.showMidpoints"));
    retval.append(getKey(cbxRelSpace, pPrefix, "bar.relativeSpace"));
    String[] linez = txaBars.getText().split("\n");
    for(int i=0; i < linez.length; i++) {
      retval.append(pPrefix).append("bar.bar.").append(i).append("=").append(linez[i]).append("\n");
    }
    return retval.toString();
  }

  public void eatPizza(Properties pPizza, String pPrefix) {
    if((pPizza == null) || pPizza.isEmpty()) return;
    String sPre = "", sLine = "";
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
    }
    setBarChart(pPizza.getProperty(sPre+"bar.GraphType", "Bar").equalsIgnoreCase("Bar"));
    setKey(pPizza,txfBarSize, pPrefix, "bar.barWidth", ""+BARSIZE+":"+BARGAP);
    setKey(pPizza,txfBarPen, pPrefix, "bar.graphPen", ""+GPEN);
    setKey(pPizza,txfTextPen, pPrefix, "bar.textPen", ""+TPEN);
    setKey(pPizza,cbxVAxis, pPrefix, "bar.VAxisPen.enabled");
    setKey(pPizza,txfVAPen, pPrefix, "bar.VAxisPen.value", ""+VAPEN);
    setKey(pPizza,cbxHAxis, pPrefix, "bar.HAxisPen.enabled");
    setKey(pPizza,txfHAPen, pPrefix, "bar.HAxisPen.value", ""+HAPEN);
    setKey(pPizza,cbxVTicks, pPrefix, "bar.verticalScale.enabled");
    setKey(pPizza,txfVTicks, pPrefix, "bar.verticalScale.value", "10");
    setKey(pPizza,cbxMidpoints, pPrefix, "bar.showMidpoints");
    setKey(pPizza,cbxRelSpace, pPrefix, "bar.relativeSpace");
    clearBars();
    int i=0;
    while((sLine = pPizza.getProperty(pPrefix+"bar.bar."+i, null)) != null) {
      txaBars.append(sLine+"\n");
      i++;
    }
  }

  private void setUpGUI() {
    setLayout(new GridBagLayout());
    GridBagConstraints gbc = new GridBagConstraints(); gbc.insets = new Insets(1,1,1,1);
    gbc.gridx = 1; gbc.gridy = 1; gbc.fill = GridBagConstraints.HORIZONTAL;
    add(cbxBars, gbc); gbc.gridx++;
    add(cbxLines, gbc); gbc.gridx++;
    add(new Label("Bar Width:Gap"), gbc); gbc.gridx++;
    add(txfBarSize, gbc); gbc.gridx=1; gbc.gridy++;
    add(cbxVAxis, gbc); gbc.gridx++;
    add(txfVAPen, gbc); gbc.gridx++;
    add(new Label("Graph Pen"), gbc); gbc.gridx++;
    add(txfBarPen, gbc); gbc.gridx=1; gbc.gridy++;
    add(cbxHAxis, gbc); gbc.gridx++;
    add(txfHAPen, gbc); gbc.gridx=1; gbc.gridy++;
    add(new Label("Text Pen"), gbc); gbc.gridx++;
    add(txfTextPen, gbc); gbc.gridx=1; gbc.gridy++;
    add(cbxVTicks, gbc); gbc.gridx++;
    add(txfVTicks, gbc); gbc.gridx=1; gbc.gridy++; gbc.gridwidth=2;
    add(cbxMidpoints, gbc); gbc.gridy++;
    add(cbxRelSpace, gbc); gbc.gridy++;
    // Add bar area
    gbc.gridx=3; gbc.gridy=3; gbc.gridheight=GridBagConstraints.REMAINDER; gbc.gridwidth=2;
    add(txaBars, gbc);
    reset();
  }

}

