package cox.jmatt.java.MathTools.demo.gpgdemo;

import java.awt.*;
import java.io.*;
import java.util.Properties;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.markup.svg.*;

/** <p>This is the base Panel for the graph paper generator panels. It defines the basic structure and function, others customize it.</p>
    <h3>Graph Persistence and Restoration</h3>
    <p>The <code>toPizza()</code> method can be used to generate a <code>java.util.Properties</code>-style snapshot of the current state of the panel. All graph 
    settings are stored and the key values are quite obvious. If desired a prefix can be prepended to the key values to allow storing more than one graph in a 
    single text file. This allows a particular graph to be saved for later use. Of course prudent use of the other script tools can even have one graph (prefix) 
    picked from a file full of them at random. Once graph data has been saved it can be loaded. There are five options for this: a <code>Properties</code> object, 
    the Global Pizza, as raw data in a String, from a file and from a resource stream.</p>
*/
public abstract class GraphPanel extends Panel {

  /** Standard constructor. Does nothing but call super(). */
  public GraphPanel() { super(); }

  /** <p>This is the primary method for this class. Subclasses override it to draw the graph paper on the <code>SVGTag</code> sent in. The desired margins are 
      also sent in and are assumed to be in user units. Subclasses may construct a <code>MathDrawSVG</code> instance but the 'pPaper' is what will ultimately be 
      rendered and printed. The default implementation prints a Debug-level message.</p>
      @param mTop The top margin of the drawable area.
      @param mBottom The bottom margin of the drawable area.
      @param mLeft The left margin.
      @param mRight The right margin.
      @param pPaper The <code>SVGTag</code> on which the drawing is to occur.
  */
  public void drawGraph(int mTop, int mBottom, int mLeft, int mRight, SVGTag pPaper) { CapCom.debug("Drawing paper"); }

  /** <p>This method provides a way for the individual generators to reset to default values. This implementation prints a debug message containing the name 
      of the class invoking it. This allows subclasses to print a simple message by calling <code>super.reset()</code>.</p>
  */
  public void reset() { CapCom.debug("Reset: "+this.getClass().getSimpleName()); }

  /** <p>This method extracts the configuration and state information from a generator and produces <code>java.util.Properties</code>-style output. A String 
      prefix can be prepended to each key. If 'pPrefix' is not null or blank it will be attached to the front of the key String, separated by a single dot. If 
      the prefix does not end with a dot one will be supplied automatically.</p>
      <p>Use of the prefix allows multiple configurations to be stored in a single file. Also, since each generator has its own unique prefix (separate from the 
      one supplied) multiple generator configs can be stored in a single file.</p>
      <p>This is primarily a scripting method. It was developed in order to allow the <i>perfect</i> settings to be saved in a (scripting-) language-agnostic 
      way that was also human-readable. The individual property names are typically Java-fied versions of their GUI labels. If MathTools is run from its JAR and 
      the standard GUI is present the instances exposed to scripts are the ones in the GUI, so setting values in the GUI will affect script output and settings 
      made by the script are reflected in the GUI.</p>
      @param pPrefix The prefix to prepend to each key. Ignored if null or blank.
  */
  public abstract String toPizza(String pPrefix);

  /** <p>This method accepts a <code>Properties</code> object and an optional String prefix and attempts to extract configuration data from it. It complements 
      <code>toPizza()</code>. If 'pPizza' is null or empty nothing happens: the method returns silently. If 'pPrefix' is not null or blank it is used as a prefix 
      to each key. If the String sent in does not end with a dot one will be supplied automatically. All of the keys requested have default valus that match the 
      ones on the GUI.</p>
      <p>Under normal cirucmstances the generator instances exposed to scripts <i>are</i> the ones in the GUI so changes using one are reflected in the other.</p>
      @param pPizza A <code>java.util.Properties</code> object containing configuration data.
      @param pPrefix An optional prefix to include with each key. Ignored if null or blank.
  */
  public abstract void eatPizza(Properties pPizza, String pPrefix);

  /** Attempt to extract a graph (with optional prefix) from the Global Pizza. */
  public void eatGlobalPizza(String pPrefix) { eatPizza(CapCom.getPizza(), pPrefix); }

  /** <p>Load a graph configuration from a file or raw String. Either way the data is parsed into a <code>Properties</code> object then pulls the data from it. 
      A prefix can be supplied or not. Any errors are reported at Error level. If the filename/raw data is null or blank the method returns silently.</p>
      <p>If 'pData' is a file and not raw data it can be loaded from the file system or via <code>getResourceAsStream()</code>. The 'pGRAS' argument controls 
      this; it is provided for Applet use.</p>
      @param pData The raw data or filename.
      @param pIsFile 'true' to load from a file, 'false' for raw data.
      @param pGRAS Controls whether to use a resource stream ('true') or a normal <code>fileInputStream</code> ('false').
      @param pPrefix The graph data prefix.
  */
  public void loadPizza(String pData, String pPrefix, boolean pIsFile, boolean pGRAS) {
    if(!Question.validString(pData)) return;
    try {
      InputStream isData = null;
      Properties pizzaTemp = new Properties();
      if(pIsFile) {
        if(pGRAS) {
          isData = this.getClass().getResourceAsStream(pData);
        } else {
          isData = new FileInputStream(pData);
        }
      } else {
        isData = new ByteArrayInputStream(pData.getBytes());
      }
      pizzaTemp.load(isData);
      eatPizza(pizzaTemp, pPrefix);
      isData.close();
    } catch(Exception ex) { CapCom.error("Error loading pizza.", ex); }
  }

  /** Convert a String to int with default values and clamping. If the String is blank, empty, or cannot be parsed the default value is returned. The minimum and 
      maximum values are enforced if defined, if not defined they are ignored. NOTE: The clamping values are <b>not</b> checked against each other!
      @param pData The String to be converted.
      @param pDefault The default value.
      @param pMin The minimum value, or null to ignore it.
      @param pMax The maximum value, ignored if null.
  */
  public int asInt(String pData, int pDefault, Integer pMin, Integer pMax) {
    int retval = pDefault;
    try { retval = Integer.parseInt(pData); } catch(Exception ex) { CapCom.debug("Integer exception parsing: "+pData, ex); }
    if((pMin != null) && (retval < pMin.intValue())) retval = pMin.intValue();
    if((pMax != null) && (retval > pMax.intValue())) retval = pMax.intValue();
    return retval;
  }

  /** Extract the contents of a <code>TextField</code> and interpret the result as a pen. If the default is less than zero it is set to zero.
      @param txfPen The TextField holding a pen value.
      @param pDefault The default value for the pen. Clamped to zero on the low end.
  */
  public int getPen(TextField txfPen, int pDefault) {
    if(pDefault < 0) pDefault = 0;
    if(txfPen == null) return pDefault;
    return asInt(txfPen.getText(), pDefault, 0, cox.jmatt.java.MathTools.demo.GPGenerator.MAX_PEN);
  }

  /** Shortcut method: defaults to pen zero. */
  public int getPen(TextField txfPen) { return getPen(txfPen, 0); }

  /** Convert a String to double with default and clamping. */
  public double asDouble(String pData, double pDefault, Double pMin, Double pMax) {
    double retval = pDefault;
    try { retval = Double.parseDouble(pData); } catch(Exception ex) { CapCom.debug("Double exception parsing: "+pData, ex); }
    if((pMin != null) && (retval < pMin.doubleValue())) retval = pMin.doubleValue();
    if((pMax != null) && (retval > pMax.doubleValue())) retval = pMax.doubleValue();
    return retval;
  }

  /** Extract the value of a <code>TextField</code> and return it as a <code>Properties</code>-formatted line. If the key value is null or blank or 'pTXF' is null 
      nothing happens and an empty String is returned.
      @param pTXF The <code>TextField</code> to supply the value.
      @param pPrefix The prefix to prepend to the key.
      @param pKey The key used to store the value.
      @return A line formatted as '[pPrefix.]pKey=pTXF.getText()/n'.
  */
  protected String getKey(TextField pTXF, String pPrefix, String pKey) {
    if((pTXF == null) || !Question.validString(pKey)) return "";
    return ((pPrefix == null)?"":pPrefix)+pKey+"="+pTXF.getText()+"\n";
  }

  /** Extract the value of a <code>Checkbox</code> and return it as a <code>Properties</code>-formatted line. If the key value is null or blank or 'pCBX' is null 
      nothing happens and an empty String is returned.
      @param pCBX The <code>Checkbox</code> to supply the value.
      @param pPrefix The prefix to prepend to the key.
      @param pKey The key used to store the value.
      @return A line formatted as '[pPrefix]pKey=pCBX.getState()/n'.
  */
  protected String getKey(Checkbox pCBX, String pPrefix, String pKey) {
    if((pCBX == null) || !Question.validString(pKey)) return "";
    return ((pPrefix == null)?"":pPrefix)+pKey+"="+pCBX.getState()+"\n";
  }

  /** Extract the value from the supplied <code>Properties</code> object and use it to set the value of the <code>TextField</code> sent in. If 'pPizza' is null, 
      pTXF is null or 'pKey' is null or blank nothing happens. A default value can be supplied; if null it is replaced by a blank.
      @param pPizza The Properties object used to extract the value.
      @param pTXF The text field to be set.
      @param pPrefix The key prefix; replaced with a blank if null.
      @param pKey The key used to extract the value.
      @param pDefault The default value if the key is not found.
  */
  protected void setKey(Properties pPizza, TextField pTXF, String pPrefix, String pKey, String pDefault) {
    if((pPizza == null) || (pTXF == null) || !Question.validString(pKey)) return;
    if(pDefault == null) pDefault = "";
    if(pPrefix == null) pPrefix = "";
    pTXF.setText(pPizza.getProperty(pPrefix+pKey, pDefault));
  }

  /** Extract the value from the supplied <code>Properties</code> object and use it to set the value of the <code>Checkbox</code> sent in. If 'pPizza' is null, 
      'pCBX' is null or 'pKey' is null or blank nothing happens.
      @param pPizza The Properties object used to extract the value.
      @param pCBX The text field to be set.
      @param pPrefix The key prefix; replaced with a blank if null.
      @param pKey The key used to extract the value.
  */
  protected void setKey(Properties pPizza, Checkbox pCBX, String pPrefix, String pKey) {
    if((pPizza == null) || (pCBX == null) || !Question.validString(pKey)) return;
    if(pPrefix == null) pPrefix = "";
    pCBX.setState("true".equalsIgnoreCase(pPizza.getProperty(pPrefix+pKey)));
  }

  /** Set or clear the <code>GPGenerator.GP_DEBUG</code> field. */
  public static final void setGPDebugMode(boolean pEnable) { cox.jmatt.java.MathTools.demo.GPGenerator.GP_DEBUG = pEnable; }
  /** Instance debug mode method. */
  public final void setGPDebug(boolean pEnable) { cox.jmatt.java.MathTools.demo.GPGenerator.GP_DEBUG = pEnable; }
  /** Finalized static method to fetch the value of <code>GPGenerator.GP_DEBUG</code>. Intended for development of new generators. */
  protected final boolean GP_DEBUG() { return cox.jmatt.java.MathTools.demo.GPGenerator.GP_DEBUG; }

}

