package cox.jmatt.java.MathTools.demo.gpgdemo;

import java.awt.*;
import java.util.Properties;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.demo.*;
import cox.jmatt.java.MathTools.markup.svg.*;

/** <p><code>HexGridPanel</code> is a quick and dirty generator for hexagonal grid paper. The hexes are generated based on the length of a side and the actual 
    hexagon is (roughly not quite) twice this so plan accordingly. The number of hexes in a column can be limited as can the number of columns; by default the 
    generator tries to fill the defined area (it may overflow so tinkering is useful). The grid can also be enumerated (top, bottom or sides, any combination) and 
    the starting values can be specified.</p>
*/

public class HexGridPanel extends GraphPanel {

  private static final double COS60 = Math.cos(Math.PI/3.0);
  private static final double SIN60 = Math.sin(Math.PI/3.0);

  private TextField txfSize = new TextField(), txfLinePen = new TextField(), txfMaxVert = new TextField(), txfMaxHoriz = new TextField();
  private TextField txfNumHoriz = new TextField(), txfNumVert = new TextField(), txfTextPen = new TextField();
  private Checkbox cbxMaxVert = new Checkbox("Max. vertical hexes"), cbxMaxHoriz = new Checkbox("Max. horiz. hexes");
  private Checkbox cbxNumTop = new Checkbox("Numbers @ top"), cbxNumBot = new Checkbox("Numbers @ bottom");
  private Checkbox cbxNumVert = new Checkbox("Number vertically");

  /** Regular constructor. Nothing to see here. */
  public HexGridPanel() {
    super();
    _setUpGUI();
    GPickle.gpgHexGrid = this;
    reset();
  }

  /** Draw the paper. */
  public void drawGraph(int pTop, int pBot, int pLeft, int pRight, SVGTag pPaper) {
    double yMin = pTop, xMin = pLeft, yMax = pBot, xMax = pRight;
    int textPen = getPen(txfTextPen, 22), linePen = getPen(txfLinePen);
    boolean numberTop = cbxNumTop.getState(), numberBot = cbxNumBot.getState(), numberSides = cbxNumVert.getState();
    double dSize = asDouble(txfSize.getText(), 25, 0.001, null);
    boolean doVMax = cbxMaxVert.getState(), doHMax = cbxMaxHoriz.getState();
    int maxYCount = asInt(txfMaxVert.getText(), 10, 0, null), maxXCount = asInt(txfMaxHoriz.getText(), 10, 0, null);
    MathDrawSVG MDraw = cox.jmatt.java.MathTools.demo.GPGenerator.getMDraw();
    MDraw.setLinePen(linePen);
    MDraw.setPathPen(linePen);
    MDraw.setTextPen(textPen);
    double alpha = dSize * SIN60; // Vertical change when drawing diagonal side.
    double beta = dSize * COS60; // Length of the line segment from a vertex to the line joining its adjascent vertices
    double gutter = (yMax - yMin) * 0.03;
    if(numberTop) yMin += gutter;
    if(numberBot) yMax -= gutter;
    if(numberSides) {
      gutter = (xMax - xMin) * 0.03;
      xMin += gutter;
      xMax -= gutter;
    }
    double curX = xMin + beta, curY = yMin;
    PathTag thePath = null;
    if(GP_DEBUG()) {
      pPaper.drawLine(xMin, yMin, xMax, yMin);
      pPaper.drawLine(xMin, yMin, xMin, yMax);
      pPaper.drawLine(xMax, yMin, xMax, yMax);
      pPaper.drawLine(xMax, yMax, xMin, yMax);
    }
    double dRawHexesPerColumn = (yMax - yMin) / (2*alpha), dNumberOfColumns = (xMax - xMin) / (2*(dSize + beta));
    int hexesPerCol = (int) dRawHexesPerColumn, numOfCols = (int) dNumberOfColumns;
    numOfCols *= 2;
    if(Math.IEEEremainder(dRawHexesPerColumn, hexesPerCol) < 0.5) hexesPerCol--;
    if(doVMax && (maxYCount < hexesPerCol)) hexesPerCol = maxYCount;
    if(doHMax && (maxXCount < numOfCols)) numOfCols = maxXCount;

    curX = xMin + beta;
    for(int i=0; i < numOfCols; i++) {
      curY = yMin;
      thePath = MDraw.getPathTag();
      for(int j=0; j < hexesPerCol; j++) {
        if((i % 2) == 0) {
          // Upper column
          thePath.MoveTo(curX, curY);
          thePath.lineTo(-beta, alpha);
          thePath.lineTo(beta, alpha);
          thePath.hLineTo(dSize);
          thePath.lineTo(beta, -alpha);
          thePath.lineTo(-beta, -alpha);
          if(j == 0) thePath.hLineTo(-dSize);
        } else {
          // Lower column
          thePath.MoveTo(curX, curY + alpha);
          thePath.hLineTo(dSize);
          if(j == (hexesPerCol - 1)) {
            thePath.moveTo(-dSize - beta, alpha);
            thePath.lineTo(beta, alpha);
            thePath.hLineTo(dSize);
            thePath.lineTo(beta, -alpha);
            if(i == (numOfCols - 1)) thePath.lineTo(-beta, -alpha);
          } else {
            if(i == (numOfCols - 1)) {
              thePath.lineTo(beta, alpha);
              thePath.lineTo(-beta, alpha);
            }
          }
        }
        curY += 2*alpha;
      }
      pPaper.add(thePath);
      curX += (dSize + beta);
    }

    // Numbering. curY and curX are at the bottom-right of the graph.
    int startX = asInt(txfNumHoriz.getText(), 0, null, null), startY = asInt(txfNumVert.getText(), 0, null, null);
    if(numberSides) {
      gutter = curX + ((xMin - pLeft)/2 - pLeft);
      for(int i = 0; i < hexesPerCol; i++) {
        pPaper.drawText((double) pLeft, curY - alpha - (2 * alpha * i), ""+(startY + i));
        pPaper.drawText(gutter, curY - (2 * alpha * i), ""+(startY + i));
      }
    }
    if(numberTop) {
      gutter = (yMin + pTop)/2;
      double xtPos = xMin + beta + dSize/2;
      for(int i=0; i < numOfCols; i++) {
        pPaper.drawText(xtPos + (dSize + beta) * i, gutter, ""+(startX + i));
      }
    }
    if(numberBot) {
      gutter = yMin + alpha + 2 * alpha * hexesPerCol;
      double xtPos = xMin + beta + dSize/2;
      for(int i=0; i < numOfCols; i++) {
        pPaper.drawText(xtPos + (dSize + beta) * i, gutter, ""+(startX + i));
      }
    }
  }

  /** Set the size of a hex side. Negative or zero values set the default value. */
  public void setHexSize(double pSize) {
    if(pSize <= 0) {
      txfSize.setText("40");
    } else {
      txfSize.setText(""+pSize);
    }
  }

  /** Set the pens: line and text. Negative values are ignored.
      @param pLine The line pen number.
      @param pText The text pen number.
  */
  public void setPens(int pLine, int pText) {
    if(pLine > -1) txfLinePen.setText(""+pLine);
    if(pText > -1) txfTextPen.setText(""+pText);
  }

  /** Set the maximum number of vertical hexes and enable or disable it. Negative values are ignored. */
  public void setMaxVert(boolean pEnable, int pMax) {
    cbxMaxVert.setState(pEnable);
    if(pMax > -1) txfMaxVert.setText(""+pMax);
  }

  /** Set the maximum number of horizontal hexes and enable or disable it. Negative values are ignored. */
  public void setMaxHoriz(boolean pEnable, int pMax) {
    cbxMaxHoriz.setState(pEnable);
    if(pMax > -1) txfMaxHoriz.setText(""+pMax);
  }

  /** Enable/disable numbering and set the starting values. */
  public void setNumbering(boolean pTop, boolean pBot, int pHoriz, boolean pSides, int pVert) {
    cbxNumTop.setState(pTop); cbxNumBot.setState(pBot); cbxNumVert.setState(pSides);
    txfNumHoriz.setText(""+pHoriz); txfNumVert.setText(""+pVert);
  }

  /** Reset everything to factory defaults. */
  public void reset() {
    setHexSize(-1);
    setPens(0, 22);
    setMaxVert(false, 10);
    setMaxHoriz(false, 10);
    setNumbering(false, false, 0, false, 0);
  }

  public String toPizza(String pPrefix) {
    StringBuffer retval = new StringBuffer("# HexGridPanel configuration data.");
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
      retval.append(" Prefix: ").append(pPrefix);
    } else {
      pPrefix = "";
    }
    retval.append("\n");
    retval.append(getKey(txfSize, pPrefix, "hex.hexSide"));
    retval.append(getKey(txfLinePen, pPrefix, "hex.linePen"));
    retval.append(getKey(txfTextPen, pPrefix, "hex.textPen"));
    retval.append(getKey(cbxMaxVert, pPrefix, "hex.maxVertical.enabled"));
    retval.append(getKey(txfMaxVert, pPrefix, "hex.maxVertical.value"));
    retval.append(getKey(cbxMaxHoriz, pPrefix, "hex.maxHorizontal.enabled"));
    retval.append(getKey(txfMaxHoriz, pPrefix, "hex.maxHorizontal.value"));
    retval.append(getKey(cbxNumVert, pPrefix, "hex.numberSides.enabled"));
    retval.append(getKey(txfNumVert, pPrefix, "hex.numberSides.value"));
    retval.append(getKey(cbxNumTop, pPrefix, "hex.numberTop.enabled"));
    retval.append(getKey(cbxNumBot, pPrefix, "hex.numberBottom"));
    retval.append(getKey(txfNumHoriz, pPrefix, "hex.numberTopBot.value"));
    return retval.toString();
  }

  public void eatPizza(Properties pPizza, String pPrefix) {
    if((pPizza == null) || pPizza.isEmpty()) return;
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
    } else {
      pPrefix = "";
    }
    setKey(pPizza, txfSize, pPrefix, "hex.hexSide", "40");
    setKey(pPizza, txfLinePen, pPrefix, "hex.linePen", "0");
    setKey(pPizza, txfTextPen, pPrefix, "hex.textPen", "22");
    setKey(pPizza, txfMaxVert, pPrefix, "hex.maxVertical.value", "10");
    setKey(pPizza, txfMaxHoriz, pPrefix, "hex.maxHorizontal.value", "10");
    setKey(pPizza, txfNumVert, pPrefix, "hex.numberSides.value", "0");
    setKey(pPizza, txfNumHoriz, pPrefix, "hex.numberTopBot", "0");
    setKey(pPizza, cbxMaxVert, pPrefix, "hex.maxVertical.enabled");
    setKey(pPizza, cbxMaxHoriz, pPrefix, "hex.maxHorizontal.enabled");
    setKey(pPizza, cbxNumVert, pPrefix, "hex.numberSides.enabled");
    setKey(pPizza, cbxNumTop, pPrefix, "hex.numberTop.enabled");
    setKey(pPizza, cbxNumBot, pPrefix, "hex.numberBottom.enabled");
  }
  

  private void _setUpGUI() {
    setLayout(new GridLayout(0,3));
    add(new Label("Hex side size")); add(txfSize); add(new Label(" (user units)"));
    add(new Label("Line pen")); add(txfLinePen); add(new Label());
    add(cbxMaxVert); add(txfMaxVert); add(new Label(" (hexes)"));
    add(cbxMaxHoriz); add(txfMaxHoriz); add(new Label(" (hexes"));
    add(new Label("Number pen")); add(txfTextPen); add(new Label());
    add(cbxNumVert); add(new Label("Start numbers at:")); add(txfNumVert);
    add(cbxNumTop); add(cbxNumBot); add(txfNumHoriz);
  }
}

