package cox.jmatt.java.MathTools.demo.gpgdemo;

import java.awt.*;
import java.util.Properties;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.demo.*;
import cox.jmatt.java.MathTools.markup.svg.*;

/** <p>This generator creates three types of isometrically-scaled paper: lined-grid isometric, diamond grid and isometric dots. Line-grid isometric divides the 
    surface of the paper into equilateral triangles: the horizontal distance between vertices equals each diagonal distance between vertices. Diamond-grid paper 
    is the same is lined isometric but without the horizontals. (To create diamond-grid uncheck the '[]Show horizontals' Checkbox!). Isometric dots don't show 
    the triangles, merely the vertices. A point is placed at each one.</p>
    <p>If desired, a border can be drawn around the grid area. The border pen is selectable and will be used if the '[]Border pen' Checkbox is enabled. The line 
    pen is used for horizontals and diagonals. Iso-dots paper uses <code>CircleTag</code>'s 'point' method.</p>
    <p>By default for lined paper the first (upper-leftmost) stroke will be diagonal down and the first iso-dot is in the upper left corner. Checking the 
    '[]Offset first row' Checkbox changes this. The first diagonal stroke will be upward and the first iso-dot is offset horizontally by half a length.</p>
*/
public class IsometricPanel extends GraphPanel {

  private static final double COS30 = Math.cos(Math.PI/6);

  private CheckboxGroup cbgType = new CheckboxGroup();
  private Checkbox cbxLines = new Checkbox("Grid with lines", cbgType, true), cbxDots = new Checkbox("Dots-only grid", cbgType, false);
  private Checkbox cbxBorders = new Checkbox("Border pen:"), cbxHorizontals = new Checkbox("Show horizontal lines", true);
  private Checkbox cbxOffset = new Checkbox("Offset first row");
  private TextField txfSize = new TextField(), txfLinePen = new TextField(), txfBorderPen = new TextField();

  /** Usual constructor. */
  public IsometricPanel() {
    super();
    _setUpGUI();
    GPickle.gpgIsometric = this;
  }

  /** Draw the paper. */
  public void drawGraph(int pTop, int pBot, int pLeft, int pRight, SVGTag pPaper) {
    double yMin = pTop, xMin = pLeft, yMax = pBot, xMax = pRight, dWidth = xMax - xMin;
    double dSize = asDouble(txfSize.getText(), 1d, 0.01, null), deltaX = dSize/2, deltaY = dSize * COS30;
    int linePen = getPen(txfLinePen), borderPen = getPen(txfBorderPen);
    MathDrawSVG MDraw = cox.jmatt.java.MathTools.demo.GPGenerator.getMDraw();
    MDraw.setPathPen(linePen);
    String sZig = "PZig"+CapCom.hitCount(), sZag = "PZag"+CapCom.hitCount();
    GroupTag gtZig = MDraw.getGroupTag(sZig), gtZag = MDraw.getGroupTag(sZag);
    sZig = "#"+sZig; sZag = "#"+sZag;
    int flipper = 1;
    double hPos = 0d, vPos = 0d, borderX = 0d, borderY = 0d;
    if(cbxLines.getState()) {
      PathTag ptZig = MDraw.getPathTag(), ptZag = MDraw.getPathTag();
      ptZig.MoveTo(0, 0); ptZag.MoveTo(0, deltaY);
      for(hPos = deltaX; hPos + deltaX < dWidth; hPos += deltaX) {
        // Grid with lines
        ptZig.lineTo(deltaX, deltaY * flipper);
        ptZag.lineTo(deltaX, -deltaY * flipper);
        flipper *= -1;
        borderX = hPos;
      }
      if(cbxHorizontals.getState()) {
        hPos -= deltaX;
        ptZig.MoveTo(hPos, deltaY); ptZig.hLineTo(-hPos);
        ptZag.MoveTo(hPos, deltaY); ptZag.hLineTo(-hPos);
        hPos += deltaX;
      }
      hPos += (dSize + deltaX);
      borderX += xMin;
      gtZig.add(ptZig); gtZag.add(ptZag);
    } else {
      for(hPos = 0d; hPos + dSize < dWidth; hPos += dSize) gtZig.drawPoint(hPos, 0d);
      borderX = hPos;
      for(hPos = deltaX; hPos + dSize < dWidth; hPos += dSize) gtZag.drawPoint(hPos, 0d);
      if(hPos > borderX) borderX = hPos;
      borderX += (xMin - dSize);
    }
    pPaper.add(gtZig); pPaper.add(gtZag);
    UseTag utLine = null;
    TransformPen tPen = null;
    boolean bZigZag = !cbxOffset.getState();
    for(vPos = yMin; vPos + deltaY < yMax; vPos += deltaY) {
      tPen = MDraw.getTransformPen();
      tPen.translate(xMin, vPos);
      utLine = MDraw.getUseTag(bZigZag?sZig:sZag);
      utLine.setTransform(tPen);
      pPaper.add(utLine);
      bZigZag = !bZigZag;
    }
    if(cbxDots.getState()) vPos -= deltaY;
    if(cbxBorders.getState()) {
      MDraw.setPathPen(borderPen);
      PathTag ptBorder = MDraw.getPathTag();
      ptBorder.MoveTo(xMin, yMin);
      ptBorder.LineTo(borderX, yMin);
      ptBorder.LineTo(borderX, vPos);
      ptBorder.LineTo(xMin, vPos);
      ptBorder.LineTo(xMin, yMin);
      pPaper.add(ptBorder);
    }
  }

  /** Restore factory values. */
  public void reset() {
    setType(true);
    setSize(25d);
    setLines(0, true);
    setBorders(false, 0);
  }

  /** Set the type of graph: lines or dots.
      @param pIsLines 'true' for a lined grid, 'false' for dots only.
  */
  public void setType(boolean pIsLines) {
    cbxLines.setState(pIsLines); cbxDots.setState(!pIsLines);
  }

  /** Set the side length/dot spacing. Any value zero or less becomes 25. */
  public void setSize(double pSize) {
    txfSize.setText((pSize <= 0)?"25.0":""+pSize);
  }

  /** Set the pen used to draw the lines and enable/disable horizontal lines.
      @param pPen The line pen used to draw the grid.
      @param pHorizontals 'true' to enable horizontal lines, 'false' to leave them off.
  */
  public void setLines(int pPen, boolean pHorizontals) {
    txfLinePen.setText("" +((pPen < 0)?0: pPen));
    cbxHorizontals.setState(pHorizontals);
  }

  /** Enable/disable first-row offset. If 'true' the lines start with an upward diagonal and the dots start offset laterally. */
  public void setOffsetFirst(boolean pOffset) { cbxOffset.setState(pOffset); }

  /** Enable/disable borders and set the border pen. */
  public void setBorders(boolean pEnable, int pPen) {
    cbxBorders.setState(pEnable);
    txfBorderPen.setText("" +((pPen < 0)?0: pPen));
  }

  private void _setUpGUI() {
    setLayout(new GridLayout(0, 2));
    add(cbxLines); add(cbxDots);
    add(new Label("Side length:")); add(txfSize);
    add(new Label("Line pen:")); add(txfLinePen);
    add(cbxOffset); add(cbxHorizontals);
    add(new Label("")); add(new Label(""));
    add(cbxBorders); add(txfBorderPen);
    add(new Label("")); add(new Label(""));
    reset();
  }

  public String toPizza(String pPrefix) {
    StringBuffer retval = new StringBuffer("# IsometricPanel configuration data.");
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
      retval.append(" Prefix: ").append(pPrefix);
    } else {
      pPrefix = "";
    }
    retval.append("\n");
    retval.append(pPrefix).append("isometric.graphType.").append(cbxLines.getState()?"Lines":"Dots").append("\n");
    retval.append(getKey(txfSize, pPrefix, "isometric.size"));
    retval.append(getKey(txfLinePen, pPrefix, "isometric.linePen"));
    retval.append(getKey(cbxHorizontals, pPrefix, "isometric.horizontals.enabled"));
    retval.append(getKey(cbxOffset, pPrefix, "isometric.offsetFirst"));
    retval.append(getKey(cbxBorders, pPrefix, "isometric.border.enabled"));
    retval.append(getKey(txfBorderPen, pPrefix, "isometric.border.pen"));
    return retval.toString();
  }

  public void eatPizza(Properties pPizza, String pPrefix) {
    if((pPizza == null) || pPizza.isEmpty()) return;
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
    } else {
      pPrefix = "";
    }
    setType("Lines".equalsIgnoreCase(pPizza.getProperty(pPrefix+"isometric.graphType")));
    setKey(pPizza, txfLinePen, pPrefix, "isometric.size", "25.0");
    setKey(pPizza, txfLinePen, pPrefix, "isometric.linePen", "0");
    setKey(pPizza, cbxHorizontals, pPrefix, "isometric.horizontals.enabled");
    setKey(pPizza, cbxOffset, pPrefix, "isometric.offsetFirst");
    setKey(pPizza, cbxBorders, pPrefix, "isometric.border.enabled");
    setKey(pPizza, txfBorderPen, pPrefix, "isometric.border.pen", "0");
  }

}


