package cox.jmatt.java.MathTools.demo.gpgdemo;

import java.awt.*;
import java.util.Properties;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.demo.*;
import cox.jmatt.java.MathTools.markup.svg.*;

/** <p>This class is designed to create lettering guides like the ones used by children when learning to write. It was inspired by <code>MusicPanel</code>. Since 
    lettering guides need lines this class offers six different types of lines, all configurable. It also offers more in the way of spacing options: more options 
    and more flexibility.</p>
    <p>The line options are solid major, solid minor, broken major, broken minor, double and inverted double. The 'major' and 'minor' versions are labels meant 
    to suggest usage. In reality any of these can be configured with any pen with one exception. The double line consists of two lines drawn with the major pen 
    above and the minor pen below. The inverted double reverses these: minor over major. The line spacing is also given in 'major' and 'minor' versions. The 
    difference between these is that the minor space is used to separate the double lines and the 'S' shortcut (see below) automatically generates one major 
    space.</p>
    <p>As with <code>MusicPanel</code> this class uses a String pattern to lay out the lines. All lines are horizontal and all stretch from margin to margin. The 
    available options are:
    <ul>
      <li>'L' - Major solid line.</li>
      <li>'M' - Minor solid line.</li>
      <li>'B' - Major broken line.</li>
      <li>'T' - Minor broken line.</li>
      <li>'D' - Double line.</li>
      <li>'U' - Inverted double line.</li>
      <li>'I' - Invisible line used to enforce trailing space.</li>
      <li>'!' - STOP the pattern at the point it occurs, regardless of remaining characters or space on the page.</li>
      <li>'1' through '0' - Variable spacing: the vertical space equals one major line space times the value of the character. '0' counts as ten.</li>
      <li>'-' - Placing a dash (minus) before a number causes a multiple of <i>minor</i> line space.</li>
      <li>'.' - Placing a dot (decimal point) before a number causes its space to be fractionalized: the space is calculated per the digit and divided by ten.</li>
      <li>'S' - Shortcut for '1' (the digit one!); a single major line space.</li>
    </ul>
    Any other character stops the pattern immediately, per '!'.</p>
    <p>The layout process will try to put as many patterns on the page as it can and it ignores trailing space. In order to enforce a minimum trailing space use 
    the 'I' character. In order to prevent infinite loops the solid and broken lines do count as 0.001 major line space. The double lines automatically count as 
    a minor space.</p>
*/

public class LetteringPanel extends GraphPanel {

  private static final String DEFAULT_PATTERN = "DSSBSU!";

  private TextField txfPattern = new TextField();
  private TextField txfSpaceMajor = new TextField(), txfSpaceMinor = new TextField(), txfLineMajor = new TextField(), txfLineMinor = new TextField();
  private TextField txfBrokenMajor = new TextField(), txfBrokenMinor = new TextField(), txfDoubleMajor = new TextField(), txfDoubleMinor = new TextField();

  /** Standard constructor. */
  public LetteringPanel() {
    super();
    _setUpGUI();
    GPickle.gpgLettering = this;
  }

  /** Lay out the paper. If the pattern is null or blank nothing happens. */
  public void drawGraph(int pTop, int pBot, int pLeft, int pRight, SVGTag pPaper) {
    String sPattern = txfPattern.getText();
    if(!Question.validString(sPattern)) return;
    double xMin = pLeft, xMax = pRight, yMin = pTop, yMax = pBot;
    sPattern = sPattern.trim().toUpperCase();
    int lineMajor = getPen(txfLineMajor), lineMinor = getPen(txfLineMinor);
    int brokenMajor = getPen(txfBrokenMajor, 10), brokenMinor = getPen(txfBrokenMinor, 15);
    int doubleMajor = getPen(txfDoubleMajor, 25), doubleMinor = getPen(txfDoubleMinor);
    double spaceMajor = asDouble(txfSpaceMajor.getText(), 10.0, 0.1, null), spaceMinor = asDouble(txfSpaceMinor.getText(), 2.0, 0.1, null);
    double dWidth = xMax - xMin, micron = spaceMajor * 0.001;
    MathDrawSVG MDraw = cox.jmatt.java.MathTools.demo.GPGenerator.getMDraw();
    UseTag utLine = null;
    TransformPen tPen = null;
    // Create the double-line group.
    String dLine = "DoubleLine_"+CapCom.hitCount(), uLine = "UDLine_"+CapCom.hitCount();
    GroupTag gt = MDraw.getGroupTag(dLine);
    dLine = "#" + dLine;
    MDraw.setLinePen(doubleMajor);
    gt.drawLine(0d, 0d, dWidth, 0d);
    MDraw.setLinePen(doubleMinor);
    gt.drawLine(0d, spaceMinor, dWidth, spaceMinor);
    pPaper.add(gt);
    gt = MDraw.getGroupTag(uLine);
    uLine = "#" + uLine;
    gt.drawLine(0d, 0d, dWidth, 0d);
    MDraw.setLinePen(doubleMajor);
    gt.drawLine(0d, spaceMinor, dWidth, spaceMinor);
    pPaper.add(gt);

    // Calculate pattern width.
    double patternSize = 0, rawPatternSize = 0, dTemp = 0;
    boolean isMinor = false, isDecimal = false;
    char ch = 'z';
    for(int i=0; i < sPattern.length(); i++) {
      ch = sPattern.charAt(i);
      switch(ch) {
        case '-': isMinor = true; break;
        case '.': isDecimal = true; break;
        case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
          if(isMinor) {
            dTemp = spaceMinor * chInt(ch);
            isMinor = false;
          } else {
            dTemp = spaceMajor * chInt(ch);
          }
          if(isDecimal) {
            dTemp *= ((double) chInt(ch) / 10d);
            isDecimal = false;
          }
          rawPatternSize += dTemp;
          break;
        case 'L': case 'M': case 'B': case 'T': case 'I':
          rawPatternSize += micron;
          patternSize = rawPatternSize;
          break;
        case 'D': case 'U':
          rawPatternSize += spaceMinor;
          patternSize = rawPatternSize;
          break;
        case 'S':
          rawPatternSize += spaceMajor;
        default:
      }
    }
    CapCom.debug("Pattern size: "+patternSize+"\tRaw size: "+rawPatternSize);

    double vPos = yMin;
    boolean trucking = true;
    isMinor = false; isDecimal = false;
    while(trucking && ((vPos + patternSize) < yMax)) {
      for(int i=0; i < sPattern.length(); i++) {
        ch = sPattern.charAt(i);
        switch(ch) {
          case '-': isMinor = true; break;
          case '.': isDecimal = true; break;
          case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            if(isMinor) {
              dTemp = spaceMinor * chInt(ch);
              isMinor = false;
            } else {
              dTemp = spaceMajor * chInt(ch);
            }
            if(isDecimal) {
              dTemp *= ((double) chInt(ch) / 10d);
              isDecimal = false;
            }
            vPos += dTemp;
            break;
          case 'S':
            vPos += spaceMajor;
            break;
          case 'L':
            MDraw.setLinePen(lineMajor);
            pPaper.drawLine(xMin, vPos, xMax, vPos);
            vPos += micron;
            break;
          case 'M':
            MDraw.setLinePen(lineMinor);
            pPaper.drawLine(xMin, vPos, xMax, vPos);
            vPos += micron;
            break;
          case 'B':
            MDraw.setLinePen(brokenMajor);
            pPaper.drawLine(xMin, vPos, xMax, vPos);
            vPos += micron;
            break;
          case 'T':
            MDraw.setLinePen(brokenMinor);
            pPaper.drawLine(xMin, vPos, xMax, vPos);
            vPos += micron;
            break;
          case 'I':
            vPos += micron;
            break;
          case 'D':
            utLine = MDraw.getUseTag(dLine);
            tPen = MDraw.getTransformPen();
            tPen.translate(xMin, vPos);
            utLine.setTransform(tPen);
            pPaper.add(utLine);
            vPos += spaceMinor;
            break;
          case 'U':
            utLine = MDraw.getUseTag(uLine);
            tPen = MDraw.getTransformPen();
            tPen.translate(xMin, vPos);
            utLine.setTransform(tPen);
            pPaper.add(utLine);
            vPos += spaceMinor;
            break;
          default:
            trucking = false;
        } // End switch
      } // End for
    } // End while
    
  }

  private int chInt(char ch) {
    int retval = 0;
    try {
      retval = Integer.parseInt(""+ch);
      if(retval == 0) retval = 10;
    } catch(Exception ignored) {}
    return retval;
  }

  /** Restore factory defaults. */
  public void reset() {
    setPattern(DEFAULT_PATTERN);
    setSpace(10, 5);
    setSolid(0, 0);
    setBroken(10, 15);
    setDouble(25, 0);
  }

  /** Set the line layout pattern. Null or blank values are ignored.
      @param pPattern The new layout pattern.
  */
  public void setPattern(String pPattern) {
    if(Question.validString(pPattern)) txfPattern.setText(pPattern);
  }

  /** Set the major and minor (vertical) line spacing. Negative or zero values are ignored. */
  public void setSpace(double pMajor, double pMinor) {
    if(pMajor > -1) txfSpaceMajor.setText("" + pMajor);
    if(pMinor > -1) txfSpaceMinor.setText("" + pMinor);
  }

  /** Set the major and minor solid line pens. Negative values are ignored. */
  public void setSolid(int pMajor, int pMinor) {
    if(pMajor > -1) txfLineMajor.setText("" + pMajor);
    if(pMinor > -1) txfLineMinor.setText("" + pMinor);
  }

  /** Set the major and minor broken (dashed) line pens. Negative values are ignored. */
  public void setBroken(int pMajor, int pMinor) {
    if(pMajor > -1) txfBrokenMajor.setText("" + pMajor);
    if(pMinor > -1) txfBrokenMinor.setText("" + pMinor);
  }

  /** Set the major and minor double line pens. Negative values are ignored. */
  public void setDouble(int pMajor, int pMinor) {
    if(pMajor > -1) txfDoubleMajor.setText("" + pMajor);
    if(pMinor > -1) txfDoubleMinor.setText("" + pMinor);
  }

  private void _setUpGUI() {
    setLayout(new GridLayout(0,3));
    add(new Label("Line pattern")); add(txfPattern); add(new Label());
    add(new Label("")); add(new Label("Major", Label.CENTER)); add(new Label("Minor", Label.CENTER));
    add(new Label("Line space:")); add(txfSpaceMajor); add(txfSpaceMinor);
    add(new Label("Solid line pen:")); add(txfLineMajor); add(txfLineMinor);
    add(new Label("Broken line pen:")); add(txfBrokenMajor); add(txfBrokenMinor);
    add(new Label("Double line pens:")); add(txfDoubleMajor); add(txfDoubleMinor);
    reset();
  }

  public String toPizza(String pPrefix) {
    StringBuffer retval = new StringBuffer("# LetteringPanel configuration data.");
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
      retval.append(" Prefix: ").append(pPrefix);
    } else {
      pPrefix = "";
    }
    retval.append("\n");
    retval.append(getKey(txfPattern, pPrefix, "letters.pattern"));
    retval.append(getKey(txfSpaceMajor, pPrefix, "letters.spacing.major"));
    retval.append(getKey(txfLineMajor, pPrefix, "letters.solid.major"));
    retval.append(getKey(txfBrokenMajor, pPrefix, "letters.broken.major"));
    retval.append(getKey(txfDoubleMajor, pPrefix, "letters.double.major"));
    retval.append(getKey(txfSpaceMinor, pPrefix, "letters.spacing.minor"));
    retval.append(getKey(txfLineMinor, pPrefix, "letters.solid.minor"));
    retval.append(getKey(txfBrokenMinor, pPrefix, "letters.broken.minor"));
    retval.append(getKey(txfDoubleMinor, pPrefix, "letters.double.minor"));
    
    return retval.toString();
  }

  public void eatPizza(Properties pPizza, String pPrefix) {
    if((pPizza == null) || pPizza.isEmpty()) return;
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
    } else {
      pPrefix = "";
    }
    setKey(pPizza, txfPattern, pPrefix, "letters.pattern", DEFAULT_PATTERN);
    setKey(pPizza, txfSpaceMajor, pPrefix, "letters.spacing.major", "10");
    setKey(pPizza, txfLineMajor, pPrefix, "letters.solid.major", "0");
    setKey(pPizza, txfBrokenMajor, pPrefix, "letters.broken.major", "10");
    setKey(pPizza, txfDoubleMajor, pPrefix, "letters.double.major", "25");
    setKey(pPizza, txfSpaceMinor, pPrefix, "letters.spacing.minor", "5");
    setKey(pPizza, txfLineMinor, pPrefix, "letters.solid.minor", "0");
    setKey(pPizza, txfBrokenMinor, pPrefix, "letters.broken.minor", "15");
    setKey(pPizza, txfDoubleMinor, pPrefix, "letters.double.minor", "0"); 
  }

}

