package cox.jmatt.java.MathTools.demo.gpgdemo;

import java.awt.*;
import java.awt.event.*;
import java.util.Properties;
import java.io.*;

import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.markup.svg.*;

/** <p>This generator creates blank music paper. There are multiple layout options, most of which involve spacing. The line space, between the lines of the staff, 
    forms the basis for all the other measurements. The basic unit of measurement is the <i>staff</i>, which is six times the line space. They layout is controlled 
    by the <i>staff pattern</i> which is a String of special characters. Unless specifically stopped the pattern will be repeated as many times as its visible 
    parts will fit between the top and bottom margins.</p>
    <p>The visible items that can be placed are musical staves, rules and lines. The musical staff consists of five evenly-spaced horizontal lines capped at the 
    left side with a vertical line. Multiple staves can be joined with a vertical line and it is also possible to place independent horizontal lines (<i>rules</i>). 
    Each of these items has its own pen: line pen, vertical pen, connector pen and rule pen. There are two types of rule(r)s. The full rule stretches from margin 
    to margin and the half-rule stretches half that distance, centered. (N.B. The rules are not specifically a part of musical notation but the page may need 
    one.)</p>
    <h3>Following the Pattern</h3>
    <p>The staff pattern is divided into visible components and spacing components. In order to calculate how many patterns will fit on the page the height of 
    the pattern is calculated from the first visible component to the last, not including vertical lines. (If a vertical line is opened and not closed it will 
    not appear at all!) When the pattern is actually put onto the paper any trailing spaces do appear. The pattern formatting characters are:
    <ul>
      <li>'S' specifies the placement of a musical staff.</li>
      <li>'X' specifies a vertical space equal in height to one staff.</li>
      <li>'D' specifies a vertical space two times the height of a staff.</li>
      <li>'H' specifies a half-staff vertical space.</li>
      <li>'Q' specifies a quarter-staff vertical space.</li>
      <li>'E' specifies an eighth of a staff vertical space.</li>
      <li>'T' specifies a vertical space equal to one line space.</li>
      <li>'L' specifies a full (margin-to-margin) horizontal rule.</li>
      <li>'R' specifies a half-rule: half the page width and centered.</li>
      <li>'I' specifies an <i>invisible line</i>. It does not appear on the page but it <i>does</i> count as a visible item; it can be used to mandate some 
        trailing space in the pattern size calculation.</li>
      <li>'!' <b>stops</b> formatting immediately. Anything previously-set will be rendered but it will not be repeated and anything afterward is ignored.</li>
    </ul>
    In addition to these characters the vertical pipe character ('|') serves to group staves and possibly spaces. These must appear in pairs: the first pipe 
    indicates the starting point of a vertical line and the second its ending point. The line characters ('L', 'R' and 'I') do not count against spacing but in 
    actuality each does incur a very small (1/1000th of a line space) vertical distance. This was done to prevent patterns like 'ISSI' from causing infinite 
    loops.</p>
    <p>Characters other than the ones listed above cause an immediate stop (per '!').</p>
*/

public class MusicPanel extends GraphPanel implements ActionListener {

  private TextField txfLineSpace = new TextField(), txfLinePen = new TextField(), txfConnPen = new TextField(), txfVertPen = new TextField();
  private TextField txfRulePen = new TextField(), txfPattern = new TextField();

  /** The usual constructor. */
  public MusicPanel() {
    super();
    _setUpGUI();
    GPickle.gpgMusic = this;
  }

  /** Draw the paper. If the staff pattern is null or blank NOTHING happens! */
  public void drawGraph(int pTop, int pBot, int pLeft, int pRight, SVGTag pPaper) {
    String sPattern = txfPattern.getText();
    if(!Question.validString(sPattern)) return;
    String staffID = "Staff_"+CapCom.hitCount(), lineID = "Rule_"+CapCom.hitCount(), ruleID = "HRule_"+CapCom.hitCount();
    sPattern = sPattern.trim().toUpperCase();
    double xMin = pLeft, xMax = pRight, yMin = pTop, yMax = pBot, lineSpace = asDouble(txfLineSpace.getText(), 0.0, 0.0, null);
    int linePen = getPen(txfLinePen), connPen = getPen(txfConnPen), rulePen = getPen(txfRulePen, 15), vertPen = getPen(txfVertPen);
    int spLen = sPattern.length();
    double dWidth = 0.98*(xMax - xMin), staffSize = 6 * lineSpace, yPos = yMin + lineSpace, micron = 0.001*dWidth;
    xMin += 0.01*dWidth; yMin += (0.01*(yMax - yMin)); // Adjust to JUST INSIDE the margins.
    MathDrawSVG MDraw = cox.jmatt.java.MathTools.demo.GPGenerator.getMDraw();
    UseTag utItem = null;
    
    // Create ONE staff and re-use it.
    GroupTag gtStaff = MDraw.getGroupTag(staffID);
    MDraw.setPathPen(linePen);
    PathTag ptLines = MDraw.getPathTag();
    ptLines.MoveTo(0, lineSpace);
    for(int i=0, iFlip = 1; i < 4; i++) {
      ptLines.hLineTo(iFlip * dWidth);
      ptLines.moveTo(0, lineSpace);
      iFlip *= -1;
    }
    ptLines.hLineTo(dWidth);
    gtStaff.add(ptLines);
    MDraw.setLinePen(vertPen);
    gtStaff.drawLine(0d, lineSpace, 0d, 5*lineSpace);
    pPaper.add(gtStaff);
    /* * * Use these two for adding staves * * */
    TransformPen tPen = null;
    staffID = "#" + staffID;
    // Create a master Line and Rule tag.
    MDraw.setPathPen(rulePen);
    gtStaff = MDraw.getGroupTag(lineID);
    ptLines = MDraw.getPathTag(); ptLines.MoveTo(0, 0); ptLines.hLineTo(dWidth);
    gtStaff.add(ptLines);
    pPaper.add(gtStaff);
    lineID = "#"+lineID;
    gtStaff = MDraw.getGroupTag(ruleID);
    ruleID = "#"+ruleID;
    ptLines = MDraw.getPathTag(); ptLines.MoveTo(dWidth/4,0); ptLines.hLineTo(dWidth/2);
    gtStaff.add(ptLines);
    pPaper.add(gtStaff);

    // Calculate the staff size.
    double patternHeight = 0, rawPatternHeight = 0;
    for(char ch : sPattern.toCharArray()) {
      switch(ch) {
        case 'S':
          rawPatternHeight += staffSize;
          patternHeight = rawPatternHeight;
          break;
        case 'L':
        case 'I':
        case 'R':
          rawPatternHeight += micron;
          patternHeight = rawPatternHeight;
          break;
        case 'D':
          rawPatternHeight += staffSize;
        case 'X':
          rawPatternHeight += staffSize;
          break;
        case 'H':
          rawPatternHeight += staffSize/4;
        case 'Q':
          rawPatternHeight += staffSize/4;
          break;
        case 'E':
          rawPatternHeight += staffSize/8;
          break;
        case 'T':
          rawPatternHeight += lineSpace;
        default:
      }
    }
    CapCom.debug("Staff height: "+staffSize+"\tPattern height: "+patternHeight+"\tRaw height: "+rawPatternHeight);

    boolean trucking = true;
    char ch;
    double connStart = -1;
    while(trucking && (yPos + patternHeight) <= yMax) {
      for(int iPos = 0; trucking && (iPos < sPattern.length()); iPos++) {
        ch = sPattern.charAt(iPos);
        switch(ch) {
          case 'S':
            utItem = MDraw.getUseTag(staffID);
            tPen = MDraw.getTransformPen().translate(xMin, yPos);
            utItem.setTransform(tPen);
            pPaper.add(utItem);
          case 'X':
            yPos += staffSize;
            break;
          case 'L':
            utItem = MDraw.getUseTag(lineID);
            tPen = MDraw.getTransformPen().translate(xMin, yPos);
            utItem.setTransform(tPen);
            pPaper.add(utItem);
            yPos += micron;
            break;
          case 'R':
            utItem = MDraw.getUseTag(ruleID);
            tPen = MDraw.getTransformPen().translate(xMin, yPos);
            utItem.setTransform(tPen);
            pPaper.add(utItem);
          case 'I':
            yPos += micron;
            break;
          case 'D':
            yPos += staffSize;
          case 'H':
            yPos += staffSize/4;
          case 'Q':
            yPos += staffSize/4;
            break;
          case 'E':
            yPos += staffSize/8;
            break;
          case 'T':
            rawPatternHeight += lineSpace;
            break;
          case '|':
            if(connStart < 0) {
              connStart = yPos;
            } else {
              pPaper.drawLine(xMin, connStart + lineSpace, xMin, yPos - lineSpace);
              connStart = -1;
            }
            break;
          default:
            trucking = false;
        }
      }
    }
    
  }

  /** Reset to original values. */
  public void reset() {
    setPattern("|SES|Q");
    setLineSpace(10);
    setPens(0, 15, 0, 25);
  }

  /** Set the staff pattern. Null or blank values are silently ignored but no other checking is done.
      @param pPattern The new staff pattern.
  */
  public void setPattern(String pPattern) {
    if(Question.validString(pPattern)) txfPattern.setText(pPattern);
  }

  /** Set the line spacing. Any value less than one is silently ignored. */
  public void setLineSpace(int pSpace) {
    if(pSpace > 0) txfLineSpace.setText(""+pSpace);
  }

  /** Set the pens to use. Any negative values are ignored: that pen will not be set.
      @param pLine The horizontal line pen.
      @param pRule The Rule pen.
      @param pVertical The vertical non-connector pen.
      @param pConnector The staff connection pen.
  */
  public void setPens(int pLine, int pRule, int pVertical, int pConnector) {
    if(pLine >-1) txfLinePen.setText(""+pLine);
    if(pRule >-1) txfRulePen.setText(""+pRule);
    if(pVertical >-1) txfVertPen.setText(""+pVertical);
    if(pConnector >-1) txfConnPen.setText(""+pConnector);
  }

  public String toPizza(String pPrefix) {
    StringBuffer retval = new StringBuffer("# MusicPanel configuration data.");
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
      retval.append(" Prefix: ").append(pPrefix);
    } else {
      pPrefix = "";
    }
    retval.append("\n");
    retval.append(getKey(txfPattern, pPrefix, "music.pattern"));
    retval.append(getKey(txfLineSpace, pPrefix, "music.lineSpace"));
    retval.append(getKey(txfLinePen, pPrefix, "music.pen.line"));
    retval.append(getKey(txfRulePen, pPrefix, "music.pen.rule"));
    retval.append(getKey(txfConnPen, pPrefix, "music.pen.connector"));
    retval.append(getKey(txfVertPen, pPrefix, "music.pen.vertical"));
    return retval.toString();
  }

  public void eatPizza(Properties pPizza, String pPrefix) {
    if((pPizza == null) || pPizza.isEmpty()) return;
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
    } else {
      pPrefix = "";
    }
    setKey(pPizza, txfPattern, pPrefix, "music.pattern", "|SXS|X");
    setKey(pPizza, txfLineSpace, pPrefix, "music.lineSpace", "25");
    setKey(pPizza, txfLinePen, pPrefix, "music.pen.line", "0");
    setKey(pPizza, txfRulePen, pPrefix, "music.pen.rule", "15");
    setKey(pPizza, txfConnPen, pPrefix, "music.pen.connector", "0");
    setKey(pPizza, txfVertPen, pPrefix, "music.pen.vertical", "25");
  }

  private void _setUpGUI() {
    setLayout(new GridLayout(0,2));
    add(new Label("Line spacing:")); add(txfLineSpace);
    add(new Label("Staff line pen:")); add(txfLinePen);
    add(new Label("Rule pen:")); add(txfRulePen);
    add(new Label("Staff connector pen:")); add(txfConnPen);
    add(new Label("Vertical line pen:")); add(txfVertPen);
    add(new Label("Staff pattern:")); add(txfPattern);
    Button btnTemp = new Button("Show Full Instructions"); btnTemp.addActionListener(this);
    add(new Label("")); add(btnTemp);
    reset();
  }

  /** All this does is handle the 'Help' button. */
  public void actionPerformed(ActionEvent aev) { showHelp(); }

  private void showHelp() {
    CapCom.println("- - - - - MusicPanel Pattern Characters and Notes - - - - -");
    CapCom.println("S - Draw a musical staff per the settings.\nX - Leave a vertical space equal to a full staff.\nH - Leave a vertical space equal to half a staff.");
    CapCom.println("Q - Leave a vertical space one quarter of a staff.\nE - Vertical space of one-eighth of a staff.\nT - Vertical space of one line-space.");
    CapCom.println("D - Leave a vertical space equal to two staves.\n| (Vertical pipe) - Toggle the staff connector line on or off.");
    CapCom.println("L - Draw a margin-to-margin non-spacing line with the Rule pen.\nR - Draw a line equal to half the width using the Rule pen.");
    CapCom.println("I - Invisible line, used for forcing trailing space.");
    CapCom.println("! - STOP!! Do not repeat the pattern.\n\nThe staves drawn stretch from the left margin to the right.");
    CapCom.println("The vertical line pen is for the initial line at the left of the staff.\nThe connector line connects multiple staves (or spaces).");
    CapCom.println("Staves include a line space above and below the top and bottom lines.");
    CapCom.println("(See the javadoc for full details.)");
  }

}

