package cox.jmatt.java.MathTools.demo.gpgdemo;

import java.awt.*;
import java.util.Properties;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.demo.*;
import cox.jmatt.java.MathTools.markup.svg.*;

/** <p>This generator produces a one-point square-perspective grid. Both scales are configurable but the perspective operates differently on each. In all cases 
    the <b>near</b> values refer to the bottom of the page; the <i>nearest</i> to the viewer. The <b>far</b> value represents the top of the page which is the 
    <i>farthest</i> from the viewer.</p>
    <p><b>NOTE:</b> In the configuration settings the <i>vertical</i> scale uses the <i>horizontal</i> line pen and vise versa. That is the vertical <i>scale</i> 
    uses horizontal <i>lines</i> and the horizontal <i>scale</i> uses vertical <i>lines</i>.</p>
    <h3>Configuration</h3>
    <p>The horizontal scale starts with two separate line spacings. The near spacing is the spacing between lines at the bottom of the page. This spacing remains 
    constant throughout the drawing process. The far spacing applies to the top of the page and it changes as the lines move outward from the centerline. With 
    each iteration the horizontal space is multiplied by the horizontal scale factor, presumably to decrease it. If the horizontal scale is set to 'Linear' then 
    both the top and bottom spacing is horizontal near.</p>
    <p>The vertical scale begins with vertical near spacing and with each subsequent line this is multiplied by the vertical scale factor so it shrinks as it 
    moves from bottom to top. If the vertical scale is 'Linear' then the spacing remains constant.</p>
    <p>If the horizontal scale is set to 'Perspective' ('Linear' box is unchecked) the vertical '[]Clip to H. scale' Checkbox comes into play. If checked 
    (default state) then the horizontal lines are all clipped to the outermost vertical lines. If not they extend margin to margin.</p>
*/
public class PerspectivePanel extends GraphPanel {

  private static final String H_NEAR = "25", H_FAR = "20", V_NEAR = "25", H_SCALE = "0.98", V_SCALE = "0.985";

  private TextField txfHLine = new TextField(), txfVLine = new TextField(), txfNearVSpace = new TextField();
  private TextField txfNearHSpace = new TextField(), txfFarHSpace = new TextField(), txfHScale = new TextField(), txfVScale = new TextField();
  private Checkbox cbxHPersp = new Checkbox("Linear", false), cbxVPersp = new Checkbox("Linear", true), cbxClipScale = new Checkbox("Clip to H. scale.", true);
  private Checkbox cbxShowHoriz = new Checkbox("Horizontal", true), cbxShowVert = new Checkbox("Vertical", true);

  /** Usual constructor. */
  public PerspectivePanel() {
    super();
    _setUpGUI();
    GPickle.gpgPerspective = this;
  }

  /** Set the horizontal and vertical line pens. Any value less than zero is ignored. */
  public void setPens(int pHorizontal, int pVertical) {
    if(pHorizontal > -1) txfHLine.setText(""+pHorizontal);
    if(pVertical > -1) txfVLine.setText(""+pVertical);
  }

  /** Set the spacing values for the grid. Any value less than 1 is ignored. */
  public void setSpacing(int pHorizNear, int pHorizFar, int pVertNear) {
    if(pHorizNear > 0) txfNearHSpace.setText(""+pHorizNear);
    if(pHorizFar > 0) txfFarHSpace.setText(""+pHorizFar);
    if(pVertNear > 0) txfNearVSpace.setText(""+pVertNear);
  }

  /** Set the horizontal and vertical scales. Any value less than 0.1 is ignored. */
  public void setScale(double pHoriz, double pVert) {
    if(pHoriz >= 0.1) txfHScale.setText(""+pHoriz);
    if(pVert >= 0.1) txfVScale.setText(""+pVert);
  }

  /** Configure one or both scales as linear. */
  public void setLinearScale(boolean pHoriz, boolean pVert) {
    cbxHPersp.setState(pHoriz); cbxVPersp.setState(pVert);
  }

  /** Enable/disable horizontal line clipping to outermost verticals in horizontal perspective mode. */
  public void setClipToScale(boolean pEnable) {
    cbxClipScale.setState(pEnable);
  }

  /** Draw the paper. */
  public void drawGraph(int pTop, int pBot, int pLeft, int pRight, SVGTag pPaper) {
    double yMin = pTop, xMin = pLeft, yMax = pBot, xMax = pRight, xMid = (xMin + xMax)/2.0;
    int vLinePen = getPen(txfVLine), hLinePen = getPen(txfHLine);
    double hNearSpace = asDouble(txfNearHSpace.getText(), 25.0, 1.0, null), hFarSpace = asDouble(txfFarHSpace.getText(), 25.0, 1.0, null);
    double vNearSpace = asDouble(txfNearVSpace.getText(), 25.0, 1.0, null);
    double vScale = asDouble(txfVScale.getText(), 0.01, 0.001, null), hScale = asDouble(txfHScale.getText(), 0.01, 0.001, null);
    MathDrawSVG MDraw = cox.jmatt.java.MathTools.demo.GPGenerator.getMDraw();
    String clipID = "ClipBox_"+CapCom.hitCount(), graphID = "Graph_"+CapCom.hitCount();
    ClipTag gtClipBox = MDraw.getClipTag(clipID);
    GroupTag gPaper = MDraw.getGroupTag(graphID);
    PathTag ptClip = MDraw.getPathTag();
    ptClip.MoveTo(xMin-1, yMin-1); ptClip.HLineTo(xMax+1); ptClip.VLineTo(yMax+1); ptClip.HLineTo(xMin-1); ptClip.VLineTo(yMin-1); ptClip.closePath();

    if(cbxShowHoriz.getState()) {
      MDraw.setLinePen(vLinePen);
      if(cbxHPersp.getState()) {
        // Horizontal scale is LINEAR!!
        double hlPos = xMid - hNearSpace, hrPos = xMid + hNearSpace;
        gPaper.drawLine(xMid, yMin, xMid, yMax);
        while((hlPos >= xMin) && (hrPos <= xMax)) {
          gPaper.drawLine(hlPos, yMin, hlPos, yMax);
          gPaper.drawLine(hrPos, yMin, hrPos, yMax);
          hlPos -= hNearSpace; hrPos += hNearSpace;
        }
      } else {
        // Horizontal scale is PERSPECTIVE!!
        gPaper.drawLine(xMid, yMin, xMid, yMax);
        double hlNear = xMid - hNearSpace, hrNear = xMid + hNearSpace, hlFar = xMid - hFarSpace, hrFar = xMid + hFarSpace;
        double hlnClip = 0, hrnClip = 0, hlfClip = 0, hrfClip = 0;
        while((hlNear >= xMin) && (hrNear <= xMax)) {
          gPaper.drawLine(hlFar, yMin, hlNear, yMax);
          gPaper.drawLine(hrFar, yMin, hrNear, yMax);
          hlnClip = hlNear - 1; hrnClip = hrNear + 1; hlfClip = hlFar - 1; hrfClip = hrFar + 1;
          hFarSpace *= hScale;
          hFarSpace = Math.max(hFarSpace, 1.0);
          hlFar -= hFarSpace; hrFar += hFarSpace;
          hlNear -= hNearSpace; hrNear += hNearSpace;
        }
        if(cbxClipScale.getState()) {
          ptClip.reset();
          ptClip.MoveTo(hlfClip, yMin-1); ptClip.HLineTo(hrfClip); ptClip.LineTo(hrnClip, yMax+1); ptClip.HLineTo(hlnClip); ptClip.LineTo(hlfClip, yMin-1);
          ptClip.closePath();
        }
      }
    }

    if(cbxShowVert.getState()) {
      MDraw.setLinePen(hLinePen);
      if(cbxVPersp.getState()) {
        // Vertical scale is LINEAR!!
        for(double vPos = yMax; vPos >= yMin; vPos -= vNearSpace) gPaper.drawLine(xMin, vPos, xMax, vPos);
      } else {
        // Vertical scale is PERSPECTIVE!!
        double vPos = yMax;
        while(vPos >= yMin) {
          gPaper.drawLine(xMin, vPos, xMax, vPos);
          vNearSpace *= vScale;
          vNearSpace = Math.max(vNearSpace, 1.0);
          vPos -= vNearSpace;
        }
      }
    }

    gtClipBox.add(ptClip);
    clipID = "#"+clipID;
    graphID = "#"+graphID;
    UseTag utGraph = MDraw.getUseTag(graphID);
    StylePen sPen = MDraw.getStylePen();
    sPen.setClipURL(clipID);
    utGraph.setStyle(sPen);

    pPaper.add(gtClipBox);
    pPaper.add(gPaper);
    pPaper.add(utGraph);

  }

  /** Restore factory values. */
  public void reset() {
    txfHLine.setText("0"); txfVLine.setText("0");
    txfNearHSpace.setText(H_NEAR); txfFarHSpace.setText(H_FAR);
    txfNearVSpace.setText(V_NEAR);
    txfHScale.setText(H_SCALE); txfVScale.setText(V_SCALE);
    cbxHPersp.setState(false); cbxVPersp.setState(false);
    cbxShowHoriz.setState(true); cbxShowVert.setState(true);
    cbxClipScale.setState(true);
  }

  private void _setUpGUI() {
    setLayout(new GridLayout(0, 3));
    add(new Label("")); add(cbxShowHoriz); add(cbxShowVert);
    add(new Label("Line pen:")); add(txfHLine); add(txfVLine);
    add(new Label("Scale type:")); add(cbxHPersp); add(cbxVPersp);
    add(new Label("Near spacing:")); add(txfNearHSpace); add(txfNearVSpace);
    add(new Label("Far spacing:")); add(txfFarHSpace); add(cbxClipScale);
    add(new Label("Scale factor:")); add(txfHScale); add(txfVScale);
    add(new Label("")); add(new Label("")); add(new Label(""));
    reset();
  }

  public String toPizza(String pPrefix) {
    StringBuffer retval = new StringBuffer("# PerspectivePanel configuration data.");
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
      retval.append(" Prefix: ").append(pPrefix);
    } else {
      pPrefix = "";
    }
    retval.append("\n");
    retval.append(getKey(cbxShowHoriz, pPrefix, "perspective.enable.horizontal"));
    retval.append(getKey(cbxShowVert, pPrefix, "perspective.enable.vertical"));
    retval.append(getKey(txfHLine, pPrefix, "perspective.pen.horizontal"));
    retval.append(getKey(txfVLine, pPrefix, "perspective.pen.vertical"));
    retval.append(getKey(txfNearHSpace, pPrefix, "perspective.near.horizontal"));
    retval.append(getKey(txfNearVSpace, pPrefix, "perspective.near.vertical"));
    retval.append(getKey(txfFarHSpace, pPrefix, "perspective.far.horizontal"));
    retval.append(getKey(txfHScale, pPrefix, "perspective.scale.horizontal"));
    retval.append(getKey(txfVScale, pPrefix, "perspective.scale.vertical"));
    retval.append(getKey(cbxHPersp, pPrefix, "perspective.linear.horizontal"));
    retval.append(getKey(cbxVPersp, pPrefix, "perspective.linear.vertical"));
    retval.append(getKey(cbxClipScale, pPrefix, "perspective.vertical.clipToScale"));
    return retval.toString();
  }

  public void eatPizza(Properties pPizza, String pPrefix) {
    if((pPizza == null) || pPizza.isEmpty()) return;
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
    } else {
      pPrefix = "";
    }
    setKey(pPizza, cbxShowHoriz, pPrefix, "perspective.enable.horizontal");
    setKey(pPizza, cbxShowVert, pPrefix, "perspective.enable.vertical");
    setKey(pPizza, txfHLine, pPrefix, "perspective.pen.horizontal", "0");
    setKey(pPizza, txfVLine, pPrefix, "perspective.pen.vertical", "0");
    setKey(pPizza, txfNearHSpace, pPrefix, "perspective.near.horizontal", H_NEAR);
    setKey(pPizza, txfNearVSpace, pPrefix, "perspective.near.vertical", V_NEAR);
    setKey(pPizza, txfFarHSpace, pPrefix, "perspective.far.horizontal", H_FAR);
    setKey(pPizza, txfHScale, pPrefix, "perspective.scale.horizontal", H_SCALE);
    setKey(pPizza, txfVScale, pPrefix, "perspective.scale.vertical", V_SCALE);
    setKey(pPizza, cbxHPersp, pPrefix, "perspective.linear.horizontal");
    setKey(pPizza, cbxVPersp, pPrefix, "perspective.linear.vertical");
    setKey(pPizza, cbxClipScale, pPrefix, "perspective.vertical.clipToScale");
  }

}


