package cox.jmatt.java.MathTools.demo.gpgdemo;

import java.awt.*;
import java.util.Properties;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.demo.*;
import cox.jmatt.java.MathTools.markup.svg.*;

/** <p>This generator produces graph paper similar to log paper. The intervals begin at the axis (horizontal or vertical) and progressively shrink (or grow) as 
    they move away from it. The settings that control this are:
    <ul>
      <li><b>Initial length</b> - The initial length of each interval.</li>
      <li><b>Minimum length</b> - The shortest allowable interval length; they will not shrink past this.</li>
      <li><b>Scale factor</b> - The value multiplied by the interval length when it changes.</li>
      <li><b>Cycle length</b> - The number of intervals at the current length before scaling occurs.</li>
    </ul>
    The other two settings are for the axis pen and the line pen.</p>
    <p>Each axis can have a linear or logarithmic scale. The '[]Linear' checkbox determines how an axis is scaled. NOTE: the <i>vertical scale</i> applies to the 
    vertical axis and the <i>horizontal scale</i> applies to the horizontal axis. The pens are opposite to their scales: the vertical <i>pens</i> affect the 
    horizontal <i>lines</i> and vise versa! The horizontal and vertical components can also be disabled, if desired.</p>
*/

public class SymmetricLogPanel extends GraphPanel {

  private Checkbox cbxVertLinear = new Checkbox("Linear", true), cbxHorizLinear = new Checkbox("Linear", false);
  private Checkbox cbxShowVert = new Checkbox("Vertical", true), cbxShowHoriz = new Checkbox("Horizontal", true);
  private TextField txfVertSpace = new TextField(), txfHorizSpace = new TextField(), txfVertScale = new TextField(), txfHorizScale = new TextField();
  private TextField txfVertMinSpace = new TextField(), txfHorizMinSpace = new TextField(), txfHorizCycle = new TextField(), txfVertCycle = new TextField();
  private TextField txfHAxisPen = new TextField(), txfVAxisPen = new TextField(), txfVLinePen = new TextField(), txfHLinePen = new TextField();

  /** Usual constructor. */
  public SymmetricLogPanel() {
    super();
    _setUpGUI();
    GPickle.gpgSymmetricLog = this;
  }

  /** Draw the thing. */
  public void drawGraph(int pTop, int pBot, int pLeft, int pRight, SVGTag pPaper) {
    double yMin = pTop, xMin = pLeft, yMax = pBot, xMax = pRight;

    double xMid = (pLeft + pRight)/2, yMid = (pTop + pBot)/2, delta = asDouble(txfVertSpace.getText(), 25d, 1d, null), scale = asDouble(txfVertScale.getText(), 0.9, 0.1, null);
    double minSize = asDouble(txfVertMinSpace.getText(), 5d, 0.1, null);
    int axisPen = getPen(txfHAxisPen), linePen = getPen(txfHLinePen);
    MathDrawSVG MDraw = cox.jmatt.java.MathTools.demo.GPGenerator.getMDraw();
    int cycleCt = 1, cycleMax = asInt(txfVertCycle.getText(), 5, 1, null);
    double curLow = yMid - delta, curHigh = yMid + delta;
    if(cbxShowVert.getState()) {
      // Show the vertical scale
      MDraw.setLinePen(axisPen);
      pPaper.drawLine(xMin, yMid, xMax, yMid);
      MDraw.setLinePen(linePen);
      if(cbxVertLinear.getState()) {
        // Linear scale
        for(; (curLow > yMin) && (curHigh < yMax); curLow -= delta, curHigh += delta) {
          pPaper.drawLine(xMin, curLow, xMax, curLow);
          pPaper.drawLine(xMin, curHigh, xMax, curHigh);
        }
      } else {
        // Log scale
        for(; (curLow > yMin) && (curHigh < yMax); curLow -= delta, curHigh += delta) {
          pPaper.drawLine(xMin, curLow, xMax, curLow);
          pPaper.drawLine(xMin, curHigh, xMax, curHigh);
          if(((cycleCt % cycleMax) == 0) && (delta >= minSize)) {
            delta *= scale;
            cycleCt = 0;
          }
          cycleCt++;
        }
      }
    }

    delta = asDouble(txfHorizSpace.getText(), 25d, 1d, null); scale = asDouble(txfHorizScale.getText(), 0.9, 0.1, null);
    minSize = asDouble(txfHorizMinSpace.getText(), 5d, 0.1, null);
    axisPen = asInt(txfVAxisPen.getText(), 0, 0, null); linePen = asInt(txfVLinePen.getText(), 0, 0, null);
    cycleCt = 1; cycleMax = asInt(txfHorizCycle.getText(), 5, 1, null);
    curLow = xMid - delta; curHigh = xMid + delta;
    if(cbxShowHoriz.getState()) {
      // Show the horizontal scale
      MDraw.setLinePen(axisPen);
      pPaper.drawLine(xMid, yMin, xMid, yMax);
      MDraw.setLinePen(linePen);
      if(cbxHorizLinear.getState()) {
        // Linear scale
        for(; (curLow > xMin) && (curHigh < xMax); curLow -= delta, curHigh += delta) {
          pPaper.drawLine(curLow, yMin, curLow, yMax);
          pPaper.drawLine(curHigh, yMin, curHigh, yMax);
        }
      } else {
        // Log scale
        for(; (curLow > xMin) && (curHigh < xMax); curLow -= delta, curHigh += delta) {
          pPaper.drawLine(curLow, yMin, curLow, yMax);
          pPaper.drawLine(curHigh, yMin, curHigh, yMax);
          if(((cycleCt % cycleMax) == 0) && (delta >= minSize)) {
            delta *= scale;
            cycleCt = 0;
          }
          cycleCt++;
        }
      }
    }
    
  }

  /** Set the scale type for each axis.
      @param pVertLinear 'true' to make the vertical scale linear, 'false' for logarithmic.
      @param pHorizLinear 'true' for a linear horizontal scale, 'false' for logarithmic.
  */
  public void setScaleType(boolean pVertLinear, boolean pHorizLinear) {
    cbxVertLinear.setState(pVertLinear); cbxHorizLinear.setState(pHorizLinear);
  }

  /** Set the initial spacing for each axis. Values less than one are ignored, allowing each value to be set individually. */
  public void setSpacing(double pVertSpace, double pHorizSpace) {
    if(pVertSpace >= 1) txfVertSpace.setText(""+pVertSpace);
    if(pHorizSpace >= 1) txfHorizSpace.setText(""+pVertSpace);
  }

  /** Set the minimum interval size for each axis. Any value less than 0.1 is ignored. */
  public void setMinSpace(double pVertMin, double pHorizMin) {
    if(pVertMin >= 0.1) txfVertMinSpace.setText(""+pVertMin);
    if(pHorizMin >= 0.1) txfHorizMinSpace.setText(""+pHorizMin);
  }

  /** Set the interval scale factor. Values less than 0.1 are ignored. */
  public void setScale(double pVertScale, double pHorizScale) {
    if(pVertScale >= 0.1) txfVertScale.setText(""+pVertScale);
    if(pHorizScale >= 0.1) txfHorizScale.setText(""+pHorizScale);
  }

  /** Set the cycle length before the scale factor is applied. Values less than one are ignored. */
  public void setCycle(int pVertCycle, int pHorizCycle) {
    if(pVertCycle >= 1) txfVertCycle.setText(""+pVertCycle);
    if(pHorizCycle >= 1) txfHorizCycle.setText(""+pHorizCycle);
  }

  /** Set the pen used for the vertical and horizontal axis. Negative values are silently ignored. */
  public void setAxisPen(int pVertPen, int pHorizPen) {
    if(pVertPen > -1) txfVAxisPen.setText(""+pVertPen);
    if(pHorizPen > -1) txfHAxisPen.setText(""+pHorizPen);
  }

  /** Set the line pen for the vertical and horizontal axis. Negative values are silently ignored. */
  public void setLinePen(int pVertPen, int pHorizPen) {
    if(pVertPen > -1) txfVLinePen.setText(""+pVertPen);
    if(pHorizPen > -1) txfHLinePen.setText(""+pHorizPen);
  }

  /** Reset to starting values. */
  public void reset() {
    setScaleType(true, false);
    setSpacing(40, 25);
    setMinSpace(5, 5);
    setScale(0.8, 0.8);
    setCycle(3, 3);
    setAxisPen(0, 0);
    setLinePen(2, 2);
  }

  private void _setUpGUI() {
    setLayout(new GridLayout(0,3));
    add(new Label("")); add(cbxShowVert); add(cbxShowHoriz);
    add(new Label("Scale type:")); add(cbxVertLinear); add(cbxHorizLinear);
    add(new Label("Initial length:")); add(txfVertSpace); add(txfHorizSpace);
    add(new Label("Minimum length:")); add(txfVertMinSpace); add(txfHorizMinSpace);
    add(new Label("Scale factor:")); add(txfVertScale); add(txfHorizScale);
    add(new Label("Cycle length:")); add(txfVertCycle); add(txfHorizCycle);
    add(new Label("Axis pen:")); add(txfVAxisPen); add(txfHAxisPen);
    add(new Label("Line pen:")); add(txfVLinePen); add(txfHLinePen);
    reset();
  }

  public String toPizza(String pPrefix) {
    StringBuffer retval = new StringBuffer("# SymmetricLogPanel configuration data.");
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
      retval.append(" Prefix: ").append(pPrefix);
    } else {
      pPrefix = "";
    }
    retval.append("\n");
    retval.append(getKey(cbxShowVert, pPrefix, "symlog.vertical.enabled"));
    retval.append(getKey(cbxShowHoriz, pPrefix, "symlog.horizontal.enabled"));
    retval.append(getKey(cbxVertLinear, pPrefix, "symlog.vertical.isLinear"));
    retval.append(getKey(cbxHorizLinear, pPrefix, "symlog.horizontal.isLinear"));
    retval.append(getKey(txfVertSpace, pPrefix, "symlog.vertical.spacing"));
    retval.append(getKey(txfVertMinSpace, pPrefix, "symlog.vertical.minSpace"));
    retval.append(getKey(txfVertScale, pPrefix, "symlog.vertical.scale"));
    retval.append(getKey(txfVertCycle, pPrefix, "symlog.vertical.cycle"));
    retval.append(getKey(txfVAxisPen, pPrefix, "symlog.vertical.axisPen"));
    retval.append(getKey(txfVLinePen, pPrefix, "symlog.vertical.linePen"));
    retval.append(getKey(txfHorizSpace, pPrefix, "symlog.horizontal.spacing"));
    retval.append(getKey(txfHorizMinSpace, pPrefix, "symlog.horizontal.minSpace"));
    retval.append(getKey(txfHorizScale, pPrefix, "symlog.horizontal.scale"));
    retval.append(getKey(txfHorizCycle, pPrefix, "symlog.horizontal.cycle"));
    retval.append(getKey(txfHAxisPen, pPrefix, "symlog.horizontal.axisPen"));
    retval.append(getKey(txfHLinePen, pPrefix, "symlog.horizontal.linePen"));
    return retval.toString();
  }

  public void eatPizza(Properties pPizza, String pPrefix) {
    if((pPizza == null) || pPizza.isEmpty()) return;
    if(Question.validString(pPrefix)) {
      if(!pPrefix.endsWith(".")) pPrefix += ".";
    } else {
      pPrefix = "";
    }
    setKey(pPizza, cbxShowVert, pPrefix, "symlog.vertical.enabled");
    setKey(pPizza, cbxShowHoriz, pPrefix, "symlog.horizontal.enabled");
    setKey(pPizza, cbxVertLinear, pPrefix, "symlog.vertical.isLinear");
    setKey(pPizza, cbxHorizLinear, pPrefix, "symlog.horizontal.isLinear");
    setKey(pPizza, txfVertSpace, pPrefix, "symlog.vertical.spacing", "40");
    setKey(pPizza, txfVertMinSpace, pPrefix, "symlog.vertical.minSpace", "40");
    setKey(pPizza, txfVertScale, pPrefix, "symlog.vertical.scale", "0.8");
    setKey(pPizza, txfVertCycle, pPrefix, "symlog.vertical.cycle", "3");
    setKey(pPizza, txfVAxisPen, pPrefix, "symlog.vertical.axisPen", "0");
    setKey(pPizza, txfVLinePen, pPrefix, "symlog.vertical.linePen", "2");
    setKey(pPizza, txfHorizSpace, pPrefix, "symlog.horizontal.spacing", "40");
    setKey(pPizza, txfVertMinSpace, pPrefix, "symlog.horizontal.minSpace", "5");
    setKey(pPizza, txfHorizScale, pPrefix, "symlog.horizontal.scale", "0.8");
    setKey(pPizza, txfHorizCycle, pPrefix, "symlog.horizontal.cycle", "3");
    setKey(pPizza, txfHAxisPen, pPrefix, "symlog.horizontal.axisPen", "0");
    setKey(pPizza, txfHLinePen, pPrefix, "symlog.horizontal.linePen", "2");
  }

}

