package cox.jmatt.java.MathTools.graphing;

import java.util.Stack;
import cox.jmatt.java.MathTools.*;

/** <p>This class exists to provide assistance and convenience when specifying limits, conditions, or constraints. A <code>Constraint</code> consists of a Requirement 
    and a value against which that requirement is applied. The Requirement is in the form of a <code>Requirement enum</code> constant and the value is an Object.</p>
    <p>A <code>Constraint</code> instance can be converted to String by its <code>toString()</code> method (of course!) but this class also contains a static method 
    to parse such a string back into an instance. The format is fairly straightforward but should not be modified or created manually! Nested <code>Constraint</code>s 
    are possible, but the innermost MUST have a non-Constraint data Object. There is no <code>toXML()</code> method since the <code>Constraint</code> class does not 
    have a pre-defined and specific place in the structure of a graph. This means implementations are responsible for handling them as needed. That is where the 
    <code>parseFromString()</code> method comes into play: it re-creates a <code>Constraint</code> from its Stringified value.</p>
*/
public class Constraint {

  /** <p>This <code>enum</code> is provided as a <code>switch()</code>-able way to designate specific conditions and their targeted value. The only value associated 
      with any Requirement is its String designation. Where not stated, the designation matches Java syntax. E.g. EQUAL is '=='.</p>
      <p>The intent behind this enum is to provide a 'human-readable' indication of what conditions and constraints should be present when graphing. The second 
      concern is a <code>switch()</code>-able way to express or generate piecewise functions or similar constructs.</p>
  */
  public static enum Requirement {
    /** String designation is: ??. */
    UNKNOWN("??"),
    LESS_THAN("<"),
    LESS_OR_EQUAL("<="),
    GREATER_THAN(">"),
    GREATER_OR_EQUAL(">="),
    NOT_EQUAL("!="),
    EQUAL("=="),
    AND("&&"),
    OR("||"),
    NOT("!"),
    /** String designation is: '&lt;--&gt;'. */
    INFINITY("<-->"),
    /** String designation is: '&lt;--'. */
    NEGATIVE_INFINITY("<--"),
    /** String designation is: '--&gt;'. */
    POSITIVE_INFINITY("-->"),
    /** String designation is '&gt;0'. */
    POSITIVE(">0"),
    /** String designation is: '&lt;0' . */
    NEGATIVE("<0"),
    /** String designation is: '!{}' . This constraint can also be interpreted as 'non-empty' or 'defined'. */
    EXIST("!{}"),
    /** String designation is: '{}' . This can also be interpreted as 'does not exist' or 'undefined'. */
    EMPTY("{}"),
    /** String designation is 'min' (lower case) . */
    MINIMUM("min"),
    /** String designation is 'max' (lower case). */
    MAXIMUM("max"),
    /** String designation is '[]' (square brackets). */
    INTERVAL("[]"),
    /** String designation is '$' (dollar sign). */
    TEXT("$"),
    /** String designation is '(C)' . */
    COMPLEX("(C)"),
    /** String designation is '(R)' . */
    REAL("(R)"),
    /** String designation is '(Q)' . */
    RATIONAL("(Q)"),
    /** String designation is '(Z)' . */
    INTEGER("(Z)"),
    /** String designation is '(W)' . */
    WHOLE("(W)"),
    /** String designation is '(N)' . */
    NATURAL("(N)");
    

    private final String myID;
    Requirement(String pID) { this.myID = pID; }

    /** Determine whether or not the given String matches this Condition. */
    public boolean matches(String pDesignation) { return myID.equalsIgnoreCase(pDesignation); }

    /** Retrieve the symbol associated with this Condition. */
    public String designation() { return myID; }
  }

  /* - - - - - - - - - - END OF ENUM - - - - - - - - - - */

  private Requirement myCondition = Requirement.UNKNOWN;
  private Object myValue = null;

  /** Create a Constraint in which the Condition is UNKNOWN and the values are null. */
  public Constraint() { super(); }

  /** Create a Constraint in which the Condition and value are both specified. */
  public Constraint(Requirement pCondition, Object pValue) {
    super();
    if(pCondition != null) myCondition = pCondition;
    myValue = pValue;
  }

  /** Create a Constraint from a String representation and an Object. */
  public Constraint(String pCondition, Object pValue) { this(match(pCondition), pValue); }

  /** Fetch the value of this Constraint's Requirement. */
  public Requirement requirement() { return myCondition; }

  /** Fetch a reference to the value Object. NOTE: This is a reference to the Object itself! No copies are made. */
  public Object value() { return myValue; }

  /** This method provides a safe way to get a Requirement. If the String does not match, Requirement.UNKNOWN is returned. Any errors are reported at Debug level. */
  public static Requirement findRequirementByName(String pName) {
    Requirement retval = Requirement.UNKNOWN;
    try { retval = Requirement.valueOf(pName); } catch(Exception ex) { CapCom.debug("Error finding Condition "+pName, ex); }
    return retval;
  }

  /** Instance version of <code>findRequirementByName()</code>. */
  public Requirement getRequirementByName(String pName) { return findRequirementByName(pName); }

  /** Match a Requirement to a String. Returns Condition.UNKNOWN if no match is found. */
  public static Requirement match(String pCondition) {
    Requirement retval = Requirement.UNKNOWN;
    for(Requirement cond : Requirement.values()) {
      if(cond.matches(pCondition)) retval = cond;
    }
    return retval;
  }

  /** Instance version of <code>match()</code>. */
  public Requirement matchRequirement(String pCondition) { return match(pCondition); }

  /** Convert this <code>Constraint</code> to a String. The format is straightforward but must not be altered if the <code>parseFromString()</code> method might 
      be needed. No XML-escaping is done so this should be done manually if it might appear in an XML document or if the value Object's String might contain 
      XML-dangerous content.
  */
  public String toString() {
    StringBuffer retval = new StringBuffer(myCondition.toString());
    if(myValue != null) {
      retval.append(" ").append(myValue.toString()).append(" ");
    }
    return retval.toString().trim();
  }

  /** <p>This method parses the <code>toString()</code> value from a <code>Constraint</code> and creates a new <code>Constraint</code> from it. It should not be 
      used for any other purpose! There is no instance version.</p>
      <p>WARNING: This method is <b>brittle</b>! Use only as directed if at all!!</p>
  */
  public static final Constraint parseFromString(String pData) {
    Constraint retval = new Constraint();
    if(!Question.validString(pData)) return retval;
    String lastValue = null;
    Requirement reqTemp = Requirement.UNKNOWN;
    Stack<Requirement> reqStack = new Stack<Requirement>();
    String[] tokenz = pData.split(" ");
    int iNdx = 0;
    for( iNdx = 0; iNdx < tokenz.length; iNdx++) {
      reqTemp = findRequirementByName(tokenz[iNdx]);
      CapCom.debug("Requirement: "+reqTemp.toString());
      if(reqTemp == Requirement.UNKNOWN) break;
      reqStack.push(reqTemp);
    }
    if(iNdx < tokenz.length) {
      lastValue = "";
      while(iNdx < tokenz.length) { lastValue = lastValue + tokenz[iNdx++] + " "; }
      CapCom.debug("Last value: "+lastValue);
    }
    if(!reqStack.isEmpty()) {
      retval = new Constraint(reqStack.pop(), lastValue);
      while(!reqStack.isEmpty()) { retval = new Constraint(reqStack.pop(), retval); }
    } else {
      retval = new Constraint(Requirement.UNKNOWN, lastValue);
    }
    return retval;
  }

}
