package cox.jmatt.java.MathTools.graphing;

import cox.jmatt.java.MathTools.*;

/** <p>This is the only concrete class in the <code>graphing.*</code> package that is absolutely required. Its purpose is to create <code>GraphEngine</code>s for 
    specific implementations. Doing so requires the fully-qualified name of the GraphEngine desired and, of course, access to it. Access requires that the implementation 
    be 'visible' to <code>GraphEngineFactory</code>, either by being in the ClassPath OR by being in CapCom's ClassLoader. GraphEngineFactory can be instructed 
    to use it via the <code>useMathClassLoader()</code> method.</p>
    <p>This class is also able to provide <code>RenderEngine</code>s and <code>Constraint</code> instances. These classes, while useful, are NOT absolutely mandatory 
    for constructing graphs!</p>
*/
public class GraphEngineFactory {

  /** String constant for constructing a <code>graphing.xml.GraphEngineXML</code> instance. */
  public static final String XML_ENGINE = "cox.jmatt.java.MathTools.graphing.xml.GraphEngineXML";

  /** String constant for constructing a <code>graphing.Graph.GraphEnginePDK</code> instance. */
  public static final String GRAPH_ENGINE = "cox.jmatt.java.MathTools.graphing.Graph.GraphEnginePDK";

  /** String constant for a no-op or 'null' <code>GraphEngine</code>. */
  public static final String NO_ENGINE = "cox.jmatt.java.MathTools.graphing.noop.NOGraphEngine";

  // Private constructor. Use newInstance() to get one.
  private GraphEngineFactory() { super(); }

  private ClassLoader myClassLoader = null;
  private static GraphEngineFactory myFactory = null;

  /** Use this method to obtain an instance of GraphEngineFactory. */
  public static GraphEngineFactory newInstance() {
    if(myFactory == null) myFactory = new GraphEngineFactory();
    return myFactory;
  }

  /** Call this method to enable or disable CapCom's ClassLoader to find <code>GraphEngine</code> instances.
      @param pEnable true to set the MathClassLoader, false to clear it.
  */
  public void setMathClassLoader(boolean pEnable) {
    myClassLoader = pEnable?CapCom.getMathClassLoader():null;
  }

  /** Clear all previous settings including the ClassLoader. */
  public void reset() {
    myClassLoader = null;
    myFactory = null;
  }

  /** Create and return a new GraphEngine tuned to a specific implementation. The fully-qualified class name must be used, and the class itself must be visible 
      in order to be created. To use an implementation from an external JAR, point ScriptPanel's URLClassLoader to it and enable it. If no ClassLoader is set 
      <code>GraphEngineFactory</code> uses the ClassLoader that created it.
      @param pEngineName The full class name of the desired GraphEngine. If blank or empty this method returns null.
  */
  public GraphEngine getGraphEngine(String pEngineName) {
    GraphEngine retval = null;
    if(!Question.validString(pEngineName)) return retval;
    if(myClassLoader == null) myClassLoader = this.getClass().getClassLoader();
    try {
      retval = (GraphEngine) Class.forName(pEngineName, true, myClassLoader).newInstance();
    } catch(Exception ex) { CapCom.error("Error getting GraphEngine", ex); }
    return retval;
  }

  /** Create and return a <code>RenderEngine</code>. */
  public RenderEngine getRenderEngine() { return new RenderEngine(); }

  /** Create and return a new <code>Constraint</code> with the given requirement and value Object.
      @param pRequirement The String designation of the desired Requirement.
      @param pValue The value Object.
      @return A fresh new Constraint.
  */
  public Constraint getConstraint(String pRequirement, Object pValue) { return new Constraint(pRequirement, pValue); }

}
