package cox.jmatt.java.MathTools.graphing;

import java.util.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.graphing.*;

/** <p>This class is designed specifically to parse <code>graphing.xml.*</code> package markup and convert it into a <code>GraphPaper</code> instance containing 
    the components described in the XML data. It should not be used for any other purpose otherwise the universe will shatter.</p>
    <p><code>GraphHandler</code> extends <code>MTXErrorHandler</code> and is used as such. The only methods defined are several from the 
    <code>org.xml.sax.ContentHandler</code> and <code>ErrorHandler</code> interfaces. The latter simply reports parser <b>warns</b> and <b>errors</b> to 
    <code>CapCom.debug()</code> and <b>fatals</b> to <code>CapCom.error()</code> If desired, an alternate <code>ErrorHandler</code> can be set without affecting 
    the main operations of this class.</p>
    <p>The <code>ContentHandler</code> methods defined are the fewest necessary to accomplish <code>GraphHandler</code>'s task: converting graphing package XML 
    into graphing components. <code>MTXProcessor</code> uses this class in its <code>parseGraph()</code> method. It simply substitutes an instance of this class 
    for its handler when parsing the XML markup.</p>
    <h3>Using <code>GraphHandler</code></h3>
    <p>To use this class from an external application, first create an instance. Next set a <code>GraphEngine</code> of the desired output type using the 
    <code>setGraphEngine()</code> method. Obtain a <code>SAXParser</code> instance, set <code>GraphHandler</code> as its ContentHandler (at LEAST), and parse the 
    XML content. If no external ErrorHandler is necessary, set this class as that and use MathTools' logging tools! When the document finishes, the graph will be 
    ready. Call <code>getResult()</code> to retrieve a complete <code>GraphPaper</code> instance from the GraphEngine sent it. This is <i>exactly</i> what 
    <code>MTXProcessor</code> does when it parses a graph!</p>
    <p>As with the <code>graphing.xml.*</code> implementation, all Object data <b>must</b> be String. If not, the <code>GraphPaper</code> <b>must</b> know how to 
    convert the Strings returned by the XML parse into whatever Objects it needs. Also, be very, VERY careful if editing the graphing XML or generating it from 
    an external source. <code>GraphHandler</code> is NOT TOLERANT of bad XML markup! A complete, albeit short, example of GraphingXML is the file <b>XMLGraph.xml</b> 
    in the <b>docs/SampleScripts/</b> folder. This shows the root, <code>Renderable</code> tags, and content tags <i>exactly as they should be!</i></p>
    <p>When the document is parsed, the three general attributes are set as follows:
    <ul>
      <li>The <b>id=</b> attribute is used to add the component to the GraphPaper: e.g. <code>GraphPaperInstance.addAxis(<u>id</u>, AxisInstanceFromTag)</code></li>
      <li>The <b>type=</b> attribute is set using <code>RenderableInstance.setType(<u>type</u>)</code> as soon as the instance is created.</li>
      <li>The <b>name=</b> attribute is used to create an instance of the component: <code>GraphableInstance = GraphEngineInstance.getGraph(<u>name</u>)</code></li>
    </ul>
    Instance-specific attributes (xmin, height, closed, scale, etc.) are set when the instances to which they apply are created. NOTE: Although Transforms are not 
    added with a String key, the <code>graphing.xml.*</code> package DOES assign a simple ID attribute to them. When parsed, this is ignored.</p>
    <p>USE CAUTION when parsing XML not generated by the <code>graphing.xml.*</code> implementation! Some Exceptions are caught and handled gracefully while others 
    are handled less so. <code>MTXProcessor.parseGraph()</code> defines an all-encompassing <code>try/catch</code> block and several methods in this class will 
    handle some errors, but it is a BAD IDEA to count on that behavior! It is a much better idea to make sure the XML is good before it is parsed!</p>
*/
public class GraphHandler extends cox.jmatt.java.MathTools.util.MTXErrorHandler {

  private GraphEngine myGraphEngine = null;
  private Renderable curTag = null;
  private GraphPaper myPaper = null;
  private String curID = null, curName = null, curType = null, curKey = null;
  private boolean curClosed = false;
  private StringBuffer sbContent = new StringBuffer();

  /** This constructor builds a standard GraphHandler. Nothing else is needed. */
  public GraphHandler() { super(); }

  /** This method sets the <code>GraphEngine</code> used to build the graph described in the XML document. If set to (or left as) null, the default engine used 
      is <code>GraphEngineXML</code>. This GraphEngine must be able to make sense of Strings in Object parameter spots as that is how they are set.
      @param pEngine The <code>GraphEngine</code> to use in constructing the graph.
  */
  public void setGraphEngine(GraphEngine pEngine) { myGraphEngine = pEngine; }

  /** Prints a debug message. If no <code>GraphEngine</code> has been set, install a default (XML) one. */
  public void startDocument() {
    CapCom.debug("Starting graph processing.");
    if(myGraphEngine == null) myGraphEngine = new cox.jmatt.java.MathTools.graphing.xml.GraphEngineXML();
  }

  /** Start an element. First determine what the element is and how to deal with it. */
  public void startElement(String uri, String localName, String qName, Attributes attrs) {
    curName = attrs.getValue("", "name");
    curType = attrs.getValue("", "type");
    curKey = attrs.getValue("", "key");
    sbContent = new StringBuffer();
    if("GraphPaper".equals(localName)) {
      myPaper = myGraphEngine.getGraphPaper(curName);
      if(Question.validString(curType)) myPaper.setType(curType);
      try {
        double xMin = Double.parseDouble(attrs.getValue("", "xmin"));
        double xMax = Double.parseDouble(attrs.getValue("", "xmax"));
        double yMin = Double.parseDouble(attrs.getValue("", "ymin"));
        double yMax = Double.parseDouble(attrs.getValue("", "ymax"));
        myPaper.setRanges(xMin, yMin, xMax, yMax);
      } catch(Exception ex) { CapCom.debug("Error setting GraphPaper ranges.", ex); }
      try {
        double dHeight = Double.parseDouble(attrs.getValue("", "height"));
        double dWidth = Double.parseDouble(attrs.getValue("", "width"));
        myPaper.setSize(dWidth, dHeight);
      } catch(Exception ex) { CapCom.debug("Error setting GraphPaper size.", ex); }
    } else if("Axis".equals(localName)) {
      curID = attrs.getValue("", "id");
      curTag = myGraphEngine.getAxis(curName);
      String sScale = attrs.getValue("", "scale");
      String sTicks = attrs.getValue("", "tickMarks");
      if(Question.validString(sScale)) ((Axis) curTag).setScale(sScale);
      if(Question.validString(sTicks)) ((Axis) curTag).setTickMarks(sTicks);
      if(Question.validString(curType)) curTag.setType(curType);
    } else if("Graphable".equals(localName)) {
      curID = attrs.getValue("", "id");
      curTag = myGraphEngine.getGraph(curName);
      if(Question.validString(curType)) curTag.setType(curType);
    } else if("Expression".equals(localName)) {
      // Pre-process Expression stuff.
    } else if("Limit".equals(localName)) {
      // Pre-process Limits
    } else if("Polygon".equals(localName)) {
      curID = attrs.getValue("", "id");
      curTag = myGraphEngine.getPolygon(curName);
      if(Question.validString(curType)) curTag.setType(curType);
    } else if("Vertices".equals(localName)) {
      curClosed = ("true").equals(attrs.getValue("", "closed"));
      ((Polygon) curTag).setClosed(curClosed);
    } else if("Transform".equals(localName)) {
      // Pre-process transforms
    } else if("Param".equals(localName)) {
      // Pre-process parameters. May happen on multiple different or nested tags.
    } else { CapCom.debug("Unknown element: "+localName); }
    if(Question.validString(curType) && (curTag instanceof Renderable)) curTag.setType(curType);
    
  }

  /** If the element is Renderable, add it to the GraphPaper. */
  public void endElement(String uri, String localName, String qName) {
  
    if("GraphPaper".equals(localName)) {
      // Post-process GraphPaper. At this point the document is complete.
    } else if("Axis".equals(localName)) {
      if(myPaper != null) myPaper.addAxis(curID, (Axis) curTag);
      curID = null;
      curTag = null;
    } else if("Graphable".equals(localName)) {
      if(myPaper != null) myPaper.addGraph(curID, (Graphable) curTag);
      curID = null;
      curTag = null;
    } else if("Expression".equals(localName)) {
      if(curTag != null) ((Graphable) curTag).addExpression(curKey, sbContent.toString());
    } else if("Limit".equals(localName)) {
      if(curTag != null) ((Graphable) curTag).addLimit(curKey, sbContent.toString());
    } else if("Polygon".equals(localName)) {
      if(myPaper != null) myPaper.addPolygon(curID, (Polygon) curTag);
      curID = null;
      curTag = null;
    } else if("Vertices".equals(localName)) {
      String [] sPoints = sbContent.toString().split(" ");
      for(int i=0; i < sPoints.length / 2; i++) {
        try{
          ((Polygon) curTag).addVertex(Double.parseDouble(sPoints[2*i]), Double.parseDouble(sPoints[2*i + 1]));
        } catch(Exception ex) { CapCom.error("Error converting point: ("+sPoints[2*i]+", "+sPoints[2*i + 1]+").", ex); }
      }
    } else if("Transform".equals(localName)) {
      if(myPaper != null) myPaper.addTransform(sbContent.toString());
    } else if("Param".equals(localName)) {
      if(curTag != null) curTag.setParameter(curKey, sbContent.toString());
      if((curTag == null) && (myPaper != null)) myPaper.setParameter(curKey, sbContent.toString());
    } else { CapCom.debug("Unknown element: "+localName); }

  }

  /** This method is used to accumulate data for tags that have character content: &lt;Expression&gt; &lt;Limit&gt; &lt;Vertices&gt; and &lt;Param&gt;. */
  public void characters(char[] pChars, int pStart, int pLen) { sbContent.append(pChars, pStart, pLen); }

  /** Clean up any mess left behind by the parsing process. */
  public void endDocument() {
    curID = null; curName = null; curType = null; curKey = null;
    myGraphEngine = null; curTag = null; curClosed = false;
    sbContent = new StringBuffer();
  }

  /** Return the <code>GraphPaper</code> instance parsed. This method is only valid AFTER the parsing process! Beforehand the return value is indeterminate. All 
      specified components will have been added per their specifications in the XML document. The GraphPaper instance will, of course, belong to the <code>GraphEngine</code> 
      set before parsing started.
  */
  public GraphPaper getResult() { return myPaper; }

}
