package cox.jmatt.java.MathTools.graphing;

/** <p><code>GraphPaper</code> is what brings together all the other graph components. It defines the 'surface' or 'paper' on which a graph is drawn and it 
    provides the machinery to generate the Object which will be transformed into the final graph. This last point is critical! Once the GraphPaper <code>render()</code> 
    method is called MathTools' responsibility <b>ends</b>. The Object returned is, as far as MathTools is concerned, the complete and final graph. No further 
    interaction is required.</p>
    <p>Since <code>GraphPaper</code> bridges the gap between the broad metaphor used by MathTools and an actual printed (or image) graph, its methods are defined 
    more in the sense of what the final presentation will look like. Thus arguments involving such things as size and boundary values are represented as actual 
    numbers (doubles) as opposed to generic Objects. The exact interpretation still depends on the implementation, but the values set should 'resemble closely' 
    the concepts that are implied.</p>
    <p>Interaction between the <code>setSize()</code> and <code>setRanges()</code> methods is not defined explicitly nor implied. Either method may be used, or 
    both, and the result is handled 'under the hood.' Likewise, the <code>addTransform()</code> method is provided as an option for implementation technology 
    capable of supporting it. The actual transformation is not defined and is specified as Object to allow maximum flexibility.</p>
*/
public interface GraphPaper extends Renderable {

  /** Set the size of the actual graph.
      @param pWidth The width of the graph.
      @param pHeight The height of the graph.
      @return true for valid settings, false otherwise.
  */
  public boolean setSize(double pWidth, double pHeight);

  /** Set the minimum and maximum horizontal and vertical values for the graph.
      @param pMinX The minimum horizontal (X) value on the graph.
      @param pMinY The minimum vertical (Y) value for the graph.
      @param pMaxX The maximum horizontal value for the graph.
      @param pMaxY The maximum vertical (Y) value on the graph.
      @return true if the implementation can support these settings, false otherwise.
  */
  public boolean setRanges(double pMinX, double pMinY, double pMaxX, double pMaxY);

  /** Use this method to apply a universal transformation to all of the graph components. Exactly what this entails, or if it is even used, depend on the 
      implementation. Implementations should keep track of the transforms applied AND the order in which they should be applied! Presumably the transform takes 
      place during rendering but this is a suggestion, not a requirement!
      @param pTransform The transformation to apply.
      @return true if the transform succeeds, false otherwise.
  */
  public boolean addTransform(Object pTransform);

  /** Add an Axis instance to the current paper. The key is provided to allow unique identification of axes.
      @param pKey The identification or description of the Axis being added.
      @param pAxis The Axis to add.
      @return true if it happened, false otherwise.
  */
  public boolean addAxis(String pKey, Axis pAxis);

  /** Add a graph to the current paper. The key is provided for identification or description and its use is optional.
      @param pKey The String key used to identify the graph.
      @param pGraph The graph to add.
      @return true if all is well, false if something bad happened.
  */
  public boolean addGraph(String pKey, Graphable pGraph);

  /** Add a polygon to the current paper. The key is provided for identification or description and its use is optional.
      @param pKey The String key used to identify the polygon.
      @param pPolygon The <code>Polygon</code> to add.
      @return true if all is well, false if something bad happened.
  */
  public boolean addPolygon(String pKey, Polygon pPolygon);

}
