package cox.jmatt.java.MathTools.graphing;

/** <p>This interface represents the actual graph drawn (rendered) onto a sheet of graph paper. The critical concepts here are the <b>expression</b> and its 
    <b>limits</b>. The expression component defines a function, expression, equation, inequality, or other mathematical <i>thing</i> that can be used to generate 
    a graph. The limits define conditions or constraints over which such generation might happen.</p>
    <p>Speaking broadly, the expression is evaluated over its limits to generate the graph itself. Exactly how these components interact depends on the actual 
    implementation, but the package is designed around the stated algorithm.</p>
    <p>For each of the String constants the value equals the fully-qualified name of the String itself. This is done to allow for easy String conversion when 
    rendering: the <code>toString()</code> value is the name of the constant itself.</p>
*/
public interface Graphable extends Renderable {

  /** This constant is provided as a convenient way to indicate the function X(t) in a parametric equation. */
  public static final String PARAMETRIC_X = "cox.jmatt.java.MathTools.graphing.Graphable.PARAMETRIC_X";

  /** This constant is provided as a convenient way to indicate the function Y(t) in a parametric equation. */
  public static final String PARAMETRIC_Y = "cox.jmatt.java.MathTools.graphing.Graphable.PARAMETRIC_Y";

  /** This constant can be used to indicate a 'standard' graphable function: y = f(x). */
  public static final String STANDARD_GRAPH = "cox.jmatt.java.MathTools.graphing.Graphable.STANDARD_GRAPH";

  /** Use this constant to key a polar function: r = f(t). */
  public static final String POLAR_GRAPH = "cox.jmatt.java.MathTools.graphing.Graphable.POLAR_GRAPH";

  /** Use this constant to key a parametric graph. */
  public static final String PARAMETRIC_GRAPH = "cox.jmatt.java.MathTools.graphing.Graphable.PARAMETRIC_GRAPH";

  /** Add an expression to the current graph. Ordinarily each graph will have only one expression, but parametric functions or piecewise graphs may require more. 
      The pKey is optional but can be used if necessary.
      @param pKey A String identifying the expression to be added. Usage is optional.
      @param pExpression The Object representing the graph-generating entity.
      @return true if the operation succeeds, false otherwise.
  */
  public boolean addExpression(String pKey, Object pExpression);

  /** Add a limit to the current graph. This may be in the form of a <code>Constraint</code> or it may be entirely implementation-defined. As with <code>addExpression()</code>, 
      a String key is provided for identification or classification. Usage is optional but strongly suggested!
      @param pKey A String used to identify the limit being added. Not required.
      @param pLimit An Object representing the limit to be enforced.
      @return true if the limit makes sense and is added successfully, false otherwise.
  */
  public boolean addLimit(String pKey, Object pLimit);

}
