package cox.jmatt.java.MathTools.graphing;

/** <p>This is the base interface for anything that might appear on a (complete!) graph. It serves as the base interface for all other graph components but will 
    not likely be included in its own right. This is an assumption, however, and not a fact!</p>
    <p>NOTE: This class has no connection with any other Java class containing the word 'render'! It is used here in the sense of 'something rendered onto a sheet 
    of graph paper.' That is, the usage context for 'render' is 'draw' or 'drawn.'</p>
    <h3>The Rendering Context</h3>
    <p>The Rendering Context Object is provided as a way for an application to pass last-second (or later) data to the rendering process. No specific definition 
    is required, the Context is supplied as a way to pass configuration data or other such information to a graph as it is being rendered.</p>
    <p>Implementations using a rendering context are responsibile for detailing what is needed and making it available to the classes requiring it. If the requirements 
    are not strict, a GraphKeys object can be used.</p>
*/
public interface Renderable {

  /** <p>This method is the raison d'etre for the Renderable interface. Calling <code>render()</code> finishes the graphing process. The Object returned is then 
      ready for graphing. That is to say, it is ready to be fed into the external application that actually <i>generates</i> the graph; once this method is called 
      MathTools' responsibility is finished. No further interaction with the returned Object is required!</p>
      @param pContext The Context Object to be used for rendering.
      @return The graph-ready Object, requiring no further MathTools interaction.
  */
  public Object render(Object pContext);

  /** Set the type of Renderable this is. The exact meaning of 'type' and how (or IF) it is used are implementation-dependent.
      @param pType The 'type' of Renderable being used.
      @return true if setting <b>this</b> renderable to <b>pType</b> makes sense, false otherwise.
  */
  public boolean setType(Object pType);

  /** This method provides a very general way to configure any type of <code>Renderable</code>. The only absolute restriction is that the name of the parameter 
      being set is not empty or null. It is the responsibility of the implementation to define what parameters are settable in which components.
      @param pKey The name of the parameter being set. Must not be null or empty.
      @param pValue The value of the parameter to set.
      @return true if the parameter is set properly, false if it did not make sense or could not be set.
  */
  public boolean setParameter(String pKey, Object pValue);

  /** Test the Renderable to see whether or not a particular condition exists. Exactly what these conditions are and how they are used depend on the implementation; 
      this method is defined to allow a uniform way to check them.
      @param pCondition The condition to test against the value.
      @param pValue The value against which the condition is to be checked.
      @return true if the test succeeds, false otherwise.
  */
  public boolean testCondition(Object pCondition, Object pValue);

}
