package cox.jmatt.java.MathTools.graphing;

import java.util.Hashtable;
import cox.jmatt.java.MathTools.*;

/** <p>This abstract class provides a partial implementation of the <code>Renderable</code> interface. It defines two constructors, one with no arguments and one 
    that takes a String. The second constructor sets a protected field (myName) that subclasses can use, if such is desirable. It also defines the <code>setType()</code> 
    method and provides a protected Object to hold its value. For subclasses that don't need condition testing, an empty <code>testCondition()</code> method is 
    defined: it always returns false.</p>
    <p>The <code>setParameter()</code> method actually provides considerable power. It makes sure the String key used to store an Object is not null or blank, 
    and it provides an intuitive way to remove previously-set parameters. The return value is true if and only if: the key String is valid, a value was added 
    successfully, or one was removed successfully. It returns false if the key is null or empty, an add fails, or there was no existing mapping to remove.</p>
    <p>This class does NOT define a <code>render()</code> method since that method cannot have a sensible default.</p>
*/
public abstract class RenderableImpl implements Renderable {

  /** This String is assigned from the one-argument constructor (String). Subclasses can access this field as needed and implementations using such functionality 
      should also define one-argument (String) constructors. The default value is null.
  */
  protected String myName = null;

  /** This is the Object set via the <code>setType()</code> method. No verification or filtering is done. The default value is null. */
  protected Object myType = null;

  /** This <code>Hashtable</code> holds parameter key-value pairs. The <code>setParameter()</code> method ensures no keys or values will be null. Upon creation 
      this field contains an empty table. Subclasses should NOT EVER assign it to be null, otherwise <code>setParameter()</code> will begin throwing nasty 
      NullPointerExceptions!
  */
  protected Hashtable<String, Object> myParameters = new Hashtable<String, Object>();

  /** Create an unnamed instance. */
  public RenderableImpl() { super(); }

  /** Create a named instance. The String name is assigned to the 'myName' field. No verification or filtering is done; the field is assigned whatever value is 
      passed into this method.
  */
  public RenderableImpl(String pName) {
    super();
    myName = pName;
  }

  /** Set the value of the 'myType' field. If its (after-assignment) value is null the return value is false. If not null it is true. */
  public boolean setType(Object pType) {
    myType = pType;
    return (myType == null);
  }

  /** This is an empty method. It always returns false. */
  public boolean testCondition(Object pCondition, Object pValue) { return false; }

  /** This method applies some processing to its function. If the String key is null or empty it returns false immediately. If the value Object is not null the 
      method returns true if it was successfully set, false if not. If pValue is null, the method attempts to remove pKey from the parameter table. If a mapping 
      existed and an Object was removed the return value is true. If no mapping existed for pKey the method returns false.
      @param pKey The name of the Object to add or remove. Must not be null or blank.
      @param pValue The Object to add or null to remove.
      @return false if the key is not valid, an add fails, or no mapping exists to remove; true otherwise.
  */
  public boolean setParameter(String pKey, Object pValue) {
    if(!Question.validString(pKey)) return false;
    Object retval = null;
    if(pValue == null) {
      retval = myParameters.remove(pKey);
    } else {
      retval = myParameters.put(pKey, pValue);
    }
    return (retval != null);
  }

}

