package cox.jmatt.java.MathTools.graphing.xml;

import java.util.*;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.graphing.*;

/** <p>This is the bring-together class of the <code>graphing.xml.*</code> package. When <code>GraphPaperXML</code> is rendered it yields an XML document containing 
    all of the graphing data. All Object arguments are <code>toString()</code>-ed to allow for proper representation, so implementations that do NOT use Strings 
    won't get the best mileage from this one.</p>
    <p>Implementations that <i>do</i> use Strings can use this package to persistently store graphs. The ONLY non-standard method used is <code>setID()</code> 
    and it is used transparently.</p>
    <p>Common XML attributes are ID, name, and type. The type is that set in Renderable, and the IDs are assigned when graphing components are added to GraphPaper. 
    The 'name=' attribute is assigned when the component is created; from <code>GraphEngine</code>. This is all handled automatically. Calling <code>GraphPaper's 
    render()</code> method both generates root tags and GraphPaper data and renders all contained elements, in order: Axes, Graphs, Polygons, Transforms, and 
    parameters. The tags are simple, straightforward, and easy to use.</p>
    <p>NOTE: None of the XML package classes need or use a rendering context Object. It is ignored in all cases.</p>
*/
public class GraphPaperXML extends RenderableXML implements GraphPaper {

  private Hashtable<String, Axis> myAxes = new Hashtable<String, Axis>();
  private Hashtable<String, Graphable> myGraphs = new Hashtable<String, Graphable>();
  private Hashtable<String, Polygon> myPolys = new Hashtable<String, Polygon>();
  private ArrayList<String> myTransforms = new ArrayList<String>();
  private double myXMin = -10, myYMin = -10, myXMax = 10, myYMax = 10, myWidth = 20, myHeight = 20;

  /** Zero-argument constructor. No name supplied. */
  public GraphPaperXML() { super(); }

  /** Constructor for named GraphPaperXML. */
  public GraphPaperXML(String pName) { super(pName); }

  /** Ranges are presented as attributes in the &lt;GraphPaper&gt; tag: xmin, ymin, xmax, ymax. Return value is always true. */
  public boolean setRanges(double pXMin, double pYMin, double pXMax, double pYMax) {
    myXMin = pXMin; myXMax = pXMax; myYMin = pYMin; myYMax = pYMax;
    return true;
  }

  /** Height and width are rendered as 'height=' and 'width=' attributes on the &lt;GraphPaper&gt; tag. Always returns true. */
  public boolean setSize(double pWidth, double pHeight) {
    myWidth = pWidth; myHeight = pHeight;
    return true;
  }

  /** Transforms, if used, MUST be representable by String! If the transform is null, it is ignored and false is returned. If non-null, its <code>toString()</code> 
      is added and the return value is true.
  */
  public boolean addTransform(Object pTransform) {
    if(pTransform == null) return false;
    myTransforms.add(pTransform.toString());
    return true;
  }

  /** Add an <code>Axis</code> to this <code>GraphPaper</code>. The key argument becomes the Axis' ID attribute, it MUST NOT be null or blank! If pAxis is null 
      it is removed, returning true if it was there and false if not.
  */
  public boolean addAxis(String pKey, Axis pAxis) {
    if(!Question.validString(pKey)) return false;
    Object retval = null;
    if(pAxis == null) {
      retval = myAxes.remove(pKey);
    } else {
      if(pAxis instanceof RenderableXML) ((RenderableXML) pAxis).setID(pKey);
      retval = myAxes.put(pKey, pAxis);
    }
    return (retval == null);
  }

  /** Add a <code>Graphable</code> to this <code>GraphPaper</code>. The key argument becomes the Axis' ID attribute, it MUST NOT be null or blank! If pGraph is null 
      it is removed, returning true if it was there and false if not.
  */
  public boolean addGraph(String pKey, Graphable pGraph) {
    if(!Question.validString(pKey)) return false;
    Object retval = null;
    if(pGraph == null) {
      retval = myGraphs.remove(pKey);
    } else {
      if(pGraph instanceof RenderableXML) ((RenderableXML) pGraph).setID(pKey);
      retval = myGraphs.put(pKey, pGraph);
    }
    return (retval == null);
  }

  /** Add an <code>Polygon</code> to this <code>GraphPaper</code>. The key argument becomes the Axis' ID attribute, it MUST NOT be null or blank! If pPoly is null 
      it is removed, returning true if it was there and false if not.
  */
  public boolean addPolygon(String pKey, Polygon pPoly) {
    if(!Question.validString(pKey)) return false;
    Object retval = null;
    if(pPoly == null) {
      retval = myPolys.remove(pKey);
    } else {
      if(pPoly instanceof RenderableXML) ((RenderableXML) pPoly).setID(pKey);
      retval = myPolys.put(pKey, pPoly);
    }
    return (retval == null);
  }

  /** Bring it all together and print it out as XML. All elements are contained within the root &lt;GraphPaper&gt; tag which has NO ID attribute. No rendering 
      context is used.
  */
  public Object render(Object pContext) {
    StringBuffer retval = new StringBuffer("<GraphPaper ");
    if(Question.validString(myName)) retval.append("name=\"").append(myName).append("\" ");
    if(Question.validString(myType)) retval.append("type=\"").append(myType).append("\" ");
    retval.append("height=\"").append(myHeight).append("\" ").append("width=\"").append(myWidth).append("\" ");
    retval.append("xmin=\"").append(myXMin).append("\" ").append("xmax=\"").append(myXMax).append("\" ");
    retval.append("ymin=\"").append(myYMin).append("\" ").append("ymax=\"").append(myYMax).append("\" >\n");
    retval.append("  <!-- Axes -->\n");
    for(Enumeration<Axis> en = myAxes.elements(); en.hasMoreElements(); ) { retval.append(en.nextElement().render(pContext).toString()); }
    retval.append("  <!-- Graphs -->\n");
    for(Enumeration<Graphable> en = myGraphs.elements(); en.hasMoreElements(); ) { retval.append(en.nextElement().render(pContext).toString()); }
    retval.append("  <!-- Polygons -->\n");
    for(Enumeration<Polygon> en = myPolys.elements(); en.hasMoreElements(); ) { retval.append(en.nextElement().render(pContext).toString()); }
    retval.append("  <!-- Transforms -->\n");
    int iNdx = 0;
    for(Iterator<String> it = myTransforms.iterator(); it.hasNext(); ) {
      retval.append("  <Transform id=\"").append(iNdx++).append("\" >").append(it.next()).append("</Transform>\n");
    }
    retval.append("  <!-- Parameters -->\n");
    retval.append(super.render(pContext).toString());
    retval.append("</GraphPaper>\n");
    return retval.toString();
  }

}
