package cox.jmatt.java.MathTools.graphing.xml;

import java.util.*;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.graphing.*;

/** <p>This class generates very simple XML markup. The &lt;Graphable&gt; tag contains the necessary 'id=' attribute along with 'name=' and 'type=' attributes, if 
    they are defined. Its children are &lt;Expression&gt; and &lt;Limit&gt; tags whose 'key=' attribute is the key value used to add them to the Graphable and whose 
    content is the XML-escaped String value of the Object added.</p>
*/
public class GraphableXML extends RenderableXML implements Graphable {

  private Hashtable<String, String> myGraphs = new Hashtable<String, String>(), myLimits = new Hashtable<String, String>();

  /** Standard constructor for an unnamed Graphable. */
  public GraphableXML() { super(); }

  /** Constructor for a named Graphable. */
  public GraphableXML(String pName) { super(pName); }

  /** Add an expression to this Graphable. If the key is null or empty the method returns false. If the value is null, it is removed; the return is true if a 
      mapping existed and false if not. If the value is not null it is added and the return is true.
  */
  public boolean addExpression(String pKey, Object pValue) {
    if(!Question.validString(pKey)) return false;
    Object retval = null;
    if(pValue == null) {
      retval = myGraphs.remove(pKey);
    } else {
      retval = myGraphs.put(pKey, pValue.toString());
    }
    return (retval != null);
  }

  /** Add a limit to this Graphable. If the key is null or empty the method returns false. If the value is null, it is removed; the return is true if a 
      mapping existed and false if not. If the value is not null it is added and the return is true.
  */
  public boolean addLimit(String pKey, Object pValue) {
    if(!Question.validString(pKey)) return false;
    Object retval = null;
    if(pValue == null) {
      retval = myLimits.remove(pKey);
    } else {
      retval = myLimits.put(pKey, pValue.toString());
    }
    return (retval != null);
  }

  /** Two things can be tested in <code>GraphableXML</code>: whether or not a specific LIMIT or EXPRESSION has been set. The condition must be either 
      <code>GraphKeysXML.HAS_LIMIT</code> or <code>GraphKeysXML.HAS_EXPRESSION</code>. The value must be the String key used for setting or clearing. If either 
      argument is null the method returns false. If the condition is NOT one of the ones specified, the return value is false.
      @param pCondition One of the GraphKeysXML constants listed above.
      @param pValue The String used to add or clear the expression or limit.
      @return true if the value is present, false if not.
  */
  public boolean testCondition(Object pCondition, Object pValue) {
    if((pCondition == null) || (pValue == null)) return false;
    boolean retval = false;
    if(pCondition == GraphKeysXML.HAS_LIMIT) retval = myLimits.containsKey(pValue.toString());
    if(pCondition == GraphKeysXML.HAS_EXPRESSION) retval = myGraphs.containsKey(pValue.toString());
    return retval;
  }

  /** Rendering a GraphableXML results in a &lt;Graphable&gt; tag containing &lt;Expression&gt; and &lt;Limit&gt; children, followed by any parameter tags. The 
      Graphable tag contains an 'id=' attribute along with 'name=' and 'type=' attributes if they are defined. Each child tag contains a 'key=' attribute and its 
      content is the expression or limit. No rendering context is required. If any parameters are set they follow the last Limit tag.
  */
  public Object render(Object pContext) {
    if(!Question.validString(myID)) myID = "Graphable"+GraphKeysXML.hitCount();
    StringBuffer retval = new StringBuffer("    <Graphable");
    retval.append(" id=\"").append(myID).append("\" ");
    if(Question.validString(myName)) retval.append("name=\"").append(myName).append("\" ");
    if(Question.validString(myType)) retval.append("type=\"").append(myType).append("\" ");
    retval.append(">\n");
    String sKey = null;
    for(Enumeration<String> en = myGraphs.keys(); en.hasMoreElements(); ) {
      sKey = en.nextElement();
      retval.append("      <Expression key=\"").append(sKey).append("\" >").append(MathXML.escapeXML(myGraphs.get(sKey))).append("</Expression>\n");
    }
    for(Enumeration<String> en = myLimits.keys(); en.hasMoreElements(); ) {
      sKey = en.nextElement();
      retval.append("      <Limit key=\"").append(sKey).append("\" >").append(MathXML.escapeXML(myLimits.get(sKey))).append("</Limit>\n");
    }
    retval.append(super.render(pContext));
    retval.append("    </Graphable>\n");
    return retval.toString();
  }

}
