package cox.jmatt.java.MathTools.graphing.xml;

import java.util.*;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.graphing.*;

/** <p>This is the base class for all XML graphing components. It provides basic parameter-handling machinery along with the foundation required for its subclasses. 
    All XML tags should have an ID. While this class does not ID-tag its parameters, it does provide machinery for the other classes to contain their IDs. It 
    also maintains a 'name' and a 'type' field. The name field is set through the one-argument constructor and the type via the <code>setType()</code> method.</p>
    <p>Although this class maintains the parameter table, it is available to subclasses as a protected ArrayList. Likewise the 'myName,' 'myType,' and 'myID' 
    fields. Since these fields are very important to subclasses, their existence can be checked with the <code>testCondition()</code> method. The <code>GraphKeysXML</code> 
    class contains constant condition Objects for verifying that these fields are not null.</p>
*/
public class RenderableXML implements Renderable {

  /** This String is used to hold the name of the object being manipulated. It is set via the one-argument constructor. */
  protected String myName = null;

  /** All XML tags should have an 'id=' attribute. This is the field set by the <code>setID()</code> method. It is protected to allow subclass access. */
  protected String myID = null;

  /** This field holds the String-ed value of the <code>setType()</code> method. Its default value is null. */
  protected String myType = null;

  /** The parameter table is automatically maintained and handled via the <code>render()</code> method. It is protected to allow subclasses access should they 
      need it.
  */
  protected Hashtable<String, String> myParams = new Hashtable<String, String>();

  /** This constructor provides a simple, blank instance. No values are initialized. */
  public RenderableXML() { super(); }

  /** This constructor is provided to set the 'myName' field upon creation. That is all it does. */
  public RenderableXML(String pName) {
    super();
    myName = pName;
  }

  /** This is <b>not</b> a mandated <code>Renderable</code> method. It is provided to allow the inclusion of an 'id=' attribute on any generated tag. */
  public void setID(String pID) { myID = pID; }

  /** Use this method to set or clear the 'myType' field. Setting it to null clears it and setting it to a non-null Object stores its <code>toString()</code>. 
      The return value is always true.
  */
  public boolean setType(Object pType) {
    if(pType == null) {
      myType = null;
    } else {
      myType = pType.toString();
    }
    return true;
  }

  /** Three conditions can be checked: name, type, and ID. Use the <code>GraphKeysXML</code> constants to check whether or not the specified value is non-null. 
      The value argument is not used. The return value is true if the value is set, false if it is null or blank.
  */
  public boolean testCondition(Object pCondition, Object pValue) {
    if(pCondition == GraphKeysXML.GOT_NAME) {
      return Question.validString(myName);
    } else if(pCondition == GraphKeysXML.GOT_ID) {
      return Question.validString(myID);
    }  else if(pCondition == GraphKeysXML.GOT_TYPE) {
      return Question.validString(myType);
    }
    return false;
  }

  /** <p>Set or clear a parameter. If the key is null or blank it is ignored and false is returned. If the key is valid and the value is not null, it is added to 
      the parameter table and the return value is true. If the key is valid and the value is null, the value stored under that key is removed from the table. If 
      the Object was successfully removed, the return is true. If the table held no mapping, the return is false.</p>
      <p>NOTE: The value Object is converted to <b>String</b> before it is added! Use with care.</p>
  */
  public boolean setParameter(String pKey, Object pValue) {
    if(!Question.validString(pKey)) return false;
    Object retval = null;
    if(pValue == null) {
      retval = myParams.remove(pKey);
    } else {
      retval = myParams.put(pKey, pValue.toString());
    }
    return (retval != null);
  }

  /** The base <code>render()</code> method simply produces &lt;Param&gt; tags. No individual ID is set. The key value for each parameter appears in a 'key=' 
      attribute and the tag contains the XML-escaped value. No rendering context is required or used.
  */
  public Object render(Object pContext) {
    if(myParams.isEmpty()) return "";
    StringBuffer retval = new StringBuffer();
    String sKey = null;
    for(Enumeration<String> en = myParams.keys(); en.hasMoreElements(); ) {
      sKey = en.nextElement();
      retval.append("      <Param key=\"").append(sKey).append("\" >").append(MathXML.escapeXML(myParams.get(sKey))).append("</Param>\n");
    }
    return retval.toString();
  }

}
