package cox.jmatt.java.MathTools.markup;

import java.util.Hashtable;

/** <p>This class provides a concrete implementation of the <code>Tag</code> class. Besides that it serves to provide MathTools with a simple way to produce proper 
    XML tags of any necessary complexity. It is similar to <code>MathXML.wrapTag()</code> but with many more features! <code>SimpleTagImpl</code> acts as a container 
    for tag information; it maintains an internal StringBuffer for content and, of course, the Hashtable for attributes. In addition to this it provides instance 
    methods to shadow each of <code>Tag</code>'s static factory methods.</p>
    <p>This class can be used to create any XML tags, up to full documents. Methods that operate on this instance return a self-reference to allow method chaining 
    and tags are easy to create and edit. <code>SimpleTagImpl</code> is not really meant to be subclassed. It is primarily meant for low-level XML tag construction. 
    This might not be the easiest way to do things, but it is easier than trying to slap together Strings manually. It is best used for quick construction of one 
    or two highly-specialized tags. For stricter, more frequent, or greater needs it is probably better to create a markup language package using <code>Tag</code> 
    and <code>StyleInfo</code>.</p>
*/
public class SimpleTagImpl extends Tag {

  private StringBuffer myContent = new StringBuffer();

  /** Standard constructor used for scripting or other purposes. Tag name is 'SimpleTag' with no (default) attributes. */
  public SimpleTagImpl() { super("SimpleTag"); }

  /** Constructor to create a named tag. If the name is null or blank it becomes 'SimpleTag'. */
  public SimpleTagImpl(String pTagName) { super((validString(pTagName)?pTagName:"SimpleTag")); }
  
  /** Fetch a <code>SimpleStyleImpl</code> if needed. */
  public static SimpleStyleImpl createStyleImpl() { return new SimpleStyleImpl(); }

  /** Static method to create a new named <code>SimpleTagImpl</code> instance. */
  public static SimpleTagImpl newInstance(String pName) { return new SimpleTagImpl(pName); }

  /** Instance method to create a new named <code>SimpleTagImpl</code> instance. */
  public SimpleTagImpl getNewInstance(String pName) { return new SimpleTagImpl(pName); }

  /** Static method to create a new blank <code>SimpleTagImpl</code> instance. */
  public static SimpleTagImpl newInstance() { return new SimpleTagImpl(); }

  /** Instance method to create a new blank <code>SimpleTagImpl</code> instance. */
  public static SimpleTagImpl getNewInstance() { return new SimpleTagImpl(); }

  /** Instance form of {@link Tag#openTag Tag.openTag()}. Arguments are per that method. */
  public String openNewTag(String pTagName, Hashtable<String, String> pAttributes, boolean isEmpty) { return openTag(pTagName, pAttributes, isEmpty); }

  /** Instance form of {@link Tag#closeTag Tag.closeTag()}. Arguments are per that method. (This method name is a bit odd but it is consistent!) */
  public String closeNewTag(String pTagName) { return closeTag(pTagName); }

  /** Instance form of {@link Tag#makeTag Tag.makeTag()}. Arguments are per that method. */
  public String makeNewTag(String pTagName, Hashtable<String, String> pAttributes, String pContent) { return makeTag(pTagName, pAttributes, pContent); }

  /** Instance form of {@link Tag#validString Tag.validString()}. Just because it's there. */
  public boolean isValidString(String pStr) { return validString(pStr); }

  /** Create and return a <code>SimpleStyleImpl</code> */
  public SimpleStyleImpl getNewStyleImpl() { return new SimpleStyleImpl(); }

  /** This method can change the tag name, just in case such a thing is needed. If the tag name sent in is null or blank, 'SimpleTag' is assigned. */
  public SimpleTagImpl setTagName(String pTagName) {
    myTagName = validString(pTagName)?pTagName:"SimpleTag";
    return this;
  }

  /** Fetch the content of the internal StringBuffer. If there is none, the return is null. */
  protected String _content() { return ((myContent.length() > 0)?myContent.toString():null); }

  /** Fetch the current content. No content is lost from this method. */
  public String getContent() { return myContent.toString(); }

  /** Set or clear the current tag contents. If 'pContent' is null or blank all content is cleared. If not, all previousl content is lost and the new content 
      is set.
  */
  protected void _setContent(String pContent) {
    if(validString(pContent)) {
      myContent = new StringBuffer(pContent);
    } else {
      myContent = new StringBuffer();
    }
  }

  /** Send a String to the <code>_setContent()</code> method. */
  public SimpleTagImpl setContent(String pContent) {
    setContent(pContent);
    return this;
  }

  /** <i>Add</i> content to what's currently there. Ignored if null or blank. */
  public SimpleTagImpl addContent(String pContent) {
    if(validString(pContent)) myContent.append(pContent);
    return this;
  }

  /** Set or remove an attribute value. */
  public SimpleTagImpl setAttribute(String pKey, String pValue) {
    _addAttribute(pKey, pValue);
    return this;
  }

  /** Return a copy of the current attributes. This is a <i>copy</i>, so changing it will not affect the original. */
  public Hashtable<String, String> getAttributes() { return _attributes(); }

}

