package cox.jmatt.java.MathTools.markup;

import java.util.*;
import cox.jmatt.java.MathTools.CapCom;

/** This class encapsulates CSS rules and 'style=' attributes. It contains an internal Hashtable for tracking this information along with methods to add, remove, 
    and clear it. It also has methods to format the information as a CSS-style rule or raw data for an XML attribute. For declarations whose value should be 
    numeric, the <code>_addNDeclaration()</code> methods offer numeric enforcement and optional value clamping.
*/
public class StyleInfo {

  private static long numHits = 0;
  private static String copyName = null;
  private static Hashtable<String, String> htbCopyDecl = null;

  private Hashtable<String, String> myDeclarations = null;
  private String myName = null;

  /** Create a blank tag, waiting for declarations to be set. */
  public StyleInfo() {
    super();
    myDeclarations = new Hashtable<String, String>();
    if(htbCopyDecl != null) myDeclarations.putAll(htbCopyDecl);
    if(Tag.validString(copyName)) myName = copyName;
    htbCopyDecl = null;
    copyName = null;
  }

  /** Create a named tag, if the name supplied is not null or blank. */
  public StyleInfo(String pRuleName) {
    super();
    myDeclarations = new Hashtable<String, String>();
    myName = pRuleName;
  }

  /** Use this method to generate a unique name rule for declaration groups. It will be 'Style_' followed by a number. This is used when the methods need a rule 
      name and none is set.
  */
  protected static final String _uniqueName() { return "Style_"+(numHits++); }

  /** Return a copy of the current declarations. This is a copy; it may be empty but will never be null. */
  protected final Hashtable<String, String> _declarations() {
    Hashtable<String, String> retval = new Hashtable<String, String>();
    if(!myDeclarations.isEmpty()) retval.putAll(myDeclarations);
    return retval;
  }

  /** Use this method to enable the various <code>copyOf()</code> methods. When called it stores all declarations in a temporary static variable; the next 
      instance automatically copies the declarations set for itself, then the temporary variable is cleared.
  */
  protected final void _enableCopy() {
    htbCopyDecl = _declarations();
    copyName = myName;
  }

  /** Add a declaration to the internal data. If the key is null or blank the method returns false. If the key and rule are both valid they are added. If the key 
      is valid and the rule is null or blank, a successful removal returns true and a non-removal returns false. This method is meant to be used as follows:<br />
      <code>addStyle('visible', 'false');</code>.
      @param pKey The declaration property name.
      @param pValue The declaration value.
  */
  protected final boolean _addDeclaration(String pKey, String pValue) {
    if(!Tag.validString(pKey)) return false;
    String retval = "Done";
    if(Tag.validString(pValue)) {
      myDeclarations.put(pKey, pValue);
    } else {
      retval = myDeclarations.remove(pKey);
    }
    return !Tag.validString(retval);
  }

  /** Add a possibly-clamped numeric declaration. The value is given as type <code>java.lang.Number</code> and the limits as <code>java.lang.Double</code>. If 
      either limit is ignored. If not, the value is replaced by the minimum if it is less than the minimum and the maximum if it is greater than the maximum. 
      CAUTION: This method does not check to see if the minimum supplied is in fact less than the maximum! Use with care. If the value is null it is removed. If 
      the key is null or blank the method returns false immediately.
      @param pKey The name of the declaration property.
      @param pValue The (numeric) value of the declaration or null to remove it.
      @param pMin The minimum value, ignored if null.
      @param pMax The maximum value, ignored if null.
  */
  protected final boolean _addNDeclaration(String pKey, Number pValue, Double pMin, Double pMax) {
    if(pValue == null) return _addDeclaration(pKey, null);
    Number newVal = pValue;
    if((pMin != null) && (pValue.doubleValue() < pMin.doubleValue())) newVal = pMin;
    if((pMax != null) && (pValue.doubleValue() > pMax.doubleValue())) newVal = pMax;
    return _addDeclaration(pKey, newVal.toString());
  }

  /** Convenience method for <code>_addNDeclaration(pKey, pValue, null, null)</code>. */
  protected final boolean _addNDeclaration(String pKey, Number pValue) { return _addNDeclaration(pKey, pValue, null, null); }

  /** Set the the (default) attribute name (attribute) or selector (css) text. Setting null clears it. */
  public void setName(String pName) { myName = pName; }

  /** This is a generic default configuration method for <code>StyleInfo</code> subclasses. This implementation is empty but is provided if needed. Any subclass 
      defining this method must document exactly what and how the information is passed (other than by the <code>Properties</code> object, which is given!). The 
      theory is that subclasses can pick through the keys, select the ones relevant to themselves and self-configure accordingly.
      @param pPizza The Properties object containing configuration information.
  */
  public void configurate(Properties pPizza) {}

  /** This method simply calls <code>configurate()</code> on the Global Pizza. */
  public void configurateGlobalPizza() { configurate(CapCom.getPizza()); }

  /** Clear all inernal data. */
  public void reset() {
    myName = null;
    myDeclarations = new Hashtable<String, String>();
  }

  /** Create and return a copy of this instance. */
  public StyleInfo copyOf() {
    _enableCopy();
    return new StyleInfo();
  }

  /** List all the declarations as a String of key-value pairs. Each individual declaration is terminated with a semicolon and they are space-separated. If no 
      declarations are set the return value is a blank but non-null String.
  */
  public String toString() {
    StringBuffer retval = new StringBuffer(" ");
    if(!((myDeclarations == null) || myDeclarations.isEmpty())) {
      for(String sKey : myDeclarations.keySet()) { retval.append(sKey).append("=").append(myDeclarations.get(sKey)).append("; "); }
    }
    return retval.toString().trim();
  }

  /** List all the declarations as a String of key-value pairs. Each individual declaration is formatted per CSS style rules and they are space-separated. If no 
      declarations are set the return value is null;
  */
  public String toRawCSS() {
    if((myDeclarations == null) || myDeclarations.isEmpty()) return null;
    StringBuffer retval = new StringBuffer(" ");
    for(String sKey : myDeclarations.keySet()) { retval.append(sKey).append(": ").append(myDeclarations.get(sKey)).append("; "); }
    return retval.toString().trim();
  }

  /** Return the declarations in a format appropriate for an XML attribute. This is basically the same as <code>toString()</code> but with quotes on both sides. 
      if the pName argument is not null or void it appears to the left, making the entire returned String usable as an attribute. If pName is null or blank no 
      name or equals sign appears.
      @param pName The attribute name to be output to the left of the rules String.
  */
  public String toAttribute(String pName) {
    StringBuffer retval = new StringBuffer();
    if(Tag.validString(pName)) retval.append(pName).append("=");
    retval.append("\"").append(toString()).append("\"");
    return retval.toString().trim();
  }

  /** This is the <code>toAttribute()</code> method with the internal name, if one is set. */
  public String toAttribute() { return toAttribute(myName); }

  /** Return the declarations in a format suitable for an internal or external CSS stylesheet. If the internal name is null or blank a default is assigned. The 
      format used is selector-style with the declarations enclosed in curly braces.
  */
  public String toCSS() {
    StringBuffer retval = new StringBuffer(Tag.validString(myName)?myName:_uniqueName());
    retval.append(" { ");
    if(!((myDeclarations == null) || myDeclarations.isEmpty())) {
      for(String sKey : myDeclarations.keySet()) { retval.append(sKey).append(": ").append(myDeclarations.get(sKey)).append("; "); }
    }
    retval.append(" }");
    return retval.toString();
  }

}

