package cox.jmatt.java.MathTools.markup;

import java.util.*;
import cox.jmatt.java.MathTools.CapCom;

/** <p>This class provides the basic machinery required to generate an XML tag. It tracks the tag name and attributes. No content is tracked but two methods are 
    provided for handling it: <code>_content() and _setcontent()</code>. The default implementations return null and do nothing, respectively. Subclasses can 
    override these methods if they have content to handle. The <code>toString()</code> method returns the data in String form, intelligently. If there is no content 
    the attributes are returned in an empty tag. If there is content the return value is an opening tag, content, and a closing tag. Attributes can be added in 
    an intelligent fashion; if the value is null or blank the value stored for that attribute key (if any) is removed. Numeric attributes have two add methods, 
    one that simply adds the number and the other that allows value-clamping. Unlike the standard methods these two ignore null numeric values and simply return 
    true.</p>
    <p>In addition to the instance methods <code>Tag</code> also contains factory methods to create an optionally-empty opening tag and closing tag. No XML 
    escaping is done so it must be handled manually.</p>
    <p>This class is meant to serve as the superclass of any XML tag. There are no external dependencies, DOM or otherwise, and the methods work by operating on 
    Strings. The internal data is handled automatically so only the tag name must be handled by the subclass. There is no public constructor and the only other 
    one is protected and absolutely requires a tag name be given. If necessary a copy of the attributes Hashtable can be returned but this is only a copy; the 
    original is never exposed directly.</p>
*/
public abstract class Tag {

  private Hashtable<String, String> myAttributes = new Hashtable<String, String>();

  /** The current tag name. This must NEVER be null and should not be changed without a very good reason! It is protected to allow such 'good reason' cases. */
  protected String myTagName = "Tag";

  private boolean enableXMLDeclTag = false;

  /** Constructor included for copy purposes. All data specified. */
  private Tag(String pName, Hashtable<String, String> pAtts, String pContent) {
    this(pName);
    if(!((pAtts == null) || pAtts.isEmpty())) {
      myAttributes = new Hashtable<String, String>();
      myAttributes.putAll(pAtts);
    }
    _setContent(pContent);
  }

  /** This constructor is used to create a named <code>Tag</code>. */
  protected Tag(String pName) {
    super();
    if(validString(pName)) myTagName = pName;
  }

  /** This method creates an opening XML tag with the name and attributes provided.
      @param pName The name of the opening tag; if null or blank it is 'Tag'.
      @param pAtts The attributes to add to the tag.
      @param isEmpty true if the tag should be self-closed, false otherwise.
  */
  public static final String openTag(String pName, Hashtable<String, String> pAtts, boolean isEmpty) {
    pName = validString(pName)?pName:"Tag";
    StringBuffer retval = new StringBuffer("<").append(pName).append(" ");
    if(!((pAtts == null) || pAtts.isEmpty())) {
      for(String sKey : pAtts.keySet()) { retval.append(sKey).append("=\"").append(pAtts.get(sKey)).append("\" "); }
    }
    retval.append(isEmpty?"/>":">");
    return retval.toString();
  }

  /** Close the tag specified. If the name is null or blank the closed tag is 'Tag'. */
  public static final String closeTag(String pName) { return ("</"+(validString(pName)?pName:"Tag")+">"); }

  /** Create a tag pair with all data supplied: name, attributes, and content. If the content is blank or null an empty tag is returned.
      @param pName The name of the tag to create.
      @param pAtts The attributes for the opening tag.
      @param pContent The String content of the tag.
  */
  public static final String makeTag(String pName, Hashtable<String, String> pAtts, String pContent) {
    String retval = openTag(pName, pAtts, !validString(pContent));
    if(validString(pContent)) retval += pContent + closeTag(pName);
    return retval;
  }

  /** This is identical to {@link cox.jmatt.java.MathTools.Question#validString Question.validString()} but is repeated here for convenience. */
  public static final boolean validString(String pStr) { return ((pStr != null) && (pStr.trim().length() > 0)); }

  /** <p>Add an attribute to the current tag, or remove one. If the key is null or blank the method returns false immediately. If the key is valid but the value 
      is null or blank the value stored under that key is removed; if an attribute was removed the method returns true, if the mapping didn't exist it returns 
      false. If both the key and value are valid the attribute is added.</p>
      <p>This method is protected to prevent automatic addition of arbitrary attributes.</p>
      @param pKey The name of the attribute to add.
      @param pValue The value of the attribute, or null to remove it.
  */
  protected final boolean _addAttribute(String pKey, String pValue) {
    if(!validString(pKey)) return false;
    String retval = null;
    if(validString(pValue)) {
      retval = "Added";
      myAttributes.put(pKey, pValue);
    } else {
      retval = myAttributes.remove(pKey);
    }
    return validString(retval);
  }

  /** Add an attribute whose value is a <code>java.lang.Number</code>. Unlike the <code>_addAttribute()</code> method this one ignores null values and returns 
      <b>true</b> when one is set. A null or blank key still returns false.
  */
  protected final boolean _addNAttribute(String pKey, Number pVal) {
    if(pVal == null) return true;
    return _addAttribute(pKey, pVal.toString());
  }

  /** Add a 'clamped' attribute; one whose value is a <code>java.lang.Number</code> and whose value must fall between a maximum and minumum, both expressed 
      as <code>Double</code>. If the max or min is null it is ignored. For non-null limits, if the Number is greater than the maximum or less than the minimum 
      it is replaced by that limit. NOTE: This method does <b>not</b> check to see if the minimum is in fact less than the maximum!
      @param pKey The name of the attribute to set. If null, the method returns false.
      @param pValue The numeric value to be clamped.
      @param pMin The minimum value allowed, ignored if null.
      @param pMax The maximum value, ignored if null.
      @param bRemoveIfNull true to remove attributes with null values, false to ignore and return true.
  */  
  protected final boolean _clampNAttribute(String pKey, Number pValue, Double pMin, Double pMax, boolean bRemoveIfNull) {
    if(pValue == null) {
      if(bRemoveIfNull) {
        return _addAttribute(pKey, null);
      } else {
        return true;
      }
    }
    Number newVal = pValue;
    if((pMin != null) && (pValue.doubleValue() < pMin.doubleValue())) newVal = pMin;
    if((pMax != null) && (pValue.doubleValue() > pMax.doubleValue())) newVal = pMax;
    return _addNAttribute(pKey, newVal);
  }

  /** Convenience method for <code>_clampNAttribute(pKey, pValue, pMin, pMax, false)</code>. */
  protected final boolean _clampNAttribute(String pKey, Number pValue, Double pMin, Double pMax) { return _clampNAttribute(pKey, pValue, pMin, pMax, false); }

  /** Return a copy of the tag's attributes. The original is left untouched. If no attributes are set an empty but non-null Hashtable is returned. */
  public final Hashtable<String, String> _attributes() {
    Hashtable<String, String> retval = new Hashtable<String, String>();
    if(!myAttributes.isEmpty()) retval.putAll(myAttributes);
    return retval;
  }

  /** Add a Hashtable full of attributes. Ignored if null or empty. */
  protected final void _addAttributes(Hashtable<String, String> pNewAtts) {
    if(pNewAtts != null) myAttributes.putAll(pNewAtts);
  }

  /** This method is meant to provide access to the tag's content, if any. The default implementation returns null; subclasses that may contain content should 
      override this method to provide appropriate functionality.
  */
  protected String _content() { return null; }

  /** If the <code>_content()</code> method is overriden this one should be as well. It provides a method to set content. This implementation is empty. Subclasses 
      defining this method must be able to handle a null argument since this is how the <code>reset()</code> method clears it.
  */
  protected void _setContent(String pContent) {}

  /** Clear all data except the tag name. Any attributes will be lost. This method does call <code>_setContent(null)</code> and reset the XML declaration to false.*/
  public void reset() {
    _setContent(null);
    enableXMLDeclTag = false;
    myAttributes =new Hashtable<String, String>();
  }

  /** <p>This is a generic default-configuration method. The implementation here is empty but it is provided for subclasses.</p>
      <p>The theory behind the method is that a <code>Tag</code> subclass can search through the <code>Properties</code> object and select information relevant 
      to itself. No input validation or other checking is done, that is the responsibility of the subclass. Subclasses defining this method must supply ALL 
      relevant details and documentation!</p>
      <p>The prefix is supplied to allow multiple configurations to be stored in a single file. If used, pizza-configurable tags will prepend the prefix to all 
      configuration keys. Overridden methods should check to see that a valid (non-null, non-blank) prefix ends with a dot character ('.') and if not, supply 
      one automatically.</p>
      @param pConfigData A Properties object holding configuration data for the subclass instance.
  */
  public void configureTag(String pPrefix, Properties pConfigData) {}

  /** Configure the tag from the Global Pizza. The prefix is passed through unaltered. */
  public void configureTagFromGlobalPizza(String pPrefix) {
    configureTag(pPrefix, CapCom.getPizza());
  }

  /** <p>This is a convenience method to format prefixes and keys. If the prefix is null or blank it is ignored, otherwise it is checked to see if it ends with 
      a dot. If not one is attached, the key is appended and the entire thing returned. If the key value is null or blank a blank String is returned.</p>
      @param pPrefix The prefix to attach to the key or ignore if null or blank.
      @param pKey The value of the key.
  */
  protected String _fixKey(String pPrefix, String pKey) {
    if(!validString(pKey)) return "";
    if(validString(pPrefix)) {
      if(!pPrefix.endsWith(">")) pPrefix += ".";
    } else {
      pPrefix = "";
    }
    return (pPrefix+pKey);
  }

  /** <p>Convenience method to extract and check a boolean key. If the (case-insensitive) value of the prefix-key combination is 'true' the method returns true, 
      otherwise it returns false.</p>
      @param pPizza The pizza to check. If null or empty the method returns false.
      @param pPrefix The key prefix.
      @param pKey The value of the key to check.
  */
  protected boolean _asBoolean(Properties pPizza, String pPrefix, String pKey) {
    if((pPizza == null) || pPizza.isEmpty()) return false;
    return ("true".equalsIgnoreCase(pPizza.getProperty(_fixKey(pPrefix, pKey), "false")));
  }

  /** <p>Convenience method to extract and return a Double (object!) from a parameter. ANY errors are silently ignored. If the key requested is not present or 
      cannot be parsed as a double the default value is returned.</p>
  */
  protected Double _asDouble(Properties pPizza, String pPrefix, String pKey, Double pDefault) {
    Double retval = pDefault;
    try {
      retval = Double.valueOf(pPizza.getProperty(pPrefix, pKey));
    } catch(Exception ignored) {}
    return retval;
  }

  /** <p>Extract and return an Integer-object or a default if it is not found. */
  protected Integer _asInt(Properties pPizza, String pPrefix, String pKey, Integer pDefault) {
    Integer retval = pDefault;
    try {
      retval = Integer.valueOf(pPizza.getProperty(pPrefix, pKey));
    } catch(Exception ignored) {}
    return retval;
  }

  /** Return 'false' if the <code>Properties</code> objet is null or empty. Laziness method. */
  public static boolean badPizza(Properties pPizza) { return ((pPizza == null) || pPizza.isEmpty()); }

  /** Set the 'id=' attribute of the tag. Included in this class since it is so common among XML tags. Setting null removes it. */
  public void setID(String pID) { _addAttribute("id", pID); }

  /** Use this method to enable or disable (default) the '&lt;?xml ... ?&gt;' declaration. If enabled it prints before the tag opens. If this is enabled it only 
      lasts for ONE <code>toString()</code> call, afterward it resets to false and must be re-enabled.
  */
  protected final void _setEnableXMLTag(boolean pEnable) { enableXMLDeclTag = pEnable; }

  /** Generate a tag or tag pair from the internal data. If there is content it is printed, enclosed by opening and closing tags. If the XML declaration tag is 
      enabled it appears before the opening tag and calling this method resets it to false.
  */
  public String toString() {
    String retval = (enableXMLDeclTag)?"<?xml version=\"1.0\" ?>\n":"\n";
    enableXMLDeclTag = false;
    return (retval + makeTag(myTagName, myAttributes, _content()));
  }

}

