package cox.jmatt.java.MathTools.markup.html;

import java.util.Hashtable;

/** <p>This is the abstract superclass for all HTML tag classes. Since HTML's structure is incredibly flexible (and possibly convoluted!) there are no 'subclass 
    chains' involved with this package. Instead each tag class implements the behavior that is absolutely required for that tag. Proper structure is up to the 
    user! The only intelligence applied innately is whether the tag should be added to a document's &lt;head&gt; or &lt;body&gt; and the <code>TagKeys enum</code> 
    handles that.</p>
    <p>As with the SVG package, HTML uses CSS (only!) to provide styling and presentation information. The tacit assumption is that an external stylesheet is 
    provided so nothing analogous to the <code>StylePen</code> class is used here.</p>
    <p>Since most of the tags represented in this package may have content rudimentary content handling is supplied. Content is held in an internal buffer until 
    needed and returned as a String when that happens. No XML-escaping is done so that must be handled manually!</p>
*/
public abstract class BaseHTag extends cox.jmatt.java.MathTools.markup.Tag {

  /** This <code>enum</code> holds information and default values for the various tag classes, all of which subclass <code>BaseHTag</code>. */
  public static enum TagKeys {
    /** This is not an HTML tag, it is used to set values for all enum constants that are. */
    GLOBAL("gt", false),
    EM_TAG("em", true),
    B_TAG("b", true),
    I_TAG("i", true),
    U_TAG("u", true),
    A_TAG("a", true),
    BQ_TAG("blockquote", true),
    SPAN_TAG("span", true),
    IMG_TAG("img", true),
    LI_TAG("li", false),
    DD_TAG("dd", false),
    DT_TAG("dt", false),
    TD_TAG("td", false),
    OL_TAG("ol", true),
    UL_TAG("ul", true),
    DL_TAG("dl", true),
    TR_TAG("tr", false),
    TABLE_TAG("table", true),
    HTML_TAG("html", false),
    DIV_TAG("div", true),
    P_TAG("p", true),
    Hn_TAG("h1", true),
    HR_TAG("hr", true);

    private String myName = null, copyCont = null, defaultName = null, defaultCSS = null, defaultClick = null, defaultStyle = null;
    private boolean inBody = true;
    private Hashtable<String, String> copyAtts = null;
    TagKeys(String pName, boolean pInBody) {
      this.myName = pName;
      this.inBody = pInBody;
    }
    /** The name of the tag as it will appear in the markup. */
    public String getName() { return myName; }
    /** Is <i>this</i> tag in the &lt;body&gt; or the &lt;head&gt; of the document? */
    public boolean isInBody() { return inBody; }
    /** Store tag attributes and content for copying. */
    public void storeCopyData(Hashtable<String, String> pAtts, String pContent) {
      this.copyAtts = pAtts; this.copyCont = pContent;
    }
    /** Retrieve the attributes stored for copying, once. */
    public Hashtable<String, String> getCopyAttributes() {
      Hashtable<String, String> retval = copyAtts;
      copyAtts = null;
      return retval;
    }
    /** Retrieve the body content stored for copying. Cleared after this method is called. */
    public String getCopyContent() {
      String retval = copyCont;
      copyCont = null;
      return retval;
    }
    /** Set the default value of the 'name=' attribute for this tag. */
    public void setDefaultName(String pStr) { defaultName = pStr; }
    /** Retrieve the default value of the 'name=' attribute. */
    public String getDefaultName() { return defaultName; }
    /** Set the default CSS class attribute for this tag. */
    public void setDefaultCSS(String pStr) { defaultCSS = pStr; }
    /** Retrieve the default value of the CSS class attribute. */
    public String getDefaultCSS() { return defaultCSS; }
    /** Set the default CSS style attribute for this tag. */
    public void setDefaultStyle(String pStr) { defaultStyle = pStr; }
    /** Retrieve the default value of the CSS style attribute. */
    public String getDefaultStyle() { return defaultStyle; }
    /** Set the default value of the 'onclick=' attribute for this tag. */
    public void setDefaultClick(String pStr) { defaultClick = pStr; }
    /** Retrieve the default value of the 'onclick=' attribute. */
    public String getDefaultClick() { return defaultClick; }
  }
  /* * * END OF ENUM * * */

  private TagKeys myKey = TagKeys.P_TAG;
  private StringBuffer myContent = null;

  /** Standard constructor for building HTML tag classes. If a copy is created its 'id=' attribute is cleared. */
  protected BaseHTag(TagKeys pKey) {
    super(pKey.getName());
    myKey = pKey;
    _addAttributes(pKey.getCopyAttributes());
    _addAttribute("id", null);
    _setContent(pKey.getCopyContent());
    setName(pKey.getDefaultName());
    setCSS(pKey.getDefaultCSS());
    setStyle(pKey.getDefaultStyle());
    setClicked(pKey.getDefaultClick());
  }

  /** Determine where to add this tag to the HTML document: true for body, false for head. Return value based on its <code>TagKeys</code> constant. */
  public boolean isInBody() { return myKey.isInBody(); }

  /** This method enables tag copying. All attributes and content are stored temporarily and the next tag instance created retrieves and clears them. */
  protected final void _enableCopy() { myKey.storeCopyData(_attributes(), _content()); }

  /** Return the String name of this tag's <code>TagKeys</code> constant. */
  public String getTagKey() { return myKey.toString(); }

  /** Return the current tag name. */
  public String getTagName() { return myTagName; }

  /** Fetch a copy of the tag's ID if one is set. If not a blank String is returned. */
  public String getID() {
    String retval = _attributes().get("id");
    return validString(retval)?retval:"";
  }

  /** This method sets the tag's 'name=' <b>attribute</b> and not the name of the tag itself! Since is is common among HTML tags it is handled here. Setting null 
      or blank clears the name.
  */
  public void setName(String pName) { _addAttribute("name", pName); }

  /** Set or clear the (CSS) 'class=' attribute for the tag. Null or blank clears it, if it was set. */
  public void setCSS(String pRuleName) { _addAttribute("class", pRuleName); }

  /** Set or clear the tag's (inline CSS) 'style=' attribute. Null or blank clears any previous value. */
  public void setStyle(String pStyle) { _addAttribute("style", pStyle); }

  /** This method is provided for (external!!) script interactions. For tags that don't allow it, DON'T USE IT!! The attribute set is 'onclick='. */
  public void setClicked(String pClickHandler) { _addAttribute("onclick", pClickHandler); }

  /** Fetch the content of the tag, unaltered and unedited. This returns ONLY the content tracked innately by this class; other classes that override the 
      <code>_content()</code> method may not dump their content via this method. Use with care!
  */
  public String getRawContent() { return myContent.toString(); }

  /** Return the contents of the internal <code>StringBuffer</code>. May be blank but will never be null. */
  protected String _content() { return myContent.toString(); }

  /** Handle the internal buffer. If pContent is null all data is cleared. If it is non-null it becomes the new content. Either way, all previous data is LOST. */
  protected void _setContent(String pContent) { myContent = new StringBuffer((validString(pContent)?pContent:"")); }

  /** Add content to the current tag. Ignored if null or blank, should not be intermixed with CDATA content. */
  public void addContent(String pContent) {
    if(validString(pContent)) myContent.append(pContent);
  }

  /** Add non-null, non-blank content enclosed in a CDATA block. This method should not be intermixed with <code>addContent()</code>. */
  public void addCDATA(String pData) {
    if(validString(pData)) myContent.append("<![CDATA\n").append(pData).append("\n]]>\n");
  }

  /** Add a tag to the content. Ignored if null. Protected to allow exclusion from tags that don't contain arbitrary other HTML tags. */
  protected void _addTag(BaseHTag pTag) {
    if(pTag != null) addContent(pTag.toString());
  }

}

