package cox.jmatt.java.MathTools.markup.html;

/** <p>This is the abstract superclass of tag classes meant to contain other tags. It provides a basic tag-adding method for any tag that can be added to the 
    &lt;body&gt; of a document although it doesn't provide that particular tag. If a &lt;head&gt;-specific tag is added it is ignored.</p>
    <p>Also provided are 'quick-add' methods designed to add simple tag-enclosed text. These methods do create tag classes so they do inherit any default values 
    for such. For more complexity use the tag classes or define a CSS rule to affect unadorned or un-CSS-class'ed tags.</p>
*/
public abstract class ContainerTag extends BaseHTag {

  /** Standard constructor for building HTML tag classes. If a copy is created its 'id=' attribute is cleared. */
  protected ContainerTag(TagKeys pKey) { super(pKey); }

  /** Quickly add <i>italicized</i> text to the current tag. Ignored if content is null or blank. */
  public void addItalic(String pContent) {
    if(validString(pContent)) _addTag(new ITag(pContent));
  }

  /** Quickly add <b>bold</b> text to the current tag. Ignored if content is null or blank. */
  public void addBold(String pContent) {
    if(validString(pContent)) _addTag(new BTag(pContent));
  }

  /** Quickly add <u>underlined</u> text to the current tag. Ignored if content is null or blank. */
  public void addUnderline(String pContent) {
    if(validString(pContent)) _addTag(new UTag(pContent));
  }

  /** Quickly add <em>emphasized</em> text to the current tag. Ignored if content is null or blank. */
  public void addEM(String pContent) {
    if(validString(pContent)) _addTag(new EMTag(pContent));
  }

  /** Quickly add a heading tag. Ignored if content is null or blank or the level is not between 1 and 6 inclusive. */
  public void addHeading(byte pLevel, String pContent) {
    if(validString(pContent) && (pLevel >= 1) && (pLevel <= 6)) _addTag(new HnTag(pLevel, pContent));
  }

  /** Add a horizontal rule ('&lt;hr&gt;'). This is the default flavor, no width specified. */
  public void addHR() { _addTag(new HRTag()); }

  /** Quickly add block-quote text to the current tag. Ignored if content is null or blank. */
  public void addBQ(String pContent) {
    if(validString(pContent)) _addTag(new BQTag(pContent));
  }

  /** Quickly add an image to the current tag. Ignored if pSrc is null or blank.
      @param pSrc The 'src=' attribute (image source).
  */
  public void addImage(String pSrc) {
    if(validString(pSrc)) _addTag(IMGTag.newInstance(pSrc));
  }

  /** Quickly add a link('&lt;a&gt;' tag) to the current tag. Ignored if pHREF and pContent are null or blank. */
  public void addAnchor(String pHREF, String pContent) { _addTag(AnchorTag.newInstance(pHREF, pContent)); }

  /** Add a &lt;br /&gt; tag. A newline is appended afterward! */
  public void addBreak() { addContent("<br />"); }

}

