package cox.jmatt.java.MathTools.markup.html;

/** This class provides the &lt;img&gt; tag. Since a valid source is mandatory these are created by a factory method that returns null if a valid 'src=' attribute 
    is not supplied.
*/
public class IMGTag extends BaseHTag {

  /** Construct a tag instance. Source is mandatory, hence the constructor is protected. */
  protected IMGTag(String pSrc) {
    super(BaseHTag.TagKeys.IMG_TAG);
    _addAttribute("src", pSrc);
    _addAttribute("alt", "Image");
  }

  /** Create an &lt;img&gt; tag instance. Since this tag does not make sense without the 'src=' attribute, one must be supplied. If not, the return value is 
      null.
  */
  public static final IMGTag newInstance(String pSrc) {
    IMGTag retval = null;
    if(validString(pSrc)) retval = new IMGTag(pSrc);
    return retval;
  }

  /** Create an &lt;img&gt; tag with source and size specified. If no source is specified the return value is null. */
  public static final IMGTag newInstance(String pSrc, double pWidth, double pHeight) {
    IMGTag retval = null;
    if(validString(pSrc)) retval = new IMGTag(pSrc);
    if(retval != null) retval.setSize(pWidth, pHeight);
    return retval;
  }

  /** Copy the current tag. */
  public IMGTag copyOf() {
    _enableCopy();
    return new IMGTag(_content());
  }

  /** Overridden to preserve the 'src=' attribute. */
  public void reset() {
    String mySrc = _attributes().get("src");
    super.reset();
    _addAttribute("src", mySrc);
  }

  /** <p>Set the width and height of the image. If null an argument is <b>ignored</b>, otherwise the arguments are interpreted specially. If the value is between 
      0.0 (inclusive) and 1.0 (exclusive) it is interpreted as a percent and a '%' is appended. If greater than or equal to 1.0 it is interpreted as a pixel 
      width and is set as-is. Negative values are made positive before processing.</p>
      @param pWidth The width of the image. Ignored if null.
      @param pHeight The height of the image. Ignored if null.
  */
  public void setSize(Double pWidth, Double pHeight) {
    double dTemp = 0;
    String sPercent = "";
    if(pWidth != null) {
      dTemp = Math.abs(pWidth.doubleValue());
      if(dTemp < 1.0) {
        sPercent = "%";
        dTemp *= 100.0;
      }
      _addAttribute("width", ""+((int) dTemp)+sPercent);
    }
    if(pHeight != null) {
      dTemp = Math.abs(pHeight.doubleValue());
      if(dTemp < 1.0) {
        sPercent = "%";
        dTemp *= 100.0;
      }
      _addAttribute("height", ""+((int) dTemp)+sPercent);
    }
  }

  /** Set or clear the alternate text for the image. The default is 'Image'. */
  public void setAlt(String pAltText) { _addAttribute("alt", validString(pAltText)?pAltText:"Image"); }

}

