package cox.jmatt.java.MathTools.markup.html;

/** <p>This class provides the &lt;ul&gt; and &lt;ol&gt; tags. Ordered or unordered is determined by the constructor, otherwise the two are identical. The only 
    content permitted is &lt;li&gt; tags (<code>LITag</code> class) so that is all that can be added. There is a method to add an item directly (defaults apply) 
    and another to add a pre-constructed instance.</p>
    <p>NOTE: Both types of list tags can contain sublists. To do this construct a separate <code>ListTag</code> for the sublist, add it to a &lt;li&gt; tag, and 
    add that tag to this one. The <code>addList()</code> method here also provides that function provided the &lt;li&gt; holding the sublist can function with 
    default settings.</p>
*/
public class ListTag extends BaseHTag {

  /** Standard constructor for an empty list tag.
      @param isOrderedList true for an &lt;ol&gt; tag, false for &lt;ul&gt;.
  */
  public ListTag(boolean isOrderedList) { super(isOrderedList?BaseHTag.TagKeys.OL_TAG:BaseHTag.TagKeys.UL_TAG); }

  /** <p>Create a list, ordered or unordered, from an Object[] array. If the array sent in is null or empty the return is null. If an element is null it is 
      skipped. The item ID prefix, if defined, is used to generate an ID for each item. The prefix is prepended to the item count number; the item number <i>not</i> 
      counting any skipped items. Non-null Object entries are <code>toString()</code>'ed for the item value.</p>
      @param pList The list item array.
      @param pItemID The item ID prefix. Ignored if null.
      @param bIsOrdered true for an ordered list, false for unordered.
  */
  public static final ListTag createList(boolean bIsOrdered, Object[] pList, String pItemID) {
    if((pList == null) || (pList.length < 1)) return null;
    ListTag retval = new ListTag(bIsOrdered);
    boolean gotID = validString(pItemID);
    int itemCount = 0;
    LITag liTemp = null;
    for(int i=0; i < pList.length; i++) {
      if(pList[i] == null) continue;
      liTemp = new LITag(pList[i].toString());
      if(gotID) liTemp.setID(pItemID+itemCount);
      itemCount++;
      retval.addItem(liTemp);
    }
    return retval;
  }

  /** Instance version of <code>createList()</code>. */
  public ListTag newList(boolean bOrdered, Object[] pList, String pID) { return createList(bOrdered, pList, pID); }

  /** Add an item directly to the tag and return a self-reference. This does create a &lt;li&gt; tag internally. Ignored if null or blank. */
  public ListTag addItem(String pItemText) {
    if(validString(pItemText)) addItem(new LITag(pItemText));
    return this;
  }

  /** Add a pre-constructed <code>LITag</code>. Ignored if null. */
  public ListTag addItem(LITag pTag) {
    _addTag(pTag);
    return this;
  }

  /** Add a sub-list to the current list. It is automatically wrapped in a default &lt;li&gt; tag. A self-reference is returned and the list is ignored if null. */
  public ListTag addList(ListTag pSubList) {
    if(pSubList != null) {
      LITag liTemp = new LITag();
      liTemp.addList(pSubList);
      _addTag(liTemp);
    }
    return this;
  }

  /** Return a copy of this tag. */
  public ListTag copyOf() {
    _enableCopy();
    return new ListTag("ol".equals(myTagName));
  }

  /** Set the 'type=' attribute of an <i>ordered</i> list. This has no effect on an unordered list. The type parameter is set as char and equals one of the 
      defined values: '1, A, a, I, i'. Case <b>does</b> matter and any value other than one of these clears the attribute.
  */
  public ListTag setType(char pType) {
    if(!("ol".equals(myTagName))) return this;
    String sType = null;
    switch(pType) {
      case '1': case 'A': case 'a': case 'I': case 'i':
        sType = "" + pType;
        break;
      default:
    }
    _addAttribute("type", sType);
    return this;
  }

  /** Set the starting number for an ordered list. Ignored for unordered lists and no value-clamping is performed. */
  public ListTag setStart(int pStart) {
    if(!("ol".equals(myTagName))) return this;
    _addAttribute("start", ""+pStart);
    return this;
  }

}

