package cox.jmatt.java.MathTools.markup.html;

import cox.jmatt.java.MathTools.util.NoteworthyParser;
import cox.jmatt.java.MathTools.markup.Tag;

/** <p>This class is to the HTML package what <code>MathDrawSVG</code> is to the SVG package. It serves to generate instances of the various tag classes and to 
    set default values. <b>Critical Note</b>: Setting the default value for a multi-purpose tag (eg <code>ListTag</code>) affects <b>all</b> instances of that 
    <i>class</i>, regardless of the <i>type</i> of tag generated. Use with caution!</p>
    <p>Concerning default values, they are set differently here than in the SVG classes. By the nature of HTML it might be desirable to set default values for 
    a range of tags, or even <i>all</i> tags. This is made easy via the <code>TagKeys.GLOBAL</code> constant. If passed into a <code>setDefault__()</code> method 
    it sets the default for <b>all</b> tags. To exclude one or two tag types re-set individuals manually.</p>
*/
public class MathGenHTML {

  /** Scripting constructor; used to supply <code>markup.html.*</code> classes to scripts. */
  public MathGenHTML() { super(); }

  /** Set or clear the default value of a certain tag's 'name=' attribute. The pTag parameter must be one of the <code>TagKeys enum</code> constants. If the tag 
      key sent in is lower case it is converted but if null or blank it is ignored. If the constant sent in doesn't match one of the constants it is ignored.
      @param pTag The name of one of the <code>TagKeyss</code> constants.
      @param pValue The default 'name=' attribute value to be assigned.
  */
  public void setDefaultName(String pTag, String pValue) {
    if(!Tag.validString(pTag)) return;
    try {
      BaseHTag.TagKeys theKey = BaseHTag.TagKeys.valueOf(pTag.toUpperCase());
      if(theKey == BaseHTag.TagKeys.GLOBAL) {
        for(BaseHTag.TagKeys tKey : BaseHTag.TagKeys.values()) tKey.setDefaultName(pValue);
      } else {
        theKey.setDefaultName(pValue);
      }
    } catch(Exception ignored) {}
  }

  /** Set or clear the default CSS class name for a certain tag. All parameters and concerns are per <code>setDefaultName()</code>. */
  public void setDefaultCSS(String pTag, String pValue) {
    if(!Tag.validString(pTag)) return;
    try {
      BaseHTag.TagKeys theKey = BaseHTag.TagKeys.valueOf(pTag.toUpperCase());
      if(theKey == BaseHTag.TagKeys.GLOBAL) {
        for(BaseHTag.TagKeys tKey : BaseHTag.TagKeys.values()) tKey.setDefaultCSS(pValue);
      } else {
        theKey.setDefaultCSS(pValue);
      }
    } catch(Exception ignored) {}
  }

  /** Set or clear the default inline CSS style for a certain tag. All parameters and concerns are per <code>setDefaultName()</code>. */
  public void setDefaultStyle(String pTag, String pValue) {
    if(!Tag.validString(pTag)) return;
    try {
      BaseHTag.TagKeys theKey = BaseHTag.TagKeys.valueOf(pTag.toUpperCase());
      if(theKey == BaseHTag.TagKeys.GLOBAL) {
        for(BaseHTag.TagKeys tKey : BaseHTag.TagKeys.values()) tKey.setDefaultStyle(pValue);
      } else {
        theKey.setDefaultStyle(pValue);
      }
    } catch(Exception ignored) {}
  }

  /** Set or clear the default 'onclick=' for a certain tag. All parameters and concerns are per <code>setDefaultName()</code>. */
  public void setDefaultClicked(String pTag, String pValue) {
    if(!Tag.validString(pTag)) return;
    try {
      BaseHTag.TagKeys theKey = BaseHTag.TagKeys.valueOf(pTag.toUpperCase());
      if(theKey == BaseHTag.TagKeys.GLOBAL) {
        for(BaseHTag.TagKeys tKey : BaseHTag.TagKeys.values()) tKey.setDefaultClick(pValue);
      } else {
        theKey.setDefaultClick(pValue);
      }
    } catch(Exception ignored) {}
  }

  /** Create a <code>NoteworthyParser</code>. */
  public NoteworthyParser getNoteParser() { return new NoteworthyParser(); }

  /** Create an empty <code>BTag</code>. */
  public BTag getBTag() { return new BTag(); }
  /** Create a <code>BTag</code> with content. */
  public BTag getBTag(String pContent) { return new BTag(pContent); }

  /** Create an empty <code>ITag</code>. */
  public ITag getITag() { return new ITag(); }
  /** Create a <code>ITag</code> with content. */
  public ITag getITag(String pContent) { return new ITag(pContent); }

  /** Create an empty <code>UTag</code>. */
  public UTag getUTag() { return new UTag(); }
  /** Create a <code>UTag</code> with content. */
  public UTag getUTag(String pContent) { return new UTag(pContent); }

  /** Create an empty <code>EMTag</code>. */
  public EMTag getEMTag() { return new EMTag(); }
  /** Create a <code>EMTag</code> with content. */
  public EMTag getEMTag(String pContent) { return new EMTag(pContent); }

  /** Create an empty &lt;h1&gt; tag. */
  public HnTag getHeadingTag() { return new HnTag(); }

  /** Create a heading tag with content. If the level is not between 1 and 6 a &lt;h1&gt; tag is created. */
  public HnTag getHeadingTag(byte pLevel, String pContent) { return new HnTag(pLevel, pContent); }

  /** Create a standard &lt;hr&gt; tag, no width specified. */
  public HRTag getHRTag() { return new HRTag(); }

  /** Create a standard &lt;hr&gt; tag with the width specified. */
  public HRTag getHRTag(double pWidth) { return new HRTag(pWidth); }

  /** Create an empty <code>BQTag</code> for a &lt;blockquote&gt; tag. */
  public BQTag getBQTag() { return new BQTag(); }
  /** Create a <code>BQTag</code> with content. */
  public BQTag getBQTag(String pContent) { return new BQTag(pContent); }

  /** Create an empty <code>SpanTag</code>. */
  public SpanTag getSpanTag() { return new SpanTag(); }
  /** Create a <code>SpanTag</code> with content. */
  public SpanTag getSpanTag(String pContent) { return new SpanTag(pContent); }

  /** Create an image tag. If the source is null or blank the return value is null. */
  public IMGTag getImageTag(String pSrc) { return IMGTag.newInstance(pSrc); }

  /** Create an image tag with source and size specified. If the source is null or blank the return value is null. */
  public IMGTag getImageTag(String pSrc, double pWidth, double pHeight) { return IMGTag.newInstance(pSrc, pWidth, pHeight); }

  /** Create a new <b>ordered</b> list. */
  public ListTag getOLTag() { return new ListTag(true); }

  /** Create a new <b>unordered</b> list. */
  public ListTag getULTag() { return new ListTag(false); }

  /** Create a list (ordered or unordered) from an Object[] array. See {@link cox.jmatt.java.MathTools.markup.html.ListTag#createList here} for details and 
      parameter info.
  */
  public ListTag createList(boolean bIsOrdered, Object[] itemList, String pItemIDPrefix) { return ListTag.createList(bIsOrdered, itemList, pItemIDPrefix); }

  /** Create an empty <code>LITag</code>. */
  public LITag getLITag() { return new LITag(); }
  /** Create a <code>LITag</code> with content. */
  public LITag getLITag(String pContent) { return new LITag(pContent); }

  /** Create a fresh &lt;dl&gt; tag. */
  public DLTag getDLTag() { return new DLTag(); }

  /** Create an empty <code>DTTag</code>. */
  public DTTag getDTTag() { return new DTTag(); }
  /** Create a <code>DTTag</code> with content. */
  public DTTag getDTTag(String pContent) { return new DTTag(pContent); }

  /** Create an empty <code>DDTag</code>. */
  public DDTag getDDTag() { return new DDTag(); }
  /** Create a <code>DDTag</code> with content. */
  public DDTag getDDTag(String pContent) { return new DDTag(pContent); }

  /** Create an empty <code>TDTag</code>. */
  public TDTag getTDTag() { return new TDTag(); }
  /** Create a <code>TDTag</code> with content. */
  public TDTag getTDTag(String pContent) { return new TDTag(pContent); }

  /** Create an empty &lt;tr&gt; tag. */
  public TRTag getTRTag() { return new TRTag(); }

  /** Create a table row (tag) around an Object[] array. Details are {@link cox.jmatt.java.MathTools.markup.html.TRTag#createRow here}.
      @param pRow The Object[] array to turn into a row.
      @param pCellID The cell ID prefix. May be null.
      @param bIsTH 'true' for a heading row, false for a standard row.
  */
  public TRTag createTRTag(Object[] pRow, String pCellID, boolean bIsTH) { return TRTag.createRow(pRow, pCellID, bIsTH); }

  /** Create a standard table row with no cell ID prefix. */
  public TRTag createTRTag(Object[] pRow) { return TRTag.createRow(pRow, null, false); }

  /** Create a table heading row with no cell ID prefix. */
  public TRTag createTHRow(Object[] pRow) { return TRTag.createRow(pRow, null, true); }

  /** Create an empty <code>TableTag</code> instance. */
  public TableTag getTableTag() { return new TableTag(); }

  /** Create a complete &lt;table&gt; around a two-dimensional Object[][] array. Parameters are {@link cox.jmatt.java.MathTools.markup.html.TableTag#createTable
      here}.
  */
  public TableTag createTable(Object[][] pTable, String rowID, String cellID, boolean bUseRowID, boolean bFirstRowTH) {
    return TableTag.createTable(pTable, rowID, cellID, bUseRowID, bFirstRowTH);
  }

  /** Create a basic &lt;p&gt; tag. */
  public PTag getPTag() { return new PTag(); }

  /** Create a basic &lt;div&gt; tag. */
  public DivTag getDivTag() { return new DivTag(); }

  /** Create an &lt;html&gt; document tag. */
  public HTMLTag getHTML() { return new HTMLTag(); }

  /** Create and return an &lt;a&gt; tag. Both arguments are mandatory.
      @param pHREF The 'href=' attribute value for the anchor.
      @param pText The text to appear within the tags.
  */
  public AnchorTag getAnchorTag(String pHREF, String pText) { return AnchorTag.newInstance(pHREF, pText); }

}

