package cox.jmatt.java.MathTools.markup.latex;

import java.util.*;
import cox.jmatt.java.MathTools.*;

/** <p>This is the abstract superclass of all other LaTeX markup-generating classes. It is primarily geared for block-style tags therefore it contains the necessary 
    equipment to generate both the initial tag and the end tag.</p>
    <p>A quick perusal of the methods here shows more than a few concerning whether or not an '/end...' tag exists, is created or is printed. This was done to 
    allow subclasses fine-grained control over the nesting of blocks. As a general rule the <code>toString()</code> method returns a block of LaTeX ready to be 
    pasted into a document. This means the classes MUST be able to handle nested blocks automatically, consistently and correctly. Of course this also allows 
    <i>users</i> that same fine-grained control while maintaining the integrity of the individual constructs. Use with care!</p>
    <p>Two methods give full information on end tags. <code>hasEndTag()</code> returns true only if a non-null, non-blank end tag is defined. If it returns false 
    then no end tag is possible because there isn't one defined. <code>isEndTagEnabled()</code> reports whether or not end-tag printing is enabled. If not then 
    <code>toString()</code> won't include an end tag, even if one exists.<code>getEndTag()</code> always creates an end tag if possible, even if end tags are 
    disabled. This plus the two copy methods allow a tag to be fully printed WITH or WITHOUT an end tag; those can be collected and stored for later use and 
    nesting.</p>
*/

public abstract class LaBaseTag {

  /** This is a Java-escaped backslash, defined here for convenience. */
  public static final String BS = "\\";
  /** Double-backslash, to prevent frustration. */
  public static final String DBS = "\\\\";
  /** Double-backslash plus a newline. */
  public static final String DBSN = "\\\\\n";

  private static boolean globalETEn = true;

  /** All LaTeX tags, statements and commands have a name. This is it. It is protected to allow subclass access but SHOULD NOT be changed without a <i>very good</i> 
      reason. The default value is 'LaTeX' for the LaTeX logo. For block-style tags this is the <b>beginning</b> tag. It defaults to 'begin' only, any options 
      or curly braces are handled later.
  */
  protected String myTagName = "begin";

  /** This is the end tag. It defaults to 'end' with curly braces added later. If this field is manually set to null or blank NO END TAG IS PRINTED!*/
  protected String myEndTag = "end";

  /** This is the content of the first set of curly braces after the '/begin' tag and also within the braces with the 'end' tag. If null or blank neither it nor 
      the braces appear when the tag is printed. The default value is null.
  */
  protected String myEnvironment = null;

  /** Options appear in square brackets between the '/begin' sequence and the first opening curly brace. If null (default) or blank nothing prints. */
  protected String myOptions = null;

  /** These are the extra curly braces that may appear after the initial set. If not null or empty each individual String is appended after the initial closing 
      curly brace and enclosed in curly braces of its own. If null (default) or empty it is ignored. If any element is null or blank an empty set of curly braces 
      ('{}') is included.
  */
  protected String[] myEnvParams = null;

  /** This is the section label. It is protected to allow subclasses to access it if necessary. Subclasses that do not allow labels may alter it but others
      should not.
  */
  protected String mySectionLabel = null;

  /** This is the title information used in <code>LaDocument</code>. It is stored here to allow proper print placement. */
  protected String myTitle = null, myDate = null, myAuthor = null;

  /** This is picture size information stored here for proper print placement. */
  protected int myPicX = -1, myPicY = -1, myPicXOrigin = -1, myPicYOrigin = -1;

  /** Flag to indicate whether or not origin coordinates were given. */
  protected boolean myPicHasOrigin = false;

  private StringBuffer myContent = new StringBuffer();
  private boolean myEndTagEnabled = getGlobalETEn();

  /** This constructor allows every aspect of a subclass to be set.
      @param pTagName The tag name OR beginning tag name. Defaults to 'begin'. Null or blank values are silently ignored.
      @param pEndTag The end tag name. Defaults to 'end'. Null or blank values are ignored.
      @param pEnvironment The initial environment. Defaults to null.
      @param pOptions Initial options. Defaults to null.
      @param pContent The content to be set. Ignored if null.
  */
  protected LaBaseTag(String pTagName, String pEndTag, String pEnvironment, String pOptions, String pContent) {
    super();
    if(Question.validString(pTagName)) myTagName = pTagName;
    if(Question.validString(pEndTag)) myEndTag = pEndTag;
    myEnvironment = pEnvironment;
    myOptions = pOptions;
    if(Question.validString(pContent)) myContent = new StringBuffer(pContent);
  }

  private static boolean getGlobalETEn() {
    boolean retval = globalETEn;
    globalETEn = true;
    return retval;
  }

  /** Use this method to allow a tag to be created with its end tag disabled. Setting this method false suppresses the end tag on THE ONE NEXT instance or 
      subclass created. The internal machinery (private static method) returns false ONCE then resets to true.
  */
  protected static final void _setETEn(boolean pEnable) { globalETEn = pEnable; }

  /** This method clears all internal content. It does not alter the tag names, environment or options, only the content and label. Subclasses that override this 
      method to clear their data should issue a call to <code>super.reset()</code> to ensure proper cleaning.
  */
  public void reset() {
    myContent = new StringBuffer();
    mySectionLabel = null;
  }

  /** This method determines whether or not this instance has a valid ending tag. It returns true if the end tag is not null and not blank, false otherwise. */
  public final boolean hasEndTag() { return Question.validString(myEndTag); }

  /** This method determines whether or not the end tag is enabled. If false it will not print. If enabled it will print if it is defined. */
  public boolean isEndTagEnabled() { return myEndTagEnabled; }

  /** Subclasses use this method to enable or disable end tag printing. */
  protected void _setEndTagEnabled(boolean pEnable) { myEndTagEnabled = pEnable; }

  /** This method creates and returns the end tag for this instance. If the end tag is not null or blank this method WILL create one. Enable/disable has no effect 
      on the results. If the end tag is null or blank the return value is an empty String.
  */
  public final String getEndTag() {
    if(!hasEndTag()) return "";
    StringBuffer retval = new StringBuffer("\\");
    retval.append(myEndTag);
    if(Question.validString(myEnvironment)) retval.append("{").append(myEnvironment).append("}");
    return retval.toString();
  }

  /** Add content to the internal buffer. No linefeeds or other formatting is done, these must be added manually. This method is protected to disallow arbitrary 
      content. Null or blank values are silently ignored.
      @param pData The content to be added to the internal buffer.
  */
  protected final void _addContent(String pData) {
    if(Question.validString(pData)) myContent.append(pData);
  }

  /** This method adds content with NO checking whatsoever. USE WITH EXTREME CAUTION! */
  protected final void _addRawContent(String pData) { myContent.append(pData); }

  /** Retrieve the raw content that has been added without any other formatting. Only the content added via <code>_addContent()</code> is returned. */
  protected final String _getContent() { return myContent.toString(); }

  /** This method allows a '/label{}' tag to appear immediately after the opening line of the block. If the label set is not null or blank it is printed during 
      any output operation. Subclasses that do not allow labels should override this method with an empty one.
      @param pLabel The label to assign or null to disable.
  */
  public void setLabel(String pLabel) { mySectionLabel = pLabel; }

  /** Print all the data within this tag as a properly-formatted String. Newlines are added after each tag command and the (entire!) content. */
  public String toString() {
    StringBuffer retval = new StringBuffer("\\");
    retval.append(myTagName);
    if(Question.validString(myOptions)) retval.append("[").append(myOptions).append("]");
    if(Question.validString(myEnvironment)) retval.append("{").append(myEnvironment).append("}");
    if((myEnvParams != null) && (myEnvParams.length > 0)) {
      for(String sEnv : myEnvParams) {
        retval.append("{").append(Question.validString(sEnv)?sEnv:"").append("}");
      }
    }
    if((myPicX > -1) && (myPicY > -1)) retval.append("(").append(myPicX).append(", ").append(myPicY).append(")");
    if(myPicHasOrigin) retval.append("(").append(myPicXOrigin).append(", ").append(myPicYOrigin).append(")");
    retval.append("\n");
    if(Question.validString(mySectionLabel)) retval.append("\\label{").append(mySectionLabel).append("}\n");
    if(Question.validString(myTitle)) {
      retval.append("\\title{").append(myTitle).append("}\n");
      if(Question.validString(myAuthor)) retval.append("\\author{").append(myAuthor).append("}\n");
      if(Question.validString(myDate)) retval.append("\\date{").append(myDate).append("}\n");
      retval.append("\\maketitle\n");
    }
    retval.append(myContent.toString()).append("\n");
    if(isEndTagEnabled()) retval.append(getEndTag());
    retval.append("\n");
    return retval.toString();
  }

  /** This <i>instance</i> method is identical to <code>toString()</code> with the end tag suppressed. SET means <b>S</b>uppress <b>E</b>nd <b>T</b>ag and the 
      end tag is explicitly disabled for the duration of this method <i>only</i>. When it returns the original state of the end tag is restored.
  */
  public String toStringSET() {
    boolean bTemp = isEndTagEnabled();
    _setEndTagEnabled(false);
    String retval = toString();
    _setEndTagEnabled(bTemp);
    return retval;
  }

  /** This method provides a copy of a <code>LaBaseTag</code> subclass. Each subclass casts its return value as its own type. */
  public abstract LaBaseTag copyOf();

  /** Create and return a copy of this tag with its end tag suppressed. */
  public abstract LaBaseTag copyOfSET();

}

