package cox.jmatt.java.MathTools.markup.latex;

import java.util.*;
import cox.jmatt.java.MathTools.*;

/** <p>This class represents the entire LaTeX document. It provides equipment to handle the preamble, the title and any other document-wide configuration that 
    might be needed. When either <code>toString()</code> method is called the result is a complete LaTeX document, ready for rendering.</p>
    <p>There are three basic types of methods here: document structure methods, editing methods and convenience methods. The structuring methods affect the preamble 
    and the overall document. The editing methods add content to the document itself. The convenience methods present shortcuts to such things as changing font 
    size or adding <code>MathGenLatex.LaSymbol</code>s to the document body.</p>
    <p>The document class is set by the {@link #setType setType()} method, see it for the types available. This can be set any time before <code>toString()</code> 
    as can any of the other structure methods. The editing (and convenience) methods all append to the end of the current content so order is important there. Since 
    this class represents the final form of a LaTeX document there are no end-tag-suppressed options. Both <code>copyOfSET()</code> and <code>toStringSET()</code> 
    return non-SET values.</p>
*/

public class LaDocument extends LaBaseTag {

  private String myType = "article", myDocOptions = null;
  private StringBuffer myPackages = new StringBuffer();

  /** Standard constructor. Default document class is 'article.' */
  public LaDocument() { super(null, null, "document", null, null); }

  private LaDocument(String pOpts, String pContent, String pType, String pTitle, String pAuthor, String pDate) {
    super(null, null, "document", null, pContent);
    myType = pType; myTitle = pTitle; myAuthor = pAuthor; myDate = pDate; myDocOptions = pOpts;
  }

  /** Set all title information. All parameters are optional but if 'pTitle' is null or blank no title information will be printed.
      @param pTitle The document title.
      @param pAuthor The document author.
      @param pDate The date for the document.
  */
  public LaDocument setTitle(String pTitle, String pAuthor, String pDate) {
    myTitle = pTitle; myAuthor = pAuthor; myDate = pDate;
    return this;
  }

  /** <p>Set the '/documentclass' of this document. Six options are available and each is represented by a single character:
      <ul>
        <li>'a' is <b>article</b> and it is the default setting.</li>
        <li>'b' is <b>book</b>.</li>
        <li>'p' is <b>proc</b>, for <i>proceedings</i>.</li>
        <li>'r' is <b>report</b>.</li>
        <li>'s' is <b>slides</b>.</li>
        <li>'m' is <b>minimal</b>, typically for debugging purposes.</li>
      </ul>
      Uppercase characters also work. Any character other than one of the ones above results in 'article' being set.</p>
  */
  public LaDocument setType(char pType) {
    switch(pType) {
      case 'b': case 'B': myType = "book"; break;
      case 'p': case 'P': myType = "proc"; break;
      case 'r': case 'R': myType = "report"; break;
      case 's': case 'S': myType = "slides"; break;
      case 'm': case 'M': myType = "minimal"; break;
      case 'a': case 'A':
      default:
        myType = "article";
    }
    return this;
  }

  /** This method sets any options that appear with the '/documentclass<b><u>[]</u></b>...' tag. Setting this to null or blank suppresses them.
      @param pOptions The options String to set on the 'documentclass' tag.
  */
  public LaDocument setDocOptions(String pOptions) {
    myDocOptions = pOptions;
    return this;
  }

  /** Overridden to clear document-specific information: title, author, date, packages etc. The document class is also reset to 'article.' */
  public void reset() {
    super.reset();
    myType = "article"; myTitle = null; myAuthor = null; myDate = null;
    myPackages = new StringBuffer();
  }

  /** This method provides the '/usepackage...' commands. The 'pOpts' parameter is optional. If not null the <i>package</i> options appear in square brackets at 
      the proper place. If the package name is null or blank it is silently ignored.
      @param pPackage The package to use.
      @param pOpts The package options to include, or null for none.
  */
  public LaDocument usePackage(String pPackage, String pOpts) {
    if(!Question.validString(pPackage)) return this;
    myPackages.append("\\usepackage");
    if(Question.validString(pOpts)) myPackages.append("[").append(pOpts).append("]");
    myPackages.append("{").append(pPackage).append("}\n");
    return this;
  }

  /** Add a <code>LaBaseTag</code> to the current content. Any tag except <code>LaDocument</code> or a subclass of it can be added. The tag is added as-is via its 
      <code>toString()</code> method so changing it after adding it has no effect on <i>this</i> content. A single newline is added at the end.
      @param pTag The tag to be added.
  */
  public LaDocument add(LaBaseTag pTag) {
    if((pTag == null) || (pTag instanceof LaDocument)) return this;
    _addContent(pTag.toString()+"\n");
    return this;
  }

  /** Add a line of text to the content. A newline is supplied automatically. If 'pLine' is null or blank <code>LaBaseTag.DBSN</code> is added for a blank line. */
  public LaDocument addLine(String pLine) {
    if(Question.validString(pLine)) _addContent(pLine+"\n");
    return this;
  }

  /** Add a blank line to the document.
      @param pHeight The height of the blank line in points (pt.). An extra newline is added for a blank line in the generated document.
  */
  public LaDocument addLine(int pHeight) {
    _addContent("\\vspace{"+Math.abs(pHeight)+"pt}\n\n");
    return this;
  }

  /** Add a '/newline' to the document. A newline <i>character</i> is appended to the end of the sequence. */
  public LaDocument newline() {
    _addContent("\\newline\n");
    return this;
  }

  /** Add a '/newpage' directive to the document. */
  public LaDocument newPage() {
    _addContent("\\newpage\n");
    return this;
  }

  /** Add a <i>single-line</i> comment to the document text. The supplied string cannot contain a newline and the comment character ('%') is prepended automatically.
      Null or blank comments are silently ignored.
      @param pComment A String comment to add. It must contain NO newline characters.
  */
  public LaDocument addComment(String pComment) {
    if(Question.validString(pComment)) _addContent("%"+pComment);
    return this;
  }

  /** <p>Include the contents of an external file in the document. Ignored if 'pFile' is null or blank. A newline is appended at the end.</p>
      <p>The 'pInclude' parameter determines which LaTeX directive is issued: '/include' (true) or '/input' (false). In the former case LaTeX places a page break 
      before the included material. In the latter case it does not. IMPORTANT NOTE: This method does <b>not</b> itself copy the file contents, it merely issues 
      the appropriate LaTeX directive.</p>
      @param pFile The filename containing the markup to be included.
      @param pInclude true to 'include,' false to 'input.'
  */
  public LaDocument include(String pFile, boolean pInclude) {
    if(Question.validString(pFile)) _addContent("\\"+(pInclude?"include":"input")+"{"+pFile+"}\n");
    return this;
  }

  /** Add raw text to the current content. Null or blank Strings are ignored but no other validation is performed. No newline is supplied. */
  public LaDocument addRaw(String pText) {
    _addRawContent(pText);
    return this;
  }

  /** Add a <code>MathGenLatex.LaSymbol</code> to the current content. A newline is included afterward automatically. If the symbol requested is null or blank 
      nothing happens.
      @param pNumDS The number of dollar signs to wrap around the result.
      @param pSymbol The symbol to add.
      @param pParams Any parameters required by the symbol.
      @param pBefore The prefix String; after the initial dollar sign(s) and before the symbol starts.
      @param pAfter The suffix String, after the Symbol and before the ending dollar signs and newline.
  */
  public LaDocument addSymbol(int pNumDS, String pBefore, String pAfter, String pSymbol, String... pParams) {
    if(!Question.validString(pSymbol)) return this;
    if(pBefore == null) pBefore = " ";
    if(pAfter == null) pAfter = " ";
    _addContent(MathGenLatex.wrapDS((pBefore+MathGenLatex.fillSymbol(pSymbol, pParams)+pAfter).trim(), pNumDS)+"\n");
    return this;
  }

  /** This is the quick version of <code>addSymbol()</code>. It takes ONLY the symbol name and parameters. NO dollar signs, prefix or suffix. */
  public LaDocument addSymbol(String pSymbol, String... pParams) {
    return addSymbol(0, null, null, pSymbol, pParams);
  }

  /** Add a '/chapter{}' tag. 'pData' is the chapter name. */
  public LaDocument addChapter(String pData) {
    if(Question.validString(pData)) _addContent("\\chapter{"+pData+"}\n");
    return this;
  }

  /** Add a '/section{}' tag. 'pData' is the section name. */
  public LaDocument addSection(String pData) {
    if(Question.validString(pData)) _addContent("\\section{"+pData+"}\n");
    return this;
  }

  /** Add a '/subsection{}' tag. 'pData' is the subsection name. */
  public LaDocument addSSection(String pData) {
    if(Question.validString(pData)) _addContent("\\subsection{"+pData+"}\n");
    return this;
  }

  /** Add a '/subsubsection{}' tag. 'pData' is the name. */
  public LaDocument addSSSection(String pData) {
    if(Question.validString(pData)) _addContent("\\subsubsection{"+pData+"}\n");
    return this;
  }

  /** Add a '/paragraph{}' tag. 'pData' is the name. */
  public LaDocument addParagraph(String pData) {
    if(Question.validString(pData)) _addContent("\\paragraph{"+pData+"}\n");
    return this;
  }

  /** Add a '/subparagraph{}' tag. 'pData' is the name. */
  public LaDocument addPParagraph(String pData) {
    if(Question.validString(pData)) _addContent("\\subparagraph{"+pData+"}\n");
    return this;
  }

  /** <p>Convenience method to alter the font size of the text. This method uses the <code>MathGenLatex.LaSymbol</code> 'FS...' constants internally. It supplies 
      the 'FS' so only the font style desired is needed. DO NOT include the 'FS' otherwise it will be repeated! A single newline is appended.</p>
      <p>If either argument is null or blank nothing happens.</p>
      @param pFontSize The ('FS'-less) name of the font effect.
      @param pText The text to be modified.
  */
  public LaDocument addFS(String pFontSize, String pText) {
    if(!Question.validString(pFontSize) || !Question.validString(pText)) return this;
    addSymbol("FS"+pFontSize, pText);
    return this;
  }

  /** Overridden to include the preamble before the 'begin...' tag. */
  public String toString() {
    StringBuffer retval = new StringBuffer("\\documentclass");
    if(Question.validString(myDocOptions)) retval.append("[").append(myDocOptions).append("]");
    retval.append("{").append(myType).append("}\n\n");
    retval.append(myPackages.toString()).append("\n");
    
    retval.append("\n");
    retval.append(super.toString());
    return retval.toString();
  }

  /** This class <b>always</b> has its end tag. This method echoes <code>toString()</code>. */
  public String toStringSET() { return toString(); }

  /** Create and return a copy of this document. */
  public LaDocument copyOf() { return new LaDocument(myDocOptions, _getContent(), myType, myTitle, myAuthor, myDate); }

  /** No document is ever without its end tag. This method simply calls <code>copyOf()</code>. */
  public LaDocument copyOfSET() { return copyOf(); }

}

