package cox.jmatt.java.MathTools.markup.latex;

import java.util.*;
import cox.jmatt.java.MathTools.*;

/** <p>This class provides the '/picture' environment complete with '/setlength{/unitlength}{}' directive. It provides both single ('/put') and multiple ('/multiput') 
    versions for all drawing commands that allow them. The Bezier drawing command ('/qbezier') has only the standard version. No value-checking is done for the 
    commands that restrict the (integer) values of the direction vectors.</p>
*/

public class LaPicture extends LaBaseTag {

  private String myUnitLength = "1pt";
  private boolean isSaveBoxOpen = false;

  /** Standard constructor. The height and width of the picture box are required. */
  public LaPicture(int pX, int pY) {
    super(null, null, "picture", null, null);
    myPicX = Math.abs(pX); myPicY = Math.abs(pY);
    myPicHasOrigin = false;
  }

  /** Constructor to specify width, height and relative origin values. */
  public LaPicture(int pX, int pY, int pX0, int pY0) {
    super(null, null, "picture", null, null);
    myPicX = Math.abs(pX); myPicY = Math.abs(pY);
    myPicXOrigin = Math.abs(pX0); myPicYOrigin = Math.abs(pY0);
    myPicHasOrigin = true;
  }
  
  /* Private copy constructor used for copyOf(). */
  private LaPicture(String pContent, int pX, int pY, int pX0, int pY0, String pUnits, boolean hasOrigin) {
    super(null, null, "picture", null, pContent);
    myPicX = Math.abs(pX); myPicY = Math.abs(pY);
    myPicXOrigin = Math.abs(pX0); myPicYOrigin = Math.abs(pY0);
    myUnitLength = pUnits;
    myPicHasOrigin = hasOrigin;
  }

  /** Overridden to reset the unit length ONLY to its default value. Size and origin values are not affected. This method does close a savebox, if one is open. */
  public void reset() {
    setUnitLength(null);
    closeSaveBox(false);
    super.reset();
  }
  
  /** Set the unit length. A null or blank resets it to the default value: 1 point ('1pt'). */
  public LaPicture setUnitLength(String pLength) {
    myUnitLength = Question.validString(pLength)?pLength:"1pt";
    return this;
  }

  /** This is the basic '/put' command used to put things into the picture. It is used for all single items except Bezier curves. If the object argument is null 
      or blank nothing happens. This is also the method to use to place text or formulas on the picture.
      @param pX The x-coordinate of the item's location.
      @param pY The y-coordinate of the location.
      @param pObject The object to be placed in the picture.
  */
  public LaPicture put(double pX, double pY, String pObject) {
    if(Question.validString(pObject)) _addContent("\\put("+pX+", "+pY+"){"+pObject.trim()+"}\n");
    return this;
  }

  /** This method implements the '/multiput' directive. A null or blank object is silently ignored.
      @param pX The x-coordinate of the item's location.
      @param pY The y-coordinate of the location.
      @param pDX The x-value of the translation vector.
      @param pDY The y-value of the translation vector.
      @param pNum The number of copies to put.
      @param pObject The object to be placed in the picture.
  */
  public LaPicture put(double pX, double pY, double pDX, double pDY, double pNum, String pObject) {
    if(Question.validString(pObject)) _addContent("\\multiput("+pX+", "+pY+")("+pDX+", "+pDY+"){"+((int) pNum)+"}{"+pObject.trim()+"}\n");
    return this;
  }

  /* Create a line or vector. */
  private String _lineCmd(int pX, int pY, double pLength, boolean isVector) {
    return "\\"+(isVector?"vector(":"line(")+pX+", "+pY+"){"+pLength+"}";
  }

  /** Issue a '/thicklines' command. */
  public LaPicture thick() {
    _addContent("\\thicklines\n");
    return this;
  }

  /** Issue a '/thinlines' command. */
  public LaPicture thin() {
    _addContent("\\thinlines\n");
    return this;
  }

  /** Set the line thickness. A null or blank value sets the value '1pt'. */
  public LaPicture setLineThickness(String pThickness) {
    _addContent("\\linethickness{"+(Question.validString(pThickness)?pThickness:"1pt")+")\n");
    return this;
  }

  /** Draw a single line.
      @param pX The x-coordinate of the line's origin.
      @param pY The y-coordiante of the origin.
      @param pDirX The x-value of the direction vector.
      @param pDirY The y-value of the direction vector.
      @param pLen The length of the line.
  */
  public LaPicture line(double pX, double pY, int pDirX, int pDirY, double pLen) { return put(pX, pY, _lineCmd(pDirX, pDirY, pLen, false)); }

  /** Draw a '/multiput' line.
  @param pX The x-coordinate of the line's origin.
      @param pY The y-coordiante of the origin.
      @param pDirX The x-value of the direction vector.
      @param pDirY The y-value of the direction vector.
      @param pLen The length of the line.
      @param pDX The '/multiput' translation vector x-value.
      @param pDY The '/multipyt' translation vector y-value.
      @param pNum The number of copies to make.
  */
  public LaPicture line(double pX, double pY, int pDirX, int pDirY, double pLen, double pDX, double pDY, int pNum) {
    return put(pX, pY, pDX, pDY, pNum, _lineCmd(pDirX, pDirY, pLen, false));
  }

  /** Draw a single vector (arrow).
      @param pX The x-coordinate of the line's origin.
      @param pY The y-coordiante of the origin.
      @param pDirX The x-value of the direction vector.
      @param pDirY The y-value of the direction vector.
      @param pLen The length of the line.
  */
  public LaPicture vector(double pX, double pY, int pDirX, int pDirY, double pLen) { return put(pX, pY, _lineCmd(pDirX, pDirY, pLen, true)); }

  /** Draw a '/multiput' arrow.
  @param pX The x-coordinate of the line's origin.
      @param pY The y-coordiante of the origin.
      @param pDirX The x-value of the direction vector.
      @param pDirY The y-value of the direction vector.
      @param pLen The length of the line.
      @param pDX The '/multiput' translation vector x-value.
      @param pDY The '/multipyt' translation vector y-value.
      @param pNum The number of copies to make.
  */
  public LaPicture vector(double pX, double pY, int pDirX, int pDirY, double pLen, double pDX, double pDY, int pNum) {
    return put(pX, pY, pDX, pDY, pNum, _lineCmd(pDirX, pDirY, pLen, true)); 
 }

  /** This command draws an unfilled circle. NOTE: The argument given is <b>diameter</b> and not radius!
      @param pX The x-coordinate of the center of the circle.
      @param pY The y-coordinate of the center.
      @param pDiameter The diameter of the circle.
  */
  public LaPicture circle(double pX, double pY, double pDiameter) {
    return put(pX, pY, "\\circle{"+pDiameter+"}");
  }

  /** The '/multiput' circle command. */
  public LaPicture circle(double pX, double pY, double pDiameter, double pDX, double pDY, int pNum) {
    return put(pX, pY, pDX, pDY, pNum, "\\circle{"+pDiameter+"}");
  }

  /** This command draws a filled circle. NOTE: The argument given is <b>diameter</b> and not radius!
      @param pX The x-coordinate of the center of the circle.
      @param pY The y-coordinate of the center.
      @param pDiameter The diameter of the circle.
  */
  public LaPicture disc(double pX, double pY, double pDiameter) {
    return put(pX, pY, "\\circle*{"+pDiameter+"}");
  }

  /** The '/multiput' disc command. */
  public LaPicture disc(double pX, double pY, double pDiameter, double pDX, double pDY, int pNum) {
    return put(pX, pY, pDX, pDY, pNum, "\\circle*{"+pDiameter+"}");
  }

  /** Place a single oval in the drawing. The parameters are standard save for the last. The position argument, omitted if null or blank, must consist of characters 
      from: 't', 'b', 'l' and 'r'. These commands draw, respectively, the top, bottom, left or right portion of the oval.
      @param pX The x-cordinate of the oval's center.
      @param pY The y-coordinate of the center.
      @param pWidth The width of the oval.
      @param pHeight The height of the oval.
      @param pPos The 'position' argument of the oval.
  */
  public LaPicture oval(double pX, double pY, double pWidth, double pHeight, String pPos) {
    return put(pX, pY, "\\oval("+pWidth+", "+pHeight+")"+(Question.validString(pPos)?"["+pPos+"]":"")+"\n");
  }

  /** Multi-put oval. */
  public LaPicture oval(double pX, double pY, double pWidth, double pHeight, String pPos, double pDX, double pDY, int pNum) {
    return put(pX, pY, pDX, pDY, pNum, "\\oval("+pWidth+", "+pHeight+")"+(Question.validString(pPos)?"["+pPos+"]":"")+"\n");
  }

  /** Put a quadratic Bezier curve into the picture. It is given as two endpoints and a single control point: (x1, y1), (x2, y2) and (cX, cY) respectively. The 
      arguments to the method are in the same order as those in the '/qbezier' command.
  */
  public LaPicture qBezier(double x1, double y1, double cX, double cY, double x2, double y2) {
    _addContent("\\qbezier("+x1+", "+y1+")("+cX+", "+cY+")("+x2+", "+y2+")\n");
    return this;
  }

  /** <p>Create and open a savebox. All subsequent drawing commands fall within the savebox until it is closed or another is opened. If the name assigned to the 
      savebox is null or blank no box is opened. The backslash IS provided automatically so the name <b>does not</b> need it.</p>
      <p>CRITICAL NOTES: The state of an open savebox is NOT preserved during a copy operation. If a <code>LaPicture</code> is copied with a savebox opened it must 
      be forcibly closed manually. Also, saveboxes cannot be nested!</p>
  */
  public LaPicture openSaveBox(String pName, double pWidth, double pHeight, String pPos) {
    if(!Question.validString(pName)) return this;
    closeSaveBox(false);
    _addContent("\\newsavebox{\\"+pName+"}\n");
    _addContent("\\savebox{\\"+pName+"}("+pWidth+", "+pHeight+")");
    if(Question.validString(pPos)) _addContent("["+pPos+"]");
    _addContent("{\n");
    isSaveBoxOpen = true;
    return this;
  }

  /** Close any open savebox. If there is not an open box (or a copy operation has occurred!) this method returns silently. If there IS an open savebox it is 
      properly closed. If the 'force' parameter is true a closing brace is emitted.
  */
  public LaPicture closeSaveBox(boolean pForce) {
    if((isSaveBoxOpen) || (pForce)) {
      _addContent("}\n");
      isSaveBoxOpen = false;
    }
    return this;
  }

  /** Use a (presumably) previously-defined savebox. If the name supplied is null or blank nothing happens.
      @param pX The x-coordinate of the used box.
      @param pY The y-coordinate of the used box.
      @param pBoxName The name of the savebox to use. The backslash before the name IS provided automatically.
  */
  public LaPicture useBox(double pX, double pY, String pBoxName) {
    if(Question.validString(pBoxName)) put(pX, pY, "\\usebox{\\"+pBoxName+"}");
    return this;
  }

  /** Copy the current picture information. */
  public LaPicture copyOf() { return new LaPicture(_getContent(), myPicX, myPicY, myPicXOrigin, myPicYOrigin, myUnitLength, myPicHasOrigin); }

  /** Copy the current picture information with no end tag. */
  public LaPicture copyOfSET() {
    _setETEn(false);
    return copyOf();
  }

  /** Overridden to set unit length prior to output. */
  public String toString() {
    closeSaveBox(false);
    return "\\setlength{\\unitlength}{"+myUnitLength+"}\n"+super.toString();
  }

  /** Overridden to set unit length first. */
  public String toStringSET() {
    closeSaveBox(false);
    return "\\setlength{\\unitlength}{"+myUnitLength+"}\n"+super.toStringSET();
  }

}

