package cox.jmatt.java.MathTools.markup.latex;

import cox.jmatt.java.MathTools.*;

/** <p>This class exists to generate the various tag classes in the MathTools <code>latex.*</code> package. It serves the same purpose as <code>MathDrawSVG</code> 
    and <code>MathGenHTML</code>. In addition to factory methods for the other <code>markup.latex.*</code> classes it contains the <code>LaSymbol enum</code>. 
    The constants in this enum all define a <code>format(String...)</code> method. In some cases it is empty but in others it is used to provide properly-formatted 
    markup without worrying about slapping Strings together manually.</p>
*/

public class MathGenLatex {

  /** <p>This <code>enum</code> holds <b>some</b> of the (millions of) LaTeX symbols and commands used with mathematics. PLEASE NOTE: This list is <b>not</b> by 
      any means complete or comprehensive! It will also grow with time. The symbols represented here should be available with a base LaTeX package and not require 
      any external packages to be included. If a symbol does require an external package this will be noted in its description.</p>
      <p>Some of the symbols are standalone, others allow parameters. Where needed the parameters are given through the <code>format()</code> methods. The order 
      of the parameters is as they appear from left to right <b>unless specified otherwise</b>.</p>
      <p>Binary relations and operations accept zero, one or two parameters. The first appears on the left of the symbol, the second on the right. The parameters 
      are enclosed within curly braces to allow compound expressions if needed. This should not affect the output.</p>
      <p>Font effect tags are named for the effect and NOT after the tag; e.g. 'ITALIC' is the '/textit...' tag. For font size effects the <code>enum</code> consants 
      all begin with 'FS'. Note the 'large' tags differentiate between themselves by the number of 'L' characters in 'LARGE'.</p>
      <p>Grouping symbol constructs all begin with <b>GS</b> and include the '/left' and '/right' directives in front of the specified variable-sized delimiters. 
      All of these constructs take one parameter which is the markup to be included between the opening and closing symbols. This also includes (without the 'GS') 
      certain 'wide' accents, such as the vector symbol, left arrow and tilde. Although these are not grouping symbols per se they can include groups of characters 
      therefore they are classified as such.</p>
  */
  public static enum LaSymbol {

    /** This is the 'not exist' symbol (slashed, backward 'E'), 0 params. This symbol is also used when another symbol cannot be found. */
    NOTEXIST("\\not\\exists", 0),
    /** This is the '/notin' symbol, 0 params. */
    NOTIN("\\notin", 0),
    /** This is the sequence for the LaTeX logo. */
    LATEX("\\LaTeX", 0),
    /** Fraction symbol, 2 params. */
    FRAC("\\frac{<0>}{<1>}", 2),
    /** Radical symbol, 1 or 2 params. First parameter is radicand, second is index. */
    SQRT("\\sqrt[<1>]{<0>}", 2),
    /** Summation symbol (uppercase sigma), 0 to 3 params. First param is lower limit, second is upper limit, third is the summand. */
    SUM("\\sum _{<0>}^{<1>} {<2>}", 3),
    /** Product symbol (uppercase pi), 0 to 3 params. First is the lower limit, second is upper limit, third is the factor expression. */
    PROD("\\prod _{<0>}^{<1>} {<2>}", 3),
    /** Single integral symbol, 0 to 3 params. First is lower limit, second is upper, third is the integrand. */
    INT("\\int _{<0>}^{<1>} {<2>}", 3),

    /** Binary relation: approximates (wavy equals). */
    APPROX("{<0>} \\approx {<1>}", 2),
    /** Binary relation: bowtie. */
    BOWTIE("{<0>} \\bowtie {<1>}", 2),
    /** Binary relation: congruent. */
    CONG("{<0>} \\cong {<1>}", 2),
    /** Binary relation: equivalent (tribar equals) */
    EQUIV("{<0>} \\equiv {<1>}", 2),
    /** Binary relation: not equal to. */
    NEQ("{<0>} \\neq {<1>}", 2),
    /** Binary relation: greater or equal. */
    GEQ("{<0>} \\geq {<1>}", 2),
    /** Binary relation: less or equal. */
    LEQ("{<0>} \\leq {<1>}", 2),
    /** Binary relation: parallel. */
    PARALLEL("{<0>} \\parallel {<1>}", 2),
    /** Binary relation: perpendicular. */
    PERP("{<0>} \\perp {<1>}", 2),
    /** Binary relation: similar. */
    SIM("{<0>} \\sim {<1>}", 2),
    /** Binary relation: similar or equal. */
    SIMEQ("{<0>} \\simeq {<1>}", 2),
    /** Binary relation: subset. */
    SUBSET("{<0>} \\subset {<1>}", 2),
    /** Binary relation: subset or equals. */
    SUBSETEQ("{<0>} \\subseteq {<1>}", 2),
    /** Binary relation: superset. */
    SUPSET("{<0>} \\supset {<1>}", 2),
    /** Binary relation: superset or equal. */
    SUPSETEQ("{<0>} \\supseteq {<1>}", 2),
    /** Binary operation: divide by. */
    DIV("{<0>} \\div {<1>}", 2),
    /** Binary operation: circle-times. */
    CIRC("{<0>} \\circ {<1>}", 2),
    /** Binary operation: center-dot multiply. */
    CDOT("{<0>} \\cdot {<1>}", 2),
    /** Binary operation: multiplication symbol. */
    TIMES("{<0>} \\times {<1>}", 2),

    /** Italicized text, 1 param. */
    ITALIC("\\textit{<0>}", 1),
    /** Bold text, 1 param. */
    BOLD("\\textbf{<0>}", 1),
    /** Slanted text, 1 param. */
    SLANT("\\textsl{<0>}", 1),
    /** Small caps, 1 param. */
    SMALLCAP("\\textsc{<0>}", 1),
    /** Monospaced (teletype) text, 1 param. */
    MONOSPACE("\\texttt{<0>}", 1),
    /** Underlined text, 1 param. */
    UNDERLINE("\\underline{<0>}", 1),
    /** Overlined text, 1 param. */
    OVERLINE("\\overline{<0>}", 1),

    /** Font size 'tiny', 1 param. */
    FSTINY("{\\tiny <0> }", 1),
    /** Font size 'scritpsize', 1 param. */
    FSSCRIPT("{\\scriptsize <0> }", 1),
    /** Font size 'footnote', 1 param. */
    FSFOOTNOTE("{\\tiny <0> }", 1),
    /** Font size 'small', 1 param. */
    FSSMALL("{\\small <0> }", 1),
    /** Font size 'normal', 1 param. */
    FSNORMAL("{\\normalsize <0> }", 1),
    /** Font size 'large', 1 param, spelling: 'L'. */
    FSLARGE("{\\large <0> }", 1),
    /** Font size '<b>L</b>arge', 1 param, spelling: 'LL'. */
    FSLLARGE("{\\Large <0> }", 1),
    /** Font size 'LARGE', 1 param, spelling: 'LLL'. */
    FSLLLARGE("{\\LARGE <0> }", 1),
    /** Font size 'huge', 1 param. */
    FSHUGE("{\\huge <0> }", 1),

    /** Grouping symbol: curly braces. */
    GSCURLY("\\left \\{ <0> \\right \\}", 1),
    /** Grouping symbol: parentheses. */
    GSPAREN("\\left ( <0> \\right )", 1),
    /** Grouping symbol: square brackets. */
    GSSQUARE("\\left [ <0> \\right ]", 1),
    /** Grouping symbol: angle brackets. */
    GSANGLE("\\left \\langle <0> \\right \\rangle", 1),
    /** Grouping symbol: /lgroup and /rgroup. */
    GSGROUP("\\left \\lgroup <0> \\right \\rgroup", 1),
    /** Overline right arrow, 1 param. */
    ORARROW("\\overrightarrow{<0>}",1),
    /** Overline left arrow, 1 param. */
    OLARROW("\\overleftarrow{<0>}", 1),

    /** For all {} in {}, 2 params. This symbol combines the '/forall' and '/in' symbols for the expression '/forall &lt;0&gt; /in &lt;1&gt;'. */
    ALLIN("\\forall {<0>} \\in {<1>}", 2);
    

    private String myTemplate = "\\notin";
    private int myParams = 0;
    LaSymbol(String pTemplate, int pParams) {
      myTemplate = pTemplate;
      myParams = pParams;
    }

    /** Return the number of parameters this symbol requires. */
    public int getParams() { return myParams; }

    /** Return this symbol's raw formatting template. */
    public String getFormat() { return myTemplate; }

    /** Return a symbol or command with all required parameters filled in. */
    public String format(String... pParams) {
      if(myParams < 1) return myTemplate;
      if(pParams == null) pParams = new String[0];
      if(pParams.length < myParams) {
        String[] arTemp = new String[myParams];
        java.util.Arrays.fill(arTemp, "");
        for(int i = 0; i < pParams.length; i++) { arTemp[i] = Question.validString(pParams[i])?pParams[i]:""; }
        pParams = arTemp;
      }
      return Question.fillTemplate(myTemplate, pParams);
    }

    /** Calls <code>format(null)</code>. */
    public String format() { return format((String[]) null); }

  }

  /** Constructor for script exposure. */
  public MathGenLatex() { super(); }

  /** Fetch a <code>LaSymbol</code> by name. The 'pSymbol' parameter is converted to uppercase so lowercase letters can be used. If the symbol cannot be 
      found <code>LaSymbol.NOTEXIST</code> is returned and a Debug-level message is logged.
      @param pSymbol The <code>LaSymbol</code> to fetch.
  */
  public static LaSymbol getSymbol(String pSymbol) {
    LaSymbol retval = LaSymbol.NOTEXIST;
    try {
      retval = LaSymbol.valueOf(pSymbol.toUpperCase());
    } catch(Exception ex) { CapCom.debug("Error locating LaTeX symbol: "+pSymbol, ex); }
    return retval;
  }

  /** Instance shadow of <code>getSymbol()</code>. */
  public LaSymbol get_Symbol(String pSymbol) { return getSymbol(pSymbol); }

  /** Fetch and format a <code>LaSymbol</code>.
      @param pSymbol The desired symbol.
      @param pParams The parameters required for the symbol.
  */
  public static String fillSymbol(String pSymbol, String... pParams) {
    return getSymbol(pSymbol).format(pParams);
  }

  /** Instance shadow. */
  public String fill_Symbol(String pSymbol, String... pParams) { return fillSymbol(pSymbol, pParams); }

  /** Wrap a non-null, non-blank String in dollar signs ('$'). If the data String is null or blank an empty String is returned.
      @param pData String to wrap.
      @param pNum Number of dollar signs to put on each side.
  */
  public static String wrapDS(String pData, int pNum) {
    if(!Question.validString(pData)) return "";
    String retval = "";
    for(int i=0; i < pNum; i++) retval = retval + "$";
    return retval+(pData.trim())+retval;
  }

  /** Wrap a String in single dollar signs. */
  public String dWrap(String pData) { return wrapDS(pData, 1); }

  /** Wrap a String in double dollar signs. Notice the double 'd' as opposed to one. */
  public String ddWrap(String pData) { return wrapDS(pData, 2); }

  /** Create and return a new LaTag with the name supplied. */
  public LaTag newLaTag(String pName) { return new LaTag(pName); }

  /** Create and return a new list tag. */
  public LaListTag newList() { return new LaListTag(); }

  /** Create and return a new equation block tag. */
  public LEquation newEquation() { return new LEquation(); }

  /** Create and return a new <code>LaDocument</code>. */
  public LaDocument newDocument() { return new LaDocument(); }

  /** Create a new <code>LaDocument</code> with the <i>amsmath</i> package pre-installed. */
  public LaDocument newAMSDoc() { return new LaDocument().usePackage("amsmath", null); }

  /** Create an empty table. */
  public LaTable newTable() { return new LaTable(); }

  /** Create an empty table with the given column alignment specifier. */
  public LaTable newTable(String pColAlign) { return new LaTable(pColAlign); }

  /** Create a <code>LaPicture</code> with only the width and height specified. */
  public LaPicture newPicture(int pX, int pY) { return new LaPicture(pX, pY); }

  /** Create and return a <code>LaPicture</code> with size and origin specified. */
  public LaPicture newPicture(int pX, int pY, int pX0, int pY0) { return new LaPicture(pX, pY, pX0, pY0); }

}

