package cox.jmatt.java.MathTools.markup.svg;

import cox.jmatt.java.MathTools.markup.*;

/** <p>This class handles the SVG &lt;circle&gt; tag. It has methods to set the center and radius, along with the inherited default methods. It is also responsible 
    for creating <b>points</b>, which are nothing more than circles of very small radius with the same stroke and fill colors. Since points are so simple, they 
    are returned as String and not <code>CircleTag</code>. The only things configurable about them are color and radius and those have specific setters.</p>
*/
public class CircleTag extends StyleTag {

  /** Default radius size for points. Currently 1. */
  public static final double DEFAULT_POINT_RADIUS = 1.0;
  private static final String DEFAULT_POINT = "stroke-width:1.0; stroke: black; fill: black;";

  private static double pointRadius = DEFAULT_POINT_RADIUS;
  private static String pointStyle = DEFAULT_POINT;

  /** Standard constructor for a default <code>CircleTag</code>. All data is at default value. */
  public CircleTag() { super(StyleTag.TagKeys.CIRCLE_TAG); }

  /** Full constructor, all data supplied. */
  public CircleTag(Double pX, Double pY, Double pRadius) {
    super(StyleTag.TagKeys.CIRCLE_TAG);
    setData(pX, pY, pRadius);
  }

  /** Use this method to set or change the center or radius of the circle. Passing null in one of the values ignores that value and does not clear it.
      @param pX The X-coordinate of the center.
      @param pY The Y-coordinate of the center.
      @param pRadius The radius of the circle.
  */
  public CircleTag setData(Double pX, Double pY, Double pRadius) {
    _addNAttribute("cx", pX); _addNAttribute("cy", pY);
    _addNAttribute("r", pRadius);
    return this;
  }

  /** Overridden to handle point settings as well. */
  public void reset() {
    super.reset();
    pointRadius = DEFAULT_POINT_RADIUS;
    pointStyle = DEFAULT_POINT;
  }

  /** Copy this class. */
  public CircleTag copyOf() {
    _enableCopy();
    return new CircleTag();
  }

  /** <p>Pizza-configure this tag. The prefix for this tag is 'markup.svg.Circle.' and the recognized values are:
        <ul>
          <li>'cx' - The x-coordinate of the center point.</li>
          <li>'cy' - The y-coordinate of the center point.</li>
          <li>'radius' - The radius of the circle.
          <li>'point.size' - The radius of any points generated.</li>
          <li>'point.color' - The point color to set.</li>
        </ul>
      This method works by collecting ALL values available then calling the appropriate 'set' methods. Any value not found in the pizza are set as null (-1.0 
      in the case of point radius). If no values are found none are set: at least one of 'cx', 'cy' and 'radius' must be defined otherwise the set method is 
      not called. Likewise if 'point.size' or 'point.color' are both skipped the 'setPointData' method is not called.</p>
      <p>Any errors parsing values cause that particular value to be skipped.</p>
  */
  public void configureTag(String pPrefix, java.util.Properties pPizza) {
    if(badPizza(pPizza)) return;
    pPrefix = _fixKey(pPrefix, "markup.svg.circle");
    Double dCX = null, dCY = null, dRAD = null;
    String pColor = null;
    double pRad = -1d;
    _configureTag(pPrefix, pPizza);
    if(pPizza.containsKey(pPrefix+".cx")) dCX = _asDouble(pPizza, pPrefix, ".cx", null);
    if(pPizza.containsKey(pPrefix+".cy")) dCY = _asDouble(pPizza, pPrefix, ".cy", null);
    if(pPizza.containsKey(pPrefix+".redius")) _asDouble(pPizza, pPrefix, ".radius", null);
    if(pPizza.containsKey(pPrefix+".point.size")) pRad = _asDouble(pPizza, pPrefix, ".point.size", -1d);
    if(pPizza.containsKey(pPrefix+".point.color")) pColor = pPizza.getProperty(pPrefix+".point.color", null);
    if(validString(pColor) || (pRad >= 0d)) setPointData(pRad, pColor);
    if(!((dCX == null) && (dCY == null) && (dRAD == null))) setData(dCX, dCY, dRAD);
  }

  /** Set the color and size of all points creates subsequently. Any size less than DEFAULT_POINT_RADIUS becomes DEFAULT_POINT_RADIUS, and setting color to null 
      or blank restores the default (black).
      @param pSize The radius of the point. Minimum value of DEFAULT_POINT_RADIUS.
      @param pColor The stroke and fill color as a String.
  */
  public static final void setPointData(double pSize, String pColor) {
    pointRadius = (pSize < DEFAULT_POINT_RADIUS)?DEFAULT_POINT_RADIUS:pSize;
    if(Tag.validString(pColor)) {
      pointStyle = "stroke-width:1.0; stroke:"+pColor.trim()+"; fill:"+pColor.trim()+";";
    } else {
      pointStyle = DEFAULT_POINT;
    }
  }

  /** Instance version. */
  public void setNewPointData(double pSize, String pColor) { setPointData(pSize, pColor); }

  /** Create and return a point as a String. Since it is a point, only the coordinates are required. The point appears at (pX, pY). */
  public static String getPoint(double pX, double pY) {
    StringBuffer retval = new StringBuffer("<circle cx=\"").append(pX).append("\" cy=\"").append(pY).append("\" ");
    retval.append("r=\"").append(pointRadius).append("\" style=\"").append(pointStyle).append("\" />");
    return retval.toString();
  }

  /** Instance method to create a point. */
  public String getNewPoint(double pX, double pY) { return getPoint(pX, pY); }

}

