package cox.jmatt.java.MathTools.markup.svg;

import java.util.Hashtable;
import cox.jmatt.java.MathTools.markup.*;

/** <p>This class provides the abstract superclass for the SVG tags that can contain other tags (container tags) and the methods to add those tags quickly, easily, 
    and conveniently (drawable). These include but are not limited to the grouping tags: &lt;g&gt;, &lt;symbol&gt;, &lt;defs&gt;, and &lt;svg&gt;. DCTag contains 
    the necessary methods to add tags where they need to be and to handle the content from <b>all</b> subclasses.</p>
    <p>The &lt;defs&gt; block (tag) is handled implicitly. Since it contains CSS stylesheets, groups, and symbols it is handled internally and never exposed. All 
    elements that should go there do.</p>
    <p>Note on default methods: This class extends <code>StyleTag</code>, therefore it inherits all the methods to set default style and transform values. <b>These 
    methods should not be used!!</b> Each <code>ContainerTag</code> should be customized individually, not by default! None of the methods is specifically blocked, 
    so they can be used if truly necessary, but such necessity should represent only the most extreme cases!</p>
*/
public abstract class DrawableContainerTag extends StyleTag {

  private StringBuffer myCSS = new StringBuffer(), myDefs = new StringBuffer(), myContent = new StringBuffer();
  private String myTitle = null, myDescription = null;

  /** Subclass constructor necessary for <code>StyleTag</code> functionality. */
  protected DrawableContainerTag(StyleTag.TagKeys pKey) {
    super(pKey);
  }

  /** This method is only implemented to <b>clear</b> content. Any value other than null or blank is ignored, and those two merely clear all content! */
  protected void _setContent(String pContent) {
    if(validString(pContent)) return;
    myCSS = new StringBuffer();
    myDefs = new StringBuffer();
    myContent = new StringBuffer();
    myTitle = null;
    myDescription = null;
  }

  /** Collect and present all the necessary content elements and other goodies in proper order and in proper form. Newlines are inserted where appropriate along 
      with descriptive comments.
  */
  protected String _content() {
    StringBuffer retval = new StringBuffer();
    if(validString(myTitle)) retval.append("<title>").append(myTitle).append("</title>\n");
    if(validString(myDescription)) retval.append("<desc>").append(myDescription).append("</desc>\n");
    if(this.canHaveDefs()) {
      retval.append("\n<!-- Definitions block -->\n<defs>\n");
      if(myCSS.length() > 0) {
        retval.append("<style type=\"text/css\" ><![CDATA[\n");
        retval.append(myCSS.toString());
        retval.append("]]></style>\n");
      }
      if(myDefs.length() > 0) retval.append(myDefs.toString());
      retval.append("</defs>\n<!-- Tag content -->");
    }
    retval.append(myContent.toString()).append("\n");
    return retval.toString();
  }

  /** Method to set the &lt;title&lt; tag, which is maintained here but protected for only those tags where it is allowed. Setting null or blank clears it.*/
  protected void _setTitle(String pTitle) { myTitle = pTitle; }

  /** Protected method to set the &lt;desc&gt; tag, which is also maintained here. */
  protected void _setDesc(String pDesc) { myDescription = pDesc; }

  /** <p>Method to set the 'viewBox=' attribute, for those tags that need it. The argument is String, so care should be used in setting it.</p>
      <p>The viewBox is set via a String of four space-separated numbers. The first two represent the minimum x- and y-values for the viewport, the last two are 
      its maxima. This establishes the 'user coordinates' used for all other tags, so set carefully! It is generally best to set the viewBox once only and on the 
      root &lt;svg&gt; tag. These values should also be set to a convenient number, like 100 user units per whatever unit is used on the root tag. So, for U.S. 
      letter-sized paper (8.5 in. by 11 in.) the viewBox is '0 0 850 1100'.</p>
  */
  protected void _setViewBox(String pBox) { _addAttribute("viewBox", pBox); }

  /** Method to set an attribute value to 'objectBoundingBox' (false) or 'userSpaceOnUse' (true). Since several attributes can have these values, the name of 
      attribute is also supplied.
      @param pAttr The attribute whose value is 'objectBoundingBox' or 'userSpaceOnUse'.
      @param isUserSpace true for 'userSpaceOnUse' false for 'objectBoundingBox'.
  */
  protected void _setOBUS(String pAttr, boolean isUserSpace) { _addAttribute(pAttr, isUserSpace?"userSpaceOnUse":"objectBoundingBox"); }

  /** <p>Set the 'preserveAspectRatio=' attribute on tags that can take it. This method expects a one-, two-, or three-character String. Each character causes 
      the following actions:
      <ul>
        <li>'0' (zero) means the 'none' specification. Other characters are unnecessary.</li>
        <li>'m'/'M' is the 'meet' specifier. Objects are scaled proportionally to their least dimension.</li>
        <li>'s'/'S' means 'slice'. The object is scaled proportionally to its greatest dimension.</li>
        <li>'n' sets 'xMin', and 'YMin' if the y-specifier hasn't been set.</li>
        <li>'d' sets 'xMid', and 'YMin' if the y-specifier hasn't been set.</li>
        <li>'x' sets 'xMax', and 'YMax' if the y-specifier hasn't been set.</li>
        <li>'N' (uppercase) sets 'YMin' for the y-specifier.</li>
        <li>'D' sets 'YMid' for the y-specifier.</li>
        <li>'X' sets 'YMax' for the y-specifier.</li>
      </ul></p>
      <p>Other characters have no effect, nor does the order in which the characters occur. The first three specifiers determine what to do when an image overlaps 
      its bounding box. The remaining characters determine what to do if 'meet' or 'slice' is specified. Setting an 'x' specifier automatically sets the same 'y' 
      specifier unless 'y' is already set. Setting a 'y' does not affect the 'x'.</p>
  */
  protected void _setPAR(String par) {
    if(!validString(par)) return;
    String xVal = "xMid", yVal = null, overlap = " meet";
    for(int i = 0; i < par.length(); i++) {
      switch(par.charAt(i)) {
        case'0': overlap = " none"; break;
        case'm': case 'M': overlap = " meet"; break;
        case's': case 'S': overlap = " slice"; break;
        case'n': xVal = "xMin"; yVal = validString(yVal)?yVal:"YMin"; break;
        case'd': xVal = "xMid"; yVal = validString(yVal)?yVal:"YMax"; break;
        case'x': xVal = "xMax"; yVal = validString(yVal)?yVal:"YMax"; break;
        case'N': yVal = "YMin"; break;
        case'D': yVal = "YMid"; break;
        case'X': yVal = "YMax"; break;
        default:
      }
    }
    yVal = validString(yVal)?yVal:"YMid";
    _addAttribute("preserveAspectRatio", xVal+yVal+overlap);
  }

  /** Add a rule to the CSS stylesheet. Protected for the &lt;svg&gt; tag. A newline is appended after each rule. Ignored if null or blank. */
  protected void _addCSS(String pRule) {
    if(validString(pRule)) myCSS.append(pRule).append("\n");
  }

  /** Add something to the &lt;defs&gt; block, followed by a newline. Ignored if null or blank. */
  protected void _addDef(String pDef) {
    if(validString(pDef)) myDefs.append(pDef).append("\n");
  }

  /** Add something to the internal content. Null or blank is ignored. */
  protected void _addContent(String pContent) {
    if(validString(pContent)) myContent.append(pContent);
  }

  /** <p>Add a <code>StyleTag</code> to the internal content. If the tag is null nothing happens. This method automatically detects tags that must appear in the 
      &lt;defs&gt; block and puts them there, others go to content. If 'pTag' is supposed to appear only in the &lt;defs&gt; block and the subclass adding it does 
      not have one the tag is ignored. The tags are converted to String when added so changing them afterward will not affect the data added.</p>
      <p>Technically, a group tag ('&lt;g&gt;') <i>should</i> appear in the definitions but it <i>can</i> appear in the body of the document. In order to do this 
      <code>groupTagInstance.toString()</code> must be added via the <code>addRaw()</code> method. This also applies to other such tags: defs-block-inclusion 
      takes priority.</p>
  */
  public DrawableContainerTag add(StyleTag pTag) {
    if(pTag == null) return this;
    if(pTag.isInDefs()) {
      if(this.canHaveDefs()) _addDef(pTag.toString());
    } else {
      _addContent(pTag.toString());
    }
    return this;
  }

  /** Add a pre-String'ed tag to the internal content. Use this method carefully: the only safety feature is disregarding null or blank arguments! */
  public DrawableContainerTag addRaw(String pTagString) {
    _addContent(pTagString);
    return this;
  }

  /** Add a &lt;line&gt; tag to the current contents. The line is from (X1, Y1) to (X2, Y2). */
  public DrawableContainerTag drawLine(Double X1, Double Y1, Double X2, Double Y2) {
    add(new LineTag(X1, Y1, X2, Y2));
    return this;
  }

  /** Add a &lt;circle&gt; tag to the current contents. The center is (X1, Y1) and the radius pRadius. */
  public DrawableContainerTag drawCircle(Double X1, Double Y1, Double pRadius) {
    add(new CircleTag(X1, Y1, pRadius));
    return this;
  }

  /** Add a point to the current graphic. It appears at (X1, Y1). Default point configuration is settable in <code>MathDrawSVG</code>. */
  public DrawableContainerTag drawPoint(double X1, double Y1) {
    _addContent(CircleTag.getPoint(X1, Y1));
    return this;
  }

  /** Add an &lt;ellipse&gt; tag to the current contents. The center is (X1, Y1) and the radii are radX and radY. */
  public DrawableContainerTag drawEllipse(Double X1, Double Y1, Double radX, Double radY) {
    add(new EllipseTag(X1, Y1, radX, radY));
    return this;
  }

  /** Add a &lt;rectangle&gt; tag to the current contents. The upper left corner is (X1, Y1) with width pWidth and height pHeight. */
  public DrawableContainerTag drawRectangle(Double X1, Double Y1, Double pWidth, Double pHeight) {
    add(new RectangleTag(X1, Y1, pWidth, pHeight));
    return this;
  }

  /** Add a rounded &lt;rectangle&gt; tag. The upper left corner is (X1, Y1) with width pWidth and height pHeight; radX and radY are the corner radii. */
  public DrawableContainerTag drawRectangle(Double X1, Double Y1, Double pWidth, Double pHeight, Double radX, Double radY) {
    add(new RectangleTag(X1, Y1, pWidth, pHeight, radX, radY));
    return this;
  }

  /** Add text to the current contents. The anchor point is (X1, Y1) and the text is 'pText'. Null or blank text is ignored. */
  public DrawableContainerTag drawText(Double X1, Double Y1, String pText) {
    if(validString(pText)) add(new TextTag(X1, Y1, pText));
    return this;
  }

  /** Add a pre-defined &lt;path&gt; to the current content. Ignored if null or blank. */
  public DrawableContainerTag drawPath(String pPath) {
    if(validString(pPath)) add(new PathTag().setPath(pPath));
    return this;
  }

  /** Add a poly<u>gon</u> to the current tag. The argument should be a String with an even number of space-separated (numeric) values. Ignored if null or blank. */
  public DrawableContainerTag drawPolygon(String pPoints) {
    if(validString(pPoints)) add(new PolygonTag().setPolygon(true).addPoints(pPoints));
    return this;
  }

  /** Add a poly<u>line</u> to the current tag. Argument per <code>addPolygon()</code>. */
  public DrawableContainerTag drawPolyline(String pPoints) {
    if(validString(pPoints)) add(new PolygonTag().setPolygon(false).addPoints(pPoints));
    return this;
  }

}

