package cox.jmatt.java.MathTools.markup.svg;

import java.util.Hashtable;
import cox.jmatt.java.MathTools.markup.*;

/** <p>This class provides both &lt;linearGradient&gt; and &lt;radialGradient&gt; tags. Tag-speciic methods affect only tags of that type. Although this class 
    does extend <code>StyleTag</code>, only the defaults and methods related to transforms should be used.</p>
*/
public class GradientTag extends StyleTag {

  private static boolean copyRadial = false;

  private boolean amIRadial = false;
  private StringBuffer sbContent = new StringBuffer();

  /** Standard constructor for a linear gradient. */
  public GradientTag() {
    super(StyleTag.TagKeys.GRADIENT_TAG);
  }

  /** By default gradient tags are linear. Use this method to change that. */
  public GradientTag setAsRadial(boolean isRadial) {
    myTagName = isRadial?"radialGradient":"linearGradient";
    amIRadial = isRadial;
    return this;
  }

  /** Implemented to handle stops and other internal content. */
  protected String _content() { return sbContent.toString(); }

  /** Implemented to receive copied content and <code>reset()</code> function. */
  protected void _setContent(String pContent) {
    if(validString(pContent)) {
      sbContent = new StringBuffer(pContent);
    } else {
      sbContent = new StringBuffer();
    }
  }

  /** Create and return a copy of this gradient. All data will be copied! */
  public final GradientTag copyOf() {
    copyRadial = amIRadial;
    _enableCopy();
    GradientTag retval = new GradientTag();
    retval.setAsRadial(copyRadial);
    copyRadial = false;
    return retval;
  }

  /** Add a &lt;stop&gt; element to the gradient. The offset automatically has a percent sign ('%') appended. The pColor argument is mandatory but 'pOpacity' 
      is not.
      @param pOffset The (percent) value at which the color stop is to occur.
      @param pColor The color value at this stop.
      @param pOpacity The opacity value at this stop or null to omit. If present, it must be between 0.0 and 1.0.
  */
  public GradientTag addStop(double pOffset, String pColor, Double pOpacity) {
    if(!validString(pColor)) return this;
    sbContent.append("<stop offset=\"").append(pOffset).append("%\" style=\"stop-color: ").append(pColor).append(";");
    if(pOpacity != null) {
      double dOpacity = pOpacity.doubleValue();
      if(dOpacity < 0.0) dOpacity = 0.0;
      if(dOpacity > 1.0) dOpacity = 1.0;
      sbContent.append(" stop-opacity: ").append(dOpacity).append(";");
    }
    sbContent.append("\" />\n");
    return this;
  }

  /** <p>Determine how a 'shortened' gradient will spread outside its defined transition area. Since there are only three values, the argument passed is char.
      <ul>
        <li>'p' sets the value to <u>p</u>ad. (Default, if this method is called.)</li>
        <li>'e' sets the value to r<u>e</u>peat.</li>
        <li>'f' sets the value to re<u>f</u>lect.</li>
      </ul>
      Uppercase will also work.</p>
  */
  public GradientTag setSpreadMethod(char pValue) {
    switch(pValue) {
    case 'e':
    case 'E':
      _addAttribute("spreadMethod", "repeat");
      break;
    case 'f':
    case 'F':
      _addAttribute("spreadMethod", "reflect");
      break;
    default:
      _addAttribute("spreadMethod", "pad");
    }
    return this;
  }

  /** Set the transition line for a linear gradient that doesn't completely fill its designated area. This method has NO EFFECT if called on a &lt;radialGradient&gt;. 
      The transition line is from the point (pX1, pY1) to (pX2, pY2).
  */
  public GradientTag setTransitionLine(Double pX1, Double pY1, Double pX2, Double pY2) {
    if(amIRadial) return this;
    _addNAttribute("x1", pX1);
    _addNAttribute("x2", pX2);
    _addNAttribute("y1", pY1);
    _addNAttribute("y2", pY2);
    return this;
  }

  /** Set the transition circle for a (radial) gradient that does not completely fill its designated area. No effect on a &lt;linearGradient&gt;.
      @param pCX The X-coordinate of the center.
      @param pCY The Y-coordinate of the center.
      @param pRad The radius of the transition area.
  */
  public GradientTag setCircle(Double pCX, Double pCY, Double pRad) {
    if(!amIRadial) return this;
    _addNAttribute("cx", pCX);
    _addNAttribute("cy", pCY);
    _addNAttribute("r", pRad);
    return this;
  }

  /** Set the focal point of the (radial) gradient if it is not the center. The new focal point will be (pFX, pFY). No effect on a &lt;linearGradient&gt;. */
  public GradientTag setFocus(Double pFX, Double pFY) {
     if(!amIRadial) return this;
    _addNAttribute("fx", pFX);
    _addNAttribute("fy", pFY);
    return this;
  }

}

