package cox.jmatt.java.MathTools.markup.svg;

import cox.jmatt.java.MathTools.markup.*;

/** <p>This class implements the &lt;marker&gt; tag. It is a &lt;defs&gt; defs-block element used to place markers along a path or line of some type. A valid ID 
    is mandatory. There is no public constructor and the factory method will NOT build a tag around a null or blank ID.</p>
*/
public class MarkerTag extends DrawableContainerTag {

  private String myID = null;

  private MarkerTag() { super(StyleTag.TagKeys.MARKER_TAG); }

  /** This is the only way to construct an instance of this class. If the ID argument is null or blank the return value is null! */
  public static MarkerTag newInstance(String pTagID) {
    if(!validString(pTagID)) return null;
    MarkerTag retval = new MarkerTag();
    retval.setID(pTagID);
    return retval;
  }

  /** Overridden to ensure the ID set is valid. */
  public void setID(String pID) {
    if(!validString(pID)) return;
    myID = pID;
    super.setID(pID);
  }

  /** Overridden to preserve the ID. To change it use <code>reset(String)</code> or re-set manually afterward. */
  public void reset() {
    super.reset();
    setID(myID);
  }

  /** Call this method to clear data <b>and</b> re-set the ID in one fell swoop. */
  public void reset(String newID) {
    super.reset();
    setID(newID);
  }

  
  /** Set the relative starting point for the marker. This point is given in reference to the <b>marker's</b> coordinate system and is the point that should be 
      used for the starting point in the <i>external</i> coordinate system. If either argument is null or blank it is ignored.
  */
  public MarkerTag setRererencePoint(Double pStartX, Double pStartY) {
    _addNAttribute("refX", pStartX);
    _addNAttribute("refY", pStartY);
    return this;
  }

  /** Set the orientation of the marker. Setting 'auto' makes it rotate to point along the path/line segment it is on. Setting a number rotates it that many 
      degrees, period. Setting null or blank clears the property.
      @param pValue The word 'auto' or a number of degrees to rotate.
  */
  public MarkerTag setOrient(String pValue) {
    _addAttribute("orient", pValue);
    return this;
  }

  /** Set the scaling of the marker. If pStrokeWidth is true it sets the marker's coordinate system so that one unit equals the stroke width. If false, the units 
      are those of the external object referencing the marker ('userSpaceOnUse').
      @param pStrokeWidth true to set the coordinates to 'strokeWidth', false for 'userSpaceOnUse'.
  */
  public MarkerTag setUnits(boolean pStrokeWidth) {
    _addAttribute("markerUnits", (pStrokeWidth?"strokeWidth":"userSpaceOnUse"));
    return this;
  }

  /** Provide viewbox configuration for this class and <code>SVGTag</code>.
      @see DrawableContainerTag#_setViewBox
  */
  public MarkerTag setViewBox(String pBox) {
    _setViewBox(pBox);
    return this;
  }

  /** Set the 'preserveAspectRatio' value for this tag.
      @see DrawableContainerTag#_setPAR
  */
  public MarkerTag setPAR(String pRatio) {
    _setPAR(pRatio);
    return this;
  }

}

