package cox.jmatt.java.MathTools.markup.svg;

import cox.jmatt.java.MathTools.markup.*;

/** <p>This class implements the &lt;mask&gt; tag. It is a &lt;defs&gt; block and is used to describe a mask. As with <code>ClipTag</code> a valid ID is mandatory, 
    so there is no public constructor. The static method will NOT create a tag around a null or blank ID.</p>
    <p>To mask a graphic, first define a <code>MaskTag</code> and use it to describe the mask to be applied. Add it to the base document. To apply the mask, 
    set the mask URL on the tag to which it applies, using the ID assigned when creating the mask.</p>
*/
public class MaskTag extends DrawableContainerTag {

  private String myID = null;

  private MaskTag() { super(StyleTag.TagKeys.MASK_TAG); }

  /** This is the only way to construct an instance of this class. If the ID argument is null or blank the return value is null! */
  public static MaskTag newInstance(String pTagID) {
    if(!validString(pTagID)) return null;
    MaskTag retval = new MaskTag();
    retval.setID(pTagID);
    return retval;
  }

  /** Overridden to ensure the ID set is valid. */
  public void setID(String pID) {
    if(!validString(pID)) return;
    myID = pID;
    super.setID(pID);
  }

  /** Overridden to preserve the ID. To change it use <code>reset(String)</code> or re-set manually afterward. */
  public void reset() {
    super.reset();
    setID(myID);
  }

  /** Call this method to clear data <b>and</b> re-set the ID in one fell swoop. */
  public void reset(String newID) {
    super.reset();
    setID(newID);
  }

  /** Set the dimensions of the &lt;mask&gt; relative to the masked object's bounding box. For user coordinates, <code>setMaskUnits(true)</code>. If any of these 
      numbers is null it is ignored.
      @param pX The X-coordinate of the clip rectangle's upper left corner.
      @param pY The Y-coordinate of the clip rectangle's upper left corner.
      @param pWidth The width of the mask's clipping rectangle.
      @param pHeight The height of the mask clipping rectangle.
  */
  public MaskTag setSize(Double pX, Double pY, Double pWidth, Double pHeight) {
    _addNAttribute("x", pX); _addNAttribute("y", pY);
    _addNAttribute("width", pWidth); _addNAttribute("height", pHeight);
    return this;
  }
  /** Set the 'maskUnits=' attribute for this tag. This is 'objectBoundingBox' or 'userSpaceOnUse'.
      @see DrawableContainerTag#_setOBUS
  */
  public MaskTag setMaskUnits(boolean pUserSpace) {
    _setOBUS("maskUnits", pUserSpace);
    return this;
  }

  /** Set the reference coordinates for elements contained within the &lt;mask&gt; tag.
      @see DrawableContainerTag#_setOBUS
  */
  public MaskTag setMaskContentUnits(boolean pUserSpace) {
    _setOBUS("maskContentUnits", pUserSpace);
    return this;
  }

}

