package cox.jmatt.java.MathTools.markup.svg;

import java.util.*;
import cox.jmatt.java.MathTools.markup.*;

/** <p>This class encapsulates the SVG &lt;path&gt; tag. It provides a quick way to set the path explicitly and commands to add elements to it. Each command comes 
    in a relative and absolute flavor. The relative versions start with lowercase letters and the absolute versions with uppercase. The arguments for each flavor 
    are the same.</p>
*/
public class PathTag extends StyleTag {

  private StringBuffer myPath = new StringBuffer();

  /** Automatically constructs a &lt;path&gt; tag. */
  public PathTag() {
    super(StyleTag.TagKeys.PATH_TAG);
  }

  /** Overridden to clear out any path data. Returns a self-reference. */
  public void reset() {
    super.reset();
    myPath = new StringBuffer();
  }

  /** Create a copy of this tag. */
  public PathTag copyOf() {
    if(myPath.length() > 0) _addAttribute("d", myPath.toString());
    _enableCopy();
    return new PathTag();
  }

  /** Overridden to add the actual path to the tag at the last instant. */
  public String toString() {
    if(myPath.length() > 0) _addAttribute("d", myPath.toString());
    return super.toString();
  }

  /** Set the path to the String supplied, or clear it with null or blank. IMPORTANT NOTE: Calling this method clears <b>all</b> previous path data. */
  public PathTag setPath(String pPath) { myPath = new StringBuffer((Tag.validString(pPath)?pPath:"")); return this; }

  /** Close the current (sub-) path. There is no absolute version of this command. */
  public PathTag closePath() { myPath.append("z"); return this; }

  /** Move to the specified point relative to the current location.
      @param pX The X-coordinate of the destination.
      @param pY The destination Y-cooridnate.
      @return A reference to this.
  */
  public PathTag moveTo(double pX, double pY) { myPath.append("m").append(pX).append(" ").append(pY); return this; }

  /** Move to the specified point, absolute coordinates.
      @param pX The X-coordinate of the destination.
      @param pY The destination Y-cooridnate.
      @return A reference to this.
  */
  public PathTag MoveTo(double pX, double pY) { myPath.append("M").append(pX).append(" ").append(pY); return this; }

  /** Draw a line to the specified point relative to the current location.
      @param pX The X-coordinate of the endpoint.
      @param pY The Y-coordinate of the endpoint.
      @return A self-reference.
  */
  public PathTag lineTo(double pX, double pY) { myPath.append("l").append(pX).append(" ").append(pY); return this; }

  /** Draw a line to the specified point, absolute version.
      @param pX The X-coordinate of the endpoint.
      @param pY The Y-coordinate of the endpoint.
      @return A self-reference.
  */
  public PathTag LineTo(double pX, double pY) { myPath.append("L").append(pX).append(" ").append(pY); return this; }

  /** Draw a horizontal line to the specified point, relative version.
      @param pEnd The (relative) endpoint of the horizontal segment.
  */
  public PathTag hLineTo(double pEnd) { myPath.append("h").append(pEnd); return this; }

  /** Draw a horizontal line to the specified point, absolute version.
      @param pEnd The (absolute) endpoint of the horizontal segment.
  */
  public PathTag HLineTo(double pEnd) { myPath.append("H").append(pEnd); return this; }

  /** Draw a vertical line to the specified point, relative version.
      @param pEnd The (relative) endpoint of the vertical segment.
  */
  public PathTag vLineTo(double pEnd) { myPath.append("v").append(pEnd); return this; }

  /** Draw a vertical line to the specified point, absolute version.
      @param pEnd The (absolute) endpoint of the vertical segment.
  */
  public PathTag VLineTo(double pEnd) { myPath.append("V").append(pEnd); return this; }

  /** Draw an elliptical arc, relative version.
      @param pRadX The X-radius of the arc.
      @param pRadY The Y-radius of the arc.
      @param pXRot The X-axis rotation.
      @param pLargeArc false (0) if the arc is less than 180 degrees, true (1) if it is greater.
      @param pSweep false (0) to draw in the negative angle direction, true (1) for the positive angle direction.
      @param pEndX The X-coordinate of the endpoint.
      @param pEndY The Y-coordinate of the endpoint.
      @return A reference to this.
  */
  public PathTag arcTo(double pRadX, double pRadY, double pXRot, boolean pLargeArc, boolean pSweep, double pEndX, double pEndY) {
    myPath.append("a").append(pRadX).append(" ").append(pRadY).append(" ").append(pXRot).append(" ");
    myPath.append(pLargeArc?"1 ":"0 ").append(pSweep?"1 ":"0 ").append(pEndX).append(" ").append(pEndY);
    return this;
  }

  /** Draw an elliptical arc, absolute version. Arguments are per the relative flavor. */
  public PathTag ArcTo(double pRadX, double pRadY, double pXRot, boolean pLargeArc, boolean pSweep, double pEndX, double pEndY) {
    myPath.append("A").append(pRadX).append(" ").append(pRadY).append(" ").append(pXRot).append(" ");
    myPath.append(pLargeArc?"1 ":"0 ").append(pSweep?"1 ":"0 ").append(pEndX).append(" ").append(pEndY);
    return this;
  }

  /** Draw a quadratic bezier curve to the specified endpoint relative to the current location.
      @param pCX The X-coordinate of the control point.
      @param pCY The Y-coordinate of the control point.
      @param pX The X-coordinate of the endpoint.
      @param pY The Y-coordinate of the endpoint.
      @return A reference to this.
  */
  public PathTag qCurveTo(double pCX, double pCY, double pX, double pY) {
    myPath.append("q").append(pCX).append(" ").append(pCY).append(" ").append(pX).append(" ").append(pY);
    return this;
  }

  /** Draw a quatratic curve, absolute version. Arguments are per the relative method. */
  public PathTag QCurveTo(double pCX, double pCY, double pX, double pY) {
    myPath.append("Q").append(pCX).append(" ").append(pCY).append(" ").append(pX).append(" ").append(pY);
    return this;
  }

  /** Draw a smooth quadratic polybezier curve; only the endpoint is required, the control point is calcuated automatically. This is the 't' command.
      @param pX The endpoint X-coordinate.
      @param pY The endpoint Y-coordinate.
      @return A reference to this to allow easy chaining.
  */
  public PathTag tCurveTo(double pX, double pY) { myPath.append("t").append(pX).append(" ").append(pY); return this; }

  /** Draw a smooth quadratic polybezier curve, absolute version. The 'T' command. */
  public PathTag TCurveTo(double pX, double pY) { myPath.append("T").append(pX).append(" ").append(pY); return this; }

  /** Draw a cubic bezier curve to the specified endpoint relative to the current location.
      @param pC1X The X-coordinate of the first control point.
      @param pC1Y The Y-coordinate of the first control point.
      @param pC2X The X-coordinate of the second control point.
      @param pC2Y The Y-coordinate of the second control point.
      @param pX The X-coordinate of the endpoint.
      @param pY The Y-coordinate of the endpoint.
      @return A reference to this.
  */
  public PathTag cCurveTo(double pC1X, double pC1Y, double pC2X, double pC2Y, double pX, double pY) {
    myPath.append("c").append(pC1X).append(" ").append(pC1Y).append(" ").append(pC2X).append(" ").append(pC2Y).append(" ").append(pX).append(" ").append(pY);
    return this;
  }

  /** Absolute version of the cubic curve. Same arguments as the relative version. */
  public PathTag CCurveTo(double pC1X, double pC1Y, double pC2X, double pC2Y, double pX, double pY) {
    myPath.append("C").append(pC1X).append(" ").append(pC1Y).append(" ").append(pC2X).append(" ").append(pC2Y).append(" ").append(pX).append(" ").append(pY);
    return this;
  }

  /** Add a smooth cubic polybezier curve. Only the second control and endpoint are required. The 's' command, relative version.
      @param pCX The control point X-coordinate.
      @param pCY The control point Y-coordinate.
      @param pX The X-coordinate for the endpoint.
      @param pY The Y-coordinate of the endpoint.
      @return A reference to this for method chaining.
  */
  public PathTag sCurveTo(double pCX, double pCY, double pX, double pY) {
    myPath.append("s").append(pCX).append(" ").append(pCY).append(" ").append(pX).append(" ").append(pY);
    return this;
  }

  /** Smooth cubic polybezier curve, absolute version. The 'S' command. */
  public PathTag SCurveTo(double pCX, double pCY, double pX, double pY) {
    myPath.append("S").append(pCX).append(" ").append(pCY).append(" ").append(pX).append(" ").append(pY);
    return this;
  }

}
