package cox.jmatt.java.MathTools.markup.svg;

import cox.jmatt.java.MathTools.markup.*;

/** <p>This class encapsulates the &lt;svg&gt; tag. By default this is a 'root' tag, meaning it will automatically have the '&lt;?xml ... ?&gt;' declaration 
    prepended when <code>toString()</code> is called. Change this with <code>setAsRoot()</code>. Similarly, since this is the ONLY tag allowed to have a &lt;defs&gt; 
    block, this is the only tag class that can add a <code>StylePen</code>.</p>
    <p>Root SVG tags must have a width and a height, therefore a String width and height are requred to create one. Since there are some conditions under which 
    an &lt;svg&gt; tag might not need them, they can be cleared. Use the <code>setSize()</code> method to do this, but do so carefully!</p>
    <h3>SVG and Units</h3>
    <p>By the SVG specification almost all numeric attributes are allowed to have units associated with them. MathTools does not allow this: all values on the 
    other tags are given in <i>user units</i>. User units are specified in a tag's 'viewBox=' attribute, for those that have it, or a parent's attribute for those 
    that don't. Since the root tag of an SVG document is an &lt;svg&gt; tag, it <b>must</b> have its viewport specified! The viewport should be specified for 
    a convenient number of user-units per unit given for the root tag's height and width.</p>
*/
public class SVGTag extends SymbolTag {

  /** This contains the SVG namespace URI. */
  public static final String NSURI = "http://www.w3.org/2000/svg";
  /** This contains the 'xmlns:xlink' namespace URI. */
  public static final String NSURI_XLINK = "http://www.w3.org/1999/xlink";

  private boolean amIRoot = true;

  private SVGTag(String pWidth, String pHeight) {
    super(StyleTag.TagKeys.SVG_TAG);
    _addAttribute("xmlns", NSURI);
    _addAttribute("xmlns:xlink", NSURI_XLINK);
    _addAttribute("width", pWidth);
    _addAttribute("height", pHeight);
  }

  /** An SVGTag must have width and height set. Period. Unless they are, this method returns null. Period. */
  public static final SVGTag newInstance(String pWidth, String pHeight) {
    if(!(validString(pWidth) && validString(pHeight))) return null;
    return new SVGTag(pWidth, pHeight);
  }

  /** By assumption any <code>SVGTag</code> generated is the root tag of a document. This means, when converted to String, that the XML declaration tag is 
      automatically prepended to the tag. If this &lt;svg&gt; tag should NOT, for whatever reason, have the beginning XML declaration use this method to disable 
      it.
  */
  public SVGTag setAsRoot(boolean isRoot) { amIRoot = isRoot; return this; }

  /** Overridden to handle root tag processing. */
  public String toString() {
    _setEnableXMLTag(amIRoot);
    return super.toString();
  }

  /** Overridden to preserve width and height during a reset. */
  public void reset() {
    String myWide = _attributes().get("width"), myHigh = _attributes().get("height");
    super.reset();
    setSize(myWide, myHigh);
  }

  /** This method can be used to re-set width and height or to clear them, by setting null. Use with care! */
  public void setSize(String pWidth, String pHeight) { _addAttribute("width", pWidth); _addAttribute("height", pHeight); }

  /** Since this is the only tag class allowed to have a &lt;defs&gt; block, this is the only class that needs to add <code>StylePen</code>s! Use this method to 
      do so. IMPORTANT NOTE: This is not the same as <i>setting</i> a pen, rather, this is the method that stores the data corresponding to that pen! This method 
      DOES prepend the dot for a class selector.
  */
  public SVGTag add(StylePen pPen) {
    if(pPen != null) _addCSS("."+pPen.toCSS());
    return this;
  }

}

