package cox.jmatt.java.MathTools.markup.svg;

import java.util.*;
import cox.jmatt.java.MathTools.markup.*;

/** <p>This is the SVG class used to encapsulate CSS rules and 'style=' attributes.</p>
*/
public class StylePen extends StyleInfo {

  public StylePen() { super(); }

  /** <code>StylePen</code>s are tracked by number; the name is handled automatically. This method does nothing. */
  public void setName(String ignored) {}

  /** Create a copy of this pen, <b>including the name!</b> Change it with <code>setPen()</code>. */
  public StylePen copyOf() {
    _enableCopy();
    return new StylePen();
  }

  /** Set the pen number used to reference this style, if it is to be a CSS 'pen' style. If the pen number is negative nothing is set. */
  public StylePen setPen(int pPen) {
    if(pPen > -1) super.setName("Pen_"+pPen);
    return this;
  }

  /** Set the stroke color. */
  public StylePen setStroke(String pValue) { _addDeclaration("stroke", pValue); return this; }

  /** Set the fill color. */
  public StylePen setFill(String pValue) { _addDeclaration("fill", pValue); return this; }

  /** Set the 'fill:' to an URL value. The value itself is set to 'url(<u>pURL</u>)'. The 'url' and parentheses are supplied, but if it is a local URL it must 
      still be prepended with the '#'. Setting null or blank clears the 'fill' setting.
  */
  public StylePen setFillURL(String pURL) {
    if(Tag.validString(pURL)) {
      _addDeclaration("fill", "url("+pURL+")");
    } else {
      _addDeclaration("fill", "");
    }
    return this;
  }

  /** Set a clip-path URL to the style specified. The 'url()' is supplied but the leading '#' for local URLs is not. Setting null clears the value. */
  public StylePen setClipURL(String pURL) {
    if(Tag.validString(pURL)) {
      _addDeclaration("clip-path", "url("+pURL+")");
    } else {
      _addDeclaration("clip-path", "");
    }
    return this;
  }

  /** Set a mask URL to the style specified. The 'url()' is supplied but the leading '#' for local URLs is not. Setting null clears the value. */
  public StylePen setMaskURL(String pURL) {
    if(Tag.validString(pURL)) {
      _addDeclaration("mask", "url("+pURL+")");
    } else {
      _addDeclaration("mask", "");
    }
    return this;
  }

  /** Set the <b>stroke</b> opacity. The value must be between 0.0 and 1.0*/
  public StylePen setOpacity(Double pValue) { _addNDeclaration("stroke-opacity", pValue, 0.0, 1.0); return this; }

  /** Set the width of the stroke. */
  public StylePen setStrokeWidth(Double pValue) { _addNDeclaration("stroke-width", pValue, 0.0, null); return this; }

  /** Set the <b>fill</b> opacity. */
  public StylePen setFillOpacity(Double pValue) { _addNDeclaration("fill-opacity", pValue, 0.0, 1.0); return this; }

  /** Set the stroke dash array. */
  public StylePen setStrokeDasharray(String pValue) { _addDeclaration("stroke-dasharray", pValue); return this; }

  /** Set the stroke dasharray offset. */
  public StylePen setStrokeDashOffset(String pValue) { _addDeclaration("stroke-dashoffset", pValue); return this; }

  /** <p>Set the stroke linecap attribute. The argument uses one <b>char</b> as there are limited valid values:
      <ul>
        <li>'b' is butt (the default value).</li>
        <li>'r' is round.</li>
        <li>'s' is square.</li>
      <ul></p>
  */
  public StylePen setStrokeLinecap(char pValue) {
    String sValue = "butt";
    switch(pValue) {
      case 'r': sValue = "round"; break;
      case 's': sValue = "square"; break;
      default:
    }
    _addDeclaration("stroke-linecap", sValue);
    return this;
  }

  /** <p>Set the stroke linejoin attribute. The argument uses one <b>char</b> as there are limited valid values:
      <ul>
        <li>'b' is bevel.</li>
        <li>'r' is round.</li>
        <li>'m' is miter (default).</li>
      <ul></p>
  */
  public StylePen setStrokeLinejoin(char pValue) {
    String sValue = "miter";
    switch(pValue) {
      case 'r': sValue = "round"; break;
      case 'b': sValue = "bevel"; break;
      default:
    }
    _addDeclaration("stroke-linejoin", sValue);
    return this;
  }

  /** Set the stroke miter limit. */
  public StylePen setStrokeMiterlimit(String pValue) { _addDeclaration("stroke-miterlimit", pValue); return this; }

  /** Set the fill-rule for this style. Since <b>nonzero</b> is the default and there are only two values, this method is named accordingly and takes a boolean 
      argument. If pNonZero is false, the fill-rule is <b>evenodd</b>.
  */
  public StylePen setFillNonzero(boolean pNonZero) { _addDeclaration("fill-rule", (pNonZero?"nonzero":"evenodd")); return this; }

}

