package cox.jmatt.java.MathTools.markup.svg;

import java.util.Hashtable;
import cox.jmatt.java.MathTools.markup.*;

/** <p>This class provides the basic machinery for the other SVG tags. It provides the mechanism for default styles and attributes the SVG implementation needs 
    beyond what <code>Tag</code> supplies.</p>
    <p>Every SVG tag class has default values for such things as style attributes, <code>StylePen</code>s, and transformations. The <code>TagKeys enum</code> keeps 
    up with this information and supplies it as needed. All the methods to set defaults are here; only the tag-critical methods are in the individual tag classes.</p>
*/
public class StyleTag extends Tag {

  /** This enum is used internally by the SVG classes to store defaults and other such information. */
  public static enum TagKeys {
    STYLE_TAG("StyleTag", false, false),
    CIRCLE_TAG("circle", false, false),
    ELLIPSE_TAG("ellipse", false, false),
    LINE_TAG("line", false, false),
    RECTANGLE_TAG("rect", false, false),
    POLYGON_TAG("polygon", false, false),
    TEXT_TAG("text", false, false),
    TSPAN_TAG("tspan", false, false),
    USE_TAG("use", false, false),
    PATH_TAG("path", false, false),
    CLIP_TAG("clipPath", false, true),
    MASK_TAG("mask", false, true),
    MARKER_TAG("marker", false, true),
    PATTERN_TAG("pattern", false, true),
    GROUP_TAG("g", false, true),
    SYMBOL_TAG("symbol", false, true),
    SVG_TAG("svg", true, false),
    GRADIENT_TAG("linearGradient", false, true);

    private String myStyle, myTagName, myTransforms, myPen, myContent;
    private Hashtable<String, String> myAttrs;
    private boolean hasDefs = false, isDef = false;
    TagKeys(String pTagName, boolean pHasDefs, boolean pIsDef) {
      this.myTagName = pTagName;
      myStyle = null;
      myTransforms = null;
      myPen = null;
      myContent = null;
      myAttrs = null;
      hasDefs = pHasDefs;
      isDef = pIsDef;
    }
    /** Can the tag have a 'defs' block or not? */
    public boolean canHaveDefs() { return hasDefs; }
    /** Does the tag belong in a 'defs' block ONLY? */
    public boolean isInDefs() { return isDef; }
    /** Set the default style. */
    public void setDefaultStyle(StylePen pStyle) {
      if(pStyle == null) {
        myStyle = null;
      } else {
        myStyle = pStyle.toRawCSS();
      }
    }
    /** Get the default style. */
    public String getDefaultStyle() { return myStyle; }
    /** Set the default pen number. */
    public void setDefaultPen(int pPen) {
      if(pPen < 0) {
        myPen = null;
      } else {
        myPen = "Pen_"+pPen;
      }
    }
    /** Get the default pen number. */
    public String getDefaultPen() { return myPen; }
    /** Set the default transforms. */
    public void setDefaultTransforms(TransformPen pTransforms) {
      if(pTransforms == null) {
        myTransforms = null;
      } else {
        myTransforms = pTransforms.toString(false);
      }
    }
    /** Get the default transforms. */
    public String getDefaultTransforms() { return myTransforms; }
    /** Get the tag name to create the SVG tags. */
    public String getTagName() { return myTagName; }
    /** Set the tag's content and attributes, if used for copying. */
    public void setTemp(Hashtable<String, String> tAtts, String tContent) {
      myAttrs = tAtts;
      myContent = tContent;
    }
    /** Fetch the attributes stored, once. */
    public Hashtable<String, String> getAttributes() {
      Hashtable<String, String> retval = myAttrs;
      myAttrs = null;
      return retval;
    }
    /** Fetch the stored content String, once. */
    public String getContent() {
      String retval = myContent;
      myContent = null;
      return retval;
    }
  }

  /* * * ENUM ENDS HERE * * */

  private TagKeys myTagKey = TagKeys.STYLE_TAG;

  /** This is the protected constructor used to build all SVG tags. The argument is a <code>TagKeys</code> constant, which contains all necessary information 
      about the tag's name and its current default settings. If subclassing, never, never, NEVER call this constructor with null!
  */
  protected StyleTag(TagKeys pTagKey) {
    super(pTagKey.getTagName());
    _addAttribute("class", pTagKey.getDefaultPen());
    _addAttribute("style", pTagKey.getDefaultStyle());
    _addAttribute("transform", pTagKey.getDefaultTransforms());
    myTagKey = pTagKey;
    _setContent(pTagKey.getContent());
    _addAttributes(pTagKey.getAttributes());
  }

  /** Set the StylePen number to be used for the tag. Any number less than -1 becomes -1 and disables the StylePen. */
  public void setPen(int pPen) {
    if(pPen < 0) {
      _addAttribute("class", null);
    } else {
      _addAttribute("class", "Pen_"+pPen);
    }
  }

  /** Set the 'style=' attribute for this tag. The argument is given as a <code>StylePen</code> but it is converted to String as soon as it is set. Changing the 
      data in the StylePen after calling this method will not affect the data here. If pPen is null the tag's StylePen is set to null and disabled.
      @param pPen The <code>StylePen</code> to apply to this tag.
  */
  public void setStyle(StylePen pPen) {
    if(pPen == null) {
      _addAttribute("style", null);
    } else {
      _addAttribute("style", pPen.toRawCSS());
    }
  }

  /** Set the transform(s) for this tag. This method calls <code>toString()</code> on the supplied <code>TransformPen</code> so any changes to the pen afterward 
      will not affect the transforms already set.
      @param pPen The <code>TransformPen</code> containing the desired transforms.
  */
  public void setTransform(TransformPen pPen) {
    if(pPen == null) {
      _addAttribute("transform", null);
    } else {
      _addAttribute("transform", pPen.toString(false));
    }
  }

  /** <p>This method establishes the pattern for pizza-configuring SVG package classes. The only configurable attributes are ID and pen number but they can apply 
      to any other tag class that calls this method. The 'pPrefix' String is for anything prepended to the prefixes described below. This allows different tag 
      classes to have different configuration lines in the overall pizza.</p>
      <p>The configuration options available from this method are:
        <ul>
          <li>'Style.ID' to set the ID attribute.</li>
          <li>'Style.pen' to set the default <code>StylePen</code> <i>number</i>. Non-parseable values are silently ignored.</li>
        </ul></p>
      <p>This is the pattern that is used for each configurable tag class. It has a tag-specific prefix for each identifier, so when the tag prefix is prepended 
      to 'Style.pen' and this method called, the pen number will be set. This is done to allow maximum flexibility in configuring these classes. If the prefix is 
      null or blank it is ignored but the keys listed above are still checked. If 'pPizza' is null or empty this method returns silently.</p>
      @param pPrefix The String prefix to prepend when extracting key values. If it does not end with a dot one will be supplied.
      @param pPizza The <code>Properties</code> object used to configure the class.
  */
  protected void _configureTag(String pPrefix, java.util.Properties pPizza) {
    if(badPizza(pPizza)) return;
    pPrefix = _fixKey(pPrefix, "Style");
    if(pPizza.containsKey(pPrefix+".ID")) setID(pPizza.getProperty(pPrefix+".ID"));
    if(pPizza.containsKey(pPrefix+".pen")) try { setPen(Integer.parseInt(pPizza.getProperty(pPrefix+".pen"))); } catch(Exception ignored) {}
  }

  /** Set the default pen number for this type of tag. This remains in effect until changed. Any number less than -1 becomes -1 and disables numbered style pens 
      by default.
  */
  public void setDefaultPen(int pPen) { myTagKey.setDefaultPen(pPen); }

  /** Set the default 'style=' attribute for this type of tag. Setting null disables it. */
  public void setDefaultStyle(StylePen pPen) { myTagKey.setDefaultStyle(pPen); }

  /** Set the default transformations for this type of tag. Setting null clears them. */
  public void setDefaultTransform(TransformPen pPen) { myTagKey.setDefaultTransforms(pPen); }

  /** <p>After calling this method, the next instantiation of a subclass (only) will instead produce a copy. This should ONLY be used in a copying method!
      <pre>
        public StyleTagSubclass copyOf() {
          _enableCopy();
          return new StyleTagSubclass();
        }
      </pre>
      Only the first instantiation after this method is called creates a copy; the data set is cleared when accessed. This method uses <code>_getContent()</code> 
      to set the content of the copied tag and the subsequent constructor calls <code>_setContent()</code>.</p>
  */
  protected final void _enableCopy() { myTagKey.setTemp(_attributes(), _content()); }

  /** Specify whether or not this subclass can have a &lt;defs&gt; block. */
  public boolean canHaveDefs() { return myTagKey.canHaveDefs(); }

  /** Specify whether or not this subclass should be within a &lt;defs&gt; block. */
  public boolean isInDefs() { return myTagKey.isInDefs(); }

}

