package cox.jmatt.java.MathTools.markup.svg;

import java.util.*;
import cox.jmatt.java.MathTools.markup.*;

/** <p>This <code>StylePen</code> subclass is designed and optimized for styling text. It contains methods to configure fonts and the way they should appear. Where 
    possible the method arguments are designed to enforce proper values. Where not possible, be careful! All MathTools SVG font configuration options are here, 
    but all <code>StylePen</code> methods are also available. Use as needed when needed.</p>
*/
public class TextPen extends StylePen {

  /** Standard constructor, nothing special. */
  public TextPen() { super(); }

  /** Return a copy of this instance. */
  public TextPen copyOf() {
    _enableCopy();
    return new TextPen();
  }

  /** This method sets the 'font-size' attribute. */
  public TextPen setFontSize(Double pValue) {
    _addNDeclaration("font-size", pValue, 0.0, null);
    return this;
  }

  /** Set the 'font-family' attribute. */
  public TextPen setFontFamily(String pValue) {
    _addDeclaration("font-family", pValue);
    return this;
  }

  /** Set the 'font-weight' attribute. Since this setting has limited, specific values it is set as a character. The characters used are:
      <ul>
        <li>'n' - <b>normal</b> weight. (Default value.)</li>
        <li>'L' - <b>lighter</b> weight.</li>
        <li>'b' - <b>bold</b> weight.</li>
        <li>'x' - <b>bolder</b> weight. (<u>X</u>-tra bold.)</li>
        <li>'1', '2', ... '9' - font weight <b>100, 200, ... 900</b>.</li>
      </ul>
      For non-numeric values the lowercase OR uppercase letters may be used.</p>
  */
  public TextPen setFontWeight(char pWeight) {
    String sWeight = "normal";
    switch(pWeight) {
      case 'l':
      case 'L':
        sWeight = "lighter";
        break;
      case 'b':
      case 'B':
        sWeight = "bold";
        break;
      case 'x':
      case 'X':
        sWeight = "bolder";
        break;
      case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
        sWeight = ""+pWeight+"00";
        break;
      case 'n':
      case 'N':
      default:
    }
    _addDeclaration("font-weight", sWeight);
    return this;
  }

  /** <p>Set the font style (inclination). There are three defined values:
      <ul>
        <li>'n' is <b>normal</b>. (Default value.)</li>
        <li>'i' is <b>italic</b>.</li>
        <li>'o' is <b>oblique</b>.</li>
      </ul>
      As with other <code>char</code>-argument methods, this one accepts both uppercase and lowercase values.</p>
  */
  public TextPen setFontStyle(char pStyle) {
    switch(pStyle) {
      case 'i':
      case 'I':
        _addDeclaration("font-style", "italic");
        break;
      case 'o':
      case 'O':
        _addDeclaration("font-style", "oblique");
        break;
      case 'n':
      case 'N':
      default:
        _addDeclaration("font-style", "normal");
    }
    return this;
  }

  /** Set the 'font-variant' attribute. Since there are only two possibilities, true sets <b>small-caps</b> and false sets <b>normal</b>.
      @param pSmallCaps true for small-caps, false for normal.
  */
  public TextPen setFontVariant(boolean pSmallCaps) {
    _addDeclaration("font-variant", (pSmallCaps?"small-caps":"normal"));
    return this;
  }

  /** <p>Set 'line-is-ness' of the text. This sets the 'text-decoration' property and valid values are:
      <ul>
        <li>'n' - <b>none</b>. (Default value.)</li>
        <li>'u' - <b>underline</b>.</li>
        <li>'o' - <b>overline</b>.</li>
        <li>'t' - <b>line-through</b>.</li>
        <li>'b' - <b>blink</b>. (YES it is valid, YES it is available, but <b>do not</b> use it!!!)</li>
      </ul>
      Both upper- and lower-case values for these letters will work.</p>
  */
  public TextPen setTextDecoration(char pValue) {
    switch(pValue) {
    case 'u':
    case 'U':
      _addDeclaration("text-decoration", "underline");
      break;
    case 'o':
    case 'O':
      _addDeclaration("text-decoration", "overline");
      break;
    case 't':
    case 'T':
      _addDeclaration("text-decoration", "line-through");
      break;
    case 'b':
    case 'B':
      _addDeclaration("text-decoration", "blink");
      break;
    case 'n':
    case 'N':
    default:
      _addDeclaration("text-decoration", "normal");
    }
    return this;
  }

  /** <p>Set the anchor (alignment) of the text relative to its coordinates. Valid values are:
      <ul>
        <li>'s' aligns the <b>start</b> of the text. (Default value.)</li>
        <li>'m' aligns the <b>middle</b> of the text.</li>
        <li>'e' aligns the <b>end</b> of the text.</li>
      </ul>
      Uppercase values also work.</p>
  */
  public TextPen setTextAnchor(char pAlign) {
    switch(pAlign) {
      case 'm':
      case 'M':
        _addDeclaration("text-anchor", "middle");
        break;
      case 'e':
      case 'E':
        _addDeclaration("text-anchor", "end");
        break;
      case 's':
      case 'S':
      default:
        _addDeclaration("text-anchor", "start");
      }
      return this;
  }

  /** Set the spacing between the letters of the text. This can be 'normal' for normal spacing, a positive number to increase the space or a negative one to 
      decrease it.
  */
  public TextPen setLetterSpacing(String pValue) {
    _addDeclaration("letter-spacing", pValue);
    return this;
  }

  /** Set the spacing between the words of the text. This is 'normal' for normal spacing, positive to increase the space or negative to decrease it. */
  public TextPen setWordSpacing(String pValue) {
    _addDeclaration("word-spacing", pValue);
    return this;
  }

}

