package cox.jmatt.java.MathTools.markup.svg;

import cox.jmatt.java.MathTools.markup.*;
import cox.jmatt.java.MathTools.MathXML;

/** <p>This class handles the SVG &lt;text&gt; tag. Some SVG tags have content, others may have child tags, but the &lt;text&gt; tag can have both, and at the 
    same time. For that reason adding content and child tags to this element requires special attention. All content, whether text or child elements, is kept and 
    output in the order in which it is entered. Child tags are converted to String as soon as they are added and may be interspersed with text content.</p>
    <p>The two SVG elements this tag can contain are &lt;tspan&gt; and &lt;textPath&gt;. The former is handled via the {@link TSpanTag TSpanTag} class and 
    has quite a few configurable options. The latter consists of text content, one mandatory argument and one optional so it is handled internally.</p>
    <p>This class has the usual default methods for transforms and styling, hence its subclass <code>TSpanTag</code> does as well. These should NOT be used in 
    the subclass, though, since it is techincally this class' responsibility. They are left unaltered, though, in case they are needed.</p>
    <h3>Text and Style</h3>
    <p>Text presentation has unique concerns outside other SVG elements. For that reason, styling text should be done via a <code>TextPen</code>. This has all the 
    options of a standard <code>StylePen</code> (which is its superclass) along with text-specific methods and options. <code>TextPen</code>s are optimized for 
    text and for this class.</p>
    <p><code>TextTag</code> is not <i>limited</i> to <code>TextPen</code>s though! If an ordinary <code>StylePen</code> will do the job there is no reason not 
    to use one! All of the <code>Pen</code> classes are really just a thin veneer over CSS styling, so any pen can work anywhere. Use as needed and where needed. 
    There are no such concerns with <code>TransformPen</code>, though. It is actually closer to structure than presentation (for MathTools' purposes) so it is 
    the only one of its kind.</p>
*/
public class TextTag extends StyleTag {

  private StringBuffer sbTagContent = new StringBuffer();

  /** Standard constructor for a blank tag. */
  public TextTag() {
    super(StyleTag.TagKeys.TEXT_TAG);
    sbTagContent = new StringBuffer();
  }

  /** Full constructor including initial text. */
  public TextTag(Double pX, Double pY, String pText) {
    super(StyleTag.TagKeys.TEXT_TAG);
    sbTagContent = new StringBuffer();
    setData(pX, pY, pText);
  }

  /** Protected constructor for subclassing. */
  protected TextTag(StyleTag.TagKeys pKey) { super(pKey); }

  /** Set all necessary data in the tag. This does clear out any previously-set <b>text</b>.*/
  public TextTag setData(Double pX, Double pY, String pText) {
    _addNAttribute("x", pX); _addNAttribute("y", pY);
    _setContent(pText);
    return this;
  }

  /** Overridden to retrieve the String content from internal storage. */
  protected String _content() { return sbTagContent.toString(); }

  /** Overridden to handle tag content. If the argument is null the internal content is cleared. Either way any previous content is lost.
      @param pContent The content to set or null to clear it.
  */
  protected void _setContent(String pContent) {
    if(pContent == null) {
      sbTagContent = new StringBuffer();
    } else {
      sbTagContent = new StringBuffer(pContent);
    }
  }

  /** Copy this class. */
  public TextTag copyOf() {
    _enableCopy();
    return new TextTag();
  }

  /* Add content to the internal StringBuffer. Null or blank ignored, no escaping done. */
  private void _addContent(String pContent) {
    if(Tag.validString(pContent)) sbTagContent.append(pContent);
  }

  /** Add raw text content to the tag. This method <b>does</b> XML-escape whatever is added. Null or blank content is ignored. */
  public TextTag addText(String pText) {
    _addContent(MathXML.escapeXML(pText));
    return this;
  }

  /** Add a <code>TSpanTag</code> to the tag content. It will be on its own line and all necessary escaping is done automatically. */
  public TextTag addText(TSpanTag pTag) {
    if(pTag != null) _addContent("\n"+pTag.toString()+"\n");
    return this;
  }

  /** <p>Add a &lt;textpath&gt; element. The path URI and text must not be null or empty. The text is XML-escaped before it is added. The pOffset parameter 
      describes where along the path the text should start. If it is null or blank it is ignored.</p>
      <p>Working with the offset requires a bit of explanation. The value given is of type double to allow both whole-number user-units and percentages. The way 
      this works is as follows: If the absolute value of 'pStartOffset' is between zero and one it is converted to an <i>integer</i> percent with the '%' sign 
      appended. If the absolute value is greater than one it is assumed to be user units and is stringified as-is.</p>
      <p>The &lt;textpant&gt; element, since it has little to configure, is added as a part of the &lt;text&gt; tag and treated as a String with benefits. No 
      attributes other than the ones specified, URI and offset, are available nor should be needed. The tag itself has a newline prepended before and appended 
      afterward so it will appear on its own line.</p>
      @param pURI The URI to the &lt;path&gt; tag along which the text is written.
      @param pText The text to write along the path. Must not be null or empty.
      @param pOffset The distance down the path the text is to start. Ignored if null or empty.
  */
  public TextTag addTextPath(String pURI, String pText, Double pOffset) {
    if(!(Tag.validString(pURI) && Tag.validString(pText))) return this;
    StringBuffer sbPath = new StringBuffer("\n<textPath xlink:href=\"");
    sbPath.append(pURI).append("\" ");
    if(pOffset != null) {
      String sOffset = pOffset.toString();
      if(Math.abs(pOffset.doubleValue()) < 1) {
        sOffset = "" + ((int) (pOffset.doubleValue() * 100)) + "%";
      }
      sbPath.append("startOffset=\"").append(sOffset).append("\" ");
    }
    sbPath.append(">").append(pText).append("</textPath>\n");
    _addContent(sbPath.toString());
    return this;
  }

  /** Set (or clear) a limit on the length of the text. Setting null or blank clears it. */
  public TextTag setTextLength(Double pValue) {
    _addNAttribute("textLength", pValue);
    if(pValue == null) _addAttribute("textLength", null);
    return this;
  }

  /** Set the adjustment method for increasing or decreasing text length. Setting true uses 'spacingAndGlyphs', false uses 'spacing'.
      @param pGlyphs true to include glyphs in length adustment, false to use spacing only.
  */
  public TextTag setLengthAdjust(boolean pGlyphs) {
    _addAttribute("lengthAdjust", (pGlyphs?"spacingAndGlyphs":"spacing"));
    return this;
  }

  /** Set the X- and Y-coordinates of this tag. If either argument is null or blank it is ignored. */
  public TextTag setLocation(Double pX, Double pY) {
   _addNAttribute("x", pX);
   _addNAttribute("y", pY);
    return this;
  }

}

