package cox.jmatt.java.MathTools.markup.svg;

import java.util.*;
import cox.jmatt.java.MathTools.markup.*;

/** <p>This pen provides an easy way to create a sequence of transformations for an SVG tag 'transform=' attribute. The operations work exactly per that standard.
    All transformations are applied IN THE ORDER THEY ARE SET, so set with care!</p>
    <p>All transform methods return a reference to this instance to allow method chaining. This, along with <code>reset() and copyOf()</code> allows easy re-use 
    of <code>TransformPen</code>s.</p>
*/
public class TransformPen  {

  private StringBuffer retval = new StringBuffer();

  /** The only constructor required. */
  public TransformPen() { super(); }

  /* Private constructor for copying. */
  private TransformPen(String pTransforms) {
    super();
    if(Tag.validString(pTransforms)) retval = new StringBuffer(pTransforms);
  }

  /** Print and format the tag as an SVG 'transform=' attribute. NOTE: All information is retained; transformations can be added after this method is called. Either 
      the complete attribute or just the data can be returned. Complete prepends 'transform=' and encloses the data in quotes.
      @param pComplete true to include the attrubute name and quotes, false for just the data.
  */
  public String toString(boolean pComplete) {
    if(pComplete) return ("transform=\""+retval.toString()+"\"");
    return retval.toString();
  }

  /** Return a copy of this <code>TransformPen</code> exactly as it is at this moment. Changing the original later will not affect the copy or vise versa. */
  public TransformPen copyOf() { return new TransformPen(retval.toString()); }

  /** Check to see if any transformations are present. If so, this method returns false. */
  public boolean isEmpty() { return (retval.length() < 1);}

  /** Clear all transforms. Returns a reference to itself to allow method chaining. */
  public TransformPen reset() {
    retval = new StringBuffer();
    return this;
  }

  /** Add a uniform scaling transform: 'scale(pScale)'.
      @param pScale The scale factor. This is applied to both the X and Y coordinates uniformly.
  */
  public TransformPen scale(double pScale) { retval.append("scale(").append(pScale).append(") "); return this; }

  /** Add a non-uniform scaling transform: 'scale(pScaleX, pScaleY)'.
      @param pScaleX The X-coordinate scale factor.
      @param pScaleY The Y scaling factor.
  */
  public TransformPen scale(double pScaleX, double pScaleY) { retval.append("scale(").append(pScaleX).append(", ").append(pScaleY).append(") "); return this; }

  /** Add a rotation about the origin: 'rotate(pTheta)'. NOTE: Positive angles increase DOWNWARD from the X-axis!!
      @param pTheta The angle of rotation. The center is (0, 0).
  */
  public TransformPen rotate(double pTheta) { retval.append("rotate(").append(pTheta).append(") "); return this; }

  /** Add a rotation about a center point other than the origin: 'rotate(pTheta, pCX, pCY)'.
      @param pTheta The angle of rotation to apply.
      @param pCX The X-coordinate of the center of rotation.
      @param pCY The Y-coordinate of the center of rotation.
  */
  public TransformPen rotate(double pTheta, double pCX, double pCY) {
    retval.append("rotate(").append(pTheta).append(", ").append(pCX).append(", ").append(pCY).append(") ");
    return this;
  }

  /** Apply a translation: 'translate(pDeltaX, pDeltaY)'.
      @param pDeltaX The X-axis translation.
      @param pDeltaY The translation in the Y direction.
  */
  public TransformPen translate(double pDeltaX, double pDeltaY) { retval.append("translate(").append(pDeltaX).append(", ").append(pDeltaY).append(") "); return this; }

  /** Apply a skew transform in the X coordinate.
      @param pTheta The angle by which all X-coordinates are to be skewed.
  */
  public TransformPen skewX(double pTheta) { retval.append("skewX(").append(pTheta).append(") "); return this; }

  /** Skew all Y-coordinates.
      @param pTheta The angle by which all X-coordinates are to be skewed.
  */
  public TransformPen skewY(double pTheta) { retval.append("skewY(").append(pTheta).append(") "); return this; }

}
