package cox.jmatt.java.MathTools.test;

import cox.jmatt.java.MathTools.*;

/** <p>This interface is designed to allow custom Object formatting for the core MathTools classes that might need it. Specifically it applies to the classes that 
    represent actual mathematical <i>things</i>. All return values are cast as generic <code>Object</code> to allow maximum flexibility. As with the 
    <code>GraphPaper.render()</code> method once a Custom Object Formatter finishes its work MathTools' is done.</p>
    <p>The <code>formatGeneric()</code> method is provided to catch anything not covered specifically that might need formatting. <code>Question</code> and 
    <code>Percent</code> instances should not need formatting but the methods are provided in case they do.</p>
    <p>In addition to the methods specified below classes implementing <code>COFormatter</code> <b>must</b> have a zero-argument constructor. This is the one 
    used by <code>COFormatterFactory</code> to create instances.</p>
    <h3>Context Objects</h3>
    <p>The context Object is treated more formally here than in the <code>graphing.*</code> package. It is created using a String key (optionally) and it is 
    assumed to contain all relevant contextual information needed to format the MathTools object. As with <code>GraphKeys</code> it is optional but provided if 
    needed.</p>
*/
public interface COFormatter {

  /** Get the formatting context Object if such is needed. A String key is provided to allow for multiple context types.
      @param pType The type of context to create.
      @return The context Object appropriate for the implementation and the task at hand.
  */
  public Object getContext(String pType);

  /** Format a <code>MFraction</code>.
      @param pFrac The MFraction to be formatted.
      @param pContext The context Object containing formatting information.
  */
  public Object format(MFraction pFrac, Object pContext);

  /** Format a <code>MEquation</code>.
      @param pEqn The MEquation to be formatted.
      @param pContext The context Object containing formatting information.
  */
  public Object format(MEquation pEqn, Object pContext);

  /** Format a <code>MRadical</code>.
      @param pRoot The MRadical to be formatted.
      @param pContext The context Object.
  */
  public Object format(MRadical pRoot, Object pContext);

  /** Format a <code>MMatrix</code> properly.
      @param pMatrix The MMatrix to be formatted.
      @param pContext The formatting context Object.
  */
  public Object format(MMatrix pMatrix, Object pContext);

  /** Format a <code>Polynomial</code>. This method must take all terms into account.
      @param pPoly The Polynomial to be formatted.
      @param pContext The context Object containing formatting information.
  */
  public Object format(Polynomial pPoly, Object pContext);

  /** Format a <code>Question</code>. The data within the question should be formatted internally but this method is provided in case it isn't.
      @param pQues The Question to format.
      @param pContext The context Object for formatting.
  */
  public Object format(Question pQues, Object pContext);

  /** Format a <code>MathStat</code> Object.
      @param pMStat The MathStat to format.
      @param pContext The Context Object.
  */
  public Object format(MathStat pMStat, Object pContext);

  /** Format a <code>MComplex</code> per the supplied context. */
  public Object format(MComplex pMCplx, Object pContext);

  /** This method is provided to format anything not covered specifically.
      @param pObject The Object to format.
      @param pContext The formatting Context to use.
  */
  public Object formatGeneric(Object pObject, Object pContext);

  /** Format XML content represented as a String. No assumptions are made other than that the raw data <i>is</i> in String form.
      @param pData The XML data to format.
      @param pContext The formatting Context Object.
  */
  public Object formatXML(String pData, Object pContext);

  /** Format XML data contained in a DOM <code>Node</code>.
      @param pData The DOM Node XML data.
      @param pContext The format Context.
  */
  public Object formatXML(org.w3c.dom.Node pData, Object pContext);

  /** This is a very generic method used to configure the <code>COFormatter</code> instance. How it is used or <i>if</i> it is used must be specified in the 
      implementation documentation. The return value should only be false if the parameter requested absolutely cannot be set as requested.
      @param pKey The name of the parameter to set.
      @param pValue The value of the parameter set.
      @return True or false, depending on the parameter, the value and the implementation.
  */
  public boolean setParameter(String pKey, Object pValue);

}

