package cox.jmatt.java.MathTools.test;

import java.util.*;
import java.io.*;
import cox.jmatt.java.MathTools.*;

/** <p>This class provides a <b>Math Test Formatting Item</b> or MTFI. At its most basic level a MTFI is a bit of text placed on a test. The 'Format Item' comes 
    from the purpose for which the 'bit of text' is used. MTFI provides five distinct types of items all explained in the <code>FIType enum</code>.</p>
    <p>Briefly, the item types are:
    <ul>
      <li><b>TEXT_LINE</b> - A single line of text added to the Problem and Answer part of a test. Think title or simple instructions.</li>
      <li><b>TEXT_BLOCK</b> - Either a very long line of text or multiple lines, even paragraphs. Think multi-step instructions. Prints with the Problems.</li>
      <li><b>TEXT_NOTE</b> - A short note that appears with the Answer part and NOT the Problems.</li>
      <li><b>GRAPH_MARK</b> - A simple text marker for where a graph, perhaps an image or other graphic, is to be inserted. Prints with the Problems.</li>
      <li><b>PAGE_BREAK</b> - Text marker for where a page break should appear.</li>
    </ul></p>
    <p>All of the items except PAGE_BREAK have content associated with them. For GRAPH_MARK it is assumed to be a short identifier for what graph should appear. 
    The 'TEXT_' elements contain text content intended to appear on the actual test, or answer key in the case of TEXT_NOTE.</p>
    <p>This documentation is provided primarily for element explanation and embedding. There should be no need to instantiate this class directly: 
    <code>MTest</code> handles it automatically.</p>
*/

public class MTFI implements MTestElement {

  /** <p>The <b>Format Item Type</b> <code>enum</code> provides the valid types a <code>MTFI</code> can be. Since the characteristics of each are slightly 
      different the <code>enum</code> elements here also store type-specific data and formatting information.</p>
      <p>Again, ordinarily there will be no need to handle this enum directly since <code>MTest</code> handles it internally.</p>
  */
  public enum FIType {

    /** <p>This is a single line of text: one sentence or two short ones. Format-wise it should not contain any newlines.</p>
        <p>This element appears with the Problem and Answer part of a test and appears in replacement tokens &lt;1&gt; and &lt;2&gt; when formatting.</p>
    */
    TEXT_LINE(false, true, true, "<1>", "<Mtfi type=\"TEXT_LINE\" ><1></Mtfi>"),

    /** <p>This item represents multiple long sentences or perhaps multiline instructions or even multiple paragraphs. It may contain any types of characters 
        needed, including newlines and other String-formatting characters.</p>
        <p>This element appears with the Problem part of a test and appears in replacement token &lt;1&gt; when formatting.</p>
    */
    TEXT_BLOCK(false, true, false, "<1>", "<Mtfi type=\"TEXT_BLOCK\" ><1></Mtfi>"),

    /** <p>This is an answer key element. Structurally it is identical to a TEXT_LINE except that it appears with the Answers.</p>
        <p>This element appears with the Answer part of a test and appears in replacement token &lt;2&gt; when formatting.</p>
    */
    TEXT_NOTE(false, false, true, "<2>", "<Mtfi type=\"TEXT_NOTE\" ><2></Mtfi>"),

    /** <p>This item is a plain text placeholder for where a graph or image should be inserted into the test. The only content is a short ID.</p>
        <p>This element has a standard, fixed format that appears with the Problems only.</p>
    */
    GRAPH_MARK(true, true, false,"** Graph(<1>) **", "<Mtfi type=\"GRAPH_MARK\"  id=\"<1>\" />"),

    /** This is a marker to indicate where a page break should occur. It has no content and appears in a standard format with the Problems. */
    PAGE_BREAK(false, true, false, "*** PAGE BREAK ***", "<Mtfi type=\"PAGE_BREAK\" />"),

    /** This is an 'Oops!' marker that <i>should</i> never be seen. It is used to catch mis-constructed MTFIs. If one is encountered it will not appear, ever, 
        but a Debug-level message is printed.
    */
    UNKNOWN(false, false, false, "-- UNKNOWN MTFI --", "<Mtfi type=\"UNKNOWN\" />");

    private boolean onID = false, onProbs = false, onAns = false;
    private String myDefaultTemplate = "", myXMLTemplate = "";
    FIType(boolean pID, boolean pProbs, boolean pAns, String pTemplate, String pXMLTemplate) {
      this.onID = pID;
      this.onProbs = pProbs;
      this.onAns = pAns;
      this.myDefaultTemplate = pTemplate;
      this.myXMLTemplate = pXMLTemplate;
    }

    /** Should this print with Problem part? */
    public boolean printProblem() { return onProbs; }
    /** Should this print with Answer part? */
    public boolean printAnswer() { return onAns; }
    /** Does this item have an XML ID attribute? (Graph marks yes, no for the others.) */
    public boolean hasID() { return onID; }
    /** Return the default template for this item type. */
    public String getTemplate() { return myDefaultTemplate; }
    /** Return the appropriate XML-formatting template. */
    public String getXMLTemplate() { return myXMLTemplate; }

  }

  /** Default 'entire' format for all MTFI types: '&lt;0:gt;&lt;1&gt; &lt;2&gt;'. */
  public static final String DEFAULT_GLOBAL_FORMAT = "<0><1> <2>";

  private static boolean globalEscapeXML = true, showID = false;
  private static String myFormat = DEFAULT_GLOBAL_FORMAT;

  private FIType myType = FIType.UNKNOWN;
  private String myContent = "";

  private MTFI() { super(); }

  /** This is the only public constructor; a type is ABSOLUTELY required, even though content may be optional. */
  public MTFI(FIType pType, String pContent) {
    super();
    if(pType != null) myType = pType;
    if(pContent != null) myContent = pContent;
  }
  
  /** This class maintains a global XML escape flag. Setting once sets it for all instances until it is reset. This method sets it. */
  public static void setEscapeXML(boolean yesNo) { globalEscapeXML = yesNo; }

  /** Set the 'entire MTFI' format, analogous to the 'entire Question' format. Setting to null or blank resets the default. */
  public static void setFormat(String pFormat) { myFormat=Question.validString(pFormat)?pFormat:DEFAULT_GLOBAL_FORMAT; }

  /** Set to 'true' to show <code>MTFI</code> type in replacement token 0 when formatting. Defaults to false. */
  public static void setShowID(boolean yesNo) { showID = yesNo; }

  /** This method is used by <code>MTXProcessor</code> to re-create format items from MathTools XML markup. If pType is not valid a TEXT_LINE with an error 
      message is created. Any errors are reported at Error level.
      @param pType The type of MTFI to create.
      @param pContent The tag content, if needed.
  */
  public static MTFI newInstance(String pType, String pContent) {
    MTFI retval = new MTFI(MTFI.FIType.UNKNOWN, "Error creating MTFI");
    try{
      retval = new MTFI(MTFI.FIType.valueOf(pType), pContent);
    } catch(Exception ex) { CapCom.error("Error creating MTFI.", ex); }
    return retval;
  }

  /** <p>This class' raison-d'etre. All other format methods echo to this one. Even thogh MTFI doesn't have Problem and Answer parts per se, the content is 
      handled similarly and consistently with <code>Question</code>'s replacement tokens: &lt;1&gt; prints with the Problems and &lt;2&gt; prints with the 
      Answers. All other tokens are blank and this is handled automatically.</p>
      <p>Default templates are supplied, they should be appropriate for most cases.</p>
      @param pTemplate The template String for formatting.
  */
  public String format(String pTemplate) { return format(pTemplate, false); }

  /* Now do the actual formatting! */
  private String format(String pTemplate, boolean isXML) {
    if(myType == FIType.UNKNOWN) CapCom.debug("UNKNOWN MTFI encountered: "+myContent);
    String[] tokenz = new String[3];
    tokenz[0] = showID?("MTFI ("+myType.toString()+"): "):"<-1>";
    tokenz[1] = myType.printProblem()?((isXML && globalEscapeXML)?MathXML.escapeXML(myContent):myContent):"<-1>";
    tokenz[2] = myType.printAnswer()?((isXML && globalEscapeXML)?MathXML.escapeXML(myContent):myContent):"<-1>";
    if(!isXML && ((myType == FIType.GRAPH_MARK) || (myType == FIType.PAGE_BREAK))) pTemplate = myType.getTemplate();
    return Question.fillTemplate(pTemplate, tokenz);
  }

  /** Format using the internal template. */
  public String format() { return format(myFormat); }

  /** Print an XML representation of this format itme. */
  public String toXML() { return format(myType.getXMLTemplate(), true); }

  /** Call <code>format()</code> when appropriate. */
  public String formatProblem() {
    return myType.printProblem()?format(myType.getTemplate()):"";
  }

  /** Call <code>format()</code> when printing answers. */
  public String formatAnswer() {
    return myType.printAnswer()?format(myType.getTemplate()):"";
  }

  /** Create a copy of this element. */
  public MTFI copyOf() {return new MTFI(myType, myContent); }

}

