package cox.jmatt.java.MathTools.test;

import java.util.*;
import java.util.regex.Matcher;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.util.MTXProcessor;

/** <p><code>MTest</code> is a simple container for <code>MTestElement</code>s. It makes provision for holding them in the order in which they are entered, 
    formatting them and presenting the results. The Problem and Answer parts can be printed individually or sequentially, the internal formats can be used or one 
    can be specified for Problem and Answer separately.</p>
    <p>Formatting via this class does not change any of the data formatted with one exception. When the element content is actually formatted all newline 
    characters are replaced with the value set in the <code>setNewline()</code> method. This occurs after the content is pulled from the test elements and before 
    it is printed so the original data doesn't change.</p>
*/

public class MTest {

  /** <p>This is the only constructor needed. 
  */
  public MTest() { super(); }

  private static String crlf = "\n";
  private static boolean myEscapeXML = true;

  private List<MTestElement> myElements = new ArrayList<MTestElement>();
  private String myLineSep = "";

  /** <p>Set the character or character sequence used for newlines. This is the only formatting 'change' MTest makes to any content. There are three choices: 
      <ul>
        <li>'n' for the newline character '/n'. (Default value.)</li>
        <li>'r' for the carriage return character '/r'.</li>
        <li>'b' for carriage return + newline, '/r/n'.</li>
      </ul></p>
      <p>Uppercase or lowercase characters work equally well but ONLY the three characters, as <b>char</b>, are valid values. Any others are ignored.</p>
  */
  public final void setNewline(char pChar) {
    switch(pChar) {
      case 'n': case 'N': crlf = "\n"; break;
      case 'r': case 'R': crlf = "\r"; break;
      case 'b': case 'B': crlf = "\r\n"; break;
      default:
    }
  }

  /** Set the (static) XML escape flag. This method also sets <code>MTFI</code>'s escape. It is assumed that any <code>Question</code>s will handle themselves. */
  public static void set_XML_escape(boolean yesNo) {
    myEscapeXML = yesNo;
    MTFI.setEscapeXML(yesNo);
  }

  /** This method replaces newline characters ('\n') with whatever value is set, other than '\n' (... which is no replacement at all!). */
  public static String replaceNewlines(String pRaw) {
    if(!crlf.equals("\n")) pRaw = pRaw.replaceAll("\n", crlf);
    return pRaw;
  }

  /** Instance shadow to set XML escape. */
  public void setEscapeXML(boolean yesNo) { set_XML_escape(yesNo); }

  /** Set the separator appended to each element's content when it is formatted. The default is an empty String ('') and this is set if pNewSep is null. */
  public void setLineSeparator(String pNewSep) {
    myLineSep = (pNewSep == null)?"":pNewSep;
  }

  /** Set the 'entire' format for MTFI. Setting null resets it to default. */
  public void setFormatMTFI(String pFormat) { MTFI.setFormat(pFormat); }

  /** Toggle MTFI ID display in formatting token 0. Defaults to false. */
  public void setShowID_MTFI(boolean yesNo) { MTFI.setShowID(yesNo); }

  /** Reset everything on the test. The newline character reverts to '\n', the XML escape reverts to 'true' (this applies to MTFI as well), all elements 
      currently stored are cleared and the default separator(s) are cleared.
  */
  public void reset() {
    myElements = new ArrayList<MTestElement>();
    setNewline('n');
    set_XML_escape(false);
    setLineSeparator(null);
    MTFI.setFormat(null);
    MTFI.setShowID(false);
  }

  /** Create and return a new <code>MTestParser</code>. */
  public static MTestParser getParser() { return new MTestParser(); }

  /** Instance method to crate a new parser. */
  public MTestParser getNewParser() { return new MTestParser(); }

  /** Add a <code>MTestElement</code> to the internal list. If mElement is null it is silently ignored.
      @param mElement The <code>MTestElement</code> to add to the test. Ignored if null.
  */
  public void add(MTestElement mElement) {
    if(mElement != null) myElements.add(mElement.copyOf());
  }

  /** Add a <code>TEXT_LINE MTFI</code> element. Silently ignored if pContent is null or blank. */
  public void addLine(String pContent) {
    if(!Question.validString(pContent)) return;
    myElements.add(new MTFI(MTFI.FIType.TEXT_LINE, pContent));
  }

  /** Add a <code>TEXT_BLOCK MTFI</code> element. Silently ignored if pContent is null or blank. */
  public void addBlock(String pContent) {
    if(!Question.validString(pContent)) return;
    myElements.add(new MTFI(MTFI.FIType.TEXT_BLOCK, pContent));
  }

  /** Add a <code>TEXT_NOTE MTFI</code> element. Silently ignored if pContent is null or blank. */
  public void addNote(String pContent) {
    if(!Question.validString(pContent)) return;
    myElements.add(new MTFI(MTFI.FIType.TEXT_NOTE, pContent));
  }

  /** Add a <code>GRAPH_MARK MTFI</code> element. If pContent is null or blank a default ID is assigned. */
  public void addGraphMark(String pContent) {
    if(!Question.validString(pContent)) pContent = "MTFI_GM"+CapCom.hitCount();
    myElements.add(new MTFI(MTFI.FIType.GRAPH_MARK, pContent));
  }

  /** Add a <code>PAGE_BREAK MTFI</code> element. */
  public void addPageBreak() {
    myElements.add(new MTFI(MTFI.FIType.PAGE_BREAK, null));
  }

  /** Format each element, in order, per the zero-argument <code>format()</code> method. The internal line separator is appended after each element and all 
      content is newline-replaced.
  */
  public String format() {
    StringBuffer retval = new StringBuffer();
    for(Iterator<MTestElement> it = myElements.iterator(); it.hasNext(); ) {
      retval.append(replaceNewlines(it.next().format())).append(replaceNewlines(myLineSep));
    }
    return retval.toString();
  }

  /** <p>This method formats each element according to the template String passed in. No validity-checking is done on the template. The line separator is appended 
      to each element and newlines are replaced.</p>
      <p>Replacement tokens are per <code>Question</code>.</p>
      @param pTemplate The template String by which each element is to be formatted.
  */
  public String format(String pTemplate) {
    StringBuffer retval = new StringBuffer();
    for(Iterator<MTestElement> it = myElements.iterator(); it.hasNext(); ) {
      retval.append(replaceNewlines(it.next().format(pTemplate))).append(replaceNewlines(myLineSep));
    }
    return retval.toString();
  }

  /** <p>This method calls <code>formatProblem()</code> on each element on the test. The line separator is appended and newlines replaced. */
  public String formatProblems() {
    StringBuffer retval = new StringBuffer();
    for(Iterator<MTestElement> it = myElements.iterator(); it.hasNext(); ) {
      retval.append(replaceNewlines(it.next().formatProblem())).append(replaceNewlines(myLineSep));
    }
    return retval.toString();
  }

  /** <p>This method calls <code>formatAnswer()</code> on each element on the test. The line separator is appended and newlines replaced. */
  public String formatAnswers() {
    StringBuffer retval = new StringBuffer();
    for(Iterator<MTestElement> it = myElements.iterator(); it.hasNext(); ) {
      retval.append(replaceNewlines(it.next().formatAnswer())).append(replaceNewlines(myLineSep));
    }
    return retval.toString();
  }

  /** This method calls <code>formatProblems()</code> and <code>formatAnswers()</code> and concatenates the results. If pKeySep is not null it is newline-replaced 
      and placed between the Problems and Answers.
      @param pKeySep A String to be placed between the <code>formatProblems()</code> and <code>formatAnswers()</code> results. Ignored if null.
  */
  public String formatTest(String pKeySep) {
    StringBuffer retval = new StringBuffer(formatProblems());
    if(pKeySep != null) retval.append(replaceNewlines(pKeySep));
    retval.append(formatAnswers());
    return retval.toString();
  }

  /** Convenience method for <code>formatTest(null)</code>. */
  public String formatTest() { return formatTest(null); }

  /** Convert all elements to XML. The line separator is appended after each element and everything is newline-escaped. No XML-escaping is done here as it is 
      assumed to have taken place already.
  */
  public String toXML() {
    StringBuffer retval = new StringBuffer("<MTest>\n");
    for(Iterator<MTestElement> it = myElements.iterator(); it.hasNext(); ) {
      retval.append(it.next().toXML()).append(myLineSep);
    }
    retval.append("</MTest>\n");
    return(replaceNewlines(retval.toString()));
  }

  /** Overridden. Simply calls <code>formatTest()</code>. */
  public String toString() { return formatTest(); }

}

