package cox.jmatt.java.MathTools.test;

import java.util.*;
import java.io.*;
import cox.jmatt.java.MathTools.*;

/** <p>This simple class is designed to parse text into a MTest. The method used to add Questions is {@link cox.jmatt.java.MathTools.Question#parseQuestion Question.parseQuestion()} 
    and all default settings apply. A very simple macro language allows setting of options and <code>MTFI</code>s. The input can come from a file, a resource 
    stream or a String of newline-separated text.</p>
    <p>All macro commands start with the '@' character and consist of a single letter and possibly text afterward. The macro commands are:
    <ul>
      <li>'@L <i>text</i>' - Add a TEXT_LINE MTFI to the MTest.</li>
      <li>'@B <i>text</i>' - Add a TEXT_BLOCK MTFI to the MTest.</li>
      <li>'@N <i>text</i>' - Add a TEXT_NOTE MTFI to the MTest.</li>
      <li>'@G <i>text</i>' - Add a GRAPH_MARK MTFI to the MTest.</li>
      <li>'@P'  - Add a PAGE_BREAK MTFI to the MTest.</li>
      <li>'@F <i>fmt1 [@@ fmt2] [@@ fmt3]'</i> - Set the default Question format(s). The break sequence is '@@' and up to 3 may be specified.</li>
      <li>'@I <i>[text]'</i> - (Re-) Set the Question ID format.</li>
      <li>'@C <i>[int]'</i> - Set the value of The Clicker (Question). If no number is supplied it defaults to zero.</li>
      <li>'@@ <i>[text]'</i> - Comment. Line is ignored.</li>
    </ul>
    Lowercase letters also work.</p>
    <p>The intent of this class is to allow quick creation of a simple MTest. Since <code>MTest</code> is very much line-oriented it lends itself particularly 
    well to LaTeX. LaTeX handles its own page layout so line-orientation suffices for the raw input. Also, since a MTest can be set before data is parsed this 
    class can also be used to configure large parts of the test that don't really change.</p>
*/

public class MTestParser {

  private static boolean autoClick = true;

  private MTest myTest = null;

  /** The only constructor needed. Ever. */
  public MTestParser() { super(); }

  /** This method sets the global autoclick value. If true, each time a Question is added its <code>click()</code> method is called. */
  public MTestParser setAutoClick(boolean yesNo) {
    autoClick = yesNo;
    return this;
  }

  /** Use this method to set the MTest instance used to receive the parsed data. If set to null a new MTest is used. */
  public MTestParser setMTest(MTest pTest) {
    myTest = pTest;
    return this;
  }

  /** Reset the parser to its default values: true for autoclick and no pre-set MTest. */
  public MTestParser reset() { return setAutoClick(true).setMTest(null); }

  /** Parse a String into a <code>MTest</code>. The parser breaks on newlines and this is not configurable.
      @param pData The String to be parsed.
  */
  public MTest parse(String pData) { return parseTest(pData, false, false); }

  /** Parse a file into a MTest. This may be on the filesystem or from a resource stream. If the latter, it will automatically use the Math ClassLoader.
      @param pFile The name of the file to parse.
      @param pGRAS true to <code>getResourceAsStream()</code>, false for a file.
  */
  public MTest parseFile(String pFile, boolean pGRAS) { return parseTest(pFile, true, pGRAS); }

  // Do the actual work.
  private MTest parseTest(String pData, boolean isFileName, boolean pGRAS) {
    if(myTest == null) myTest = new MTest();
    if(!Question.validString(pData)) return myTest;
    BufferedReader myReader = null;
    try {
      if(isFileName) {
        if(pGRAS) {
          myReader = new BufferedReader(new InputStreamReader(CapCom.getMathClassLoader().getResourceAsStream(pData)));
        } else {
          myReader = new BufferedReader(new FileReader(pData));
        }
      } else {
        myReader = new BufferedReader(new StringReader(pData));
      }
      String sLine = null;
      while((sLine = myReader.readLine()) != null) {
        if(!Question.validString(sLine)) continue;
        sLine += "  ";
        if(sLine.startsWith("@")) {
          switch(sLine.charAt(1)) {
          case 'l':
          case 'L':
            myTest.addLine(sLine.substring(2));
            break;
          case 'b':
          case 'B':
            myTest.addBlock(sLine.substring(2));
            break;
          case 'n':
          case 'N':
            myTest.addNote(sLine.substring(2));
            break;
          case 'g':
          case 'G':
            myTest.addGraphMark(sLine.substring(2));
            break;
          case 'p':
          case 'P':
            myTest.addPageBreak();
            break;
          case 'f':
          case 'F':
            Question.setDefaultFormat(sLine.substring(2).split("@@"));
            break;
          case 'c':
          case 'C':
            try {
              Question.setClicker(Integer.parseInt(sLine.substring(2).trim()));
            } catch(Exception ignored) {
              Question.setClicker(0);
            }
            break;
          case 'i':
          case 'I':
            Question.setIDFormat(sLine.substring(2));
            break;
          case '@': break;
          default:

          }
        } else {
          myTest.add(Question.parseQuestion(sLine));
          if(autoClick) Question.incrementClicker();
        }
      }
    } catch(Exception ex){
      CapCom.debug("MTestParser Exception", ex);
    }
    return myTest;
  }

}

