package cox.jmatt.java.MathTools.test;

import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.sql.ResultSet;
import java.text.DateFormat;
import cox.jmatt.java.MathTools.*;

/** <p>This class provides a more sophisticated templating engine than <code>Question.fillTemplate()</code>. In terms of capability it lies between that method 
    and Apache Velocity, but far closer to the former than the latter. It uses a Velocity-like syntax but serves mostly as a text-replacement engine with a few 
    extras.</p>
    <h3>So why not just use Velocity?</h3>
    <p>Mainly because this is included within the <code>MathTools</code> package and requires no external equipment, no subclassing, no re-invention of a GUI, 
    and no custom-scripting-application construction. But if you like Velocity there is <i>no reason whatsoever</i> not to use it!</p>
    <h3>Okay, so why bother with this?</h3>
    <p><code>MathTestFormatter</code> is written specifically to work with the other <code>MathTools</code> classes. It is a part of the package and it has NO 
    external dependencies. It is included, it's easy, and it works with the other tools.</p>
    <p>Speaking more philosophically, <code>MathTestFormatter</code> offers a very flexible test-centric way to format and present data. <code>MTest</code> does 
    the same thing but it is much more question-centered. This class is designed to represent the test as a whole. While <code>Question</code>s are an essential 
    part they are not the entirety of the test.</p>
    <p>In addition to dealing with Question instances, this class has ample machinery to handle other aspects of the test. It can load a template String or Properties 
    from an external source, add any other necessary data, and merge the whole thing. It can row-format a <code>java.sql.ResultSet</code> generated by 
    <code>MathDBC</code> (think grade reports) and, since the templating machinery doesn't use any XML-dangerous characters intrinsically, generate XML output 
    easily. <code>MathTestFormatter</code> offers maximum flexibility in producing a print-ready test without needing any middleware.</p>
    <p>Finally, there is the abundance of <code>format()</code> methods. The basic <code>fillTemplate()</code> method in this class accepts a template String and 
    a Properties object for data but there is more machinery under the hood. The formatting methods are:
    <ul>
      <li><code>format()</code> applies the internal data to the internal template.</li>
      <li><code>format(String)</code> applies the internal data to the template supplied.</li>
      <li><code>format(java.sql.ResultSet)</code> applies the internal data to the results of a database query.</li>
      <li><code>format(java.util.List)</code> applies the internal data to a <code>List</code> of template Strings.</li>
      <li><code>format(String[])</code> applies the internal data to an array of template Strings.</li>
      <li><code>format(Properties[], boolean)</code> applies each Properties object to the internal template. The internal data may also be applied.</li>
      <li><code>reFormat(String)</code> formats the internal template with the capture groups from a regular expression.</li>
      <li><code>reFormat(String[])</code> applies <code>reFormat()</code> to an array of Strings.</li>
    </ul></p>
    <p>The only tradeoff is complexity. <code>MTest</code> is considerably easier to use than this class. That is not to say using this class is <i>hard</i>, merely 
    that it is not quite as easy as MTest. If the trade is worth it, use it!</p>
    <h3>Now, about <code>Question.fillTemplate()</code>...</h3>
    <p>The <code>Question.fillTemplate()</code> method was designed around the <i>internal</i> structure and formatting of a test question. Its replacement token 
    syntax is both different and simpler than that used here. This was a deliberate choice meant to separate test <i>question</i> formatting from <i>test</i> 
    formatting. One is 'internal' in intent; focusing primarily on how to deal with the data before it's ready to go on the test. The other is more 'external' 
    in that it assumes that, by the time it is used, the internal data handling and formatting is done. The question is ready for the test and only needs to be 
    placed appropriately.</p>
    <p>This class is a more 'real' templating engine in that the syntax <b>is</b> different from <code>Question.fillTemplate()</code>. The templates here use an 
    identifier-based token syntax rather than array-position-based. This makes designing the template for a whole test easier and more 'natural'; to wit, the 
    template 'Chapter Exam ${version} ${date}' is easier to understand than 'Chapter Exam &lt;4&gt; &lt;7&gt;'.</p>
*/
public class MathTestFormatter {

  /** This is the default Problem key format: '@{QID}_Prob'. */
  public static final String DEFAULT_PROBLEM_KEY = "@{QID}_Prob";

  /** This is the default Answer key format: '@{QID}_Ans'. */
  public static final String DEFAULT_ANSWER_KEY = "@{QID}_Ans";

  private static String globalTimestampFormat = null, defaultNewline = "\n";

  private String myTemplate = null, myProbKey = DEFAULT_PROBLEM_KEY, myAnsKey = DEFAULT_ANSWER_KEY;
  private Properties myData = null;
  private Pattern myPattern = null;
  private ClassLoader myLoader = null;
  private long myClicker = 0;

  /** Standard constructor for scripts. */
  public MathTestFormatter() { super(); }

  /** <p>This is the foundation method for the entire class. It does the actual work of filling out a template. It accepts as input a String template containing 
      replacement tokens and a <code>java.util.Properties</code> object containing the values for the tokens. If either is null or empty the template String is 
      returned unaltered. Any errors are reported at Error level.</p>
      <p>Standard replacement tokens are of the form '${<i>someKeyName</i>}' where <i>someKeyName</i> is a key within the <code>Properties</code> object. When 
      invoked this method replaces the entire replacement token with the value of the key between the curly braces. For example if the template is 'I like ${food}!' 
      and the Properties contained the mapping 'food = gyros' the completed template would be 'I like gyros!'.</p>
      <p>NOTE: This method makes excessive use of <code>String replaceAll()</code> which uses regular expressions for key-matching. This can cause problems for 
      key-value mappings whose key contains forbidden characters. Since the dot is common it and it alone is escaped. Using a dot in a property key will work, no 
      other collision characters will! Use with care.</p>
      <p>In addition to the data fields defined by the Properties object, <code>MathTestFormatter</code> defines a group of special tokens (or fields). These are 
      handled internally and automatically, though some can be configured. The special field tokens are:
      <ul>
        <li>'@{time}' yields the current time (<code>DateFormat.SHORT)</code>.</li>
        <li>'@{fullTime}' yields the current time (default format).</li>
        <li>'@{date}' yields the current date (default format).</li>
        <li>'@{timestamp}' allows a user-defined format.</li>
        <li>'@{count}' is used as an index counter in some methods (these are noted in the method documentation).</li>
        <li>'@{newline}' holds a carriage return, linefeed (default), or CRLF sequence.</li>
        <li>'@{blank}' is replaced with a single space. This occurs after all other tokens are replaced.</li>
        <li>'#{..<i>text</i>..}#' is a comment.</li>
      </ul></p>
      <p>The @{timestamp} field is configured via the <code>setDefaultTimestamp()</code> methods. The formatting String is per the <code>SimpleDateFormat</code> 
      class; it might not be available. If not the hard-wired timestamp format is <code>DateFormat.FULL</code>. The @{newline} field can be configured through the 
      <code>setDefaultNewline()</code> methods. The only values accepted are newline ('/n'), carriage return ('/r'), or both ('/r/n'). Once set these parameters 
      remain in effect until they are changed. Text within the comment container ('#{' to '}#') does not appear when the template is processed. Also note that 
      unlike other tokens the comment delimiter ('#') appears on <b>both sides</b> of the curly braces.</p>
      <p>The '@{blank}' token is replaced with a single space character. This token is replaced after <b>all</b> other tokens have been handled, whether regular, 
      special, or numeric. This allows templates to be 'blanked out' if desired by setting a the value of one or more keys in the Properties object to '@{blank}'. 
      First the standard token is replaced with the value '@{blank}' which is then replaced with a single space. For those interested, the order in which tokens 
      are replaced is:
      <ol>
        <li>Standard tokens: values from the <code>Properties</code> object 'pData'.</li>
        <li>Numeric tokens: <code>ResultSet</code> fields or regex capture groups.</li>
        <li>Pre-defined tokens: those beginning with '@{'.</li>
        <li>Comments; '#{<i>Comment text</i>}#'.
        <li>The '@{blank}' token.
      </ol></p>
      <p>Token order replacement should not be relevant other than for blanking, but occasional edge conditions may make it relevant.</p>
      <p>Any errors occurring during template merging are reported at Error level. The process halts at that point.</p>
      @param pTemplate A String template containing one or more replacement tokens.
      @param pData A Properties object whose keys match the replacement characters and whose mapped values will replace their template tokens.
      @return A new String consisting of the template with all tokens in pData replaced.
  */
  public static final String fillTemplate(String pTemplate, Properties pData) { return fillTemplate(pTemplate, pData, null, 0); }

  //
  // Do the actual work. The real method also allows a String array to replace tokens of the form '${#}' where # is a number.
  //
  private static final String fillTemplate(String pTemplate, Properties pData, String[] pArray, int pCount) {
    if(!Question.validString(pTemplate)) return pTemplate;
    String retval = pTemplate, sKeyTemp = null, sCount = ""+pCount;
    Calendar rightNow = Calendar.getInstance();
    String curDate = DateFormat.getDateInstance().format(rightNow.getTime());
    String fullTime = DateFormat.getTimeInstance().format(rightNow.getTime());
    String curTime = DateFormat.getTimeInstance(DateFormat.SHORT).format(rightNow.getTime());
    String timeStamp = DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.FULL).format(rightNow.getTime());
    if(Question.validString(globalTimestampFormat)) {
      try {
        timeStamp = (new java.text.SimpleDateFormat(globalTimestampFormat)).format(rightNow.getTime());
      } catch(Exception ex) { CapCom.debug("Unable to format timestamp", ex); }
    }
    try {
      if((pData != null) && !pData.isEmpty()) {
        for(String sKey : pData.stringPropertyNames()) {
          if(!Question.validString(sKey)) continue;
          sKeyTemp = sKey.replaceAll("[.]", "[.]");
          retval = retval.replaceAll("[$][{]"+sKeyTemp+"[}]", Matcher.quoteReplacement(pData.getProperty(sKey)));
        }
      }
      if((pArray != null) && (pArray.length > 0)) {
        for(int i=0; i < pArray.length; i++) {
          if(Question.validString(pArray[i])) retval = retval.replaceAll("[$][{]"+i+"[}]", Matcher.quoteReplacement(pArray[i]));
        }
      }
      retval = retval.replaceAll("@[{]date[}]", curDate);
      retval = retval.replaceAll("@[{]time[}]", curTime);
      retval = retval.replaceAll("@[{]fullTime[}]", fullTime);
      retval = retval.replaceAll("@[{]timestamp[}]", timeStamp);
      retval = retval.replaceAll("@[{]count[}]", sCount);
      retval = retval.replaceAll("@[{]newline[}]", defaultNewline);
      retval = retval.replaceAll("#[{].*?[}]#", "");
      retval = retval.replaceAll("@[{]blank[}]", " ");
    } catch(Exception ex) { CapCom.error("MTF.fillTemplate() error.", ex); }
    return retval;
  }

  /** Set the default format for the @{timestamp} field. This is per <code>SimpleDateFormat</code> and may not work. If not, the timestamp is formatted per 
      <code>DateFormat.FULL</code> for both date and time.
      @param pFormat The formatting String for the timestamp field.
  */
  public static void setDefaultTimestamp(String pFormat) { globalTimestampFormat = pFormat; }

  /** Instance version. */
  public void setNewDefaultTimestamp(String pFormat) { globalTimestampFormat = pFormat; }

  /** <p>This method sets the character sequence used for the @{newline} field. Since there are only three choices they are represented by char:
      <ul>
        <li>'n' sets a single newline character (Unix)</li>
        <li>'r' sets a single carriage return character (Mac)</li>
        <li>'b' sets a carriage return followed by a newline (DOS-based)</li>
      </ul></p>
      <p>Upper-case letters also work. Any character other than the ones listed are silently ignored.</p>
  */
  public static void setDefaultNewline(char pNewline) {
    switch(pNewline) {
    case 'n': case 'N': defaultNewline = "\n"; break;
    case 'r': case 'R': defaultNewline = "\r"; break;
    case 'b': case 'B': defaultNewline = "\r\n"; break;
    default:
    }
  }

  /** Instance version. */
  public static void setNewDefaultNewline(char pNewline) { setDefaultNewline(pNewline); }

  /** Reset the internal template and Properties object to null. This affecs only instance data, not global default values. */
  public void reset() {
    myTemplate = null;
    myData = null;
    setQuestionKeyFormat(null, null);
    myClicker = 0;
    myPattern = null;
  }

  /** Set (or clear) the internal template String. */
  public void setTemplate(String pTemplate) { myTemplate = pTemplate; }

  /** Add a Properties object to the internal one. If pNewProps is null or empty this method returns silently. */
  public void addProperties(Properties pNewProps) {
    if((pNewProps == null) || (pNewProps.isEmpty())) return;
    if(myData == null) myData = new Properties();
    myData.putAll(pNewProps);
  }

  /** Add Properties to the internal Properties object. The data is represented as a String containing property (line) and key-value separators. It is first split 
      by the property separator and then each line is split (maximum 2) on the key-value separator. The resulting property is then added. If the data String is 
      null or empty nothing happens. If either the property separator or key-value separator is null the default values are used.
      @param pData The String containing the properties data to be added.
      @param pLineSep The properties (line) separator to be used. Defaults to a newline.
      @param pKVSep The String used to separate the key from the value. Defaults to an equals sign.
  */
  public void addProperties(String pData, String pLineSep, String pKVSep) {
    if(!Question.validString(pData)) return;
    String sKVSep = Question.validString(pKVSep)?pKVSep:"=";
    String[] propz = pData.split(Question.validString(pLineSep)?pLineSep:"\n"), tokenz = null;
    for(int i=0; i < propz.length; i++) {
      tokenz = propz[i].split(sKVSep, 2);
      if(tokenz.length == 2) addProperty(tokenz[0], tokenz[1]);
    }
  }

  /** Add the entire contents of the Global Pizza to the internal Properties object. */
  public void addGlobalPizza() { addProperties(CapCom.getPizza()); }

  /** Add a single property to the internal Properties object. If pKey is null or blank nothing happens. If pValue is null or blank the property is removed.
      @param pKey The key used to store or remove an internal mapping.
      @param pValue The value to store under pKey, or null to remove any stored value.
  */
  public void addProperty(String pKey, String pValue) {
    if(!Question.validString(pKey)) return;
    if(myData == null) myData = new Properties();
    if(Question.validString(pValue)) {
      myData.put(pKey.trim(), pValue.trim());
    } else {
      myData.remove(pKey.trim());
    }
  }

  /** Add a <code>java.util.Map</code> to the properties. This method uses the widest cast on the <code>Map</code> interface to allow for maximum flexibility. 
      Null or invalid keys or values are silently ignored.
      @param pMap The Map to be added to the internal properties.
  */
  public void addMap(Map pMap) {
    if((pMap == null) || pMap.isEmpty()) return;
    Object oVal = null;
    for(Object oKey : pMap.keySet()) {
      if(oKey == null) continue;
      oVal = pMap.get(oKey);
      if(oVal != null) addProperty(oKey.toString(), oVal.toString());
    }
  }

  /** Enable or disable the Math ClassLoader for loading properties files or templates. This method causes immediate action so changing it between load calls WILL 
      affect them.
      @param pEnable true to use CapCom's ClassLoader, false not to.
  */
  public void setMathClassLoader(boolean pEnable) { myLoader = (pEnable?CapCom.getMathClassLoader():null); }

  //
  // Create an InputStream based on GRAS and ClassLoader settings.
  //
  private InputStream getLoadStream(String pFileName, boolean pGRAS) throws Exception {
    InputStream retval = null;
    if(pGRAS) {
      if(myLoader == null) {
        retval = this.getClass().getResourceAsStream(pFileName);
      } else {
        retval = myLoader.getResourceAsStream(pFileName);
      }
    } else {
      retval = new FileInputStream(pFileName);
    }
    return retval;
  }

  /** Load a Properties object from a file and add it to the currently-set internal properties. Any errors are reported at Error level and nothing is loaded or 
      added if one occurs. If the filename supplied is null or empty the method returns silently.
      @param pFileName The name of the file to load.
      @param pGRAS true to <code>getResourceAsStream()</code>, false to open as a standard <code>File</code>.
  */
  public void loadProperties(String pFileName, boolean pGRAS) {
    if(!Question.validString(pFileName)) return;
    Properties loadedProps = new Properties();
    try {
      InputStream isLoad = getLoadStream(pFileName, pGRAS);
      loadedProps.load(isLoad);
      isLoad.close();
      addProperties(loadedProps);
    } catch(Exception ex) { CapCom.error("Error loading Properties", ex); }
  }

  /** Load a template String from a file or resource. Parameters are per <code>loadProperties()</code> as is error reporting. */
  public void loadTemplate(String pFileName, boolean pGRAS) {
    if(!Question.validString(pFileName)) return;
    try {
      InputStreamReader isrLoad = new InputStreamReader(getLoadStream(pFileName, pGRAS));
      StringBuffer sbTemplate = new StringBuffer();
      int inChar = 0;
      while((inChar = isrLoad.read()) > -1) { sbTemplate.append((char) inChar); }
      isrLoad.close();
      myTemplate = sbTemplate.toString();
    } catch(Exception ex) { CapCom.error("Error loading template", ex); }
  }

  /** Format the internal template String using the internal Properties data. */
  public String format() { return fillTemplate(myTemplate, myData); }

  /** Format the supplied template String using the internal data. */
  public String format(String pTemplate) { return fillTemplate(pTemplate, myData); }

  /** <p>This method formats each row in a <code>java.sql.ResultSet</code>. The internal Properties object is available as are all the automatic fields. The column 
      data from the <code>ResultSet</code> is accessed via a special set of fields defined in this method: '${<i>X</i>}' where <i>X</i> is the number of the column 
      containing the data. Since SQL indexing is one-based the first data value is '${1}' for column 1, etc. The '${0}' token is the <b>row count</b> which starts 
      at one and moves up by one with each row. The '@{count}' token is used here: it starts at <i>zero</i> and moves up by one.</p>
      <p>In use, this method calls the <code>ResultSet getObject()</code> method to retrieve the data then casts it as String, or blank if the value is null. Each 
      row in the <code>ResultSet</code> is processed and the result for each appended to a <code>StringBuffer</code>, whose String value is then returned. There 
      is <b>no</b> end-of-line processing done so the template String should include an embedded newline or the ${newline} field.</p>
      <p>Any errors are reported at Error level and the first such will terminate the entire process. Use with care. If pResult is null or closed the method 
      returns null and nothing else happens.</p>
      @param pResults The <code>ResultSet</code> to be processed.
  */
  public String format(ResultSet pResults) {
    String retval = null;
    try {
      if((pResults == null) || (pResults.isClosed())) return null;
      int colMax = pResults.getMetaData().getColumnCount() + 1, iCount = 1;
      Object objTemp = null;
      StringBuffer sbReturns = new StringBuffer();
      String[] cellVals = new String[colMax];
      while(pResults.next()) {
        cellVals[0] = "" + iCount;
        for(int i=1; i < colMax; i++) {
          objTemp = pResults.getObject(i);
          cellVals[i] = (objTemp == null)?"":objTemp.toString();
        }
        sbReturns.append(fillTemplate(myTemplate, myData, cellVals, (iCount - 1)));
        iCount++;
      }
      retval = sbReturns.toString();
    } catch(Exception ex) { CapCom.error("Error formatting ResultSet", ex); }
    return retval;
  }

  /** <p>This method formats a <code>java.util.List</code> of template Strings into a <code>List</code> of filled templates using the same internal data for all. 
      If any template is null or blank the returned String will be empty but not null. If the <code>List</code> sent in is null or empty an empty (but non-null) 
      <code>List</code> is returned.</p>
      <p>The '@{count}' token is used here. It holds the (zero-based) number of the current template. That is, the first template is '0', the 
      second is '1', etc. NOTE: '@{count}' <b>is</b> available to the other <code>format()</code> methods but its value will always be zero!</p>
      @param pTemplates A <code>List&lt;String&gt;</code> of templates.
      @return A <code>List&lt;String&gt;</code> resulting from filling pTemplates with the current data.
  */
  public List<String> format(List<String> pTemplates) {
    ArrayList<String> retval = new ArrayList<String>();
    if((pTemplates == null) || (pTemplates.isEmpty())) return retval;
    String sTemp = "";
    for(String sTemplate : pTemplates) {
      sTemp = fillTemplate(sTemplate, myData, null, 0);
      retval.add(Question.validString(sTemp)?sTemp:"");
    }
    return retval;
  }

  /** This method operates exactly per the <code>format(List)</code> method except that it operates on a String[] array and the returned values may be null. The 
      '@{count} token equals the array index of the current template String. If the array sent in is null or empty an empty array is returned.
  */
  public String[] format(String[] pTemplates) {
    String[] retval = {};
    if((pTemplates == null) || (pTemplates.length < 1)) return retval;
    retval = new String[pTemplates.length];
    for(int i=0; i < pTemplates.length; i++) { retval[i] = fillTemplate(pTemplates[i], myData, null, i); }
    return retval;
  }

  /** <p>Use the internal template String for an array of Properties objects, with or without the internal data. If any element in the Properties[] array is null 
      or empty a blank (but not null) String is returned for that entry. The returned array will be one element longer than the one sent in: element zero is the 
      result from the internal data or blank if it is not used. If the array sent in is null or empty an empty array is returned. If the internal template is null 
      or empty the result is an empty array.</p>
      <p>The '@{count}' token for this method equals the index of the <b>result</b> array. That is, element 0 is the result of the internal data or blank, element 
      1 is from the first (element 0) Properties object in the array sent in, etc.</p>
      @param pProps The Properties[] array to apply to the internal template, in order.
      @param pInternal true to include the internal data, false to exclude it.
      @return A String[] array one element longer than the data sent in.
  */
  public String[] format(Properties[] pProps, boolean pInternal) {
    String[] retval = {};
    if((pProps == null) || (pProps.length < 1) || !Question.validString(myTemplate)) return retval;
    retval = new String[pProps.length + 1];
    retval[0] = (pInternal?format():"");
    for(int i=1; i < retval.length; i++) {
      if((pProps[i] == null) || (pProps[i].isEmpty())) {
        retval[i] = "";
      } else {
        retval[i] = fillTemplate(myTemplate, pProps[i], null, i);
      }
    }
    return retval;
  }

  /** Use this method to set a regular expression for the <code>reFormat()</code> methods. When set the regex is compiled into a <code>Pattern</code> which is 
      stored internally. If the regex cannot be compiled it is reported at Debug level and no action is taken.
      @param pRegex The regular expression to compile. If null or blank false is returned immediately.
      @return true if it compiled, false otherwise.
  */
  public boolean setRegex(String pRegex) {
    if(!Question.validString(pRegex)) return false;
    myPattern = null;
    try { myPattern = Pattern.compile(pRegex); } catch(Exception ex) { CapCom.debug("Failed to compile regex", ex); }
    return (myPattern != null);
  }

  /** Use this method to set a regular expression AND the two flags that cannot be set via embedded flag expressions.
      @param pRegex The regex to compile into a Pattern.
      @param ynLiteral true to set the <code>Pattern.LITERAL</code> flag, false to leave it out.
      @param ynCanonEq true to set <code>Pattern.CANON_EQ</code>, false not to.
  */
  public boolean setRegex(String pRegex, boolean ynLiteral, boolean ynCanonEq) {
    if(!Question.validString(pRegex)) return false;
    myPattern = null;
    int iFlags = 0 + (ynLiteral?Pattern.LITERAL:0) + (ynCanonEq?Pattern.CANON_EQ:0);
    try { myPattern = Pattern.compile(pRegex, iFlags); } catch(Exception ex) { CapCom.debug("Failed to compile regex", ex); }
    return (myPattern != null);
  }

  /** <p>This is the basic regular-expression formatting method. It uses the previously-set regex, applies it to the data String sent in, and formats the internal 
      template against it. It is tacitly assumed that the regex contains one or more capture groups. These groups are accessed by the '${<i>N</i>}' fields where 
      the number <i>N</i> is the (one-based) number of the capture group. The '${0}' field matches the entire data String. The replacement fields are guaranteed 
      to be non-null; if a capture group did not match its (null) value becomes a blank String.</p>
      <p>NOTE: All internal properties as well as the special fields are also available. If the regex does not match the only valid field will be '${0}'. If pData 
      is null or empty '' is returned. If no regex is set the return value is the same as <code>format()</code>; the '${0}' field will <b>not</b> be replaced.</p>
      @param pData The data String to match against the internal regex.
      @return The formatted internal template with capture groups replaced.
  */
  public String reFormat(String pData) { return _reFormat(pData, 0); }

  /** This method applies <code>reFormat()</code> to a String[] array. This is essentially the same method but the '@{count}' field is also available; it holds 
      the index of the element being formatted. If the array sent in is null or empty an empty array is returned. The elements of the array returned may be blank 
      but will never be null. Each String is trimmed before it is sent in.
      @param pData The String[] array to apply the <code>reFormat()</code> method to.
      @return A non-null String[] array holding the results of the formatting.
  */
  public String[] reFormat(String[] pData) {
    String[] retval = {};
    if((pData == null) || (pData.length < 1)) return retval;
    retval = new String[pData.length];
    for(int i=0; i < pData.length; i++) { retval[i] = _reFormat(pData[i].trim(), i); }
    return retval;
  }

  private String _reFormat(String pData, int iCount) {
    if(!Question.validString(pData)) return "";
    if(myPattern == null) return format();
    String retval = "";
    Matcher reMatch = myPattern.matcher(pData);
    String[] tokenz = new String[1];
    tokenz[0] = pData;
    if(reMatch.matches() && (reMatch.groupCount() > 0)) {
      tokenz = new String[reMatch.groupCount() + 1];
      tokenz[0] = pData;
      for(int i=1; i <= reMatch.groupCount(); i++) {
        tokenz[i] = reMatch.group(i);
        tokenz[i] = Question.validString(tokenz[i])?tokenz[i]:"";
      }
    }
    return fillTemplate(myTemplate, myData, tokenz, iCount);
  }

  /** <p>This method sets the key String format used when adding a <code>Question</code> to the internal Properties. Each format is a template String but these are 
      considerably simpler than the ones available to the <code>Question</code> class. Each template has exactly two special fields available to it:
      <ul>
        <li>'@{QID}' is the <code>Question</code>'s ID.</li>
        <li>'@{INDX}' is the value of the internal index clicker.</li>
      </ul></p>
      <p>Each format has a default value. Setting either to null or blank resets it to its default value. CRITICAL NOTE: These formats <b>must</b> be set properly 
      or the data will not be accessible from the template!</p>
      <h3>Format Setting Guide</h3>
      <p>There are two ways to handle <code>Question</code> formatting. The first way is to set the default format for the <code>Question</code> class before 
      generating any instances and using its Clicker. Alternately the ID may be set explicitly for each instance. This places the responsibility for proper 
      identification on the <code>Question</code> but, since two elements are added to the internal pizza for each <code>Question</code> added. The default key 
      formats handle this by appending '_Prob' or '_Ans' to the Question's ID. This is the best procedure if the Question IDs are descriptive in nature.</p>
      <p>The second way to handle things is to <i>ignore</i> the Question's ID and handle everything internally here. Set the formats to something like 
      'Q@{INDX}_P' and 'Q@{INDX}_A' and use the internal Clicker. When using this method the Clicker MUST be initialized to the first number used on the 
      template and the <code>click()</code> method MUST be called (on <code>MathTestFormatter</code> and NOT the <code>Question</code>!!!) after each Question 
      is added.</p>
      @param pProbFormat The template String used to format the key under which the Problem is added.
      @param pAnsFormat The template used to format the key for the Answer.
  */
  public void setQuestionKeyFormat(String pProbFormat, String pAnsFormat) {
    myProbKey = Question.validString(pProbFormat)?pProbFormat:DEFAULT_PROBLEM_KEY;
    myAnsKey = Question.validString(pAnsFormat)?pAnsFormat:DEFAULT_ANSWER_KEY;
  }

  /** As does <code>Question</code>, this class maintains an internal long usable to index the key values for Questions added to the internal data. Use this method 
      to set it. Any value less than zero becomes zero.
      @param pClickerValue The new value for the internal index clicker.
  */
  public void setClicker(long pClickerValue) { myClicker = (pClickerValue < 0)?0:pClickerValue; }

  /** Increment the internal clicker by +1. Returns a self-reference to allow method chaining. */
  public MathTestFormatter click() {
    myClicker++;
    return this;
  }

  /** <p>Add the data from a <code>Question</code> to the internal Properties. The Problem and Answer components are each added under their respective keys, set 
      using the {@link #setQuestionKeyFormat setQuestionKeyFormat()} method, and the @{QID} and @{INDX} fields will be set appropriately. A self-reference is 
      returned to allow method chaining. If the pQuestion is null this method returns silently. If the Problem or Answer is null or blank it is not added.</p>
      <p>The <code>Question</code> data is extracted via <code>formatProblem()</code> and <code>formatAnswer()</code> so it is critical to make sure the formats 
      cooperate with each other. The best thing to do is set the Question ID format to identify the Question (as it <b>should</b>) and nothing else. Set the 
      Problem and Answer formats to '&lt;1&gt;' and '&lt;2&gt;' respectively and don't worry about key-value-formatting. Include the '@{QID}' token when setting 
      Problem- and Answer-key formats here.</p>
      @param pQuestion The <code>Question</code> to be added to the internal data.
      @return A reference to this.
  */
  public MathTestFormatter addQuestion(Question pQuestion) {
    if(pQuestion == null) return this;
    String qID = ""+pQuestion.getID(), qProb = pQuestion.formatProblem(), qAns = pQuestion.formatAnswer();
    if(Question.validString(qProb)) myData.put(myProbKey.replaceAll("@[{]QID[}]", Matcher.quoteReplacement(qID)).replaceAll("@[{]INDX[}]", ""+myClicker), qProb);
    if(Question.validString(qAns)) myData.put(myAnsKey.replaceAll("@[{]QID[}]", Matcher.quoteReplacement(qID)).replaceAll("@[{]INDX[}]", ""+myClicker), qAns);
    return this;
  }

  /** This is a debugging tool. It prints all internal data at the Debug level. */
  public void dumpDebug() {
    CapCom.debug("- - - Dumping MathFormatter Data - - -\nTemplate:\t"+myTemplate+"\nProblem key:\t"+myProbKey+"\nAnswer key:\t"+myAnsKey+"\nClicker:\t"+myClicker);
    CapCom.debug("RegEx:\t"+((myPattern == null)?"null":myPattern.toString()));
    if(myData == null) {
      CapCom.debug("Internal Properties object is null");
    } else {
      CapCom.debug("Properties set:\t"+myData.size());
      for(String sKey : myData.stringPropertyNames()) {
        CapCom.debug(sKey+" = "+myData.getProperty(sKey));
      }
    }
    CapCom.debug("- - - Dump finished - - -");
  }

  /** Overriden to call <code>format()</code>. If nothing is set a default String is printed. */
  public String toString() {
    String retval = format();
    return Question.validString(retval)?retval:"MathTestFormatter(empty)";
  }

}

