package cox.jmatt.java.MathTools.util;

import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.event.*;
import java.io.*;
import cox.jmatt.java.MathTools.*;

/** <p>This class provides a MathPrinter implementation with all required methods and several extension points. The basic log and print functions are 
    covered and extra machinery is provided to extract the stack trace from a Throwable and return it as a String. In addition, BMP provides a simple GUI with 
    a TextArea to receive messages, Checkboxes to selectively enable/disable the three log levels, a Checkbox to enable/disable stack trace printing, and a 
    Button to clear the console (txaConsole). The standard GUI arranges the buttons and checkboxes to the right of the console. This is accomplished via the 
    standardGUI() method which returns a reference to the MathPrinter itself; the constructor DOES NOT establish the GUI!</p>
    <p>The extension machinery comes via the required processMathObject() method and the processGoButton() method. MathTools-using applications can use the first 
    method as required per the purpose of the application. The latter method is used if BMP is subclassed as the application's GUI. The processGoButton() method 
    is called whenever the [Go!] button is clicked. Default behavior is to test the log and print methods but it is primarily provided to allow for easy 
    subclassing: applications with minimal GUI requirements can simply subclass BMP and provide what they need there. (The default version also checks for version 
    information in the resource 'version.txt' in the top level of the JAR file.)</p>
    <p>For subclasses with more complex GUI requirements, standardGUI() is NOT mandatory! Each necessary component, txaConsole and the control Panel, has a 
    finalized public to build it. For a complex GUI, subclass BMP, set the layout as required and add txaConsole and getControlPanel() wherever they are 
    convenient. The getTitle() method is used in the standardGUI() to put a title in the NORTH position. If the GUI requirement is more complex still, implement 
    the MathPrinter interface manually.</p>
    <p>All MathPrinter methods except processMathObject() are finalized. This is done to prevent infinite loops: BMP is designed to RECEIVE from CapCom, not to 
    SEND to it. If overriding the processMathObject() method DO NOT call CapCom.processMathObject(), otherwise bad things will happen! Default behavior for the 
    processMathObject() method is to dump the <code>toString()</code>'ed contents of the argument to the clipboard if this is enabled and the argument is not 
    null.</p>
*/
public class BasicMathPrinter extends Panel implements MathPrinter, ActionListener, ClipboardOwner {

  /** This is the TextArea used for displaying information. The standard format has it in the center of a BorderLayout but subclasses may wish to change its 
      location. If so, it is critical that it be added somewhere, otherwise the printed information will not display properly!
  */
  public final TextArea txaConsole = new TextArea();

  private Checkbox cbxDebug = new Checkbox("Debug");
  private Checkbox cbxInfo = new Checkbox("Info", true);
  private Checkbox cbxError = new Checkbox("Error", true);
  private Checkbox cbxSTrace = new Checkbox("S Trace");
  private Checkbox cbxClip = new Checkbox("Clip", true);
  private Button btnClear = new Button("Clr");
  private Button btnGo = new Button("Go!");

  /** This method sets up the standard GUI: BorderLayout with getTitle() in the NORTH position, txaConsole in the CENTER, and the control panel at EAST. 
      It returns a reference to the MathPrinter itself to allow for customization. Subclasses can rearrange things but when only a MathPrinter is needed 
      the construct (new BasicMathPrinter()).standardGUI() can be used in place of the constructor.
  */
  public final MathPrinter standardGUI() {
    setLayout(new BorderLayout());
    add(new Label(getTitle(), Label.CENTER), BorderLayout.NORTH);
    add(txaConsole, BorderLayout.CENTER);
    add(getControlPanel(true), BorderLayout.EAST);
    return this;
  }
  /** The control Panel holds the [Go!] and [Clr] buttons as well as the log enable\disable checkboxes. It can be presented vertically or horizontally.
      @param pVertical true for a vertical bar, false for horizontal.
  */
  public final Panel getControlPanel(boolean pVertical) {
    Panel pnlTemp = new Panel(new GridLayout(1,0));
    if(pVertical) pnlTemp.setLayout(new GridLayout(0,1));
    pnlTemp.add(new Label(" "));
    btnGo.addActionListener(this);
    pnlTemp.add(btnGo);
    pnlTemp.add(new Label(" "));
    pnlTemp.add(cbxInfo);
    pnlTemp.add(cbxDebug);
    pnlTemp.add(cbxError);
    pnlTemp.add(cbxSTrace);
    pnlTemp.add(cbxClip);
    pnlTemp.add(new Label(" "));
    btnClear.addActionListener(this);
    pnlTemp.add(btnClear);
    pnlTemp.add(new Label(" "));
    return pnlTemp;
  }

  /** Logged messages and stack traces appear in the TextArea if their levels are enabled. */
  public final void mathLog(byte pLevel, String pMesg, Throwable pErr) {
    String prefix = null;
    switch(pLevel) {
      case MathPrinter.LOG_INFO: prefix = (cbxInfo.getState()?"[I] ":null); break;
      case MathPrinter.LOG_DEBUG: prefix = (cbxDebug.getState()?"[D] ":null); break;
      case MathPrinter.LOG_ERROR: prefix = (cbxError.getState()?"[E] ":null); break;
      default:
    }
    if(prefix == null) return;
    mathPrintln(prefix+pMesg);
    printStackTrace(pErr);
  }
  /** Extract and print the Stack Trace from a throwable. If 'S Trace' is enabled and pThr is not null. */
  public void printStackTrace(Throwable pThr) {
    if((pThr != null) && (cbxSTrace.getState())) mathPrintln(cox.jmatt.java.MathTools.CapCom.getStackTrace(pThr));
  }

  /** These messages go straight to the TextArea. */
  public final void mathPrintln(String pMesg) {
    if((pMesg == null) || (pMesg.trim().length() < 1)) return;
    txaConsole.append(pMesg+"\n");
  }

  /** This method provides access to the system Clipboard if the '[]Clip' Checkbox is enabled. If the Object passed in is not null and the clipboard is enabled, 
      the Object's <code>toString()</code> content is wrapped in the appropriate <code>Transferable</code> and sent to the clipboard. Any errors are reported at 
      Debug level. The Object sent in is returned unaltered.
  */
  public Object processMathObject(Object pObj) {

    if(cbxClip.getState() && (pObj != null)) {
      try {
        Clipboard theClip = Toolkit.getDefaultToolkit().getSystemClipboard();
        if(theClip == null) {
          mathLog(MathPrinter.LOG_DEBUG, "Failed to acquire clipboard.", null);
          return pObj;
        }
        theClip.setContents(new StringSelection(pObj.toString()), this);
      } catch(Exception ex) { mathLog(MathPrinter.LOG_DEBUG, "Clipboard error.", ex); }
    }
    return pObj;
  }

  /** Required for the <code>ClipboardOwner</code> interface. Default behavior prints a Debug-level message when ownership is lost. */
   public void lostOwnership(Clipboard pClip, Transferable pData) {
     mathLog(MathPrinter.LOG_DEBUG, "Clipboard ownership lost.", null);
   }

  /** This method provides an extension point for simple subclasses. It is called whenever the [Go!] button is clicked. The default implementation provides a basic 
      test of the logging and print methods, including setting itself as the Global Printer. It will also look for the resource '/version.txt' and print its 
      contents, if found. This file is in the top level of the JAR and stores, well, version information.
  */
  public void processGoButton() {
    CapCom.setPrinter(this);
    CapCom.println("Global Printer set to: "+this.toString());
    mathPrintln("Checking all logging levels...");
    mathLog(MathPrinter.LOG_INFO, "Info level is enabled.", null);
    mathLog(MathPrinter.LOG_DEBUG, "Debug level is enabled.", null);
    mathLog(MathPrinter.LOG_ERROR, "Error level is enabled.", null);
    printStackTrace(new Exception("Testing printStackTrace()."));
    CapCom.println("- - Version  Info - -");
    try {
      BufferedReader brVersion = new BufferedReader(new InputStreamReader(this.getClass().getResourceAsStream("/version.txt")));
      String sLine = null;
      while((sLine = brVersion.readLine()) != null) { CapCom.println(sLine); }
      brVersion.close();
    } catch(Exception ignored) { CapCom.println("Unavailable"); }
      CapCom.println("- - - - - - - - - - -");
    mathPrintln("Press [Clr] to clear the screen.");
  }

  /** Override this method to change BMP's title to something other than 'Basic MathPrinter'. */
  public String getTitle() { return "Basic MathPrinter"; }
  /** Clear the TextArea. */
  public void clearScreen() { txaConsole.setText(""); }

  /** This method checks to see if the [Clr] or [Go!] buttos have been pressed. Subclasses overriding this method should call super.actionPerformed() to allow for 
      proper clearing of the screen and [Go!] button processing.
  */
  public void actionPerformed(ActionEvent aev) {
    if(aev.getSource() == btnClear) clearScreen();
    if(aev.getSource() == btnGo) processGoButton();
  }
}

