package cox.jmatt.java.MathTools.util;

import java.util.logging.*;
import cox.jmatt.java.MathTools.CapCom;

/** <p>This class is designed to complement <code>CCLogHandler</code>. It sends any logging errors to <code>CapCom.error()</code>. It is also worth noting that 
    this class is NOT necessary for the CapCom <code>Handler</code> class; log system error management is built in! Rather, this class exists to allow error 
    management for non-MathTools <code>Handler</code> subclasses.</p>
    <p>By the first error is logged and the rest are ignored, per its superclass. This behavior can be overridden to log all errors. This is accomplished via 
    constructor or the <code>setLogAlways()</code> method.<p>
    <p>It is important to note that this class is used to report error conditions that happen as a result of logging an application error. The original error that 
    caused the log request to be generated is NOT reported here, only any errors that occur as a result of the logging call. Six failure codes are represented. 
    They are: 'Generic,' 'Write,' 'Flush,' 'Close,' 'Open' and 'Format.' If the integer fail code does not correspond to one of these the fail code is 'Unknown.' 
    If an Exception is generated it is also printed.</p>
*/

public class CCLogErrorManager extends ErrorManager {

  private static final String[] ERROR_CODES = {"Generic", "Write", "Flush", "Close", "Open", "Format"};

  private boolean gotPrinter = false, printAlways = false, firstError = true;

  /** Standard constructor. */
  public CCLogErrorManager() {
    super();
    gotPrinter = CapCom.isPrinterSet();
  }

  /** Constructor used to change the 'always print' status to true or explicitly leave it disabled. */
  public CCLogErrorManager(boolean pPrintAlways) {
    super();
    printAlways = pPrintAlways;
    gotPrinter = CapCom.isPrinterSet();
  }

  /** <p>This method is called when the <i>logging</i> system generates an error. If no Global Printer is set it returns immediately. If default behavior is 
      in effect the second and subsequent errors are silently ignored. If 'print all' is enabled then all such errors will be printed.</p>
      <p>The error code is printed if recognized, otherwise the value is 'Unknown'.</p>
  */
  public void error(String pMesg, Exception pErr, int pCode) {
    if(!gotPrinter) return;
    if(!(printAlways || firstError)) return;
    String failCode = "Unknown ";
    if((pCode >= 0) && (pCode < ERROR_CODES.length)) failCode = ERROR_CODES[pCode];
    CapCom.error("Log error ["+failCode+"]: "+pMesg, pErr);
    firstError = false;
  }

  /** <p>Use this method to enable/disable default behavior. By default only the first error message is printed. Call this method with 'true' to print ALL 
      error messages.</p>
  */
  public void setLogAlways(boolean pEnable) { printAlways = pEnable; }

}

