package cox.jmatt.java.MathTools.util;

import java.util.logging.*;
import cox.jmatt.java.MathTools.CapCom;

/** <p>This class allows the <code>java.util.logging</code> framework to be attached to the MathTools logging methods. It is intended to allow third-party logging 
    frameworks to take advantage of MathTools logging. Since most such packages make provision for the Java logging package it allows for easy integration.</p>
    <p>Each <code>java.util.logging.Level</code> constant is mapped to one of the <code>MathPrinter</code> log levels. This is done numerically to allow custom 
    Java levels to map appropriately. If the int value of the <code>Level</code> is greater than or equal to WARNING the message is logged at Error level. If it 
    between INFO (inclusive) and WARNING (exclusive) it goes to Info level and anything less than INFO goes to Debug.</p>
    <p>This <code>Handler</code> logs messages naively. The <code>publish()</code> method does not check to see whether or not its <code>Logger</code> accepts 
    a particular Level so the <code>Logger</code> must do this manually if it is desired. The only concession made is in the <code>isLoggable()</code> method. If 
    no Global Printer is set that method returns false. Likewise, <code>publish()</code> returns immediately no printer exists.</p>
    <p>To use this class:
      <ol>
        <li>Create or obtain a <code>java.util.logging.Logger</code> instance.</li>
        <li>Create an instance of this class.</li>
        <li>If desired, set a <code>java.util.logging.Formatter</code> to format the logged output. The default is a <code>SimpleFormatter</code>.</li>
        <li>Add this instance to the <code>Logger</code>.</li>
      </ol></p>
    <p>To use CapCom logging with a third-party package follow the steps above but modified for the desired package. Most logging packages provide adapters for 
    <code>java.util.logging</code> that fit it, hence CapCom, into themselves.</p>

    <h3>Error Management</h3>
    <p>This concerns errors generated by the <i>logging system</i> and not the application. These errors are, by default, reported at CapCom Error level along 
    with a failure code, exception and optional message. The <code>reportError()</code> method exactly mirrors the <code>error()</code> method in the 
    <code>CCLogErrorManager</code> class.</p>
    <p>The <code>CCLogErrorManager</code> can be used with this class but doing so would be redundant. It is designed to allow <i>other</i> <code>Handler</code> 
    classes access to CapCom logging. All of its behavior is mirrored internally here and enabled by default.</p>
*/

public class CCLogHandler extends Handler {

  private static final String[] ERROR_CODES = {"Generic", "Write", "Flush", "Close", "Open", "Format"};

  private boolean gotPrinter = true, printAlways = false, firstError = false, useErrorManager = false;
  private Formatter myFormatter = null, myDefaultFormatter = new SimpleFormatter();

  /** This constructor creates an instance and determines whether or not a Global MathPrinter is set. If so, logging messages will be printed. If not, not. */
  public CCLogHandler() {
    super();
    gotPrinter = CapCom.isPrinterSet();
  }

  /** Empty method. No buffering occurs. */
  public void flush() {}

  /** This method removes all internal Formatters. Calling <code>publish()</code> after this method returns <b>will</b> result in a NullPointerException. */
  public void close() {
    myFormatter = null;
    myDefaultFormatter = null;
  }

  /** <p>Send <code>CapCom</code> a logging call. If no <code>Formatter</code> is set a <code>SimpleFormatter</code> is used. If 'pLogRec' is null or if its 
      Level is null this method returns silently and no logging takes place. If logging does occur, the <code>java.util.logging</code> package levels are mapped 
      numerically to <code>CapCom</code> as follows: SEVERE and WARNING map to Error level; INFO maps to Info level and anything less maps to Debug.</p>
      <p>This method publishes records naively. It does not check loggability; rather, it publishes any non-null record with a non-null Level that is passed 
      in. <code>Logger</code> instances requiring discrimination of log levels should implement it manually. It does check to see if a MathPrinter is set (or was 
      set when this instance was created). If not it returns silently.</p>
  */
  public void publish(LogRecord pLogRec) {
    if(!gotPrinter) return;
    if((pLogRec == null) || (pLogRec.getLevel() == null)) return;
    myFormatter = getFormatter();
    if(myFormatter == null) myFormatter = myDefaultFormatter;
    String theRecord = myFormatter.format(pLogRec);
    int howBad = pLogRec.getLevel().intValue();
    if(howBad >= Level.WARNING.intValue()) {
      CapCom.error(theRecord);
    } else if(howBad >= Level.INFO.intValue()) {
      CapCom.info(theRecord);
    } else {
      CapCom.debug(theRecord);
    }
  }

  /** <p>This method ANDs <code>super.isLoggable()</code> with the presence of a Global Printer. The latter depends on whether or not a printer was set when 
      this instance was created. If no printer was set this method returns false. Period.</p>
  */
  public boolean isLoggable(LogRecord pLogRec) { return (gotPrinter && super.isLoggable(pLogRec)); }

  /** <p>This method is overridden to use <code>CapCom</code>. If an <code>ErrorManager</code> should be used set one and call 'setUseErrorManager(true)'.</p>
      <p>Unless 'logAlways' is set true this method ignores any calls after the first. If no Global Printer is set then no CapCom logging occurs but if an 
      external manager is set it will still be called.</p>
  */
  protected void reportError(String pMesg, Exception pErr, int pCode) {
    if(!(printAlways || firstError)) return;
    firstError = false;
    if(useErrorManager) {
      super.reportError(pMesg, pErr, pCode);
    } else {
      if(!gotPrinter) return;
      String failCode = "Unknown ";
      if((pCode >= 0) && (pCode < ERROR_CODES.length)) failCode = ERROR_CODES[pCode];
      CapCom.error("Log error ["+failCode+"]: "+pMesg, pErr);
    }
  }

  /** <p>By default only the first log system error message is displayed; use this method to change that. Calling this method with 'true' causes all log system 
      messages to be displayed. Any such appear at Error level.</p>
  */
  public void setLogAlways(boolean pEnable) { printAlways = pEnable; }

  /** Use this method to enable an external <code>ErrorManager</code>. By default this method reports all errors generated by the logging system to CapCom. Calling 
      this method with 'true' causes the external manager to be used instead.
  */
  public void setUseErrorManager(boolean pEnable) { useErrorManager = pEnable; }

}

