package cox.jmatt.java.MathTools.util;

/** <p>This interface specifies a generic way to authenticate at three levels: student, proctor, and instructor. It is of use if MathTools is deployed as a part 
    of an independent testing package. The specific authentication methodology used is not specified. Rather, authentication is handled very generically at one of 
    the three levels. One method is provided to initialize an implementation ant the other is the acutal authentication method.</p>
    <p>The pre-defined authentication levels are:
    <ol>
      <li><b>Student</b> level represents a person taking a test or whose authentication should be the most limited.</li>
      <li><b>Proctor</b> access is for a person giving or proctoring a test; it should be higher than student level.</li>
      <li><b>Instructor</b> level is for the person creating the test and ultimately responsible for its handling and disposition.</li>
    </ol>
    Two debugging and development levels are also provided: ADMIN and NONE. The former is for development and program testing, the latter is used to indicate 
    that <b>no</b> authentication level is given/required.</p>
    <p><code>CapCom</code> provides a method to specify a global Authenticator along with a method to check authentication using it. It is set using the full 
    name of the class implementing this interface along with the context Object used to initialize the impementation. Classes implementing the <code>MAuthenticator</code> 
    interface must provide a zero-argument contstructor; <code>CapCom</code> uses it when creating the Authenticator.</p>
    <h3>Generating Authentication Codes</h3>
    <p><code>MAuthenticator</code> specifies the <code>generateCodes()</code> method to generate authentication codes. This method is <i>optional</i> but is provided 
    if needed. Implementations not using it should simply return an empty array. Implementations that do use it must accept an authenticaion level for the codes 
    generated and the number of codes to generate. An <code>Object...</code> parameter is also supplied for implementation-specific use.</p>
*/
public interface MAuthenticator {

  /** Static constant used to indicate authentication at Student level. */
  public static final byte AUTH_STUDENT = 1;
  /** Static constant used to indicate authentication at Proctor level. */
  public static final byte AUTH_PROCTOR = 2;
  /** Static constant used to indicate authentication at Instructor level. */
  public static final byte AUTH_INSTRUCTOR = 3;
  /** Static constant used to indicate authentication at Admin level. This is intended primarily for debugging and development and should not be used as a deployment 
      level.
  */
  public static final byte AUTH_ADMIN = -2;
  /** Static constant used to indicate no authentication level. */
  public static final byte AUTH_NONE = 0;

  /** Implement this method to initialize a <code>MAuthenticator</code> implementation. By intent this method should be called once but can be called more than 
      once. Each call should not only initialize and configure whatever the authenticator needs to function, it should also clear out any old data left from prior 
      use. More specifically, calling this method with the same context Object should yield the same authenticator regardless of whether it is the first, tenth, 
      or hundredth call.
      @param pContext The context Object containing information necessary to initialize the instance.
  */
  public void initMAuthenticator(Object pContext);

  /** The raison d'etre for this interface. The return value indicates whether or not the supplied authentication code is valid for the level specified.
      @param pLevel The AUTH_ constant for the desired authentication level.
      @param pCode The authentication code to check at the specified level.
      @return true if the code is valid, false otherwise.
  */
  public boolean mAuthenticate(byte pLevel, String pCode);

  /** This method, if defined, is used to generate authentication codes at the specified level. Implementations that do not use this method should return an empty 
      array but may return null instead. The mandatory arguments are the AUTH_ level of the codes to generate and the number of codes to generate. The 
      <code>Object...</code> argument is supplied for implementations to use as required and must be specified in the documentation!
      @param pLevel The AUTH_ level constant for the desired codes.
      @param pNum The number of codes to generate.
      @param pParams Implementation-specific parameters for generating the codes.
  */
  public String[] generateCodes(byte pLevel, int pNum, Object... pParams);

}

