package cox.jmatt.java.MathTools.util;

import java.io.*;
import java.util.*;
import cox.jmatt.java.MathTools.*;

/** <p>This is a helper method for creating MMatrices. The MMatrix class is deliberately light on constructors to ease the burden of scripting. The methods in this 
    class are designed in pairs: one method to create a double[][] array and another to create a MMatrix from the same signature.</p>
*/
public class MMatrixBuilder {

  public MMatrixBuilder() { super(); }

  /** This method creates a simple 3x3 MMatrix with the elements '1 2 3/4 5 6/7 8 9.' As implied by its name this MMatrix is useful for testing purposes. The 
      default variables are 'x,' 'y,' and 'z.'
  */
  public static MMatrix TEST_MATRIX() {
    MMatrix retval = createMMatrix("1,2,3:4,5,6:7,8,9", ":", ",");
    retval.setVariables("x y z");
    return retval;
  }

  /** Use this method to create a MMatrix directly. */
  public static MMatrix createMMatrix(double[][] pArray) { return new MMatrix(pArray); }

  /** Create a MMatrix filled with a particular constant. */
  public static MMatrix createMMatrix(int pRows, int pCols, double pFill) { return new MMatrix(createArray(pRows, pCols, pFill)); }

  /** Create a non-ragged double[][] filled with the supplied value. If the pRows or pCols is zero, it is increased to 1. 
      @param pRows The number of rows in the returned array.
      @param pCols The number of columns in the returned array.
      @param pFill The fill value for the new array.
      @return A fresh double[][] filled with the specified value.
  */
  public static double[][] createArray(int pRows, int pCols, double pFill) {
    if(pRows < 1) pRows = 1;
    if(pCols < 1) pCols = 1;
    double[][] retval = new double[pRows][pCols];
    for(int i=0; i < pRows; i++) {
      for(int j=0; j < pCols; j++) { retval[i][j] = pFill; }
    }
    return retval;
  }
  
  /** Create a non-ragged int[][] filled with the supplied value. If the pRows or pCols is zero, it is increased to 1. This method does NOT have a MMatrix creator 
      associated with it: MMatrix operates internally on floating-point (double) values and the integer version is for presentation ONLY!
      @param pRows The number of rows in the returned array.
      @param pCols The number of columns in the returned array.
      @param pFill The fill value for the new array.
      @return A fresh double[][] filled with the specified value.
  */
  public static int[][] createIntArray(int pRows, int pCols, int pFill) {
    if(pRows < 1) pRows = 1;
    if(pCols < 1) pCols = 1;
    int[][] retval = new int[pRows][pCols];
    for(int i=0; i < pRows; i++) {
      for(int j=0; j < pCols; j++) { retval[i][j] = pFill; }
    }
    return retval;
  }

  /** Create a MMatrix with the same number of rows and columns, filled with the pFill value. */
  public static MMatrix createSquareMMatrix(int pDimension, double pFill) { return new MMatrix(createArray(pDimension, pDimension, pFill)); }

  /** Create a square double[][] filled with the supplied value. If the dimension is zero it is increased to 1. */
  public static double[][] createSquareArray(int pDimension, double pFill) { return createArray(pDimension, pDimension, pFill); }

  /** Create a MMatrix with the specified 'background' pFill and pDiagonal down its main diagonal. */
  public static MMatrix createDiagonalMMatrix(int pDimension, double pFill, double pDiagonal) { return new MMatrix(createDiagonalArray(pDimension, pFill, pDiagonal)); }

  /** Create a square double[][] filled with the pFill value and pDiagonal along its main diagonal.
      @param pDimension The number of rows and columns in the returned array.
      @param pFill The fill value ('background') value in the array.
      @param pDiagonal The diagonal value in the resulting array.
      @return A double[][] with pFill in all non-diagonal entries and pDiagonal along the main diagonal.
  */
  public static double[][] createDiagonalArray(int pDimension, double pFill, double pDiagonal) {
    if(pDimension < 1) pDimension = 1;
    double[][] retval = createArray(pDimension, pDimension, pFill);
    for(int i = 0; i < pDimension; i++) { retval[i][i] = pFill; }
    return retval;
  }

  /** Create a MMatrix filled with random values. Rows, columns, min and max element values and 'allow zero element' can all be specified. */
  public static MMatrix createRandMMatrix(int pRows, int pCols, int pMin, int pMax, boolean pAllowZero) {
    return new MMatrix(createRandArray(pRows, pCols, pMin, pMax, pAllowZero));
  }

  /** Create an array of the given dimension filled with random values. This method uses the <code>MathEngine.random()</code> methods to generate values so the 
      minimum element-value is included and the maximum value is excluded. The 'allowZero' argument determines which method to use. If either dimension is zero 
      or less an empty array is returned.
      @param pRows The number of rows to create.
      @param pCols The number of columns in each row.
      @param pMin The minimum value for each element. This is included in the possible values.
      @param pMax The maximum value for each element. Not included in possible values.
      @param pAllowZero true to allow zero as a possible value, false to exclude it.
  */
  public static double[][] createRandArray(int pRows, int pCols, int pMin, int pMax, boolean pAllowZero) {
    double[][] retval = {};
    if((pRows < 1) || (pCols < 1)) return retval;
    retval = new double[pRows][pCols];
    for(int i = 0; i < pRows; i++) {
      for(int j=0; j < pCols; j++) {
        if(pAllowZero) {
          retval[i][j] = (double) MathEngine.random(pMin, pMax);
        }  else {
          retval[i][j] = (double) MathEngine.nzRandom(pMin, pMax);
        }
      }
    }
    return retval;
  }

  /** Create a non-ragged copy of the double[][] passed in. The number of rows stays the same but the number of columns will be that of the longest row. Any 
      padded elements will be zero. If the array sent in is null or has length zero, the zero array is returned.
  */
  public static double[][] smoothCopy(double[][] pAry) {
    double[][] retval = {{0}};
    if((pAry == null) || (pAry.length == 0)) return retval;
    int nRows = pAry.length, nCols = 0;
    for(int i=0; i < nRows; i++) {
      if((pAry[i] != null) && (pAry[i].length > nCols)) nCols = pAry[i].length;
    }
    retval = createArray(nRows, nCols, 0);
    for(int i=0; i < nRows; i++) {
      if((pAry[i] == null) || (pAry[i].length < 1)) continue;
      for(int j = 0; j < pAry[i].length; j++) { retval[i][j] = pAry[i][j]; }
    }
    return retval;
  }

  /** Create a MMatrix from a String[] array with each String split by pColSplit. */
  public static MMatrix createMMatrix(String[] pRows, String pColSplit) { return new MMatrix(createArray(pRows, pColSplit)); }

  /** Create a double[][] from a String[] array. Each element in the array is split() on pColSplit and the results are parsed into a double[] which is then added 
      to the outer double[]. The return array is a smoothCopy() of the original values. If pColSplit is null or empty a single space is used. If the String[] array 
      is null or empty the zero array is returned. If any single entry cannot be parsed as a double that entry is zero.
  */
  public static double[][] createArray(String[] pRows, String pColSplit) {
    double[][] retval = MMatrix.ZERO();
    if((pRows == null) || (pRows.length < 1)) return retval;
    String pSplitter = ((pColSplit != null) && (pColSplit.length() > 0))?pColSplit:" ";
    String[] sRowTemp = null;
    double[] daTemp = new double[1];
    retval = new double[pRows.length][];
    for(int i=0; i < retval.length; i++) {
      if((pRows[i] == null) || (pRows[i].length() < 1)) {
        retval[i] = new double[1]; retval[i][0] = 0;
        continue;
      }
      sRowTemp = pRows[i].split(pSplitter);
      daTemp = new double[sRowTemp.length];
      for(int j = 0; j < daTemp.length; j++) {
        try{ daTemp[j] = Double.parseDouble(sRowTemp[j]); } catch(Exception ex) { CapCom.error("Error parsing ["+i+","+j+"]= "+sRowTemp[j], ex); daTemp[j] = 0; }
      }
      retval[i] = daTemp;
    }
    return smoothCopy(retval);
  }

  /** Create a MMatrix from a single String, broken by rows and then columns. */
  public static MMatrix createMMatrix(String pArray, String pRowSplit, String pColSplit) { return new MMatrix(createArray(pArray, pRowSplit, pColSplit)); }

  /** <p>This method creates a double[][] from a (single!) String, broken by rows and then columns by pRowSplit and pColSplit respectively. If null, pRowSplit 
      defaults to a newline character and pColSplit defaults to a single space.</p>
      <p>NOTE: The row- and column-split defaults are different from the <code>MMatrix.createMMatrix()</code> method! This method, by default, parses a format 
      more 'natural' to reading. That is, space-separated element values organized (newline) into visual rows.</p>
  */
  public static double[][] createArray(String pArray, String pRowSplit, String pColSplit) {
    double[][] retval = {{0}};
    if(!Question.validString(pArray)) return retval;
    String splitter = Question.validString(pRowSplit)?pRowSplit:"\n";
    return createArray(pArray.trim().split(splitter), pColSplit);
  }

  /** Convenience method for createMMatrix(pArray, null, null). */
  public static MMatrix createMMatrix(String pArray) { return new MMatrix(createArray(pArray, null, null)); }

  /** Convenience method for createArray(pArray, null, null). */
  public static double[][] createArray(String pArray) { return createArray(pArray, null, null); }

  /** This method replaces one element in a MMatrix with the specified value. It does <b>not</b> change the MMatrix, it creates an altered copy! If the MMatrix 
      to change is null or empty a default MMatrix is returned. If the element location is outside the MMatrix nothing happens and a simple copy is returned.
      @param pMatrix The MMatrix to 'change.'
      @param pRow The row of the element to change.
      @param pCol The column of the element to change.
      @param pVal The value to insert.
      @return A copy of pMatrix with the alteration made.
  */
  public static MMatrix changeElement(MMatrix pMatrix, int pRow, int pCol, double pVal) {
    if(pMatrix == null) return new MMatrix();
    double[][] coefs = pMatrix.elements();
    if((pRow < coefs.length) && (pCol < coefs[pRow].length)) coefs[pRow][pCol] = pVal;
    return new MMatrix(coefs);
  }

  /** Change an entire MMatrix row. As with <code>changeElement()</code> this copies the MMatrix and works on the copy. The row sent in is copied to its length 
      or that of the row it replaces. Any elements left in the MMatrix are unalterd, any leftovers in the replacement row are ignored. If pMatrix is null a zero 
      MMatrix is returned. If the replacement row index is greater than pMatrix.rows() a simple copy is returned. If the replacement row is null or empty a simple 
      copy is returned.
      @param pMatrix The MMatrix to change.
      @param pRow The row to replace.
      @param pVals The double[][] array to replace pRow.
      @return A copy of pMatrix with the row replacement done.
  */
  public static MMatrix replaceRow(MMatrix pMatrix, int pRow, double[] pVals) {
    if((pMatrix == null) || (pRow < 0)) return new MMatrix();
    double[][] coefs = pMatrix.elements();
    if((pVals != null) && (pRow < coefs.length)) {
      for(int i = 0; i < Math.min(coefs[i].length, pVals.length); i++) coefs[pRow][i] = pVals[i];
    }
    return new MMatrix(coefs);
  }

  /** Replace some or all of a MMatrix column with new values. A copy is created and changed. As much of the 'column' sent in is copied as possible. Leftovers in 
      the MMatrix are unchaged, leftovers in the array sent in are ignored. If pMatrix is null a zero MMatrix is returned. If pVals is null or the column to replace 
      is out of bounds a simple copy is returned.
      @param pMatrix The MMatrix to edit.
      @param pCol The column to replace.
      @param pVals The column replacement values.
      @return A copy of pMatrix with the changes made.
  */
  public static MMatrix replaceColumn(MMatrix pMatrix, int pCol, double[] pVals) {
    if((pMatrix == null) || (pCol < 0)) return new MMatrix();
    double[][] coefs = pMatrix.elements();
    if((pVals != null) && (pCol < pMatrix.columns())) {
      for(int i=0; i < Math.min(coefs.length, pVals.length); i++) coefs[i][pCol] = pVals[i];
    }
    return new MMatrix(coefs);
  }

  /** <p>Convert a double[][] array into a String[][] array. The row prefix is prepended to each row and the row suffix is appended. An element prefix and suffix 
      can be specified for each element by index. If the element array runs out before the row does any remaining elements are left unadorned. If the element prefix 
      or suffix arrays have leftovers they are ignored. Null prefixes and suffixes are ignored but blanks are not.</p>
      <p>This method makes no assumption about the size or smoothness of the array. If pAry is ragged so will be the returned array. If pAry is null or empty an 
      empty array is returned.</p>
      @param pAry The double[][] array to convert to String[][].
      @param rowPfx The prefix added to each row. Ignored if null.
      @param rowSfx The suffix appended after each row. Ignored if null.
      @param elPfx The array of prefixes to prepend to row elements. Ignored if null, individual elemenents also ignored if null.
      @param elSfx The array of element suffixes appended. Ignored if null as are individual elements.
      @return The newly String-ified double[][] array.
  */
  public static String[][] toStringArray(double[][] pAry, String rowPfx, String rowSfx, String[] elPfx, String[] elSfx) {
    String[][] retval = {{}};
    if((pAry == null) || (pAry.length < 1)) return retval;
    retval = new String[pAry.length][];
    rowPfx = Question.validString(rowPfx)?rowPfx:""; rowSfx = Question.validString(rowSfx)?rowSfx:"";
    StringBuffer sbElt;
    for(int iNdx = 0; iNdx < pAry.length; iNdx++) {
      retval[iNdx] = new String[0];
      if((pAry[iNdx] == null) || (pAry[iNdx].length < 1)) continue;
      retval[iNdx] = new String[pAry[iNdx].length];
      sbElt = new StringBuffer();
      for(int jNdx = 0; jNdx < pAry[iNdx].length; jNdx++) {
        if(jNdx == 0) sbElt.append(rowPfx);
        if((elPfx != null) && (jNdx < elPfx.length) && (elPfx[jNdx] != null) ) sbElt.append(elPfx[jNdx]);
        sbElt.append(pAry[iNdx][jNdx]);
        if((elSfx != null) && (jNdx < elSfx.length) && (elSfx[jNdx] != null) ) sbElt.append(elSfx[jNdx]);
        if(jNdx == (retval[iNdx].length - 1)) sbElt.append(rowSfx);
        retval[iNdx][jNdx] = sbElt.toString();
        sbElt = new StringBuffer();
      }
    }
    return retval;
  }

  /** This (debugging!) method prints the contents of a double[][] at Debug level. */
  public static void dump(double[][] pAry) {
    if((pAry == null) || (pAry.length < 1)) {
      CapCom.debug("Array was null or empty.");
      return;
    }
    CapCom.debug("Dumping double[][].");
    StringBuffer sbTemp = null;
    for(int i=0; i < pAry.length; i++) {
      sbTemp = new StringBuffer("["+i+"]\t");
      if((pAry[i] == null) || (pAry[i].length < 1)) {
        sbTemp.append("Row was null or empty");
        CapCom.debug(sbTemp.toString());
        continue;
      }
      for(int j=0; j < pAry[i].length; j++) { sbTemp.append(pAry[i][j]).append("\t"); }
      CapCom.debug(sbTemp.toString());
    }
    CapCom.debug("--- End of Dump ---");
  }
  /** Dump a double[] at Debug level. */
  public static void dump(double[] pAry) {
    if((pAry == null) || (pAry.length < 1)) {
      CapCom.debug("Array was null or empty.");
      return;
    }
    CapCom.debug("Dumping double[]");
    for(int i=0; i < pAry.length; i++) { CapCom.debug("["+i+"]= "+pAry[i]); }
    CapCom.debug("--- End of Dump ---");
  }

  /** Print MMatrix information. The rows, columns, order, and element array are all printed and the method is <code>CapCom.println()</code>. */
  public static void printMMatrix(MMatrix pMatrix) {
    if(pMatrix == null) {
      CapCom.println("Matrix was null");
      return;
    }
    CapCom.println("-- Printing MMatrix --\nRows: "+pMatrix.rows()+"\nCols: "+pMatrix.columns()+"\nOrder: "+pMatrix.order());
    double[][] pAry = pMatrix.elements();
    StringBuffer sbTemp = null;
    for(int i=0; i < pAry.length; i++) {
      sbTemp = new StringBuffer("["+i+"]\t");
      if((pAry[i] == null) || (pAry[i].length < 1)) {
        sbTemp.append("Row was null or empty");
        CapCom.println(sbTemp.toString());
        continue;
      }
      for(int j=0; j < pAry[i].length; j++) { sbTemp.append(pAry[i][j]).append("\t"); }
      CapCom.println(sbTemp.toString());
    }
    CapCom.println("Data String: "+pMatrix.toString(null, null));
    CapCom.println("---");
  }

  /** Transpose an array by interchanging its rows and columns. A smooth copy is made first so null or empty arrays return the zero array.
      @param pAry A M x N array.
      @return The N x M transposed array.
  */
  public static double[][] transposeArray(double[][] pAry) {
    pAry = smoothCopy(pAry);
    double[][] retval = new double[pAry[0].length][pAry.length];
    for(int i=0; i < pAry.length; i++) {
      for(int j = 0; j < pAry[i].length; j++) retval[j][i] = pAry[i][j];
    }
    return retval;
  }

  /** Return a SCROB instance of this class. */
  public cox.jmatt.java.MathTools.scrob.MMatrixBuilder_SCROB getSCROB() { return new cox.jmatt.java.MathTools.scrob.MMatrixBuilder_SCROB(); }

}
