package cox.jmatt.java.MathTools.util;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.io.*;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.util.*;
import cox.jmatt.java.MathTools.test.*;

/** <p>This Applet is designed to allow MathTools scripting via a web browser and to act as CapCom's MathPrinter on an (X)HTML page. The GUI consists of a TextArea 
    used for displaying log and print messages along with Checkboxes to enable/disable log levels, a Checkbox to enable the Clipboard, a [Calc] button to show a 
    <code>QuickRPNCalculator</code>, a [Clr] button to clear the screen. When using MTApplet make sure to leave room for these components!</p>
    <p>All MathPrinter method are finalized method because they are not designed to be subclassed. To script MathTools using a browser, embed a MTApplet somewhere 
    in the page. Obtain a reference to it from the scripting language, call setAsPrinter() if desired, and program away!</p>
    <p>NOTE: All create__() methods in this Applet return <b>static</b> versions of their class. To obtain the SCROB aliases call the (instance) getSCROB() method 
    on the Object returned.</p>
    <p>Applets generally do not have filesystem access so <code>MTApplet</code> will not create and cannot access the <code>MathFileIO</code> class. It can, however, 
    provide Applet-appropriate access via <code>Class.getResourceAsStream()</code>. This functionality is available through the <code>getResourceStream()</code> 
    and <code>getResourceReader()</code> methods. They both operate from <code>MTApplet</code>'s Classloader (<b>not</b> the Math ClassLoader) and must work within 
    the security and access restrictions imposed on any Applet. Use as needed but USE WITH CARE!</p>
    <p>Security permitting, the <code>processMathObject()</code> method provides access to the system clipboard. If the Object passed in is not null, the '[]CB' 
    Checkbox is enabled, and the security manager allows it the String content of the Object passed in is dumped to the clipboard.</p>
    <p>HTML documents using MTApplet on a local machine may need access to Capcom's MAuthenticator or Accessory creation methods it is possible to set the appropriate 
    flags using &lt;param&gt; tags. The two parameters are <code>control.MAuthenticator</code> for setting MAuthenticator access and <code>control.Accessory</code> to 
    allow accessory creation. Setting either of these to 'true' enables the correspoinding field in CapCom.</p>
    <h3>The Calculator</h3>
    <p><code>MTApplet</code> includes a <code>QuickRPNCalculator</code> usable from a dialog. Since calculators are not always a good idea, it can be enabled or 
    disabled via the <code>setCalculatorEnabled()</code>. The <code>showCalc()</code> method can bring up the calculator from within a script, if it is enabled. 
    By default the calculator is disabled. To enable by default use a &lt;param&gt; tag: set <code>control.Calculator</code> to true.</p>
*/
public class MTApplet extends Applet implements ActionListener, MathPrinter, ClipboardOwner {


  private TextArea txaConsole = new TextArea();
  private Checkbox cbxInfo = new Checkbox("I");
  private Checkbox cbxDebug = new Checkbox("D");
  private Checkbox cbxError = new Checkbox("E");
  private Checkbox cbxStackTrace = new Checkbox("ST");
  private Checkbox cbxClip = new Checkbox("CB", true);
  private Button btnClear = new Button("Clr"), btnCalc = new Button("Calc");
  private boolean amInited = false, calcEnabled = false;
  private static java.util.Vector<InputStream> openStreams = new java.util.Vector<InputStream>();
  private Dialog dlgCalc = null;

  /** Calls <code>setUpGUI()</code>. If subclassing this must be done manually! */
  public void init() { setUpGUI(); }

  /** Set up the GUI. Finalized to allow subclassing. */
  public final void setUpGUI() {
    if(amInited) return;
    amInited = true;
    setLayout(new BorderLayout());
    add(txaConsole, BorderLayout.CENTER);
    btnClear.addActionListener(this);
    add(btnClear, BorderLayout.EAST);
    Panel pnlTemp = new Panel(new GridLayout(0,1));
    pnlTemp.add(cbxDebug); pnlTemp.add(cbxError);
    pnlTemp.add(cbxInfo); pnlTemp.add(cbxStackTrace);
    pnlTemp.add(cbxClip);
    pnlTemp.add(btnCalc); btnCalc.addActionListener(this);
    add(pnlTemp, BorderLayout.WEST);
    CapCom.allowConsoleAuthenticatorAccess = ("true".equalsIgnoreCase(getParameter("control.MAuthenticator")));
    CapCom.allowAccessories = ("true".equalsIgnoreCase(getParameter("control.Accessory")));
    calcEnabled = ("true".equalsIgnoreCase(getParameter("control.Calculator")));
  }

  /** Overridden to call <code>MathDBC.closeAll()</code> and <code>this.closeAll()</code>. If overriding this method they must be called manually. */
  public void destroy() {
    MathDBC.closeAll();
    this.closeAll();
  }

  /** ActionListener method. All it does is clear the screen or show the calculator (if enabled). */
  public void actionPerformed(ActionEvent aev) {
    if(aev.getSource() == btnClear) {
      clearScreen();
    } else if(aev.getSource() == btnCalc) {
      showCalc();
    } else {}
  }

  /** Clear the TextArea used for messages. */
  public void clearScreen() { txaConsole.setText(""); }

  /** Display a <code>QuickRPNCalculator</code> if it is enabled. If not, nothing happens. */
  public final void showCalc() {
    if(!calcEnabled) return;
    if(dlgCalc == null) {
      dlgCalc = new Dialog(null, false);
      dlgCalc.setTitle("Calculator");
      dlgCalc.setLayout(new BorderLayout());
      dlgCalc.add(new QuickRPNCalculator(), BorderLayout.CENTER);
      dlgCalc.pack();
      dlgCalc.addWindowListener(new WindowAdapter() {
        public void windowClosing(WindowEvent wev) {
          dlgCalc.setVisible(false);
          dlgCalc.dispose();
        }
      });
    }
    dlgCalc.setVisible(true);
  }
  /** Enable/disable the calculator. If the calculator is visible when it is disabled it is hidden. Returns a self-reference to allow the calculator to be
      enabled and made visible (if desired) with one statement.
      @param isEnabled true to enable the calculator, false to disable and possibly hide it.
  */
  public final MTApplet setCalculatorEnabled(boolean isEnabled) {
    calcEnabled = isEnabled;
    if((!calcEnabled) && (dlgCalc != null)) {
      dlgCalc.setVisible(false);
      dlgCalc.dispose();
    }
    return this;
  }

  /** This method may be called directly or via a MathConsole from createConsole()!  */
  public final void mathLog(byte pLevel, String pMesg, Throwable pThr) {
    String prefix = null;
    switch(pLevel) {
      case MathPrinter.LOG_INFO: prefix = (cbxInfo.getState()?"[I] ":null); break;
      case MathPrinter.LOG_DEBUG: prefix = (cbxDebug.getState()?"[D] ":null); break;
      case MathPrinter.LOG_ERROR: prefix = (cbxError.getState()?"[E] ":null); break;
      default:
    }
    if(prefix == null) return;
    mathPrintln(prefix+pMesg);
    if(cbxStackTrace.getState() && (pThr != null)) mathPrintln(CapCom.getStackTrace(pThr));
  }
  /** Send a String to the TextArea 'screen.' Can be called directly or through a MathConsole. */
  public final void mathPrintln(String pMesg) { txaConsole.append(pMesg+"\n"); }

  /** Overridden for clipboard access, security restrictions permitting. If the '[]CB' Checkbox is enabled and pObj is not null its String content is wrapped in 
      a <code>Transferable</code> and sent to the clipboard.
  */
  public final Object processMathObject(Object pObj) {

    if(cbxClip.getState() && (pObj != null)) {
      try {
        Clipboard theClip = Toolkit.getDefaultToolkit().getSystemClipboard();
        if(theClip == null) {
          mathLog(MathPrinter.LOG_DEBUG, "Failed to acquire clipboard.", null);
          return pObj;
        }
        theClip.setContents(new StringSelection(pObj.toString()), this);
      } catch(Exception ex) { mathLog(MathPrinter.LOG_DEBUG, "Clipboard error.", ex); }
    }
    return pObj;
  }

  /** Required for the <code>ClipboardOwner</code> interface. Default behavior prints a Debug-level message when ownership is lost. */
   public void lostOwnership(Clipboard pClip, Transferable pData) {
     mathLog(MathPrinter.LOG_DEBUG, "Clipboard ownership lost.", null);
   }

  /** Call this method to set THIS instance of MTApplet as the Global Printer. */
  public final void setAsPrinter() {
    CapCom.setPrinter(this);
    CapCom.println("Testing from MTApplet.");
  }

  /** This method prints the Class name of the Object passed in, or 'null' if it is null. */
  public void printClassInfo(Object pObj) {
    if(pObj == null) { CapCom.println("Class: null"); return; }
    Class pClass = pObj.getClass();
    CapCom.println("Class: "+pClass.toString());
  }

  /** <p>No scripting tool allows Global Pizza access past getting parameters from it. Applications can set the Global Pizza before script execution, so methods 
      like MathEngine's <code>blines()</code> can function. This method provides MTApplet with the capability to set the Global Pizza, presumably before execution 
      of a script requiring such access.</p>
      <p>This method splits the supplied String first on pLineSep and then each 'line' on pKeySep. The result is then added to a Properties object which, when the 
      lines run out, is set as the Global Pizza. The default values for pLineSep and pKeySep are a newline and an equals sign, respectively. The second split has 
      an upper limit of two, so the key separator can be used afterward and will not lose data.</p>
      @param pPizza The String containing the Properties-formatted data for the Global Pizza.
      @param pLineSep The separator to split the data into 'lines.' Defaults to a single newline.
      @param pKeySep The String used to separate the key from the value.
  */
  public void setGlobalPizza(String pPizza, String pLineSep, String pKeySep) {
    if(!Question.validString(pPizza)) return;
    if(!Question.validString(pLineSep)) pLineSep = "\n";
    if(!Question.validString(pKeySep)) pKeySep = "=";
    java.util.Properties thePizza = new java.util.Properties();
    String[] linez = pPizza.split(pLineSep), tokenz;
    for(int i=0; i < linez.length; i++) {
      tokenz = linez[i].split(pKeySep, 2);
      if((tokenz.length == 2) && Question.validString(tokenz[0]) && Question.validString(tokenz[1])) thePizza.put(tokenz[0], tokenz[1]);
    }
    CapCom.setPizza(thePizza);
  }
  /** Convenience method for setGlobalPizza(pPizza, null, null). */
  public void setGlobalPizza(String pPizza) { setGlobalPizza(pPizza, null, null); }

  /** This method allows access to <code>Class.getResourceAsStream()</code>, which provides <code>MTApplet</code> with the closest thing to filesystem access it 
      can get. The access is only allowed within the bounds of the <code>Applet</code> security constraints so all of its restrictions are in effect. The Math 
      ClassLoader is <b>not</b> accessed so only the resources available to a normal Applet can be fetched. Use with care! Any errors are reported at Error level 
      and the return value for such is null.
      @param pName The name of the resource to be fetched.
  */
  public final InputStream getResourceStream(String pName) {
    if(!Question.validString(pName)) return null;
    InputStream retval = null;
    try {
      retval = this.getClass().getResourceAsStream(pName);
      if(retval != null) openStreams.add(retval);
    } catch(Exception ex) { CapCom.error("Error getting resource stream", ex); }
    return retval;
  }

  /** This method works per <code>getResourceStream()</code> (and does use that method) but wraps the resulting stream in a <code>BufferedReader</code>. Restrictions 
      are per <code>getResourceStream()</code> and if a stream cannot be created the return value is null. Any errors are reported at Error level.
      @param pName The name of the resource to be read.
  */
  public final BufferedReader getResourceReader(String pName) {
    if(!Question.validString(pName)) return null;
    BufferedReader retval = null;
    InputStream isTemp = getResourceStream(pName);
    if(isTemp == null) return null;
    try {
      retval = new BufferedReader(new InputStreamReader(isTemp));
    } catch(Exception ex) { CapCom.error("Error creating BufferedReader", ex); }
    return retval;
  }

  /** <code>MTApplet</code> maintains an internal list of resource InputStreams and Readers. This method closes them and reports any errors from attempting to do 
      so. The <code>destroy()</code> method calls this one automatically but subclasses overriding it might not. If there is any doubt, call it from the script! 
      Multiple calls will have no ill effects.
  */
  public final void closeAll() {
    if(openStreams.isEmpty()) return;
    for(InputStream is : openStreams) {
      try { is.close(); } catch(Exception ex) { CapCom.error("Error closing stream", ex); }
    }
  }

  /** Check an authentication {@link cox.jmatt.java.MathTools.util.MAuthenticator code} at Student level. */
  public final boolean authenticateStudent(String pCode) { return CapCom.authenticate(MAuthenticator.AUTH_STUDENT, pCode); }
  /** Check an authentication {@link cox.jmatt.java.MathTools.util.MAuthenticator code} at Proctor level. */
  public final boolean authenticateProctor(String pCode) { return CapCom.authenticate(MAuthenticator.AUTH_PROCTOR, pCode); }
  /** Check an authentication {@link cox.jmatt.java.MathTools.util.MAuthenticator code} at Instructor level. */
  public final boolean authenticateInstructor(String pCode) { return CapCom.authenticate(MAuthenticator.AUTH_INSTRUCTOR, pCode); }

  /** Create and return a blank Question. */
  public Question createQuestion() { return new Question(); }
  /** Create and return a Question with all data provided. */
  public Question createQuestion(String pID, String pProb, String pAns) { return new Question(pID, pProb, pAns); }
  /** Create and return an instance of MathEngine. */
  public MathEngine createMEngine() { return new MathEngine(); }
  /** Create and return a MathConsole. This method uses the zero-argument constructor so the Console returned reflects all method calls through CapCom. */
  public MathConsole createConsole() { return new MathConsole(); }
  /** Create and return an Arithmetic instance. */
  public Arithmetic createArithmetic() { return new Arithmetic(); }
  /** Create and return a MathXML object. */
  public MathXML createMathXML() { return new MathXML(); }
  /** Create a default (zero) Polynomial. */
  public Polynomial createPolynomial() { return new Polynomial(); }
  /** Create a Polynomial from a format String.
      @param pPoly A String conforming to the Polynomial.createPolynomial() method. Which this one calls!
  */
  public Polynomial createPolynomial(String pPoly) { return Polynomial.createPolynomial(pPoly); }
  /** Create an MFraction from a formatting String. */
  public MFraction createMFraction(String pFormat) { return MFraction.createMFraction(pFormat); }
  /** Create and return a new MMatrixBuilder. */
  public cox.jmatt.java.MathTools.util.MMatrixBuilder createMMBuilder() { return new cox.jmatt.java.MathTools.util.MMatrixBuilder(); }
  /** Create and return a Percent instance. */
  public Percent createPercent() { return new Percent(); }
  /** Create and return a MEquation. */
  public MEquation createMEquation() { return new MEquation(); }
  /** Create and return a MRadical. */
  public MRadical createMRadical() { return new MRadical(); }
  /** Create and return a MathStat. */
  public MathStat createMStat() { return new MathStat(); }
  /** Create and return a MTXMLProcessor. IMPORTANT NOTE: If used from a web browser, Applet security constraints may restrict this class! */
  public MTXProcessor createMTXPro() { return new MTXProcessor(); }
  /** Return a new GraphEngineFactory. */
  public cox.jmatt.java.MathTools.graphing.GraphEngineFactory createGraphEngineFactory() { return cox.jmatt.java.MathTools.graphing.GraphEngineFactory.newInstance(); }
  /** Return a MathDrawSVG object. */
  public cox.jmatt.java.MathTools.markup.svg.MathDrawSVG createMDraw() { return new cox.jmatt.java.MathTools.markup.svg.MathDrawSVG(); }
  /** Return a MathGenHTML object. */
  public cox.jmatt.java.MathTools.markup.html.MathGenHTML createMathHTM() { return new cox.jmatt.java.MathTools.markup.html.MathGenHTML(); }
  /** Return a SimpleTagImpl. */
  public cox.jmatt.java.MathTools.markup.SimpleTagImpl createSimpleTag() { return new cox.jmatt.java.MathTools.markup.SimpleTagImpl(); }
  /** Create a MathDBC instance. */
  public MathDBC createMathDBC() { return new MathDBC(); }
  /** Create an empty MTest. */
  public MTest createMTest() { return new MTest(); }
  /** Create a new MathTestFormatter. */
  public MathTestFormatter createMTFormatter() { return new MathTestFormatter(); }
  /** Create a COFormatterFactory */
  public COFormatterFactory createCOFormatter() { return new COFormatterFactory(); }
  /** Create a <code>GraphPanelFactory</code>. */
  public cox.jmatt.java.MathTools.demo.gpgdemo.GraphPanelFactory createNotebook() { return new cox.jmatt.java.MathTools.demo.gpgdemo.GraphPanelFactory(); }
  /** Create a <code>MathGenLatex</code>. */
  public cox.jmatt.java.MathTools.markup.latex.MathGenLatex createMLatex() { return new cox.jmatt.java.MathTools.markup.latex.MathGenLatex(); }
  /** Create a <code>MComplex</code> instance. */
  public MComplex createMComplex() { return new MComplex(); }
  /** Create a real-number <code>MathSet</code>. */
  public MathSet createMathSet() { return MathSet.realNumbers(); }

}

