package cox.jmatt.java.MathTools.util;

import java.io.*;
import java.util.*;
import java.net.URI;
import java.net.URL;
import java.net.URLClassLoader;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.xpath.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.w3c.dom.*;

import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.test.MTFI;
import cox.jmatt.java.MathTools.graphing.GraphHandler;
import cox.jmatt.java.MathTools.graphing.GraphEngine;
import cox.jmatt.java.MathTools.graphing.GraphPaper;

/** <p>The MathTools XML Processor processes the XML tags generated by the MathTools classes with a <code>toXML()</code> method. In order to function, this class 
    MUST have appropriate XML equipment available. Specifically, MTXPro requires a SAXParserFactory capable of creating namespace-aware SAXParsers. Fortunately 
    this is easy to check with the <code>isAvailable()</code> method. If it returns true, sufficient machinery is in place and MTXPro will function properly. If 
    not, simply place a SAXParserFactory provider in the CLASSPATH or set the appropriate system properties; MTXPro uses the discovery mechanism to create a 
    SAXParserFactory which it then uses to create parsers.</p>

    <h3>MTXProcessor is...</h3>
    <p>MTXPro is written to process the XML tags generated by the MathTools classes that have a <code>toXML()</code> of some description. It is NOT designed as an 
    all-purpose XML or even MathML processor! Some MathTools classes, such as Question, natively generate XML. This XML markup is MathTools-specific and this class 
    parses it as such.</p>
    <p>MTXProcessor itself extends <code>MTXErrorHandler</code> and overrides methods germane to processing MathTools XML. It can be used as a 
    DefaultHandler without modification. Simply create an instance and feed it to any SAXParser's <code>parse()</code> method as the DefaultHandler. When parsing 
    is complete, call the <code>getReturnList()</code> method for a List of whatever was processed.</p>
    <p>In addition to the DefaultHandler methods, MTXPro contains all the machinery necessary to actually parse MathTools XML content. The source can be a String, 
    a file, or an <code>org.xml.sax.InputSource</code>. For file or String input, MTXPro wraps an InputSource around either a FileReader or StringReader and calls 
    <code>processInput()</code>. It then creates a SAXParser, sets itself as the DefaultHandler and parses the document.</p>
    <p>The XML markup recognized by MTXPro is, as stated previously, that generated by the MathTools classes. The XML itself must be well-formed but MTXPro is not 
    picky past that. There <b>must</b> be a root tag but exactly <i>what</i> that tag is does not matter. An &lt;MTest&gt; tag will work, but so will any other 
    properly-opened and -closed tag. The default tag, provided as a static String constant, is &lt;MathTools&gt;. If present, the 'id=' attribute on any parsed 
    tag is sent to the <code>process__()</code> methods, as is any 'data=' tag. Since the Question tag is the only one with children, it is handled differently. 
    Other tags are handled from the <code>startElement()</code> method. As soon as MTXPro has the necessary attributes it calls the appropriate <code>process__()</code> 
    method. For Question tags, processing waits until the closing tag to present the id, Problem, and Answer components. MFormatElement tags, which may have 
    content but no children, are also handled via <code>endElement()</code>.</p>
    <p>Default behavior for MTXPro parsing is to create an instance of the MathTools core class specified by each tag. If creation succeeds and the resulting 
    instance is not null, it is added to the internal return List. This list is returned from the various <code>parse__()</code> methods or can be retrieved via 
    <code>getReturnList()</code> if MTXPro is used as a handler.</p>
    <p>Per <code>MTXErrorHandler</code> all SAX or XSLT warns and errors are reported at Debug level and fatals are reported at Error level. No error processing 
    is done as this is beyond the scope of this class.</p>

    <h3>Loading Factories</h3>
    <p><code>MTXPro</code> can use CapCom's Math ClassLoader to create <code>SAXParserFactory, TransformerFactory, DocumentBuilderFactory,</code> and 
    <code>XPathFactory</code> instances. To do so two things must happen. First and foremost <code>setMathClassLoader(true)</code> sets the <code>CapCom ClassLoader</code> 
    in place. Passing <code>false</code> clears it and both happen immediately.</p>
    <p>The second step is to set the class name of the desired factory type. This is critical: the name of the class <b>must</b> be set or no custom factory will 
    be created! If the Math ClassLoader is in place and the desired factory name is not null or blank, the next call to the appropriate <code>isAvailable()</code> 
    method <i>will</i> try to instantiate the specified class using the Math ClassLoader. This is done on a per-type basis so it is quite possible to use a custom 
    XPath factory while still using the standard ones for the rest. Simply enable the MathClassLoader and set the name of the <code>XPathFactory</code> while <i>not</i> 
    setting or clearing the others.</p>
    <h4>Configuring SAX</h4>
    <p>SAX configuration options are minimal but present. <code>MTXProcessor</code> maintains global flags for SAX validation and namespace awareness. Once set 
    these values take effect on the next call to the (basic) 'isAvailable()' method. They apply to the <code>SAXParserFactory</code> and remain at the values set 
    until they are manually changed.</p>
    <p>To set something other than validation or namespaces or to have the change take effect immediately use the 'setSAXFeature()' method. This must happen 
    <b>after</b> at least one call to 'isAvailable()' otherwise no parser will be set. If a parser is set then the feature value is set immediately. If it cannot 
    be set or if attempting to set it throws an Exception this is reported at Debug level and the method returns false, otherwise it returns true.</p>

    <h3>MTXProcessor and XSLT</h3>
    <p>MTXPro has a basic XSLT transformation capacity. It is present more for the sake of completeness than mathematical nececssity, though it can be useful in 
    preparing mathematical XML documents. The transformer itself depends on two basic methods, both named <code>transformData()</code>. The actual transformation 
    method accepts a <code>javax.xml.transform.Source</code> for the XML data and stylesheet and a <code>Result</code> for the output. If the stylesheet is null 
    the identity transform is applied. Values for any &lt;xsl:param &gt; tags can be set with the <code>setParameter()</code> method. It accepts a String key (that 
    must match the 'name=' attribute in the stylesheet tag) and a String value. When the transformation is done, all parameters set will be placed into the 
    Transformer.</p>
    <p>Scripts use the <code>transformData()</code> that accepts two Strings and two booleans. The first String is the XML data source, the second is the stylesheet. 
    The booleans control whether to interpret the Strings as raw data (true) or as a filename (false). The result of the transform is returned as a String.</p>
    <p>The <code>isTransformAvailable()</code> checks to see if XSLT equipment is in place, so it should be checked at least once per configuration or location 
    change. If it returns false the MathClassLoader can be enabled and used.</p>

    <h3>DOM Processing</h3>
    <p>MTXProcessor's DOM processing capability is, like XLST, added for the sake of completeness. It allows scriptable XML generation beyond what <code>MathXML</code>
    provides. Specifically, it allows the creation of a blank DOM Document from which Elements and other things can be added. DOM Documents can also be parsed from 
    Strings or files, manipulated, transformed, and printed out.</p>
    <p>As with the SAX and XSLT methods, the <code>isDOMAvailable()</code> method checks the DOM equipment and prepares it; and the <code>resetDOM()</code> 
    method clears it.</p>
    <p>If necessary, DOM Attributes and Features can be set via the <code>setDOMFeature()</code> and <code>setDOMAttribute()</code> methods. These methods are 
    called on the DocumentBuilderFactory before a DocumentBuilder is created so they will be in effect for all subsequent DOM activity. Calling <code>resetDOM()</code> 
    clears all Attributes and Features, plus the current DocumentBuilder and DocumentBuilderFactory.</p>
    <h4>DOM and XPath</h4>
    <p>The DOM equipment includes basic XPath capability. An XPath can be applied to a DOM Document or Node with the result returned as a Node. The XPath equipment 
    includes a factory class name and settable XPath features. As with DOM and transform equipment setting the XPath factory classname to a non-null, non-blank 
    value uses whatever <code>ClassLoader</code> is set, if one is. The <code>isXPathAvailable()</code> method creates the necessary infrastructure and must be 
    called before XPath can be used.</p>
    <p>The <code>applyXPath()</code> method does the work. It accepts a String XPath and a DOM Node and returns a DOM Node. The method itself attempts to apply 
    the XPath to the data Node. If an error occurs it is reported and the return value is null. The <code>XPath</code> instance itself is never directly exposed 
    so any operations beyond the bare basics are beyond the scope of this class.</p>

    <h3>Oh yeah. Graphing!</h3>
    <p>The <code>graphing.xml.*</code> package produces simplified XML markup from graphing commands. The <code>parseGraph()</code> method here can convert that 
    markup back into graphing classes. See the {@link cox.jmatt.java.MathTools.graphing.GraphHandler graphing} package for exact details.</p>
*/
public class MTXProcessor extends MTXErrorHandler {

  /** String constant for the MathTools namespace URI: 'cox.jmatt.java/MathTools'. */
  public static final String NSURI = "cox.jmatt.java/MathTools";
  /** Default opening tag for a MathTools XML document. &lt;MathTools&gt; with the MathTools namespace. */
  public static final String MT_OPEN = "<MathTools xmlns=\""+NSURI+"\" >";
  /** Default closing tag for a MathTools XML document. */
  public static final String MT_CLOSE = "</MathTools>";

  private SAXParserFactory myFactory = null;
  private SAXParser myParser = null;
  private TransformerFactory myTFactory = null;
  private Transformer myTransformer = null;
  private DocumentBuilderFactory myDOMFactory = null;
  private DocumentBuilder myDOMBuilder = null;
  private XPathFactory myXFactory = null;
  private StringBuffer sbContent = new StringBuffer();
  private ArrayList<Object> myList = new ArrayList<Object>();
  private String curQID = null, curProblem = null, curAnswer = null;
  private ClassLoader myLoader = null;
  private String myPFactoryClassName = "", myTFactoryClassName = "", myDOMFactoryClassName = "", myXFactoryClassName = "";
  private Hashtable<String, String> transPizza = new Hashtable<String, String>(), transProps = new Hashtable<String, String>();
  private Hashtable<String, Boolean> domFeatures = new Hashtable<String, Boolean>();
  private Hashtable<String, Object> domAttributes = new Hashtable<String, Object>();
  private GraphHandler myGraphHandler = new GraphHandler();
  private boolean graphMode = false, checkMode = false, saxValidating = false, saxNamespace = true;

  /** Standard constructor. Used for scripting and for instantiating a default SAXParserFactory using the built-in (to Java!) lookup procedure. */
  public MTXProcessor() { super(); }

  /** This method checks to see if a SAXParser can be created. <b>ALWAYS check this method once!</b> If a namespace-aware SAXParser cannot be created for whatever 
      reason this class will not function! Use this method to check before a lot of extra work is done. If it returns true, all is well. If not, there either is 
      no SAXParserFactory or no available SAXParser. Any Errors or Exceptions thrown during factory or parser creation are reported at debug level.
  */
  public boolean isAvailable() {
    try {
      if((myLoader == null) || !Question.validString(myPFactoryClassName)) {
        myFactory = SAXParserFactory.newInstance();
      } else {
        myFactory = SAXParserFactory.newInstance(myPFactoryClassName, myLoader);
      }
      myFactory.setNamespaceAware(saxNamespace);
      myFactory.setValidating(saxValidating);
      myParser = myFactory.newSAXParser();
    } 
    catch(FactoryConfigurationError ex) {
      CapCom.debug("SAXParserFactory cannot be created/instantiated", ex);
      return false;
    }
    catch(ParserConfigurationException ex) {
      CapCom.debug("Error creating SAXParser", ex);
      return false;
    }
    catch(Exception ex) {
      CapCom.debug("Error creating XML equipment", ex);
      return false;
    }
    return (myParser != null);
  }

  /** Set the validation flag for <code>SAXParser</code> instances created by the internal factory. Changing this will not take effect until AFTER a call to 
      <code>isAvailable()</code>. The default value is false.
  */
  public void setValidating(boolean pEnable) { saxValidating = pEnable; }

  /** Enable/disable SAX Namespace awareness. Changing the value takes effect after the next call to <code>isAvailable()</code>. By default namespace awareness 
      is enabled.
  */
  public void setNamespaceAware(boolean pEnable) { saxNamespace = pEnable; }

  /** Enable/disable a SAX Feature on the <b>current</b> parser. The feature is set (or not) immediately and remains in effect until the next call to 
      <code>isAvailable()</code>. If, for whatever reason, a feature cannot be set (Exception thrown) this is reported at Debug level and the return value is 
      false. If the feature is set the return value is true. No string validation is done; use with care! If no parser is set the method returns false immediately.
      @param pFeature The name of the SAX feature to set.
      @param pEnable The value to set for the specified feature.
  */
  public boolean setSAXFeature(String pFeature, boolean pEnable) {
    if(myParser == null) return false;
    boolean retval = true;
    try {
      myParser.getXMLReader().setFeature(pFeature, pEnable);
    } catch(Exception ex) {
      CapCom.debug("Can't set feature: "+pFeature, ex);
      retval = false;
    }
    return retval;
  }

  /** This method specifies the name of the factory class used to instantiate a SAXParserFactory. Setting it to null or blank clears it. */
  public void setPFactoryClassName(String pName) { myPFactoryClassName = Question.validString(pName)?pName:""; }

  /** This method specifies the name of the factory class used to instantiate a TransformerFactory. Setting it to null or blank clears it. */
  public void setTFactoryClassName(String pName) { myTFactoryClassName = Question.validString(pName)?pName:""; }

  /** This method specifies the name of the factory class for <code>XPATHFactory</code> instances. Setting it to blank or null clears it. */
  public void setXFactoryClassName(String pName) { myXFactoryClassName = Question.validString(pName)?pName:""; }

  /** Enable/disable <code>CapCom</code>'s MathClassLoader. The loader is set or cleared instantly but will not be used until the next call to one of the 
      <code>isAvailable()</code> methods. NOTE: setting this to false <b>does not</b> clear any previously-set factory class names.
      @param pEnable true to set the MathClassLoader, false to disable it.
  */
  public void setMathClassLoader(boolean pEnable) {
    myLoader = pEnable?CapCom.getMathClassLoader():null;
  }

  /** This method prints a debug-level message and initializes the internal return List. */
  public void startDocument() {
    CapCom.debug("Starting document processing");
    myList = new ArrayList<Object>();
  }

  /** This method prints a debug-level message and nothing else. */
  public void endDocument() { CapCom.debug("Ending document processing"); }

  /** This method is used only by Problem, Answer, and some MFormatElement tags. It accumulates content until the tag closes. */
  public void characters(char[] pData, int pStart, int pLen) { sbContent.append(pData, pStart, pLen); }

  /** Process the element properly. Questions accumulate the Problem and Answer tags, MFormatElements accumulate content, all others are empty. */
  public void startElement(String uri, String pName, String qName, Attributes pAtts) {
    String sID = pAtts.getValue("", "id");
    String sData = pAtts.getValue("", "data");
    CapCom.debug("Got tag: "+pName+" with ID="+sID+" and data="+sData);
    Object oTemp = null;
    if("Question".equals(pName)) {
      curQID = sID;
    } else if("Problem".equals(pName)) {
      sbContent = new StringBuffer();
    } else if("Answer".equals(pName)) {
      sbContent = new StringBuffer();
    } else if("MMatrix".equals(pName)) {
      oTemp = processMMatrix(sID, sData);
      if(oTemp != null) myList.add(oTemp);
    } else if("Polynomial".equals(pName)) {
      oTemp = processPolynomial(sID, sData);
      if(oTemp != null) myList.add(oTemp);
    } else if("MFraction".equals(pName)) {
      oTemp = processMFraction(sID,  pAtts.getValue("", "numerator"),  pAtts.getValue("", "denominator"));
      if(oTemp != null) myList.add(oTemp);
    } else if("MRadical".equals(pName)) {
      try {
        oTemp = processMRadical(sID,  Integer.parseInt(pAtts.getValue("", "radicand")), Integer.parseInt(pAtts.getValue("", "index")));
        if(oTemp != null) myList.add(oTemp);
      } catch(Exception ex) { CapCom.debug("Error parsing MRadical", ex); }
    } else if("Mtfi".equals(pName)) {
      curQID = pAtts.getValue("", "type");
      sbContent = new StringBuffer();
      if(Question.validString(sID)) sbContent.append(sID);
    } else if("MEquation".equals(pName)) {
      oTemp = processMEquation(sData);
      if(oTemp != null) myList.add(oTemp);
    } else if("MathStat".equals(pName)) {
      oTemp = processMathStat(sData, pAtts.getValue("", "data2"));
      if(oTemp != null) myList.add(oTemp);
    } else if("MComplex".equals(pName)) {
      oTemp = processMComplex(sData);
      if(oTemp != null) myList.add(oTemp);
    }
  }

  /** Handle processing of Question and its children and MFormatElement. All other types are instantiated in <code>startElement()</code>. */
  public void endElement(String uri, String pName, String qName) {
    Object oTemp = null;
    if("Question".equals(pName)) {
      oTemp = processQuestion(curQID, curProblem, curAnswer);
      if(oTemp != null) myList.add(oTemp);
    } else if("Problem".equals(pName)) {
      curProblem = sbContent.toString();
    } else if("Answer".equals(pName)) {
      curAnswer = sbContent.toString();
    } else if("Mtfi".equals(pName)) {
      oTemp = processMTFI(curQID, sbContent.toString());
      if(oTemp != null) myList.add(oTemp);
    }
  }

  /** Parse an XML file. This method constructs an InputSource around pString and calls <code>parseInput()</code> on it. Return values and Exceptions are per 
      that method. If called from MTApplet this method may fail due to security considerations. Use with care!
      @param pFile The name of the file to parse. If null or empty the method returns null.
  */
  public List parseFile(String pFile) throws IOException, SAXException {
    if(!Question.validString(pFile)) return null;
    FileInputStream fsIn = new FileInputStream(pFile);
    List retval =  parseInput(new InputSource(fsIn));
    fsIn.close();
    return retval;
  }

  /** Parse MathTools XML contained in a String. Exceptions and return value per <code>parseInput()</code>
      @param pData A String containing MathTools XML markup.
  */
  public List parseString(String pData) throws IOException, SAXException {
    if(!Question.validString(pData)) return null;
    StringReader srIn = new StringReader(pData);
    List retval = parseInput(new InputSource(srIn));
    srIn.close();
    return retval;
  }

  /** Process XML from an InputSource. All other parsing methods reflect to this one. If pSource is null the return value is null. If <code>isAvailable()</code> 
      returns false, the return value is null.
      @param pSource InputSource containing MathTools XML markup.
      @return A List of non-null Objects resulting from the various <code>process__()</code> methods.
      @throws IOException If the underlying IO operations do.
      @throws SAXException If the parser does.
  */
  public List parseInput(InputSource pSource) throws IOException, SAXException {
    if(pSource == null) return null;
    if(!isAvailable()) return null;
    myList = new ArrayList<Object>();
    if(graphMode) {
      graphMode = false;
      myParser.parse(pSource, myGraphHandler);
    } else if(checkMode) {
      checkMode = false;
      myParser.parse(pSource, new MTXErrorHandler());
    } else {
      myParser.parse(pSource, this);
    }
    return myList;
  }

  /** Use this method to check a document ONLY. It parses XML without creating any MathTools or Graphing components, all it does is check for errors. The pData 
      parameter can be raw XML or a filename. Any unhandled exceptions are reported at Error level.
      @param pData The data or file to check.
      @param pRawData true if pData is raw XML, false if it is a filename.
  */
  public void checkXML(String pData, boolean pRawData) {
    checkMode = true;
    try {
      if(pRawData) {
        parseString(pData);
      } else {
        parseFile(pData);
      }
    } catch(Exception ex) { CapCom.error("Error checking XML document", ex); }
    checkMode = false;
  }

  /** This method returns the internal List used to hold the components created by the various <code>process__()</code> methods. It may be empty but will always 
      be defined and will not contain any null entries. With that said, the ONLY time it is specifically valid is <b>after</b> an XML document has been parsed 
      with MTXPro or a subclass as the DefaultHandler! This method is provided to allow use of MTXPro as a handler but <b>without</b> using its built-in <code>
      parse__()</code> methods.
  */
  public List getReturnList() { return myList; }

  /** The parser has completed a Question tag. Override this method to do something with it. The default implementation creates a Question object around the data 
      sent in and returns it. By default, <code>endElement()</code> calls this method and, if the return is not null, adds it to the list it returns. If the return 
      is null, it is NOT added to the list.
      @param pID The ID from the Question tag. May be null.
      @param pProblem The content of the Problem tag. May be null.
      @param pAnswer The content of the Answer tag. May be null.
      @return An Object to add to the return list, or null to add nothing.
  */
  public Object processQuestion(String pID, String pProblem, String pAnswer) {
    Question retval = new Question(pProblem, pAnswer);
    if(Question.validString(pID)) retval.setID(pID);
    return retval;
  }

  /** The processor has detected a MMatrix tag. Override this method to do something with it. The default implementation creates and returns a MMatrix based on 
      the data sent in. The ID is ignored, and if the data is null so is the return value.
      @param pID The value of the tag's 'id=' attribute, if present.
      @param pData The value of the tag's 'data=' attribute. If null, this method returns null.
  */
  public Object processMMatrix(String pID, String pData) {
    if(!Question.validString(pData)) return null;
    return MMatrixBuilder.createMMatrix(pData, ";", ",");
  }

  /** The processor has detected a Polynomial tag. Override this method to do something with it. The default implementation creates and returns a Polynomial 
      based on the data sent in. The ID is ignored, and if the data is null so is the return value.
      @param pID The value of the tag's 'id=' attribute, if present.
      @param pData The value of the tag's 'data=' attribute. If null, this method returns null.
  */
  public Object processPolynomial(String pID, String pData) {
    if(!Question.validString(pData)) return null;
    return new Polynomial(pData);
  }

  /** The processor found a MFraction tag. The numerator and denominator attributes are passed as String; the default implementation concatenates them around a 
      slash character and creates a MFraction to return. If either is null or empty the return is null.
      @param pID The value of the tag's id attribute, if present.
      @param pNumerator The String value of the tag's 'numerator=' attribute.
      @param pDenominator The String value of the tag's 'denominator=' attribute.
  */
  public Object processMFraction(String pID, String pNumerator, String pDenominator) {
    if(!Question.validString(pNumerator) || !Question.validString(pDenominator)) return null;
    return new MFraction(pNumerator+"/"+pDenominator);
  }

  /** The processor has a MathStat tag. This method creates a new MathStat instance from it.
      @param pData The primary data set as a space-separated String.
      @param pData2 The second data set, space-separated.
  */
  public Object processMathStat(String pData, String pData2) { return new MathStat(MathStat.sparse(pData), MathStat.sparse(pData2)); }

  /** The processor has data for a MEquation. The raw data string is sent and converted here. If the data String is not properly constructed, null is returned. */
  public Object processMEquation(String pData) {
    if(!Question.validString(pData)) return null;
    String [] partz = pData.split(":");
    if(partz.length != 5) return null;
    return MEquation.createMEquation(partz[0], partz[1], partz[2], partz[3], partz[4]);
  }

  /** Process a MRadical tag. The radicand and index are converted when the tag is parsed; if either is not a valid int, this method is not called. */
  public Object processMRadical(String pID, int pRadicand, int pIndex) { return new MRadical(pRadicand, pIndex); }

  /** Process a MComplex tag. The return value will never be null. */
  public Object processMComplex(String pCplx) { return MComplex.reParse(pCplx); }

  /** Process a MFormatElement. This is the ONE valid use of the class constructor! Proper types and content are not enforced; USE WITH CAUTION!
      @param pType The MFormatElement type. If null or empty the return value is null.
      @param pContent The element content.
  */
  public Object processMTFI(String pType, String pContent) {
    if(!Question.validString(pType)) return null;
    return MTFI.newInstance(pType, pContent);
  }

  /** Check to see if an XSLT Transformer is avaiable. <b>Always check this method once!</b> If it returns true, XSLT machinery is in place. If not, download the 
      JAR, add the file's URL, set the TransformerFactory class, and try again.
  */
  public boolean isTransformAvailable() {
    try {
      if((myLoader == null) || !Question.validString(myTFactoryClassName)) {
        myTFactory = TransformerFactory.newInstance();
      } else {
        myTFactory = TransformerFactory.newInstance(myTFactoryClassName, myLoader);
      }
      myTFactory.setErrorListener(this);
      myTransformer = myTFactory.newTransformer();
    }
    catch(TransformerFactoryConfigurationError ex) {
      CapCom.debug("Error creating TransformerFactory", ex);
      return false;
    }
    catch(TransformerConfigurationException ex) {
      CapCom.debug("Exception creating Transformer", ex);
      return false;
    }
    catch(Exception ex) {
      CapCom.debug("Error creating XSLT equipment.", ex);
      return false;
    }
    return (myTransformer != null);
  }

  /** This method is the basic XLST Transform method. It transforms pData by means of pStyle into pResult. All other transform methods use this one. If the stylesheet 
      is null the identity transform is used.
      @param pData The XML data to be transformed.
      @param pStyle The XLST Stylesheet to use for the transformation.
      @param pResult The result of transforming pData by pStyle.
      @throws TransformerConfigurationException If something goes wrong instantiating a Transformer.
      @throws TransformerException If there is a problem transforming the data.
  */
  public void transformData(Source pData, Source pStyle, Result pResult) throws TransformerConfigurationException, TransformerException {
    if((pData == null) || (pResult == null)) return;
    if(!isTransformAvailable()) return;
    if(pStyle == null) {
      myTransformer = myTFactory.newTransformer();
      CapCom.debug("Using identity transform.");
    } else {
      myTransformer = myTFactory.newTransformer(pStyle);
    }
    myTransformer.setErrorListener(this);
    String sKey = null;
    for(Enumeration en = transPizza.keys(); en.hasMoreElements(); ) {
      sKey = (String) en.nextElement();
      myTransformer.setParameter(sKey, transPizza.get(sKey));
    }
    for(Enumeration en = transProps.keys(); en.hasMoreElements(); ) {
      sKey = (String) en.nextElement();
      try { myTransformer.setOutputProperty(sKey, transProps.get(sKey)); } catch(Exception ex) { CapCom.debug("Problem with output property: "+sKey, ex); }
    }
    
    CapCom.debug("Transform starting.");
    myTransformer.transform(pData, pResult);
    CapCom.debug("Transform finished.");
  }

  /** This is the scripting transform method. It can use a file or String as the data or the stylesheet and it returns the result as a String.
      @param pData The raw XML data or the name of the file where it lives.
      @param pStyle The XSLT Stylesheet or the name of the file containing it.
      @param pRawData true to interpret pData as raw data, false to interpret as a filename.
      @param pRawStyle true to interpret pStyle as the raw stylesheet, false for a filename.
      @return The result of the transform as a String.
  */
  public String transformData(String pData, String pStyle, boolean pRawData, boolean pRawStyle) {
    if(!Question.validString(pData)) return null;
    Source srcData = null, srcStyle = null;
    InputStream isData = null, isStyle = null;
    ByteArrayOutputStream baOut = null;
    FileInputStream fiData = null, fiStyle = null;
    try {
      if(Question.validString(pStyle)) {
        if(pRawStyle) {
          isStyle = new ByteArrayInputStream(pStyle.getBytes());
        } else {
          isStyle = new FileInputStream(pStyle);
        }
        srcStyle = new StreamSource(isStyle);
      }
      if(pRawData) {
        isData = new ByteArrayInputStream(pData.getBytes());
      } else {
        isData = new FileInputStream(pData);
      }
      srcData = new StreamSource(isData);
      baOut = new ByteArrayOutputStream();
      Result resOut = new StreamResult(baOut);
      transformData(srcData, srcStyle, resOut);
      isData.close();
      if(isStyle != null) isStyle.close();
      baOut.flush();
      baOut.close();
    } catch(Exception ex) { CapCom.error("Exception transforming document.", ex); }
    return (baOut !=null)?baOut.toString():"";
  }

  /** This method transforms based on two filenames. That is, <code>transformData(String, String, false, false)</code>.
      @param pData The XML data file to be used.
      @param pStyle The XSLT stylesheet to transform the data.
      @return A String containing the result of the transformation.
  */
  public String transformFiles(String pData, String pStyle) { return transformData(pData, pStyle, false, false); }

  /** Set a stylesheet parameter used for the transformation. Null or empty keys or values are silently ignored. */
  public void setParameter(String pKey, String pValue) {
    if(Question.validString(pKey) && Question.validString(pValue)) transPizza.put(pKey, pValue);
  }

  /** Add the Global Pizza to the parameters fed to the transform. NOTE: The Global Pizza <b>cannot</b> be un-added! Use with caution. Any Exceptions are reported 
      at Debug level.
  */
  public void setParametersFromGlobalPizza() {
    try {
      Properties gPizza = CapCom.getPizza();
      String sKey = null;
      for(Enumeration en = gPizza.propertyNames(); en.hasMoreElements(); ) {
        sKey = (String) en.nextElement();
        transPizza.put(sKey, (String) gPizza.get(sKey));
      }
    } catch(Exception ex) { CapCom.debug("Problem adding Global Pizza.", ex); }
  }

  /** Remove all previously-set parameters. This does NOT affect default values in the stylesheet, only those from <code>setParameter()</code>. */
  public void clearParameters() { transPizza = new Hashtable<String, String>(); }

  /** <p>Set an Output Property to be used by the transformer. Output properties set with this method are collected and added when the Transformer is created. Any 
      Exceptions thrown are reported at Debug level. Add properties carefully; if the Transformer chokes on multiple entries there will be multiple stack traces 
      printed! Empty or null keys or values are silently ignored.</p>
      <p>Regardless of badd Output Property values, the transform will continue.</p>
  */
  public void setOutputProperty(String pKey, String pValue) {
    if(Question.validString(pKey) && Question.validString(pValue)) transProps.put(pKey, pValue);
  }

  /** Clear all previously-set Transformer Output Properties. */
  public void clearOutputProperties() { transProps = new Hashtable<String, String>(); }

  /** <p>Check to see if XML DOM equipment is available. Details are reported at Debug level. This method should be checked at least once when the MathTools 
      environmentor configuration changes.</p>
      <p>Calling this method creates a DOM DocumentBuilderFactory and a DocumentBuilder. These instances remain in effect until cleared by a call to <code>resetDOM()</code>. 
      Any DOM Features or Attributes must be set <b>before</b> calling this method, otherwise they will be ignored. Care should also be exercised in calling 
      <code>resetDOM()</code> as this will ALSO clear all Features and Attributes. The ideal sequence of events to ensure proper setting is:
      <ol>
        <li>Issue a call to <code>resetDOM()</code>. If the script is just starting this is not necessary.</li>
        <li>Set all Attributes and Features.</li>
        <li>Call <code>isDOMAvailable()</code> to create a DocumentBuilderFactory with all Attributes and Features set.</li>
      </ol>
      If an Attribute or Feature cannot be set that Attribute or Feature <b>only</b> is reported at Debug level. Attempts to set all remaining Features and 
      Attributes will be made.</p>
  */
  public boolean isDOMAvailable() {
    if(myDOMBuilder != null) return true;
    try {
      if((myLoader == null) || !Question.validString(myDOMFactoryClassName)) {
        myDOMFactory = DocumentBuilderFactory.newInstance();
      } else {
        myDOMFactory = DocumentBuilderFactory.newInstance(myDOMFactoryClassName, myLoader);
      }
      String sKey = null;
      for(Enumeration<String> enKey = domFeatures.keys(); enKey.hasMoreElements(); ) {
        sKey = enKey.nextElement();
        try { myDOMFactory.setFeature(sKey, domFeatures.get(sKey)); } catch(Exception ex) { CapCom.debug("Error setting DOM Feature: "+sKey, ex); }
      }
      for(Enumeration<String> enKey = domAttributes.keys(); enKey.hasMoreElements(); ) {
        sKey = enKey.nextElement();
        try { myDOMFactory.setAttribute(sKey, domAttributes.get(sKey)); } catch(Exception ex) { CapCom.debug("Error setting DOM Attribute: "+sKey, ex); }
      }
      myDOMBuilder = myDOMFactory.newDocumentBuilder();
      myDOMBuilder.setErrorHandler(this);
    }
    catch(FactoryConfigurationError fce) {
      CapCom.debug("Error instantiating DOM Factory", fce);
      return false;
    }
    catch(ParserConfigurationException ex) {
      CapCom.debug("Error configuring DOM Builder", ex);
      return false;
    }
    catch(Exception ex) {
      CapCom.debug("Error creating DOM equipment", ex);
      return false;
    }
    return (myDOMBuilder != null);
  }

  /** Set (or clear) the class name used to create a DOM DocumentBuilderFactory. If a ClassLoader is set and this name is not null or empty, the full version of 
      <code>DocumentBuilderFactory.newInstance()</code> is used. ClassLoader rules are per the SAX parser and Transformers; they share the same ClassLoader.
      @param pName The fully-qualified class name of the DOM DocumentBuilderFactory to be used, or null to clear it.
  */
  public void setDOMFactoryClassName(String pName) { myDOMFactoryClassName = pName; }

  /** <p>Clear and reset all DOM-related machinery <b>except</b> the DocumentBuilderFactory class name. If the name is set it must be cleared manually.</p>
      <p>NOTE: Calling this method <b>does</b> clear all previously-set Features and Attributes.</p>
  */
  public void resetDOM() {
    myDOMFactory = null;
    myDOMBuilder = null;
    domAttributes = new Hashtable<String, Object>();
    domFeatures = new Hashtable<String, Boolean>();
  }

  /** Create a new, blank DOM document. This method does call <code>isDOMAvailable()</code> and returns null if it is not. */
  public Document newDocument() {
    if(!isDOMAvailable()) return null;
    return myDOMBuilder.newDocument();
  }

  /** Transform a DOM Document (or any Node) via XSLT. If pStyle is null the identity transform is used.
      @param pSource The DOM Node to transform. If null, an empty String is returned.
      @param pStyle The XSLT Stylesheet to use for the transformation.
      @param pRawStyle true to interpret the stylesheet as raw data, false to read it as a file.
      @return The String containing the transformed data.
  */
  public String transformDOM(Node pSource, String pStyle, boolean pRawStyle) {
    if(pSource == null) return "";
    Source srcData = null, srcStyle = null;
    InputStream isStyle = null;
    ByteArrayOutputStream baOut = null;
    FileInputStream fiData = null, fiStyle = null;
    try {
      if(Question.validString(pStyle)) {
        if(pRawStyle) {
          isStyle = new ByteArrayInputStream(pStyle.getBytes());
        } else {
          isStyle = new FileInputStream(pStyle);
        }
        srcStyle = new StreamSource(isStyle);
      }
      srcData = new javax.xml.transform.dom.DOMSource(pSource);
      baOut = new ByteArrayOutputStream();
      Result resOut = new StreamResult(baOut);
      transformData(srcData, srcStyle, resOut);
      if(isStyle != null) isStyle.close();
      baOut.flush();
      baOut.close();
    } catch(Exception ex) { CapCom.error("Exception transforming document.", ex); }
    return (baOut !=null)?baOut.toString():"";
  }

  /** Transform a DOM Node into another Node via XSLT. Any Exceptions are reported at Error level and the result will be null.
      @param pSource The DOM Node to transform. If null, an empty String is returned.
      @param pStyle The XSLT Stylesheet to use for the transformation or null for the identity transform.
      @param pRawStyle true to interpret the stylesheet as raw data, false to read it as a file.
      @return The transformed Node, or null if some catastrophe occurred.
  */
  public Node transformNode(Node pSource, String pStyle, boolean pRawStyle) {
    if(pSource == null) return null;
    Source srcData = null, srcStyle = null;
    InputStream isStyle = null;
    javax.xml.transform.dom.DOMResult resOut = null;
    FileInputStream fiData = null, fiStyle = null;
    try {
      if(Question.validString(pStyle)) {
        if(pRawStyle) {
          isStyle = new ByteArrayInputStream(pStyle.getBytes());
        } else {
          isStyle = new FileInputStream(pStyle);
        }
        srcStyle = new StreamSource(isStyle);
      }
      srcData = new javax.xml.transform.dom.DOMSource(pSource);
      resOut = new javax.xml.transform.dom.DOMResult();
      transformData(srcData, srcStyle, resOut);
      if(isStyle != null) isStyle.close();
    } catch(Exception ex) { CapCom.error("Exception transforming document.", ex); }
    return (resOut != null)?resOut.getNode():null;
  }

  /** Parse XML into a DOM tree. The source data String can be interpreted as raw data or as a filename. MTXPro sets itself as the ErrorHandler and will receive 
      all parsing error messages at the appropriate level. (Per the other <code>parse__()</code> methods. The only difference is that a DOM Document is returned.
      Any non-parsing errors are reported at Error level, and if they occur the return value is null.
      @param pData The XML to parse into a DOM tree. If blank or empty, this method returns null.
      @param pRawData true to interpret pSource as raw data, false to interpret as a filename.
      @return A fresh DOM Document, or null if an error occurred.
  */
  public Document parseDOM(String pData, boolean pRawData) {
    Document retval = null;
    if(!Question.validString(pData) || !isDOMAvailable()) return retval;
    try {
      InputStream isData = null;
      if(pRawData) {
        isData = new ByteArrayInputStream(pData.getBytes());
      } else {
        isData = new FileInputStream(pData);
      }
      retval = myDOMBuilder.parse(isData);
      isData.close();
    } catch(Exception ex) { CapCom.error("Error parsing DOM tree", ex); }
    return retval;
  }

  /** <p>Use this method to set a DOM Feature on the DocumentBuilderFactory. These features are saved and set when the DocumentBuilder is actually created, so any 
      incompatibilities will be reported then. NOTE: calling this method after <code>isDOMAvailable()</code> will NOT work! Features must be set before the factory 
      itself is instantiated. Calling <code>resetDOM()</code> before setting features or attributes clears everything and guarantees that any feature set before 
      the next call to <code>isDOMAvailable()</code> will attempt to set the features stored.</p>
      <p>If a feature cannot be set it is reported at the Debug level but does NOT interrupt setting of other features. If pName is null or empty it is ignored 
      silently.</p>
      @param pName The name of the feature to set. Ignored if null or empty.
      @param pValue The boolean value to set the feature to.
  */
  public void setDOMFeature(String pName, boolean pValue) {
    if(Question.validString(pName)) domFeatures.put(pName, pValue);
  }

  /** This method sets a DOM Attribute on the current DOM DocumentBuilderFactory. As with features, these are stored until a new factory is instantiated. Set them 
      before a call to <code>isDOMAvailable()</code> but after <code>resetDOM()</code>. If the Attribute name is null or empty it is ignored. If the Object value 
      is null it is ignored.
      @param pName The name of the Attribute to set.
      @param pValue The Object value to be set.
  */
  public void setDOMAttribute(String pName, Object pValue) {
    if(Question.validString(pName) && (pValue != null)) domAttributes.put(pName, pValue);
  }

  /** Check XPath availability. As with other methods of this type calling it attempts to create the necessary XPath equipment, so it should be checked at minimum 
      once per configuration change. Any creation errors are reported at DEBUG level.
      @return true if an <code>XPathFactory</code> is successfully instantiated, false otherwise.
  */
  public boolean isXPathAvailable() {
    if(myXFactory != null) return true;
    try {
      if(Question.validString(myXFactoryClassName) && (myLoader != null)) {
        myXFactory = XPathFactory.newInstance(XPathFactory.DEFAULT_OBJECT_MODEL_URI, myXFactoryClassName, myLoader);
      } else {
        myXFactory = XPathFactory.newInstance();
      }
    } catch(Exception ex) { CapCom.debug("Error creating XPath equipment", ex); }
    return (myXFactory != null);
  }

  /** Set an XPath feature. If pName is null or blank this method returns silently. If no <code>XPathFactory</code> instance has been created this method returns 
      silently; call <code>isXPathAvailable()</code> to create the factory. Any exceptions thrown are reported at DEBUG level.
      @param pName The name of the feature to set.
      @param pValue true to enable the feature, false to disable.
  */
  public void setXFeature(String pName, boolean pValue) {
    if(!Question.validString(pName) || (myXFactory == null)) return;
    try {
      myXFactory.setFeature(pName, pValue);
    } catch(Exception ex) { CapCom.debug("XPathFactory configuration error", ex); }
  }

  /** Reset the current XPath configuration. This does NOT change the class name; if that is set it must be cleared manually. */
  public void resetXPath() { myXFactory = null; }

  /** <p>This method applies the specified XPath to the supplied DOM <code>Node</code>. If pXPath is null or blank or pNode is null the return value is null. If 
      an error occurs it is reported at ERROR level and the return value is null. This method does not check the availability method but does ensure that an XPath 
      factory is defined; if not it returns null.</p>
      <p>Internally the XPath application specifies a <code>NodeList</code>. If this list contains one element it is returned. If it contains more than one element 
      each <code>Node</code> is added to a <code>DocumentFragment</code> which is cast as <code>Node</code> and returned. The <code>DocumentFragment</code> is 
      created from the first node's owning <code>Document</code>, if it exists. If not then the return value is the first node in the list. If the list contains 
      only one <code>Node</code> then it is returned.</p>
      <p>The XPath context object and method return values are both DOM <code>Node</code> instances. This allows multiple successive XPath applications. The final 
      result can then be transformed into whatever form or format is desired.</p>
      @param pXPath The XPath string to evaluate against the Node supplied.
      @param pNode The DOM Node the XPath is applied against.
      @return The result, as a DOM Node, of applying the XPath to the Node sent in.
  */
  public Node applyXPath(String pXPath, Node pNode) {
    Node retval = null;
    NodeList rezultz = null;
    if(!Question.validString(pXPath) || (pNode == null) || (myXFactory == null)) return retval;
    try {
      rezultz = ((NodeList) myXFactory.newXPath().evaluate(pXPath, pNode, XPathConstants.NODESET));
      switch(rezultz.getLength()) {
      case 0: break;
      case 1:
        retval = rezultz.item(0);
        break;
      default:
        retval = rezultz.item(0).getOwnerDocument();
        if(retval == null) {
          retval = rezultz.item(0);
        } else {
          retval = ((Document) retval).createDocumentFragment();
          for(int i=0; i < rezultz.getLength(); i++) retval.appendChild(rezultz.item(i));
        }
      }
    } catch(Exception ex) { CapCom.error("Error applying XPath", ex); }
    return retval;
  }

  /** This method is used to parse <code>cox.jmatt.java.MathTools.graphing.xml</code> package XML markup into a <code>GraphPaper</code> instance. Specific details 
      are {@link cox.jmatt.java.MathTools.graphing.GraphHandler here}. If pGraphSource is null or blank the method returns null immediately. Any parse errors are 
      reported at Error level, although some internal errors are reported at Debug.
      @param pEngine The <code>GraphEngine</code> to use in building the graph. If null, <code>GraphEngineXML</code> is used.
      @param pGraphSource The source of the XML to convert.
      @param pRawData true to parse the source as raw data, false to interpret it as a filename.
      @return A <code>GraphPaper</code> from the supplied engine, complete with components properly added.
  */
  public GraphPaper parseGraph(GraphEngine pEngine, String pGraphSource, boolean pRawData) {
    if(!Question.validString(pGraphSource)) return null;
    try {
      graphMode = true;
      myGraphHandler.setGraphEngine(pEngine);
      if(pRawData) {
        parseString(pGraphSource);
      } else {
        parseFile(pGraphSource);
      }
    } catch(Exception ex) { CapCom.error("Error parsing graph: "+pGraphSource, ex); }
    graphMode = false;
    return myGraphHandler.getResult();
  }

}

