package cox.jmatt.java.MathTools.util;

import cox.jmatt.java.MathTools.util.MathPrinter;
import cox.jmatt.java.MathTools.CapCom;
import cox.jmatt.java.MathTools.Question;
import java.util.Properties;

/** This class exists to provide scripting languages access to the MathTools logging framework and Global Pizza. If subclassing, DO NOT implement the MathPrinter 
    interface and set the subclass as either the Global Printer OR the subclass' MathPrinter. The universe will implode if this happens!
*/
public class MathConsole {

  /** Static constant for generating Student codes. */
  public static final byte STUDENT = MAuthenticator.AUTH_STUDENT;
  /** Static constant for Proctor codes. */
  public static final byte PROCTOR = MAuthenticator.AUTH_PROCTOR;
  /** Static constant for Instructor authentication codes. */
  public static final byte INSTRUCTOR = MAuthenticator.AUTH_INSTRUCTOR;
  /** String constant for a <code>SimpleAuthenticator</code>. */
  public static final String AUTH_SIMPLE = SimpleAuthenticator.MYSELF;

  private MathPrinter myPrinter = null;
  private Properties myPizza = null;

  /** The basic MathConsole constructor. It reflects all method calls to CapCom and fetches all data from CapCom. */
  public MathConsole() { super(); }
  
  /** Build a MathConsole around a MathPrinter instance. This may be the one set in CapCom or it may be different, but it must be specified! */
  public MathConsole(MathPrinter pPrinter) {
    super();
    myPrinter = pPrinter;
  }
  /** Build a MathConsole completely separate from CapCom. The Properties object specified will be used for all getProperty() calls, which makes the Global Pizza 
      inaccessible.
      @param pPrinter The MathPrinter to use for logging/printing calls. If null, CapCom is used.
      @param pPizza The Properties object to use for getProperty() calls. If null, the Global Pizza is used.
  */
  public MathConsole(MathPrinter pPrinter, Properties pPizza) {
    super();
    myPrinter = pPrinter;
    myPizza = pPizza;
  }

  /** <p>This method intercepts all logging calls made on this method directly OR on its convenience methods. If this console is built around a MathPrinter, that 
      printer receives all calls. If not, this method reflects to the CapCom mathLog() method.</p>
      <p>As with CapCom, MathConsole provides convenience methods for Debug, Info and Error: one method accepting a String and another that accepts a String and 
      a Throwable.</p>
  */
  public void mathLog(byte pLevel, String pMessage, Throwable pThr) {
    if(myPrinter == null) {
      CapCom.mathLog(pLevel, pMessage, pThr);
    } else {
      myPrinter.mathLog(pLevel, pMessage, pThr);
    }
  }
  public void error(String pMesg) { mathLog(MathPrinter.LOG_ERROR, pMesg, null); }
  public void error(String pMesg, Throwable pThr) { mathLog(MathPrinter.LOG_ERROR, pMesg, pThr); }
  public void debug(String pMesg) { mathLog(MathPrinter.LOG_DEBUG, pMesg, null); }
  public void debug(String pMesg, Throwable pThr) { mathLog(MathPrinter.LOG_DEBUG, pMesg, pThr); }
  public void info(String pMesg) { mathLog(MathPrinter.LOG_INFO, pMesg, null); }
  public void info(String pMesg, Throwable pThr) { mathLog(MathPrinter.LOG_INFO, pMesg, pThr); }

  /** Print a message outside the logging framework. If MathConsole has its own MathPrinter that is where the message goes. If not, it hits CapCom. The regular 
      println() method echoes to this one.
      @param pMessage The message to be printed.
  */
  public void mathPrintln(String pMessage) {
    if(myPrinter == null) {
      CapCom.mathPrintln(pMessage);
    } else {
      myPrinter.mathPrintln(pMessage);
    }
  }
  /** Laziness method aliasing mathPrintln(). */
  public void println(String pMessage) { mathPrintln(pMessage); }

  /** Process an Object on a MathPrinter. MathConsole's, if one is defined, CapCom's otherwise. */
  public Object processMathObject(Object pData) {
    if(myPrinter == null) {
      return CapCom.processMathObject(pData);
    }
      return myPrinter.processMathObject(pData);
  }

  /** <p>Sometimes a script needs to generate an error and terminate. This method does that. If the message is null or blank a default is supplied. When 
      called, this method throws a generic <code>Exception</code> which will be caught by <code>ScriptPanel</code>. The message should always be specified and 
      it should always be descriptive of the error condition.</p>
      @param pMesg The message to include in the Exception.
  */
  public void abort(String pMesg) throws Exception {
    throw new Exception(Question.validString(pMesg)?pMesg:"Unspecified script error.");
  }

  /** Fetch a property from a Properties object. The first time this method is called, it checks to see if MathConsole has a Properties object set. If so, that 
      is the one asked for the property. If no internal Properties object is currently loaded it fetches a copy of the Global Pizza and checks that.
      @param pKey The key of the property to be fetched. If not valid, the default is returned.
      @param pDefault The String (or null) to return if the specified property can't be found.
  */
  public String getProperty(String pKey, String pDefault) {
    if(!Question.validString(pKey)) return pDefault;
    if(myPizza == null) myPizza = CapCom.getPizza();
    return myPizza.getProperty(pKey, pDefault);
  }
  /** Convenience method for getProperty(pKey, null). */
  public String getProperty(String pKey) { return getProperty(pKey, null); }
  /** Get an Object from the CapCom shelf. If the required Object is present, it is returned. If not, null is returned. If the key value is null or empty, null 
      is returned.
    @param pKey The String key under which the Object was stored.
    @return The Object if it is there, null if not.
  */
  public Object getObject(String pKey) { return CapCom.getObject(pKey); }
  /** Store something on the CapCom shelf. If the key is null or empty or the Object is null the request is silently ignored.
      @param pKey The String key under which the Object is to be shelved.
      @param pData The Object to be stored.
  */
  public void storeObject(String pKey, Object pData) { CapCom.storeObject(pKey, pData); }

  /** <p>This method is used to explicitly store a return Object from a script. When called it calls {@link cox.jmatt.java.MathTools.CapCom#setScriptReturnObject} 
      and returns whatever value was sent in. This is done to facilitate the following:<br />
      <pre>Console.returns(someObject)</pre><br />
      as the final statement in a script. This construct works regardless of whether or not the <code>ScriptEngine.eval()</code> method actually returns a value.</p>
      @param pRetObject The return value from the script
      @return Whatever was sent in.
  */
  public final Object returns(Object pRetObject) {
    CapCom.setScriptReturnObject(pRetObject);
    return pRetObject;
  }
  /** Check an authentication code at Student level. */
  public final boolean authenticateStudent(String pCode) { return CapCom.authenticate(MAuthenticator.AUTH_STUDENT, pCode); }
  /** Check an authentication {@link cox.jmatt.java.MathTools.util.MAuthenticator code} at Proctor level. This and all other Authenticator methods reflect to 
      <code>CapCom</code>.
  */
  public final boolean authenticateProctor(String pCode) { return CapCom.authenticate(MAuthenticator.AUTH_PROCTOR, pCode); }
  /** Check an authentication code at Instructor level. */
  public final boolean authenticateInstructor(String pCode) { return CapCom.authenticate(MAuthenticator.AUTH_INSTRUCTOR, pCode); }
  /** Install an Authenticator. By default this is allowed but setting <code>CapCom.allowConsoleAuthenticatorAccess</code> to false blocks it.
      @param pName The full name of the <code>MAuthenticator</code> to install.
      @param pCtx The context Object, if required.
  */
  public final void setAuthenticator(String pName, Object pCtx) {
    if(CapCom.allowConsoleAuthenticatorAccess) CapCom.setMAuthenticator(pName, pCtx);
  }
  /** Generate a set of authentication codes, if access is permitted. Parameters are per {@link cox.jmatt.java.MathTools.util.MAuthenticator#generateCodes generateCodes()} 
      in the <code>MAuthenticator</code> interface. As with setting the authenticator, this method can be disabled from CapCom. If access is not allowed the return 
      value is null. The authentication levels are reflected in static constants here for convenience.
  */
  public final String[] generateCodes(byte pLevel, int pNumberOf, Object... pParams) {
    if(!CapCom.allowConsoleAuthenticatorAccess) return null;
    return CapCom.generateAuthenticationCodes(pLevel, pNumberOf, pParams);
  }

}
