package cox.jmatt.java.MathTools.util;

import java.util.*;
import java.sql.*;
import javax.sql.DataSource;
import javax.naming.Context;
import javax.naming.InitialContext;

import cox.jmatt.java.MathTools.*;

/** <p>This class provides MathTools access to JDBC-compliant databases. Since MathTools is a single-user desktop application the majority of the equipment here 
    utilizes <code>java.sql.Driver</code>s and requires direct access to the JDBC drivers. This is done via the MathClassLoader which is used automatically 
    and transparently any time a driver is set. Setting a driver via the <code>setDriver()</code> method causes the driver to be created. Once that is done a 
    Connection can be established using nothing but the URL, the URL and two Strings, or the URL and a Properties object.</p>
    <p>Successfully-created Connections are added to a static internal List. Once the application is finished a call to <code>closeAll()</code> or its instance 
    version closes ALL the Connections. <code>ScriptPanel</code> does this automatically as does <code>MTApplet</code> but other scripting apps or Applets that 
    override MTApplet's <code>destroy()</code> method must do this manually.</p>
    <p>Establishing the Connection is the extent of what <code>MathDBC</code> does; once the database connection is made the user takes over. No control or safety 
    measures are established for committing transactions; this <b>must</b> be handled manually. This is true for Connections established using a Driver or 
    DataSource; the only post-creation work done is from <code>closeAll()</code>.</p>
    <p>As stated previously CapCom's MathClassLoader is used for locating Drivers. MathTools has no native JDBC drivers so the URLLoaderPanel must be pointed to 
    the JAR containing one or the MathClassLoader set manually.</p>
    <h3>Connecting and Resetting</h3>
    <p>Creating a <code>Connection</code> absolutely requires a String URL. Period! This is the only globally mandatory parameter, though. A Connection can be 
    created with only this. There is also a method that uses the URL, a username and a password. Or the URL and a Properties object can be used.</p>
    <p>In order to use a Properties object to get a Connection, properties must be set. The <code>setProperty()</code> method does this. After initial creation 
    or after the <code>reset()</code> method is called there is <b>no</b> internal Properties object. The first time a property is set, even if its value is null 
    (remove), the internal Properties object is created and will be used. Individual properties can be removed by setting their value to null but the internal 
    Properties object remains or is defined, even if all content is removed.</p>
    <p>The <code>reset()</code> method does exactly that. It clears all internally-tracked things other than Connections: the current ClassLoader, the Properties 
    object used for connecting, the Driver, and the internal DataSource. This method is, in fact, the ONLY way to get rid of the Properties object or a DataSource. 
    Use with caution!</p>
    <p>NOTE: The <code>reset()</code> method does NOT close Connections! That is the <code>closeAll()</code> method's function. Nor does <code>closeAll()</code> 
    clear any internal settings, it simply closes all current Connections and clears the internal static list.</p>

    <h2>JNDI and DataSources</h2>
    <p>Establishing JNDI is well outside the scope of MathTools! If, however, JNDI is present it may be used for database access via <code>DataSource</code>s. The 
    <code>isAvailable()</code> method accepts a String argument used to look up the DataSource in the JNDI <code>Context lookup()</code> call. If this call succeeds 
    and returns a non-null DataSource it is maintained internally until the next call to <code>isAvailable()</code>.</p>
    <p>Once established, the internal DataSource is used in both <code>getDSConnection()</code> methods. The Connection returned is identical to those returned 
    by the DriverManager methods and each is tracked internally and is subject to <code>closeAll()</code>. The <code>reset()</code> method DOES clear out any 
    DataSources and can be used to start from scratch if necessary.</p>
    <p>By default <code>MathDBC</code> uses <code>CapCom.getNamingContext()</code> if it is set. If not, or if the CapCom's Context is set to null it uses the 
    zero-argument <code>InitialContext</code>. The <code>reset()</code> method resets to the CapCom Context but the <code>setDSContext()</code> method can be used 
    to set the Context to null and force the <code>IntialContext()</code> option. Of course <code>setDSContext()</code> can also be used to <i>set</i> a naming 
    Context if one is available!</p>
*/
public class MathDBC {

  private static ArrayList<Connection> myConnections = new ArrayList<Connection>();

  private ClassLoader myLoader = null;
  private Properties myProps = null;
  private Driver myDriver = null;
  private DataSource myDataSource = null;
  private Context myContext = CapCom.getNamingContext();

  /** Standard scripting constructor. Nothing special. */
  public MathDBC() { super(); }

  /** This method resets everything BUT existing Connections. Any previously-created Drivers are still available. */
  public void reset() {
    myLoader = null;
    myProps = null;
    myDriver = null;
    myDataSource = null;
    myContext = CapCom.getNamingContext();
  }

  /** <p>This method attempts to create the specified JDBC Driver. Any Exceptions are reported at DEBUG level. Since MathTools has no intrinsic JDBC Drivers this 
      method automatically uses the MathClassLoader. The first time it is called, or after <code>reset()</code>, it will re-load the MathClassLoader. If the 
      supplied String is null or blank nothing else happens.</p>
      <p>CRITICAL NOTE: <code>MathDBC</code> handles drivers manually, without <code>DriverManager</code>. (Because it did not want to behave!!) One and only ONE 
      <code>Driver</code> is active at any time; setting a new one replaces the previous one! Use carefully!!</p>
      @param pDriver The name of the JDBC Driver to create.
      @return true if the <code>Driver</code> was created, false otherwise.
  */
  public boolean setDriver(String pDriver) {
    if(!Question.validString(pDriver)) return false;
    if(myLoader == null) myLoader = CapCom.getMathClassLoader();
    try {
      myDriver = (Driver) Class.forName(pDriver, true, myLoader).newInstance();
    } catch(Exception ex) { CapCom.debug("Error creating JDBC Driver", ex); }
    return (myDriver != null);
  }

  /** Close ALL connections. Any errors are reported at DEBUG level. */
  public static final void closeAll() {
    for(Connection con : myConnections) {
      try { con.close(); } catch(Exception ex) { CapCom.debug("Error closing DB connection", ex); }
    }
    myConnections = new ArrayList<Connection>();
  }

  /** Instance version of <code>closeAll()</code>. */
  public final void close_All() { closeAll(); }

  /** Set a property to be used in the <code>getConnection(String, <u>Properties</u>)</code> method. If the internal Properties object is null it will be 
      created. If pKey is null or blank this method returns silently. If pValue is null or blank the current value for the key given, if it exists, is removed. 
      CRITICAL NOTE: Even if pValue is null this method creates a Properties object if there was not one before.
      @param pKey The name of the property to be added.
      @param pValue The value of the property to set, or null to remove it.
  */
  public void setProperty(String pKey, String pValue) {
    if(!Question.validString(pKey)) return;
    if(myProps == null) myProps = new Properties();
    if(Question.validString(pValue)) {
      myProps.setProperty(pKey, pValue);
    } else {
      myProps.remove(pKey);
    }
    return;
  }

  /** Add the contents of a <code>Properties</code> object to the internal connection properties.
      @param pNewProps The Properties object to add.
  */
  public void addProperties(Properties pNewProps) {
    if(pNewProps == null) return;
    if(myProps == null) myProps = new Properties();
    myProps.putAll(pNewProps);
  }

  /** Attempt to get a connection from a database URL, a username and a password. Any errors are reported at ERROR level and no string validation is done. If a 
      Connection is successfully created it is added to an internal List of open connections. If no <code>Driver</code> is set this method prints an Error level 
      message and returns null.
      @param pURL The URL of the database to connect to.
      @param pName The username to use.
      @param pPassword The password to use.
      @return The requested Connection or null if it failed.
  */
  public Connection getConnection(String pURL, String pName, String pPassword) {
    setProperty("user", pName);
    setProperty("password", pPassword);
    return getConnection(pURL);
  }

  /** Connect to the specified database URL. Any properties set will be sent in. If no <code>Driver</code> is set this method prints an Error level message and 
      returns null.
      @param pURL The URL of the database desired.
  */
  public Connection getConnection(String pURL) {
    Connection retval = null;
    if(myDriver == null) {
      CapCom.info("No JDBC driver set.");
      return retval;
    }
    try {
      if(myProps == null) myProps = new Properties();
      retval = myDriver.connect(pURL, myProps);
      if(retval != null) myConnections.add(retval);
    } catch(Exception ex) { CapCom.error("Error creating DB connection", ex); }
    return retval;
  }

  /** Use this method to test whether or not the JNDI <code>DataSource</code> API is available. If the supplied string is null or blank the method returns false 
      and no checking is actually done. Any errors are reported at DEBUG level. This method DOES clear any previously-set DataSource before attempting to create 
      a new one.
      @param pDatabase The Database to attempt using JNDI.
  */
  public boolean isDataSourceAvailable(String pDatabase) {
    if(!Question.validString(pDatabase)) return false;
    myDataSource = null;
    try {
      if(myContext == null) myContext = new InitialContext();
      myDataSource = (DataSource) myContext.lookup(pDatabase);
    } catch(Exception ex) { CapCom.debug("Error creating JNDI DataSource", ex); }
    return myDataSource != null;
  }

  /** Use this method to set the JNDI <code>Context</code> used for DataSource lookup. Setting it to null clears it and the next call to <code>isDSAvailable()</code> 
      results in a zero-argument <code>InitialContext</code> being used. To reset to the default (<code>CapCom.getNamingContext()</code>) use the <code>reset()</code> 
      method.
      @param pContext The <code>javax.naming.Context</code> used for <code>DataSource</code> instances, or null to clear it.
  */
  public void setDSContext(Context pContext) { myContext = pContext; }

  /** Get a Connection from a <code>DataSource</code>. This is the zero-argument method. Any errors are reported at Error level. If no DataSource has been 
      established this method returns null.
  */
  public Connection getDSConnection() {
    Connection retval = null;
    if(myDataSource == null) return retval;
    try {
      retval = myDataSource.getConnection();
      if(retval != null) myConnections.add(retval);
    } catch(Exception ex) { CapCom.error("Error establishing DataSource connection", ex); }
    return retval;
  }

  /** Get a Connection from a <code>DataSource</code> using the provided username and password. Any errors are reported at Error level. If no DataSource has been 
      established this method returns null. No String-checking is done before attempting the connection.
      @param pName The username for the connection.
      @param pPass The password for the connection.
  */
  public Connection getDSConnection(String pName, String pPass) {
    Connection retval = null;
    if(myDataSource == null) return retval;
    try {
      retval = myDataSource.getConnection(pName, pPass);
      if(retval != null) myConnections.add(retval);
    } catch(Exception ex) { CapCom.error("Error establishing DataSource connection", ex); }
    return retval;
  }

}

