package cox.jmatt.java.MathTools.util;

import java.io.*;
import java.util.*;
import java.util.regex.*;

import cox.jmatt.java.MathTools.*;

/** <p>This class provides MathTools with file IO capability. Since Applets do not routinely have filesystem access, <code>MTApplet</code> does not have access 
    to this class, nor will <code>MathEngine</code> create one. The zero-argument constructor of this class is private; at any given time there must be <b>only 
    one</b> base instance of this class. Use the <code>getInstance()</code> method to retrieve it.</p>
    <p><code>MathFileIO</code> works by wrapping an I/O Object of some flavor: InputStream, OutputStream, Reader, or Writer. It handles all read/write calls and 
    wraps any <code>Exceptions</code> thrown. More importantly, it keeps track of the open file references. When any or all need for file IO is gone, a call to 
    <code>closeAll()</code> closes any open MathFileIO instances and makes them eligible for garbage collection. <code>ScriptPanel</code> handles this automatically 
    but applications embedding and using this class must do it manually.</p>
    <p><code>MathFileIO</code> is designed to hide the grisly details of low-level file I/O from scripts, primarily, and from applications as needed. The class 
    itself wraps around an IO Object and contains all methods necessary to interact with it. All Exceptions thrown are caught gracefully and reported at Error 
    level, and anything opened is closed automatically, thus freeing resources when the script ends. (This is standard behavior for <code>ScriptPanel</code> but 
    should be implemented in applications!)</p>
    <p>As implied by its name, <code>MathFileIO</code> is designed to deal with <i>files</i>. By intent it was designed to deal with <b>text</b> files, as that is 
    what MathTools generate, but basic machinery is supplied to work with files of other types. Byte[] and char[] arrays can be created and used for both input and 
    output operations. This should be enough to allow binary or close-to-binary file manipulation. If a <code>graphing.*</code> implementation needs such machinery 
    it is here.</p>
    <p>In addition to simple files, <code>MathFileIO</code> can create <code>InputStream</code>s connected to resources. Scripts can use this to fetch files from 
    the MathTools JAR or from a remote URL. The only downside is these methods create an <code>InputStream</code>: use <code>toReader()</code> to change this.</p>
    <p>For advanced users <code>MathFileIO</code> provides the <code>newRandomAccessFile()</code> method. It requires a non-null, non-empty filename and a valid 
    mode: 'r', 'rw', 'rws', or 'rwd'. Unlike the kinder gentler creation methods this one returns its <code>RandomAccessFile</code> live and raw. The only protection 
    offered is ensuring that the filename and mode are not null or empty and proper closure from the <code>closeAll()</code> method. See the <code>java.io.RandomAccessFile</code> 
    javadoc for further information. <b>USE WITH CAUTION!</b></p>
    <p>The final piece of equipment this class provides is directory processing. This works on file names only but comes with a great deal of power and flexibility. 
    The basic process scans a specified directory and provides LOTS of information on the files it finds. This information appears as a rich set of formatting 
    tokens applied to a template sent in as an argument, with the resulting Strings returned in an array. This processing can be done recursively and it allows 
    for prefix-, suffix- and regex-filtering of the files it finds. In regex mode, the capture groups from the regular expression are also available as formatting 
    tokens.</p>
*/
public class MathFileIO {

  private static MathFileIO mySelf = null;
  private static ArrayList<MathFileIO> openFiles = null;
  private static ArrayList<RandomAccessFile> openRAFs = null;
  private static long procDirIndex = 0;
  private static String myProcDirTemplate = "<0>";

  private Reader myReader = null;
  private Writer myWriter = null;
  private InputStream myInputStream = null;
  private OutputStream myOutputStream = null;

  private MathFileIO(Reader pReader) {
    this();
    myReader = pReader;
    if(pReader != null) openFiles.add(this);
  }

  private MathFileIO(Writer pWriter) {
    this();
    myWriter = pWriter;
    if(pWriter != null) openFiles.add(this);
  }

  private MathFileIO(InputStream pInputStream) {
    this();
    myInputStream = pInputStream;
    if(pInputStream != null) openFiles.add(this);
  }

  private MathFileIO(OutputStream pOutputStream) {
    this();
    myOutputStream = pOutputStream;
    if(pOutputStream != null) openFiles.add(this);
  }

  private MathFileIO() {
    super();
    if(openFiles == null) openFiles = new ArrayList<MathFileIO>();
    if(mySelf == null) mySelf = this;
  }

  /** <p>This method provides powerful and flexible machinery for processing the filenames within a directory. It does not process the individual files, but rather 
      allows their names to be taken and formatted per <code>Question.fillTemplate()</code> with a rich set of file- and filename-related replacement tokens. File 
      names can be filtered three ways: prefix, suffix and regular expression. The first two use the String <code>startsWith()</code> and <code>endsWith</code> 
      methods. The third filters by regular expression match AND provides extra formatting tokens based on the capture groups. These filtering methods are not 
      mutually exclusive and can be combined. The logic used is exclusive, so prefix matching and regex matching MUST satisfy both conditions to be processed. 
      Filtering is enabled by non-null values for the corresponding Strings; sending in null disables that type of matching.</p>
      <p>If 'pRecursive' is set to 'true' all subdirectories of the one specified are processed as well. That is, they are processed <i>recursively</i>. In either 
      recursive or non-recursive mode there is no guarantee made as to the order in which the filenames will appear. This method does keep a count of the number 
      of files processed. This count is available as a number (long) via token #6. Tokens #7 and #8 also provide this counter as upper- and lowercase letters.</p>
      <p>The standard formatting tokens are available for all files, regardless of filtering. Two <code>java.io.File</code> constants are provided for convenience. 
      The 'index' tokens utilize a long value that begins at zero and increments by one with each file actually processed. The standard tokens are:
      <ul>
        <li>&lt;0&gt; - The pathname per <code>File.getPath()</code>.</li>
        <li>&lt;1&gt; - The full filename per <code>File.getName()</code>.</li>
        <li>&lt;2&gt; - The file's parent, or an empty String if there is no parent.</li>
        <li>&lt;3&gt; - The absolute pathname per <code>File.getAbsolutePath()</code>. (SX)</li>
        <li>&lt;4&gt; - The canonical pathname per <code>File.getCanonicalPath()</code>. (SX)</li>
        <li>&lt;5&gt; - The file URL per <code>File.toURI().toURL().toString()</code>. (SX)</li>
        <li>&lt;6&gt; - The current index as a number.</li>
        <li>&lt;7&gt; - The current index as an uppercase letter.</li>
        <li>&lt;8&gt; - The current index as a lowercase letter.</li>
        <li>&lt;9&gt; - The constant <code>java.io.File.separator</code>.</li>
        <li>&lt;10&gt; - The constant <code>java.io.File.pathSeparator</code>.</li>
        <li>&lt;11&gt; - The 'last modified' date, expressed as a long value. (SX)</li>
        <li>&lt;12&gt; - The file size in bytes. (SX)</li>
        <li>&lt;13&gt; - Execute, Read, Write and Hidden flags, in that order. If the corresponding 'can__()' method returns true, this String contains 'E', 'R', 
          'W' or 'H' in the first, second, third or fourth position. If the method returns false, that position is '-'. (SX)</li>
        <li>&lt;14&gt; - The base directory. Same as &lt;2&gt; for non-recursive processing.</li>
        <li>&lt;15&gt; - The relative path from the base (starting) directory to this file. Blank for non-recursive processing.</li>
        <li>&lt;16&gt; - The implied separator. Equals <code>File.separator</code> if #15 is not blank. Blank otherwise.</li>
        <li>&lt;17&gt; - The name of the file (#1) without its extension. This assumes the extension, if present, follows the final dot. If no dot is present 
          this is the same as token #1.</li>
        <li>&lt;18&gt; - The <i>smart dot</i>. If the filename has an extension this token contains a single dot ('.'), otherwise it is blank.</li>
        <li>&lt;19&gt; - The file extension: any characters following the final dot, if present. If not, this token is blank.</li>
        <li>&lt;20&gt; and up - Regex tokens, defined only during regex filtering.</li>
      </ul>
      Tokens marked 'SX' depend on <code>File</code> methods that may throw a security exception. If that happens they will contain the literal String 'Exception' 
      but the Exception itself is caught but not reported.</p>
      <p>Tokens #14, #15 and #16 are provided for recursive processing. #14 contains the path (<code>getPath()</code>) of the directory specified with the 
      method call. Token #15 contains the relative path to files 'beneath' the base directory. It consists of the file's <code>getPath()</code> String with the 
      base directory removed. It does NOT include the separator between them! This token is blank for files processed in the base directory. Token #16 contains 
      the <code>java.io.File.separator</code> constant when processing files in subdirectories and is blank for files within the base directory.</p>
      <p>Tokens #17 through #19 are based on the ever-classic 'filename-dot-extension' naming convention. If a dot is present in the filename, token #17 holds 
      everything before but not including the last dot and #19 holds everything after it. Token #18 holds a dot if both #17 and #19 contain non-empty Strings. 
      If there is no dot or if it appears as the first character, #17 holds the entire filename and the other two tokens are blank.</p>
      <p>In addition to the standard tokens defined above, which are valid for ALL files and matches, this method defines a set of <i>Regular Expression Tokens</i>. 
      These Regex tokens begin with token 20 ('&lt;20&gt;)' and contain any matched capture groups. Token 20 is group zero and the values move up from there. 
      These tokens result from the <code>group(int)</code> method <code>Matcher</code> defines with a (replacement token) offset of +20. This means token #20 
      (capture group zero) contains the entire pattern, #21 contains capture group 1 etc. These tokens are only defined during regular-expression matching! Any 
      tokens with null values instead contain empty Strings.</p>
      @param pDir The directory to process. If the String describes a file, its parent will be taken instead.
      @param pTemplate The template used to format the matched results.
      @param pRecursive Set 'true' to process the specified directory recursively, otherwise any subdirectories are ignored.
      @param pStartsWith Select only files that start with this String. If null it is ignored and no prefix-matching is done.
      @param pEndsWith Select only files that end with this. Ignored if null.
      @param pRegex If not null then all files are matched against this Regular Expression and capture groups are available for formatting. Ignored if null.
  */
  public static final String[] processDirectory(String pDir, String pTemplate, boolean pRecursive, String pStartsWith, String pEndsWith, String pRegex) {
    String[] retval = new String[0];
    if(!Question.validString(pDir)) return retval;
    pDir = pDir.trim();
    File rootFile = new File(pDir);
    if(!rootFile.isDirectory()) rootFile = rootFile.getParentFile();
    if(rootFile == null) return retval;
    File[] theFilez = rootFile.listFiles();
    if((theFilez == null) || (theFilez.length < 1)) return retval;
    pTemplate = Question.validString(pTemplate)?pTemplate:myProcDirTemplate;
    String sRoot = rootFile.getPath();
    CapCom.debug("Processing directory: "+sRoot);
    _setFileIndex(0);
    ArrayList<String> arlFilez = new ArrayList<String>();
    Pattern patRegex = null;
    if(Question.validString(pRegex)) {
      try {
        patRegex = Pattern.compile(pRegex);
      } catch(Exception ex) {
        CapCom.error("Error compiling regex: "+pRegex, ex);
        patRegex = null;
      }
    }
    for(File fil : theFilez) _processFiles(fil, pTemplate, pRecursive, arlFilez, pStartsWith, pEndsWith, patRegex, sRoot);
    retval = new String[arlFilez.size()];
    int iNdx = 0;
    for(String sFil : arlFilez) retval[iNdx++] = sFil;
    return retval;
  }

  /** Convenience method that only takes the first three parameters. No file filtering is done and all such arguments are null. */
  public static final String[] processDirectory(String pDir, String pTemplate, boolean pRecursive) { return processDirectory(pDir, pTemplate, pRecursive, null, null, null); }

  /** Instance shadow for the full method. */
  public String[] process_Directory(String pDir, String pTemplate, boolean pRecursive, String pStartsWith, String pEndsWith, String pRegex) {
    return processDirectory(pDir, pTemplate, pRecursive, pStartsWith, pEndsWith, pRegex);
  }

  /** Instance shadow for the short method. */
  public String[] process_Directory(String pDir, String pTemplate, boolean pRecursive) {
    return processDirectory(pDir, pTemplate, pRecursive, null, null, null);
  }

  /* Process the files in a directory. */
  private static final void _processFiles(File pDir, String pTemplate, boolean pRecursive, ArrayList<String> pFiles, String pStartsWith, String pEndsWith, Pattern pPattern, String pBasePath) {
    if(pRecursive && pDir.isDirectory()) {
      File[] theFilez = pDir.listFiles();
      for(File fil : theFilez) _processFiles(fil, pTemplate, pRecursive, pFiles, pStartsWith, pEndsWith, pPattern, pBasePath);
    }
    if(pDir.isFile()) {
      String sTemp = pDir.getName().trim();
      if((pStartsWith != null) && !sTemp.startsWith(pStartsWith)) return;
      if((pEndsWith != null) && !sTemp.endsWith(pEndsWith)) return;
      if(pPattern == null) {
        pFiles.add(Question.fillTemplate(pTemplate, _makeFormatTokens(pDir, pBasePath)));
      } else {
        Matcher matRegex = pPattern.matcher(sTemp);
        if(!matRegex.matches()) return;
        String[] stdTokens = _makeFormatTokens(pDir, pBasePath);
        int iStart = stdTokens.length, iMax = matRegex.groupCount()+1;
        String[] tokenz = new String[iStart + iMax];
        for(int i=0; i < iStart; i++) tokenz[i] = stdTokens[i];
        for(int i = iStart; i < tokenz.length; i++) {
          tokenz[i] = matRegex.group(i - iStart);
          if(tokenz[i] == null) tokenz[i] = "";
        }
        pFiles.add(Question.fillTemplate(pTemplate, tokenz));
      }
    }
  }

  /* Create formatting tokens. */
  private static final String[] _makeFormatTokens(File pDir, String pBasePath) {
    String[] retval = new String[0];
      if(pDir == null) return retval;
      retval = new String[20];
      retval[0] = pDir.getPath(); retval[1] = pDir.getName(); retval[2] = pDir.getParent(); retval[2] = Question.validString(retval[2])?retval[2]:"";
      try {
        retval[3] = pDir.getAbsolutePath(); retval[4] = pDir.getCanonicalPath(); retval[5] = pDir.toURI().toURL().toString();
        retval[11] = "" + pDir.lastModified(); retval[12] = "" + pDir.length();
        retval[13] = (pDir.canExecute()?"E":"-") + (pDir.canRead()?"R":"-") + (pDir.canWrite()?"W":"-") + (pDir.isHidden()?"H":"-");
      } catch(Exception ignored) {
        retval[3] = "Exception"; retval[4] = "Exception"; retval[5] = "Exception";
        retval[11] = "Exception"; retval[12] = "Exception"; retval[13] = "Exception";
      }
      long pIndex = _getFileIndex(); retval[6] = ""+pIndex; retval[7] = Question.numberToLetter(pIndex); retval[8] = retval[7].toLowerCase();
      retval[9] = File.separator; retval[10] = File.pathSeparator;
      retval[14] = pBasePath; retval[15] = (retval[2]+"    ").substring(pBasePath.length()+1).trim();
      retval[16] = Question.validString(retval[15])?File.separator:"";
      int theDot = retval[1].lastIndexOf(".");
      if(theDot > 0) {
        retval[17] = retval[1].substring(0, theDot);
        retval[19] = (retval[1]+"    ").substring(theDot + 1).trim();
        retval[18] = ".";
      } else {
        retval[17] = retval[1]; retval[18] = ""; retval[19] = "";
      }
    return retval;
  }

  /** Set a default directory processing template. If pTemplate is null or blank, '&lt;0&gt;' is set. */
  public static final void setDirProcTemplate(String pTemplate) { myProcDirTemplate = Question.validString(pTemplate)?pTemplate:"<0>"; }

  /** Instance template-setting method. */
  public final void setNewDirProcTemplate(String pTemplate) { setDirProcTemplate(pTemplate); }

  private static long _getFileIndex() { return procDirIndex++; }

  private static void _setFileIndex(long pVal) { procDirIndex = (pVal > -1)?pVal:0; }

  /** This method returns the base instance of <code>MathFileIO</code>. It is a factory method: if no instance exists it creates one but if an instance does 
      already exist this method returns a reference to it. This is done to allow proper closing and disposal of resources.
  */
  public static final MathFileIO newInstance() {
    if(mySelf == null) mySelf = new MathFileIO();
    return mySelf;
  }

  /** <p>This method returns a new MathFileIO wrapped around whatever I/O Object and file are requested. The options are:
      <ul>
        <li>'I' sets a BufferedInputStream.</li>
        <li>'R' sets a BufferedReader.</li>
        <li>'G' is used for <code>getResourceAsStream()</code>.
        <li>'M' is per 'G' but using <code>CapCom.getMathClassLoader()</code>.</li>
        <li>'O' sets a BufferedOutputStream.</li>
        <li>'W' sets a BufferedWriter.</li>
      </ul></p>
      <p>Attempting to set anything other than one of these results in a null return. Also, if the filename is null or empty, null is returned. Any Exceptions are 
      reported at Error level and the return value for such is null.</p>
      @param pFileName The name of the file to be I/O'ed.
      @param pType One of the characters above, upper or lower case.
      @return A fresh <code>MathFileIO</code> ready for I/O, or null if something awful happened.
  */
  public static final MathFileIO newInstance(String pFileName, char pType) {
    MathFileIO retval = null;
    if(!Question.validString(pFileName)) return retval;
    try {
      switch(pType) {
        case 'I':
        case 'i':
          retval = new MathFileIO(new BufferedInputStream(new FileInputStream(pFileName)));
          break;
        case 'O':
        case 'o':
          retval = new MathFileIO(new BufferedOutputStream(new FileOutputStream(pFileName)));
          break;
        case 'R':
        case 'r':
          retval = new MathFileIO(new BufferedReader(new FileReader(pFileName)));
          break;
        case 'W':
        case 'w':
          retval = new MathFileIO(new BufferedWriter(new FileWriter(pFileName)));
          break;
        case 'G':
        case 'g':
          retval = new MathFileIO(Class.forName("cox.jmatt.java.MathTools.util.MathFileIO").getResourceAsStream(pFileName));
          break;
        case 'M':
        case 'm':
          retval = new MathFileIO(Class.forName("cox.jmatt.java.MathTools.util.MathFileIO", true, CapCom.getMathClassLoader()).getResourceAsStream(pFileName));
          break;
        default:
      }
    } catch(Exception ex) { CapCom.error("Error creating file: "+pFileName, ex); }
    return retval;
  }

  /** Create an InputStream MathFileIO. */
  public MathFileIO newInputStream(String pFileName) { return newInstance(pFileName, 'I'); }

  /** Create an OutputStream MathFileIO. */
  public MathFileIO newOutputStream(String pFileName) { return newInstance(pFileName, 'O'); }

  /** Create a MathFileIO Reader. */
  public MathFileIO newReader(String pFileName) { return newInstance(pFileName, 'R'); }

  /** Create a MathFileIO Writer. */
  public MathFileIO newWriter(String pFileName) { return newInstance(pFileName, 'W'); }

  /** Use MathFileIO to get an InputStream connected to a specific resource. (Simple version.) */
  public MathFileIO newResourceStream(String pFileName) { return newInstance(pFileName, 'G'); }

  /** Use MathFileIO to get an InputStream connected to a specific resource using <code>CapCom.getMathClassLoader()</code>. */
  public MathFileIO newLoaderStream(String pFileName) { return newInstance(pFileName, 'M'); }

  /** Create and return a <code>java.io.RandomAccessFile</code>. Unlike the other creation methods this one returns a <i>live</i> <code>RandomAccessFile</code>. 
      There is no Exception-proofing or wrapping, the file itself is returned unwrapped. It is still tracked, so the <code>closeAll()</code> method will dispose 
      of it but there is no other protection on it. <b>USE WITH CAUTION!!</b> If the filename or mode are empty or null the return value is null. If the mode is 
      not valid the return value is null. Any creation errors are reported at the Error level.
      @param pFileName The name of the file to create or open.
      @param pMode The mode in which the file is to operate.
  */
  public RandomAccessFile newRandomAccessFile(String pFileName, String pMode) {
    RandomAccessFile retval = null;
    if(!Question.validString(pFileName) || !Question.validString(pMode)) return retval;
    if(openRAFs == null) openRAFs = new ArrayList<RandomAccessFile>();
    try {
      retval = new RandomAccessFile(pFileName, pMode);
      openRAFs.add(retval);
    } catch(Exception ex) { CapCom.error("Error creating RandomAccessFile", ex); }
    return retval;
  }

  /** Convert an <code>InputStream</code> instance to a <code>Reader</code>. This has no effect on an output instance or an instance that already encapsulates 
      a <code>Reader</code>. USE WITH CAUTION: This is a one-way trip and the underlying stream is lost in the conversion! Any errors are reported at Error level 
      and the underlying stream is not affected. This method returns a self-reference.
  */
  public MathFileIO toReader() {
    if((myInputStream == null) || (myReader != null)) return this;
    try {
      myReader = new BufferedReader(new InputStreamReader(myInputStream));
      myInputStream = null;
    } catch(Exception ex) {
      CapCom.error("Error converting MathFileIO to Reader", ex);
      myReader = null;
    }
    return this;
  }

  /** Convert an <code>OutputStream</code> instance to a <code>Writer</code>. This has no effect on an input instance or an instance that already encapsulates 
      a <code>Writer</code>. USE WITH CAUTION: This is a one-way trip and the underlying stream is lost in the conversion! Any errors are reported at Error level 
      and the underlying stream is not affected. This method returns a reference to itself.
  */
  public MathFileIO toWriter() {
    if((myOutputStream == null) || (myWriter != null)) return this;
    try {
      myWriter = new BufferedWriter(new OutputStreamWriter(myOutputStream));
      myOutputStream = null;
    } catch(Exception ex) {
      CapCom.error("Error converting MathFileIO to Writer", ex);
      myWriter = null;
    }
    return this;
  }

  /** Flush the OutputStream or Writer encapsulated by this MFileIO. If it is not an output instance, nothing happens. Multiple calls to this method will have 
      no bad effects, even if it is an input Object. Any Exceptions thrown are reported at the Error level.
  */
  public void flush() {
    try {
      if(myWriter != null) myWriter.flush();
      if(myOutputStream != null) myOutputStream.flush();
    } catch(Exception ex) { CapCom.error("Error flushing.",ex); }
  }

  /** Close this MFileIO instance. Any exceptions are reported at Error level. Note: This method does NOT automatically <code>flush()</code> OutputStreams or 
      Writers!
  */
  public void close() {
    try {
      if(myReader != null) myReader.close();
      myReader = null;
      if(myWriter != null) myWriter.close();
      myWriter = null;
      if(myInputStream != null) myInputStream.close();
      myInputStream = null;
      if(myOutputStream != null) myOutputStream.close();
      myOutputStream = null;
      openFiles.remove(this);
    } catch(Exception ex) { CapCom.error("Error closing IO", ex); }
  }

  /** Close ALL currently-open <code>MathFileIO</code> instances. */
  public static final void closeAll() {
    if(openFiles != null) {
      Object[] curOpen = openFiles.toArray();
      for(int i=0; i < curOpen.length; i++) {
        ((MathFileIO) curOpen[i]).flush();
        ((MathFileIO) curOpen[i]).close();
      }
      openFiles = new ArrayList<MathFileIO>();
    }
    if(openRAFs != null) {
      for(Iterator<RandomAccessFile> it = openRAFs.iterator(); it.hasNext(); ) {
        try {
          it.next().close();
        } catch(Exception ex) { CapCom.error("Error closing RandomAccessFile", ex); }
      }
      openRAFs = null;
    }
  }

  /** Instance version of <code>closeAll()</code>. */
  public void close_All() { closeAll(); }

  /** Create and return a new byte[] array. If the requested length is less than one, the array '{0}' is returned. NOTE: The array is created but NOT filled! */
  public byte[] newByteArray(int pLength) {
    byte[] retval = {0};
    if(pLength > 0) retval = new byte[pLength];
    return retval;
  }

  /** Create and return a new char[] array. If the requested length is less than one, '{0}' is returned. NOTE: The array is created but NOT filled! */
  public char[] newCharArray(int pLength) {
    char[] retval = {0};
    if(pLength > 0) retval = new char[pLength];
    return retval;
  }

  /** Write a String to the OutputStream or Writer. No newline is appended, the String is sent as-is! If the String is null it is ignored. If empty but not null 
      it is written. Any Exceptions are reported at Error level. Calling this method on an Input object has no effect.
      @param pData The String to send out.
  */
  public void write(String pData) {
    if(pData == null) return;
    byte[] baData = pData.getBytes();
    try {
      if(myOutputStream != null) myOutputStream.write(baData, 0, baData.length);
      if(myWriter != null) myWriter.write(pData, 0, pData.length());
    } catch(Exception ex) { CapCom.error("Output exception.", ex); }
  }

  /** Write a single byte (OutputStream) or char (Writer) to the Output object. No effect on Input objects. Any errors are reported at Error level.
      @param pData The byte or char to write, the selection is automatic.
  */
  public void write(int pData) {
    try {
      if(myOutputStream != null) myOutputStream.write(pData);
      if(myWriter != null) myWriter.write(pData);
    } catch(Exception ex) { CapCom.error("Write exception.", ex); }
  }

  /** Write a String per the plain <code>write()</code> method BUT append a newline (OutputStream) or call <code>newLine()</code> (Writer). */
  public void writeLine(String pLine) {
    if(pLine == null) return;
    byte[] baData = (pLine + "\n").getBytes();
    try {
      if(myOutputStream != null) myOutputStream.write(baData, 0, baData.length);
      if(myWriter != null) {
        myWriter.write(pLine, 0, pLine.length());
        ((BufferedWriter) myWriter).newLine();
      }
    } catch(Exception ex) { CapCom.error("Output exception.", ex); }
  }

  /** Issue a write to the underlying OutputStream. This method DOES NOTHING if the IO Object is NOT an OutputStream!
      @param pData The array of bytes to write.
      @param pOffset The starting point within the array.
      @param pLength The number of bytes to write.
  */
  public void write(byte[] pData, int pOffset, int pLength) {
    if(myOutputStream != null) { try { myOutputStream.write(pData, pOffset, pLength); } catch(Exception ex) { CapCom.error("Error writing to OutputStream.", ex); } }
  }

  /** Issue a write to the underlying Writer. This method has NO EFFECT if the IO Object is NOT a Writer!
      @param pData The char[] array to write.
      @param pOffset The starting point within the array.
      @param pLength The number of bytes to write.
  */
  public void write(char[] pData, int pOffset, int pLength) {
    if(myWriter != null) { try { myWriter.write(pData, pOffset, pLength); } catch(Exception ex) { CapCom.error("Error writing to Writer.", ex); } }
  }

  /** Read a line. Return is null if the IO object is not a Reader, there is no data, or an error occurs! */
  public String readLine() {
    String retval = null;
    if(myReader != null) {
      try{
        retval = ((BufferedReader) myReader).readLine();
      } catch(Exception ex) { CapCom.error("Error reading a line.", ex); }
    }
    return retval;
  }

  /** Read a <b>byte</b> or <b>char</b> of data, depending on the IO Object. Returns -1 at the end of file or if an error occurs. Calling this method on an Output 
      object always returns -1.
  */
  public int read() {
    int retval = -1;
    try {
      if(myInputStream != null) retval = myInputStream.read();
      if(myWriter != null) retval = myReader.read();
    } catch(Exception ex) { CapCom.error("Error reading file.", ex); }
    return retval;
  }

  /** <p>Read a sequence of bytes into an array. The return value equals the number of bytes read UNLESS:
      <ul>
        <li>The IO Object is <b>not</b> an InputStream.</li>
        <li>The end-of-file has been reached.</li>
        <li>An error occurs</li>
      </ul></p>
      <p>If any of these occur, the return value is -1</p>
      @param bData Byte array into which to read the data.
      @param pOffset Location in the array to start writing.
      @param pLength The number of bytes to read.
      @return The number of bytes read, or -1.
  */
  public int read(byte[] bData, int pOffset, int pLength) {
    int retval = -1;
    if(myInputStream != null) {
      try {
        retval = ((BufferedInputStream) myInputStream).read(bData, pOffset, pLength);
      } catch(Exception ex) { CapCom.error("Error reading InputStream", ex); }
    }
    return retval;
  }

  /** <p>Read a sequence of <code>char</code> into an array. The return value equals the number read UNLESS:
      <ul>
        <li>The IO Object is <b>not</b> a Reader.</li>
        <li>The end-of-file has been reached.</li>
        <li>An error occurs</li>
      </ul></p>
      <p>If any of these occur, the return value is -1</p>
      @param bData Char array into which to read the data.
      @param pOffset Location in the array to start writing.
      @param pLength The number of bytes to read.
      @return The number of <code>char</code> read, or -1.
  */
  public int read(char[] bData, int pOffset, int pLength) {
    int retval = -1;
    if(myReader != null) {
      try {
        retval = ((BufferedReader) myReader).read(bData, pOffset, pLength);
      } catch(Exception ex) { CapCom.error("Error reading Reader", ex); }
    }
    return retval;
  }

}

