package cox.jmatt.java.MathTools.util;

/** <p>This interface describes the mechanism by which the MathTools classes communicate with external applications. It provides a very basic logging framework, 
    a general-purpose print method, and a generic processMathObject() method.</p>
    <p>The logging framework is not intended to be a comprehensive logging package. Rather, it represents the minimum equipment needed to debug MathTools apps 
    and to report error conditions within the MathTools classes themselves.</p>
    <p>The mathPrintln() and processMathObject() methods provide a non-logging way to communicate with applications. mathPrintln() is designed to print messages 
    outside the logging framework. The processMathObject() method is designed to provide a very generic way for a customized MathPrinter to process something 
    defined by the application implementing it.</p>
    <p>Operationally, a MathPrinter works in conjunction with CapCom. CapCom implements all MathPrinter methods (but NOT the interface itself!) and the MathTools 
    classes that generate logging messages all report to CapCom. CapCom contains a method to set a global MathPrinter and, if one is set, all log, print and 
    processMathObject() calls are directed to it. If no MathPrinter is set, CapCom silently ignores all such calls. Most of the classes in the demo package also 
    utilize the Global Printer for their output, so adding them to other applications will require a Global Printer!</p>
*/
public interface MathPrinter {

  /** Info-level logging is for messages of interest but not necessarily a program-threatening nature. */
  public static final byte LOG_INFO = 0;
  /** Debug is used to report internal conditions within MathTools classes. It is primarily intended for debugging applications. */
  public static final byte LOG_DEBUG = 1;
  /** Error level represents a serious problem that may terminate program execution. */
  public static final byte LOG_ERROR = 2;

  /** This method provides a connection for MathTools to log errors and information. What is done with this data is implementation-specific.
      @param pLevel One of the LOG_ levels defined above.
      @param pMesg A String message describing the event.
      @param pThr A Throwable generated by the event. May be null.
  */
  public void mathLog(byte pLevel, String pMesg, Throwable pThr);

  /** Print a message to something that will be read either immediately or later.
      @param pMesg The String message to be printed.
  */
  public void mathPrintln(String pMesg);

  /** This method provides a basic, generic extension point for custom MathPrinter implementations. Its use is not specified other than that, and none of the 
      MathTools classes actually use it. All other details are implementation-specific. A non-implemented message should simply return the Object it is passed.
      @param pData The data to be processed.
      @return The result of processing the data given.
  */
  public Object processMathObject(Object pData);
}
