package cox.jmatt.java.MathTools.util;

import java.io.*;
import java.util.*;
import cox.jmatt.java.MathTools.*;
import cox.jmatt.java.MathTools.markup.html.*;

/** <p>This class is designed to take a Noteworthy-formatted document and convert it to HTML. To use the parser first create an instance. Next create and 
    configure a <code>MathGenHTML</code> for the parser to use. This is handy if some of the tags generated need default attributes set. Finally, feed the 
    <code>parse()</code> method a newline-separated String of Noteworthy-formatted text. Unless the universe collapses the method will return an 
    <code>HTMLTag</code> containing the HTML-formatted document. This can then be manipulated as needed.</p>
    <h2>The Noteworthy Markup Format</h2>
    <p><b>Noteworthy</b> is a set of rules and a parser designed to convert plain text lecture notes into HTML documents. Specifically, it allows both unparsed 
    notes <i>and</i> parsed markup to be mixed in a single document. The basic parse-marker is the double quote (N.B. This is the single-key double-quote character, 
    not a double apostrophe!) followed by a special formatting character. The parser extracts these lines, converts them to HTML, and returns them in a 
    <code>HTMLTag</code> instance. Unquoted lines are ignored.</p>
    <p>Noteworthy input is, as stated previously, plain text. Lines that begin with special (and <i>logical</i>) formatting characters are formatted per their 
    function. Other parsed lines are assumed to be part of a paragraph, division, or directly within the &lt;body&gt; of the document. Some nesting of tag types 
    is possible but it is tricky at best.</p>
    <p>Any tag, list, paragraph or division can be closed with a single blank line. One blank line closes the closest previous open 'thing.' So if a list follows 
    a paragraph the first blank line closes the list. In the case of lists, though, <i>any</i> line starting with a non-list character terminates the list. If 
    the list immediately follows an open paragraph (no blank lines between) it will be included in the paragraph. The best way to gain a feel for this is to 
    experiment!</p>
    <h3>Noteworthy Structure</h3>
    <p>The structural philosophy of Noteworthy is fairly rigid. The ultimate top-level container is the &lt;body&gt; tag. All elements are contained within it 
    and it doesn't close until the end of the document. The body can hold divisions, paragraphs, or lists. Divisions can hold paragraphs and lists. Heading 
    elements of any level can be included in the body or a division.</p>
    </p>Opening a new division ('&lt;div&gt;' tag) closes the previous division, if one was defined, which automatically closes all other open things within it. 
    A division can also be explicitly closed by issuing the <i>close all</i> command: '!'.</p>
    <p>The text structure is very simple. Special characters must be the first non-whitespace on the line but all lines are trimmed before parsing. The data 
    portion of the line, which is anything <i>after</i> a special character or the entire line for paragraphs, is also trimmed before it is used. This allows 
    hard-formatting using spaces or tabs, which allows and encourages organized and readable input.</p>
    <p>Headings are preceded by one or more hash marks ('#'). The number of these characters indicates the heading level. Any hash marks at the end of the line 
    are ignored and will be stripped before the heading is created, so symmetry is allowed or not, although it is more visually appealing.</p>
    <h4>Tables</h4>
    <p>Tables are started with a single pipe character, '|'. The first such indicates that a table is to be created, subsequent lines become the rows of the 
    table. Table cell elements ('&lt;td&gt;') are separated by a double pipe with no space between. If the first line of the table contains cell elements they 
    become the headers ('&lt;th&gt;'), otherwise the table has none. Subsequent rows all describe standard cells within the table, '&lt;td&gt;'.</p>
    <h4>ID Attributes</h4>
    <p>All major structural elements within the document are assigned 'id=' attributes automatically: &lt;div&gt;, &lt;p&gt;, &lt;ol&gt;, &lt;ul&gt; and &lt;dl&gt;. 
    Other tags do not receive IDs automatically. The IDs are very straightforward: tag type plus a count appended to the end. For example the second paragraph in 
    the document receives the ID 'par2'.</p>
    <h4>Raw Mode</h4>
    <p>When the automatic behavior simply will not suffice, <i>raw mode</i> is available. Raw mode is toggled by three equal signs alone on a line, '==='. 
    Everything from the line after raw mode is activated to the line before it ends is added to the document without formatting. The lines are trimmed and any 
    blank lines are ignored, but any non-blank lines are included exactly as written.</p>
    
    <h3>Formatting Characters</h3>
    <p>The special formatting characters and their actions are:
      <table border="1" cellpadding="8" >
        <tr><th>Character</th><th>Action</th></tr>
        <tr><td>/ or /#...</td><td>Open a division. This closes <b>all</b> open tags back to &lt;body&gt; and starts a fresh division.<br />The second 
          version ('/#') opens a division and includes a header of the given level.</td></tr>
        <tr><td>&quot;0, &quot;1, &quot;2, &quot;3, &quot;4,<br />&quot;5, &quot;6, &quot;7, &quot;8, &quot;9</td><td>Start an ordered list or add an item to an 
          existing list or a &lt;dd&gt; to a definition list.</td></tr>
        <tr><td>&quot;*,  &quot;<b>.</b>, &quot;+, &quot;-</td><td>Start an ordered list or add an item to an existing list or a &lt;dd&gt; to a definition list.</td></tr>
        <tr><td> &quot;: </td><td>Start a definition list or ad a &lt;dt&gt; to an existing one.</td></tr>
        <tr><td> &quot;# </td><td> Add a heading tag to any eligible element. The number of hashes indicates the heading level.</td></tr>
        <tr><td> &quot;| </td><td>Start a table or add a row to an existing one.<br />Individual elements are double-pipe separated.<br />The first row (if defined) 
          becomes the table headers, subsequent rows are &lt;td&gt; elements.</td></tr>
        <tr><td> &quot;(blank line) </td><td> A blank line closes the <b>last</b> tag opened.</td></tr>
        <tr><td> &quot;= </td><td>A single equals sign is used to escape a line that begins with an otherwise-significant character.<br />To begin a line with an 
          actual equals sign, double it! To start a line with a double-equal use '= =='.</td></tr>
        <tr><td> &quot;! </td><td> Forcibly close all tags back to the &lt;body&gt; tag.</td></tr>
        <tr><td> &quot;{ </td><td> Execute a macro command (see below).</td></tr>
        <tr><td> '&quot;===' </td><td> Toggle <i>raw mode</i>. In raw mode all content is added as-is to whatever tag is currently open.<br />Null or blank lines are 
          ignored and the lines added are trimmed.<br />The triple-equals must be the only thing on the line.</td></tr>
        <tr><td> Any other character </td><td>Other characters either form &lt;p&gt; tags or add directly to existing &lt;p&gt; or &lt;div&gt; tags.</td></tr>
      </table></p>
      <p>Ordered, unordered and definition lists opened immediately beneath a paragraph (no blank lines between) are included in the paragraph. Lines beginning 
      immediately after this are also included in the same paragraph. Whitespace before and after the formatting characters is trimmed as is any at the end of 
      each line. Where necessary space is replaced.</p>

      <h3>Macro Commands</h3>
      <p>Macro commands are directives that do not specifically affect formatting. They may affect procesing or inline configuration.</p>
      <p>All macro commands consist of a left curly brace followed (no space!) by one or more other letters or symbols. Some commands have arguments; these are 
      separated from the command itself by at least one space. The closing brace is not required. Available macro commands are:
      <table border="1" cellpadding="8" >
        <tr><th>Command</th><th>Result</th></tr>
        <tr><td> &quot;{t <i>Page Title</i></td><td>Set the title of the HTML document; <code>HTMLTag setTitle()</code>.</td></tr>
        <tr><td> &quot;{i <i>fileName</i><br />&quot;{g <i>fileName</i></td><td>Copy the contents of an external file into the document (rawly!).<br />The first 
          version uses a standard <code>FileReader</code> and the second uses <code>getResourceAsStream()</code>.<br />Any exceptions are reported at Error level.</td></tr>
      </table></p>
      <p>Macro commands are still considered standard directives; they will not be processed in raw mode!</p>
*/
public class NoteworthyParser {

  private MathGenHTML theGenerator = null;
  private HTMLTag mainBody = null;
  private BaseHTag curTag = null;
  private Stack<BaseHTag> tagStack = null;
  private int divCount = 1, pCount = 1, ulCount = 1, olCount = 1, dlCount = 1, tableCount = 1, opcode = -1;
  private boolean rawMode = false;
  private String opData = null;

  /** Standard constructor used for scripting and other instances. */
  public NoteworthyParser() { super(); }

  private static boolean validString(String pStr) {
    return ((pStr != null) && (pStr.trim().length() > 0));
  }

  /** Set the <code>MathGenHTML</code> instance used to generate tag classes. Null clears it and one will be generated automatically when needed. */
  public void setGenerator(MathGenHTML pGen) { theGenerator = pGen; }

  /** This is the reason this class exists. It takes a String of Noteworthy data and converts it into HTML. If the string sent in is null or blank it is returned 
      unaltered.
      @param pNoteworthy The Noteworthy-formatted data String.
      @return An <code>HTMLTag</code> containing the proper markup.
  */
  public HTMLTag parse(String pNoteworthy) {
    if(!validString(pNoteworthy)) return new HTMLTag();
    initParse();
    for(String sLine : pNoteworthy.split("\n")) parseLine(sLine);
    closeAllTags();
    return mainBody;
  }

  /* Initialize parsing equipment. */
  private void initParse() {
    if(theGenerator == null) theGenerator = new MathGenHTML();
    mainBody = theGenerator.getHTML();
    tagStack = new Stack<BaseHTag>();
    curTag = null;
    divCount = 1; pCount = 1; ulCount = 1; olCount = 1; dlCount = 1; tableCount = 1;
    rawMode = false;
  }

  /* Parse a single line. */
  private void parseLine(String pLine) {
    boolean gotLine = validString(pLine);
    if(rawMode) {
      if(!gotLine) return;
      if(("===".equals(pLine.trim())) || ("\"===".equals(pLine.trim())))  rawMode = false;
      if(rawMode) {
        if(curTag == null) {
          mainBody.addContent(pLine.trim());
        } else {
          curTag.addContent(pLine.trim());
        }
      }
      return;
    }
    if((gotLine) && (pLine.trim().startsWith("\""))) {
      pLine = (pLine.length() > 1?pLine.substring(1):"");
    } else {
      return;
    }
    if(!validString(pLine)) {
      closeCurTag();
      return;
    }
    pLine = pLine.trim();
    if("===".equals(pLine)) {
      rawMode = true;
      return;
    }
    opcode = pLine.charAt(0);
    opData = (pLine.length() > 1)?pLine.substring(1).trim():null;
    opData = validString(opData)?opData:null;
    if(opData != null) {
      opData = decorate(opData);
      pLine = decorate(pLine);
    }
    HnTag hnTemp = null;
    String[] rowCellz = null;
    String idTemp = null;
    switch(opcode) {
      case '"': break; // Comment!
      case '/':
        // Open a div tag.
        closeAllTags();
        curTag = theGenerator.getDivTag();
        curTag.setID("div"+(divCount++));
        if(opData != null) {
          hnTemp = createHeader(opData);
          if(hnTemp != null) ((DivTag) curTag).add(hnTemp);
        }
      break;
      case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
        // Open ordered list or add list item / dd
        if(curTag == null) {
          curTag = theGenerator.getOLTag();
          curTag.setID("ol"+(olCount++));
        }
        if(curTag instanceof ListTag) {
          ((ListTag) curTag).addItem(opData);
        } else if(curTag instanceof DLTag) {
          ((DLTag) curTag).addDef(opData);
        } else {
          tagStack.push(curTag);
          curTag = theGenerator.getOLTag();
          curTag.setID("ol"+(olCount++));
          ((ListTag) curTag).addItem(opData);
        }
      break;
      case '*': case '.': case '+': case '-':
        // Open unordered list or add list item / dd
        if(curTag == null) {
          curTag = theGenerator.getULTag();
          curTag.setID("ul"+(ulCount++));
        }
        if(curTag instanceof ListTag) {
          ((ListTag) curTag).addItem(opData);
        } else if(curTag instanceof DLTag) {
          ((DLTag) curTag).addDef(opData);
        } else {
          tagStack.push(curTag);
          curTag = theGenerator.getULTag();
          curTag.setID("ul"+(ulCount++));
          ((ListTag) curTag).addItem(opData);
        }
      break;
      case ':':
        // Open def. list and/or add def term.
        if(curTag == null) {
          curTag = theGenerator.getDLTag();
          curTag.setID("dl"+(dlCount++));
        }
        if(curTag instanceof DLTag) {
          ((DLTag) curTag).addTerm(opData);
        } else {
          tagStack.push(curTag);
          curTag = theGenerator.getDLTag();
          curTag.setID("dl"+(dlCount++));
          ((DLTag) curTag).addTerm(opData);
        }
      break;
      case '|':
        // Open table w/ headers or add row.
        rowCellz = (opData == null)?null:opData.split("[|][|]");
        if(curTag == null) {
          curTag = theGenerator.getTableTag();
          curTag.setID("table" + (tableCount++));
          if((rowCellz != null) && (rowCellz.length > 0)) ((TableTag) curTag).addRow(TRTag.createRow((Object[]) rowCellz, null, true));
        } else if(curTag instanceof TableTag) {
          if((rowCellz != null) && (rowCellz.length > 0)) ((TableTag) curTag).addRow(TRTag.createRow((Object[]) rowCellz, null, false));
        } else {
          tagStack.push(curTag);
          curTag = theGenerator.getTableTag();
          curTag.setID("table" + (tableCount++));
          if((rowCellz != null) && (rowCellz.length > 0)) ((TableTag) curTag).addRow(TRTag.createRow((Object[]) rowCellz, null, true));
        }
      break;
      case '#':
        // Add a heading to the current tag.
        if(curTag == null) {
          curTag = createHeader(pLine);
          closeCurTag();
          break;
        }
        hnTemp = createHeader(pLine);
        if((curTag instanceof ContainerTag) && (hnTemp != null)) curTag.addContent(hnTemp.toString());
      break;
      case '!':
        // Force closeAllTags()
        closeAllTags();
      break;
      case '{':
        // Execute a macro command
        executeMacro(opData);
      break;
      case '=':
        // Ignore and print an otherwise-special character.
        pLine = opData;
      default:
        // Start or continue a paragraph.
        while((curTag != null) && ((curTag instanceof ListTag) || (curTag instanceof DLTag) || (curTag instanceof TableTag)) ) closeCurTag();
        if((curTag == null) || (curTag instanceof DivTag)) {
          if(curTag != null) tagStack.push(curTag);
          curTag = theGenerator.getPTag();
          curTag.setID("par"+(pCount++));
        }
        curTag.addContent(pLine + "\n");
    }
  }

  /* Close ALL open tags except the main body. */
  private void closeAllTags() {
    while(!tagStack.empty()) closeCurTag();
    closeCurTag();
  }

  /* Close the current tag and pop the stack into it. */
  private void closeCurTag() {
    String sCurTag = null;
    if(curTag != null) sCurTag = curTag.toString();
    if(tagStack.empty()) {
      if(validString(sCurTag)) mainBody.addContent(sCurTag + "\n");
      curTag = null;
    } else {
      curTag = tagStack.pop();
      if(validString(sCurTag) && (curTag != null)) {
        if(curTag instanceof ListTag) {
          ((ListTag) curTag).addItem(sCurTag);
        } else if(curTag instanceof DLTag) {
          ((DLTag) curTag).addDef(sCurTag);
        } else if(curTag instanceof TableTag) {
          // Process TableTag
        } else {
          curTag.addContent(sCurTag + "\n");
        }
      }
    }
  }

  /* Create a heading tag. */
  private HnTag createHeader(String pHeader) {
    if(!validString(pHeader)) return null;
    HnTag retval = null;
    pHeader = pHeader.trim();
    byte pLevel = 0;
    while(pHeader.startsWith("#")) {
      pHeader = pHeader.substring(1);
      if(pLevel < 7) pLevel++;
    }
    while(pHeader.endsWith("#")) pHeader = pHeader.substring(0, pHeader.length() - 1);
    if(validString(pHeader)) retval = theGenerator.getHeadingTag(pLevel, pHeader);
    return retval;
  }

  /* Decorate a line of text with italics, bold, or underline tags. */
  private String decorate(String pLine) {
    String retval = pLine;
    if(!validString(retval)) return retval;
    
    return retval;
  }

  /* Execute a macro command. */
  private void executeMacro(String pCmd) {
    if(!validString(pCmd)) return;
    pCmd = pCmd.trim();
    int macOpcode = pCmd.charAt(0);
    String macData = (pCmd.length() > 1)?pCmd.substring(1).trim():null;
    switch(macOpcode) {
      case 't':
        // Set document title
        if(validString(macData)) mainBody.setTitle(macData);
      break;
      case 'i':
        // Include external file.
        includeFile(macData, false);
      break;
      case 'g':
        // Include external file.
        includeFile(macData, true);
      break;
      default:
    }
  }

  /* Include an external file. */
  private void includeFile(String pFile, boolean pGRAS) {
    if(!validString(pFile)) return;
    try {
      InputStream isIn = null;
      if(pGRAS) {
        isIn = this.getClass().getResourceAsStream(pFile);
      } else {
        isIn = new FileInputStream(pFile);
      }
      BufferedReader brIn = new BufferedReader(new InputStreamReader(isIn));
      String brLine = null;
      BaseHTag hotTag = (curTag == null)?mainBody:curTag;
      while((brLine = brIn.readLine()) != null) {
        if(brLine.trim().length() > 0) hotTag.addContent(brLine.trim() + "\n");
      }
      brIn.close();
    } catch(Exception ex) { CapCom.error("Error including Noteworthy file", ex); }
  }

  /** Reset the parser to its initial state. All settings return to default values. */
  public void reset() {
    initParse();
    theGenerator = null;
    mainBody = null;
  }

}

