package cox.jmatt.java.MathTools.util;

import cox.jmatt.java.MathTools.*;

/** <p>This class provides a fully-functional <code>MAuthenticator</code> implementation. It can be used as-is for authentication code generation or it can serve 
    as an example of how an implementation should work. To function properly <code>SimpleAuthenticator</code> must be initialized with a prime number. There is a 
    default value but another can be used if desired. The number should be at least three digits but no more than five and it must be prime otherwise duplicate 
    codes may be generated.</p>
    <p>Three levels of codes are generated: Student, Proctor, and Instructor. ADMIN and NONE automatically authenticate to false. The code formats are:
    <ul>
      <li><b>Student:</b> <i>nnn</i> where the <i>n</i> characters represent numbers. The actual length may vary.</li>
      <li><b>Proctor:</b> <i>X</i>-<i>nnn</i> where X is an uppercase letter and n is numbers.</li>
      <li><b>Instructor:</b> <i>nnn</i>-<i>X</i>-<i>mm</i> where the 'n' and 'm' characters represent a number and 'X' is a single letter.</li>
    </ul>
    The same basic mechanism is used to generate all codes but since the formats are different cross-authentication will not work!</p>
    <p>All necessary information is contained within the authentication code itself. Nothing outside this class is required to generate or use the codes. More 
    complex or secure implementations might look up information from a database or a dedicated server; but such is beyond the scope of this class.</p>
    <p><b>IMPORTANT NOTE:</b> Although this implementation <i>is</i> self-contained, codes will not work if the authenticator is initialized with a value different 
    from the one used to generate them!</p>
*/
public class SimpleAuthenticator implements MAuthenticator {

  /** String constant holding the fully-qualified class name. */
  public static final String MYSELF = "cox.jmatt.java.MathTools.util.SimpleAuthenticator";

  private static final String LETTERZ = "ABCDEFGHJKLMNPQRSTUVWXYZ";
  private static final int DEFAULT_MODULUS = 113;

  private long modulus = DEFAULT_MODULUS;
  private int modSep = 1000;

  /** Zero-argument constructor required for <code>CapCom</code> use. */
  public SimpleAuthenticator() { super(); }

  /** All authentication codes operate around a non-negative prime integer. It should be at least 2 digits, more is better. To change the default value pass a 
      String containing the new value. Any errors are reported at Debug level. Passing null resets the default value.
      @param pPrimeInt A non-negative prime integer (int or long), sent in as a String.
  */
  public final void initMAuthenticator(Object pPrimeInt) {
    modulus = DEFAULT_MODULUS;
    if(pPrimeInt != null) {
      try { modulus = Long.parseLong(pPrimeInt.toString()); } catch(Exception ex) { CapCom.debug("Cannot set Authenticator prime", ex); } 
    }
    modSep = (int) Math.pow(10, (""+modulus).length());
  }

  /** Authenticate a code. Each level has a specific format so the pLevel argument is critical! No code will cross-authenticate at a different level, even if it
      is fundamentally identical to another at a different level. (I.e. the base numbers used to generate them were identical.)
  */
  public final boolean mAuthenticate(byte pLevel, String pCode) {
    if(Question.validString(pCode)) {
      String[] partz = pCode.split("-");
      for(int i=0; i < partz.length; i++) { CapCom.debug("["+i+"] = "+partz[i]); }
      switch(pLevel) {
        case AUTH_STUDENT: return authStudent(partz);
        case AUTH_PROCTOR: return authProctor(partz);
        case AUTH_INSTRUCTOR: return authTeacher(partz);
        default:
      }
    }
    return false;
  }

  // Authenticate a student code
  private boolean authStudent(String[] pCode) {
    boolean retval = false;
    if((pCode != null) && (pCode.length == 1)) {
      try {
        long iCode = Long.parseLong(pCode[0]);
        long iMod = iCode % modSep;
        iCode /= modSep;
        retval = (iMod == iCode % modulus);
      } catch(Exception ignored) {}
    }
    return retval;
  }
  // Authenticate a proctor code
  private boolean authProctor(String[] pCode) {
    boolean retval = false;
    if((pCode != null) && (pCode.length == 2)) {
      String [] pTemp = { pCode[1] };
      retval = authStudent(pTemp);
      if(retval) retval = (""+pCode[0].trim()).equalsIgnoreCase(""+LETTERZ.charAt(((int) Long.parseLong(pCode[1])) % LETTERZ.length()));
    }
    return retval;
  }
  // Authenticate an instructor code
  private boolean authTeacher(String[] pCode) {
    boolean retval = false;
    if((pCode != null) && (pCode.length == 3)) {
      String [] pTemp = { pCode[1], pCode[0] };
      retval = authProctor(pTemp);
    }
    if(retval) {
      long lTemp = modulus - (Long.parseLong(pCode[0]) % modulus);
      retval = (""+lTemp).equals(pCode[2].trim());
    }
    return retval;
  }

  /** Generate codes. The only parameters used are pLevel and pNumberOf. This class neither requires nor uses the Object params.
      @param pLevel The level of authentication codes to generate.
      @param pNumberOf The number of codes to generate. If less than one, one is generated.
      @param pIgnored Not used.
  */
  public final String[] generateCodes(byte pLevel, int pNumberOf, Object... pIgnored) {
    if(pNumberOf < 1) pNumberOf = 1;
    String[] retval = new String[pNumberOf];
    long tBase = MathEngine.nzRandom(10000, 50000), tMod = 0, tComp = 0, tCode = 0;
    String tLett = "";
    for(int i=0; i < pNumberOf; i++) {
      tMod = tBase % modulus;
      tCode = tBase * modSep + tMod;
      tLett = ""+LETTERZ.charAt((int) tCode % LETTERZ.length());
      tComp = modulus - (tCode % modulus);
      switch(pLevel) {
      case AUTH_INSTRUCTOR:
        retval[i] = ""+tCode+"-"+tLett+"-"+tComp;
        break;
      case AUTH_PROCTOR:
        retval[i] = tLett+"-"+tCode;
        break;
      default:
        retval[i] = ""+tCode;
      }
      tBase = tBase + 53 * (i+1);
    }
    return retval;
  }

}

